/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CApplication.cpp

				CApplication Class

 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <iostream>
#include <fstream>

#include "CExceptions.h"
#include "CStringReader.h"
#include "CCommandRecord.h"
#include "CReportRecord.h"
#include "CEngineCaller.h"
#include "CTextWriter.h"
#include "CTextReader.h"
#include "CProgress.h"
#include "CReporter.h"
#include "CTextReporter.h"
#include "CHTMLReporter.h"
#include "CApplication.h"

// ---------------------------------------------------------------------------------
//		 CApplication
// ---------------------------------------------------------------------------------

CApplication::CApplication()
{
	mEngineCaller = NULL;
	
	mResultsReporter = NULL;
	mLogReporter = NULL;
	mErrorReporter = NULL;
	mProgress = NULL;
	
	strcpy(mPrompt, "App> ");

	mReader = new CStringReader();
	mHelper = NULL;
}

CApplication::~CApplication()
{
	if (mEngineCaller != NULL)
		mEngineCaller->release();

	if (mResultsReporter != NULL)
		mResultsReporter->release();

	if (mLogReporter != NULL)
		mLogReporter->release();

	if (mErrorReporter != NULL)
		mErrorReporter->release();

	if (mProgress != NULL)
		mProgress->release();
		
	mReader->release();

	if (mHelper != NULL)
		mHelper->release();
}

// ---------------------------------------------------------------------------------
//		 CApplication::writeTitle
// ---------------------------------------------------------------------------------

void CApplication::writeTitle() 
{
}

// ---------------------------------------------------------------------------------
//		 CApplication::setApplicationHelper
// ---------------------------------------------------------------------------------

void CApplication::setApplicationHelper(CApplicationHelper *inHelper) 
{
	if (mHelper != NULL)
		mHelper->release();
		
	mHelper = inHelper; 
	if (mHelper != NULL)
		mHelper->retain(); 
}

// ---------------------------------------------------------------------------------
//		 CApplication::setResultsReporter
// ---------------------------------------------------------------------------------

void CApplication::setResultsReporter(CReporter* inReporter) 
{
	if (mResultsReporter != NULL)
		mResultsReporter->release();
		
	mResultsReporter = inReporter; 
	if (mResultsReporter != NULL)
		mResultsReporter->retain(); 
}

// ---------------------------------------------------------------------------------
//		 CApplication::setLogReporter
// ---------------------------------------------------------------------------------

void CApplication::setLogReporter(CReporter* inReporter) 
{
	if (mLogReporter != NULL)
		mLogReporter->release();
		
	mLogReporter = inReporter; 
	if (mLogReporter != NULL)
		mLogReporter->retain(); 
}

// ---------------------------------------------------------------------------------
//		 CApplication::setErrorReporter
// ---------------------------------------------------------------------------------

void CApplication::setErrorReporter(CReporter* inReporter) 
{
	if (mErrorReporter != NULL)
		mErrorReporter->release();
		
	mErrorReporter = inReporter; 
	if (mErrorReporter != NULL)
		mErrorReporter->retain(); 
}

// ---------------------------------------------------------------------------------
//		 CApplication::setProgress
// ---------------------------------------------------------------------------------

void CApplication::setProgress(CProgress *inProgress) 
{
	if (mProgress != NULL)
		mProgress->release();
		
	mProgress = inProgress; 
	if (mProgress != NULL)
		mProgress->retain(); 
}

// ---------------------------------------------------------------------------------
//		 CApplication::run
// ---------------------------------------------------------------------------------

void CApplication::run() 
{
	ThrowIfNull(mEngineCaller);
	
	char commandString[1024];
	int len;
	char *P;
	bool done=false;
	
	while (!done) {
		try {
			std::cout << mPrompt;
			std::cin.getline(commandString, 1022);
			std::cout << std::endl;
			len = strlen(commandString);
			
			P = &commandString[len-1];
			while (P > commandString && isspace(*P))
				P--;
			
			if (P >= commandString) {
				if (*P != ';') {
					P++; 
					*P = ';';
					P++;
					*P = '\0';
				}
							
				done = processCommand(commandString, true);
			}
		} catch (CEndOfFile) {
			done=true;
		}
	}
}
		
// ---------------------------------------------------------------------------------
//		 CApplication::parseCommand
// ---------------------------------------------------------------------------------

bool CApplication::processCommand(char *inCommandString, bool inHandleErrors)
{
	CCommandRecord *command = NULL;
	CReportRecord *report = NULL;
	bool shouldQuit = false;
	
	ThrowIfNull(mEngineCaller);
	
	try {
		command = parseCommand(inCommandString);
		
		report = mEngineCaller->execute(command, shouldQuit);
		
		command->release();
	} catch (CErrorInFile e) {
		if (inHandleErrors && mErrorReporter != NULL) {
			CReportRecord *errorReport = new CReportRecord();
			
			CListElement *list = new CListElement(kDefinitionList);
			CParagraphElement *para = new CParagraphElement(kHeadingLevel2);
			para->add(new CTextElement("Error parsing command in file (line "));
			para->add(new CIntegerElement(e.getLineNumber()));
			para->add(new CTextElement("):"));
			list->add(para);
			para = new CParagraphElement(kHeadingLevel2);
			CException e1 = e.getException();
			if (e1.getReportElement1() != NULL)
				para->add(e1.getReportElement1());
			if (e1.getReportElement2() != NULL) {
				para->addLineBreak();
				para->add(e1.getReportElement2());
			}
			list->add(para);
			errorReport->add(list);
			
			mErrorReporter->writeReport(errorReport);
			
			errorReport->release();
		} else
			throw e;
	} catch (CException e) {
		if (inHandleErrors && mErrorReporter != NULL) {
			CReportRecord *errorReport = new CReportRecord();
			
			//CParagraphElement *para = new CParagraphElement(kHeadingLevel2);
			//para->add(new CTextElement("Error parsing command:"));
			//errorReport->add(para);
			CListElement *list = new CListElement(kDefinitionList);
			list->add(new CTextElement("Error parsing command:"));
			CParagraphElement *para = new CParagraphElement(kHeadingLevel2);
			if (e.getReportElement1() != NULL)
				para->add(e.getReportElement1());
			if (e.getReportElement2() != NULL)
				para->add(e.getReportElement2());
	
			list->add(para);
			errorReport->add(list);
			
			mErrorReporter->writeReport(errorReport);
			
			errorReport->release();
		} else
			throw e;
	}
	
	if (report != NULL) {
		if (mResultsReporter != NULL) {
			mResultsReporter->writeReport(report);	
		}
		if (mLogReporter != NULL) {
			mLogReporter->writeReport(report);	
		}
		report->release();
	}
	
	return shouldQuit;
}

// ---------------------------------------------------------------------------------
//		 CApplication::parseCommand
// ---------------------------------------------------------------------------------

CCommandRecord* CApplication::parseCommand(char *inCommandString)
{
	char token[256], token2[1024], ch;
	
	mReader->setString(inCommandString);
	
	ch = mReader->readToken(token, 255, "?;", true, false);
	
	CCommandRecord *command;
	
	if (isspace(ch)) {
		mReader->skipSpace();
		ch = mReader->readChar();
		if (ch != '?' && ch != ';') 
			mReader->unreadChar(ch);
	}
	
	if (ch == '?') { // Create a help command
		if (strlen(token) > 0)
			NewClass(command, CCommandRecord("HELP", token));
		else
			NewClass(command, CCommandRecord("HELP", NULL));
		return command;
	}
	
	if (ch == ';') { // Create a empty command
		NewClass(command, CCommandRecord(token, NULL));
		return command;
	}
	
	bool hasCommandValue = true;
	
	mReader->markPosition();
	
	mReader->skipSpace();
	ch = mReader->readChar();
	if (ch == '(') { // is a list
		ch = mReader->readString(token2, 1023, ")=;", false, true);
		if (ch != ')')
			throw CParserError("Expected character ')' missing from file");
			
		mReader->skipSpace();
		ch = mReader->readChar();
		if (ch != ';')
			mReader->unreadChar(ch);
	} else {
		mReader->unreadChar(ch);
		ch = mReader->readToken(token2, 255, "=;", false);
		if (ch == '=')
			hasCommandValue = false;
	}
	
	if (hasCommandValue) {
		NewClass(command, CCommandRecord(token, token2));
	} else {
		NewClass(command, CCommandRecord(token, NULL));
		ch = '\0';
		mReader->restorePosition();
	}
	
	while (ch != ';') {
		ch = mReader->readToken(token, 255, "=;", true);
		
		if (ch == '=') {
			mReader->skipSpace();
			ch = mReader->readChar();
			
			if (ch == '(') { // is a list
				ch = mReader->readString(token2, 1023, ");", false);
				if (ch != ')')
					throw CParserError("Expected character ')' missing from file");
				mReader->skipSpace();
				ch = mReader->nextChar();
			} else {
				mReader->unreadChar(ch);
				ch = mReader->readToken(token2, 1023, "=;", false);

				if (ch == '=')
					throw CParserError("Unexpected character '=' in file");
			}
				
			command->addSubCommand(token, token2);
		} else
			throw CParserError("Expected character '=' missing from file");
	}
	
	return command;	
}

// ---------------------------------------------------------------------------------
//		 CApplication::createLogReporter
// ---------------------------------------------------------------------------------

CReporter* CApplication::createLogReporter(CWriter* inWriter, bool inUseHTML)
{
 	if (inUseHTML)
   		return new CHTMLReporter(inWriter, true);
   	else
   		return new CTextReporter(inWriter, true);
}

// ---------------------------------------------------------------------------------
//		 CApplication::closeLogReporter
// ---------------------------------------------------------------------------------

void CApplication::closeLogReporter()
{
	if (mLogReporter != NULL)
		delete mLogReporter;
	mLogReporter = NULL;
}

// ---------------------------------------------------------------------------------
//		 CApplication::createFileWriter
// ---------------------------------------------------------------------------------

CWriter *CApplication::createFileWriter(CFileRequest *ioFileRequest) 
{
	if (mHelper == NULL)
		mHelper = new CApplicationHelper();
	
	return mHelper->createFileWriter(ioFileRequest);
}

// ---------------------------------------------------------------------------------
//		 CApplication::createFileReader
// ---------------------------------------------------------------------------------

CReader *CApplication::createFileReader(CFileRequest *ioFileRequest) 
{
	if (mHelper == NULL)
		mHelper = new CApplicationHelper();
	
	return mHelper->createFileReader(ioFileRequest);
}

