/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CAxis.cpp
 
		CAxis Class
	 
 ******************************************************************************/

#include <math.h>

#include "CDoubleArray.h"
#include "CAxis.h"

#pragma mark === Construction & Destruction ===

// ---------------------------------------------------------------------------------
//		 CAxis()
// ---------------------------------------------------------------------------------

CAxis::CAxis( )
{
	InitAxis(kAtMajorTick, kAtMajorTick, false);
}


CAxis::CAxis(AxisFlag inMinAxisFlag, AxisFlag inMaxAxisFlag)
{
	InitAxis(inMinAxisFlag, inMaxAxisFlag, false);
}


CAxis::CAxis(AxisFlag inMinAxisFlag, AxisFlag inMaxAxisFlag, bool inIsDiscrete)
{
	InitAxis(inMinAxisFlag, inMaxAxisFlag, inIsDiscrete);
}


// ---------------------------------------------------------------------------------
//		CAxis::InitAutomatic()
// ---------------------------------------------------------------------------------

void CAxis::InitAxis(AxisFlag inMinAxisFlag, AxisFlag inMaxAxisFlag, bool inIsDiscrete)
{
	minData=0.0;
	maxData=0.0;

	minAxisFlag=inMinAxisFlag;
	maxAxisFlag=inMaxAxisFlag;
	
	prefNumMajorTicks=5;
	prefNumMinorTicks=2;

	isDiscrete=inIsDiscrete;
	labelFirst=false;
	labelLast=false;

	isCalibrated=false;
}


#pragma mark === Accessing ===

// ---------------------------------------------------------------------------------
//		CAxis::SetRange()
// ---------------------------------------------------------------------------------

void CAxis::SetRange(CDoubleArray *inData)
{
	maxData=inData->getMaxValue();
	minData=inData->getMinValue();

	isCalibrated=false;
}

// ---------------------------------------------------------------------------------
//		CAxis::AddRange()
// ---------------------------------------------------------------------------------

void CAxis::AddRange(CDoubleArray *inData)
{
	if (inData->getMaxValue()>maxData)
		maxData=inData->getMaxValue();
	if (inData->getMinValue()<minData)
		minData=inData->getMinValue();
	
	isCalibrated=false;
}

// ---------------------------------------------------------------------------------
//		CAxis::SetRange()
// ---------------------------------------------------------------------------------

void CAxis::SetRange(double inMinValue, double inMaxValue)
{
	maxData=inMaxValue;
	minData=inMinValue;

	isCalibrated=false;
}

// ---------------------------------------------------------------------------------
//		CAxis::AddRange()
// ---------------------------------------------------------------------------------

void CAxis::AddRange(double inMinValue, double inMaxValue)
{
	if (inMaxValue>maxData)
		maxData=inMaxValue;
	if (inMinValue<minData)
		minData=inMinValue;

	isCalibrated=false;
}

// ---------------------------------------------------------------------------------
//		CAxis::GetLabelFirst()
// ---------------------------------------------------------------------------------

bool CAxis::GetLabelFirst()
{
	if (GetNumMinorTicks(-1)==0) // The first tick is a label anyway
		return false;
	else
		return labelFirst;
}

// ---------------------------------------------------------------------------------
//		CAxis::GetLabelLast()
// ---------------------------------------------------------------------------------

bool CAxis::GetLabelLast()
{
	if (GetNumMinorTicks(numMajorTicks-1)==0) // The last tick is a label anyway
		return false;
	else
		return labelLast;
}

//-------------------------------------------------------------------------------------
// 		CAxis::Transform
//-------------------------------------------------------------------------------------

double CAxis::Transform(double inValue)
{
	return inValue;	// a linear transform !
}
	
//-------------------------------------------------------------------------------------
// 		CAxis::ScaleValue
//-------------------------------------------------------------------------------------

double CAxis::ScaleValue(double inValue)
{
	if (!isCalibrated)
		Calibrate();
		
	double f=(Transform(inValue)-Transform(minAxis))/(Transform(maxAxis)-Transform(minAxis));
	return f;
}

//-------------------------------------------------------------------------------------
// 		CAxis::GetNumMajorTicks
//-------------------------------------------------------------------------------------

int CAxis::GetNumMajorTicks()
{
	if (!isCalibrated)
		Calibrate();
		
	return numMajorTicks;
}
	
//-------------------------------------------------------------------------------------
// 		CAxis::GetNumMinorTicks
// 		
// 		Returns the number of minor tick marks within each major one
// 		By default all major ticks have the same number of minor ticks
// 		except the last which has none.
//-------------------------------------------------------------------------------------

int CAxis::GetNumMinorTicks(int majorTickNo)
{
	if (!isCalibrated)
		Calibrate();
		
	if (majorTickNo==numMajorTicks-1)
		return (int)((maxAxis-maxTick)/minorTick);
	else if (majorTickNo==-1)
		return (int)((minTick-minAxis)/minorTick);
	else
		return numMinorTicks; 
}
	
//-------------------------------------------------------------------------------------
// 		CAxis::GetMajorTick
// 		
// 		Returns the value of the majorTickNo'th major tick.
//-------------------------------------------------------------------------------------

double CAxis::GetMajorTick(int majorTickNo)
{
	if (!isCalibrated)
		Calibrate();
		
	return (majorTickNo*majorTick)+minTick;
}
	
//-------------------------------------------------------------------------------------
// 		CAxis::GetMinorTick
// 		
// 		Returns the value of the minorTickNo'th minor tick.
//-------------------------------------------------------------------------------------

double CAxis::GetMinorTick(int minorTickNo, int majorTickNo)
{
	if (!isCalibrated)
		Calibrate();
		
	// get minorTickNo+1 to skip the major tick
	if (majorTickNo==-1)
		return minTick-((minorTickNo+1)*minorTick);
	else
		return ((minorTickNo+1)*minorTick)+GetMajorTick(majorTickNo);
}

//-------------------------------------------------------------------------------------
// 		CAxis::GetMinAxis
//-------------------------------------------------------------------------------------

double CAxis::GetMinAxis()
{
	if (!isCalibrated)
		Calibrate();
		
	return minAxis;
}

//-------------------------------------------------------------------------------------
// 		CAxis::GetMaxAxis
//-------------------------------------------------------------------------------------

double CAxis::GetMaxAxis()
{
	if (!isCalibrated)
		Calibrate();
		
	return maxAxis;
}

#pragma mark === Calibrating ===

//-------------------------------------------------------------------------------------
// 		CAxis::Calibrate
//-------------------------------------------------------------------------------------

void CAxis::Calibrate()
{
	double range=maxData-minData;
	epsilon=range * 1.0E-10;

	// We must find the optimum minMajorTick and maxMajorTick so 
	// that they contain the data range (minData to maxData) and
	// are in the right order of magnitude
	
	if (range<1.0E-30) {
		if (minData<0.0) {
			majorTick=pow(10.0, floor(log10(fabs(minData))));
			minTick=floor(minData/majorTick)*majorTick;
			maxTick=0.0;
		} else if (minData>0.0) {
			majorTick=pow(10.0, floor(log10(fabs(minData))));
			minTick=0.0;
			maxTick=ceil(maxData/majorTick)*majorTick;
		} else {
			minTick=-1.0;
			maxTick=1.0;
			majorTick=1.0;
		}
	} else {
		// First find order of magnitude below the data range...
		majorTick=pow(10.0, floor(log10(range)));

		SetMinTick();
		SetMaxTick();
	}
		
	SetMajorTick();
	SetMinorTick();
			
	minAxis=minTick;
	maxAxis=maxTick;
	
	HandleAxisFlags();

	isCalibrated=true;
}

//-------------------------------------------------------------------------------------
// 		CAxis::SetMinTick
//-------------------------------------------------------------------------------------

void CAxis::SetMinTick()
{
	// Find the nearest multiple of majorTick below minData/minAxis
	if (minData==0.0)
		minTick=0;
	else
		minTick=floor(minData/majorTick)*majorTick;
}
	
//-------------------------------------------------------------------------------------
// 		CAxis::SetMaxTick
//-------------------------------------------------------------------------------------

void CAxis::SetMaxTick()
{
	// Find the nearest multiple of majorTick above maxData
	if (maxData==0)
		maxTick=0;
	else
		maxTick=ceil(maxData/majorTick)*majorTick;
}
	
//-------------------------------------------------------------------------------------
// 		CAxis::SetMajorTick
//-------------------------------------------------------------------------------------

void CAxis::SetMajorTick()
{
	fraction=kUnit;
	
	// make sure that there are at least prefNumMajorTicks major ticks
	// by dividing up into halves, quarters, fifths or tenths 
	double u=majorTick;
	double r=maxTick-minTick;
	numMajorTicks=(int)(r/u);
	
	while (numMajorTicks<prefNumMajorTicks) {
		u=majorTick/2;	// Try using halves
		if (!isDiscrete || u==floor(u)) { // u is an integer
			numMajorTicks=(int)(r/u);
			fraction=kHalfs;
			if (numMajorTicks>=prefNumMajorTicks)
				break;
		}
						
		u=majorTick/4;	// Try using quarters
		if (!isDiscrete || u==floor(u)) { // u is an integer
			numMajorTicks=(int)(r/u);
			fraction=kQuarters;
			if (numMajorTicks>=prefNumMajorTicks)
				break;
		}
						
		u=majorTick/5;	// Try using fifths
		if (!isDiscrete || u==floor(u)) { // u is an integer
			numMajorTicks=(int)(r/u);
			fraction=kFifths;
			if (numMajorTicks>=prefNumMajorTicks)
				break;
		}
									
		if (isDiscrete && (majorTick/10)!=floor(majorTick/10)) {
			// majorTick/10 is not an integer so no point in further subdivision
			u=majorTick;
			numMajorTicks=(int)(r/u);
			break;
		}
		
		majorTick/=10;	// finally just divide by ten
		u=majorTick;	// and go back to whole units
		numMajorTicks=(int)(r/u);
		fraction=kUnit;
		
	}
	majorTick=u;
	
	if (isDiscrete && majorTick<1.0) {
		majorTick=1.0;
		numMajorTicks=(int)(r/majorTick);
		fraction=kUnit;
	}
	
	numMajorTicks++;	// Add 1 to give the final tick
	
	// Trim down any excess major ticks either side of the data range
	// Epsilon allows for any inprecision in the calculation
	while (minTick+majorTick-epsilon<=minData) {
		minTick+=majorTick;
		numMajorTicks--;
	}
	while (maxTick-majorTick+epsilon>=maxData) {
		maxTick-=majorTick;
		numMajorTicks--;
	}
}

//-------------------------------------------------------------------------------------
// 		CAxis::SetMinorTick
//-------------------------------------------------------------------------------------

void CAxis::SetMinorTick()
{
	minorTick=majorTick; // start with minorTick the same as majorTick
	double u=minorTick;
	double r=majorTick;
	numMinorTicks=(int)(r/u);
	
	while (numMinorTicks<prefNumMinorTicks) {
		// if the majorTick was divided as quarters, then we can't
		// divide the minor ticks into halves or quarters.
		if (fraction!=kQuarters) { 
			u=minorTick/2;	// Try using halves
			if (!isDiscrete || u==floor(u)) { // u is an integer
				numMinorTicks=(int)(r/u);
				if (numMinorTicks>=prefNumMinorTicks)
					break;
			}
								
			u=minorTick/4;	// Try using quarters
			if (!isDiscrete || u==floor(u)) { // u is an integer
				numMinorTicks=(int)(r/u);
				if (numMinorTicks>=prefNumMinorTicks)
					break;
			}
		}
				
		u=minorTick/5;	// Try using fifths
		if (!isDiscrete || u==floor(u)) { // u is an integer
			numMinorTicks=(int)(r/u);
			if (numMinorTicks>=prefNumMinorTicks)
				break;
		}
						
		if (isDiscrete && (minorTick/10)!=floor(minorTick/10)) {
			// minorTick/10 is not an integer so no point in further subdivision
			u=minorTick;
			numMinorTicks=(int)(r/u);
			break;
		}
		
		minorTick/=10;	// finally just divide by ten
		u=minorTick;	// and go back to whole units
		numMinorTicks=(int)(r/u);
	}
	minorTick=u;
			
	numMinorTicks--;
}

//-------------------------------------------------------------------------------------
// 		CAxis::HandleAxisFlags
//-------------------------------------------------------------------------------------

void CAxis::HandleAxisFlags()
{
	// Now we must honor the min/maxAxisFlag settings
	if (minAxisFlag==kAtMajorTickPlus || minAxisFlag==kAtMinorTickPlus) {
		if (minAxis+epsilon>=minData) {
			numMajorTicks++;
			minTick-=majorTick;
			minAxis=minTick;
		}
	}
	
	if (minAxisFlag==kAtMinorTickPlus) {
		if ((minAxis+minorTick)<minData) {
			numMajorTicks--;
			minTick+=majorTick;
			while ((minAxis+minorTick)<minData) {
				minAxis+=minorTick;
			}
		}
	} else if (minAxisFlag==kAtMinorTick) {
		if ((minAxis+minorTick-epsilon)<=minData) {
			numMajorTicks--;
			minTick+=majorTick;
			while ((minAxis+minorTick-epsilon)<=minData) {
				minAxis+=minorTick;
			}
		}
	} else if (minAxisFlag==kAtValue) {
		if (minTick<minData) { // in case minTick==minData
			numMajorTicks--;
			minTick+=majorTick;
			minAxis=minData;
		}
	} else if (minAxisFlag==kAtZero) {
		numMajorTicks+=(int)(minTick/majorTick);
		minTick=0;
		minAxis=0;
	}

	if (maxAxisFlag==kAtMajorTickPlus || maxAxisFlag==kAtMinorTickPlus) {
		if (maxAxis-epsilon<=maxData) {
			numMajorTicks++;
			maxTick+=majorTick;
			maxAxis=maxTick;
		}
	}
	
	if (maxAxisFlag==kAtMinorTickPlus) {
		if ((maxAxis-minorTick)>maxData) {
			numMajorTicks--;
			maxTick-=majorTick;
			while ((maxAxis-minorTick)>maxData) {
				maxAxis-=minorTick;
			}
		}
	} else if (maxAxisFlag==kAtMinorTick) {
		if ((maxAxis-minorTick+epsilon)>=maxData) {
			numMajorTicks--;
			maxTick-=majorTick;
			while ((maxAxis-minorTick+epsilon)>=maxData) {
				maxAxis-=minorTick;
			}
		}
	} else if (maxAxisFlag==kAtValue) {
		if (maxTick>maxData) { // in case maxTick==maxData
			numMajorTicks--;
			maxTick-=majorTick;
			maxAxis=maxData;
		}
	} else if (maxAxisFlag==kAtZero) {
		if (maxTick > 0.0)
			throw /* some exception or another */;
			
		numMajorTicks+=(int)(-maxTick/majorTick);
		maxTick=0;
		maxAxis=0;
	}
}

