/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CCommandRecord.cpp

				CCommandRecord Class

 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"
#include "CStringReader.h"

#include "CCommandRecord.h"

// ---------------------------------------------------------------------------------
//		 SubCommandRecord::SubCommandRecord
// ---------------------------------------------------------------------------------

SubCommandRecord::SubCommandRecord(char *inLabel, char *inValue) 
{
	if (inLabel != NULL) {
		NewArray(mLabel, char, strlen(inLabel) + 1);
		strcpy(mLabel, inLabel);
	} else
		mLabel = NULL;
		
	if (inValue != NULL) {
		NewArray(mValue , char, strlen(inValue) + 1);
		strcpy(mValue, inValue);
	} else
		mValue = NULL;
}

// ---------------------------------------------------------------------------------
//		 SubCommandRecord::~SubCommandRecord
// ---------------------------------------------------------------------------------

SubCommandRecord::~SubCommandRecord() {
	if (mLabel != NULL)
		delete [] mLabel;
		
	if (mValue != NULL)
		delete [] mValue;
}

// ---------------------------------------------------------------------------------
//		 SubCommandRecord::compareLabel
// ---------------------------------------------------------------------------------

bool SubCommandRecord::compareLabel(char *inString, int inMinLen, bool inCaseSensitive)
{
	return compareTokens(mLabel, inString, inMinLen, inCaseSensitive);
}

// ---------------------------------------------------------------------------------
//		 SubCommandRecord::compareValue
// ---------------------------------------------------------------------------------

bool SubCommandRecord::compareValue(char *inString, int inMinLen, bool inCaseSensitive)
{
	return compareTokens(mValue, inString, inMinLen, inCaseSensitive);
}

// ---------------------------------------------------------------------------------
//		 SubCommandRecord::compareTokens
// ---------------------------------------------------------------------------------

// This function compares a token with a full string. The token must match the
// string for all its characters but can be shorter (minimum inMinLen).

bool SubCommandRecord::compareTokens(char *inToken, char *inString, int inMinLen, bool inCaseSensitive)
{
	char *P = inToken;
	char *Q = inString;
	int i = 0;
	
	if (inToken == NULL)
		return false;
	
	if (inString == NULL)
		return false;
		
	if (inCaseSensitive) {
		while (*P != '\0' && *Q != '\0' && *P == *Q) {
			i++;
			P++;
			Q++;
		}

		if (*P == '\0' && *Q != '\0') { // token shorter than string
			if (i < inMinLen) // Too short? 
				throw CAmbiguousField(inToken);
				
			return true;
		}
			
		if (*P != '\0' && *Q == '\0') // token longer than string
			return false;
			
		if (*P != *Q) // mismatch
			return false;
	} else {
		while (*P != '\0' && *Q != '\0' && toupper(*P) == toupper(*Q)) {
			i++;
			P++;
			Q++;
		}
		
		if (*P == '\0' && *Q != '\0') { // token shorter than string
			if (i < inMinLen) // Too short? 
				throw CAmbiguousField(inToken);
				
			return true;
		}
			
		if (*P != '\0' && *Q == '\0') // token longer than string
			return false;
			
		if (toupper(*P) != toupper(*Q)) // mismatch
			return false;
	}
		
	return true; // match
}


// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsDouble
// ---------------------------------------------------------------------------------
// Converts the subcommand value to an double, throws a CFailedConversion if it fails.

double SubCommandRecord::getValueAsDouble() 
{
	double value;
	
	if (sscanf(mValue, "%lf", &value) != 1)
		throw CFailedConversion(mLabel, mValue);
		
	return value;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsInteger
// ---------------------------------------------------------------------------------
// Converts the subcommand value to an integer, throws a CFailedConversion if it fails.

int SubCommandRecord::getValueAsInteger() 
{
	int value;
	
	if (sscanf(mValue, "%d", &value) != 1)
		throw CFailedConversion(mLabel, mValue);
		
	return value;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsBoolean
// ---------------------------------------------------------------------------------
// Converts the subcommand value to an double, throws a CFailedConversion if it fails.
// Recognizes 'true', 'yes', 'y', '1', 'false', 'no', 'n', '0'

bool SubCommandRecord::getValueAsBoolean() 
{
	bool value;
	
	if (compareValue("TRUE", 1, false) || 
				compareValue("YES", 1, false) || 
				compareValue("1", 1, false)) {
		value = true;
	} else if (compareValue("FALSE", 1, false) || 
				compareValue("NO", 1, false) || 
				compareValue("0", 1, false)) {
		value = false;
	} else
		throw CFailedConversion(mLabel, mValue);
	
	return value;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsOption
// ---------------------------------------------------------------------------------
// Looks for the subcommand value in inList and returns an index of it if it
// is there. Throws a CFailedConversion if it fails to find it.

int SubCommandRecord::getValueAsOption(char **inList, int *inMinLength, int inNumItems, bool inCaseSensitive) 
{
	int value;
	bool found = false;
	int i = 0; 
	while (!found && i < inNumItems) {
		found = compareValue(inList[i], inMinLength[i], inCaseSensitive);
		if (!found)
			i++;
	}
	
	if (!found)
		throw CFailedConversion(mLabel, mValue);
	
	return i;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getNumValuesInList
// ---------------------------------------------------------------------------------

int SubCommandRecord::getNumValuesInList() 
{
	CStringReader reader;
	reader.setString(mValue);
	
	int i = 0;
	try {
		char token[256];
		
		while (1) {
			reader.readToken(token, 255, ",");
			i++;
		}
	} catch (CEndOfFile) { }	
	
	return i;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsDoubleList
// ---------------------------------------------------------------------------------
// Converts the subcommand value into a list of doubles. It only stores a maximum
// of inMaxItems but returns the number that were actually there.

int SubCommandRecord::getValueAsDoubleList(double *outList, int inMaxItems) 
{
	CStringReader reader;
	reader.setString(mValue);
	
	int i = 0;
	try {
		double value;
		
		while (i < inMaxItems) {
			reader.readDouble(&value, ",");
			if (i < inMaxItems)
				outList[i] = value;
			i++;
		}
	} catch (CEndOfFile) { }	
	
	return i;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsIntegerList
// ---------------------------------------------------------------------------------
// Converts the subcommand value into a list of integers. It only stores a maximum
// of inMaxItems but returns the number that were actually there.

int SubCommandRecord::getValueAsIntegerList(int *outList, int inMaxItems) 
{
	CStringReader reader;
	reader.setString(mValue);
	
	int i = 0;
	try {
		int value;
		
		while (i < inMaxItems) {
			reader.readInteger(&value, ",");
			if (i < inMaxItems)
				outList[i] = value;
			i++;
		}
	} catch (CEndOfFile) { }	
	
	return i;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsStringList
// ---------------------------------------------------------------------------------
// Converts the subcommand value into a list of strings. It only stores a maximum
// of inMaxItems but returns the number that were actually there.

int SubCommandRecord::getValueAsStringList(char **outList, int inMaxItems) 
{
	CStringReader reader;
	reader.setString(mValue);
	
	int i = 0;
	try {
		char token[256];
		
		while (i < inMaxItems) {
			reader.readToken(token, 255, ",");
			if (i < inMaxItems)
				strcpy(outList[i], token);
			i++;
		}
	} catch (CEndOfFile) { }	
	
	return i;
}
		
// ---------------------------------------------------------------------------------
//		 SubCommandRecord::getValueAsRange
// ---------------------------------------------------------------------------------
// Converts the subcommand value into a range. This is in the form of "1-5" or
// just a single digit "4". Throws a CFailedConversion if it fails.

void SubCommandRecord::getValueAsRange(int &outFirst, int &outLast) 
{
	CStringReader reader;
	reader.setString(mValue);
	
	int value;
	char ch = reader.readInteger(&value, "-");
	
	outFirst = value;
	
	if (ch == '-')
		reader.readInteger(&value, "-");
	
	outLast = value;
}
		
// ---------------------------------------------------------------------------------
//		 CCommandRecord::CCommandRecord
// ---------------------------------------------------------------------------------

CCommandRecord::CCommandRecord(char *inLabel, char *inValue) :
			SubCommandRecord(inLabel, inValue)
{
	numSubCommands = 0;
}

// ---------------------------------------------------------------------------------
//		 CCommandRecord::~CCommandRecord
// ---------------------------------------------------------------------------------

CCommandRecord::~CCommandRecord() 
{
	for (int i = 0; i < numSubCommands; i++)
		delete subCommands[i];
}

// ---------------------------------------------------------------------------------
//		 CCommandRecord::addSubCommand
// ---------------------------------------------------------------------------------

void CCommandRecord::addSubCommand(char *inLabel, char *inValue) 
{
	if (numSubCommands >= kMaxSubCommands)
		throw CIndexOutOfBounds();
		
	subCommands[numSubCommands] = new SubCommandRecord(inLabel, inValue);
	numSubCommands++;
}
		
// ---------------------------------------------------------------------------------
//		 CCommandRecord::getSubCommand
// ---------------------------------------------------------------------------------

SubCommandRecord *CCommandRecord::getSubCommand(int index) 
{
	if (index < 0 || index >= numSubCommands)
		throw CIndexOutOfBounds();
		
	return subCommands[index];
}

// ---------------------------------------------------------------------------------
//		 CCommandRecord::getSubCommandLabel
// ---------------------------------------------------------------------------------

char *CCommandRecord::getSubCommandLabel(int index) 
{ 
	if (index < 0 || index >= numSubCommands)
		throw CIndexOutOfBounds();
		
	return subCommands[index]->getLabel(); 
}

// ---------------------------------------------------------------------------------
//		 CCommandRecord::getSubCommandValue
// ---------------------------------------------------------------------------------

char *CCommandRecord::getSubCommandValue(int index) 
{
	if (index < 0 || index >= numSubCommands)
		throw CIndexOutOfBounds();
		
	return subCommands[index]->getValue();
}

