/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CDoubleArray.cpp
 
		CDoubleArray class 

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of doubles

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"

#include "CDoubleArray.h"

#pragma mark === CDoubleArray ===

// ---------------------------------------------------------------------------------
//		CDoubleArray::CDoubleArray
// ---------------------------------------------------------------------------------

//: Constructs an empty CDoubleArray
//  The array will only be allocated when a double is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
CDoubleArray::CDoubleArray(int inInitial, int inIncrement)
{
	capacity=0;
	numValues=0;
	values=NULL;
	iterator=0;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::~CDoubleArray
// ---------------------------------------------------------------------------------

//: The destructor
CDoubleArray::~CDoubleArray()
{
	if (values!=NULL)
		delete[] values;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void CDoubleArray::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void CDoubleArray::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void CDoubleArray::setIncrement(int inIncrement)
{
	increment=inIncrement;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::getValue
// ---------------------------------------------------------------------------------

//: Returns the double that is at an index.
double CDoubleArray::getValue(int index)
{
	TestBounds(index, numValues);
	
	return values[index];
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::getSum
// ---------------------------------------------------------------------------------

//: Returns the sum of all the doubles
double CDoubleArray::getSum()
{
	int i;
	double sum = 0.0;
	for (i = 0; i < numValues; i++)
		sum += values[i];
		
	return sum;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::getMaxValue
// ---------------------------------------------------------------------------------

//: Returns the maximum double in the list
double CDoubleArray::getMaxValue()
{
	if (numValues < 1)
		return 0.0;
		
	int i;
	double max = values[0];
	for (i = 1; i < numValues; i++)
		if ( values[i] > max)
			max = values[i];
		
	return max;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::getMinValue
// ---------------------------------------------------------------------------------

//: Returns the minimum double in the list
double CDoubleArray::getMinValue()
{
	if (numValues < 1)
		return 0.0;
		
	int i;
	double min = values[0];
	for (i = 1; i < numValues; i++)
		if ( values[i] < min)
			min = values[i];
		
	return min;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::setObject
// ---------------------------------------------------------------------------------

//: Sets the double at an index
//  If index>numValues then numValues becomes index+1 with memory 
//  allocated if needed.
void CDoubleArray::setValue(int index, double inValue)
{
	ThrowIf(index<0);

	if (index>=numValues) {
		numValues=index+1;
		checkCapacity(numValues);
	} 
		
	values[index]=inValue;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::add
// ---------------------------------------------------------------------------------

//: Adds an double to the end of the list
//  This expands the list's capacity if necessary.
int CDoubleArray::add(double inValue)
{
	checkCapacity(numValues+1);

	values[numValues]=inValue;
	numValues++;
	
	return numValues-1;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::insert
// ---------------------------------------------------------------------------------

//: Inserts an double into the list at a position
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new double. If the double's new position equals
//  numObjects then this adds it to the end of the list.
void CDoubleArray::insert(int index, double inValue)
{
	ThrowIf(index<0 || index>numValues+1);

	checkCapacity(numValues+1);
		
	for (int i=numValues; i>index; i--)
		values[i]=values[i-1];
	
	values[index]=inValue;
	numValues++;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::remove
// ---------------------------------------------------------------------------------

//: Removes the double at index and moves others up 
//  This function returns the double being removed.
double CDoubleArray::remove(int index)
{
	TestBounds(index, numValues);

	double tmp=values[index];

	for (int i=index; i<numValues-1; i++)
		values[i]=values[i+1];
		
	numValues--;
	
	return tmp;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the doubles in the list
void CDoubleArray::removeAll()
{
	numValues=0;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::swap
// ---------------------------------------------------------------------------------

//: Swaps a value at index1 with that at index2
void CDoubleArray::swap(int index1, int index2)
{
	TestBounds(index1, numValues);
	TestBounds(index2, numValues);

	double val = values[index1];
	values[index1] = values[index2];
	values[index2] = val;
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::sort
// ---------------------------------------------------------------------------------

//: Sorts the values in the list 
//  By default this is in ascending order unless inAscending is false.
void CDoubleArray::sort(bool inAscending)
{
	if (numValues > 1) {
	
			// Shellsort Algorithm
		
		int	hh = 1;						// Find starting "h" value
		int	stopH = numValues / 9;
		while (hh < stopH) {
			hh = 3 * hh + 1;
		}
		
		double tmp;
		
		int direction = (inAscending ? 1 : -1);
		
		for ( ; hh > 0; hh /= 3) {
				
			for (int step = hh + 1; step <= numValues; step++) {
				tmp=getValue(step - 1);
				
				long i;
				for (i = step - hh; i > 0; i -= hh) {
					if ((tmp > getValue(i-1) ? 1 : -1) * direction > 0) {
						break;
					}
					setValue(i + hh - 1, getValue(i-1));
				}
					
				setValue(i + hh - 1, tmp);
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		CDoubleArray::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the double list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new double array of this
//  size, copies all the old doubles into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void CDoubleArray::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity>capacity) {
		double *newValues;
		int i;
		
		// Increase capacity
		while (capacity<requiredCapacity)
			capacity+=increment;
		
		// Allocate the new list.
		NewArray(newValues, double, capacity);
		
		if (values!=NULL) {
			// Copy the existing list into the new list.
			for (i=0; i<numValues; i++)
				newValues[i]=values[i];
	
			// Free the old list.
			delete[] values;
		}
		
		// Point the old list to the new one.
		values=newValues;
	}
}

