/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  CExceptions.h

				CException classes

 ******************************************************************************/
 
#ifndef _CEXCEPTIONS_H_
#define _CEXCEPTIONS_H_

#include "CReportRecord.h"

#pragma mark === CException Classes ===

#pragma mark === Class CException ===

//: A base general-purpose exception class
//  This class provides a base for all MolEvol library exceptions.
//  It provides various constructors that allow the setting of 
//  various forms of messages, and numbers.
class CException {
	public:
		CReportElement *element1, *element2;
		
		CException() { element1 = NULL; element2 = NULL;}
			//: Default constructor - no message
			
		CException(const char *inMsg) 
				{ element1 = new CTextElement(inMsg); element2 = NULL;}
			//: Constructor with simple message
					
		CException(int inData) 
				{ element1 = new CIntegerElement(inData); element2 = NULL;}
			//: Constructor with numerical data
					
		CException(const char *inMsg, int inData) 
				{ element1 = new CTextElement(inMsg); element2 = new CIntegerElement(inData); }
			//: Constructor with message and number
					
		CException(const char *inMsg1, const char *inMsg2)
				{ element1 = new CTextElement(inMsg1); element2 = new CTextElement(inMsg2);}
			//: Constructor with two message strings
					
		CException(const char *inMsg1, const char *inMsg2, int inData)
				{ 	element1 = new CTextElement(inMsg1); 
					CComplexElement *e = new CComplexElement(); 
					e->add((char *)inMsg2); e->add(inData); 
					element2 = e;
				}
			//: Constructor with two message strings
					
		CException(CReportElement *inElement)
					{ element1 = inElement; element2 = NULL; }
			//: Constructor with report element
					
		CException(CReportElement *inElement1, CReportElement *inElement2)
					{ element1 = inElement1; element2 = inElement2; }
			//: Constructor with report element
										
		CReportElement *getReportElement1() { return element1; }
			//: Returns report element1
		
		CReportElement *getReportElement2() { return element2; }
			//: Returns report element1
			
		char *getMessage(char *ioMessage, int inMaxLength);
			//: Returns report element1
};


#pragma mark === Utility Exceptions ===

//: Thrown when a pointer is unexpectedly NULL
//  This is generally a bug in the program.
class CNullPointer : public CException { 
	public:
		CNullPointer(const char *inMsg, int inData) : CException("Null pointer:", inMsg, inData) { };
};

//: Thrown when a class allocation failed
//  Probably due to lack of memory.
class CNewClassFailed : public CException {  
	public:
		CNewClassFailed(const char *inMsg, int inData) : CException("Failed to create Class: ", inMsg, inData) { };
};

//: Thrown when an array allocation failed
//  Probably due to lack of memory.
class CNewArrayFailed : public CException {  
	public:
		CNewArrayFailed(const char *inMsg, int inData) : CException("Failed to create Class: ", inMsg, inData) { };
};

//: Thrown when the assert macros (ThrowIf, ThrowIfNot) fail
//  Usually used to test for conditions which should never happen (i.e. bugs).
class CAssertFailed : public CException {  
	public:
		CAssertFailed(const char *inMsg, int inData) : CException("Assert failed: ", inMsg, inData) { };
};

//: Thrown by the TestBounds macro
class CIndexOutOfBounds : public CException {  
	public:
		CIndexOutOfBounds() : CException("Index out of bounds") { };
		CIndexOutOfBounds(const char *inMsg, int inData) : CException("Index out of bounds: ", inMsg, inData) { };
};

//: Can be thrown in cases where the user has interrupted a long process.
class CInterrupted : public CException {  
	public:
		CInterrupted() : CException("User Interrupted") { };
		CInterrupted(const char *aMsg) : CException(aMsg) { };
};



#pragma mark === Input Parsing Exceptions ===

//: An exceptions to package another exception with a line number
class CErrorInFile : public CException { 
	protected:
		CException mException;
		int mLineNumber;
		
	public:
		CErrorInFile(CException inException, int inLineNumber) : 
			CException(),
			mException(inException),
			mLineNumber(inLineNumber)
			{ };

		CException getException() { return mException; }
		int getLineNumber() { return mLineNumber; }
};

//: The base to a number of parser related exceptions
class CParserError : public CException { 
	public:
		CParserError(const char *aMsg) : CException(aMsg) { };
		CParserError(const char *aMsg1, const char *aMsg2) : CException(aMsg1, aMsg2) { };
};

//: Throw when the input file format is unknown
class CUnknownFormat : public CParserError { 
	public:
		CUnknownFormat() : CParserError("Unknown File Format") { };
};

//: Thrown when the input file format is not yet supported
class CUnsupportedFormat : public CParserError { 
	public:
		CUnsupportedFormat(const char *aMsg) : CParserError("Unsupported File Format: ", aMsg) { };
};

//: Thrown when the data type (e.g. alignment, tree, etc.) is not supported in the current file format
class CUnsupportedData : public CParserError { 
	public:
		CUnsupportedData(const char *aMsg) : CParserError("Unsupported Data Type: ", aMsg) { };
};

//: Thrown when a required field in the file format is missing
class CMissingField : public CParserError { 
	public:
		CMissingField(const char *aMsg) : CParserError("Missing field: ", aMsg) { };
};

//: Thrown when a field in the file format is duplicated
class CDuplicateField : public CParserError { 
	public:
		CDuplicateField(const char *aMsg) : CParserError("Duplicated field: ", aMsg) { };
};

//: Thrown when a field in the file format is duplicated
class CAmbiguousField : public CParserError { 
	public:
		CAmbiguousField(const char *aMsg) : CParserError("Ambiguous field: ", aMsg) { };
};

//: Thrown when a field in the file format is unknown
class CUnknownField : public CParserError { 
	public:
		CUnknownField(const char *aMsg) : CParserError("Unknown field: ", aMsg) { };
};

//: A general problem in the file format
class CBadFormat : public CParserError { 
	public:
		CBadFormat(const char *aMsg) : CParserError("Error in input format: ", aMsg) { };
};

#pragma mark === Command Parsing Exceptions ===

//: Thrown when a field in the command format is unknown
class CUnknownModel: public CException { 
	public:
		CUnknownModel(const char *aMsg) : CException("Unknown model: ", aMsg) { };
};

//: Thrown when a field in the command format is unknown
class CUnknownModelType: public CException { 
	public:
		CUnknownModelType(const char *aMsg) : CException("Unknown model type: ", aMsg) { };
};

//: Thrown when a field in the command format is unknown
class CUnknownCommand: public CException { 
	public:
		CUnknownCommand(const char *aMsg) : CException("Unknown command: ", aMsg) { };
};

//: Thrown when a field in the command format is unknown
class CUnknownSubCommand: public CException { 
	public:
		CUnknownSubCommand(const char *aMsg) : CException("Unknown subcommand: ", aMsg) { };
};

//: Thrown when a field in the command format is unknown
class CUnknownValue: public CException { 
	public:
		CUnknownValue(const char *aMsg) : CException("Unknown command value: ", aMsg) { };
};

//: Thrown when a field in the command format is unknown
class CFieldMissing: public CException { 
	public:
		CFieldMissing(const char *aMsg) : CException(aMsg) { };
};


#pragma mark === File I/O Exceptions ===

//: A base class that provides exceptions for the file in/out
//  These are throw by the file using classes (MEStream, MEReader and CWriter) 
//  and their descendants.
class CFileError : public CException { 
	public:
		CFileError(const char *inMsg) : CException(inMsg) { };
		CFileError(const char *inMsg, int inData) : CException(inMsg, inData) { };
		CFileError(const char *inMsg1, const char *inMsg2) : CException(inMsg1, inMsg2) { };
};

//: A file was not opened due to an error
//  If opening a file for reading, then this will most likely be due to it not 
//  existing. Otherwise it will be some sort of operating system error.
class COpenFailed : public CFileError { 
	public:
		COpenFailed(const char *inMsg) : CFileError("Failed to open file, ", inMsg) { };
};

//: Thrown when the end of file is reached during a read operation
//  This can be caught in order to detect the end of file and finish the reading process.
class CEndOfFile : public CFileError {
	public:
		CEndOfFile() : CFileError("Unexpected end of file") { };
};

//: Thrown when attempting to read an integer but failed
class CIllegalInt : public CFileError {
	public:
		CIllegalInt() : CFileError("Unable to read integer") { };
};

//: Thrown when attempting to read a double but failed
class CIllegalDbl : public CFileError {
	public:
		CIllegalDbl() : CFileError("Unable to read double") { };
};

//: Thrown when some sort of write error occurred
//  These will probably be operating system level errors.
class CWriteError : public CFileError {
	public:
		CWriteError() : CFileError("Error writing file") { };
		CWriteError(const char *inMsg) : CFileError("Error writing file: ", inMsg) { };
		CWriteError(int inData) : CFileError("Error writing file: ", inData) { };
};

//: Thrown when some sort of read error occurred
//  These will probably be operating system level errors.
class CReadError : public CFileError {
	public:
		CReadError() : CFileError("Error writing file") { };
		CReadError(const char *inMsg) : CFileError("Error reading file: ", inMsg) { };
		CReadError(int inData) : CFileError("Error reading file: ", inData) { };
};

#pragma mark === CException Macros ===

// Macros to throw these exceptions

//: Used to throw a general purpose exception
//	Throws CException. 
//  The exception includes the name and line of the source file. 
//  This should be used to detect bugs and bad conditions in the 
//  program (i.e. if the program works, they should never be called).
#ifdef Throw
	#undef Throw
#endif

#define	Throw(message)												\
	do {															\
		throw CException(message									\
			", in file: " __FILE__ ", line ", __LINE__);			\
	} while (false)

//: Allocates a class using new and checks that the result is not NULL 
//	Throws MENewClassFailed. 
//  Usage: 
//	<PRE>NewClass(myPtr, MyClass(...));</PRE>
//  where <CODE>MyClass(...)</CODE> is the constructor.
//  The <CODE>delete</CODE> command should be used to free the memory.
#ifdef NewClass
	#undef NewClass
#endif

#define NewClass(ptr, constructor)									\
	do {															\
		ptr = new constructor;										\
		if ((ptr) == NULL) {											\
			throw CNewClassFailed(#ptr "=" #constructor 			\
			", in file: " __FILE__ ", line ", __LINE__);			\
		}															\
	} while (false)
	
//: Allocates an array using new and checks that the result is not NULL 
//	Throws MENewArrayFailed. 
//  Usage: 
//	<PRE>NewArray(myPtr, double, 100);</PRE>
//  which would allocate an array of 100 doubles to myPtr.
//  The <CODE>delete[]</CODE> command should be used to free the memory.
#ifdef NewArray
	#undef NewArray
#endif

#define NewArray(ptr, type, number)									\
	do {															\
		ptr = new type[number];										\
		if ((ptr) == NULL) {											\
			throw CNewArrayFailed(#ptr "=" #type "[" #number 		\
			"], in file: " __FILE__ ", line ", __LINE__);			\
		}															\
	} while (false)
	
#ifdef CORE_LIB_DEBUG_MACROS
// These versions of the macros are for whilst debugging

//: Checks to see if a pointer is NULL
//	Throws MENullPointer.
#ifdef ThrowIfNull
	#undef ThrowIfNull
#endif

#define	ThrowIfNull(ptr)												\
	do {															\
		if ((ptr) == NULL) throw CNullPointer(#ptr					\
			", in file: " __FILE__ ", line ", __LINE__);			\
	} while (false)

//: Checks to see if a pointer is NULL
//  Throws MENullPointer.
#ifdef ThrowIfNULL
	#undef ThrowIfNULL
#endif

#define	ThrowIfNULL(ptr)											\
	do {															\
		if ((ptr) == NULL) throw CNullPointer(#ptr					\
			", in file: " __FILE__ ", line ", __LINE__);			\
	} while (false)
	
//: Checks to see if an expression is true
//  Throws MEAssertFailed.
#ifdef ThrowIf
	#undef ThrowIf
#endif

#define	ThrowIf(test)												\
	do {															\
		if (test) throw CAssertFailed(#test						\
			", in file: " __FILE__ ", line ", __LINE__);			\
	} while (false)

//: Checks to see if an expression is false
//  Throws MEAssertFailed.
#ifdef ThrowIfNot
	#undef ThrowIfNot
#endif

#define	ThrowIfNot(test)											\
	do {															\
		if (!(test)) CAssertFailed(#test							\
			", in file: " __FILE__ ", line ", __LINE__);			\
	} while (false)

//: Checks to see if 0 <= index < bounds
//  Throws MEIndexOutOfBounds.
#ifdef TestBounds
	#undef TestBounds
#endif

#define	TestBounds(index, bounds)									\
	do {															\
		if (index<0) {												\
			throw CIndexOutOfBounds(#index 						\
			" is out of bounds (<0), in file: " __FILE__ 			\
			", line ", __LINE__);									\
		} else if (index>=bounds) {									\
			throw CIndexOutOfBounds(#index 						\
			" is out of bounds (>=" #bounds							\
			"), in file: " __FILE__ ", line ", __LINE__);			\
		}															\
	} while (false)

#else

#define	ThrowIfNull(ptr)
#define	ThrowIfNULL(ptr)
#define	ThrowIf(test)
#define	ThrowIfNot(test)
#define	TestBounds(index, bounds)

#endif

#endif /* _CEXCEPTIONS_H_ */
