/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CGraph.cpp
 
		CGraph Class
	 
 ******************************************************************************/

#include <stdio.h>
#include <string.h>

#include "CGraph.h"
#include "CDataSeries.h"
#include "CAxis.h"
#include "CLogAxis.h"
#include "CPlot.h"

#include "CGraphicsRecord.h"

#include "CExceptions.h"
#include "CDoubleArray.h"

#pragma mark === Construction & Destruction ===

//int CGraph::mMaxPlots = 16;

// ---------------------------------------------------------------------------------
//		 CGraph()
// ---------------------------------------------------------------------------------

CGraph::CGraph( )
{
	isCalibrated=false;
	
	xRangeManual = yRangeManual = false;
	xLogarithmic = yLogarithmic = false;
	
	xMinValue=0.0;
	xMaxValue=1.0;
	yMinValue=0.0;
	yMaxValue=1.0;
	
	mMajorTickSize = 8;
	mMinorTickSize = 4;
	
	mTitleText = NULL;
	mXAxisLabelText = NULL;
	mYAxisLabelText = NULL;
	
	mTitleFontFace = NULL;
	mAxisLabelFontFace = NULL;
	mTickLabelFontFace = NULL;
	
	mTitleFontSize = 20;
	mAxisLabelFontSize = 15;
	mTickLabelFontSize = 10;

	mXAxis = mYAxis = NULL;
	
	mNumPlots = 0;
}


// ---------------------------------------------------------------------------------
//		 ~CGraph()
// ---------------------------------------------------------------------------------

CGraph::~CGraph( )
{
	if (mXAxis != NULL)
		delete mXAxis;
	if (mYAxis != NULL)
		delete mYAxis;

	if (mTitleText != NULL)
		delete [] mTitleText;
	if (mXAxisLabelText != NULL)
		delete [] mXAxisLabelText;
	if (mYAxisLabelText != NULL)
		delete [] mYAxisLabelText;

	if (mTitleFontFace != NULL)
		delete [] mTitleFontFace;
	if (mAxisLabelFontFace != NULL)
		delete [] mAxisLabelFontFace;
	if (mTickLabelFontFace != NULL)
		delete [] mTickLabelFontFace;
		
	for (int i = 0; i < mNumPlots; i++)
		delete mPlots[i];
}

#pragma mark === Accessing ===

//-------------------------------------------------------------------------------------
// 		CGraph::SetAxes
//-------------------------------------------------------------------------------------

void CGraph::SetAxes(CAxis *inXAxis, CAxis *inYAxis)
{
	mXAxis=inXAxis;
	mYAxis=inYAxis;
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::AddPlot
//-------------------------------------------------------------------------------------

int CGraph::AddPlot(CPlot *inPlot)
{
	ThrowIf(mNumPlots == mMaxPlots);
	
	mPlots[mNumPlots] = inPlot;
	mNumPlots++;
	
	Invalidate();
	
	return mNumPlots - 1;
}


//-------------------------------------------------------------------------------------
// 		CGraph::GetPlot
//-------------------------------------------------------------------------------------

CPlot* CGraph::GetPlot(int index)
{
	TestBounds(index, mNumPlots);
	
	return mPlots[index];
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetPlotData
//-------------------------------------------------------------------------------------

void CGraph::SetPlotData(int index, CDataSeries *inData)
{
	TestBounds(index, mNumPlots);
	
	mPlots[index]->SetData(inData);
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetTitle
//-------------------------------------------------------------------------------------

void CGraph::SetTitle(char *inText)
{
	if (mTitleText != NULL)
		delete [] mTitleText;
		
	mTitleText = NULL;
	
	if (inText != NULL) {
		int len = strlen(inText);
		if (len > 0) {
			NewArray(mTitleText, char, len+1);
			strcpy(mTitleText, inText);
		}
	}
}

//-------------------------------------------------------------------------------------
// 		CGraph::SetXAxisLabel
//-------------------------------------------------------------------------------------

void CGraph::SetXAxisLabel(char *inText)
{
	if (mXAxisLabelText != NULL)
		delete [] mXAxisLabelText;
		
	mXAxisLabelText = NULL;
	
	if (inText != NULL) {
		int len = strlen(inText);
		if (len > 0) {
			NewArray(mXAxisLabelText, char, len+1);
			strcpy(mXAxisLabelText, inText);
		}
	}
}

//-------------------------------------------------------------------------------------
// 		CGraph::SetYAxisLabel
//-------------------------------------------------------------------------------------

void CGraph::SetYAxisLabel(char *inText)
{
	if (mYAxisLabelText != NULL)
		delete [] mYAxisLabelText;
		
	mYAxisLabelText = NULL;
	
	if (inText != NULL) {
		int len = strlen(inText);
		if (len > 0) {
			NewArray(mYAxisLabelText, char, len+1);
			strcpy(mYAxisLabelText, inText);
		}
	}
}

//-------------------------------------------------------------------------------------
// 		CGraph::SetTitleFont
//-------------------------------------------------------------------------------------

void CGraph::SetTitleFont(int inFontSize, char *inFontFace)
{
	if (mTitleFontFace != NULL)
		delete [] mTitleFontFace;
		
	mTitleFontFace = NULL;
	
	if (inFontFace != NULL) {
		int len = strlen(inFontFace);
		if (len > 0) {
			NewArray(mTitleFontFace, char, len+1);
			strcpy(mTitleFontFace, inFontFace);
		}
	}
	
	mTitleFontSize = inFontSize;
}

//-------------------------------------------------------------------------------------
// 		CGraph::SetAxisLabelFont
//-------------------------------------------------------------------------------------

void CGraph::SetAxisLabelFont(int inFontSize, char *inFontFace)
{
	if (mAxisLabelFontFace != NULL)
		delete [] mAxisLabelFontFace;
		
	mAxisLabelFontFace = NULL;
	
	if (inFontFace != NULL) {
		int len = strlen(inFontFace);
		if (len > 0) {
			NewArray(mAxisLabelFontFace, char, len+1);
			strcpy(mAxisLabelFontFace, inFontFace);
		}
	}
	
	mAxisLabelFontSize = inFontSize;
}

//-------------------------------------------------------------------------------------
// 		CGraph::SetTickLabelFont
//-------------------------------------------------------------------------------------

void CGraph::SetTickLabelFont(int inFontSize, char *inFontFace)
{
	if (mTickLabelFontFace != NULL)
		delete [] mTickLabelFontFace;
		
	mTickLabelFontFace = NULL;
	
	if (inFontFace != NULL) {
		int len = strlen(inFontFace);
		if (len > 0) {
			NewArray(mTickLabelFontFace, char, len+1);
			strcpy(mTickLabelFontFace, inFontFace);
		}
	}
	
	mTickLabelFontSize = inFontSize;
}



//-------------------------------------------------------------------------------------
// 		CGraph::SetXAxisLogarithmic
//-------------------------------------------------------------------------------------

void CGraph::SetXAxisLogarithmic(bool inIsLogarithmic)
{
	if (xLogarithmic != inIsLogarithmic) {
		if (mXAxis != NULL)
			delete mXAxis;
		mXAxis = NULL;
	}
	xLogarithmic = inIsLogarithmic;
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetYAxisLogarithmic
//-------------------------------------------------------------------------------------

void CGraph::SetYAxisLogarithmic(bool inIsLogarithmic)
{
	if (yLogarithmic != inIsLogarithmic) {
		if (mYAxis != NULL)
			delete mYAxis;
		mYAxis = NULL;
	}
	yLogarithmic = inIsLogarithmic;
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetXRangeManual
//-------------------------------------------------------------------------------------

void CGraph::SetXRangeManual(double minValue, double maxValue)
{
	xMinValue=minValue;
	xMaxValue=maxValue;
	xRangeManual=true;
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetXRangeAutomatic
//-------------------------------------------------------------------------------------

void CGraph::SetXRangeAutomatic()
{
	xRangeManual=false;
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetYRangeManual
//-------------------------------------------------------------------------------------

void CGraph::SetYRangeManual(double minValue, double maxValue)
{
	yMinValue=minValue;
	yMaxValue=maxValue;
	yRangeManual=true;
	
	Invalidate();
}


//-------------------------------------------------------------------------------------
// 		CGraph::SetYRangeAutomatic
//-------------------------------------------------------------------------------------

void CGraph::SetYRangeAutomatic()
{
	yRangeManual=false;
	
	Invalidate();
}


#pragma mark === Calibrating ===

//-------------------------------------------------------------------------------------
// 		CGraph::Calibrate
//	This is called by DrawSelf (and only by DrawSelf) whenever isCalibrated is false.
//-------------------------------------------------------------------------------------

void CGraph::Calibrate(CComplexGraphicElement *ioFigure)
{
	
	CalibrateXAxis();
	CalibrateYAxis();
	
	for (int i = 0; i < mNumPlots; i++) {
		mPlots[i]->SetAxes(mXAxis, mYAxis);
		mPlots[i]->SetGraph(this);
	}
	
	if (mTitleText != NULL)
		mPlotFrameTop = mTitleFontSize * 2;
	else
		mPlotFrameTop = 0.0;
		
	if (mXAxisLabelText != NULL)
		mPlotFrameLeft = mAxisLabelFontSize * 1.5;
	else
		mPlotFrameLeft = 0.0;
		
	int n=mYAxis->GetNumMajorTicks();
	int len1 = GetYTickLabelWidth(mYAxis->GetMajorTick(0));
	int len2 = GetYTickLabelWidth(mYAxis->GetMajorTick(n - 1));
	if (len2 > len1)
		len1 = len2;
		
	// rough guess at the width of the y axis tick labels...
	// Guessing that the average character width is 0.6 of the font height
	mPlotFrameLeft += mMajorTickSize + (len1 * mTickLabelFontSize * 0.6);
		
	if (mXAxisLabelText != NULL)
		mPlotFrameBottom = ioFigure->getHeight() - (mAxisLabelFontSize * 1.5);
	else
		mPlotFrameBottom = ioFigure->getHeight();
		
	mPlotFrameBottom -= mMajorTickSize + mTickLabelFontSize;

	n=mXAxis->GetNumMajorTicks();
	len1 = GetYTickLabelWidth(mYAxis->GetMajorTick(n - 1));
		
	// The last tick label may stick out so add space on the right...
	mPlotFrameRight = ioFigure->getWidth() - (len1 * mTickLabelFontSize * 0.3);

	mHeight = mPlotFrameBottom - mPlotFrameTop;
	mWidth = mPlotFrameRight - mPlotFrameLeft; 

	xMin = mXAxis->Transform(mXAxis->GetMinAxis());
	yMin = mYAxis->Transform(mYAxis->GetMinAxis());
	xScale = mWidth / (mXAxis->Transform(mXAxis->GetMaxAxis()) - xMin);
	yScale = mHeight / (mYAxis->Transform(mYAxis->GetMaxAxis()) - yMin);
	
	isCalibrated=true;
}


//-------------------------------------------------------------------------------------
// 		CGraph::CalibrateXAxis
//-------------------------------------------------------------------------------------

void CGraph::CalibrateXAxis()
{
	CDoubleArray *col;
	
	if (mXAxis == NULL) {
		if (xLogarithmic)	
			mXAxis = new CLogAxis();
		else
			mXAxis = new CAxis();
	}

	CPlot *plot;
	
	if (xRangeManual) {
		mXAxis->SetRange(xMinValue, xMaxValue);
	} else {
		col=mPlots[0]->GetData()->GetXData();
		mXAxis->SetRange(col);
		
		for (int i = 1; i < mNumPlots; i++) {
			col=mPlots[i]->GetData()->GetXData();
			mXAxis->AddRange(col);
		}
	}
}



//-------------------------------------------------------------------------------------
// 		CGraph::CalibrateYAxis
//-------------------------------------------------------------------------------------

void CGraph::CalibrateYAxis()
{
	CDoubleArray *col;
	
	if (mYAxis == NULL) {
		if (yLogarithmic)	
			mYAxis = new CLogAxis();
		else
			mYAxis = new CAxis();
	}

	CPlot *plot;
	
	if (yRangeManual) {
		mYAxis->SetRange(yMinValue, yMaxValue);
	} else {
		col=mPlots[0]->GetData()->GetYData();
		mYAxis->SetRange(col);
		
		if (mPlots[0]->GetData()->GetHasLimits()) {
			col=mPlots[0]->GetData()->GetYHigh();
			mYAxis->AddRange(col);
			col=mPlots[0]->GetData()->GetYLow();
			mYAxis->AddRange(col);
		}

		for (int i = 1; i < mNumPlots; i++) {
			col=mPlots[i]->GetData()->GetYData();
			mYAxis->AddRange(col);
		
			if (mPlots[i]->GetData()->GetHasLimits()) {
				col=mPlots[i]->GetData()->GetYHigh();
				mYAxis->AddRange(col);
				col=mPlots[i]->GetData()->GetYLow();
				mYAxis->AddRange(col);
			}
		}
	}	
	
}

#pragma mark === Drawing ===

//-------------------------------------------------------------------------------------
// 		CGraph::ConvertX
//-------------------------------------------------------------------------------------

double CGraph::ConvertX(double x)
{
	return mPlotFrameLeft + ((mXAxis->Transform(x) - xMin) * xScale); 
}

//-------------------------------------------------------------------------------------
// 		CGraph::ConvertY
//-------------------------------------------------------------------------------------

double CGraph::ConvertY(double y)
{
	return mPlotFrameBottom - ((mYAxis->Transform(y) - yMin) * yScale); 
}

#pragma mark === Drawing with CGraphicElement ===

//-------------------------------------------------------------------------------------
// 		CGraph::GetYTickLabelWidth
//-------------------------------------------------------------------------------------

int CGraph::GetYTickLabelWidth(double y)
{
	char label[64];
	if (mYAxis->GetIsDiscrete())
		sprintf(label, "%ld", (long)y);
	else if (mYAxis->GetUseScientific())
		sprintf(label, "%G", y);
	else
		sprintf(label, "%f", y);

	return strlen(label);
}
	
//-------------------------------------------------------------------------------------
// 		CGraph::DrawMajorTickY
//-------------------------------------------------------------------------------------

void CGraph::DrawMajorTickY(CComplexGraphicElement *ioGraphic, double y)
{
	double yPos=ConvertY(y);
	CLineElement *line = new CLineElement(mPlotFrameLeft, yPos, mPlotFrameLeft - mMajorTickSize, yPos);
	line->setStrokeStyle(1.0, kSquareLineCap);
	line->setStrokeColour(kBlackColour);
	ioGraphic->add(line);

	char label[64];
	if (mYAxis->GetIsDiscrete())
		sprintf(label, "%ld", (long)y);
	else if (mYAxis->GetUseScientific())
		sprintf(label, "%G", y);
	else
		sprintf(label, "%f", y);

	CTextGraphicElement *text = new CTextGraphicElement(mPlotFrameLeft - (mMajorTickSize * 1.25), yPos + 3,
															label, kRightTextAnchor);
	text->setFontStyle(mTickLabelFontSize, mTickLabelFontFace);
	ioGraphic->add(text);
}
	
//-------------------------------------------------------------------------------------
// 		CGraph::DrawMinorTickY
//-------------------------------------------------------------------------------------

void CGraph::DrawMinorTickY(CComplexGraphicElement *ioGraphic, double y) {
	double yPos=ConvertY(y);
	CLineElement *line = new CLineElement(mPlotFrameLeft, yPos, mPlotFrameLeft - mMinorTickSize, yPos);
	line->setStrokeStyle(0.75, kSquareLineCap);
	line->setStrokeColour(kBlackColour);
	ioGraphic->add(line);
}
	
//-------------------------------------------------------------------------------------
// 		CGraph::DrawYAxis
//-------------------------------------------------------------------------------------

void CGraph::DrawYAxis(CComplexGraphicElement *ioGraphic)
{
	int n1=mYAxis->GetNumMajorTicks();
	int n2, i, j;
	
	n2=mYAxis->GetNumMinorTicks(-1);
	if (mYAxis->GetLabelFirst()) { // Draw first minor tick as a major one (with a label)
		DrawMajorTickY(ioGraphic, mYAxis->GetMinorTick(0, -1));
		for (j=1; j<n2; j++) 
			DrawMinorTickY(ioGraphic, mYAxis->GetMinorTick(j, -1));
	} else {
		for (j=0; j<n2; j++) 
			DrawMinorTickY(ioGraphic, mYAxis->GetMinorTick(j, -1));
	}
	for (i=0; i<n1; i++) {
		DrawMajorTickY(ioGraphic, mYAxis->GetMajorTick(i));
		n2=mYAxis->GetNumMinorTicks(i);
		if (i==(n1-1) && mYAxis->GetLabelLast()) { // Draw last minor tick as a major one
			DrawMajorTickY(ioGraphic, mYAxis->GetMinorTick(0, i));
			for (j=1; j<n2; j++) 
				DrawMinorTickY(ioGraphic, mYAxis->GetMinorTick(j, i));
		} else {
			for (j=0; j<n2; j++) 
				DrawMinorTickY(ioGraphic, mYAxis->GetMinorTick(j, i));
		}
	}
}

//-------------------------------------------------------------------------------------
// 		CGraph::GetXTickLabelWidth
//-------------------------------------------------------------------------------------

int CGraph::GetXTickLabelWidth(double x)
{
	char label[64];
	if (mXAxis->GetIsDiscrete())
		sprintf(label, "%ld", (long)x);
	else if (mYAxis->GetUseScientific())
		sprintf(label, "%G", x);
	else
		sprintf(label, "%f", x);

	return strlen(label);
}
	
//-------------------------------------------------------------------------------------
// 		CGraph::DrawMajorTickX
//-------------------------------------------------------------------------------------

void CGraph::DrawMajorTickX(CComplexGraphicElement *ioGraphic, double x)
{
	double xPos=ConvertX(x);
	CLineElement *line = new CLineElement(xPos, mPlotFrameBottom, xPos, mPlotFrameBottom + mMajorTickSize);
	line->setStrokeStyle(1.0, kSquareLineCap);
	line->setStrokeColour(kBlackColour);
	ioGraphic->add(line);

	char label[64];
	if (mXAxis->GetIsDiscrete())
		sprintf(label, "%ld", (long)x);
	else if (mXAxis->GetUseScientific())
		sprintf(label, "%G", x);
	else
		sprintf(label, "%G", x);

	CTextGraphicElement *text = new CTextGraphicElement(xPos, mPlotFrameBottom + (mMajorTickSize * 1.25) + 10,
														label, kMiddleTextAnchor);
	text->setFontStyle(mTickLabelFontSize, mTickLabelFontFace);
	ioGraphic->add(text);
}
	
//-------------------------------------------------------------------------------------
// 		CGraph::DrawMinorTickX
//-------------------------------------------------------------------------------------

void CGraph::DrawMinorTickX(CComplexGraphicElement *ioGraphic, double x) {
	double xPos=ConvertX(x);
	CLineElement *line = new CLineElement(xPos, mPlotFrameBottom, xPos, mPlotFrameBottom + mMinorTickSize);
	line->setStrokeStyle(0.75, kSquareLineCap);
	line->setStrokeColour(kBlackColour);
	ioGraphic->add(line);
}
	
//-------------------------------------------------------------------------------------
// 		CGraph::DrawXAxis
//-------------------------------------------------------------------------------------

void CGraph::DrawXAxis(CComplexGraphicElement *ioGraphic)
{
	int n1=mXAxis->GetNumMajorTicks();
	int n2, i, j;
	
	n2=mXAxis->GetNumMinorTicks(-1);
	if (mXAxis->GetLabelFirst()) { // Draw first minor tick as a major one (with a label)
		DrawMajorTickX(ioGraphic, mXAxis->GetMinorTick(0, -1));
		for (j=1; j<n2; j++) 
			DrawMinorTickX(ioGraphic, mXAxis->GetMinorTick(j, -1));
	} else {
		for (j=0; j<n2; j++) 
			DrawMinorTickX(ioGraphic, mXAxis->GetMinorTick(j, -1));
	}
	for (i=0; i<n1; i++) {
		DrawMajorTickX(ioGraphic, mXAxis->GetMajorTick(i));
		n2=mXAxis->GetNumMinorTicks(i);
		if (i==(n1-1) && mXAxis->GetLabelLast()) { // Draw last minor tick as a major one
			DrawMajorTickX(ioGraphic, mXAxis->GetMinorTick(0, i));
			for (j=1; j<n2; j++) 
				DrawMinorTickX(ioGraphic, mXAxis->GetMinorTick(j, i));
		} else {
			for (j=0; j<n2; j++) 
				DrawMinorTickX(ioGraphic, mXAxis->GetMinorTick(j, i));
		}
	}
}


//-------------------------------------------------------------------------------------
// 		CGraph::DrawPlot
//-------------------------------------------------------------------------------------

void CGraph::DrawGraph(CComplexGraphicElement *ioGraphic, 
						double inXTolerance, 
						double inYTolerance)
{
	if (mNumPlots == 0) 
		return;
		
	if (!isCalibrated) {
		Calibrate(ioGraphic);
	}
	
	CTextGraphicElement *text;
	
	if (mTitleText != NULL) {
		text = new CTextGraphicElement(ioGraphic->getWidth() / 2.0, mTitleFontSize, 
										mTitleText, kMiddleTextAnchor);
		text->setFontStyle(mTitleFontSize, mTitleFontFace);
		ioGraphic->add(text);
	}
	
	CRectElement *frame = new CRectElement(mPlotFrameLeft, mPlotFrameTop, mWidth, mHeight);
	frame->setStrokeStyle(1.0, kDefaultLineCap, kDefaultLineJoin);
	frame->setStrokeColour(kBlackColour);
	frame->setFillColour(kNoColour);
	
	ioGraphic->add(frame);
	
	for (int i = 0; i < mNumPlots; i++)
		mPlots[i]->DrawPlot(ioGraphic, inXTolerance, inYTolerance);

	CGroupElement *group = new CGroupElement();
	DrawXAxis(group);
	ioGraphic->add(group);

	if (mXAxisLabelText != NULL) {
		text = new CTextGraphicElement(mPlotFrameLeft + (mWidth / 2.0), 
										ioGraphic->getHeight() - mAxisLabelFontSize * 0.2, 
										mXAxisLabelText, kMiddleTextAnchor);
		text->setFontStyle(mAxisLabelFontSize, mAxisLabelFontFace);
		ioGraphic->add(text);
	}
	
	group = new CGroupElement();
	DrawYAxis(group);
	ioGraphic->add(group);

	if (mXAxisLabelText != NULL) {
		text = new CTextGraphicElement(mAxisLabelFontSize, mPlotFrameTop + (mHeight / 2.0),
										mYAxisLabelText, kMiddleTextAnchor);
		text->setFontStyle(mAxisLabelFontSize, mAxisLabelFontFace);
		text->setRotation(-90.0, mAxisLabelFontSize, mPlotFrameTop + (mHeight / 2.0));
		ioGraphic->add(text);
	}
}	

