/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CGraphicsRecord.h

		Interface for the CGraphicsRecord Class
	
 ******************************************************************************/

#ifndef _GRAPHICS_RECORD_H_
#define _GRAPHICS_RECORD_H_

#include "CReportRecord.h"

class CDataSeries;

enum GraphicType {
	kLineGraphicElement,
	kRectGraphicElement,
	kRoundedRectGraphicElement,
	kCircleGraphicElement,
	kEllipseGraphicElement,
	kTextGraphicElement,
	kPathGraphicElement,
	kGroupGraphicElement,
	kViewPortGraphicElement,
	kFigureGraphicElement
};

enum PageUnitsType {
	kNoPageUnits,
	kPixelPageUnits,
	kCMPageUnits,
	kInchPageUnits,
	kPointPageUnits
};

// ---------------------------------------------------------------------------------
//		class CGraphicElement
// ---------------------------------------------------------------------------------

enum StrokeLineCapType {
	kDefaultLineCap,
	kButtLineCap,
	kRoundLineCap,
	kSquareLineCap
};

enum StrokeLineJoinType {
	kDefaultLineJoin,
	kMiterLineJoin,
	kRoundLineJoin,
	kBevelLineJoin
};

enum ColourType {
	kNoColour,
	kRGBColour,
	kBlackColour,
	kBlueColour,
	kBrownColour,
	kCyanColour,
	kDarkGrayColour,
	kGrayColour,
	kGreenColour,
	kLightGrayColour,
	kMagentaColour,
	kOrangeColour,
	kPurpleColour,
	kRedColour,
	kWhiteColour,
	kYellowColour
};

#pragma mark CColourRec
struct CColourRec {
	public:
		ColourType					mType;
		double						mRed, mGreen, mBlue, mAlpha;
};


#pragma mark CGraphicElement
class CGraphicElement : public CReportElement {
	protected:
		double						mX, mY;
		double 						mWidth, mHeight;

		double						mStrokeWidth;
		bool						mHasStrokeWidth;
		
		StrokeLineCapType			mLineCap;
		bool						mHasLineCap;
		
		StrokeLineJoinType			mLineJoin;
		bool						mHasLineJoin;
		
		CColourRec					mStrokeColour;
		bool						mHasStrokeColour;
		
		CColourRec					mFillColour;
		bool						mHasFillColour;
		
		bool						mHasRotation;
		double						mRotation;
		double						mRotationCentreX, mRotationCentreY;
		
		bool						mHasScale;
		double						mScaleX, mScaleY;
		double						mScaleCentreX, mScaleCentreY;
				
		bool						mHasTranslation;
		double						mTransX, mTransY;
				
	public:
									CGraphicElement();
									CGraphicElement(const CGraphicElement &inSource);
		virtual						~CGraphicElement();		

		virtual void				setStyle(const CGraphicElement &inSource);
		virtual void				clearStyle();
		
		virtual void				setPosition(double inX, double inY)
										{ setTopLeftPosition(inX, inY); }
		virtual void				setTopLeftPosition(double inX, double inY) = 0;
		virtual void				setCentrePosition(double inX, double inY) = 0;
		virtual void				setSize(double inWidth, double inHeight)
										{ mWidth = inWidth; mHeight = inHeight; }
										
		virtual double				getX() { return mX; }
		virtual double				getY() { return mY; }
		virtual double				getWidth() { return mWidth; }
		virtual double				getHeight() { return mHeight; }
		
		virtual double				getLeft() { return mX; }
		virtual double				getTop() { return mY; }
		virtual double				getRight() { return mX + mWidth; }
		virtual double				getBottom() { return mY + mHeight; }
		 
		virtual void				setStrokeStyle(	double inStrokeWidth, 
													StrokeLineCapType inLineCap = kButtLineCap, 
													StrokeLineJoinType inLineJoin = kMiterLineJoin);
												
		virtual void				setStrokeWidth(double inStrokeWidth);
		virtual bool				getHasStrokeWidth() { return mHasStrokeWidth; }
		virtual double				getStrokeWidth() { return mStrokeWidth; }

		virtual void				setStrokeLineCap(StrokeLineCapType inLineCap);
		virtual bool				getHasStrokeLineCap() { return mHasLineCap; }
		virtual StrokeLineCapType	getStrokeLineCap() { return mLineCap; }

		virtual void				setStrokeLineJoin(StrokeLineJoinType inLineJoin);
		virtual bool				getHasStrokeLineJoin() { return mHasLineJoin; }
		virtual StrokeLineJoinType	getStrokeLineJoin() { return mLineJoin; }
		
		virtual void				setStrokeColour(const CColourRec &inColour);
		virtual void				setStrokeColour(ColourType inColour, double inAlpha = 1.0);
		virtual void				setStrokeColour(double inRed, double inGreen, double inBlue, double inAlpha = 1.0);
		virtual bool				getHasStrokeColour() { return mHasStrokeColour; }
		virtual ColourType			getStrokeColourType() { return mStrokeColour.mType; }
		virtual double				getStrokeColourRed() { return mStrokeColour.mRed; }
		virtual double				getStrokeColourGreen() { return mStrokeColour.mGreen; }
		virtual double				getStrokeColourBlue() { return mStrokeColour.mBlue; }
		virtual double				getStrokeColourAlpha() { return mStrokeColour.mAlpha; }
		
		virtual void				setFillColour(const CColourRec &inColour);
		virtual void				setFillColour(ColourType inColour, double inAlpha = 1.0);
		virtual void				setFillColour(double inRed, double inGreen, double inBlue, double inAlpha = 1.0);
		virtual bool				getHasFillColour() { return mHasFillColour; }
		virtual ColourType			getFillColourType() { return mFillColour.mType; }
		virtual double				getFillColourRed() { return mFillColour.mRed; }
		virtual double				getFillColourGreen() { return mFillColour.mGreen; }
		virtual double				getFillColourBlue() { return mFillColour.mBlue; }
		virtual double				getFillColourAlpha() { return mFillColour.mAlpha; }
		
		virtual void				setRotation(double inRotation, 
												double inCentreX = 0.0,
												double inCentreY = 0.0) 
										{ 	mHasRotation = true; 
											mRotation = inRotation; 
											mRotationCentreX = inCentreX;
											mRotationCentreY = inCentreY; }
		virtual void				clearRotation()
										{ 	mHasRotation = false; } 

		virtual bool				getHasRotation() { return mHasRotation; }
		virtual double				getRotation() { return mRotation; }
		virtual double				getRotationCentreX() { return mRotationCentreX; }
		virtual double				getRotationCentreY() { return mRotationCentreY; }

		virtual void				setScale(double inScaleX, double inScaleY,
												double inCentreX = 0.0,
												double inCentreY = 0.0) 
										{ 	mHasScale = true; 
											mScaleX = inScaleX; 
											mScaleY = inScaleY; 
											mScaleCentreX = inCentreX;
											mScaleCentreY = inCentreY; }
		virtual void				clearScale()
										{ 	mHasScale = false; } 

		virtual bool				getHasScale() { return mHasScale; }
		virtual double				getScaleX() { return mScaleX; }
		virtual double				getScaleY() { return mScaleY; }
		virtual double				getScaleCentreX() { return mScaleCentreX; }
		virtual double				getScaleCentreY() { return mScaleCentreY; }

		virtual void				setTranslation(double inTransX, double inTransY) 
										{ 	mHasTranslation = true; 
											mTransX = inTransX; 
											mTransY = inTransY; }
		virtual void				clearTranslation()
										{ 	mHasTranslation = false; } 

		virtual bool				getHasTranslation() { return mHasTranslation; }
		virtual double				getTranslationX() { return mTransX; }
		virtual double				getTranslationY() { return mTransY; }
		
		virtual ElementType			getElementType() { return kGraphicElement; }	
		
		virtual GraphicType			getGraphicType() = 0;
};

// ---------------------------------------------------------------------------------
//		class CLineElement
// ---------------------------------------------------------------------------------

#pragma mark CLineElement
class CLineElement : public CGraphicElement {
	protected:
		double						mX2, mY2;
		
	public:
									CLineElement(double inX1, double inY1, 
													double inX2, double inY2);
									CLineElement(const CLineElement &inSource);

		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);
		virtual void				setSize(double inWidth, double inHeight);
		
		virtual double				getX1() { return mX; }
		virtual double				getY1() { return mY; }
		virtual double				getX2() { return mX2; }
		virtual double				getY2() { return mY2; }
		
		virtual double				getLeft();
		virtual double				getTop();
		virtual double				getRight();
		virtual double				getBottom();
		 
		virtual GraphicType			getGraphicType() { return kLineGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CRectElement
// ---------------------------------------------------------------------------------

#pragma mark CRectElement
class CRectElement : public CGraphicElement {
	public:
									CRectElement(double inX, double inY, 
													double inWidth, double inHeight);
									CRectElement(const CRectElement &inSource);

		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);

		virtual GraphicType			getGraphicType() { return kRectGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CRoundedRectElement
// ---------------------------------------------------------------------------------

#pragma mark CRoundedRectElement
class CRoundedRectElement : public CRectElement {
	protected:
		double						mRadiusX, mRadiusY;

	public:
									CRoundedRectElement(double inX, double inY, 
														double inWidth, double inHeight,
														double inRadiusX, double inRadiusY);
									CRoundedRectElement(const CRoundedRectElement &inSource);

		virtual double				getRadiusX() { return mRadiusX; }
		virtual double				getRadiusY() { return mRadiusY; }

		virtual GraphicType			getGraphicType() { return kRoundedRectGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CCircleElement
// ---------------------------------------------------------------------------------

#pragma mark CCircleElement
class CCircleElement : public CGraphicElement {
	protected:
		double						mRadius;
		
	public:
									CCircleElement(double inX, double inY, double inRadius);
									CCircleElement(const CCircleElement &inSource);

		virtual void				setPosition(double inX, double inY);
		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);
		virtual void				setSize(double inWidth, double inHeight);

		virtual double				getRadius() { return mRadius; }

		virtual double				getLeft();
		virtual double				getTop();
		virtual double				getRight();
		virtual double				getBottom();
		 
		virtual GraphicType			getGraphicType() { return kCircleGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CEllipseElement
// ---------------------------------------------------------------------------------

#pragma mark CEllipseElement
class CEllipseElement : public CGraphicElement {
	protected:
		double						mRadiusX, mRadiusY;
		
	public:
									CEllipseElement(double inX, double inY, 
													double inRadiusX, double inRadiusY);
									CEllipseElement(const CEllipseElement &inSource);

		virtual void				setPosition(double inX, double inY);
		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);
		virtual void				setSize(double inWidth, double inHeight);

		virtual double				getRadiusX() { return mRadiusX; }
		virtual double				getRadiusY() { return mRadiusY; }

		virtual double				getLeft();
		virtual double				getTop();
		virtual double				getRight();
		virtual double				getBottom();
		 
		virtual GraphicType			getGraphicType() { return kEllipseGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CTextGraphicElement
// ---------------------------------------------------------------------------------

enum TextAnchorType {
	kDefaultTextAnchor,
	kLeftTextAnchor,
	kMiddleTextAnchor,
	kRightTextAnchor
};

#pragma mark CTextGraphicElement
class CTextGraphicElement : public CGraphicElement {
	protected:
		char						*mText;
		
		bool						mHasTextAnchor;
		TextAnchorType				mTextAnchor;
		
		bool						mHasFontSize;
		double						mFontSize;
		
		bool						mHasFontFace;
		char*						mFontFace;
		
	public:
									CTextGraphicElement(double inX, double inY, 
														char *inText,
														TextAnchorType inAnchor = kDefaultTextAnchor);
									CTextGraphicElement(const CTextGraphicElement &inSource);
		virtual						~CTextGraphicElement();		

		virtual void				setPosition(double inX, double inY);
		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);

		virtual double				getLeft();
		virtual double				getTop();
		virtual double				getRight();
		virtual double				getBottom();
		 
		virtual void				setText(char *inText);
		virtual char*				getText() { return mText; }

		virtual void				setTextAnchor(TextAnchorType inAnchor = kDefaultTextAnchor);
		virtual bool				getHasTextAnchor() { return mHasTextAnchor; }
		virtual TextAnchorType		getTextAnchor() { return mTextAnchor; }

		virtual void				setFontStyle(double inFontSize, char *inFontFace);
		virtual void				clearFontStyle();

		virtual void				setFontSize(double inFontSize);
		virtual bool				getHasFontSize() { return mHasFontSize; }
		virtual double				getFontSize() { return mFontSize; }
		
		virtual void				setFontFace(char *inFontFace);
		virtual bool				getHasFontFace() { return mHasFontFace; }
		virtual char*				getFontFace() { return mFontFace; }

		virtual GraphicType			getGraphicType() { return kTextGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CPathSubelement
// ---------------------------------------------------------------------------------

enum PathSubelementType {
	kMoveToSubelement,
	kMoveBySubelement,
	kLineToSubelement,
	kLineBySubelement,
	kHLineToSubelement,
	kHLineBySubelement,
	kVLineToSubelement,
	kVLineBySubelement,
	kCurveToSubelement,
	kCurveBySubelement,
	kArcSubelement
//	kArcToSubelement,
//	kArcBySubelement
};

#pragma mark CPathSubelement
class CPathSubelement {
	public:
		PathSubelementType			mType;
		double						mX1, mY1, mX2, mY2, mX3, mY3;

									CPathSubelement();
									CPathSubelement(PathSubelementType inType, double inX, double inY);
									CPathSubelement(PathSubelementType inType, double inX1, double inY1,
													double inX2, double inY2);
									CPathSubelement(PathSubelementType inType, double inX1, double inY1,
													double inX2, double inY2, double inX3, double inY3);
};

// ---------------------------------------------------------------------------------
//		class CPathElement
// ---------------------------------------------------------------------------------

#pragma mark CPathElement
class CPathElement : public CGraphicElement {
	protected:
		bool						mPathClosed;
		
		double						mLeft, mTop, mRight, mBottom;
		double						mLastX, mLastY;
		
		CPathSubelement*			mSubelements;
		int							mNumSubelements;
		int							mCapacity;
		
	public:
									CPathElement(double inX, double inY);
									CPathElement(const CPathElement &inSource);
		virtual						~CPathElement();		

		virtual GraphicType			getGraphicType() { return kPathGraphicElement; }	

		virtual void				setPosition(double inX, double inY);
		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);

		virtual void				setPathClosed(bool inPathClosed) { mPathClosed = inPathClosed; }
		virtual bool				getPathClosed() { return mPathClosed; }
		
		virtual double				getLeft() { return mLeft; }
		virtual double				getTop() { return mTop; }
		virtual double				getRight() { return mRight; }
		virtual double				getBottom() { return mBottom; }
		 
		virtual void				moveTo(double inX, double inY);
		virtual void				moveBy(double inDX, double inDY);
		
		virtual void				lineTo(double inX, double inY);
		virtual void				lineBy(double inDX, double inDY);
		
		virtual void				horizLineTo(double inX);
		virtual void				horizLineBy(double inDX);
		
		virtual void				vertLineTo(double inY);
		virtual void				vertLineBy(double inDY);
		
		virtual void				curveTo(double inX, double inY, 
											double inCX1, double inCY1,
											double inCX2, double inCY2);
		virtual void				curveBy(double inDX, double inDY, 
											double inCX1, double inCY1,
											double inCX2, double inCY2);
		
		virtual void				arc(double inCX, double inCY, 
											double inAngle1, double inAngle2);
//		virtual void				arcTo(double inX, double inY, 
//											double inRX, double inRY, double inRotate,
//											bool inLargeArcFlag, bool inSweepFlag);
//		virtual void				arcBy(double inDX, double inDY, 
//											double inRX, double inRY, double inRotate,
//											bool inLargeArcFlag, bool inSweepFlag);
	protected:
		virtual void				calcBounds(double inX, double inY);
		virtual void				addSublement(CPathSubelement inSubelement);
		
	public:
		virtual int					getNumSubelements() { return mNumSubelements; }
		virtual CPathSubelement		getSubelement(int index);
						
	protected:
		virtual void				checkCapacity(int requiredCapacity);
};

// ---------------------------------------------------------------------------------
//		class CComplexGraphicElement
// ---------------------------------------------------------------------------------

#pragma mark CComplexGraphicElement
class CComplexGraphicElement : public CGraphicElement {
	protected:
		CGraphicElement**			mElements;
		int							mNumElements;
		int							mCapacity;
				
		double						mLeft, mTop, mRight, mBottom;

	public:
									CComplexGraphicElement();
									CComplexGraphicElement(const CComplexGraphicElement &inSource);
		virtual						~CComplexGraphicElement();
							
		virtual void				setTopLeftPosition(double inX, double inY);
		virtual void				setCentrePosition(double inX, double inY);
		virtual void				setSize(double inWidth, double inHeight);

		virtual double				getLeft() { return mLeft; }
		virtual double				getTop() { return mTop; }
		virtual double				getRight() { return mRight; }
		virtual double				getBottom() { return mBottom; }
		 
		virtual void				add(CGraphicElement *inElement);
		
		virtual int					getNumElements() { return mNumElements; }
		virtual CGraphicElement*	getElement(int index);
						
	protected:
		virtual void				calcBounds(CGraphicElement *inElement);
		
		virtual void				checkCapacity(int requiredCapacity);
};

// ---------------------------------------------------------------------------------
//		class CGroupElement
// ---------------------------------------------------------------------------------

#pragma mark CGroupElement
class CGroupElement : public CComplexGraphicElement {
	public:
									CGroupElement();
									CGroupElement(const CGroupElement &inSource);
		virtual						~CGroupElement();
					
		
		virtual GraphicType			getGraphicType() { return kGroupGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CViewPortElement
// ---------------------------------------------------------------------------------

enum AspectRatioType {
	kNoneAspect,
	kXMinYMinAspect,
	kXMidYMinAspect,
	kXMaxYMinAspect,
	kXMinYMidAspect,
	kXMidYMidAspect,
	kXMaxYMidAspect,
	kXMinYMaxAspect,
	kXMidYMaxAspect,
	kXMaxYMaxAspect
};

#pragma mark CViewPortElement
class CViewPortElement : public CComplexGraphicElement {
	protected:
		bool						mHasViewBox;
		double						mViewBoxMinX, mViewBoxMinY;
		double						mViewBoxWidth, mViewBoxHeight;
		
		AspectRatioType				mPreserveAspectRatio;
		bool						mSliceViewBox;
		
		PageUnitsType				mPageUnits;
		
	public:
									CViewPortElement();
									CViewPortElement(const CViewPortElement &inSource);
		virtual						~CViewPortElement();
					
		virtual void				setSize(double inWidth, double inHeight);
		virtual void				setSize(PageUnitsType inPageUnits, double inWidth, double inHeight);
		virtual PageUnitsType		getPageUnits() { return mPageUnits; }

		virtual void				setViewBox(double inMinX, double inMinY, 
												double inWidth, double inHeight);
		virtual void				setPreserveAspectRatio(AspectRatioType inType)
										{ mPreserveAspectRatio = inType; }
		virtual void				setSliceViewBox(bool inSlice) 
										{ mSliceViewBox = inSlice; }
		
		virtual double				getHasViewBox() { return mHasViewBox; }
		virtual double				getViewBoxMinX() { return mViewBoxMinX; }
		virtual double				getViewBoxMinY() { return mViewBoxMinY; }
		virtual double				getViewBoxWidth() { return mViewBoxWidth; }
		virtual double				getViewBoxHeight() { return mViewBoxHeight; }
		
		virtual AspectRatioType		getPreserveAspectRatio() { return mPreserveAspectRatio; }
		virtual bool				getSliceViewBox() { return mSliceViewBox; }
		
		virtual GraphicType			getGraphicType() { return kViewPortGraphicElement; }	
};

// ---------------------------------------------------------------------------------
//		class CFigureElement
// ---------------------------------------------------------------------------------

#pragma mark CFigureElement
class CFigureElement : public CViewPortElement {
	protected:
		char*						mName;
		char*						mTitle;
	
	public:
									CFigureElement(char *inName = NULL, char *inTitle = NULL);
									CFigureElement(const CFigureElement &inSource);
		virtual						~CFigureElement();		
		
		virtual GraphicType			getGraphicType() { return kFigureGraphicElement; }	
		
		virtual void				setName(char *inName = NULL);				
		virtual char*				getName() { return mName; }				
					
		virtual void				setTitle(char *inTitle = NULL);				
		virtual char*				getTitle() { return mTitle; }	
					
		virtual void				fitFigureInBounds();	
};

#endif /* _GRAPHICS_RECORD_H_ */
