/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CIntArray.cpp
 
		CIntArray Class
	 
 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of integers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"

#include "CIntArray.h"

#pragma mark === CIntArray ===

// ---------------------------------------------------------------------------------
//		CIntArray::CIntArray
// ---------------------------------------------------------------------------------

//: Constructs an empty CIntArray
//  The array will only be allocated when an integer is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
CIntArray::CIntArray(int inInitial, int inIncrement)
{
	capacity=0;
	numValues=0;
	values=NULL;
	iterator=0;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		CIntArray::~CIntArray
// ---------------------------------------------------------------------------------

//: The destructor
CIntArray::~CIntArray()
{
	if (values!=NULL)
		delete[] values;
}

// ---------------------------------------------------------------------------------
//		CIntArray::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void CIntArray::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		CIntArray::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void CIntArray::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		CIntArray::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void CIntArray::setIncrement(int inIncrement)
{
	increment=inIncrement;
}

// ---------------------------------------------------------------------------------
//		CIntArray::getValue
// ---------------------------------------------------------------------------------

//: Returns the integer that is at an index.
int CIntArray::getValue(int index)
{
	TestBounds(index, numValues);
	
	return values[index];
}

// ---------------------------------------------------------------------------------
//		CIntArray::getSum
// ---------------------------------------------------------------------------------

//: Returns the sum of all the integers
//  This function has no protection against overflow.
int CIntArray::getSum()
{
	int i;
	int sum = 0;
	for (i = 0; i < numValues; i++)
		sum += values[i];
		
	return sum;
}

// ---------------------------------------------------------------------------------
//		CIntArray::setObject
// ---------------------------------------------------------------------------------

//: Sets the integer at an index
//  If index>numValues then numValues becomes index+1 with memory 
//  allocated if needed.
void CIntArray::setValue(int index, int inValue)
{
	ThrowIf(index<0);

	if (index>=numValues) {
		numValues=index+1;
		checkCapacity(numValues);
	} 
		
	values[index]=inValue;
}

// ---------------------------------------------------------------------------------
//		CIntArray::add
// ---------------------------------------------------------------------------------

//: Adds an integer to the end of the list
//  This expands the list's capacity if necessary.
int CIntArray::add(int inValue)
{
	checkCapacity(numValues+1);

	values[numValues]=inValue;
	numValues++;
	
	return numValues-1;
}

// ---------------------------------------------------------------------------------
//		CIntArray::insert
// ---------------------------------------------------------------------------------

//: Inserts an integer into the list at a position
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new integer. If the integer's new position equals
//  numObjects then this adds it to the end of the list.
void CIntArray::insert(int index, int inValue)
{
	ThrowIf(index<0 || index>numValues+1);

	checkCapacity(numValues+1);
		
	for (int i=numValues; i>index; i--)
		values[i]=values[i-1];
	
	values[index]=inValue;
	numValues++;
}

// ---------------------------------------------------------------------------------
//		CIntArray::remove
// ---------------------------------------------------------------------------------

//: Removes the integer at index and moves others up 
//  This function returns the integer being removed.
int CIntArray::remove(int index)
{
	TestBounds(index, numValues);

	int tmp=values[index];

	for (int i=index; i<numValues-1; i++)
		values[i]=values[i+1];
		
	numValues--;
	
	return tmp;
}

// ---------------------------------------------------------------------------------
//		CIntArray::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the integers in the list
void CIntArray::removeAll()
{
	numValues=0;
}

// ---------------------------------------------------------------------------------
//		CIntArray::findValue
// ---------------------------------------------------------------------------------

//: Finds the index of an integer in the list
int CIntArray::findValue(int inValue)
{
	int i = 0;
	while (i < numValues && values[i]!=inValue) i++;
	
	if (i == numValues)
		return -1;
			
	return i;
}

// ---------------------------------------------------------------------------------
//		CIntArray::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the integer list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new integer array of this
//  size, copies all the old integers into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void CIntArray::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity>capacity) {
		int *newValues;
		int i;
		
		// Increase capacity
		while (capacity<requiredCapacity)
			capacity+=increment;
		
		// Allocate the new list.
		NewArray(newValues, int, capacity);
		
		if (values!=NULL) {
			// Copy the existing list into the new list.
			for (i=0; i<numValues; i++)
				newValues[i]=values[i];
	
			// Free the old list.
			delete[] values;
		}
		
		// Point the old list to the new one.
		values=newValues;
	}
}

