/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CList.cpp

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//:	Implements a dynamically resizing list of CObjects
//  CList is derived itself from CObject so it is possible to have lists of lists.
//  The list keeps track of the number of objects and its current capacity. When adding items
//  to the list, it checks to see if it has the capacity and, if not, reallocates
//  its array, incrementing the capacity by increment. The object pointers
//  are copied into the new array and the old array is deleted. This means that
//  the capacity of CList can increase but not decrease.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"

#include "CObject.h"
#include "CList.h"

#pragma mark === CList ===

// ---------------------------------------------------------------------------------
//		CList::CList
// ---------------------------------------------------------------------------------

//: Constructs an empty CList
//  The array will only be allocated when an object is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
CList::CList(int inInitial, int inIncrement) : CObject()
{
	mCapacity = 0;
	mNumObjects = 0;
	mObjects = NULL;
	mIterator = 0;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		CList::~CList
// ---------------------------------------------------------------------------------

//: The destructor
//  Delete the list of objects and call removeReference for each object
CList::~CList()
{
	removeAll();
	if (mObjects != NULL) {
		delete[] mObjects;
	}
}

// ---------------------------------------------------------------------------------
//		CList::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void CList::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		CList::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void CList::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		CList::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void CList::setIncrement(int inIncrement)
{
	mIncrement = inIncrement;
}

// ---------------------------------------------------------------------------------
//		CList::createObject
// ---------------------------------------------------------------------------------

//: Creates a new object of a type appropriate to the list
//  By default this just returns a new instance of CObject. By overriding
//  this, descendents (e.g. TreeList) can create instances appropriate
//  to the list (e.g. Tree).
CObject* CList::createObject()
{
	CObject *obj;
	NewClass(obj, CObject());
	return obj;
}

// ---------------------------------------------------------------------------------
//		CList::getObject
// ---------------------------------------------------------------------------------

//: Returns the CObject that is at an index.
CObject* CList::getObject(int index)
{
	TestBounds(index, mNumObjects);
	
	return mObjects[index];
}

// ---------------------------------------------------------------------------------
//		CList::setObject
// ---------------------------------------------------------------------------------

//: Sets the object at an index
//  If an object already exists then the reference to it is lost. 
//  If index>numObjects then numObjects becomes index+1 with memory 
//  allocated if needed.
void CList::setObject(int index, CObject *inObject)
{
	ThrowIf(index < 0);

	if (index >= mNumObjects) {
		mNumObjects = index + 1;
		checkCapacity(mNumObjects);
	} 
	
	if (mObjects[index] != NULL)
		mObjects[index]->release();

	if (inObject != NULL)
		inObject->retain();
	mObjects[index] = inObject;
}

// ---------------------------------------------------------------------------------
//		CList::add
// ---------------------------------------------------------------------------------

//: Adds an object to the end of the list
//  This expands the list's capacity if necessary.
int CList::add(CObject *inObject)
{
	checkCapacity(mNumObjects + 1);

	if (inObject != NULL)
		inObject->retain();
	mObjects[mNumObjects] = inObject;
	mNumObjects++;
	
	return mNumObjects - 1;
}

// ---------------------------------------------------------------------------------
//		CList::insert
// ---------------------------------------------------------------------------------

//: Inserts an object into the list at a position
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new object. If the object's new position equals
//  numObjects then this adds it to the end of the list.
void CList::insert(int index, CObject *inObject)
{
	ThrowIf(index < 0 || index > mNumObjects + 1);

	checkCapacity(mNumObjects + 1);
		
	for (int i = mNumObjects; i > index; i--)
		mObjects[i] = mObjects[i-1];
	
	if (inObject != NULL)
		inObject->retain();
	mObjects[index] = inObject;
	mNumObjects++;
}

// ---------------------------------------------------------------------------------
//		CList::swap
// ---------------------------------------------------------------------------------

//: Swaps an object at index1 with that at index2
//  Both objects must already be in the list.
void CList::swap(int index1, int index2)
{
	TestBounds(index1, mNumObjects);
	TestBounds(index2, mNumObjects);

	CObject *obj = mObjects[index1];
	mObjects[index1] = mObjects[index2];
	mObjects[index2] = obj;
}

// ---------------------------------------------------------------------------------
//		CList::remove
// ---------------------------------------------------------------------------------

//: Removes the object at index and moves others up 
//  This function calls release so if no other references to the object exit,
//  it will be deleted. If the object is required elsewhere it should be retained
//  before removal from the list.

void CList::remove(int index)
{
	TestBounds(index, mNumObjects);

	CObject* object = mObjects[index];

	for (int i = index; i < mNumObjects - 1; i++)
		mObjects[i] = mObjects[i+1];
		
	mNumObjects--;
	mObjects[mNumObjects] = NULL;

	object->release();
}

// ---------------------------------------------------------------------------------
//		CList::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the objects in the list
//  This function calls release for each so if no other references to the objects exit,
//  they will be deleted.

void CList::removeAll()
{
	for (int i = 0; i < mNumObjects; i++) {
		mObjects[i]->release();
		mObjects[i] = NULL;
	}
	mNumObjects = 0;
}

// ---------------------------------------------------------------------------------
//		CList::getFirst
// ---------------------------------------------------------------------------------

//: Gets the first object in the list
//  Sets the iterator variable to 0 and returns the first object.
CObject* CList::getFirst()
{
	mIterator = 0;
	return getNext();
}

// ---------------------------------------------------------------------------------
//		CList::getNext
// ---------------------------------------------------------------------------------

//: Gets the next object in the list
//  Increments the iterator variable and returns the object at that position.
CObject* CList::getNext()
{
	CObject *obj = NULL;
	if (mIterator < mNumObjects)
		obj = mObjects[mIterator];
	mIterator++;
	
	return obj;
}

// ---------------------------------------------------------------------------------
//		CList::compare
// ---------------------------------------------------------------------------------

//: Compares two objects for sorting
//  By default this is an alphabetical comparison of the names.
int CList::compare(CObject *inObject1, CObject *inObject2)
{
	char *name1 = inObject1->getName();
	char *name2 = inObject2->getName();
	
	if (name1 == NULL || name2 == NULL)
		return 0;
		
	return strcmp(name1, name2);
}

// ---------------------------------------------------------------------------------
//		CList::compare
// ---------------------------------------------------------------------------------

//: Sorts the objects in the list 
//  By default this is in ascending order unless inAscending is false.
void CList::sort(bool inAscending)
{
	if (mNumObjects > 1) {
	
			// Shellsort Algorithm
		
		int	hh = 1;						// Find starting "h" value
		int	stopH = mNumObjects / 9;
		while (hh < stopH) {
			hh = 3 * hh + 1;
		}
		
		CObject *tmp;
		
		int direction = (inAscending ? 1 : -1);
		
		for ( ; hh > 0; hh /= 3) {
				
			for (int step = hh + 1; step <= mNumObjects; step++) {
				tmp = getObject(step - 1);
				
				long i;
				for (i = step - hh; i > 0; i -= hh) {
					if ((compare(tmp, getObject(i-1)) * direction) >= 0) {
						break;
					}
					setObject(i + hh - 1, getObject(i-1));
				}
					
				setObject(i + hh - 1, tmp);
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		CList::getObjectName
// ---------------------------------------------------------------------------------

//: Returns the name string from the object at an index
char* CList::getObjectName(int index)
{
	TestBounds(index, mNumObjects);
		 
	ThrowIfNull(mObjects[index]);

	return mObjects[index]->getName();
}

// ---------------------------------------------------------------------------------
//		CList::findObject
// ---------------------------------------------------------------------------------

//: Searches the list for object referenced by pointer
//  Returns index of the object or -1 if not found.
int	CList::findObject(CObject *inObject)
{
	int i = 0;
	
	while (i < mNumObjects && mObjects[i] != inObject) i++;
	
	if (i == mNumObjects)
		return -1;
	else
		return i;
}
		
// ---------------------------------------------------------------------------------
//		CList::findByName
// ---------------------------------------------------------------------------------

//: Searches the objects for a name
//  Returns the index of the object or -1 if not found.
int	CList::findByName(char *inName, bool inIgnoreCase)
{
	int i = 0;
	
	if (inIgnoreCase) {
		bool found = false;
		char *P, *Q;
		
		while (i < mNumObjects && !found) {
			P = getObjectName(i);
			Q = inName;
			while (*P != '\0' && *Q != '\0' && toupper(*P) == toupper(*Q)) {
				P++;
				Q++;
			}
			if (*P == '\0' && *Q == '\0')
				found = true;
			else
				i++;
		}
	} else
		while (i < mNumObjects && strcmp(getObjectName(i), inName) != 0)
			i++;
	
	if (i == mNumObjects)
		return -1;
	else
		return i;
}
		
// ---------------------------------------------------------------------------------
//		CList::findByContents
// ---------------------------------------------------------------------------------

//: Searches the list for an item who's contents equals object
//  Returns index of the object or -1 if not found.
int	CList::findByContents(CObject *inObject)
{
	int i = 0;
	
	while (i < mNumObjects && compare(mObjects[i], inObject) != 0) i++;
	
	if (i == mNumObjects)
		return -1;
	else
		return i;
}
		
// ---------------------------------------------------------------------------------
//		CList::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the objects list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new CObject pointer list of this
//  size, copies all the old object pointers into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void CList::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity > mCapacity) {
		CObject **newObjects;
		int i;
		
		// Increase capacity
		while (mCapacity < requiredCapacity)
			mCapacity += mIncrement;
		
		// Allocate the new list.
		NewArray(newObjects, CObject*, mCapacity);
		
		if (mObjects != NULL) {
			// Copy the existing list into the new list.
			for (i = 0; i < mNumObjects; i++)
				newObjects[i] = mObjects[i];
	
			// Free the old list.
			delete[] mObjects;

			// Set the remaining pointers to NULL.
			for (i = mNumObjects; i < mCapacity; i++)
				newObjects[i] = NULL;
		} else {
			// Set the all the pointers to NULL.
			for (i = 0; i < mCapacity; i++)
				newObjects[i] = NULL;
		}
		
		// Point the old list to the new one.
		mObjects = newObjects;
	}
}

