/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CList.h

		Interface for the CList Class
					
 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global
 
 
#ifndef _CLIST_H_
#define _CLIST_H_

#include "CObject.h"

#define INITIAL_SIZE 100
#define INCREMENT_SIZE 100

class MEStream;

/******************************************************************************
	class CList
 ******************************************************************************/
#pragma mark CList class

//:	Implements a dynamically resizing list of CObjects
//  CList is derived itself from CObject so it is possible to have lists of lists.
//  The list keeps track of the number of objects and its current capacity. When adding items
//  to the list, it checks to see if it has the capacity and, if not, reallocates
//  its array, incrementing the capacity by increment. The object pointers
//  are copied into the new array and the old array is deleted. This means that
//  the capacity of CList can increase but not decrease.

class CList : public CObject {
	public:
		virtual char*		getClassName() const { return "List"; }
		//: Returns the name of this class - "List"
		
	protected:
		int 				mNumObjects; //: The number of objects in the list
		int					mCapacity;	//: The capacity of the list
		int					mIncrement;	//: The amount by which capacity increments
		int					mIterator;	//: Used to iterate through the list

		CObject 			**mObjects;	//: An array of CObject pointers

	public:
							CList(int inInitial = INCREMENT_SIZE, 
									int inIncrement = INCREMENT_SIZE);	//: The constructor
	 	virtual				~CList();	//: The destructor		
	 	
		virtual void		setInitial(int inInitial, int inIncrement = INCREMENT_SIZE);
		//: Set the initial capacity and (optionally) the increment

		int					getNumObjects() { return mNumObjects; }
		//: Returns the number of objects in the list
		
		virtual CObject**	getObjects() { return mObjects; }
		//: Returns the list of CObject pointers
		
		virtual void		setCapacity(int inCapacity);
		//: Set the capacity
		
		virtual int			getCapacity() { return mCapacity; }
		//: Returns the capacity

		virtual int			getIncrement() { return mIncrement; }
		//: Returns the increment

		virtual void		setIncrement(int inIncrement);
		//: Sets the amount that capacity increments by
		
		virtual CObject*	createObject();
		//: Creates a new object of a type appropriate to the list
		
		virtual CObject*	getObject(int index);
		//: Returns the CObject that is at an index
		
		virtual void		setObject(int index, CObject *inObject);
		//: Sets the object at an index

		virtual int			add(CObject *inObject);
		//: Adds an object to the end of the list
		
		virtual void		insert(int index, CObject *inObject);
		//: Inserts an object into the list at a position
		
		virtual void		swap(int index1, int index2);
		//: Swaps an object at index1 with that at index2
		
		virtual void		remove(int index);
		//: Removes the object at index and moves others up 
		
		virtual void		removeAll();
		//: Removes all the objects in the list

		virtual CObject*	getFirst();
		//: Gets the first object in the list

		virtual CObject*	getNext();
		//: Gets the next object in the list
		
		virtual int			compare(CObject *inObject1, CObject *inObject2);
		//: Compares two objects for sorting
		
		virtual void		sort(bool inAscending=true);
		//: Sorts the objects in the list 
		
		virtual char* 		getObjectName(int index);
		//: Returns the name string from the object at an index

		virtual int			findObject(CObject *inObject);		
		//: Searches the list for object referenced by pointer
		
		virtual int			findByName(char *inName, bool inIgnoreCase = false);
		//: Searches the objects for a name

		virtual int			findByContents(CObject *inObject);		
		//: Searches the list for an item whos contents equals object
				
		virtual CObject*	operator[](int idx) { return getObject(idx); }

	protected:
		virtual void		checkCapacity(int requiredCapacity);
		//: Checks that the objects list has enough space and resizes it as necessary
		
};

#endif /* _CLIST_H_ */
