/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CLogAxis.cpp
 
		CLogAxis Class
	 
 ******************************************************************************/

#include <math.h>

#include "CExceptions.h"
#include "CDoubleArray.h"
#include "CLogAxis.h"

#pragma mark === Construction & Destruction ===

// ---------------------------------------------------------------------------------
//		 CLogAxis()
// ---------------------------------------------------------------------------------

CLogAxis::CLogAxis( ) : CAxis( )
{
}


CLogAxis::CLogAxis(AxisFlag inMinAxisFlag, AxisFlag inMaxAxisFlag) : 
	CAxis(inMinAxisFlag, inMaxAxisFlag)
{
}


CLogAxis::CLogAxis(AxisFlag inMinAxisFlag, AxisFlag inMaxAxisFlag, bool inIsDescrete) :
	CAxis(inMinAxisFlag, inMaxAxisFlag, inIsDescrete)
{
}


#pragma mark === Accessing ===

// ---------------------------------------------------------------------------------
//		CLogAxis::SetRange()
// ---------------------------------------------------------------------------------

void CLogAxis::SetRange(CDoubleArray *inData)
{
	maxData=inData->getMaxValue();
	minData=inData->getMinValue();
	if (minData<=0.0) {
		int i, n=inData->getNumValues();
		double d;
		for (i=0; i<n; i++) {
			d=inData->getValue(i);
			if (d>0.0 && (d<minData || minData<=0.0))
				minData=d;
		}
		
		if (minData <= 0.0)
			throw /* some exception or another */;
	}
	
	isCalibrated=false;	
}

// ---------------------------------------------------------------------------------
//		CLogAxis::AddRange()
// ---------------------------------------------------------------------------------

void CLogAxis::AddRange(CDoubleArray *inData)
{
	if (inData->getMaxValue()>maxData)
		maxData=inData->getMaxValue();
		
	double min=inData->getMinValue();
	if (min<=0.0) {
		int i, n=inData->getNumValues();
		double d;
		for (i=0; i<n; i++) {
			d=inData->getValue(i);
			if (d>0.0 && (d<min || min<=0.0))
				min=d;
		}
		
		if (min<=0.0)
			min=minData;
	}
	
	if (min<minData)
		minData=min;
		
	isCalibrated=false;	
}

// ---------------------------------------------------------------------------------
//		CLogAxis::SetRange()
// ---------------------------------------------------------------------------------

void CLogAxis::SetRange(double inMinValue, double inMaxValue)
{
	if (inMinValue <= 0.0)
		throw /* some exception or another */;
	CAxis::SetRange(inMinValue, inMaxValue);
}

// ---------------------------------------------------------------------------------
//		CLogAxis::AddRange()
// ---------------------------------------------------------------------------------

void CLogAxis::AddRange(double inMinValue, double inMaxValue)
{
	if (inMinValue <= 0.0)
		throw /* some exception or another */;
	CAxis::AddRange(inMinValue, inMaxValue);
}

//-------------------------------------------------------------------------------------
// 		CLogAxis::Transform
//-------------------------------------------------------------------------------------

double CLogAxis::Transform(double inValue)
{
	if (inValue<=0.0)
		throw CIllegalLogValue();
		
	return log10(inValue);
}
	
//-------------------------------------------------------------------------------------
// 		CLogAxis::GetNumMinorTicks
//-------------------------------------------------------------------------------------

int CLogAxis::GetNumMinorTicks(int majorTickNo)
{
	if (!isCalibrated)
		Calibrate();
		
	if (majorTickNo==numMajorTicks-1)
		return (int)((maxAxis-maxTick)/maxTick);
	else if (majorTickNo==-1) {
		return (int)((minTick-minAxis)/(minTick/10));
	} else
		return numMinorTicks; 
}
	
//-------------------------------------------------------------------------------------
// 		CLogAxis::GetMajorTick
// 		
// 		Returns the value of the majorTickNo'th major tick.
//-------------------------------------------------------------------------------------

double CLogAxis::GetMajorTick(int majorTickNo)
{
	if (!isCalibrated)
		Calibrate();
		
	if (majorTickNo==numMajorTicks-1)
		return maxTick;
	else
		return pow(10.0, majorTickNo)*minTick;
}
	
//-------------------------------------------------------------------------------------
// 		CLogAxis::GetMinorTick
// 		
// 		Returns the value of the minorTickNo'th minor tick.
//-------------------------------------------------------------------------------------

double CLogAxis::GetMinorTick(int minorTickNo, int majorTickNo)
{
	if (!isCalibrated)
		Calibrate();
		
	if (majorTickNo==-1)
		return minTick-((minorTickNo+1)*(minTick/10));
	else
		return (minorTickNo+2)*GetMajorTick(majorTickNo);
		
	//	This last line is the equivalent of:
	//	return ((minorTickNo+1)*getMajorTick(majorTickNo))+getMajorTick(majorTickNo);
}

//-------------------------------------------------------------------------------------
// 		CLogAxis::SetMinTick
//-------------------------------------------------------------------------------------

void CLogAxis::SetMinTick()
{
	minTick=1;
	// Find the nearest multiple of majorTick below minData
	if (minData>1) { // work upwards
		while ((minTick*10)<minData)
			minTick*=10;
	} else if (minData<1) { // work downwards
		while ((minTick-minData) > (minData * epsilon))
			minTick/=10;
	}
}
	
//-------------------------------------------------------------------------------------
// 		CLogAxis::SetMaxTick
//-------------------------------------------------------------------------------------

void CLogAxis::SetMaxTick()
{
	maxTick=1;
	// Find the nearest multiple of majorTick above maxData
	if (maxData>1) { // work upwards
		while (maxTick<maxData)
			maxTick*=10;
	} else if (maxData<1) { // work downwards
		while ((maxTick/10)>maxData)
			maxTick/=10;
	}
}
	
//-------------------------------------------------------------------------------------
// 		CLogAxis::SetMajorTick
//-------------------------------------------------------------------------------------

void CLogAxis::SetMajorTick()
{
	majorTick=10;
	numMajorTicks=(int)(log10(maxTick/minTick))+1;
}

//-------------------------------------------------------------------------------------
// 		CLogAxis::SetMinorTick
//-------------------------------------------------------------------------------------

void CLogAxis::SetMinorTick()
{
		minorTick=1;
		numMinorTicks=8; // 1 and 10 are major ticks
}

//-------------------------------------------------------------------------------------
// 		CLogAxis::HandleAxisFlags
//-------------------------------------------------------------------------------------

void CLogAxis::HandleAxisFlags()
{
	if (minAxisFlag == kAtValue)
		throw /* some exception or another */;
	if (minAxisFlag == kAtZero)
		throw /* some exception or another */;
	if (maxAxisFlag == kAtValue)
		throw /* some exception or another */;
	if (maxAxisFlag == kAtZero)
		throw /* some exception or another */;
	
	if (minAxisFlag!=kAtMajorTick && minAxisFlag!=kAtMajorTickPlus) {
		if (minAxis+minTick<=minData) {
			while ((minAxis+minTick)<=minData) {
				minAxis+=minTick;
			}
			numMajorTicks--;
			minTick*=10;
		}
	}

	if (maxAxisFlag!=kAtMajorTick && maxAxisFlag!=kAtMajorTickPlus) {
		if (maxAxis-(maxTick/10)>=maxData) {
			numMajorTicks--;
			maxTick/=10;
			while ((maxAxis-maxTick)>=maxData) {
				maxAxis-=maxTick;
			}
		}
	}
}

