/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CPlot.cpp
 
		CPlot Class
	 
 ******************************************************************************/

#include <stdio.h>

#include "CPlot.h"

#include "CGraph.h"
#include "CAxis.h"

#include "CDataSeries.h"
#include "CDoubleArray.h"

#include "CGraphicsRecord.h"

#pragma mark === Construction & Destruction ===

// ---------------------------------------------------------------------------------
//		 CPlot()
// ---------------------------------------------------------------------------------

CPlot::CPlot( )
{
	mXAxis = mYAxis = NULL;
	mData = NULL;
	mGraph = NULL;

	mShowOrigin=false;

	mMarkGraphic = NULL;
	mLineGraphic = NULL;
}


// ---------------------------------------------------------------------------------
//		 ~CPlot()
// ---------------------------------------------------------------------------------

CPlot::~CPlot( )
{
	if (mData != NULL)
		delete mData;
		
	if (mMarkGraphic != NULL)
		delete mMarkGraphic;
	if (mLineGraphic != NULL)
		delete mLineGraphic;
}

#pragma mark === Accessing ===

//-------------------------------------------------------------------------------------
// 		CPlot::SetAxes
//-------------------------------------------------------------------------------------

void CPlot::SetAxes(CAxis *inXAxis, CAxis *inYAxis)
{
	mXAxis=inXAxis;
	mYAxis=inYAxis;
}


//-------------------------------------------------------------------------------------
// 		CPlot::SetGraph
//-------------------------------------------------------------------------------------

void CPlot::SetGraph(CGraph *inGraph)
{
	mGraph = inGraph;
}


//-------------------------------------------------------------------------------------
// 		CPlot::SetData
//-------------------------------------------------------------------------------------

void CPlot::SetData(CDataSeries *inData)
{
	mData=inData;
}


// ---------------------------------------------------------------------------------
//		 CPlotElement::SetLineStyle
// ---------------------------------------------------------------------------------

void CPlot::SetLineStyle(double inWidth, StrokeLineCapType inLineCap, StrokeLineJoinType inLineJoin,
							CColourRec &inColourRec)
{
	if (mLineGraphic == NULL)
		mLineGraphic = new CLineElement(0.0, 0.0, 0.0, 0.0);
	
	mLineGraphic->setStrokeStyle(inWidth, inLineCap, inLineJoin);
	mLineGraphic->setStrokeColour(inColourRec);
	mLineGraphic->setFillColour(kNoColour);
}

// ---------------------------------------------------------------------------------
//		 CPlot::SetMarkStyle
// ---------------------------------------------------------------------------------

void CPlot::SetMarkStyle(PlotMarkType inMarkType,
							double inMarkWidth, double inStrokeWidth, 
							CColourRec &inStrokeColour, CColourRec &inFillColour)
{
	if (mMarkGraphic != NULL)
		delete mMarkGraphic;
	
	if (inMarkType == kNoMarks) {
		mMarkGraphic = NULL;
		return;
	}
	
	double w = inMarkWidth / 2.0;
	CPathElement *path;
	
	switch (inMarkType) {
		case kCrossMarks:
			path = new CPathElement(0.0, 0.0);
			path->moveTo(-w, -w);
			path->lineTo(w, w);
			path->moveTo(w, -w);
			path->lineTo(-w, w);
			path->setStrokeStyle(inStrokeWidth);
			path->setStrokeColour(inStrokeColour);
			mMarkGraphic = path;
		break;
		case kPlusMarks:
			path = new CPathElement(0.0, 0.0);
			path->moveTo(-w, 0);
			path->lineTo(w, 0);
			path->moveTo(0, -w);
			path->lineTo(0, w);
			path->setStrokeStyle(inStrokeWidth);
			path->setStrokeColour(inStrokeColour);
			mMarkGraphic = path;
		break;
		case kCircleMarks:
			mMarkGraphic = new CCircleElement(0.0, 0.0, w);
			mMarkGraphic->setStrokeStyle(inStrokeWidth);
			mMarkGraphic->setStrokeColour(inStrokeColour);
			mMarkGraphic->setFillColour(inFillColour);
		break;
		case kSquareMarks:
			mMarkGraphic = new CRectElement(-w, -w, inMarkWidth, inMarkWidth);
			mMarkGraphic->setStrokeStyle(inStrokeWidth);
			mMarkGraphic->setStrokeColour(inStrokeColour);
			mMarkGraphic->setFillColour(inFillColour);
		break;
		case kDiamondMarks:
			path = new CPathElement(0.0, 0.0);
			path->moveTo(0, -w);
			path->lineTo(w, 0);
			path->lineTo(0, w);
			path->lineTo(-w, 0);
			path->setPathClosed(true);
			path->setStrokeStyle(inStrokeWidth);
			path->setStrokeColour(inStrokeColour);
			path->setFillColour(inFillColour);
			mMarkGraphic = path;
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 CPlot::SetMarkGraphic
// ---------------------------------------------------------------------------------

void CPlot::SetMarkGraphic(CGraphicElement* inMarkGraphic)
{
	if (mMarkGraphic != NULL)
		delete mMarkGraphic;
	
	mMarkGraphic = inMarkGraphic;
}

#pragma mark === Drawing ===

//-------------------------------------------------------------------------------------
// 		CPlot::DrawMark
//-------------------------------------------------------------------------------------

void CPlot::DrawMark(CComplexGraphicElement *ioGraphic, CGraphicElement *inMarkGraphic,
						double x, double y)
{
	CGraphicElement *point;

	if (inMarkGraphic == NULL)
		return;
		
	switch (inMarkGraphic->getGraphicType()) {
		case kRectGraphicElement:
			point = new CRectElement(*dynamic_cast<CRectElement*>(inMarkGraphic));
		break;
		case kRoundedRectGraphicElement:
			point = new CRoundedRectElement(*dynamic_cast<CRoundedRectElement*>(inMarkGraphic));
		break;
		case kCircleGraphicElement:
			point = new CCircleElement(*dynamic_cast<CCircleElement*>(inMarkGraphic));
		break;
		case kEllipseGraphicElement:
			point = new CEllipseElement(*dynamic_cast<CEllipseElement*>(inMarkGraphic));
		break;
		case kPathGraphicElement:
			point = new CPathElement(*dynamic_cast<CPathElement*>(inMarkGraphic));
		break;
		default:
			point = new CCircleElement(0.0, 0.0, 5.0);
		break;
	}
	
	point->setPosition(	mGraph->ConvertX(x), 
						mGraph->ConvertY(y));
						
	ioGraphic->add(point);
}

//-------------------------------------------------------------------------------------
// 		CPlot::DrawLine
//-------------------------------------------------------------------------------------

void CPlot::DrawLine(CComplexGraphicElement *ioGraphic, CGraphicElement *inLineStyle,
						double x1, double y1, double x2, double y2) 
{	
	try {
		CLineElement *line = new CLineElement(mGraph->ConvertX(x1), mGraph->ConvertY(y1),
												mGraph->ConvertX(x2), mGraph->ConvertY(y2));
								
		line->setStyle(*inLineStyle);
		ioGraphic->add(line);
	} catch (...) {
	}			
}

//-------------------------------------------------------------------------------------
// 		CPlot::DrawRect
//-------------------------------------------------------------------------------------

void CPlot::DrawRect(CComplexGraphicElement *ioGraphic, CGraphicElement *inRectStyle,
						double x1, double y1, double x2, double y2) 
{
	try {
		CRectElement *rect = new CRectElement(mGraph->ConvertX(x1), mGraph->ConvertY(y1),
												mGraph->ConvertX(x2), mGraph->ConvertY(y2));
								
		rect->setStyle(*inRectStyle);
		ioGraphic->add(rect);
	} catch (...) {
	}			
}


//-------------------------------------------------------------------------------------
// 		CPlot::DrawPlot
//-------------------------------------------------------------------------------------

void CPlot::DrawPlot(CComplexGraphicElement *ioGraphic, 
						double inXTolerance, 
						double inYTolerance)
{
	if (mGraph == NULL || mData== NULL || mXAxis == NULL || mYAxis == NULL)
		return;
		
	CGroupElement *group = new CGroupElement();
	DrawDataSeries(group, mData, inXTolerance, inYTolerance);
	ioGraphic->add(group);
		
	if (mShowOrigin) {
		double minX=mXAxis->GetMinAxis();
		double maxX=mXAxis->GetMaxAxis();
		double minY=mYAxis->GetMinAxis();
		double maxY=mYAxis->GetMaxAxis();
		
		CGroupElement *group = new CGroupElement();
		CLineElement *line;
		
		if (minX<0.0 && maxX>0.0) {
			line = new CLineElement(0.0, minY, 0.0, maxY);
			line->setStrokeStyle(0.5);
			line->setStrokeColour(kBlackColour);
			group->add(line);
		}
					
		if (minY<0.0 && maxY>0.0) {
			line = new CLineElement(minX, 0.0, maxX, 0.0);
			line->setStrokeStyle(0.5);
			line->setStrokeColour(kBlackColour);
			group->add(line);
		}
		ioGraphic->add(group);
	}
}


