/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CReader.cpp

				CReader Class

 ******************************************************************************/

// CReader provides a common base to all classes reading in sequences or
// trees from standard file formats. It implements token reading and comment 
// skipping (and much else). Also reads Newick format trees. It is an abstract 
// class and must be overridden to implement the low-level reading.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"

#include "CReader.h"
#include "CWriter.h"

#pragma mark === CReader ===

// ---------------------------------------------------------------------------------
//		 CReader
// ---------------------------------------------------------------------------------

CReader::CReader()
{
	initReader();
}

// ---------------------------------------------------------------------------------
//		 CReader::initReader
// ---------------------------------------------------------------------------------

void CReader::initReader()
{
	startComment='\0';
	stopComment='\0';
	lineComment='\0';
	writeComment='\0';
	
	mLineNumber = 1;
	
	commentWriter = NULL;
}

// ---------------------------------------------------------------------------------
//		 ~CReader
// ---------------------------------------------------------------------------------

CReader::~CReader()
{
}

// ---------------------------------------------------------------------------------
//		 CReader::setCommentDelimiters
// ---------------------------------------------------------------------------------
 
// Sets the strings that delimit a multi-line comment, plus the string that
// proceeds a line comment. A zero will prevent comments.

void CReader::setCommentDelimiters(char start, char stop, char line, char write) 
{
	startComment=start;
	stopComment=stop;
	lineComment=line;
	writeComment=write;
}


#pragma mark === Higher Level Reading ===

// ---------------------------------------------------------------------------------
//		 CReader::skipComments
// ---------------------------------------------------------------------------------
 
// Skips over any comments. The first opening delimiter will have been read.

void CReader::skipComments()
{
	char ch;
	int n=1;
	bool write = false;
	
	do {
		ch=readChar();
		if (ch==startComment)
			n++;
		else if (ch==stopComment) {
			if (write && commentWriter != NULL)
				commentWriter->writeChar('\r');
			n--;
		} else if (write && commentWriter != NULL)
			commentWriter->writeChar(ch);
		else if (ch==writeComment)
			write = true;
	} while (n>0);
}


// ---------------------------------------------------------------------------------
//		 CReader::skipToEndOfLine
// ---------------------------------------------------------------------------------
 
// Skips to the end of the line. If a comment is found then this is read.

void CReader::skipToEndOfLine()
{
	char ch;
	
	do {
		ch=readChar();
	} while (ch!='\n' && ch!='\r' && ch!=startComment);
	
	skipWhile("\n\r");
	
	if (ch==startComment)
		skipComments();
}


// ---------------------------------------------------------------------------------
//		 CReader::skipSpace
// ---------------------------------------------------------------------------------
 
// Skips over any space (plus tabs and returns) in the file. Will also skip
// comments. Returns number of characters skipped.

int CReader::skipSpace()
{
	char ch;
	int i=-1;
	
	do {
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
	} while (isspace(ch));
	
	unreadChar(ch);
	
	return i;
}


// ---------------------------------------------------------------------------------
//		 CReader::skipSpace
// ---------------------------------------------------------------------------------

// Skips over any space or any of the characters in delimiters.  Returns number 
// of characters skipped.

int CReader::skipSpace(char *delimiters)
{
	char ch;
	int i=-1;
	
	do {
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
	} while (isspace(ch) || strchr(delimiters, ch)!=NULL);
	
	unreadChar(ch);

	return i;
}


// ---------------------------------------------------------------------------------
//		 CReader::skipSpace
// ---------------------------------------------------------------------------------

// Skips over any space or any of the characters in delimiters.  Returns the 
// first delimiter character found. The boolean, 'returnDelimiter' is a flag 
// for this overloaded function (i.e. adding the extra parameter uses this
// version and not the other).

char CReader::skipSpace(char *delimiters, bool /* returnDelimiter */)
{
	char ch, term;
	bool isDelim;
	int i=-1;
	
	term='\0';
	do {
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
		isDelim=(strchr(delimiters, ch)!=NULL);
		if (isDelim)
			term=ch;
	} while (isspace(ch) || isDelim);
	
	unreadChar(ch);

	return term;
}


/// ---------------------------------------------------------------------------------
//		 CReader::skipSpace
// ---------------------------------------------------------------------------------

// Skips over any space or any of the characters in delimiters1 or 2.  Returns 
// number of characters skipped.

int CReader::skipSpace(char *delimiters1, char *delimiters2)
{
	char ch;
	int i=-1;
	
	do {
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
	} while (isspace(ch) || strchr(delimiters1, ch)!=NULL || strchr(delimiters2, ch)!=NULL);
	
	unreadChar(ch);

	return i;
}


// ---------------------------------------------------------------------------------
//		 CReader::skipSpace
// ---------------------------------------------------------------------------------

// Skips over any space or any of the characters in delimiters1 or 2.  Returns 
// the first delimiter character found. The boolean, 'returnDelimiter' is a 
// flag for this overloaded function.

char CReader::skipSpace(char *delimiters1, char *delimiters2, bool /*returnDelimiter*/)
{
	char ch, term;
	bool isDelim;
	int i=-1;
	
	term='\0';
	do {
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
		isDelim=(strchr(delimiters1, ch)!=NULL || strchr(delimiters2, ch)!=NULL);
		if (isDelim)
			term=ch;
	} while (isspace(ch) || isDelim);
	
	unreadChar(ch);

	return ch;
}


// ---------------------------------------------------------------------------------
//		 CReader::skipUntil
// ---------------------------------------------------------------------------------
 
// Skips over the file until thr character delimiter is found. Returns the  
// delimiter.

char CReader::skipUntil(char delimiter)
{
	char delims[2];
	
	delims[0]=delimiter;
	delims[1]='\0';
	
	return skipUntil(delims);
}


// ---------------------------------------------------------------------------------
//		 CReader::skipUntil
// ---------------------------------------------------------------------------------

// Skips over the file until a character from delimiters is found. Returns 
// the delimiter found.

char CReader::skipUntil(char *delimiters)
{
	char ch;
	int i=-1;
	
	do {
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
	} while (strchr(delimiters, ch)==NULL);
	
	return ch;
}


// ---------------------------------------------------------------------------------
//		 CReader::skipWhile
// ---------------------------------------------------------------------------------

// Skips over the file while characters from delimiters are found. Returns last
// delimiter found.

char CReader::skipWhile(char *delimiters)
{
	char ch='\0', last;
	int i=-1;
	
	do {
		last=ch;
		ch=readChar();
		if (ch==startComment) {
			skipComments();
			ch=readChar();
		} else if (ch==lineComment) {
			skipToEndOfLine();
			ch=readChar();
		}
		i++;
	} while (strchr(delimiters, ch)!=NULL);
	
	unreadChar(ch);
	
	return last;
}


// ---------------------------------------------------------------------------------
//		 CReader::readLine
// ---------------------------------------------------------------------------------
 
// Reads a line into 'line' up to a maximum of maxLen characters. Reads a line up 
// to a EOL character. If the line begins with a quote marks (double or single) 
// then the line is read until the matching quote is found. Within the quotes, 
// multiple whitespace is compressed into a single space, including EOLs. If the 
// line is quoted then we skip to the next EOL.

int CReader::readLine(char *line, int maxLen, bool upper, bool skipBlank)
{
	int space=0, n=0;
	char *P, ch='\0', quoteChar='\0', eolChar='\0';
	bool done=false, quoted=false;
	
	P=line;
	try {
		while (!done) {
			ch=readChar();
			
			if (quoted && ch==quoteChar) { // Found the closing quote
				done=true;
				quoted=false;
			} else if (n==0 && (ch=='\'' || ch=='"')) { 	// if the opening character is a quote
				quoted=true;								// read everything upto the closing quote
				quoteChar=ch;
				space=0;
			} else {
				if (quoted) { // compress multiple spaces into one
					if (isspace(ch)) {
						space++;
						ch=' ';
					} else
						space=0;
					
					if (space<2 && n<maxLen) {
						if (upper)
							*P=toupper(ch);
						else
							*P=ch;
						P++;
						n++;
					}
				} else if (!done && ch!='\n' && ch!='\r') {
					if (n<maxLen) {
						if (upper)
							*P=toupper(ch);
						else
							*P=ch;
						P++;
						n++;
					}
				} else {
					eolChar = ch;
					done=true;
				}
			}
		}
		
		if (quoteChar!='\0')
			eolChar = skipUntil("\n\r");
			
		if (skipBlank) 
			skipWhile("\n\r");
		else {
			// just check that this is not a DOS text file...
			ch = readChar();
			if (eolChar == '\n' && ch != '\r' || 
				eolChar == '\r' && ch != '\n')
				unreadChar(ch);
		}
	} catch (CEndOfFile) { }
	*P='\0';
	
	return n;
}


// ---------------------------------------------------------------------------------
//		 CReader::readInteger
// ---------------------------------------------------------------------------------

int CReader::readInteger()
{
	char str[256];
	int d;
	
	readToken(str, 255);
	if (sscanf(str, "%d", &d)!=1)
		throw CIllegalInt();
		
	return d;
}


// ---------------------------------------------------------------------------------
//		 CReader::readInteger
// ---------------------------------------------------------------------------------

char CReader::readInteger(int *d, char *delimiters)
{
	char str[256], ch;
	
	ch=readToken(str, 255, delimiters);
	if (sscanf(str, "%d", d)!=1)
		throw CIllegalInt();
		
	return ch;
}


// ---------------------------------------------------------------------------------
//		 CReader::readDouble
// ---------------------------------------------------------------------------------

double CReader::readDouble()
{
	char str[256];
	double f;
	
	readToken(str, 255);
	if (sscanf(str, "%lf", &f)!=1)
		throw CIllegalDbl();
		
	return f;
}


// ---------------------------------------------------------------------------------
//		 CReader::readDouble
// ---------------------------------------------------------------------------------

char CReader::readDouble(double *f, char *delimiters)
{
	char str[256], ch;
	
	ch=readToken(str, 255, delimiters);
	if (sscanf(str, "%lf", f)!=1)
		throw CIllegalDbl();
		
	return ch;
}


// ---------------------------------------------------------------------------------
//		 CReader::readToken
// ---------------------------------------------------------------------------------
 
// Reads a token into 'token' up to a maximum of maxLen characters. Will stop 
// when any space is found. Will then skip any further space.If upper is true, 
// then the token will be forced to upper case. A comment will also delimit the 
// token. Returns character that delimited the token.

char CReader::readToken(char *token, int maxLen, bool upper, bool skipDelimiters)
{
	return readToken(token, maxLen, "", upper, skipDelimiters);
}


// ---------------------------------------------------------------------------------
//		 CReader::readToken
// ---------------------------------------------------------------------------------
 
// Reads a token into 'token' up to a maximum of maxLen characters. Will stop 
// when any character in delimiters is found. If upper is true, then the token 
// will be forced to upper case. A comment will also delimit the token. Returns
// the character that delimited the token. If the token begins with a quote char
// then all characters will be included in token until a matching quote is found.

char CReader::readToken(char *token, int maxLen, char *delimiters, bool upper, bool skipDelimiters)
{
	int n=0, space=0;
	char ch, ch2, *P, quoteChar='\0';
	bool done=false, quoted=false;
	
	skipSpace();
	
	P=token;	
	while (!done) {
		try {
			ch=readChar();
		
			if (quoted && ch==quoteChar) { // Found the closing quote
				ch2=readChar();
					
				if (ch==ch2) {	// A repeated quote character so add this to the token
					if (n<maxLen) {
						*P=ch;
						P++;
						n++;
					}
				} else {		// otherwise it terminates the token
					unreadChar(ch2);
					done=true;
					quoted=false;
					ch=' '; // pretend the token was terminated with white space
				}
			} else if (n==0 && (ch=='\'' || ch=='"')) { 	// if the opening character is a quote
				quoted=true;						// read everything upto the closing quote
				quoteChar=ch;
				space=0;
			} else {
				if (quoted) { // compress multiple spaces into one
					if (isspace(ch)) {
						space++;
						ch=' ';
					} else
						space=0;
					
					// Even if we run out of space
					// we must read until the close quote
					if (space<2 && n<maxLen) { 
						if (upper)
							*P=toupper(ch);
						else
							*P=ch;
						P++;
						n++;
					}
				} else if (!done && !isspace(ch) && strchr(delimiters, ch)==NULL && ch!=startComment && ch!=lineComment) {
					// Even if we run out of space
					// we must read until a delimiter is found
					if (n<maxLen) {
						if (upper)
							*P=toupper(ch);
						else
							*P=ch;
						P++;
						n++;
					}
				} else
					done=true;
			}
		} catch (CEndOfFile e) {
			done=true;
		}
	}
	*P='\0';
	
	if (isspace(ch) && strlen(delimiters) > 0) {
		try {
			skipSpace();
			ch2=readChar();
			if (strchr(delimiters, ch2)!=NULL) 
				ch = ch2;
			else
				unreadChar(ch2);
		} catch (CEndOfFile e) {
		}
	}
	
	if (skipDelimiters) {
		try {
			ch2 = ch;
			while (strchr(delimiters, ch2)!=NULL || isspace(ch2))
				ch2=readChar();
				
			unreadChar(ch2);
		} catch (CEndOfFile e) {
		}
	}
		
	return ch;
}


// ---------------------------------------------------------------------------------
//		 CReader::readString
// ---------------------------------------------------------------------------------
 
// Reads a string into 'str' up to a maximum of maxLen characters. Will stop 
// when any character in delimiters is found. If upper is true, then the string 
// will be forced to upper case. A comment will also delimit the string. Returns
// the character that delimited the string. If the string begins with a quote char
// then all characters will be included in string until a matching quote is found.

char CReader::readString(char *token, int maxLen, char *delimiters, bool upper, bool ignoreQuotes)
{
	int n=0;
	char ch, ch2, *P, quoteChar='\0';
	bool done=false, quoted=false;
	
	skipSpace();
	
	P=token;	
	while (!done) {
		ch=readChar();
		
		if (quoted && ch==quoteChar) { // Found the closing quote
			ch2=readChar();
				
			if (ch==ch2) {	// A repeated quote character so add this to the token
				if (n<maxLen) {
					*P=ch;
					P++;
					n++;
				}
			} else {		// otherwise it terminates the token
				unreadChar(ch2);
				done=true;
				quoted=false;
			}
		} else if (!ignoreQuotes && n==0 && (ch=='\'' || ch=='"')) { 	// if the opening character is a quote
			quoted=true;						// read everything upto the closing quote
			quoteChar=ch;
		} else {
			if (quoted) { // compress multiple spaces into one
				if (ch=='\'' || ch=='"') {
					ch2=readChar();
						
					if (ch==ch2) {	// A repeated quote character so add this to the token
						if (n<maxLen) {
							*P=ch;
							P++;
							n++;
						}
					} else
						unreadChar(ch2);
				}
				
				// Even if we run out of space
				// we must read until the close quote
				if (n<maxLen) { 
					if (upper)
						*P=toupper(ch);
					else
						*P=ch;
					P++;
					n++;
				}
			} else if (!done && strchr(delimiters, ch)==NULL && ch!=startComment && ch!=lineComment) {
				// Even if we run out of space
				// we must read until a delimiter is found
				if (n<maxLen) {
					if (upper)
						*P=toupper(ch);
					else
						*P=ch;
					P++;
					n++;
				}
			} else
				done=true;
		}
	}
	*P='\0';
	if (ch==quoteChar) {
		do {
			ch=readChar();
		} while (isspace(ch));
	}	
		
	return ch;
}


// ---------------------------------------------------------------------------------
//		 CReader::findToken
// ---------------------------------------------------------------------------------
 
// Will find a token or throw an end of file. 

void CReader::findToken(char *token, bool upper)
{
	char st[256];

	do {
		readToken(st, 255, upper);
	} while (strcmp(st, token)!=0);
}


// ---------------------------------------------------------------------------------
//		 CReader::findToken
// ---------------------------------------------------------------------------------
 
// Will find a token or throw an end of file. 

void CReader::findToken(char *token, char *delimiters, bool upper)
{
	char st[256];

	do {
		readToken(st, 255, delimiters, upper);
	} while (strcmp(st, token)!=0);
}


// ---------------------------------------------------------------------------------
//		 CReader::readTaxonLabel
// ---------------------------------------------------------------------------------
 
// Reads a taxon name into taxon, upto a maximum of maxLen characters. Will skip 
// any space, characters in 'ignore' or comments. If the label is completely numeric 
// and a positive integer less than maxIndex, then this will be returned.

int CReader::readTaxonLabel(char *token, int maxLen, int maxIndex)
{
	int i, n=-1;
	bool isNumeric=true;
	char *P;
	
	readToken(token, maxLen);
	P=token;
	while (*P!='\0' && isNumeric) {
		isNumeric=isdigit(*P);
		P++;
	}
	
	if (isNumeric && sscanf(token, "%d", &i)==1 && i>0 && i<=maxIndex) {
		n=i-1;
	} 
	
	return n;
}


