/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CReportRecord.h

		Interface for the CReportRecord Class

 ******************************************************************************/

#include "CObject.h"
#include "CList.h"

#include <stdio.h>

#ifndef _REPORT_RECORD_H_
#define _REPORT_RECORD_H_

class CString;

enum ElementType {
	kReport,
	kTextElement,
	kLineBreakElement,
	kIntegerElement,
	kRealElement,
	kTokenElement,
	kComplexElement,
	kParagraphElement,
	kPreformattedElement,
	kTableElement,
	kListElement,
	kGraphicElement
};

enum HeadingLevel {
	kNormalText,
	kHeadingLevel1,
	kHeadingLevel2,
	kHeadingLevel3,
	kHeadingLevel4,
	kHeadingLevel5,
	kHeadingLevel6
};

enum ElementAlignment {
	kAlignElementDefault,
	kAlignElementLeft,
	kAlignElementCentre,
	kAlignElementRight,
	kAlignElementJustify
};

enum ListType {
	kBullettedList,
	kNumberedList,
	kDefinitionList
};

// ---------------------------------------------------------------------------------
//		class CReportElement
// ---------------------------------------------------------------------------------

#pragma mark CReportElement
class CReportElement : public CObject {
	protected:
		CString					*mLink;
		
	public:
								CReportElement();
		virtual					~CReportElement();
						
		virtual ElementType		getElementType() = 0;				
		virtual bool			isComplexElement() { return false; }				
		
		virtual	void			setLink(const char *inLink);
		virtual	char*			getLink();				
};

// ---------------------------------------------------------------------------------
//		class CStyledElement
// ---------------------------------------------------------------------------------

#pragma mark CStyledElement
class CStyledElement : public CReportElement  {
	protected:
		bool					mIsItalic;
		bool					mIsBold;
	
	public:
								CStyledElement();
		virtual					~CStyledElement();
						
		virtual void			setStyle(bool inIsItalic, bool inIsBold); 
		
		bool					isItalic() { return mIsItalic; }
		bool					isBold() { return mIsBold; }
};

// ---------------------------------------------------------------------------------
//		class CSimpleElement
// ---------------------------------------------------------------------------------

#pragma mark CSimpleElement
class CSimpleElement : public CStyledElement {
	public:
								CSimpleElement(bool inIsItalic = false, bool inIsBold = false);
		virtual					~CSimpleElement();
						
};

// ---------------------------------------------------------------------------------
//		class CTextElement
// ---------------------------------------------------------------------------------

#pragma mark CTextElement
class CTextElement : public CSimpleElement {
	protected:
		CString					*mText;
		
	public:
								CTextElement(const char *inText, bool inIsItalic = false, bool inIsBold = false);
								CTextElement(const char *inText, const char *inLink, bool inIsItalic = false, bool inIsBold = false);
		virtual					~CTextElement();		
						
		virtual	void			setText(const char *inText);
		virtual	char*			getText();				
						
		virtual ElementType		getElementType() { return kTextElement; }				
};

// ---------------------------------------------------------------------------------
//		class CTextElement
// ---------------------------------------------------------------------------------

#pragma mark CLineBreakElement
class CLineBreakElement : public CSimpleElement {		
	public:
								CLineBreakElement() : CSimpleElement() { }
						
		virtual ElementType		getElementType() { return kLineBreakElement; }				
};

// ---------------------------------------------------------------------------------
//		class CIntegerElement
// ---------------------------------------------------------------------------------

#pragma mark CIntegerElement
class CIntegerElement : public CSimpleElement {
	protected:
		int						mValue;
		
	public:
								CIntegerElement(int inValue, bool inIsItalic = false, bool inIsBold = false);
								CIntegerElement(int inValue, const char *inLink, bool inIsItalic = false, bool inIsBold = false);
								CIntegerElement(long inValue, bool inIsItalic = false, bool inIsBold = false);
								CIntegerElement(long inValue, const char *inLink, bool inIsItalic = false, bool inIsBold = false);
		virtual					~CIntegerElement();		

		virtual	int				getValue() { return mValue; }				

		virtual ElementType		getElementType() { return kIntegerElement; }				
};

// ---------------------------------------------------------------------------------
//		class CRealElement
// ---------------------------------------------------------------------------------

#pragma mark CRealElement
class CRealElement : public CSimpleElement {
	protected:
		double					mValue;
		int						mPrecision;
		bool					mSciFormat;
		
	public:
								CRealElement(double inValue, bool inIsItalic = false, bool inIsBold = false);
								CRealElement(double inValue, int inPrecision, bool inIsItalic = false, bool inIsBold = false);
								CRealElement(double inValue, int inPrecision, bool inSciFormat, bool inIsItalic = false, bool inIsBold = false);

								CRealElement(double inValue, const char *inLink, bool inIsItalic = false, bool inIsBold = false);
								CRealElement(double inValue, int inPrecision, const char *inLink, bool inIsItalic = false, bool inIsBold = false);
								CRealElement(double inValue, int inPrecision, bool inSciFormat, const char *inLink, bool inIsItalic = false, bool inIsBold = false);

		virtual					~CRealElement();		

		virtual	double			getValue() { return mValue; }
		
		virtual	void 			setPrecision(int inPrecision, bool inSciFormat = false) 
									{ mPrecision = inPrecision; mSciFormat = inSciFormat; }				
		virtual	int 			getPrecision()  { return mPrecision; }
		virtual	bool 			getIsSciFormat()  { return mSciFormat; }

		virtual ElementType		getElementType() { return kRealElement; }				
};

// ---------------------------------------------------------------------------------
//		class CComplexElement
// ---------------------------------------------------------------------------------

#pragma mark CComplexElement
class CComplexElement : public CStyledElement {
	protected:
		CList						*mElements;
		int							mNumElements;
		int							mCapacity;
				
		int							mIndentLevel;
		ElementAlignment			mAlignment;	
		
		CString						*mAnchor;
		
	public:
									CComplexElement(ElementAlignment inAlignment = kAlignElementDefault,
													int inIndentLevel = 0);
		virtual						~CComplexElement();
					
		virtual ElementType			getElementType() { return kComplexElement; }				

		void						setAlignment(ElementAlignment inAlignment) { mAlignment = inAlignment; }	
		ElementAlignment			getAlignment() { return mAlignment; }	
					
		void						setIndentLevel(int inIndentLevel) { mIndentLevel = inIndentLevel; }				
		int							getIndentLevel() { return mIndentLevel; }				

		
		virtual	void				setAnchor(const char *inAnchor);
		virtual	char*				getAnchor();
					
	protected:
		virtual void				addElement(CReportElement *inElement);
		
	public:
		virtual void				add(CReportElement *inElement);
		virtual void				add(char *inText);
		virtual void				add(int inValue);
		virtual void				add(double inValue);
		
		virtual int					getNumElements() { return mElements->getNumObjects(); }
		virtual CReportElement*		getElement(int index);
		
		virtual bool				isComplexElement() { return true; }
};

// ---------------------------------------------------------------------------------
//		class CTokenElement
// ---------------------------------------------------------------------------------

#pragma mark CTokenElement
class CTokenElement : public CComplexElement {
	public:
								CTokenElement(bool inIsItalic = false, bool inIsBold = false);
		virtual					~CTokenElement();		

		virtual void			add(CSimpleElement *inElement);
		virtual void			add(char *inText);
		virtual void			add(int inValue);
		virtual void			add(double inValue);

		virtual ElementType		getElementType() { return kTokenElement; }				
};

// ---------------------------------------------------------------------------------
//		class CParagraphElement
// ---------------------------------------------------------------------------------

#pragma mark CParagraphElement
class CParagraphElement : public CComplexElement {
	protected:
		HeadingLevel			mHeadingLevel;
	
	public:
								CParagraphElement(	HeadingLevel inHeadingLevel = kNormalText, 
													ElementAlignment inAlignment = kAlignElementDefault,
													int inIndentLevel = 0);
								CParagraphElement(	char *inText,
													HeadingLevel inHeadingLevel = kNormalText, 
												 	ElementAlignment inAlignment = kAlignElementDefault,
													int inIndentLevel = 0);
								CParagraphElement(	int inValue,
													HeadingLevel inHeadingLevel = kNormalText, 
												 	ElementAlignment inAlignment = kAlignElementDefault,
													int inIndentLevel = 0);
								CParagraphElement(	double inValue,
													HeadingLevel inHeadingLevel = kNormalText, 
												 	ElementAlignment inAlignment = kAlignElementDefault,
													int inIndentLevel = 0);
		virtual					~CParagraphElement();		

		virtual void			add(CReportElement *inElement);
		virtual void			add(char *inText);
		virtual void			add(int inValue);
		virtual void			add(double inValue);
		virtual void			addLineBreak();

		virtual ElementType		getElementType() { return kParagraphElement; }				

		HeadingLevel			getHeadingLevel() { return mHeadingLevel; }				
};

// ---------------------------------------------------------------------------------
//		class CPreformattedElement
// ---------------------------------------------------------------------------------

#pragma mark CPreformattedElement
class CPreformattedElement : public CComplexElement {
	public:
								CPreformattedElement(int inIndentLevel = 0);
								CPreformattedElement(char *inText,
													int inIndentLevel = 0);
		virtual					~CPreformattedElement();		

		virtual void			addLineBreak();

		virtual ElementType		getElementType() { return kPreformattedElement; }				
};

// ---------------------------------------------------------------------------------
//		class CTableElement
// ---------------------------------------------------------------------------------

#pragma mark CTableElement
class CTableElement : public CComplexElement {
	protected:
		int						mNumRows, mNumCols;
		int						mNumTitleRows, mNumTitleCols;
		
	public:
								CTableElement(int inNumRows, int inNumCols, 
												int inNumTitleRows = 0, int inNumTitleCols = 0,
												ElementAlignment inAlignment = kAlignElementDefault,
												int inIndentLevel = 0);
		virtual					~CTableElement();		

		virtual ElementType		getElementType() { return kTableElement; }				
						
		virtual void			add(CReportElement *inElement);
		virtual void			add(char *inText);
		virtual void			add(int inValue);
		virtual void			add(double inValue);
		
		virtual int				getNumRows() { return mNumRows; }
		virtual int				getNumCols() { return mNumCols; }

		virtual int				getNumTitleRows() { return mNumTitleRows; }
		virtual int				getNumTitleCols() { return mNumTitleCols; }

		virtual void			setCell(int inRow, int inCol, CReportElement* inElement,
										ElementAlignment inAlignment = kAlignElementDefault);
		virtual void			setCell(int inRow, int inCol, char *inText,
										ElementAlignment inAlignment = kAlignElementDefault);
		virtual void			setCell(int inRow, int inCol, int inValue,
										ElementAlignment inAlignment = kAlignElementDefault);
		virtual void			setCell(int inRow, int inCol, double inValue,
										ElementAlignment inAlignment = kAlignElementDefault);
		virtual CComplexElement* getCell(int inRow, int inCol);
};

// ---------------------------------------------------------------------------------
//		class CListElement
// ---------------------------------------------------------------------------------

#pragma mark CListElement
class CListElement : public CTableElement {
	protected:
		ListType					mListType;
	
	public:
									CListElement(ListType inListType = kBullettedList,
												int inIndentLevel = 0);
		virtual						~CListElement();		
		
		virtual void				add(CReportElement *inElement);
		virtual void				add(char *inText);
		virtual void				add(int inValue);
		virtual void				add(double inValue);

		virtual ElementType			getElementType() { return kListElement; }				

		ListType					getListType() { return mListType; }				

		virtual int					getNumCols() { return 1; }
		virtual int					getNumTitleRows() { return 0; }
		virtual int					getNumTitleCols() { return 0; }
		
		virtual CComplexElement*	getCell(int inRow);
};

// ---------------------------------------------------------------------------------
//		class CReportRecord
// ---------------------------------------------------------------------------------

#pragma mark CReportRecord
class CReportRecord : public CComplexElement {
	protected:
		CString					*mTitle;
	
	public:
								CReportRecord(char *inTitle = NULL);
		virtual					~CReportRecord();		

		virtual ElementType		getElementType() { return kReport; }	
					
		virtual void			setTitle(char *inTitle = NULL);				
		virtual char*			getTitle();
					
		virtual void			add(CReportElement *inElement);
};

#endif /* _REPORT_RECORD_H_ */
