/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CSVGReporter.cpp

				CSVGReporter Class

 ******************************************************************************/

#include <string.h>
#include <math.h>

#include "CExceptions.h"

#include "CWriter.h"
#include "CReportRecord.h"
#include "CGraphicsRecord.h"
#include "CSVGReporter.h"

#include "CGraph.h"
#include "CLinePlot.h"
#include "CScatterPlot.h"
#include "CDataSeries.h"

#pragma mark === CSVGReporter ===

#define PI 3.14159265359

// ---------------------------------------------------------------------------------
//		 CSVGReporter
// ---------------------------------------------------------------------------------

CSVGReporter::CSVGReporter(CWriter *inWriter, bool inOwner) : CReporter(inWriter, inOwner)
{
	isReporting = false;
}

CSVGReporter::~CSVGReporter()
{
	if (isReporting)
		finishReport();
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::startReport
// ---------------------------------------------------------------------------------

void CSVGReporter::startReport(char *inTitle)
{
	writer->writeString("<?xml version=\"1.0\" standalone=\"no\"?>\n");
	writer->writeString("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20001102//EN\" ");
	writer->writeString("\"http://www.w3.org/TR/2000/CR-SVG-20001102/DTD/svg-20001102.dtd\">\n");
		
	isReporting = true;
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::finishReport
// ---------------------------------------------------------------------------------

void CSVGReporter::finishReport()
{
	isReporting = false;
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeReport
// ---------------------------------------------------------------------------------

void CSVGReporter::writeReport(CReportRecord *inReportRecord)
{
	ThrowIfNull(inReportRecord);
	
	char *title = inReportRecord->getTitle();
	
	if (!isReporting)
		startReport(title);

	CReporter::writeReport(inReportRecord);
	
	finishReport();
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeFigure
// ---------------------------------------------------------------------------------

void CSVGReporter::writeFigure(CFigureElement *inElement)
{
	char *title = inElement->getTitle();
	
	writer->writeString("<svg");

	PageUnitsType pageUnits = inElement->getPageUnits();
	double width = inElement->getWidth();
	double height = inElement->getHeight();
	if (width !=0 && height != 0) {
		writer->writeString(" xml:space=\"preserve\" width=\"");
		switch (pageUnits) {
			case kPixelPageUnits:
				writer->writeDouble(width);
				writer->writeString("px\" height=\"");
				writer->writeDouble(height);
				writer->writeString("px\"");
			break;
			
			case kCMPageUnits:
				writer->writeDouble(width);
				writer->writeString("cm\" height=\"");
				writer->writeDouble(height);
				writer->writeString("cm\"");
			break;
			
			case kInchPageUnits:
				writer->writeDouble(width);
				writer->writeString("in\" height=\"");
				writer->writeDouble(height);
				writer->writeString("in\"");
			break;
		}
	} 
	
	writeViewBox(inElement);
	writer->writeString(">\n");
	
	if (title != NULL && strlen(title) > 0) {
		writer->writeString("<title>"); 
		writer->writeString(title); 
		writer->writeString("</title>\n"); 
	}
	
	writer->writeString("<g");
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString(">\n");
	
	int i, n = inElement->getNumElements();
	for (i = 0; i < n; i++) 
		writeElement(inElement->getElement(i));

	writer->writeString("</g>\n");

	writer->writeString("</svg>\n");
}



// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeElement
// ---------------------------------------------------------------------------------

void CSVGReporter::writeElement(CReportElement *inElement)
{
	ThrowIfNull(inElement);
	
	ElementType eType = inElement->getElementType();
	
	if (eType != kGraphicElement) {
		CReporter::writeElement(inElement);
		return;
	}
	
	GraphicType gType = dynamic_cast<CGraphicElement*>(inElement)->getGraphicType();
	
	switch (gType) {
		case kLineGraphicElement:
			writeLine(dynamic_cast<CLineElement*>(inElement));
		break;
		case kRectGraphicElement:
			writeRect(dynamic_cast<CRectElement*>(inElement));
		break;
		case kRoundedRectGraphicElement:
			writeRoundedRect(dynamic_cast<CRoundedRectElement*>(inElement));
		break;
		case kCircleGraphicElement:
			writeCircle(dynamic_cast<CCircleElement*>(inElement));
		break;
		case kEllipseGraphicElement:
			writeEllipse(dynamic_cast<CEllipseElement*>(inElement));
		break;
		case kTextGraphicElement:
			writeTextGraphic(dynamic_cast<CTextGraphicElement*>(inElement));
		break;
		case kPathGraphicElement:
			writePath(dynamic_cast<CPathElement*>(inElement));
		break;
		case kGroupGraphicElement:
			writeGroup(dynamic_cast<CGroupElement*>(inElement));
		break;
		case kViewPortGraphicElement:
			writeViewPort(dynamic_cast<CViewPortElement*>(inElement));
		break;
		case kFigureGraphicElement:
			writeFigure(dynamic_cast<CFigureElement*>(inElement));
		break;
	}	
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeAtt
// ---------------------------------------------------------------------------------

void CSVGReporter::writeAtt(char *name, double value)
{
	writer->writeString(" ");
	writer->writeString(name);
	writer->writeString("=\"");
	writer->writeDouble(value);
	writer->writeString("\"");
}
	
// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeViewBox
// ---------------------------------------------------------------------------------

void CSVGReporter::writeViewBox(CViewPortElement *inElement)
{
	writer->writeString(" preserveAspectRatio=\"");
	switch (inElement->getPreserveAspectRatio()) {
		case kNoneAspect: writer->writeString("none"); break;
		case kXMinYMinAspect: writer->writeString("xMinYMin"); break;
		case kXMidYMinAspect: writer->writeString("xMidYMin"); break;
		case kXMaxYMinAspect: writer->writeString("xMaxYMin"); break;
		case kXMinYMidAspect: writer->writeString("xMinYMid"); break;
		case kXMidYMidAspect: writer->writeString("xMidYMid"); break;
		case kXMaxYMidAspect: writer->writeString("xMaxYMid"); break;
		case kXMinYMaxAspect: writer->writeString("xMinYMax"); break;
		case kXMidYMaxAspect: writer->writeString("xMidYMax"); break;
		case kXMaxYMaxAspect: writer->writeString("xMaxYMax"); break;
	}
	
	if (inElement->getSliceViewBox())
		writer->writeString(" slice");
	else
		writer->writeString(" meet");	
	writer->writeString("\"");

		
	if (inElement->getHasViewBox()) {
		writer->writeString(" viewBox=\"");
		writer->writeDouble(inElement->getViewBoxMinX());
		writer->writeString(" ");
		writer->writeDouble(inElement->getViewBoxMinY());
		writer->writeString(" ");
		writer->writeDouble(inElement->getViewBoxWidth());
		writer->writeString(" ");
		writer->writeDouble(inElement->getViewBoxHeight());
		writer->writeString("\"");
	}
}


// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeTransformations
// ---------------------------------------------------------------------------------

void CSVGReporter::writeTransformations(CGraphicElement *inElement)
{
	if (inElement->getHasRotation() || inElement->getHasScale() || 
				inElement->getHasTranslation())
		writer->writeString(" transform=\"");
		
	if (inElement->getHasRotation()) {
		double r = inElement->getRotation();
		double cx = inElement->getRotationCentreX();
		double cy = inElement->getRotationCentreY();
		
		if (cx != 0.0 || cy != 0.0) {
			writer->writeString("translate(");
			writer->writeDouble(cx);
			writer->writeString(",");
			writer->writeDouble(cy);
			writer->writeString(") ");
		}
		writer->writeString("rotate(");
		writer->writeDouble(inElement->getRotation());
		writer->writeString(")");
		if (cx != 0.0 || cy != 0.0) {
			writer->writeString(" translate(");
			writer->writeDouble(-cx);
			writer->writeString(",");
			writer->writeDouble(-cy);
			writer->writeString(") ");
		}
	}

	if (inElement->getHasScale()) {
		double x = inElement->getScaleX();
		double y = inElement->getScaleY();
		double cx = inElement->getScaleCentreX();
		double cy = inElement->getScaleCentreY();
		
		if (cx != 0.0 || cy != 0.0) {
			writer->writeString("translate(");
			writer->writeDouble(cx);
			writer->writeString(",");
			writer->writeDouble(cy);
			writer->writeString(") ");
		}
		
		if (x == y) {
			writer->writeString("scale(");
			writer->writeDouble(x);
			writer->writeString(")");
		} else {
			writer->writeString("scale(");
			writer->writeDouble(x, y);
			writer->writeString(")");
		}
		
		if (cx != 0.0 || cy != 0.0) {
			writer->writeString(" translate(");
			writer->writeDouble(-cx);
			writer->writeString(",");
			writer->writeDouble(-cy);
			writer->writeString(") ");
		}
	}
	
	if (inElement->getHasTranslation()) {
		writer->writeString(" translate(");
		writer->writeDouble(inElement->getTranslationX());
		writer->writeString(",");
		writer->writeDouble(inElement->getTranslationY());
		writer->writeString(") ");
	}
	
	if (inElement->getHasRotation() || inElement->getHasScale() || 
			inElement->getHasTranslation())
		writer->writeString("\"");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeColour
// ---------------------------------------------------------------------------------

void CSVGReporter::writeColour(ColourType inType, double inR, double inG, double inB)
{
	switch (inType) {
		case kRGBColour:
			writer->writeString("rgb(");
			writer->writeDouble(inR);
			writer->writeString(",");
			writer->writeDouble(inG);
			writer->writeString(",");
			writer->writeDouble(inB);
			writer->writeString(")");
		break;
		
		case kNoColour:writer->writeString("none"); break;
		case kBlackColour:writer->writeString("black"); break;
		case kBlueColour:writer->writeString("blue"); break;
		case kBrownColour:writer->writeString("brown"); break;
		case kCyanColour:writer->writeString("cyan"); break;
		case kDarkGrayColour:writer->writeString("darkgray"); break;
		case kGrayColour:writer->writeString("gray"); break;
		case kGreenColour:writer->writeString("green"); break;
		case kLightGrayColour:writer->writeString("lightgray"); break;
		case kMagentaColour:writer->writeString("magenta"); break;
		case kOrangeColour:writer->writeString("orange"); break;
		case kPurpleColour:writer->writeString("purple"); break;
		case kRedColour:writer->writeString("red"); break;
		case kWhiteColour:writer->writeString("white"); break;
		case kYellowColour:writer->writeString("yellow"); break;
	}
}


// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeStrokeStyle
// ---------------------------------------------------------------------------------

void CSVGReporter::writeStrokeStyle(CGraphicElement *inElement)
{
	bool hasStyle = false;
	
	if (inElement->getHasStrokeColour()) {
		writer->writeString(" style=\"stroke:");
		writeColour(inElement->getStrokeColourType(), inElement->getStrokeColourRed(),
					inElement->getStrokeColourGreen(), inElement->getStrokeColourBlue());
		
		if (inElement->getStrokeColourAlpha() < 1.0) {
			writer->writeString("; stroke-opacity:");
			writer->writeDouble(inElement->getStrokeColourAlpha());
		}
		
		hasStyle = true;
	}

	if (inElement->getHasStrokeWidth()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-width:");
		else
			writer->writeString("; stroke-width:");
		writer->writeDouble(inElement->getStrokeWidth());
		hasStyle = true;
	}
	if (inElement->getHasStrokeLineCap()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-linecap:");
		else
			writer->writeString("; stroke-linecap:");
		switch (inElement->getStrokeLineCap()) {
			case kButtLineCap: writer->writeString("butt"); break;
			case kRoundLineCap: writer->writeString("round"); break;
			case kSquareLineCap: writer->writeString("square"); break;
		}
		hasStyle = true;
	}
	if (inElement->getHasStrokeLineJoin()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-linejoin:");
		else
			writer->writeString("; stroke-linejoin:");
		switch (inElement->getStrokeLineJoin()) {
			case kMiterLineJoin: writer->writeString("miter"); break;
			case kRoundLineJoin: writer->writeString("round"); break;
			case kBevelLineJoin: writer->writeString("bevel"); break;
		}
		hasStyle = true;
	}
	if (hasStyle)
		writer->writeString("\"");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeStrokeStyle
// ---------------------------------------------------------------------------------

void CSVGReporter::writeStrokeFillStyle(CGraphicElement *inElement)
{
	bool hasStyle = false;
	
	if (inElement->getHasStrokeColour()) {
		writer->writeString(" style=\"stroke:");
		writeColour(inElement->getStrokeColourType(), inElement->getStrokeColourRed(),
					inElement->getStrokeColourGreen(), inElement->getStrokeColourBlue());
		
		if (inElement->getStrokeColourAlpha() < 1.0) {
			writer->writeString("; stroke-opacity:");
			writer->writeDouble(inElement->getStrokeColourAlpha());
		}
		
		hasStyle = true;
	}

	if (inElement->getHasStrokeWidth()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-width:");
		else
			writer->writeString("; stroke-width:");
		writer->writeDouble(inElement->getStrokeWidth());
		hasStyle = true;
	}
	if (inElement->getHasStrokeLineCap()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-linecap:");
		else
			writer->writeString("; stroke-linecap:");
		switch (inElement->getStrokeLineCap()) {
			case kButtLineCap: writer->writeString("butt"); break;
			case kRoundLineCap: writer->writeString("round"); break;
			case kSquareLineCap: writer->writeString("square"); break;
		}
		hasStyle = true;
	}
	if (inElement->getHasStrokeLineJoin()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-linejoin:");
		else
			writer->writeString("; stroke-linejoin:");
		switch (inElement->getStrokeLineJoin()) {
			case kMiterLineJoin: writer->writeString("miter"); break;
			case kRoundLineJoin: writer->writeString("round"); break;
			case kBevelLineJoin: writer->writeString("bevel"); break;
		}
		hasStyle = true;
	}
	if (inElement->getHasFillColour()) {
		if (!hasStyle)
			writer->writeString(" style=\"fill:");
		else
			writer->writeString("; fill:");
		writeColour(inElement->getFillColourType(), inElement->getFillColourRed(),
					inElement->getFillColourGreen(), inElement->getFillColourBlue());
		
		if (inElement->getFillColourAlpha() < 1.0) {
			writer->writeString("; fill-opacity:");
			writer->writeDouble(inElement->getFillColourAlpha());
		}
		
		hasStyle = true;
	}

	if (hasStyle)
		writer->writeString("\"");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeTextStyle
// ---------------------------------------------------------------------------------

void CSVGReporter::writeTextStyle(CTextGraphicElement *inElement)
{
	bool hasStyle = false;
	
	if (inElement->getHasStrokeColour()) {
		writer->writeString(" style=\"stroke:");
		writeColour(inElement->getStrokeColourType(), inElement->getStrokeColourRed(),
					inElement->getStrokeColourGreen(), inElement->getStrokeColourBlue());
		
		if (inElement->getStrokeColourAlpha() < 1.0) {
			writer->writeString("; stroke-opacity:");
			writer->writeDouble(inElement->getStrokeColourAlpha());
		}
		
		hasStyle = true;
	}
	if (inElement->getHasStrokeWidth()) {
		if (!hasStyle)
			writer->writeString(" style=\"stroke-width:");
		else
			writer->writeString("; stroke-width:");
		writer->writeDouble(inElement->getStrokeWidth());
		hasStyle = true;
	}
	if (inElement->getHasFillColour()) {
		if (!hasStyle)
			writer->writeString(" style=\"fill:");
		else
			writer->writeString("; fill:");
		writeColour(inElement->getFillColourType(), inElement->getFillColourRed(),
					inElement->getFillColourGreen(), inElement->getFillColourBlue());
		
		if (inElement->getFillColourAlpha() < 1.0) {
			writer->writeString("; fill-opacity:");
			writer->writeDouble(inElement->getFillColourAlpha());
		}
		
		hasStyle = true;
	}
	if (inElement->getHasFontSize()) {
		if (!hasStyle)
			writer->writeString(" style=\"font-size:");
		else
			writer->writeString("; font-size:");
		writer->writeDouble(inElement->getFontSize());
		hasStyle = true;
	}
	if (inElement->getHasTextAnchor()) {
		if (!hasStyle)
			writer->writeString(" style=\"text-anchor:");
		else
			writer->writeString("; text-anchor:");
		switch (inElement->getTextAnchor()) {
			case kLeftTextAnchor: writer->writeString("start"); break;
			case kMiddleTextAnchor: writer->writeString("middle"); break;
			case kRightTextAnchor: writer->writeString("end"); break;
		}
		hasStyle = true;
	}
	if (inElement->getHasFontFace()) {
		if (!hasStyle)
			writer->writeString(" style=\"font-family:");
		else
			writer->writeString("; font-family:");
		writer->writeString(inElement->getFontFace());
		hasStyle = true;
	}

	if (hasStyle)
		writer->writeString("\"");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeLine
// ---------------------------------------------------------------------------------

void CSVGReporter::writeLine(CLineElement *inElement)
{
	writer->writeString("<line");
	writeAtt("x1", inElement->getX1());
	writeAtt("y1", inElement->getY1());
	writeAtt("x2", inElement->getX2());
	writeAtt("y2", inElement->getY2());
	writeStrokeStyle(inElement);
	writeTransformations(inElement);
	writer->writeString("/>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeRect
// ---------------------------------------------------------------------------------

void CSVGReporter::writeRect(CRectElement *inElement)
{
	writer->writeString("<rect");
	writeTransformations(inElement);
	writeAtt("x", inElement->getX());
	writeAtt("y", inElement->getY());
	writeAtt("width", inElement->getWidth());
	writeAtt("height", inElement->getHeight());
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString("/>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeRoundedRect
// ---------------------------------------------------------------------------------

void CSVGReporter::writeRoundedRect(CRoundedRectElement *inElement)
{
	writer->writeString("<rect");
	writeAtt("x", inElement->getX());
	writeAtt("y", inElement->getY());
	writeAtt("width", inElement->getWidth());
	writeAtt("height", inElement->getHeight());
	writeAtt("rx", inElement->getRadiusX());
	writeAtt("ry", inElement->getRadiusY());
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString("/>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeCircle
// ---------------------------------------------------------------------------------

void CSVGReporter::writeCircle(CCircleElement *inElement)
{
	writer->writeString("<circle");
	writeAtt("cx", inElement->getX());
	writeAtt("cy", inElement->getY());
	writeAtt("r", inElement->getRadius());
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString("/>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeEllipse
// ---------------------------------------------------------------------------------

void CSVGReporter::writeEllipse(CEllipseElement *inElement)
{
	writer->writeString("<ellipse");
	writeAtt("cx", inElement->getX());
	writeAtt("cy", inElement->getY());
	writeAtt("rx", inElement->getRadiusX());
	writeAtt("ry", inElement->getRadiusY());
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString("/>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeTextGraphic
// ---------------------------------------------------------------------------------

void CSVGReporter::writeTextGraphic(CTextGraphicElement *inElement)
{
	writer->writeString("<text");
	writeAtt("x", inElement->getX());
	writeAtt("y", inElement->getY());
	writeTextStyle(inElement);
	writeTransformations(inElement);
	writer->writeString(">\n");

	writer->writeString(inElement->getText());
	writer->writeString("\n</text>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeCode
// ---------------------------------------------------------------------------------

void CSVGReporter::writeCode(char code, double value) 
{
	writer->writeChar(code);
	writer->writeDouble(value);
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeCode
// ---------------------------------------------------------------------------------

void CSVGReporter::writeCode(char code, double value1, double value2) 
{
	writer->writeChar(code);
	writer->writeDouble(value1);
	writer->writeChar(',');
	writer->writeDouble(value2);
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writePath
// ---------------------------------------------------------------------------------

void CSVGReporter::writePath(CPathElement *inElement)
{
	int i, n = inElement->getNumSubelements();
	
	writer->writeString("<path ");
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString(" d=\"");

	writeCode('M', inElement->getX(), inElement->getY());

	CPathSubelement e;
	int count = 0;
	for (i = 0; i < n; i++) {
		e = inElement->getSubelement(i);
		switch (e.mType) {
			case kMoveToSubelement:
				writeCode('M', e.mX1, e.mY1);
			break;
			case kMoveBySubelement:
				writeCode('m', e.mX1, e.mY1);
			break;
			case kLineToSubelement:
				writeCode('L', e.mX1, e.mY1);
			break;
			case kLineBySubelement:
				writeCode('l', e.mX1, e.mY1);
			break;
			case kHLineToSubelement:
				writeCode('H', e.mX1);
			break;
			case kHLineBySubelement:
				writeCode('h', e.mX1);
			break;
			case kVLineToSubelement:
				writeCode('V', e.mY1);
			break;
			case kVLineBySubelement:
				writeCode('v', e.mY1);
			break;
			case kCurveToSubelement:
				writeCode('C', e.mX1, e.mY1);
				writeCode(' ', e.mX2, e.mY2);
				writeCode(' ', e.mX3, e.mY3);
			break;
			case kCurveBySubelement:
				writeCode('c', e.mX1, e.mY1);
				writeCode(' ', e.mX2, e.mY2);
				writeCode(' ', e.mX3, e.mY3);
			break;
			case kArcSubelement: {
			
				double theta1 = (e.mX2 * PI) / 180.0;
				double theta2 = ((e.mX2 + e.mY2) * PI) / 180.0;
				double x0 = e.mX1 + (e.mX3 * cos(theta1));
				double y0 = e.mY1 + (e.mX3 * sin(theta1));
				double x1 = e.mX1 + (e.mX3 * cos(theta2));
				double y1 = e.mY1 + (e.mX3 * sin(theta2));
				char largeArc = (e.mY2 > 180.0 ? '1' : '0');
				char sweep = (e.mY2 > 0.0 ? '1' : '0');
				
				writeCode('M', x0, y0);
				writeCode('A', e.mX3, e.mX3);
				writeCode(' ', 0.0);
				writer->writeChar(largeArc);
				writer->writeChar(' ');
				writer->writeChar(sweep);
				writeCode(' ', x1, y1);
			} break;
/*			case kArcToSubelement:
				writeCode('A', e.mX1, e.mY1);
				writeCode(' ', e.mZ);
				writer->writeChar(e.mFlag1 ? '1' : '0');
				writer->writeChar(' ');
				writer->writeChar(e.mFlag2 ? '1' : '0');
				writeCode(' ', e.mX2, e.mY2);
			break;
			case kArcBySubelement:
				writeCode('a', e.mX1, e.mY1);
				writeCode(' ', e.mZ);
				writer->writeChar(e.mFlag1 ? '1' : '0');
				writer->writeChar(' ');
				writer->writeChar(e.mFlag2 ? '1' : '0');
				writeCode(' ', e.mX2, e.mY2);
			break;
*/
		}
		count++;
		if (count > 4) {
			writer->writeString("\n");
			count = 0;
		}
	}
	if (inElement->getPathClosed())
		writer->writeChar('z');
		 
	writer->writeString("\"/>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeGroup
// ---------------------------------------------------------------------------------

void CSVGReporter::writeGroup(CGroupElement *inElement)
{
	writer->writeString("<g");
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString(">\n");
	
	int i, n = inElement->getNumElements();
	for (i = 0; i < n; i++) 
		writeElement(inElement->getElement(i));

	writer->writeString("</g>\n");
}

// ---------------------------------------------------------------------------------
//		 CSVGReporter::writeViewPort
// ---------------------------------------------------------------------------------

void CSVGReporter::writeViewPort(CViewPortElement *inElement)
{
	writer->writeString("<svg");
	writeViewBox(inElement);
	writer->writeString(">\n");
	
	writer->writeString("<g");
	writeStrokeFillStyle(inElement);
	writeTransformations(inElement);
	writer->writeString(">\n");
	
	int i, n = inElement->getNumElements();
	for (i = 0; i < n; i++) 
		writeElement(inElement->getElement(i));

	writer->writeString("</g>\n");

	writer->writeString("</svg>\n");
}


