/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CString.cpp
 
		CString Class
	 
 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of strings

#include <string.h>

#include "CExceptions.h"

#include "CString.h"

#pragma mark === CString ===

// ---------------------------------------------------------------------------------
//		CString::CString
// ---------------------------------------------------------------------------------

//: Constructs an empty CString
//  The array will only be allocated when a string is added. The  
//  initial mCapacity will be zero.
CString::CString()
{
	mCapacity = 0;
	mLength = 0;
	mString = NULL;
	mIncrement = 20;
}

// ---------------------------------------------------------------------------------
//		CString::CString
// ---------------------------------------------------------------------------------

//: Constructs an empty CString but allocates initial memory
CString::CString(int inInitial)
{
	mCapacity = 0;
	mLength = 0;
	mString = NULL;
	
	setInitial(inInitial, 20);
}

// ---------------------------------------------------------------------------------
//		CString::CString
// ---------------------------------------------------------------------------------

//: Constructs an CString with inString
CString::CString(const char *inString)
{
	mCapacity = 0;
	mLength = 0;
	mString = NULL;
	mIncrement = 20;
	
	setString(inString);
}

// ---------------------------------------------------------------------------------
//		CString::~CString
// ---------------------------------------------------------------------------------

//: The destructor
CString::~CString()
{
	if (mString != NULL)
		delete [] mString;
}

// ---------------------------------------------------------------------------------
//		CString::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial mCapacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void CString::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		CString::setCapacity
// ---------------------------------------------------------------------------------

//: Set the mCapacity
// If the mCapacity is not already greater than inCapacity, then this increases it.
void CString::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		CString::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that mCapacity increments by
//  If mCapacity needs to increase then it will do so by multiples of increment.
void CString::setIncrement(int inIncrement)
{
	mIncrement = inIncrement;
}

// ---------------------------------------------------------------------------------
//		CString::getString
// ---------------------------------------------------------------------------------

//: Copies the string into outString.
void CString::getString(char *outString)
{
	ThrowIfNull(outString);
	
	if (mString != NULL)
		strcpy(outString, mString);
	else
		outString[0] = '\0';
}

// ---------------------------------------------------------------------------------
//		CString::getChar
// ---------------------------------------------------------------------------------

//: Returns the char that is at an index.
char CString::getChar(int index)
{
	TestBounds(index, mLength);
	
	return mString[index];
}

// ---------------------------------------------------------------------------------
//		CString::setObject
// ---------------------------------------------------------------------------------

//: Sets the string
void CString::setString(const char *inString)
{
	if (inString != NULL) {
		int len = strlen(inString);
		checkCapacity(len + 1);
		strcpy(mString, inString);
		mLength = len;
	}
}

// ---------------------------------------------------------------------------------
//		CString::append
// ---------------------------------------------------------------------------------

//: Adds a string to the end of the string
//  This expands the string's capacity if necessary.
void CString::append(const char *inString)
{
	if (inString != NULL) {
		int len = strlen(inString);
		checkCapacity(mLength + len + 1);
		strcat(mString, inString);
		mLength += len;
	}
}

// ---------------------------------------------------------------------------------
//		CString::clear
// ---------------------------------------------------------------------------------

//: Clears the string
void CString::clear()
{
	if (mString != NULL) {
		mString[0] = '\0';
		mLength = 0;
	}
}

// ---------------------------------------------------------------------------------
//		CString::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the string  has enough space and resizes it as necessary
//  Keeps increasing the mCapacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new string  of this
//  size, copies all the old string into it and sets the rest to NULL.
//  Finally it deletes the old string, and replaces it with the new.
void CString::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity > mCapacity) {
		char *newString;
		int i;
		
		// Increase mCapacity
		while (mCapacity<requiredCapacity)
			mCapacity += mIncrement;
		
		// Allocate the new list.
		NewArray(newString, char, mCapacity);
		newString[0] = '\0';
		
		if (mString != NULL) {
			// Copy the existing string into the new list.
			strcpy(newString, mString);
				
			// Free the old string.
			delete [] mString;
		}
		
		// Point the old string to the new one.
		mString = newString;
	}
}

