/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CStringArray.cpp
 
		CStringArray Class
	 
 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of strings

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"

#include "CStringArray.h"

#pragma mark === CStringArray ===

// ---------------------------------------------------------------------------------
//		CStringArray::CStringArray
// ---------------------------------------------------------------------------------

//: Constructs an empty CStringArray
//  The array will only be allocated when a string is added. The increment 
//  amount will be set to a default value, and the initial mCapacity will be zero.
CStringArray::CStringArray(int inInitial, int inIncrement)
{
	mCapacity=0;
	mNumStrings=0;
	mStrings=NULL;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		CStringArray::~CStringArray
// ---------------------------------------------------------------------------------

//: The destructor
CStringArray::~CStringArray()
{
	if (mStrings!=NULL) {
		for (int i = 0; i < mNumStrings; i++) {
			if (mStrings[i] != NULL)
				delete [] mStrings[i];
		}
		delete [] mStrings;
	}
}

// ---------------------------------------------------------------------------------
//		CStringArray::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial mCapacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void CStringArray::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		CStringArray::setCapacity
// ---------------------------------------------------------------------------------

//: Set the mCapacity
// If the mCapacity is not already greater than inCapacity, then this increases it.
void CStringArray::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		CStringArray::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that mCapacity increments by
//  If mCapacity needs to increase then it will do so by multiples of increment.
void CStringArray::setIncrement(int inIncrement)
{
	mIncrement = inIncrement;
}

// ---------------------------------------------------------------------------------
//		CStringArray::getValue
// ---------------------------------------------------------------------------------

//: Returns the string that is at an index.
char *CStringArray::getString(int index)
{
	TestBounds(index, mNumStrings);
	
	return mStrings[index];
}

// ---------------------------------------------------------------------------------
//		CStringArray::setObject
// ---------------------------------------------------------------------------------

//: Sets the string at an index
//  If index>mNumStrings then mNumStrings becomes index+1 with memory 
//  allocated if needed.
void CStringArray::setString(int index, char *inString)
{
	ThrowIf(index<0);

	if (index>=mNumStrings) {
		checkCapacity(index+1);
		mNumStrings=index+1;
	} 
		
	if (mStrings[index] != NULL)
		delete [] mStrings[index];
		
	NewArray(mStrings[index], char, strlen(inString) + 1);
	strcpy(mStrings[index], inString);
}

// ---------------------------------------------------------------------------------
//		CStringArray::add
// ---------------------------------------------------------------------------------

//: Adds a string to the end of the list
//  This expands the list's mCapacity if necessary.
int CStringArray::add(char *inString)
{
	setString(mNumStrings, inString);

	return mNumStrings - 1;
}

// ---------------------------------------------------------------------------------
//		CStringArray::insert
// ---------------------------------------------------------------------------------

//: Inserts a string into the list at a position
//  This expands the list's mCapacity if necessary and moves others downwards
//  to make space for the new string. If the string's new position equals
//  numObjects then this adds it to the end of the list.
void CStringArray::insert(int index, char *inString)
{
	ThrowIf(index < 0 || index > mNumStrings + 1);

	checkCapacity(mNumStrings + 1);
		
	for (int i = mNumStrings; i > index; i--)
		mStrings[i] = mStrings[i-1];
		
	mNumStrings++;
	mStrings[index] = NULL;
	
	setString(index, inString);
}

// ---------------------------------------------------------------------------------
//		CStringArray::remove
// ---------------------------------------------------------------------------------

//: Removes the string at index and moves others up 
//  This function returns the string being removed.
char* CStringArray::remove(int index)
{
	TestBounds(index, mNumStrings);

	char *tmp = mStrings[index];

	for (int i=index; i<mNumStrings-1; i++)
		mStrings[i] = mStrings[i+1];
		
	mNumStrings--;
	
	return tmp;
}

// ---------------------------------------------------------------------------------
//		CStringArray::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the strings in the list
void CStringArray::removeAll()
{
	for (int i = 0; i < mNumStrings; i++) {
		delete [] mStrings[i];
		mStrings[i] = NULL;
	}
	mNumStrings=0;
}

// ---------------------------------------------------------------------------------
//		CStringArray::findString
// ---------------------------------------------------------------------------------

//: Finds the index of a string in the list
int CStringArray::findString(char *inString)
{
	int i = 0;
	while (i < mNumStrings && strcmp(mStrings[i], inString) != 0) 
		i++;
	
	if (i == mNumStrings)
		return -1;
			
	return i;
}

// ---------------------------------------------------------------------------------
//		CStringArray::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the string list has enough space and resizes it as necessary
//  Keeps increasing the mCapacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new string array of this
//  size, copies all the old strings into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void CStringArray::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity > mCapacity) {
		char **newStrings;
		int i;
		
		// Increase mCapacity
		while (mCapacity<requiredCapacity)
			mCapacity += mIncrement;
		
		// Allocate the new list.
		NewArray(newStrings, char*, mCapacity);
		
		if (mStrings != NULL) {
			// Copy the existing list into the new list.
			for (i=0; i<mNumStrings; i++)
				newStrings[i] = mStrings[i];
			for (i=mNumStrings; i<mCapacity; i++)
				newStrings[i] = NULL;
				
			// Free the old list.
			delete [] mStrings;
		} else {
			for (i=0; i<mCapacity; i++)
				newStrings[i] = NULL;
		}
		
		// Point the old list to the new one.
		mStrings = newStrings;
	}
}

