/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CTextReporter.cpp

				CTextReporter Class

 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "CExceptions.h"

#include "CWriter.h"
#include "CReportRecord.h"
#include "CGraphicsRecord.h"
#include "CTextReporter.h"

#pragma mark === CTextReporter ===

// ---------------------------------------------------------------------------------
//		 CTextReporter
// ---------------------------------------------------------------------------------

CTextReporter::CTextReporter(CWriter *inWriter, bool inOwner) : CReporter(inWriter, inOwner)
{
	mIsWrapping = true;
	mIsPreformatted = false;
	mPageWidth = 80;
	mCurrentColumn = 0;
	mInitialIndent = 0;
	mParagraphIndent = 0;
	mTabSize = 4;
	mLineBetweenParagraphs = true;

	mLine = NULL;
}

CTextReporter::~CTextReporter()
{
	if (mLine != NULL)
		delete [] mLine;
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::setPageWidth
// ---------------------------------------------------------------------------------

void CTextReporter::setPageWidth(int inPageWidth)
{
	mPageWidth = inPageWidth;
	if (mLine != NULL) {
		delete [] mLine;
		mLine = NULL;
	}
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::setLineBetweenParagraphs
// ---------------------------------------------------------------------------------

void CTextReporter::setLineBetweenParagraphs(bool inLineBetweenParagraphs)
{
	mLineBetweenParagraphs = inLineBetweenParagraphs;
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeReport
// ---------------------------------------------------------------------------------

void CTextReporter::writeReport(CReportRecord *inReportRecord)
{
	ThrowIfNull(inReportRecord);
	
	char *title = inReportRecord->getTitle();
	if (title != NULL && strlen(title) > 0) {
		writer->writeString(title);
		writer->writeChar('\n');
		for (int i = 0; i < strlen(title); i++)
			writer->writeChar('=');
		writer->writeChar('\n');
		writer->writeChar('\n');
	}

	CReporter::writeReport(inReportRecord);
	
	if (mIsWrapping && mCurrentColumn > 0)
			flushToEndOfLine();
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::getTokenLength
// ---------------------------------------------------------------------------------

int CTextReporter::getTokenLength(char *inText)
{
	int len = strlen(inText);
	char *P = inText;
	int i = 0;
	while (i < len && !isspace(*P)) {
		P++;
		i++;				
	}
	return i;
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeText
// ---------------------------------------------------------------------------------

void CTextReporter::writeText(char *inText)
{
	if (inText == NULL)
		return;
		
	int len = strlen(inText);
	
	int i = 0;
	int k, n;
	
	if (mIsWrapping) {
		if (mLine == NULL) {
			NewArray(mLine, char, mPageWidth + 1);
			memset(mLine, ' ', mPageWidth);
			mLine[mPageWidth] = '\0';
		}
		
		while (i < len) {
			while (i < len && isspace(inText[i]))
				i++;

			n = getTokenLength(&inText[i]);
			
			if (n > 0) {
				if (n + mCurrentColumn > mPageWidth) { // Wrap next token onto next line
					if (n > mPageWidth - mParagraphIndent) { // can't fit next token on a single line
						for (k = 0; k < mPageWidth; k++) {
							mLine[mCurrentColumn] = inText[i];
							i++;
							mCurrentColumn++;
						}
					} 
					
					flushLine();
				} else { // Write next token to line
					if (mCurrentColumn > 0) { 	// if not first token on line,
									// write space.
						mLine[mCurrentColumn] = ' ';
						mCurrentColumn++;
					} else {		// Indent
						for (k = 0; k < mParagraphIndent; k++) {
							mLine[mCurrentColumn] = ' ';
							mCurrentColumn++;
						}
					}
					
					for (k = 0; k < n; k++) {
						mLine[mCurrentColumn] = inText[i];
						i++;
						mCurrentColumn++;
					}
				}
			} else
				i++;
		}
	} else {
		while (i < len) {
			if (!mIsPreformatted) {
				while (i < len && isspace(inText[i]))
					i++;
				n = getTokenLength(&inText[i]);
			} else {
				n = len;
			}
			
			if (n > 0) {
				if (mCurrentColumn > 0) { 	// if not first token on line,
											// write space.
					if (!mIsPreformatted) {
						writer->writeChar(' ');
						mCurrentColumn++;
					}
				} else {		// Indent
					for (k = 0; k < mParagraphIndent; k++) {
						writer->writeChar(' ');
						mCurrentColumn++;
					}
				}
				for (k = 0; k < n; k++) {
					writer->writeChar(inText[i]);
					i++;
					mCurrentColumn++;
				}
			} else
				i++;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeText
// ---------------------------------------------------------------------------------

void CTextReporter::writeText(CTextElement *inElement)
{
	writeText(inElement->getText());
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::flushLine
// ---------------------------------------------------------------------------------

void CTextReporter::flushLine()
{
	mLine[mCurrentColumn] = '\0';
	writer->writeString(mLine);
	writer->writeChar('\n');
	mCurrentColumn = 0;
	memset(mLine, ' ', mPageWidth);
	mLine[mPageWidth] = '\0';
}
	
// ---------------------------------------------------------------------------------
//		 CTextReporter::flushToEndOfLine
// ---------------------------------------------------------------------------------

void CTextReporter::flushToEndOfLine()
{
	if (mCurrentColumn > 0) {
		mLine[mCurrentColumn] = '\0';
		writer->writeString(mLine);
		mCurrentColumn = 0;
		memset(mLine, ' ', mPageWidth);
		mLine[mPageWidth] = '\0';
	}
}
	
// ---------------------------------------------------------------------------------
//		 CTextReporter::writeLineBreak
// ---------------------------------------------------------------------------------

void CTextReporter::writeLineBreak(CLineBreakElement *inElement)
{
	if (mIsWrapping) {
		if (mLine == NULL) {
			NewArray(mLine, char, mPageWidth + 1);
			memset(mLine, ' ', mPageWidth);
			mLine[mPageWidth] = '\0';
		}
		
		if (mCurrentColumn > 0)
			flushLine();
	} else {
		writer->writeChar('\n');
		mCurrentColumn = 0;
	}
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeInteger
// ---------------------------------------------------------------------------------

void CTextReporter::writeInteger(CIntegerElement *inElement)
{
	char text[64];
	sprintf(text, "%d", inElement->getValue());
	writeText(text);
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeReal
// ---------------------------------------------------------------------------------

void CTextReporter::writeReal(CRealElement *inElement)
{
	char text[64], fmt[16];
	int precision = inElement->getPrecision();

	if (inElement->getIsSciFormat()) {
		if (precision < 0)
			sprintf(fmt, "%%lE");
		else
			sprintf(fmt, "%%.%dlE", precision);
	} else {
		if (precision < 0)
			sprintf(fmt, "%%lf");
		else
			sprintf(fmt, "%%.%dlf", precision);
	}
	sprintf(text, fmt, inElement->getValue());
	writeText(text);
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeComplexElement
// ---------------------------------------------------------------------------------

void CTextReporter::writeComplexElement(CComplexElement *inElement)
{
	int i, n = inElement->getNumElements();
	for (i = 0; i < n; i++) {
		writeElement(inElement->getElement(i));
//		if (i < n-1)
//			writer->writeChar(' ');
	}
}


// ---------------------------------------------------------------------------------
//		 CTextReporter::writeToken
// ---------------------------------------------------------------------------------

void CTextReporter::writeToken(CTokenElement *inElement)
{
	char text[1024];
	char text2[64];
	int i, n = inElement->getNumElements();
	CReportElement *element;
	ElementType type;
	
	strcpy(text, "");
	
	for (i = 0; i < n; i++) {
		element = inElement->getElement(i);
		type = element->getElementType();
	
		switch (type) {
			case kReport: break; // shouldn't occur
			
			case kTextElement:
				strcat(text, ((CTextElement *)element)->getText());
			break;
			case kIntegerElement:
				sprintf(text2, "%d", ((CIntegerElement *)element)->getValue());
				strcat(text, text2);
			break;
			case kRealElement:
				sprintf(text2, "%G", ((CRealElement *)element)->getValue());
				strcat(text, text2);
			break;
		}
	}
	writeText(text);
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeParagraph
// ---------------------------------------------------------------------------------
// Should wrap text here... (or in writeText)?

void CTextReporter::writeParagraph(CParagraphElement *inElement)
{
	mParagraphIndent = inElement->getIndentLevel() * mTabSize;
	mInitialIndent = mParagraphIndent;
	
	mIsWrapping = true;
	mIsPreformatted = false;
//	writer->writeSpaces(mInitialIndent);
//	mCurrentColumn = mInitialIndent;
	writeComplexElement(inElement);
	
	writeLineBreak();
	
	if (mLineBetweenParagraphs)
		writer->writeChar('\n');
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeParagraph
// ---------------------------------------------------------------------------------

void CTextReporter::writePreformatted(CPreformattedElement *inElement)
{
	mParagraphIndent = inElement->getIndentLevel() * mTabSize;
	mInitialIndent = mParagraphIndent;
	
	mIsWrapping = false;
	mIsPreformatted = true;
	writeComplexElement(inElement);
	
	writeLineBreak();
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeTable
// ---------------------------------------------------------------------------------

void CTextReporter::writeTable(CTableElement *inElement)
{
	CComplexElement *cell;
	
	int i, j;
	mIsWrapping = false;
	mIsPreformatted = false;
	mParagraphIndent = inElement->getIndentLevel() * mTabSize;
	mInitialIndent = mParagraphIndent;
	
	for (i = 0; i < inElement->getNumRows(); i++) {
		for (j = 0; j < inElement->getNumCols(); j++) {
			cell = inElement->getCell(i, j);
			if (cell != NULL)
				writeComplexElement(cell);
			if (j < inElement->getNumCols() - 1)
				writer->writeChar('\t');
		}
		writer->writeChar('\n');
		
		if (i == inElement->getNumTitleRows() - 1) {
			for (j = 0; j < mCurrentColumn; j++)
				writer->writeChar('=');
			writer->writeChar('\n');
		}
		
		mCurrentColumn = 0;
	}
	writeLineBreak();
	if (mLineBetweenParagraphs)
		writer->writeChar('\n');
}

// ---------------------------------------------------------------------------------
//		 CTextReporter::writeList
// ---------------------------------------------------------------------------------

void CTextReporter::writeList(CListElement *inElement)
{
	char text[64];
	int i, n = inElement->getNumElements();

	i = 0;
	while (i < n) {
		if (inElement->getListType() == kDefinitionList) {
			mParagraphIndent = inElement->getIndentLevel() * mTabSize;
			mInitialIndent = mParagraphIndent;
			
			writeComplexElement((CComplexElement *)inElement->getElement(i));
			
			i++;
			if (i < n) {	
				writeLineBreak();
				
				mParagraphIndent = (inElement->getIndentLevel() + 1) * mTabSize;
				mInitialIndent = mParagraphIndent;
			
				writeComplexElement((CComplexElement *)inElement->getElement(i));
			}
		} else if (inElement->getListType() == kBullettedList) {
			mParagraphIndent = inElement->getIndentLevel() * mTabSize;
			mInitialIndent = mParagraphIndent;
			
			writeText("*");
			writeComplexElement((CComplexElement *)inElement->getElement(i));
		} else { // kNumberedList
			mParagraphIndent = inElement->getIndentLevel() * mTabSize;
			mInitialIndent = mParagraphIndent;
			
			sprintf(text, "%d", i+1);
			writeText(text);
			writeComplexElement((CComplexElement *)inElement->getElement(i));
		}

		writeLineBreak();

		i++;
	}
	
	if (mLineBetweenParagraphs)
		writer->writeChar('\n');
}


// ---------------------------------------------------------------------------------
//		 CTextReporter::writeFigure
// ---------------------------------------------------------------------------------

void CTextReporter::writeFigure(CFigureElement *inElement)
{
	if (inElement->getName() != NULL)
		writeText(inElement->getName());
	if (inElement->getTitle() != NULL)
		writeText(inElement->getTitle());
	writeLineBreak();
	writeText("[Figure]");
	writeLineBreak();
	if (mLineBetweenParagraphs)
		writer->writeChar('\n');
}


