/******************************************************************************

ZooCore C++ Class Library (ZooCoreLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of ZooCoreLib C++ Class Library.

    ZooCoreLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ZooCoreLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 CVariate.cpp
 
		CVariate class 

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of doubles

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "CExceptions.h"

#include "CVariate.h"

#pragma mark === CVariate ===

// ---------------------------------------------------------------------------------
//		CVariate::CVariate
// ---------------------------------------------------------------------------------

//: Constructs an empty CVariate
CVariate::CVariate(int inInitial, int inIncrement) : CDoubleArray(inInitial, inIncrement)
{
	weight = 0.0;
}

// ---------------------------------------------------------------------------------
//		CVariate::~CVariate
// ---------------------------------------------------------------------------------

//: The destructor
CVariate::~CVariate()
{
}

// ---------------------------------------------------------------------------------
//		CVariate::getSum
// ---------------------------------------------------------------------------------

double CVariate::getSum()
{
	int i;
	double sum = 0.0;
	for (i = 0; i < numValues; i++) {
		sum += getValue(i);
	}
	return sum;
}

// ---------------------------------------------------------------------------------
//		CVariate::getSumOfSquares
// ---------------------------------------------------------------------------------

double CVariate::getSumOfSquares()
{
	int i;
	double d, sum = 0.0;
	for (i = 0; i < numValues; i++) {
		d = getValue(i);
		sum += pow(d, 2);
	}
	return sum;
}

// ---------------------------------------------------------------------------------
//		CVariate::getMean
// ---------------------------------------------------------------------------------

double CVariate::getMean()
{
	if (weight > 0.0)
		return getSum() / weight;
	else
		return getSum() / getNumValues();
}

// ---------------------------------------------------------------------------------
//		CVariate::getMedian
// ---------------------------------------------------------------------------------

double CVariate::getMedian()
{
	sort(true);
	
	double median;
	int n = getNumValues();
	if (numValues % 2 == 1)		// n is odd
		median = getValue(n / 2);
	else				// n is even
		median = (getValue((n / 2) - 1) + getValue(n / 2)) / 2.0;

	return median;
}

// ---------------------------------------------------------------------------------
//		CVariate::getVariance
// ---------------------------------------------------------------------------------

double CVariate::getVariance()
{
	double n = getNumValues();
	if (weight > 0.0)
		n = weight;
	double m = getMean();
	return ( getSumOfSquares() - (n * m * m) ) / (n - 1);
}

// --------------------------------------------------------------------------------
//    CVariate::getStandardDeviation
// --------------------------------------------------------------------------------

double CVariate::getStandardDeviation()
{
	return sqrt(getVariance());
}

// --------------------------------------------------------------------------------
//    CVariate::getStandardError
// --------------------------------------------------------------------------------

double CVariate::getStandardError()
{
	double n = getNumValues();
	if (weight > 0.0)
		n = weight;
	return getSD() / sqrt(n);
}

// --------------------------------------------------------------------------------
//    CVariate::getPercentile
// --------------------------------------------------------------------------------
double CVariate::getPercentile(double inPercentile)
{
	sort(true);
	
	int numDivs = getNumValues() - 1;
	double division = 100.0 / numDivs;
	int below = (int)floor(inPercentile / division);
	int above = below + 1;

	double belowPer = below * division;
	double abovePer = above * division;

	double belowVal = getValue(below);
	double aboveVal = getValue(above);

	double fraction = (inPercentile - belowPer) / division;

	return belowVal + ((aboveVal - belowVal) * fraction);
}

// --------------------------------------------------------------------------------
//    CVariate::getSampleVariance
// --------------------------------------------------------------------------------

double CVariate::getSampleVariance()
{
	double n = getNumValues();
	if (weight > 0.0)
		n = weight;
	double ex2 = getSumOfSquares();
	double m = getMean();
	double m2 = m * m;
	double ex1 = n * m2;
	double tex1 = getSum() * m;
	double sVar = (ex2 - ex1) / (n - 1);
	return ( getSumOfSquares() - getSampleSize() * m2 ) / (n - 1);
}

// --------------------------------------------------------------------------------
//    CVariate::getSampleStandardDeviation
// --------------------------------------------------------------------------------

double CVariate::getSampleStandardDeviation()
{
	return sqrt(getSampleVariance());
}

// --------------------------------------------------------------------------------
//    CVariate::getCoefficientOfVariation
// --------------------------------------------------------------------------------

double CVariate::getCoefficientOfVariation()
{
	return getSampleStandardDeviation() * 100.0 / getMean();
}

