/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEAlignment.cpp

				MEAlignment Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEObject.h"
#include "MEBasicAlignment.h"
#include "MEFeature.h"
#include "MEFeatureList.h"
#include "MESequence.h"
#include "MEAlignment.h"
#include "MEBitset.h"
#include "MEStream.h"

#pragma mark === MEAlignment ===

// ---------------------------------------------------------------------------------
//		 MEAlignment
// ---------------------------------------------------------------------------------

MEAlignment::MEAlignment() : MEBasicAlignment()
{
	featureList = NULL;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment
// ---------------------------------------------------------------------------------

// This special constructor upgrades an MEBasicAlignment to an MEAlignment.
MEAlignment::MEAlignment(MEBasicAlignment *src) : MEBasicAlignment()
{
	alignmentType = src->getAlignmentType();
	featureList = NULL;
	
	numObjects = getNumObjects();
	capacity = getCapacity();
	increment = getIncrement();
	objects = src->getObjects();
	src->removeAll();

	MEBasicSequence *seq0;
	MESequence *seq1 = NULL;
	
	for (int i = 0; i < numObjects; i++) {	
		seq0 = (MEBasicSequence *)getObject(i);
		NewClass(seq1, MESequence());
		seq1->clone(seq0);
		setObject(i, seq1);
		delete seq0;
	}
}

// ---------------------------------------------------------------------------------
//		 ~MEAlignment
// ---------------------------------------------------------------------------------

MEAlignment::~MEAlignment()
{
	if (featureList != NULL && featureList->getOwner() == this)
		delete featureList;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::clone
// ---------------------------------------------------------------------------------

void MEAlignment::clone(MEObject *src)
{
	MEBasicAlignment::clone(src);
	
	MEAlignment *align = (MEAlignment *)src;
	if (align->getFeatureList() != NULL) {
		NewClass(featureList, MEFeatureList());
		featureList->clone(align->getFeatureList());
	}
}

// ---------------------------------------------------------------------------------
//		MEAlignment::createObject
// ---------------------------------------------------------------------------------

MEObject* MEAlignment::createObject()
{
	MEObject *obj;
	NewClass(obj, MESequence());
	return obj;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::setAlignmentType
// ---------------------------------------------------------------------------------
// Alignment type specifies the preferred type of sequence in the alignment.
// The default is kUnknownAlignment which is compatible with all sequences.
// If another type is set, then this function checks that all sequences are
// compatible:
//     kAminoAcidAlignment is compatible with amino acid, DNA or RNA sequences.
//     kCodonAlignment is compatible with DNA or RNA sequences.
//     kNucleotideAlignment is compatible with DNA or RNA sequences.

void MEAlignment::setAlignmentType(AlignmentType inType)
{
	int i;
	MESequence *seq;

	int n = getNumSequences();
	for (i = 0; i < n; i++) {
		seq = (MESequence*)getSequence(i);
		if (!seq->isSequenceCompatible(inType))
			throw MEIncompatibleSequence();
		seq->setAlignmentType(inType);
	}
	
	alignmentType = inType;
	numSitesKnown = false;
	
	switch (alignmentType) {
		case kNucleotideAlignment:
			numStates = 4; 
		break;
		case kAminoAcidAlignment:
			numStates = 20; 
		break;
		case kCodonAlignment:
			numStates = 64; 
		break;
		case kDiscreteAlignment:
		case kUnknownAlignment:
			// Allow all letters and digits
			numStates = 65;
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::setAlignmentType
// ---------------------------------------------------------------------------------
// This specifies the alignmentType by the sequenceType (i.e. as inferred by calling
// MEBasicAlignment's inferSequenceType() function).

void MEAlignment::setAlignmentType(SequenceType inType)
{
	switch (inType) {
		case kUnknownSequence:
			return;
		case kDNASequence:
		case kRNASequence:
			setAlignmentType(kNucleotideAlignment);
		break;
		case kAminoAcidSequence:
			setAlignmentType(kAminoAcidAlignment);
		break;
		default:
			throw MEIncompatibleSequence();
	}
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::setSequenceTypes
// ---------------------------------------------------------------------------------
// This sets the sequenceType for each sequence.

void MEAlignment::setSequenceTypes(SequenceType inType)
{
	int i, n = getNumSequences();
	MESequence *seq;

	for (i = 0; i < n; i++) {
		seq = (MESequence*)getSequence(i);
		seq->setSequenceType(inType);
	}
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getNumSites
// ---------------------------------------------------------------------------------
// This seems to be have to be overridden because MEBasicAlignment::getNumSites(int)
// is overridden - is this a bug?

int MEAlignment::getNumSites()
{
	if (!numSitesKnown)
		numSites = getMaxNumSites();
		
	numSitesKnown = true;

	return numSites;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getNumSites
// ---------------------------------------------------------------------------------
// This gets the number of sites for the alignment if it were of type inAlignmentType
// which can be different from the alignment's own type.

int MEAlignment::getNumSites(AlignmentType inAlignmentType)
{
	int len, maxLen=0;
	MESequence* seq;
	
	int n=getNumSequences();
	for (int i=0; i<n; i++) {
		seq = (MESequence *)getSequence(i);
		len = seq->getNumSites(inAlignmentType);
		if (len>maxLen)
			maxLen=len;
	}
	
	return maxLen;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getNumStates
// ---------------------------------------------------------------------------------

int MEAlignment::getNumStates()
{
	return numStates;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::newSequence
// ---------------------------------------------------------------------------------

// Returns a new sequence. 
MEBasicSequence* MEAlignment::newSequence(SequenceType inType)
{
	MEBasicSequence* seq;
	
	NewClass(seq, MESequence());
	seq->setSequenceType(inType);
	return seq;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::insertGap
// ---------------------------------------------------------------------------------

void MEAlignment::insertGap(int inTaxon, int inSite, int inLength)
{
	MESequence *seq = (MESequence*)getSequence(inTaxon);
	int site1 = seq->getAlign2GapSite(inSite);
	int site2 = seq->getAlign2GapSite(inSite + inLength);
	seq->insertGap(site1, site2 - site1);
	numSitesKnown = false;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::removeSites
// ---------------------------------------------------------------------------------

void MEAlignment::removeSites(int inTaxon, int inSite, int inLength)
{
	MESequence *seq = (MESequence*)getSequence(inTaxon);
	int site1 = seq->getAlign2GapSite(inSite);
	int site2 = seq->getAlign2GapSite(inSite + inLength);
	seq->removeSites(site1, site2 - site1);
	numSitesKnown = false;
}

// In these conversions, the sequence (seq) index refers to the underlying 
// sequence as if all gaps had been removed. This index can be used to refer 
// to sites independently of any gaps being added or removed. The sequence
// index also reflects the current reading frame - it can be less than 0 if
// the sequence is not in the first reading frame. Gapped (gap)
// index also refers to the underlying sequence but including any gaps. 
// Alignment (align) index refers to the sites in the alignment, in whatever
// alignment type and reading frame it is in (so it may be amino acids although
// the underlying sequence is nucleotides).

// ---------------------------------------------------------------------------------
//		 MEAlignment::getSeq2AlignSite
// ---------------------------------------------------------------------------------

int MEAlignment::getSeq2AlignSite(int inTaxon, int inSequenceSite)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	int index=seq->getSeq2GapSite(inSequenceSite);
	index=seq->getGap2AlignSite(index);
	
	return index;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getSeq2GapSite
// ---------------------------------------------------------------------------------

int MEAlignment::getSeq2GapSite(int inTaxon, int inSequenceSite)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getSeq2GapSite(inSequenceSite);	
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getGap2AlignSite
// ---------------------------------------------------------------------------------

int MEAlignment::getGap2AlignSite(int inTaxon, int inGappedSite)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getGap2AlignSite(inGappedSite);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getGap2AlignLength
// ---------------------------------------------------------------------------------

int MEAlignment::getGap2AlignLength(int inTaxon, int inGappedLength)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getGap2AlignLength(inGappedLength);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getAlign2SeqSite
// ---------------------------------------------------------------------------------

int MEAlignment::getAlign2SeqSite(int inTaxon, int inAlignmentSite)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	int index=seq->getAlign2GapSite(inAlignmentSite);
	index=seq->getGap2SeqSite(index);
	return index;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getAlign2GapSite
// ---------------------------------------------------------------------------------

int MEAlignment::getAlign2GapSite(int inTaxon, int inAlignmentSite)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getAlign2GapSite(inAlignmentSite);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getAlign2GapLength
// ---------------------------------------------------------------------------------

int MEAlignment::getAlign2GapLength(int inTaxon, int inAlignmentLength)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getAlign2GapLength(inAlignmentLength);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getGap2SeqSite
// ---------------------------------------------------------------------------------

int MEAlignment::getGap2SeqSite(int inTaxon, int inGappedSite)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getGap2SeqSite(inGappedSite);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getNumSites
// ---------------------------------------------------------------------------------

// Returns the length of the sequence at inTaxon in characters. Thus if the
// sequence is in triplets this will be 3 times longer than getNumSites.

int MEAlignment::getNumSites(int inTaxon)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getNumSites();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getStringLength
// ---------------------------------------------------------------------------------

// Returns the length of the sequence at inTaxon in characters. Thus if the
// sequence is in triplets this will be 3 times longer than getSequenceLength.

int MEAlignment::getStringLength(int inTaxon)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getStringLength();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getStringLength
// ---------------------------------------------------------------------------------

// Returns the length of the a theoretical portion sequence at inTaxon of inNumSites
// in characters. 

int MEAlignment::getStringLength(int inTaxon, int inNumSites)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	return seq->getStringLength(inNumSites);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getSequenceString
// ---------------------------------------------------------------------------------

// Copies the sequence at inTaxon into a string. Must be long enough.
// This will be translated into the appropriate alignment type.
void MEAlignment::getSequenceString(int inTaxon, char *outStr)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	seq->getSequenceString(outStr);
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getSequenceString
// ---------------------------------------------------------------------------------

// Copies the sequence at inTaxon into a string. Must be long enough.
// This will be translated into the appropriate alignment type. If 
// inNumSites is less than the number of sites in this sequence then 
// then the whole sequence is written. If greater then the output 
// is padded with gaps.

void MEAlignment::getSequenceString(int inTaxon, char *outStr, int inNumSites)
{
	MESequence *seq=(MESequence *)getSequence(inTaxon);
	int len=seq->getNumSites();
	
	seq->getSequenceString(outStr);

	char *P=outStr+seq->getStringLength();
	if (len<inNumSites)
		seq->getGapString(P, inNumSites-len);

	outStr[inNumSites]='\0';
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getSequenceString
// ---------------------------------------------------------------------------------

// Copies part of the sequence at inTaxon into a string. Must be long enough.
// This will be translated into the appropriate alignment type. If 
// inNumSites is less than the number of sites in this sequence then 
// then the whole sequence is written. If greater then the output 
// is padded with gaps.

void MEAlignment::getSequenceString(int inTaxon, char *outStr, int inStartSite, int inNumSites)
{
	MESequence *seq=(MESequence *)getSequence(inTaxon);
	seq->getSequenceString(outStr, inStartSite, inNumSites);
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getSequenceCodes
// ---------------------------------------------------------------------------------

// Copies the sequence codes at inTaxon into a string. Must be long enough.
// This will be translated into the appropriate alignment type.
void MEAlignment::getSequenceCodes(int inTaxon, short *outCodes)
{
	MESequence *seq=(MESequence*)getSequence(inTaxon);
	seq->getSequenceCodes(outCodes);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getGapString
// ---------------------------------------------------------------------------------

// Gets a string of gaps equivalent to inNumSites in the current alignment type.

void MEAlignment::getGapString(int inTaxon, char *outStr, int inNumSites)
{
	MESequence *seq=(MESequence *)getSequence(inTaxon);
	seq->getGapString(outStr, inNumSites);
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getSiteIsGapped
// ---------------------------------------------------------------------------------

// If the site contains a gap in any taxon, this returns true.

bool MEAlignment::getSiteIsGapped(int inSite)
{
	MESequence *seq = (MESequence *)getFirstSequence();
	
	while (seq != NULL) {
		if (seq->getSiteIsGap(inSite))
			return true;
			
		seq = (MESequence *)getNextSequence();
	}
	
	return false;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getSiteIsConstant
// ---------------------------------------------------------------------------------

// If the site the same in every taxon, this returns true.

bool MEAlignment::getSiteIsConstant(int inSite)
{
	MESequence *seq = (MESequence *)getFirstSequence();
	if (seq == NULL)
		return true;
		
	int state = seq->getSiteCode(inSite);
	
	seq = (MESequence *)getNextSequence();
	while (seq != NULL) {
		if (seq->getSiteCode(inSite) != state)
			return false;
			
		seq = (MESequence *)getNextSequence();
	}
	
	return true;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::setSequenceFrame
// ---------------------------------------------------------------------------------

void MEAlignment::setSequenceFrame(int inTaxon, int inFrame)
{
	((MESequence*)getSequence(inTaxon))->setReadingFrame(inFrame);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getSequenceFrame
// ---------------------------------------------------------------------------------

int MEAlignment::getSequenceFrame(int inTaxon)
{
	return ((MESequence*)getSequence(inTaxon))->getReadingFrame();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getState
// ---------------------------------------------------------------------------------

// Returns the state of site 'site' in sequence 'inTaxon'.
char MEAlignment::getState(int inTaxon, int inSite)
{
	MESequence *seq=(MESequence *)getSequence(inTaxon);
	return seq->getSite(inSite);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::putSplit -- MAC
// ---------------------------------------------------------------------------------
void MEAlignment::putSplit(MEBitset *b, int idx, char *splitMap)
// This function puts the split represented by the idx-th site with mapping splitMap.
{
	char
		ch;
	int
		i, n = getNumSequences();
	b->resize(n);
	b->clear();
	for (i = 0; i < n; i++) {
		ch = getState(i, idx);
		if (strchr(splitMap, ch) != NULL)	// see THINK Reference for defn. of strchr
			b->add(i);
	}
	if (b->contains(n-1))
		b->setComplement(n);
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getPattern
// ---------------------------------------------------------------------------------
// This function gets the pattern string for a particular site. This is stored
// in inPattern which should be allocated enough space (numTaxa*numSites).
void MEAlignment::getPattern(int inSite, char *outPattern)
{
	int i, n = getNumTaxa();
	char *P = outPattern;
	
	for (i = 0; i < n; i++) {
		getPattern(i, inSite, P);
		P += getNumStates();
	}
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getPattern
// ---------------------------------------------------------------------------------
// This function gets the pattern string for a particular site and a particular 
// taxon. This is stored in inPattern which should be allocated enought 
// space (numSites). 
void MEAlignment::getPattern(int inTaxon, int inSite, char *outPattern)
{
	char *P=outPattern;
	MESequence *seq=(MESequence*)getSequence(inTaxon);
			
	switch (alignmentType) {
		case kNucleotideAlignment:
			seq->getNucleotideSite(inSite, P);
			break;
		case kAminoAcidAlignment:
			seq->getAminoAcidSite(inSite, P);
			break;
		case kCodonAlignment:
			seq->getCodonSite(inSite, P);
			break;
		default:
			throw MEIncompatibleSequence();
	}
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getEmpiricalFrequencies
// ---------------------------------------------------------------------------------
// This function estimates the state frequencies for the alignment. This consists of
// one frequency for each state returned in the provided array (which should
// be allocated enough space).
double *MEAlignment::getEmpiricalFrequencies(double *inFrequencies)
{
	int i, j, k;
	double total, sum, x, difference;
	double *tempFreq;
	char *pattern;

	int nStates=getNumStates();
	int nTaxa=getNumTaxa();
	int nSites=getNumSites();
	
	double *freqs = getEqualFrequencies(inFrequencies);
		
	NewArray(tempFreq, double, nStates);
	NewArray(pattern, char, nStates*nTaxa);
	
	do {
		for (i=0; i<nStates; i++)
			tempFreq[i]=0.0;
			
		total=0.0;
		for (i=0; i<nTaxa; i++) {
			for (k=0; k<nSites; k++) {
				getPattern(i, k, pattern);
				sum=0.0;
				for (j=0; j<nStates; j++)
					sum+=freqs[j]*pattern[j];
					
				for (j=0; j<nStates; j++) {
					x=(freqs[j]*pattern[j])/sum;
					tempFreq[j]+=x;
					total+=x;
				}
			}
		}
		difference=0.0;
		for (i=0; i<nStates; i++) {
			difference+=fabs((tempFreq[i]/total)-freqs[i]);
			freqs[i]=tempFreq[i]/total;
		}
	} while (difference>kTolerance);
	
	delete[] tempFreq;
	delete[] pattern;

	return freqs;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::getEqualFrequencies
// ---------------------------------------------------------------------------------
// This function returns the equal state frequencies for the alignment. 
double *MEAlignment::getEqualFrequencies(double *inFrequencies)
{
	int nStates=getNumStates();
	
	double *freqs = inFrequencies;
	if (freqs == NULL)
		NewArray(freqs, double, nStates);
	
	for (int i=0; i<nStates; i++)
		freqs[i]=1.0/nStates;
		
	return freqs;
}
	
// ---------------------------------------------------------------------------------
//		 MEAlignment::setFeatureList
// ---------------------------------------------------------------------------------

void MEAlignment::setFeatureList(MEFeatureList *inFeatureList, bool isOwner)
{
	featureList=inFeatureList;
	if (featureList==NULL || isOwner)
		featureList->setOwner(this);
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::newFeatureList
// ---------------------------------------------------------------------------------

MEFeatureList *MEAlignment::newFeatureList()
{
	MEFeatureList *list;
	NewClass(list, MEFeatureList());
	return list;
}


// ---------------------------------------------------------------------------------
//		 MEAlignment::newFeature
// ---------------------------------------------------------------------------------

MEFeature *MEAlignment::newFeature()
{
	if (featureList==NULL)
		setFeatureList(newFeatureList(), true);

	return (MEFeature *)featureList->createObject();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getNumFeatures
// ---------------------------------------------------------------------------------

int MEAlignment::getNumFeatures()
{
	if (featureList==NULL)
		return 0;
		
	return featureList->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getFeature
// ---------------------------------------------------------------------------------

MEFeature* MEAlignment::getFeature(int index)
{
	ThrowIfNull(featureList);
	
	return (MEFeature *)featureList->getObject(index);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getFirstFeature
// ---------------------------------------------------------------------------------

MEFeature* MEAlignment::getFirstFeature()
{
	ThrowIfNull(featureList);
	
	return (MEFeature *)featureList->getFirst();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getNextFeature
// ---------------------------------------------------------------------------------

MEFeature* MEAlignment::getNextFeature()
{
	ThrowIfNull(featureList);
	
	return (MEFeature *)featureList->getNext();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::addFeature
// ---------------------------------------------------------------------------------

void MEAlignment::addFeature(MEFeature *inFeature)
{
	if (featureList==NULL)
		setFeatureList(newFeatureList(), true);

	featureList->add(inFeature);
	featureList->setSortType(kSortByPosition);
	featureList->sort();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::insertFeature
// ---------------------------------------------------------------------------------

void MEAlignment::insertFeature(int index, MEFeature *inFeature)
{
	if (featureList==NULL)
		setFeatureList(newFeatureList(), true);

	featureList->insert(index, inFeature);
	featureList->setSortType(kSortByPosition);
	featureList->sort();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::removeFeature
// ---------------------------------------------------------------------------------

MEFeature* MEAlignment::removeFeature(int index)
{
	return (MEFeature*)featureList->remove(index);
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::removeAllFeatures
// ---------------------------------------------------------------------------------

void MEAlignment::removeAllFeatures()
{
	featureList->removeAll();
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::readObject
// ---------------------------------------------------------------------------------

void MEAlignment::readObject(MEStream *stream)
{
	MEBasicAlignment::readObject(stream);
	featureList=(MEFeatureList*)stream->get("Features");
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::writeObject
// ---------------------------------------------------------------------------------

void MEAlignment::writeObject(MEStream *stream)
{
	MEBasicAlignment::writeObject(stream);
	stream->put("Features", featureList);	
}


