/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEBasicAlignment.cpp

				MEBasicAlignment Class
				
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEObject.h"
#include "METaxonList.h"
#include "MEBasicSequence.h"
#include "MEBasicAlignment.h"

#include "MEStream.h"

#pragma mark === MEBasicAlignment ===

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment
// ---------------------------------------------------------------------------------

MEBasicAlignment::MEBasicAlignment() : METaxonList()
{
	numSites=0;
	numWeights = 0;
	numSitesKnown=false;
	numStates=4;
	alignmentType=kUnknownAlignment;
	
	siteWeights=NULL;
}

// ---------------------------------------------------------------------------------
//		 ~MEBasicAlignment
// ---------------------------------------------------------------------------------

MEBasicAlignment::~MEBasicAlignment()
{
	if (siteWeights!=NULL)
		delete [] siteWeights;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::clone
// ---------------------------------------------------------------------------------

void MEBasicAlignment::clone(MEObject *src)
{
	METaxonList::clone(src);
	
	MEBasicAlignment *align=(MEBasicAlignment *)src;
	numSites=align->getNumSites();
	setAllSiteWeights(align->getSiteWeights());
	setAlignmentType(align->getAlignmentType());
}

// ---------------------------------------------------------------------------------
//		MEBasicAlignment::createObject
// ---------------------------------------------------------------------------------

MEObject* MEBasicAlignment::createObject()
{
	MEObject *obj;
	NewClass(obj, MEBasicSequence());
	return obj;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getNumSites
// ---------------------------------------------------------------------------------

int MEBasicAlignment::getNumSites()
{
	if (!numSitesKnown)
		numSites=getMaxNumSites();
		
	numSitesKnown=true;

	return numSites;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getNumSites
// ---------------------------------------------------------------------------------

// Returns the number of sites of the sequence at index.
int MEBasicAlignment::getNumSites(int inTaxon)
{
	return getSequence(inTaxon)->getNumSites();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getMaxNumSites
// ---------------------------------------------------------------------------------

int MEBasicAlignment::getMaxNumSites()
{
	int len, maxLen=0;
	
	int n=getNumSequences();
	for (int i=0; i<n; i++) {
		len=getNumSites(i);
		if (len>maxLen)
			maxLen=len;
	}
	
	return maxLen;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getMinNumSites
// ---------------------------------------------------------------------------------

int MEBasicAlignment::getMinNumSites()
{
	int n=getNumSequences();
	if (n==0)
		return 0;
		
	int len, minLen=getNumSites(0);
	for (int i=1; i<n; i++) {
		len=getNumSites(i);
		if (len<minLen)
			minLen=len;
	}
	
	return minLen;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::hasEqualNumSites
// ---------------------------------------------------------------------------------

bool MEBasicAlignment::hasEqualNumSites()
{
	return getNumSites()==getMinNumSites();
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getNumStates()
// ---------------------------------------------------------------------------------

// Returns the number of different states in the alignment.
int MEBasicAlignment::getNumStates()
{
	numStates=countNumStates();
		
	return numStates;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::countNumStates()
// ---------------------------------------------------------------------------------

// Counts the number of different states in the alignment.
int MEBasicAlignment::countNumStates()
{
	int i, j, n=0;
	MEBasicSequence *seq;
	int numSeqs=getNumSequences();
	
	for (i=0; i<numSeqs; i++) {
		seq=getSequence(i);
		j=seq->getNumStates();
		if (j>n)
			n=j;
	}
	
	return n;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::setAlignmentType
// ---------------------------------------------------------------------------------
// Alignment type specifies the preferred type of sequence in the alignment.

void MEBasicAlignment::setAlignmentType(AlignmentType inType)
{
	int i;
	MEBasicSequence *seq;

	int n=getNumSequences();
	for (i=0; i<n; i++) {
		seq=getSequence(i);
		seq->setAlignmentType(inType);
	}
	
	alignmentType=inType;
	numSitesKnown=false;
	
	switch (alignmentType) {
		case kNucleotideAlignment:
			numStates = 4; 
		break;
		case kAminoAcidAlignment:
			numStates = 20; 
		break;
		case kCodonAlignment:
			numStates = 64; 
		break;
		case kDiscreteAlignment:
		case kUnknownAlignment:
			// Allow all letters and digits
			numStates = 65;
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequence
// ---------------------------------------------------------------------------------

MEBasicSequence* MEBasicAlignment::getSequence(int inTaxon)
{
	return (MEBasicSequence *)getObject(inTaxon);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceLabel
// ---------------------------------------------------------------------------------
// By default this just returns the name. But this behaviour can be overridden...

char* MEBasicAlignment::getSequenceLabel(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getName();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceInfo
// ---------------------------------------------------------------------------------
// By default this just returns the description. But this behaviour can be overridden...

char* MEBasicAlignment::getSequenceInfo(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getDescription();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceName
// ---------------------------------------------------------------------------------

char* MEBasicAlignment::getSequenceName(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getName();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceAccession
// ---------------------------------------------------------------------------------

char* MEBasicAlignment::getSequenceAccession(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getAccession();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceDescription
// ---------------------------------------------------------------------------------

char* MEBasicAlignment::getSequenceDescription(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getDescription();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getFirstSequence
// ---------------------------------------------------------------------------------

MEBasicSequence* MEBasicAlignment::getFirstSequence()
{
	return (MEBasicSequence *)getFirst();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getNextSequence
// ---------------------------------------------------------------------------------

MEBasicSequence* MEBasicAlignment::getNextSequence()
{
	return (MEBasicSequence *)getNext();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::addSequence
// ---------------------------------------------------------------------------------

int MEBasicAlignment::addSequence(MEBasicSequence *inSequence)
{
	numSitesKnown=false;
	inSequence->setAlignmentType(alignmentType);
	return add(inSequence, inSequence->getOwner()==NULL);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::insertSequence
// ---------------------------------------------------------------------------------

void MEBasicAlignment::insertSequence(int inTaxon, MEBasicSequence *inSequence)
{
	numSitesKnown=false;
	inSequence->setAlignmentType(alignmentType);
	insert(inTaxon, inSequence, inSequence->getOwner()==NULL);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::removeSequence
// ---------------------------------------------------------------------------------

MEBasicSequence* MEBasicAlignment::removeSequence(int inTaxon)
{
	numSitesKnown=false;
	return (MEBasicSequence *)remove(inTaxon);
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::removeAll
// ---------------------------------------------------------------------------------

void MEBasicAlignment::removeAll()
{
	numSitesKnown=false;
	MEList::removeAll();
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getIndelRange
// ---------------------------------------------------------------------------------

// Gets the range of the indel (either a run of states or gaps) that contains
// the site 'site' in sequence 'taxon'.
void MEBasicAlignment::getIndelRange(int inTaxon, int inSite, int *outFirst, int *outLast)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getIndelRange(inSite, outFirst, outLast);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getVerticalIndelRange
// ---------------------------------------------------------------------------------

// Gets the vertical range of the indel (either a run of states or gaps) that contains
// the site 'site' in sequence 'taxon'. 
void MEBasicAlignment::getVerticalIndelRange(int inTaxon, int inSite, int *outFirst, int *outLast)
{
	int fst, lst;
	fst=lst=inTaxon;
	
	if (getState(inTaxon, inSite)=='-') {
		while (fst>0 && getState(fst-1, inSite)=='-')
			fst--;
		while (lst<numObjects-1 && getState(lst+1, inSite)=='-')
			lst++;
	} else {
		while (fst>0 && getState(fst-1, inSite)!='-')
			fst--;
		while (lst<numObjects-1 && getState(lst+1, inSite)!='-')
			lst++;
	}
	
	*outFirst=fst;
	*outLast=lst;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::appendSequence
// ---------------------------------------------------------------------------------

void MEBasicAlignment::appendSequence(int inTaxon, char *inSequence, int inLength)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->appendSequence(inSequence, inLength);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::appendSequence
// ---------------------------------------------------------------------------------

void MEBasicAlignment::appendSequence(int inTaxon, MEBasicSequence *inSequence)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->appendSequence(inSequence);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::appendSite
// ---------------------------------------------------------------------------------

void MEBasicAlignment::appendSite(int inTaxon, char inSite)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->appendSite(inSite);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::insertSequence
// ---------------------------------------------------------------------------------

void MEBasicAlignment::insertSequence(int inTaxon, char *inSequence, int inLength, int inSite)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->insertSequence(inSequence, inLength, inSite);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::insertSequence
// ---------------------------------------------------------------------------------

void MEBasicAlignment::insertSequence(int inTaxon, MEBasicSequence *inSequence, int inSite)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->insertSequence(inSequence, inSite);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::insertGap
// ---------------------------------------------------------------------------------

void MEBasicAlignment::insertGap(int inTaxon, int inSite, int inLength)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->insertGap(inSite, inLength);
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::removeSites
// ---------------------------------------------------------------------------------

void MEBasicAlignment::removeSites(int inTaxon, int inSite, int inLength)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->removeSites(inSite, inLength);
	numSitesKnown=false;
}


// ---------------------------------------------------------------------------------
//		 PPAlignment::countInitialGap
// ---------------------------------------------------------------------------------
// Returns the length of any initial gap in taxon 'inTaxon'.

int MEBasicAlignment::countInitialGap(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	int gap=0;
	int n=seq->getNumSites();
	char ch=getState(inTaxon, gap);
	
	while (ch=='-' && gap<n) {
		gap++;
		ch=getState(inTaxon, gap);
	}
	
	return gap;
}


// ---------------------------------------------------------------------------------
//		 PPAlignment::removeInitialGaps
// ---------------------------------------------------------------------------------
// Removes any initial gap in taxon 'inTaxon'.

int MEBasicAlignment::removeInitialGap(int inTaxon)
{
	int gap=countInitialGap(inTaxon);
	if (gap>0)
		removeSites(inTaxon, 0, gap);
	return gap;
}


// ---------------------------------------------------------------------------------
//		 PPAlignment::trimGapsFromEnd
// ---------------------------------------------------------------------------------

void MEBasicAlignment::trimGapsFromEnd()
{
	MEBasicSequence *seq=(MEBasicSequence*)getFirst();
	while (seq!=NULL) {
		seq->trimGapsFromEnd();
		seq=(MEBasicSequence*)getNext();
	}
	
	numSitesKnown=false;
}


// ---------------------------------------------------------------------------------
//		 PPAlignment::removeGaps
// ---------------------------------------------------------------------------------

void MEBasicAlignment::removeGaps()
{
	MEBasicSequence *seq=getFirstSequence();
	while (seq!=NULL) {
		seq->removeGaps();
		seq=getNextSequence();
	}
	
	numSitesKnown=false;
}


// ---------------------------------------------------------------------------------
//		 PPAlignment::removeGaps
// ---------------------------------------------------------------------------------

void MEBasicAlignment::removeGaps(int inTaxon)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->removeGaps();
	
	numSitesKnown=false;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getState
// ---------------------------------------------------------------------------------

// Returns the state of site 'inSite' in sequence 'inTaxon'.
char MEBasicAlignment::getState(int inTaxon, int inSite)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getSite(inSite);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getStateNumber
// ---------------------------------------------------------------------------------

// Returns the state number of site 'site' in sequence 'taxon'.
char MEBasicAlignment::getStateNumber(int inTaxon, int inSite)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	
	switch (seq->getSequenceType()) {
		kDNASequence:
		kRNASequence:
			return kNucleotideNumbers[seq->getSite(inSite)];
		kAminoAcidSequence:
			return kAminoAcidNumbers[seq->getSite(inSite)];
		default:
			throw MEIncompatibleSequence();
	}
	return 0;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::setState
// ---------------------------------------------------------------------------------

// Sets the state of site 'inSite' in sequence 'taxon'.
void MEBasicAlignment::setState(int inTaxon, int inSite, char inState)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	seq->setSite(inSite, inState);
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSiteString
// ---------------------------------------------------------------------------------

// Returns the states of site 'inSite' across all the taxa.
void MEBasicAlignment::getSiteString(int inSite, char *outSiteString)
{
	int i, n=getNumSequences();
	char *P=outSiteString;
	
	for (i=0; i<n; i++) {
		*P=getState(i, inSite);
		P++;
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::setSiteString
// ---------------------------------------------------------------------------------

// Sets the state of site 'inSite' across all the taxa.
void MEBasicAlignment::setSiteString(int inSite, char *inSiteString)
{
	int i, n=getNumSequences();
	char *P=inSiteString;
	
	for (i=0; i<n; i++) {
		setState(i, inSite, *P);
		P++;
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::appendSiteString
// ---------------------------------------------------------------------------------

// Appends a site from 'inSiteString' to each of the taxa.
void MEBasicAlignment::appendSiteString(char *inSiteString)
{
	char *P=inSiteString;
	MEBasicSequence *seq=getFirstSequence();
	while (seq!=NULL) {
		seq->appendSite(*P);
		P++;
		seq=getNextSequence();
	}
	numSitesKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getBinarySiteString
// ---------------------------------------------------------------------------------

// Returns the states of site 'site' across all the taxa, mapped into binary states.
void MEBasicAlignment::getBinarySiteString(int inSite, char *inStateMap, char *outSiteString)
{
	int i, n=getNumSequences();
	char ch;
	char *P=outSiteString;
	
	for (i=0; i<n; i++) {
		ch=getState(i, inSite);
		if (strchr(inStateMap, ch) != NULL)
			*P='1';
		else
			*P='0';
		P++;
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getPattern
// ---------------------------------------------------------------------------------

void MEBasicAlignment::getPattern(int inSite, char *outPattern)
{
	getSiteString(inSite, outPattern);
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getIsMismatch
// ---------------------------------------------------------------------------------

bool MEBasicAlignment::getIsMismatch(int inTaxon, int inSite, char inState)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	return seq->getIsMismatch(inSite, inState);
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::newSequence
// ---------------------------------------------------------------------------------

// Returns a new sequence of an appropriate type. 
MEBasicSequence* MEBasicAlignment::newSequence(SequenceType inType)
{
	MEBasicSequence* seq;
	NewClass(seq, MEBasicSequence());
	seq->setSequenceType(inType);
	return seq;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getStringLength
// ---------------------------------------------------------------------------------

// Returns the required string length for sequence.
int MEBasicAlignment::getStringLength(int inTaxon)
{
	return getSequence(inTaxon)->getNumSites();
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getStringLength
// ---------------------------------------------------------------------------------

// Returns the required string length for sequence of length inLength.
// This may seem strange but in MEAlignment, this may depend on whether
// the sequences are in a triplet format (codons, aa triplets). inTaxon
// is ignored.

int MEBasicAlignment::getStringLength(int /*inTaxon*/, int inNumSites)
{
	if (alignmentType==kCodonAlignment)
		return inNumSites*3;
		
	return inNumSites;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceString
// ---------------------------------------------------------------------------------

// Copies the sequence at inTaxon into a string. Must be long enough
// plus 1 extra for the null termination.
void MEBasicAlignment::getSequenceString(int inTaxon, char *outStr)
{
	getSequenceString(inTaxon, outStr, 0);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSequenceString
// ---------------------------------------------------------------------------------

// Copies the sequence at inTaxon into a string. Must be long enough
// plus 1 extra for the null termination. If inNumSites is less than the
// number of sites in this sequence then the whole sequence is written. 
// If greater then the output is padded with gaps.

void MEBasicAlignment::getSequenceString(int inTaxon, char *outStr, int inNumSites)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	int len=seq->getNumSites();
	
	char *str=seq->getSequence();
	memcpy(outStr, str, len);
	
	char *P=outStr+len;
	while (len<inNumSites) {
		*P='-';
		P++;
		len++;
	}
	*P='\0';
}

// ---------------------------------------------------------------------------------
//		 MEAlignment::getSequenceString
// ---------------------------------------------------------------------------------

// Copies part of the sequence at inTaxon into a string. Must be long enough.
// This will be translated into the appropriate alignment type. If 
// inNumSites is less than the number of sites in this sequence then 
// then the whole sequence is written. If greater then the output 
// is padded with gaps.

void MEBasicAlignment::getSequenceString(int inTaxon, char *outStr, int inStartSite, int inNumSites)
{
	MEBasicSequence *seq=getSequence(inTaxon);
	char *str = seq->getSequence();
	int len=seq->getNumSites();
	
	char *P=outStr;
	if (inStartSite < 0) {
		memset(P, '-', -inStartSite);
		P -= inStartSite;
		inNumSites += inStartSite;
		inStartSite = 0;
	}
	
	int endSite = inStartSite + inNumSites;
	if (endSite > len) {
		memcpy(P, str + inStartSite, len - inStartSite);
		P += len - inStartSite;
		memset(P, '-', endSite - len);
		P += endSite - len;
	} else {
		memcpy(P, str + inStartSite, inNumSites);
		P += inNumSites;
	}
	*P='\0';
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSiteWeight
// ---------------------------------------------------------------------------------

double MEBasicAlignment::getSiteWeight(int site)
{
	if (siteWeights==NULL)
		return 1.0;
		
	TestBounds(site, numSites);
		
	return siteWeights[site];
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getSiteWeights
// ---------------------------------------------------------------------------------

double *MEBasicAlignment::getSiteWeights()
{
	return siteWeights;
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getMaxSiteWeight
// ---------------------------------------------------------------------------------

double MEBasicAlignment::getMaxSiteWeight()
{
	double d, e;
	int n=getNumSites();
	
	d=getSiteWeight(0);
	for (int i = 1; i < n; i++) {
		e = getSiteWeight(i);
		if (e > d)
			d = e;
	}
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::getMinSiteWeight
// ---------------------------------------------------------------------------------

double MEBasicAlignment::getMinSiteWeight()
{
	double d, e;
	int n=getNumSites();
	
	d=getSiteWeight(0);
	for (int i = 1; i < n; i++) {
		e = getSiteWeight(i);
		if (e < d)
			d = e;
	}
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::setSiteWeight
// ---------------------------------------------------------------------------------

void MEBasicAlignment::setSiteWeight(int site, double weight)
{
	resizeSiteWeights();
	
	TestBounds(site, numWeights);
		
	siteWeights[site]=weight;
}	

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::setAllSiteWeights
// ---------------------------------------------------------------------------------

void MEBasicAlignment::setAllSiteWeights(double weight)
{
	resizeSiteWeights();

	int i;
	double *P=siteWeights;
	for (i=0; i<numSites; i++) {
		*P=weight;
		P++;
	}
}
	
// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::setAllSiteWeights
// ---------------------------------------------------------------------------------

void MEBasicAlignment::setAllSiteWeights(double *inWeights)
{
	if (inWeights == NULL)
		return;
		
	resizeSiteWeights();

	int i;
	double *P=siteWeights;
	double *Q=inWeights;
	for (i=0; i<numSites; i++) {
		*P=*Q;
		P++;
		Q++;
	}
}	

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::resizeSiteWeights()
// ---------------------------------------------------------------------------------

// Update the array containing the site weights
void MEBasicAlignment::resizeSiteWeights()
{
	if (siteWeights==NULL || getNumSites() > numWeights) {
		int k = numWeights;
		while (k < numSites)
			k += 100;
		double *tempWeights;
		NewArray(tempWeights, double, k);

		if (siteWeights!=NULL) {
			for (int i = 0; i < numWeights; i++)
				tempWeights[i] = siteWeights[i];
			delete [] siteWeights;
		}
		
		siteWeights = tempWeights;
		numWeights = k;
	}
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::inferSequenceType
// ---------------------------------------------------------------------------------

SequenceType MEBasicAlignment::inferSequenceType()
{
	SequenceType seqType;
	MEBasicSequence *sequence;
	int nTaxa=getNumSequences();
	
	if (nTaxa==0)
		return kUnknownSequence;
	
	int nNuc, nT, nU, nAA;
	double fNuc;
		
	nNuc=nT=nU=nAA=0;
	for (int i=0; i<nTaxa; i++) {
		sequence=getSequence(i);
		nNuc+=sequence->countStates(kNucleotides);
		nT+=sequence->countStates("T");
		nU+=sequence->countStates("U");
		nAA+=sequence->countStates(kAminoAcids);
	}

	fNuc=((double)nNuc)/(nAA);

	if (fNuc>=0.95) { 	// Unlikely that 95% of AAs would be ACGT etc.
		if (nT>nU)		// More T's than U's so make DNA mode
			seqType=kDNASequence;
		else			// else RNA mode
			seqType=kRNASequence;
	} else
		seqType=kAminoAcidSequence;
	
	return seqType;
}


// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::sortBySiteWeights
// ---------------------------------------------------------------------------------

void MEBasicAlignment::sortBySiteWeights(bool down)
{
	// set up the ShellSort: (see Algorithms, by Robert Sedgewick, Addison Wesley, pg 108)
	char
		ch;
	double
		d;
	int
		h = 1, 
		i, j, 
		jh,
		k = getNumSites(),
		t;
	do {
		h = 3*h + 1;
	} while (h <= k);
	if (down) {
		do {
			h = (div(h, 3)).quot;	// see THINK Reference for definition of div
			for (i = h; i < k; i++) {
				for (j = i-h; 
						getSiteWeight(j+h) > getSiteWeight(j); 
						j -= h) {
					jh = j + h;
					// move the characters one by one:
					for (t = 0; t < numObjects; t++) {
						ch = getState(t, j);
						setState(t, j, getState(t, jh));
						setState(t, jh, ch);
					}
					// swap the weights:
					d = getSiteWeight(j);
					setSiteWeight(j, getSiteWeight(jh));
					setSiteWeight(jh, d);
					if (j < h)
						break;
				}
			}
		} while (h > 1);
	}
	else {
		do {
			h = (div(h, 3)).quot;	// see THINK Reference for definition of div
			for (i = h; i < k; i++) {
				for (j = i-h; 
						getSiteWeight(j+h) < getSiteWeight(j); 
						j -= h) {
					jh = j + h;
					// move the characters one by one:
					for (t = 0; t < numObjects; t++) {
						ch = getState(t, j);
						setState(t, j, getState(t, jh));
						setState(t, jh, ch);
					}
					// swap the weights:
					d = getSiteWeight(j);
					setSiteWeight(j, getSiteWeight(jh));
					setSiteWeight(jh, d);
					if (j < h)
						break;
				}
			}
		} while (h > 1);
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::readObject
// ---------------------------------------------------------------------------------

void MEBasicAlignment::readObject(MEStream *stream)
{
	METaxonList::readObject(stream);
	
	numSitesKnown=stream->get("NumSites", &numSites);
	
	int i;
	if (stream->get("Type", &i, kAlignmentTypeNames, numAlignmentTypes))
		alignmentType=(AlignmentType)i;
}

// ---------------------------------------------------------------------------------
//		MEBasicAlignment::readData
// ---------------------------------------------------------------------------------
// Overridden in order to set the alignmentType for all the sequences
void MEBasicAlignment::readData(MEStream *stream)
{
	MEList::readData(stream);
	
	setAlignmentType(alignmentType);
}

// ---------------------------------------------------------------------------------
//		 MEBasicAlignment::writeObject
// ---------------------------------------------------------------------------------

void MEBasicAlignment::writeObject(MEStream *stream)
{
	METaxonList::writeObject(stream);

	stream->put("NumSites", getNumSites());
	stream->put("Type", kAlignmentTypeNames[alignmentType]);
}


