/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEBasicSequence.cpp

				MEBasicSequence Class
				
 ******************************************************************************/

// MEBasicSequence is derived from METaxon, and adds an accession number and the actual
// sequence to the fields. This acts as a base for MESequence and its descendents but
// can also be used in its own right (usually with MEBasicAlignment) if simplicity is 
// required.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"
#include "MEStream.h"
#include "METaxon.h"
#include "MEBasicSequence.h"

#pragma mark === MEBasicSequence ===

// ---------------------------------------------------------------------------------
//		 MEBasicSequence
// ---------------------------------------------------------------------------------

MEBasicSequence::MEBasicSequence() :
	METaxon()
{
	desc=acc=seq=NULL;
	length=0;
	capacity=0;
	
	seqType = kUnknownSequence;
	alignmentType = kUnknownAlignment;

	numSites = 0;
	numSitesKnown = false;
		
	numStates = 0;
	
	seqChanged = true;
}

// ---------------------------------------------------------------------------------
//		 ~MEBasicSequence
// ---------------------------------------------------------------------------------

MEBasicSequence::~MEBasicSequence()
{
	if (acc!=NULL)
		delete[] acc;
	if (seq!=NULL)
		delete[] seq;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::clone
// ---------------------------------------------------------------------------------

void MEBasicSequence::clone(MEObject *src)
{
	METaxon::clone(src);
	
	MEBasicSequence *seq=(MEBasicSequence *)src;
	setAccession(seq->getAccession());
	setSequenceType(seq->getSequenceType());
	setSequence(seq->getSequence(), seq->getLength());
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::cloneExceptSequence
// ---------------------------------------------------------------------------------

void MEBasicSequence::cloneExceptSequence(MEBasicSequence *src)
{
	METaxon::clone(src);
	
	setAccession(src->getAccession());
	setSequenceType(src->getSequenceType());
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getNumStates()
// ---------------------------------------------------------------------------------

// Returns the number of different states in the sequence. In MEBasicSequence
// this is the same as the number of observed states.
int MEBasicSequence::getNumStates()
{
	return numStates;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::setAccession
// ---------------------------------------------------------------------------------

// Sets the accession field. Allocates memory if needed. An inAccession which equals 
// NULL will set acc to NULL without deleting the array.

void MEBasicSequence::setAccession(char *inAccession)
{
	if (inAccession!=NULL) {	
		if (acc==NULL)
			NewArray(acc, char, MAX_ACC_LEN+1);
			
		strncpy(acc, inAccession, MAX_ACC_LEN);
		acc[MAX_ACC_LEN]='\0';
	} else 
		acc=NULL;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::setSequenceType
// ---------------------------------------------------------------------------------
void MEBasicSequence::setSequenceType(SequenceType inSequenceType)
{ 
	seqType=inSequenceType; 
	numSitesKnown = false;
	seqChanged = true;
}


// ---------------------------------------------------------------------------------
//		 MESequence::setAlignmentType
// ---------------------------------------------------------------------------------

void MEBasicSequence::setAlignmentType(AlignmentType inType)
{	
	alignmentType = inType;
	numSitesKnown = false;
	seqChanged = true;
}


// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getAccession
// ---------------------------------------------------------------------------------

char* MEBasicSequence::getAccession()
{
	return acc;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getAccession
// ---------------------------------------------------------------------------------

void MEBasicSequence::getAccession(char *inAccession)
{
	if (acc!=NULL)
		strcpy(inAccession, acc);
	else
		inAccession[0]='\0';
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::setSite
// ---------------------------------------------------------------------------------

// Sets the state of site 'inSite'.
void MEBasicSequence::setSite(int inSite, char state)
{
	TestBounds(inSite, length);
	seq[inSite]=state;
	seqChanged = true;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getSite
// ---------------------------------------------------------------------------------

// Returns the state of site 'inSite'.
char MEBasicSequence::getSite(int inSite)
{
	ThrowIf(inSite<0);

	if (inSite>=length)
		return '-';
	else
		return seq[inSite];
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getSites
// ---------------------------------------------------------------------------------

// Copies a portion of the sequence into outStr. If inStart=0 and inLength=0
// then the whole sequence is copied. If any of the required portion lies
// outside the actual sequence then the copy is padded with gaps.
void MEBasicSequence::getSites(int inStartSite, int inNumSites, char *outStr)
{
	char *Q=outStr;
	
	int seqPos=inStartSite;
	while (seqPos<0) {
		*Q='-';
		Q++;
		seqPos++;
		inNumSites--;
	}
	
	int seqLen=length-seqPos;
	if (seqLen < 0)
		seqLen = 0;
	if (inNumSites>0 && inNumSites<seqLen)
		seqLen=inNumSites;
		
	if (seqLen > 0) {
		strncpy(Q, seq+seqPos, seqLen);
		Q+=seqLen;
	}
	
	while (seqLen<inNumSites) {
		*Q='-';
		Q++;
		seqLen++;
	}
	*Q='\0';
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getIsMismatch
// ---------------------------------------------------------------------------------

bool MEBasicSequence::getIsMismatch(int site, char state)
{
	TestBounds(site, length);
	return seq[site]!=state;
}


// ---------------------------------------------------------------------------------
//		 MEBasicSequence::setSequence
// ---------------------------------------------------------------------------------

// Sets the sequence and the sequence buffer length at the same time. Calls setLength
// first to allocate memory.
void MEBasicSequence::setSequence(char *inSequence, int inLength)
{
	if (inSequence!=NULL && inLength>0) {
		setLength(inLength);
			
		memcpy(seq, inSequence, length);
	} else {
		seq=NULL;
		length=0;
		capacity=0;
		numSitesKnown = false;
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getSequence
// ---------------------------------------------------------------------------------

char* MEBasicSequence::getSequence()
{
	return seq;
}


// ---------------------------------------------------------------------------------
//		 MEBasicSequence::setLength
// ---------------------------------------------------------------------------------

void MEBasicSequence::setLength(int inLength)
{
	if (inLength>capacity) {
		while (inLength>capacity)
			capacity+=SEQ_INCREMENT_SIZE;
		
		char *newSeq;
		NewArray(newSeq, char, capacity+1);
		newSeq[0] = '\0';
		
		if (seq!=NULL && length>0) {
			memcpy(newSeq, seq, length);
			delete[] seq;
		}
		
		seq=newSeq;
	}
	length=inLength;
	numSitesKnown = false;
	seqChanged = true;
}


// ---------------------------------------------------------------------------------
//		 MEBasicSequence::appendSequence
// ---------------------------------------------------------------------------------

// Appends inSequence to the sequence. This may have to reallocate the sequence
// to make space.
void MEBasicSequence::appendSequence(char *inSequence, int inLength)
{
	ThrowIfNull(inSequence);
	
	setLength(inLength+length);
	strcat(seq, inSequence);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::appendSequence
// ---------------------------------------------------------------------------------

// Appends inSequence to the sequence. This may have to reallocate the sequence
// to make space.
void MEBasicSequence::appendSequence(MEBasicSequence *inSequence)
{
	ThrowIfNull(inSequence);
	appendSequence(inSequence->getSequence(), inSequence->getLength());
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::appendSite
// ---------------------------------------------------------------------------------

// Appends inSite to the sequence. This may have to reallocate the sequence
// to make space.
void MEBasicSequence::appendSite(char inSite)
{
	setLength(length+1);
	seq[length-1]=inSite;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::insertSequence
// ---------------------------------------------------------------------------------

// Inserts inSequence to the sequence at position inSite. This may have to 
// reallocate the sequence to make space.
void MEBasicSequence::insertSequence(char *inSequence, int inLength, int inSite)
{
	ThrowIfNull(inSequence);
	
	if (inSite >= length) {
		int oldLen = length;
		setLength(inSite + inLength);
		memset(seq + oldLen, '-', inSite - oldLen);
	} else if (inSite < length) {
		int len=length-inSite;
		setLength(inLength+length);
		memmove(seq+inSite+inLength, seq+inSite, len);
	} else {
		setLength(inLength+length);
	}
	memcpy(seq+inSite, inSequence, inLength);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::insertSequence
// ---------------------------------------------------------------------------------

// Inserts inSequence to the sequence at position inSite. This may have to 
// reallocate the sequence to make space.
void MEBasicSequence::insertSequence(MEBasicSequence *inSequence, int inSite)
{
	ThrowIfNull(inSequence);
	insertSequence(inSequence->getSequence(), inSequence->getLength(), inSite);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::insertGap
// ---------------------------------------------------------------------------------

// Inserts a gap of inLength to the sequence at position inSite. This may have to 
// reallocate the sequence to make space.
void MEBasicSequence::insertGap(int inSite, int inLength)
{
	int len = length - inSite;
	if (len < 0) {
		int oldLength = length;
		
		len = inSite - length + inLength;
		setLength(len + length);
		
		memset(seq + oldLength, '-', len);
	} else {
		setLength(inLength + length);
		if (len > 0)
			memmove(seq + inSite + inLength, seq+inSite, len);
		memset(seq+inSite, '-', inLength);
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::removeSites
// ---------------------------------------------------------------------------------

void MEBasicSequence::removeSites(int inSite, int inLength)
{
	if ( (inSite+inLength)>length )
		inLength=length - inSite;
		
	if (inLength > 0) {
		int len=length-(inSite + inLength) + 1;
		if (len>0)
			memmove(seq+inSite, seq + inSite + inLength, len);
		setLength(length-inLength);
	}
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::trimGapsFromEnd
// ---------------------------------------------------------------------------------

void MEBasicSequence::trimGapsFromEnd()
{
	int site=length;
	char *P=seq+site-1;
	
	while (site>0 && *P=='-') {
		P--;
		site--;
	}
	setLength(site);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::removeGaps
// ---------------------------------------------------------------------------------

// Removes gaps from the sequence. 
void MEBasicSequence::removeGaps()
{
	removeGaps(0, getLength() - 1);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::removeGaps
// ---------------------------------------------------------------------------------

// Removes gaps from the sequence between sites 'inFirstSite' and 'inLastSite', inclusive. 
void MEBasicSequence::removeGaps(int inFirstSite, int inLastSite)
{
	char* ahead = seq+inFirstSite;
	char* behind = seq+inFirstSite;
	int numGaps = 0;
	int n = getLength();
	if (inLastSite > n)
		inLastSite = n;
		
	for (int i = inFirstSite; i <= inLastSite; i++) {
		if (*ahead != '-') {
			*behind = *ahead;
			behind++;
		} else
			numGaps++;
		ahead++;
	}
	
	for (int i = inLastSite + 1; i < n; i++) {
		*behind = *ahead;
		behind++;
		ahead++;
	}
	setLength(getLength() - numGaps);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::getIndelRange
// ---------------------------------------------------------------------------------

// Gets the range of the indel (either a run of states or gaps) that contains
// the site given.
void MEBasicSequence::getIndelRange(int site, int *first, int *last)
{
	TestBounds(site, length);

	int fst, lst;
	fst=lst=site;
	
	if (getSite(site)=='-') {
		while (fst>0 && getSite(fst-1)=='-')
			fst--;
		while (lst<length-1 && getSite(lst+1)=='-')
			lst++;
	} else {
		while (fst>0 && getSite(fst-1)!='-')
			fst--;
		while (lst<length-1 && getSite(lst+1)!='-')
			lst++;
	}
	
	*first=fst;
	*last=lst;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::filterSequence
// ---------------------------------------------------------------------------------

// Finds characters in sequence that match those in src and changes them to dest.
// Returns the number of matches found.
int MEBasicSequence::filterSequence(const char *src, char dest)
{
	char *P;
	int n=0;
	
	P=seq;
	while (*P!='\0') {
		if (strchr(src, *P)!=NULL) {
			*P=dest;
			n++;
		}
		P++;
	}
	seqChanged = true;

	return n;
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::countState
// ---------------------------------------------------------------------------------

// Returns the number of inState characters in sequence.
int MEBasicSequence::countState(char inState)
{
	char *P;
	int n=0;
	
	P=seq;
	while (*P!='\0') {
		if (*P==inState)
			n++;
		P++;
	}
	
	return n;
}


// ---------------------------------------------------------------------------------
//		 MEBasicSequence::countStates
// ---------------------------------------------------------------------------------

// Returns the number of characters in sequence that match those in inStates.
int MEBasicSequence::countStates(const char *inStates)
{
	char *P;
	int n=0;
	
	P=seq;
	while (*P!='\0') {
		if (strchr(inStates, *P)!=NULL)
			n++;
		P++;
	}
	
	return n;
}


// ---------------------------------------------------------------------------------
//		 MEBasicSequence::filterMatches
// ---------------------------------------------------------------------------------

// Finds characters in the sequence that match those in match and changes them to the
// the homologous characters in src.
// Returns the number of matches found.
int MEBasicSequence::filterMatches(const char *match, const char *src)
{
	int n=0;
	char *P;
	const char *Q;
	
	Q=src;
	P=seq;
	while (*P!='\0' && *Q!='\0') {
		if (strchr(match, *P)!=NULL) {
			*P=*Q;
			n++;
		}
		P++;
		Q++;
	}
	seqChanged = true;

	return n;
}
// ---------------------------------------------------------------------------------
//		 MEBasicSequence::readObject
// ---------------------------------------------------------------------------------

void MEBasicSequence::readObject(MEStream *stream)
{
	METaxon::readObject(stream);
	
	char inAcc[MAX_ACC_LEN+1];
	if (stream->get("Accession", inAcc, MAX_ACC_LEN))
		setAccession(inAcc);
	
	seqType=kUnknownSequence;
	
	int i;
	stream->get("Type", &i, kSequenceTypeNames, numSequenceTypes);
	seqType=(SequenceType)i;
	
	int len=0;
	stream->get("Length", &len);
	setLength(len);
	
	stream->get("Sequence", seq, len);
}

// ---------------------------------------------------------------------------------
//		 MEBasicSequence::writeObject
// ---------------------------------------------------------------------------------

void MEBasicSequence::writeObject(MEStream *stream)
{
	METaxon::writeObject(stream);

	stream->put("Accession", acc);
	stream->put("Type", kSequenceTypeNames[seqType]);
	stream->put("Length", length);
	stream->put("Sequence", seq, length);
}


