/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEBinaryStream.cpp

				MEBinaryStream Class
				
 ******************************************************************************/

// MEBinaryStream overrides MEStream to provide a binary based format for
// streaming objects. This is an abstract class and must be overridden to
// write to a particular type of file or memory.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"
#include "MEUndead.h"
#include "MEObject.h"
#include "MEList.h"
#include "MEIntArray.h"
#include "MEDoubleArray.h"
#include "MELink.h"
#include "METree.h"
#include "MEBinaryStream.h"

#pragma mark === MEBinaryStream ===

// ---------------------------------------------------------------------------------
//		 MEBinaryStream
// ---------------------------------------------------------------------------------

MEBinaryStream::MEBinaryStream() : MEStream()
{
}

// ---------------------------------------------------------------------------------
//		 ~MEBinaryStream
// ---------------------------------------------------------------------------------

MEBinaryStream::~MEBinaryStream()
{
}


#pragma mark === Low Level Reading ===

// All are abstract functions

#pragma mark === Data Object Reading ===
// These functions are used for the streaming of data objects.

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getLabel
// ---------------------------------------------------------------------------------


bool MEBinaryStream::getLabel(char *label)
{
	char str[256];
	
	readString(str);
	
	// is it the label we want?
	if (strcmp(label, str)!=0)
		throw MEStreamError("Field missing in binary stream", label);
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

bool MEBinaryStream::get(char *label, int *i)
{
	if (getLabel(label)) {
		*i=readInteger();
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

bool MEBinaryStream::get(char *label, double *d)
{
	if (getLabel(label)) {
		*d=readDouble();
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

bool MEBinaryStream::get(char *label, bool *b)
{
	if (getLabel(label)) {
		*b=(bool)readChar();
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

bool MEBinaryStream::get(char *label, char *ch)
{
	if (getLabel(label)) {
		*ch=readChar();
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

bool MEBinaryStream::get(char *label, char *str, int /*maxLen*/)
{
	if (getLabel(label)) {
		readString(str);
		if (strlen(str)==0)
			return false;
			
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

bool MEBinaryStream::get(char *label, ObjectID *id)
{
	if (getLabel(label)) {
		id->chars[0]=readChar();
		id->chars[1]=readChar();
		id->chars[2]=readChar();
		id->chars[3]=readChar();
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

// Find the index of the parameter in stringList and returns it in d.

bool MEBinaryStream::get(char *label, int *d, const char **stringList, const int numStrings)
{
	if (getLabel(label)) {
		char str[256];
		readString(str);
		
		int n=-1;
		for (int i=0; i<numStrings; i++)
			if (strcmp(str, stringList[i])==0)
				n=i;
				
		if (n>-1)
			*d=n;
			
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::get
// ---------------------------------------------------------------------------------

MEObject * MEBinaryStream::get(char *label)
{
	MEObject *obj=NULL;
	
	if (getLabel(label)) {
		int ref=readInteger();
		if (ref==0) { // This object is given in full
			obj=getObject();
		} else if (ref>1) {
			obj=getObjectRef(ref-1);
		} // otherwise assume obj is null
	}
	
	return obj;
}
		
// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getList
// ---------------------------------------------------------------------------------

void MEBinaryStream::getList(MEList *inList)
{
	int count=readInteger();
	
	if (count>0) {
		MEObject *obj;
		int ref;
		for (int i=0; i<count; i++) {
			ref=readInteger();
			if (ref==0) { // This object is given in full
				obj=getObject();
			} else
				obj=getObjectRef(ref-1);
			inList->add(obj);	
		}
	}
}


// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getIntArray
// ---------------------------------------------------------------------------------

void MEBinaryStream::getIntArray(MEIntArray *inList)
{
	int count=readInteger();
	
	if (count>0) {
		int val;
		for (int i=0; i<count; i++) {
			val=readInteger();
			inList->add(val);	
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getDoubleArray
// ---------------------------------------------------------------------------------

void MEBinaryStream::getDoubleArray(MEDoubleArray *inList)
{
	int count=readInteger();
	
	if (count>0) {
		double val;
		for (int i=0; i<count; i++) {
			val=readDouble();
			inList->add(val);	
		}
	}
}


#pragma mark === Tree Reading ===

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getTree
// ---------------------------------------------------------------------------------

// This reads in a tree. It can cope with white-space but otherwise conform
// to a strict format. Tips should be numbers, and branch lengths should be
// present (even if just 0.0).

void MEBinaryStream::getTree(METree *inTree)
{
	MELink *link, *link1;
	
	int n=readInteger(); // number of branches in polytomy
	link=getNode(inTree, n);

	// prune out the temporary MELink that pointed to the root
	link1=link;
	while (link->next!=link1) 
		link=link->next;
		
	link->next=link1->next;
	inTree->removeLink(link1);
	
	// move on to the first branch in the node
	link=link->next;
	link->setIsStem(true);
	
	if (link->getNumBranches()<3) {
		// if the root node has less than 3 branches then this node is
		// removed. This makes the tree unrooted but we store the position
		// of the root by setting METree's root1 and root2 to point to the
		// nodes either side of the root.
		inTree->setIsRooted(true);
		
		// Set the tree's root pointers
		MELink* root1=link->branch;
		MELink* root2=link->next->branch;
		inTree->setRoot1(root1);
		inTree->setRoot2(root2);
		
		// Set the tree's root branch lengths
		inTree->setRootDistances(inTree->getBranchLength(root1->branchNo),
									inTree->getBranchLength(root2->branchNo));
		
		// Set the nodes either side of the root to point to each other
		root1->branch=root2;
		root2->branch=root1;
		
		// Remove root2's branch length from the tree's array
		inTree->removeBranchLength(root2->branchNo);
		
		// Set the new branches branchNo and the corresponding branchLength
		// This just uses root1's old branch length slot and sets it to the
		// sum of the two old branch lengths
		root2->branchNo=root1->branchNo;
		inTree->setBranchLength(root1->branchNo, 
								inTree->getRootDistance1() + inTree->getRootDistance2());
		
		inTree->removeLink(link->next);
		inTree->removeLink(link);
		
		inTree->setNumInternalNodes(inTree->getNumInternalNodes()-1);
		inTree->setNumNodes(inTree->getNumNodes()-1);
	} else {
		inTree->setIsRooted(false);
		inTree->setRoot1(link);
		inTree->setRoot2(link->branch);
	}
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getTip
// ---------------------------------------------------------------------------------
 
MELink *MEBinaryStream::getTip(METree *inTree)
{
	int tipNo=readInteger();
	return inTree->newTip(tipNo-1);
}


// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getNode
// ---------------------------------------------------------------------------------
 
//	First this function creates a new MELink that will point back up at
//	the ancestoral node. Then it works through each branch in a node and 
//	calls readBranch to read them. By passing the root MELink, each branch
//	can hook on to the linked list and form a polytomy.

MELink *MEBinaryStream::getNode(METree *inTree, int numBranches)
{
	MELink *link;
	
	// This MELink is the one that points back up to the root
	link=inTree->newNode();
	
	for (int i=0; i<numBranches; i++)
		getBranch(inTree, link);				
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getBranch
// ---------------------------------------------------------------------------------
 
// Reads a branch in. This could be a node or a tip (calls getNode or getTip
// accordingly). It then reads the branch length and creates an MELink that will
// point at the new node or tip. It connects these two up and sets them to use
// the same branch number (an index to an array of branch lengths in the METree).
// Finally it inserts the new branch MELink into the polytomy given by inLink.

MELink *MEBinaryStream::getBranch(METree *inTree, MELink *inLink)
{
	double len;
	MELink *link, *branchLink;

	int n=readInteger(); // number of branches in polytomy
	if (n>0)	// is a node
		link=getNode(inTree, n);
	else		// is a tip 
		link=getTip(inTree);

	len=readDouble();
	
	branchLink=inTree->newBranch(inLink, len);
	link->branchNo=branchLink->branchNo;
	link->branch=branchLink;
	branchLink->branch=link;
	
	return branchLink;
}	


#pragma mark === Matrix Reading ===

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::getMatrix
// ---------------------------------------------------------------------------------

void MEBinaryStream::getMatrix(MEMatrix *inMatrix)
{
	ThrowIfNull(inMatrix);

	throw MEException("No implemented yet");
}


#pragma mark === Low Level Writing ===

// All are abstract functions

#pragma mark === Data Object Writing ===

// These functions are used to put data objects onto a stream

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, int i)
{
	writeString(label);
	writeInteger(i);
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, double d)
{
	writeString(label);
	writeDouble(d);
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, bool b)
{
	writeString(label);
	writeChar((char)b);
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, char ch)
{
	writeString(label);
	writeChar(ch);
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, const char *str)
{
	writeString(label);
	if (str!=NULL)
		writeString(str);
	else
		writeString("");
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, const char *str, int length)
{
	writeString(label);
	if (str!=NULL && length > 0)
		writeText(str, length);
	else
		writeString("");
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, const ObjectID id)
{
	writeString(label);
	writeChar(id.chars[0]);
	writeChar(id.chars[1]);
	writeChar(id.chars[2]);
	writeChar(id.chars[3]);
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::put
// ---------------------------------------------------------------------------------

void MEBinaryStream::put(char *label, MEObject *inObject)
{
	writeString(label);
	
	if (inObject!=NULL) {
		int index=cache->findIndex(inObject);
			
		if (index<0) { // This object is not in the cache...
			writeInteger(0);
			putObject(inObject);
			writeChar(0xFF); // end of object flag
		} else { // This object is in the cache...
			// ...so just refer to it by its index
			writeInteger(index+1);
		}
	} else
		writeInteger(-1);
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putList
// ---------------------------------------------------------------------------------

void MEBinaryStream::putList(MEList *inList)
{
	int i, n;
	
	ThrowIfNull(inList);
	
	n=inList->getNumObjects();
	writeInteger(n);
	if (n>0) {
		for (i=0; i<n; i++) {
			int index=cache->findIndex(inList->getObject(i));
			
			if (index<0) { // This object is not in the cache...
				writeInteger(0);
				// ...so write it out in full.
				putObject(inList->getObject(i));
				writeChar(0xFF); // end of object flag
			} else { // This object is in the cache...
				// ...so just refer to it by its index
				writeInteger(index+1);
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putIntArray
// ---------------------------------------------------------------------------------

void MEBinaryStream::putIntArray(MEIntArray *inList)
{
	int i, n;
	
	ThrowIfNull(inList);
	
	n=inList->getNumValues();
	writeInteger(n);
	if (n>0) {
		for (i=0; i<n; i++)
			writeInteger(inList->getValue(i));
	}
}

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putDoubleArray
// ---------------------------------------------------------------------------------

void MEBinaryStream::putDoubleArray(MEDoubleArray *inList)
{
	int i, n;
	
	ThrowIfNull(inList);
	
	n=inList->getNumValues();
	writeInteger(n);
	if (n>0) {
		for (i=0; i<n; i++)
			writeDouble(inList->getValue(i));
	}
}

#pragma mark === Tree Writing ===

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putTree
// ---------------------------------------------------------------------------------

void MEBinaryStream::putTree(METree *inTree)
{
	ThrowIfNull(inTree);
	
	if (inTree->getIsRooted()) {
		writeInteger(2);
		putNode(inTree, inTree->getRoot1());
		writeDouble(inTree->getRootDistance1());
		
		putNode(inTree, inTree->getRoot2());
		writeDouble(inTree->getRootDistance2());
	} else {
		MELink *root=inTree->getRoot1();
		writeInteger(root->getNumBranches());

		MELink *link=root->next;
		while (link!=root) {
			putBranch(inTree, link->branch);
			link=link->next;
		}
		putBranch(inTree, link->branch);
	}
}


// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putBranch
// ---------------------------------------------------------------------------------

void MEBinaryStream::putBranch(METree *inTree, MELink *inLink)
{
	putNode(inTree, inLink);
	writeDouble(inTree->getBranchLength(inLink->branchNo));
}


// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putNode
// ---------------------------------------------------------------------------------

void MEBinaryStream::putNode(METree *inTree, MELink *inLink)
{
	MELink *link=inLink->next;
	
	if (inLink->tipNo==-1) {
		writeInteger(inLink->getNumBranches());
		while (link!=inLink) {
			putBranch(inTree, link->branch);
			link=link->next;
		}
	} else {
		writeInteger(0);
		writeInteger(inLink->taxonNo+1);
	}
}

#pragma mark === Matrix Writing ===

// ---------------------------------------------------------------------------------
//		 MEBinaryStream::putMatrix
// ---------------------------------------------------------------------------------

void MEBinaryStream::putMatrix(MEMatrix *inMatrix)
{
	ThrowIfNull(inMatrix);

	throw MEException("No implemented yet");
}


