/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEBirthDeathSimulator.cpp

				MEBirthDeathSimulator Class
				
 ******************************************************************************/

#include <stdio.h>

#include "CProgress.h"
#include "MEBirthDeathSimulator.h"

#include "MEMaths.h"

#pragma mark === MEBirthDeathSimulator ===

// =========================================================================
//	MEBirthDeathSimulator::MEBirthDeathSimulator
// =========================================================================

MEBirthDeathSimulator::MEBirthDeathSimulator() :
	MEPhylogenySimulator()
{
	setPureBirth(1.0);
	
	mExtantStop = 100;
	mTimeStop = 0.0;		
}

// =========================================================================
//	MEBirthDeathSimulator::~MEBirthDeathSimulator
// =========================================================================

MEBirthDeathSimulator::~MEBirthDeathSimulator()
{
}
		
// =========================================================================
//	MEBirthDeathSimulator::setPureBirth
// =========================================================================

void MEBirthDeathSimulator::setPureBirth(double inBirthRate)
{
	mBirthRate = inBirthRate;
	mDeathRate = 0.0;
	mSumRate = mBirthRate;
	mRelativeBirthRate = 1.0;
	
	mPureBirth = true;
}

// =========================================================================
//	MEBirthDeathSimulator::setBirthDeathRates
// =========================================================================

void MEBirthDeathSimulator::setBirthDeathRates(double inBirthRate, double inDeathRate)
{
	if (inDeathRate == 0.0)
		setPureBirth(inBirthRate);
	else {
		mBirthRate = inBirthRate;
		mDeathRate = inDeathRate;
		mSumRate = mBirthRate + mDeathRate;
		mRelativeBirthRate = mBirthRate / mSumRate;
	
		mPureBirth = false;
	}
}

// =========================================================================
//	MEBirthDeathSimulator::setExtantStop
// =========================================================================

void MEBirthDeathSimulator::setExtantStop(long inExtantStop)
{
	mExtantStop = inExtantStop;
}

// =========================================================================
//	MEBirthDeathSimulator::setTimeStop
// =========================================================================

void MEBirthDeathSimulator::setTimeStop(long inTimeStop)
{
	mTimeStop = inTimeStop;
}

// =========================================================================
//	MEBirthDeathSimulator::simulate
// =========================================================================

void MEBirthDeathSimulator::simulate()
{
	double dt;
	long extant;
	
	if (mExtantStop < 2 && mTimeStop <= 0.0)
		throw CException("Bad stop conditions for birth-death simulation");	
	
	initialize();
	mPhylogeny->setBirthTime(0, mCurrentTime);
	
	if (mProgress != NULL && mExtantStop > 1) {
		mProgress->description("Generating phylogeny...");
		mProgress->reset();
		mProgress->progress(0, (int)mExtantStop);
		mProgress->setSkip((int)mExtantStop / 100);
	}
		
	while (checkStopConditions()) {
		dt = timeToNextEvent();
		
		mCurrentTime += dt;
		
		event();
			
		if (mPhylogeny->getNumExtantLineages() < 1)
			throw MEAllExtinctTreeSimError();

		if (mProgress != NULL && mExtantStop > 1)
			mProgress->progress(mPhylogeny->getNumExtantLineages(), (int)mExtantStop);
	}
	
	int i, n = mPhylogeny->getNumExtantLineages();
	for (i = 0; i < n; i++)
		mPhylogeny->setDeathTime(mPhylogeny->getExtantLineage(i), mCurrentTime);
}

// =========================================================================
//	MEBirthDeathSimulator::event
// =========================================================================

bool MEBirthDeathSimulator::event()
{
	long extant = mPhylogeny->getRandomExtantIndex();
	
	if (mPureBirth || eventIsBirth())
		return lineageBirth(extant, 2);
	else
		return lineageDeath(extant);
}

// =========================================================================
//	MEBirthDeathSimulator::checkStopConditions
// =========================================================================

bool MEBirthDeathSimulator::checkStopConditions()
{
	bool go = true;
	
	if (mExtantStop > 0 && mPhylogeny->getNumExtantLineages() >= mExtantStop) {
		// draw one more waiting time
		mCurrentTime += timeToNextEvent();
		go = false;
	}
	
	if (mTimeStop > 0.0 && mCurrentTime >= mTimeStop) {
		mCurrentTime = mTimeStop;
		go = false;
	}

	return go;
}

// =========================================================================
//	MEBirthDeathSimulator::timeToNextEvent
// =========================================================================

double MEBirthDeathSimulator::timeToNextEvent()
{
	return -log((double)MERandom::fRan()) / 
			(double(mPhylogeny->getNumExtantLineages()) * mSumRate);
}

// =========================================================================
//	MEBirthDeathSimulator::eventIsBirth
// =========================================================================

bool MEBirthDeathSimulator::eventIsBirth()
{
	return MERandom::bRan(mRelativeBirthRate);
}

// ---------------------------------------------------------------------------------
//		 MEBirthDeathSimulator::describe
// ---------------------------------------------------------------------------------

void MEBirthDeathSimulator::describe(CReportRecord *ioReport)
{
	CParagraphElement *heading = new CParagraphElement(kHeadingLevel2);
	heading->add("Birth-Death model");
	ioReport->add(heading);
	
	CParagraphElement *para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
	para->add("Settings:");
	ioReport->add(para);

	CListElement *list = new CListElement(kBullettedList, 2);
	para = new CParagraphElement();
	para->add("Birth rate = ");
	para->add(mBirthRate);
	list->add(para);
	if (!mPureBirth) {
		para = new CParagraphElement();
		para->add("Death rate = ");
		para->add(mDeathRate);
		list->add(para);
	}
	if (mExtantStop > 0) {
		para = new CParagraphElement();
		para->add("Extant stop = ");
		para->add((int)mExtantStop);
		list->add(para);
	}
	if (mTimeStop > 0.0) {
		para = new CParagraphElement();
		para->add("Time stop = ");
		para->add(mTimeStop);
		list->add(para);
	}
	if (mExtantStop <= 0 && mTimeStop <= 0.0) {
		list->add("No stopping conditions specified");
	}	
	
	ioReport->add(list);
}
