/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEBitset.cpp

		MEBitset Class
					
 ******************************************************************************/
// MEBitset.cpp - some code taken with permission from R. D. M. Page's VClass code;
// vastly extended by M. A. Charleston.  Copyright held by R.D.M.Page and M.A.Charleston, 1997.

//  includes:
// C/C+:
#include <stdio.h>
// Project specific:
#include "MEGlobal.h"
#include "MEExceptions.h"
#include "MEList.h"
#include "MEMaths.h"
#include "MEBitset.h"
#include "MEPermutation.h"

using namespace std;

//  variables:
char
	bsmsg[kLongStringLength];


// LONGBITSET is defined in MEBitset.h at present (5/12/97)

#ifdef LONGBITSET
unsigned long 
	bits[kBitsPerCell] =
		{0x00000001,0x00000002,0x00000004,0x00000008,
		 0x00000010,0x00000020,0x00000040,0x00000080,
 		 0x00000100,0x00000200,0x00000400,0x00000800,
		 0x00001000,0x00002000,0x00004000,0x00008000,
		 0x00010000,0x00020000,0x00040000,0x00080000,
		 0x00100000,0x00200000,0x00400000,0x00800000,
		 0x01000000,0x02000000,0x04000000,0x08000000,
		 0x10000000,0x20000000,0x40000000,0x80000000};
#else
unsigned short 
	bits[kBitsPerCell] =
		{0x0001,0x0002,0x0004,0x0008,
		 0x0010,0x0020,0x0040,0x0080,
 		 0x0100,0x0200,0x0400,0x0800,
		 0x1000,0x2000,0x4000,0x8000};
#endif

// macros:
//#ifdef _MEDEBUG_
#define ThrowBadBitsetCell(ix) \
	if ((ix < 0) || (ix >= size)) { \
		sprintf(bsmsg, "local cell index %d is out of range [0, %d]", ix, size-1); \
		throw MEBadBitsetCell(bsmsg); }
#define ThrowBadBitsetElement(ix) \
	if ((ix < 0) || (ix >= getMaxCapacity())) { \
		sprintf(bsmsg, "local MEBitset element index %d is out of range [0, %d]", \
			ix, getMaxCapacity()-1); \
		throw MEBadBitsetElement(bsmsg); }
//#else
//#define ThrowBadBitsetCell(ix) { } // Do nothing
//#define ThrowBadBitsetElement(ix) { } // Do nothing
//#endif

//  methods:

// MEBitset
MEBitset::MEBitset(int m)
{
	maxElements = abs(m);
	size = (int) (maxElements / kBitsPerCell) + 1;
	card = 0;
	Count = 0;
	CurrentElement = 0;

	NewArray(set, cell, size);
	idx = -1;					// because most lists start from 0
	for (int i = 0; i < size; i++)
		setCell(i, 0);
}


// ~MEBitset
MEBitset::~MEBitset()
{
	delete [] set;
}


#pragma mark -

// accommodate
void MEBitset::accommodate(int i)
//: Make sure there is room in the bitset for this element, resizing if necessary
{
	if (i < size*kBitsPerCell)
		return;

	int newSize = (int) (i / kBitsPerCell) + 1;
	resize(newSize);
}


// add
MEBitset& MEBitset::add(int i) // Add element i to the set
{
	return add((long int) i);
}
MEBitset& MEBitset::add(long int i) // Add element i to the set
{
	accommodate(i);
	long int m, j;
	m = i % kBitsPerCell;
	j = i / kBitsPerCell;
	if ((getCell(j) & bits[m]) > 0) {
		// don't do anything
	}
	else {
		addToCell(j, bits[m]);
		card++;
	}	
	return *this;
}
MEBitset& MEBitset::add(const MEBitset &b)	// Add another set to the set
{
	ThrowIfNULL(&b);
	resize(b.getSize());
	int i, k = b.getSize();
	for (i = 0; i < k; i++)
		setCell(i, getCell(i) | b.getCell(i));
	return *this;
}
MEBitset& MEBitset::add(char *str)
{
	int i, j = strlen(str);
	for (i = j-1; i >= 0; i--) {
		if (str[i] == '1' || str[i] == '\1')
			add(i);
	}
	return *this;
}
MEBitset& MEBitset::add(char *str, int length)
{
	int i;
	for (i = length-1; i >= 0; i--) {
		if (str[i] == '1' || str[i] == '\1')
			add(i);
	}
	return *this;
}


// MEBitset::addToCell
MEBitset& MEBitset::addToCell(int i, cell x)
{
	cell temp = getCell(i);
	temp += x;
	setCell(i, temp);
	return *this;
}


#pragma mark -


// MEBitset::checkcard
bool MEBitset::checkcard() const
{
	int i = getCard();
	countElements();
	if (i != getCard())
		return false;
	return true;
}


// clear
MEBitset& MEBitset::clear()	// Clears the set
{
	for (int i = 0; i < size; i++)
		setCell(i, 0);
	card = 0;
	_cardKnown = true;
	Count = 0;
	CurrentElement = 0;
	return *this;
}


// contains
bool MEBitset::contains(MEBitset &b) const
{
	ThrowIfNull(&b);
	return isSupersetOf(b);
}
bool MEBitset::contains(int i) const // true if i in set
{
	if ((i < 0) || (i >= getMaxCapacity()))
		return false;
		// it is impossible for this set to contain this number.
		// NOTE: this test used to throw an exception! (MAC 1/11/99)
	ThrowBadBitsetElement(i);
	int m = i % kBitsPerCell;
	int j = i / kBitsPerCell;
	if ((getCell(j) & bits[m]) == bits[m])
		return true;
	else
		return false;
}


// copyBits
MEBitset& MEBitset::copyBits(long i)
{
	clear();
	setCell(0, i);
	return *this;
}


// copyFrom
MEBitset& MEBitset::copyFrom(const MEBitset &bs)
{
	ThrowIfNull(&bs);
	resize(max(size, bs.getSize()));
	int i, k = bs.getSize();
	for (i = 0; i < k; i++)
		setCell(i, bs.getCell(i));
	for ( ; i < size; i++)
		setCell(i, 0);
	CurrentElement = bs.CurrentElement;
	Count = bs.Count;
	card = bs.card;
	return *this;
}
MEBitset& MEBitset::copyFrom(const permutation &p)
{
	int i, k = p.getSize();
	resize(k);
	for (i = 0; i < k; i++)
		add(p[i]);
	CurrentElement = getFirstElement();
	return *this;
}
MEBitset& MEBitset::copyFrom(char *str)
{
	clear();
	add(str);
	return *this;
}
MEBitset& MEBitset::copyFrom(char *str, int length)
{
	clear();
	add(str, length);
	return *this;
}

// ------------------------------------------------------------------------------------
//    MEBitset::copyFromSplitIndex
// ------------------------------------------------------------------------------------
MEBitset& MEBitset::copyFromSplitIndex(unsigned long int l)
{
	clear();
	int i, k = 8*sizeof(unsigned long int);
	for (i = 0; i < k; i++) {
		if (l & (1 << i))
			add(i);
	}
	return *this;
}


// countElements
int MEBitset::countElements() const // Return the number of elements in the set
{
	int count = 0, i, j;
	for (i = 0; i < size; i++)
		if (getCell(i) != 0)
			for (j = 0; j < kBitsPerCell; j++)
				if (getCell(i) & bits[j])
					count++;
	card = count;
	return card;
}


#pragma mark -


// del
MEBitset& MEBitset::del(int i)	// Delete element i from the set
{
	ThrowBadBitsetElement(i);
	int m, j;
	m = i % kBitsPerCell;
	j = i / kBitsPerCell;
	if (getCell(j) & bits[m]) {
		delFromCell(j, bits[m]);
		card--;
	}
	return *this;
}
MEBitset& MEBitset::del(const MEBitset &b)	// Subtract set b from this set
{
	ThrowIfNull(&b);
	cell c;
	int i, k = b.getSize();
	for (i = 0; i < k; i++) {
		//setCell(i, getCell(i) - b.getCell(i));
		c = b.getCell(i);
		set[i] -= (set[i] & c);
		// e.g. (101100 -= (101100 & 001010)) = (101100 -= 001000) = 100100
	}
	_cardKnown = false;
	return *this;
}
MEBitset& MEBitset::del(char *str)
{
	int i, k = strlen(str);
	for (i = 0; i < k; i++) {
		if (str[k-i-1] == '1')
			del(i);
	}
	return *this;
}


// delFromCell
MEBitset& MEBitset::delFromCell(int i, cell x)
{
	cell temp = getCell(i);
	temp -= (temp & x);
	return setCell(i, temp);
}


#pragma mark -


// equals
bool MEBitset::equals(const MEBitset &b) const
{
	ThrowIfNull(&b);
	int i, k = b.getSize();
	for (i = 0; i < k; i++) {
		if (getCell(i) != b.getCell(i))
			return false;
	}
	for ( ; i < size; i++) {
		if (getCell(i) > 0)
			return false;
	}
	return true;
}
bool MEBitset::equals(char *str) const
{
	int i, k = strlen(str);
	for (i = 0; i < k; i++) {
		if (contains(i)) {
			if (str[k-i-1] == '0')
				return false;
		}
		else {
			if (str[k-i-1] == '1')
				return false;
		}
	}
	return true;
}


#pragma mark -


// fill
MEBitset& MEBitset::fill(int n)
{
//	ThrowBadBitsetElement(n);
	if (n < 0)
		throw MEBitsetError("Can't fill bitset with negative maximum");
	for (int i = 0; i < n; i++)
		add(i);
	return *this;
}
MEBitset& MEBitset::fill(char *str, char *map)
{
	int i, k = strlen(str);
	accommodate(k);
	clear();
	if (map) {
		for (i = 0; i < k; i++) {
			if (strchr(map, str[i]) != NULL)	// see THINK Reference for defn. of strchr
				add(i);
		}
	}
	else {
		char ch = str[0];
		for (i = 0; i < k; i++) {
			if (ch != str[i])
				add(i);
		}
	}
	if (contains(k-1))
		setComplement(k);
	return *this;
}


MEBitset& MEBitset::fillFrom(const permutation &p)
{
	int i, k = min(getMaxCapacity(), p.getSize());
	for (i = 0; i < k; i++)
		add(p[i]);
	return *this;
}


MEBitset& MEBitset::fillRandom(int k)
{
	if (k < 0)
		k = getSize()*kBitsPerCell;
	permutation *pi;
	NewClass(pi, permutation(k));
	pi->randomize();
	clear();
	for (int i = 0; i < k; i++)
		add(pi->getIthElement(i));
	delete pi;
	return *this;
}


MEBitset& MEBitset::fillRandomUpTo(int k, double pr)
{
	clear();
	pr = (double) pr - ((int)pr);
	for (int i = 0; i < k; i++) {
		if (MERandom::fRan() < pr)
			add(i);
	}
	return *this;
}


#pragma mark -


// getCard
int MEBitset::getCard() const
{
	if (!_cardKnown)
		countElements();
	return card;
}


// getbit
bool MEBitset::getBit(int x) const
{
	ThrowBadBitsetElement(x);
	int m = x % kBitsPerCell;
	int j = x / kBitsPerCell;
	if ((getCell(j) & bits[m]) > 0)
		return true;
	else
		return false;
}


// getCell
cell MEBitset::getCell(int i) const
{
//	ThrowBadBitsetCell(i);
	if ((i < 0) || (i >= size)) {
		sprintf(bsmsg, "local cell index %d is out of range [0, %d]", i, size-1);
		throw MEBadBitsetCell(bsmsg); 
	}
	return set[i];
}


// getFirstElement
int	MEBitset::getFirstElement() const
{
	countElements();
	if (card == 0)
		return (-1);
	Count = 0;
	for (int i = 0; i < size*kBitsPerCell; i++)
		if (contains(i)) {
			CurrentElement = i;
			Count++;
			return i;
		}
	return -1;
}


// getLastElement
int MEBitset::getLastElement() const
{
	int i, k = size*kBitsPerCell;
	for (i = k; i >= 0; i--)
		if (contains(i)) {
			CurrentElement = i;
			Count = getCard();
			return i;
		}
	return -1;
}


// getIthElement
int MEBitset::getIthElement(int i) const
{
	ThrowBadBitsetElement(i);
	int count = 0;
	for (int j = 0; j < size*kBitsPerCell; j++) {
		if (contains(j)) {
			if (count == i)
				return j;
			count++;
		}
	}
	return -1;
}


// getNextElement
int MEBitset::getNextElement() const
{
	if (Count >= getCard())
		return -1;	// no more elements!
	do {
		CurrentElement++;
	} while (!contains(CurrentElement));
	Count++;
	return (CurrentElement);
}


// getNextLargestFrom
int MEBitset::getNextLargestFrom(int x) const
{
	int k = x;
	do {
		k--;
	} while (!contains(k));
	return getPreviousElement();
}


// getPreviousElement
int MEBitset::getPreviousElement() const
{
	if (Count <= 0)
		return -1;	// no more elements!
	do {
		CurrentElement--;
	} while (!contains(CurrentElement));
	Count--;
	return (CurrentElement);
}


#pragma mark -


// intersects
bool MEBitset::intersects(const MEBitset &b) const
{
	int i, k = min(getSize(), b.getSize());
	for (i = 0; i < k; i++)
		if (getCell(i) & b.getCell(i))
			return true;
	return false;
}


// intersectWith
MEBitset& MEBitset::intersectWith(const MEBitset &bs)
{
	cell ts;
	for (int i = 0; i < min(size, bs.getSize()); i++) {
		ts = getCell(i) & bs.getCell(i);
		setCell(i, ts);
	}
	countElements();
	return *this;
}


// isDisjointFrom
bool MEBitset::isDisjointFrom(const MEBitset &b) const
{
	if (&b == NULL)
		throw MENullPointer("MEBitset::isDisjointFrom(b): MEBitset b is null", 668);
	MEBitset *i;
	NewClass(i, MEBitset(size*kBitsPerCell));
	i->setIntersectionOf(*this, b);
	bool _disjoint = i->isNull();
	delete i;
	return _disjoint;
}


// isLexAfter
bool MEBitset::isLexAfter(MEBitset &b) const
// return true iff the "Hendy index" of this set is a larger integer than that of <b>.
{
	// if the largest element in <this> is greater than the largest element in <b> then <this> has a higher index:
	if (size > b.getSize() && getCell(size-1) > 0)
		return true;
	// ensure strict inequality:
	if (equals(b))
		return false;
	// now the bitsets form a canonical ordering, treating the cells as coefficients of a large integer
	// in base (2^(size of a long integer)):
	int i, k = max(size, b.getSize());
	for (i = k-1; i >= 0; i--)
		if (getCell(i) > b.getCell(i))
			return true;
	return false;
}


// isLexBefore
bool MEBitset::isLexBefore(MEBitset &b) const
// return true iff the "Hendy index" of this set is a smaller integer than that of <b>.
{
	// if the largest element in <this> is greater than the largest element in <b> then <this> has a higher index:
	if (size < b.getSize() && b.getCell(b.getSize()-1) > 0)
		return true;
	// ensure strict inequality:
	if (equals(b))
		return false;
	// now the bitsets form a canonical ordering, treating the cells as coefficients of a large integer
	// in base (2^(size of a long integer)):
	int i, k = max(size, b.getSize());
	for (i = k-1; i >= 0; i--)
		if (getCell(i) < b.getCell(i))
			return true;
	return false;
}


// isSubsetOf
bool MEBitset::isSubsetOf(MEBitset &b) const // true if set is subset of b
{
	ThrowIfNULL(&b);
	MEBitset *u;
	NewClass(u, MEBitset(size*kBitsPerCell));
	u->copyFrom(*this);
	u->add(b);					// u = *this U b
	b.countElements();
	bool _subset = (b.getCard() == u->getCard());
	delete u;
	return _subset;
		// $ A \subseteq B \iff |A U B| = |B| $
}


// isSupersetOf
bool MEBitset::isSupersetOf(MEBitset &b) const	// true if set is superset of b
{
	ThrowIfNULL(&b);
	MEBitset *u;
	NewClass(u, MEBitset(size*kBitsPerCell));
	u->copyFrom(*this);
	u->add(b);					// u = *this U b
	countElements();
	bool _superset = (getCard() == u->getCard());
	delete u;
	return _superset;
		// $ A \supseteq B \iff |A U B| = |A| $
}


#pragma mark -


void MEBitset::putAsString(char *str)
//:  Puts the bitset as a string of 1's and 0's into $str
// Note that the most significant bit corresponds to the 
{
	int i, k = strlen(str);
	for (i = 0; i < k; i++) {
		if (contains(i))
			str[i] = '1';
		else
			str[i] = '0';
	}
}


#pragma mark -


// randomSubset
MEBitset* MEBitset::randomSubset(int x, MEBitset *S)
//: Select 'x' elements from this subset and put them into S.
{
	if (S == NULL)
		NewClass(S, MEBitset());
	if (x >= getCard()) {
		S->copyFrom(*this);
		return S;
	}
	permutation *p;
	NewClass(p, permutation(getCard()));
	p->copyFrom(*this);
	p->randomize();
	for (int i = 0; i < x; i++)
		S->add((*p)[i]);
	delete p;
	return S;
}


// resize
MEBitset& MEBitset::resize(int newSize)
{
	// newsize is the number of cells
	if (newSize <= size)
		return *this;
	cell *tempcells;
	NewArray(tempcells, cell, newSize);
	for (int i = 0; i < size; i++)
		tempcells[i] = set[i];
	delete [] set;
	
	NewArray(set, cell, newSize);
	int i;
	for (i = 0; i < size; i++)
		set[i] = tempcells[i];
	for ( ; i < newSize; i++)
		set[i] = 0;
	delete [] tempcells;
	countElements();
	size = newSize;
	return *this;
}


#pragma mark -


// setcard
MEBitset& MEBitset::setCard(int c)
{
	ThrowBadBitsetElement(c);
	card = c;
	return *this;
}


// setCell
MEBitset& MEBitset::setCell(int i, cell ts)
{
	ThrowBadBitsetCell(i);
	set[i] = (cell) ts;// ts
	_cardKnown = false;
	return *this;
}


// setCount
void MEBitset::setCount(int c) const
{
	if ((c < 0) || (c >= card)) {
		sprintf(bsmsg, "MEBitset::setCount(c): int c=%d is out of range [0, %d]",
			c, card-1);
		throw MEException(bsmsg, 668);
	}
	Count = c;
//	return *this;
}


// ------------------------------------------------------------------------------------
//    MEBitset::setComplement
// ------------------------------------------------------------------------------------
//: MEBitset::setComplement(n)
//	 Set the current bitset to its complement, with respect to the set {0,...,n-1}.
//	 For example, after initializing a set "B" with "MEBitset B(\"01101\");", "B"
//	 would be {0,2,3}.  After applying "B.setComplement(5);" "B" is
//	 {0,1,2,3,4} \ {0,2,3} = {1,4,5}.
MEBitset& MEBitset::setComplement(int n)
{
	MEBitset
		bs;
	bs.fill(n);
	bs -= *this;		// copy to bs
	copyFrom(bs);
	return *this;
}


// setIntersectionOf
MEBitset& MEBitset::setIntersectionOf(const MEBitset& bs1, const MEBitset& bs2)
{
	//  NOTE  this destroys the original!
	ThrowIfNULL(&bs1);
	ThrowIfNULL(&bs2);
	clear();
	resize(max(bs1.getSize(), bs2.getSize()));
	cell ts;
	for (int i = 0; i < min(bs1.getSize(), bs2.getSize()); i++) {
		ts = bs1.getCell(i) & bs2.getCell(i);
		setCell(i, ts);
	}
	return *this;
}


// setIntersectionWith
MEBitset& MEBitset::setIntersectionWith(const MEBitset& bs)
{
	//  NOTE  this destroys the original!
	ThrowIfNULL(&bs);
	resize(max(size, bs.getSize()));
	cell ts;
	for (int i = 0; i < min(size, bs.getSize()); i++) {
		ts = getCell(i) & bs.getCell(i);
		setCell(i, ts);
	}
	return *this;
}


// setNextLex
MEBitset& MEBitset::setNextLex()
//: Set the next lexicographic bitset.
{
	int
		b = getNumCells(), 
		i = 0, 
		k = 0,
		lastMax;
	// find the first cell which isn't 'full' of ones:
	for (k = 0; k < size-1; k++) {
		if ((long) getCell(k) != -1)
			break;
	}
	if (k < size-1) {
		// all the 'full' cells are "clocked" back to zero:
		for (i = 0; i < k; i++)
			set[i] = 0;
		// and the first non-full cell is incremented:
		set[k]++;
	}
	else {
		lastMax = 1 << (maxElements % kBitsPerCell) - 1;
		if ((long) getCell(k) == lastMax)
			setCell(k, 0);
	}
	// if the cells are like this,
	//                           |maxElements = 15
	// cells == 1111 1111 1100 0000
	// then we get k <- 2,
	//                           |maxElements = 15
	// cells <- 0000 0000 0010 0000
	//          ^^^^ ^^^^ ^^^^- incrementing the k-th cell
	//          |       |-- clocking the previous cells.
	// whereas if we had
	//                     |maxElements = 10
	// cells == 1111 1111 1100 0000
	// we get
	// cells <- 0000 0000 0000 0000.
	return *this;
}


// setunionof
MEBitset& MEBitset::setUnionOf(const MEBitset& bs1, const MEBitset& bs2)
{
	clear();
	add(bs1);
	add(bs2);
	return *this;
}


// setunionwith
MEBitset& MEBitset::setUnionWith(const MEBitset& bs)
{
	resize(max(size, bs.getSize()));
	add(bs);
	return *this;
}


#pragma mark -


// write
char* MEBitset::write(char *str) const
{
	strcpy(bsmsg, "{ ");
	char tempstr[16];
	int i, j, k=0;
	for (i = 0; i < size; i++)
		for (j = 0; j < kBitsPerCell; j++)
			if ((getCell(i) & bits[j]) == bits[j]) {
				sprintf(tempstr, "%d ", i * kBitsPerCell + j);
				strcat(bsmsg, tempstr);
				k++;
			}
	strcat(bsmsg, "}");
	if (str)
		strcpy(str, bsmsg);
	return bsmsg;
}


#pragma mark -


// ----- MEBitset functions --------------------------------------------------------------------------

// requiredspace_bitset
int requiredspace_bitset(int mel)
{
	// return the amount of space required in bytes of a MEBitset with up to mel elements
	int s = 0;
	s += 4*sizeof(int);	// card, CurrentElement, Count, size
	s += sizeof(cell) * mel;
	return s;
}


