/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEBitsetList.cpp

				MEBitsetList Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEBitsetList.h"
#include "MEBitset.h"

#include "MEStream.h"


#pragma mark === MEBitsetList ===

// ---------------------------------------------------------------------------------
//		 MEBitsetList
// ---------------------------------------------------------------------------------

MEBitsetList::MEBitsetList() : 
	MEList()
//: Construct an MEBitsetList.
// This calls the base class (MEList) constructor.
{
}

// ---------------------------------------------------------------------------------
//		 ~MEBitsetList
// ---------------------------------------------------------------------------------

MEBitsetList::~MEBitsetList()
//: Destroy this list.
{
}


// -----------------------------------------------------------------------------
//    MEBitsetList::copyFrom
// -----------------------------------------------------------------------------
MEBitsetList& MEBitsetList::copyFrom(MEBitsetList &bsl)
//: Copy the entire contents of the given list into this one. 
{
	int i, k = bsl.getNumBitsets();
	for (i = 0; i < k; i++)
		add(bsl.getBitset(i));
	return *this;
}


// ---------------------------------------------------------------------------------
//		MEBitsetList::createObject
// ---------------------------------------------------------------------------------
MEObject* MEBitsetList::createObject()
//: Create a new MEBitset for the list.
{
	MEObject *obj;
	NewClass(obj, MEBitset());
	return obj;
}


// ---------------------------------------------------------------------------------
//		 MEBitsetList::getBitset
// ---------------------------------------------------------------------------------
MEBitset* MEBitsetList::getBitset(int index)
//: Retrieves a bitset from the list.
// This also checks to see that the passed index is in the right range.
{
	return (MEBitset *)getObject(index);
}


// ------------------------------------------------------------------------------------
//    MEBitsetList::getEmptySet
// ------------------------------------------------------------------------------------
MEBitset* MEBitsetList::getEmptySet()
//: Get an empty set from the list.
// If there is no such set to begin with, this function creates an empty set and adds it to the list.
{
	MEBitset *b;
	static int emptySetIndex = -1;
	if (emptySetIndex >= 0)
		return getBitset(emptySetIndex);
	int i, k = getNumObjects();
	for (i = 0; i < k; i++) {
		b = getBitset(i);
		if (b->getCard() == 0) {
			emptySetIndex = i;
			return b;
		}
	}
	if (!_owningSets)
		return NULL;
	NewClass(b, MEBitset());
	b->clear();
	add(b);
	emptySetIndex = b->getIndex();
	return b;
}


// ---------------------------------------------------------------------------------
//		 MEBitsetList::addBitset
// ---------------------------------------------------------------------------------
int MEBitsetList::addBitset(MEBitset *inBitset, bool isOwner)
//: Add a bitset to the MEBitsetList and return its index in the list.
{
	return add(inBitset, isOwner);
}


// -----------------------------------------------------------------------------
//    MEBitsetList::clear
// -----------------------------------------------------------------------------
void MEBitsetList::clear()
 //: Clear all the elements from the list.  Delete them elsewhere!
{
	int i, k = getNumObjects();
	for (i = 0; i < k; i++)					// delete the MEBitsets themselves
		if (((void *)objects[i]->getOwner() == (void *)this)
			|| ((void*)objects[i]->getOwner() == NULL))	// ... but only if this is the owner!
			delete objects[i];
	delete [] objects;
	objects = NULL;
	// ... to here.
	capacity = numObjects = 0;
}


// ---------------------------------------------------------------------------------
//		 MEBitsetList::insertBitset
// ---------------------------------------------------------------------------------
void MEBitsetList::insertBitset(int index, MEBitset *inBitset, bool isOwner)
//: Insert a bitset into the list, advancing the positions of the lower-order bitsets if necessary.
// This is a simple wrapper for the base method.
{
	insert(index, inBitset, isOwner);
}


// ---------------------------------------------------------------------------------
//		 MEBitsetList::removeBitset
// ---------------------------------------------------------------------------------
MEBitset* MEBitsetList::removeBitset(int index)
//: Remove a bitset from the list, decrementing the positions of the lower-order bitsets if necessary.
// This is a simple wrapper for the base method.
{
	return (MEBitset *)remove(index);
}


// ---------------------------------------------------------------------------------
//		MEBitsetList::compare
// ---------------------------------------------------------------------------------
int MEBitsetList::compare(MEObject *inObject1, MEObject *inObject2)
//: Compare the two bitsets in lexicographic order and return their relationship.
// The function returns 0 if the sets are the same, -1 if the first one passed is
// before the second in lexicographic order, and 1 otherwise.
{
	if ( ((MEBitset*)inObject1)->equals(*((MEBitset*)inObject2)) )
		return 0;
	if ( ((MEBitset*)inObject1)->isLexBefore(*((MEBitset*)inObject2)) )
		return -1;
	return 1;
}


#pragma mark === Object I/O ===


// ---------------------------------------------------------------------------------
//		 MEBitsetList::readObject
// ---------------------------------------------------------------------------------
void MEBitsetList::readObject(MEStream *stream)
{
	MEList::readObject(stream);
}


// ---------------------------------------------------------------------------------
//		 MEBitsetList::writeObject
// ---------------------------------------------------------------------------------
void MEBitsetList::writeObject(MEStream *stream)
{
	MEList::writeObject(stream);
}


