/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MECAICParser.cpp

				MECAICParser Class
	
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CReader.h"
#include "MEParser.h"
#include "MECAICParser.h"
#include "METaxon.h"
#include "METaxonList.h"
#include "MELink.h"
#include "METree.h"
#include "METreeList.h"


// ---------------------------------------------------------------------------------
//		 MECAICParser
// ---------------------------------------------------------------------------------

MECAICParser::MECAICParser(CReader *inReader) : MEParser(inReader)
{
	reader->setCommentDelimiters('\0', '\0', '\0');
}


// ---------------------------------------------------------------------------------
//		 MECAICParser::readTree
// ---------------------------------------------------------------------------------

void MECAICParser::readTree(METree *inTree)
{
	char tok[16];
	reader->markPosition();
	reader->readToken(tok, 15);
	if (tok[0]!='#')
		reader->restorePosition();
		
	char line1[1025];
	char line2[1025];
	METaxon *taxon;
	METaxonList *taxonList;
			
	NewClass(taxonList, METaxonList());
	bool done = false;
		
	do {
		if (reader->readLine(line1, 1025, false) > 0 &&
			reader->readLine(line2, 1025, false) > 0) {
			
			NewClass(taxon, METaxon());
			taxon->setDescription(line1);
			taxon->setName(line2);
			
			taxonList->addTaxon(taxon);
		} else 
			done = true;

	} while (!done);
	
	inTree->setTaxonList(taxonList, true);
	
	MELink *link = constructNode(inTree, taxonList, "");
	
	// prune out the temporary MELink that pointed to the root
	MELink *link1=link;
	while (link->next!=link1) 
		link=link->next;
		
	link->next=link1->next;
	inTree->removeLink(link1);
	
	// move on to the first branch in the node
	link=link->next;
	
	if (link->getNumBranches()<3) {
		// if the root node has less than 3 branches then this node is
		// removed. This makes the tree unrooted but we store the position
		// of the root by setting METree's root1 and root2 to point to the
		// nodes either side of the root.
		inTree->setIsRooted(true);
		
		// Set the tree's root pointers
		MELink* root1=link->branch;
		MELink* root2=link->next->branch;
		inTree->setRoot1(root1);
		inTree->setRoot2(root2);
		
		// Set the tree's root branch lengths
		inTree->setRootDistances(inTree->getBranchLength(root1->branchNo),
									inTree->getBranchLength(root2->branchNo));
		
		// Set the nodes either side of the root to point to each other
		root1->branch=root2;
		root2->branch=root1;
		
		// Remove root2's branch length from the tree's array
		inTree->removeBranchLength(root2->branchNo);
		
		// Set the new branches branchNo and the corresponding branchLength
		// This just uses root1's old branch length slot and sets it to the
		// sum of the two old branch lengths
		root2->branchNo=root1->branchNo;
		inTree->setBranchLength(root1->branchNo, 
								inTree->getRootDistance1() + inTree->getRootDistance2());
		
		inTree->removeLink(link->next);
		inTree->removeLink(link);
		
		inTree->setNumInternalNodes(inTree->getNumInternalNodes()-1);
		inTree->setNumNodes(inTree->getNumNodes()-1);
	} else {
		inTree->setIsRooted(false);
		inTree->setRoot1(link);
		inTree->setRoot2(link->branch);
	}
}

// ---------------------------------------------------------------------------------
//		 MECAICParser::readTreeList
// ---------------------------------------------------------------------------------

void MECAICParser::readTreeList(METreeList *inTreeList)
{
	METree *tree=inTreeList->newTree();

	readTree(tree);
	inTreeList->addTree(tree);
}

// ---------------------------------------------------------------------------------
//		 MECAICParser::constructNode
// ---------------------------------------------------------------------------------

MELink* MECAICParser::constructNode(METree *inTree, METaxonList *inTaxonList, char *inCode)
{
	MELink *link;
	int codeLen1, codeLen2, minCodeLen, codeMatch, bestMatch, bestCodeLen, bestTaxon;
	char *taxonCode;
	
	codeLen1 = strlen(inCode);
	bestMatch = -1;
	
	METaxon *taxon = (METaxon *)inTaxonList->getFirst();
	int taxonNo = 0;
	while (taxon != NULL) {
		taxonCode = taxon->getDescription();
		codeLen2 = strlen(taxonCode);
		
		minCodeLen = codeLen2;
		if (codeLen1 < codeLen2)
			minCodeLen = codeLen1;
			
		codeMatch = 0;
		while (codeMatch < minCodeLen && inCode[codeMatch] == taxonCode[codeMatch])
			codeMatch++;
			
		if (codeMatch > bestMatch) {
			bestMatch = codeMatch;
			bestCodeLen = codeLen2;
			bestTaxon = taxonNo;
		}
		
		taxon = (METaxon *)inTaxonList->getNext();
		taxonNo++;
	}
	
	if (codeLen1>0 && bestMatch == bestCodeLen) {
		link = inTree->newTip(bestTaxon);
	} else {
		link = inTree->newNode();
		
		bool found = false;
		taxonNo = 0;
		int numTaxa = inTaxonList->getNumTaxa();
		while (taxonNo < numTaxa && !found) {
			taxon = inTaxonList->getTaxon(taxonNo);
			taxonCode = taxon->getDescription();
			codeLen2 = strlen(taxonCode);
			if (codeLen1 == 0 ||
				(codeLen2 > codeLen1 && strncmp(taxonCode, inCode, codeLen1)==0) )
				found = true;
			taxonNo++;
		}	
		
		ThrowIf(!found);
		
		char newCode[1025];
		strncpy(newCode, taxonCode, codeLen1+1);
		newCode[codeLen1+1] = '\0';

		MELink *link1 = constructNode(inTree, inTaxonList, newCode);
		MELink *link2 = inTree->newBranch(link, 2.0);
		link1->branchNo=link2->branchNo;
		link2->branch=link1;
		link1->branch=link2;
		
		while (taxonNo < numTaxa) {
			taxon = inTaxonList->getTaxon(taxonNo);
			taxonCode = taxon->getDescription();
			codeLen2 = strlen(taxonCode);
			if (codeLen2 > codeLen1 && strncmp(taxonCode, inCode, codeLen1)==0) {
				if (taxonCode[codeLen1] != newCode[codeLen1]) {
					newCode[codeLen1] = taxonCode[codeLen1];

					link1 = constructNode(inTree, inTaxonList, newCode);
					link2 = inTree->newBranch(link, 2.0);
					link1->branchNo=link2->branchNo;
					link2->branch=link1;
					link1->branch=link2;
				}
			}
			taxonNo++;
		}	
		
	}
	
	return link;
}

