/******************************************************************************


Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 MECodeTable.cpp

				MECodeTable Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MECodeTable.h"

#pragma mark === MECodeTable ===

// ---------------------------------------------------------------------------------
//		 MECodeTable
// ---------------------------------------------------------------------------------

MECodeTable::MECodeTable() : MEObject()
{
	setCode(kUniversalCode);
}


// ---------------------------------------------------------------------------------
//		 MECodeTable
// ---------------------------------------------------------------------------------

MECodeTable::MECodeTable(GeneticCode inCode) : MEObject()
{
	setCode(inCode);
}

// ---------------------------------------------------------------------------------
//		 ~MECodeTable
// ---------------------------------------------------------------------------------

MECodeTable::~MECodeTable()
{
}

// ---------------------------------------------------------------------------------
//		 MECodeTable::setCode
// ---------------------------------------------------------------------------------

void MECodeTable::setCode(GeneticCode inCode)
{
	TestBounds(inCode, numGeneticCodes);
	code=inCode;
	
	// Set the code from the built-in code tables using the
	// nucleotides order of T, C, A, G - the order given by the Genbank web site
	setTable((char *)kCodeTable[code], "TCAG");
	setName((char *)kGeneticCodeNames[code]);
}


// ---------------------------------------------------------------------------------
//		 MECodeTable::translateCodon
// ---------------------------------------------------------------------------------

int MECodeTable::translateCodon(char first, char second, char third)
{
	return table[kNucleotideNumbers[first]]
				[kNucleotideNumbers[second]]
				[kNucleotideNumbers[third]];
}

// ---------------------------------------------------------------------------------
//		 MECodeTable::translateCodon
// ---------------------------------------------------------------------------------

int MECodeTable::translateCodon(int first, int second, int third)
{
	return table[first][second][third];
}


// ---------------------------------------------------------------------------------
//		 MECodeTable::getUnambiguousCodon
// ---------------------------------------------------------------------------------

int MECodeTable::getUnambiguousCodon(const char *fst, const char *sec, const char *thd)
{
	int i, j, k;
	
	int n1=strlen(fst);
	int n2=strlen(sec);
	int n3=strlen(thd);
	
	bool identical=true;
	int aa=table[kNucleotideNumbers[fst[0]]][kNucleotideNumbers[sec[0]]][kNucleotideNumbers[thd[0]]];
	
	for (i=1; i<n1; i++) {
		for (j=0; j<n2; j++) {
			for (k=0; k<n3; k++) {
				if (aa!=table[kNucleotideNumbers[fst[i]]][kNucleotideNumbers[sec[j]]][kNucleotideNumbers[thd[k]]]) 
					identical=false;
			}
		}
	}
	
	if (identical)
		return aa;
	else
		return kAminoAcidUnknown;
}

// ---------------------------------------------------------------------------------
//		 MECodeTable::setTable
// ---------------------------------------------------------------------------------

void MECodeTable::setTable(char *inTable, char *inOrder)
{
	int i, j, k, n=0;
	
	numStopCodons=0;
		
	// First fill in the simple 4x4x4 table for just the standard ACGT using
	// the order given by inOrder.
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			for (k=0; k<4; k++) {
				table[kNucleotideNumbers[inOrder[i]]]
						[kNucleotideNumbers[inOrder[j]]]
						[kNucleotideNumbers[inOrder[k]]]=kAminoAcidNumbers[inTable[n]];
				if (kAminoAcidNumbers[inTable[n]]==kAminoAcidStop)
					numStopCodons++;
					
				n++;
			}
		}
	}
	
	// Constructs the full table making use of ambiguity
	// i.e. if there is a Y at a codon position where both pyramidines produce
	// the same amino acid then use that aa otherwise ?.
	for (i=0; i<kNumNucleotides; i++) {
		for (j=0; j<kNumNucleotides; j++) {
			for (k=0; k<kNumNucleotides; k++) {
				if (i>3 || j>3 || k>3) { // The simple 4x4x4 is already set
					if (i==kNucleotideGap || j==kNucleotideGap || k==kNucleotideGap) {
						// If there is a gap in all three codon positions then the
						// translation gives a gap, otherwise the aa is unknown.
						if (i==kNucleotideGap && j==kNucleotideGap && k==kNucleotideGap)
							table[i][j][k]=kAminoAcidGap;
						else 
							table[i][j][k]=kAminoAcidUnknown;
					} else
						// If the codon can be unambiguously translated then fine, 
						// otherwise the aa is unknown.
						table[i][j][k]=getUnambiguousCodon( kNucleotideMaps[i],
															kNucleotideMaps[j],
															kNucleotideMaps[k]);
				}
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MECodeTable::getTable
// ---------------------------------------------------------------------------------

void MECodeTable::getTable(char *outTable)
{
	int i, j, k, n=0;
	
	// First fill in the simple 4x4x4 table for just the standard ACGT
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			for (k=0; k<4; k++) {
				outTable[n]=kAminoAcids[table[i][j][k]];
				n++;
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MECodeTable::readObject
// ---------------------------------------------------------------------------------

void MECodeTable::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
	
	// Reads the code table in the form: AAA=K, AAC=N, ... UUU=F
	// These are read over the base table so only differences are required
	// The codons can be DNA or RNA
	int i, j, k, n=0;
	char codon[4], aa;
	codon[3]='\0';
	
	// Read the base table
	i=kUniversalCode; // Default to universal
	if (stream->get("Code", &i, kGeneticCodeNames, numGeneticCodes))
		code=(GeneticCode)i;
	
	// Now read in those codons that differ
	char uniTable[65];
	strcpy(uniTable, kCodeTable[code]);
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			for (k=0; k<4; k++) {
				codon[0]=kRNATranslateTable[i];
				codon[1]=kRNATranslateTable[j];
				codon[2]=kRNATranslateTable[k];
				if (stream->get(codon, &aa))
					uniTable[n]=aa;
				n++;
			}
		}
	}
	
	setTable(uniTable);
}

// ---------------------------------------------------------------------------------
//		 MECodeTable::writeObject
// ---------------------------------------------------------------------------------

void MECodeTable::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
	
	// Writes the code table in the form: AAA=K, AAC=N, ... UUU=F
	// The codons are written as RNA
	char codon[4];
	codon[3]='\0';
	
	// First write the code on which this one is based
	stream->put("Code", kGeneticCodeNames[code]);
	
	// We only write differences from the above...
	int i, j, k;
	const char *P=kCodeTable[code];
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			for (k=0; k<4; k++) {
				codon[0]=kRNATranslateTable[i];
				codon[1]=kRNATranslateTable[j];
				codon[2]=kRNATranslateTable[k];
				if (*P!=kAminoAcids[table[i][j][k]])
					stream->put(codon, kAminoAcids[table[i][j][k]]);
				P++;
			}
		}
	}
}


