/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MECreator.cpp

				MECreator Class
	
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEExceptions.h"
#include "CWriter.h"
#include "MEBasicAlignment.h"
#include "MELink.h"
#include "METree.h"
#include "MEPhylogeny.h"
#include "MECreator.h"

// Whether the format is supported for output
const bool kOutputFormatSupported[numFormatTypes] = { 
	false,	//	kFormatUnknown
	true,	//	kFormatNEXUS
	true, 	//	kFormatXML
	true, 	//	kFormatPHYLIP
	true, 	//	kFormatMOLPHY
	true, 	//	kFormatMEGA
	true, 	//	kFormatFLAT
	true,	//	kFormatNBRF
	true, 	//	kFormatFASTA
	true, 	//	kFormatCLUSTAL
	false, 	//	kFormatGCG
	false, 	//	kFormatGENBANK
	false, 	//	kFormatPAG
	true, 	//	kFormatCAIC
	false, 	//	kFormatCOMPARE
	true, 	//	kFormatTAB
	true, 	//	kFormatTree
	true, 	//	kFormatPAMLTree
	true 	//	kFormatPretty
}; 

#pragma mark === MECreator ===

// ---------------------------------------------------------------------------------
//		 MECreator
// ---------------------------------------------------------------------------------

MECreator::MECreator()
{
	setWriter(NULL);
	setIsInterleaved(false);
	setNumColumns(0);
	setRepeatTaxa(true);
}

MECreator::MECreator(CWriter *inWriter)
{
	setWriter(inWriter);
	setIsInterleaved(false);
	setNumColumns(0);
	setRepeatTaxa(true);
}

MECreator::~MECreator()
{
}

void MECreator::setWriter(CWriter *inWriter)
{
	writer=inWriter;
}

// ---------------------------------------------------------------------------------
//		 MECreator::writeSequence
// ---------------------------------------------------------------------------------

// Writes a sequence of length maxLen. Wraps on 'width' characters, indenting
// subsequent lines by 'indent' and placing a space every 'space' characters. 
// An 'indent' less than zero will indent with a tab. Use a width or space
// less than zero not to wrap or put spaces in.

void MECreator::writeSequence(char *inSequence, int maxLen, int width, int indent, int space)
{
	char *P=inSequence;
	int i, n, m;
	
	n=0; m=0;
	for (i=0; i<maxLen; i++) {
		writer->writeChar(*P);
		P++; n++; m++;
		if (i<maxLen-1) {
			if (n==width) {
				writer->writeChar('\r');
				if (indent<0)
					writer->writeChar('\t');
				else if (indent>0)
					writer->writeSpaces(indent);
				n=0;
			} else if (m==space) {
				writer->writeChar(' ');
				m=0;
			}
		}
	}
}


void MECreator::writeSequenceData(
		MEBasicAlignment *inAlignment, 
		int inIndent, 
		int inTruncateNameLen,
		bool inUseTaxonNumbers,
		bool inIncludeInfo,
		bool inIncludeSpace,
		bool inDrawRuler,
		bool inRulerInComment,
		bool inUseMatchChar,
		int inSpaceEvery)
{
	int i, n, m, nameLen, maxNameLen, infoLen;
	
	n=inAlignment->getNumTaxa();
	m=inAlignment->getNumSites();
	
	char *name, *info;
	char nameStr[256];

	maxNameLen=0;
	for (i=0; i<n; i++) {
		nameLen = 0;
		
		if (inUseTaxonNumbers) {
			sprintf(nameStr, "%d", i + 1);
		} else {
			name = inAlignment->getSequenceLabel(i);
			if (name != NULL)
				strcpy(nameStr, name);
			else
				nameStr[0] = '\0';
		}
					
		nameLen = writer->getLabelLength(nameStr, inTruncateNameLen > 0);
	
		if (inTruncateNameLen > 0 && nameLen > inTruncateNameLen)
			nameLen = inTruncateNameLen;
		
		if (inIncludeInfo) {
			info=inAlignment->getSequenceInfo(i);
			if (info != NULL) {
				infoLen = strlen(info);
				if (infoLen > 0)
					infoLen += 3;
				nameLen += infoLen;
			}
		}
					
		if (nameLen > maxNameLen)
			maxNameLen = nameLen;
	}
		
	char *seqStr1, *seqStr2;
	char str[64];
	
	if (interleaved) {
		seqStr1 = new char[numColumns + 1];
		seqStr2 = new char[numColumns + 1];
	} else {
		seqStr1 = new char[m + 1];
		seqStr2 = new char[m + 1];
	}
		
	int b = 0, c, d, e, j;
	while (b < m) {
		if (interleaved) {
			c = numColumns;
			if (b + c > m)
				c = m - b;
		} else if (numColumns > 0)
			c = numColumns;
		else
			c = m;
					
		if (inDrawRuler) {
			if (inIndent > 0)
				writer->writeSpaces(inIndent);
			else if (inIndent < 0)
				writer->writeTabs(-inIndent);
				
			if (inRulerInComment) {
				writer->writeString("[");
				writer->writeSpaces(maxNameLen);
			} else
				writer->writeSpaces(maxNameLen + 1);
				
			d = b;
			while (d < b + c) {
				sprintf(str, "%d", d + 1); 
				writer->writeString(str);
				
				e = 10 - strlen(str);
				if ((d + e) > (b + c))
					e = (b + c) - d;
					
				writer->writeSpaces(e);

				d += 10;
			}
			if (inRulerInComment)
				writer->writeString("]\r");
			else
				writer->writeString("\r");
			
			if (inIndent > 0)
				writer->writeSpaces(inIndent);
			else if (inIndent < 0)
				writer->writeTabs(-inIndent);

			if (inRulerInComment) {
				writer->writeString("[");
				writer->writeSpaces(maxNameLen);
			} else
				writer->writeSpaces(maxNameLen + 1);
				
			d = b;
			while (d < b + c) {
				writer->writeString("|");
				
				e = 9;
				if ((d + e) > (b + c))
					e = (b + c) - d;
					
				writer->writeSpaces(e);
				
				d += 10;
			}
			if (inRulerInComment)
				writer->writeString("]\r");
			else
				writer->writeString("\r");
		}
					
		for (i=0; i<n; i++) {
			if (inIndent > 0)
				writer->writeSpaces(inIndent);
			else if (inIndent < 0)
				writer->writeTabs(-inIndent);

			if (b == 0 || repeatTaxa) {
				if (inUseTaxonNumbers) {
					sprintf(nameStr, "%d", i + 1);
				} else {
					name = inAlignment->getSequenceLabel(i);
					if (name != NULL)
						strcpy(nameStr, name);
					else
						nameStr[0] = '\0';
				}
				
				nameLen = writer->getLabelLength(nameStr, inTruncateNameLen > 0);
					
				if (inTruncateNameLen > 0) {
					if (nameLen > inTruncateNameLen)
						nameLen = inTruncateNameLen;
					nameStr[nameLen] = '\0';	
					writer->writeLabel(nameStr, true);
				} else {
					if (nameLen > 0)
						writer->writeLabel(nameStr);
				}
						
				info=inAlignment->getSequenceInfo(i);
				if (info != NULL)
					infoLen = strlen(info);
				else
					infoLen = 0;
					
				if (infoLen > 0) {
					writer->writeChar(' ');
					writer->writeChar('[');
					writer->writeString(info);
					writer->writeChar(']');
					nameLen += infoLen + 3;
				}

				writer->writeSpaces(maxNameLen - nameLen);
			} else {
				writer->writeSpaces(maxNameLen);
			}
			
			if (inIncludeSpace)
				writer->writeSpaces(1);
				
			if (interleaved) {
				if (i == 0 || !inUseMatchChar) {
					inAlignment->getSequenceString(i, seqStr1, b, c);
					writeSequence(seqStr1, c, -1, 0, inSpaceEvery);
				} else {
					inAlignment->getSequenceString(i, seqStr2, b, c);
					char *P = seqStr1;
					char *Q = seqStr2;
					for (j = 0; j < c; j++) {
						if (*P == *Q)
							*Q = '.';
						P++;
						Q++;
					}
					writeSequence(seqStr2, c, -1, 0, inSpaceEvery);
				}
			} else {
				if (i == 0 || !inUseMatchChar) {
					inAlignment->getSequenceString(i, seqStr1, 0, m);
					writeSequence(seqStr1, m, c, 0, inSpaceEvery);
				} else {
					inAlignment->getSequenceString(i, seqStr2, 0, m);
					char *P = seqStr1;
					char *Q = seqStr2;
					for (j = 0; j < m; j++) {
						if (*P == *Q)
							*Q = '.';
						P++;
						Q++;
					}
					writeSequence(seqStr2, m, c, 0, inSpaceEvery);
				}
			}
			writer->writeChar('\r');
		}
		writer->writeString("\r");
		
		if (interleaved)
			b += c;
		else
			b += m;
	}
	
	delete [] seqStr1;
	delete [] seqStr2;
}

#pragma mark === Tree Writing ===

// ---------------------------------------------------------------------------------
//		 MECreator::writeBranch
// ---------------------------------------------------------------------------------

void MECreator::writeBranch(METree *inTree, MELink *inLink)
{
	double length;
	MELink *link=inTree->getDescendantNode(inLink, &length);
	writeNode(inTree, link);
		
	if (inTree->getHasLengths()) {
		writer->writeChar(':');	
		writer->writeDouble(length, 8);
	}
}

// ---------------------------------------------------------------------------------
//		 MECreator::writeNode
// ---------------------------------------------------------------------------------

void MECreator::writeNode(METree *inTree, MELink *inLink)
{
	if (inLink != NULL && inLink->getIsTip()) {
		if (inTree->getUseNumericalTips())
			writer->writeInteger(inTree->getTipNumber(inLink)+1);
		else
			writer->writeLabel(inTree->getTipName(inLink));
	} else {
		writer->writeChar('(');	
		
		MELink *link=inTree->getFirstBranch(inLink);
		while (link != NULL) {
			writeBranch(inTree, link);
			link=inTree->getNextBranch(inLink, link);
			if (link!=NULL)
				writer->writeChar(',');	
		}
		
		writer->writeChar(')');	
	}
}


// ---------------------------------------------------------------------------------
//		 MECreator::writeTree
// ---------------------------------------------------------------------------------

void MECreator::writeTree(METree *inTree)
{
	writeNode(inTree, NULL);
	writer->writeChar(';');	
}

// ---------------------------------------------------------------------------------
//		 MECreator::writeLineage
// ---------------------------------------------------------------------------------

void MECreator::writeLineage(MEPhylogeny *inPhylogeny, long inNode, long &ioTip)
{
	long descendant = inPhylogeny->getDescendant(inNode);
	
	if (descendant == -1) {
		writer->writeInteger(ioTip);
		ioTip++;
	} else {
		writer->writeChar('(');	
		
		writeLineage(inPhylogeny, descendant, ioTip);

		double length = inPhylogeny->getDeathTime(descendant) - inPhylogeny->getBirthTime(descendant);
		writer->writeChar(':');	
		writer->writeDouble(length, 8);
		writer->writeChar(',');	

		long sibling = inPhylogeny->getSibling(descendant);
		long d1;
		
		while (sibling != descendant) {
			writeLineage(inPhylogeny, sibling, ioTip);			
			length = inPhylogeny->getDeathTime(sibling) - inPhylogeny->getBirthTime(sibling);
			writer->writeChar(':');	
			writer->writeDouble(length, 8);

			sibling = inPhylogeny->getSibling(sibling);
			
			if (sibling != descendant)
				writer->writeChar(',');	
		}
		
		writer->writeChar(')');	
	}
	
}

// ---------------------------------------------------------------------------------
//		 MECreator::writePhylogeny
// ---------------------------------------------------------------------------------

void MECreator::writePhylogeny(MEPhylogeny *inPhylogeny)
{
	long tip = 1;
	writeLineage(inPhylogeny, 0, tip);
	writer->writeChar(';');	
}


void MECreator::createAlignment(MEBasicAlignment *inAlignment)
{
	throw MEUnsupportedData("MEBasicAlignment");
}

void MECreator::createTree(METree *inTree)
{
	throw MEUnsupportedData("METree");
}

void MECreator::createTreeList(METreeList *inTreeList)
{	
	throw MEUnsupportedData("METreeList");
}

void MECreator::createPhylogeny(MEPhylogeny *inPhylogeny)
{	
	throw MEUnsupportedData("MEPhylogeny");
}

void MECreator::createPhylogenyList(MEPhylogenyList *inPhylogenyList)
{	
	throw MEUnsupportedData("MEPhylogenyList");
}

void MECreator::createDistanceMatrix(MEDistanceMatrix *inDistanceMatrix)
{
	throw MEUnsupportedData("MEDistanceMatrix");
}