/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEDataColumn.cpp

		MEDataColumn represents a column of data in an MEDataMatrix. It is 
		derived from MEVariate.

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MEObject.h"
#include "MEDataColumn.h"

#pragma mark === MEDataColumn ===

// ---------------------------------------------------------------------------------
//	MEDataColumn::MEDataColumn()
// ---------------------------------------------------------------------------------

MEDataColumn::MEDataColumn() : MEVariate(NULL, true)
{
	isLogged=isDiscrete=isDichotomous=false;
	characteristicsKnown=false;
}

// ---------------------------------------------------------------------------------
//	@destructor
//	@description
//		Delete everything that isn't NULL, as long as it is owned by this.
// ---------------------------------------------------------------------------------
MEDataColumn::~MEDataColumn()
{
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::setIsLogged
// ---------------------------------------------------------------------------------

void MEDataColumn::setIsLogged(bool inIsLogged)
{
	if (isLogged==inIsLogged)
		return;
		
	if (inIsLogged && getMin()<=0.0)
		throw MENegativeLog();
	isLogged=inIsLogged;
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::getIsLogged
// ---------------------------------------------------------------------------------

bool MEDataColumn::getIsLogged()
{
	return isLogged;
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::getIsDiscrete
// ---------------------------------------------------------------------------------

bool MEDataColumn::getIsDiscrete()
{
	if (!characteristicsKnown)
		calcCharacteristics();
		
	return isDiscrete;
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::getIsDichotomous
// ---------------------------------------------------------------------------------

bool MEDataColumn::getIsDichotomous()
{
	if (!characteristicsKnown)
		calcCharacteristics();
		
	return isDichotomous;
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::getRow
// ---------------------------------------------------------------------------------

double MEDataColumn::getRow(int index)
{
	double f=getVal(index);
	if (isLogged) {
		ThrowIf(f<=0.0);
		return log(f);
	} else
		return f;
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::setRow
// ---------------------------------------------------------------------------------

void MEDataColumn::setRow(int index, double f)
{
	setVal(index, f);
	characteristicsKnown=false;
}

// ---------------------------------------------------------------------------------
//		MEDataColumn::calcCharacteristics
// ---------------------------------------------------------------------------------

void MEDataColumn::calcCharacteristics()
{
	int numRows=getN();
	
	int i=0;
	while (i<numRows && getRow(i) == NAN)
		i++;
		
	if (i==numRows) {
		isDiscrete=isDichotomous=false;
		characteristicsKnown=true;
		return;
	}
	
	isDiscrete=isDichotomous=true;
	int value1, value2;
	bool foundTwo=false;
	
	double d=getRow(i);
	if (d!=floor(d)) {
		isDiscrete=isDichotomous=false;
		characteristicsKnown=true;
		return;
	}
	
	value1=floor(d);
	while (i<numRows) {
		while (i<numRows && getRow(i) == NAN)
			i++;
		
		if (i<numRows) {
			d=getRow(i);
			if (isDiscrete) {
				if (d!=floor(d)) {
					isDiscrete=isDichotomous=false;
					characteristicsKnown=true;
					return;
				}
				if (!foundTwo) {
					if (floor(d)!=value1) {
						value2=floor(d);
						foundTwo=true;
					}
				} else {
					if (floor(d)!=value1 && floor(d)!=value2) 
						isDichotomous=false;
				}
			}
			i++;
		}
	}
	
	characteristicsKnown=true;
}

