/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEDataMatrix.cpp

				MEDataMatrix Class
					
	This matrix form is for storing columns of data with one row per taxon.
	It is descended from METaxonMatrix, but implements the storage of the
	data differently by overriding setSize, getElement and setElement. The
	columns are stored as MEDataColumns in a MEList. This allows the dynamic
	adding and removal of columns.
	
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "MEExceptions.h"

#include "MEDataColumn.h"
#include "MEDataMatrix.h"
#include "METaxon.h"
#include "METaxonList.h"
#include "MEList.h"

#include "MEStream.h"

#pragma mark === MEDataMatrix Class ===

// ---------------------------------------------------------------------------------
//		 MEDataMatrix
// ---------------------------------------------------------------------------------

MEDataMatrix::MEDataMatrix() : METaxonMatrix()
{
	NewClass(columns, MEList());
}


// ---------------------------------------------------------------------------------
//		 MEDataMatrix
// ---------------------------------------------------------------------------------

MEDataMatrix::MEDataMatrix(int inDimension) : METaxonMatrix(inDimension)
{
	NewClass(columns, MEList());
}


// ---------------------------------------------------------------------------------
//		 MEDataMatrix
// ---------------------------------------------------------------------------------

MEDataMatrix::MEDataMatrix(int inNumRows, int inNumCols) : 
	METaxonMatrix(inNumRows, inNumCols)
{
	NewClass(columns, MEList());
}


// ---------------------------------------------------------------------------------
//		 ~MEDataMatrix
// ---------------------------------------------------------------------------------

MEDataMatrix::~MEDataMatrix()
{
	if (columns!=NULL)
		delete columns;
}


// ---------------------------------------------------------------------------------
//		 MEDataMatrix::setSize
// ---------------------------------------------------------------------------------

// This resizes the matrix if neccessary. If it does then the contents are lost.

void MEDataMatrix::setSize(int inNumRows, int inNumCols)
{
	ThrowIf(inNumRows<0 || inNumCols<0);
	
	if (inNumRows==numRows && inNumCols==numCols)
		return;
		
	numRows=inNumRows;
	numCols=inNumCols;
	
	numElements=numRows*numCols;
		
	columns->removeAll();
		
	if (numElements>0) {
		MEDataColumn *column;
		for (int i=0; i<numCols; i++) {
			NewClass(column, MEDataColumn());
			columns->add(column);
		}
	}
	
	isSymmetric=false;
	isSymmetricKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::getColumnList
// ---------------------------------------------------------------------------------

MEList *MEDataMatrix::getColumnList()
{
	return columns;
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::getColumnObject
// ---------------------------------------------------------------------------------

MEObject *MEDataMatrix::getColumnObject(int col)
{
	if (columns!=NULL) {
		MEObject *obj=columns->getObject(col);
		return obj;
	} else
		return NULL;
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::setElement
// ---------------------------------------------------------------------------------

// Sets the element of the matrix at posistion (row, col) with value f.
void MEDataMatrix::setElement(int row, int col, double f)
{
	TestBounds(row, numRows);
	TestBounds(col, numCols);
	ThrowIfNull(columns);
		
	MEDataColumn *column=(MEDataColumn*)columns->getObject(col);
	column->setRow(row, f);		
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::setElement
// ---------------------------------------------------------------------------------

// Sets the index'th element of the matrix with value f.

void MEDataMatrix::setElement(int index, double f)
{
	int row=index / numCols;
	int col=index % numCols;
	setElement(row, col, f);		
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::getElement
// ---------------------------------------------------------------------------------

// Returns the value of the element of the matrix at posistion (row, col).

double MEDataMatrix::getElement(int row, int col)
{
	TestBounds(row, numRows);
	TestBounds(col, numCols);
	ThrowIfNull(columns);
		
	MEDataColumn *column=(MEDataColumn*)columns->getObject(col);
	return column->getRow(row);		
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::getElement
// ---------------------------------------------------------------------------------

// Returns the value of the index'th element of the matrix.

double MEDataMatrix::getElement(int index)
{
	int row=index / numCols;
	int col=index % numCols;
	return getElement(row, col);		
}

#pragma mark -
#pragma mark === Column/Row ===

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::getColumn
// ---------------------------------------------------------------------------------

MEDataColumn *MEDataMatrix::getColumn(int index)
{
	return (MEDataColumn *)columns->getObject(index);
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::addColumn
// ---------------------------------------------------------------------------------

void MEDataMatrix::addColumn()
{
	MEDataColumn *column;
	NewClass(column, MEDataColumn());
	columns->add(column, true);
	numCols=columns->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::addColumn
// ---------------------------------------------------------------------------------

void MEDataMatrix::addColumn(char *inName)
{
	MEDataColumn *column;
	NewClass(column, MEDataColumn());
	if (inName!=NULL)
		column->setName(inName);
	columns->add(column, true);
	numCols=columns->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::addColumn
// ---------------------------------------------------------------------------------

void MEDataMatrix::addColumn(MEDataColumn *inColumn, bool isOwner)
{
	ThrowIf(inColumn->getNumRows() != numRows);
	columns->add(inColumn, isOwner);
	numCols=columns->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::addRow
// ---------------------------------------------------------------------------------

void MEDataMatrix::addRow()
{
	setTaxonIndex(numRows, numRows);
	numRows++;
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::addRow
// ---------------------------------------------------------------------------------

void MEDataMatrix::addRow(char *inName)
{
	if (taxonList==NULL) {
		NewClass(taxonList, METaxonList());
		taxonList->setOwner(this);
	}
	
	int index=taxonList->findName(inName);
	if (index<0) { // not in taxon list
		if (taxonList->getOwner()==this) { 
			// if we own the taxon list then add the taxon
			METaxon *taxon;
			NewClass(taxon, METaxon());
			taxon->setName(inName);
			index=taxonList->addTaxon(taxon);	
		} else
			throw METaxonError();
	}
	
	setTaxonIndex(numRows, index);

	numRows++;
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::setColumnLogged
// ---------------------------------------------------------------------------------

void MEDataMatrix::setColumnLogged(int index, bool inIsLogged)
{
	((MEDataColumn*)columns->getObject(index))->setIsLogged(inIsLogged);
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::getIsColumnLogged
// ---------------------------------------------------------------------------------

bool MEDataMatrix::getIsColumnLogged(int index)
{
	return ((MEDataColumn*)columns->getObject(index))->getIsLogged();
}

#pragma mark -
#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::readObject
// ---------------------------------------------------------------------------------

void MEDataMatrix::readObject(MEStream *stream)
{
	METaxonMatrix::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::readData
// ---------------------------------------------------------------------------------

void MEDataMatrix::readData(MEStream *stream)
{
	columns=(MEList*)stream->get("Columns");
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::writeObject
// ---------------------------------------------------------------------------------

void MEDataMatrix::writeObject(MEStream *stream)
{
	METaxonMatrix::writeObject(stream);
}

// ---------------------------------------------------------------------------------
//		 MEDataMatrix::writeData
// ---------------------------------------------------------------------------------

void MEDataMatrix::writeData(MEStream *stream)
{
	stream->put("Columns", columns);
}
