/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEDistanceMatrix.cpp

				MEDistanceMatrix Class
					
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "METaxonList.h"
#include "METree.h"
#include "MELink.h"
#include "MEAlignment.h"
#include "MEDistanceMatrix.h"
#include "MEPatternList.h"

#include "MEStream.h"

const char* kDistanceMethodNames[numDistanceMethods] =
	{ "Uncorrected", "P", "Poisson", "LogDet" };

#pragma mark === MEDistanceMatrix Class ===

// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix
// ---------------------------------------------------------------------------------

MEDistanceMatrix::MEDistanceMatrix() : METaxonMatrix()
{
	sourceAlignment = NULL;
	sourceMatrix = NULL;
	sourcePatterns = NULL;
}


// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix
// ---------------------------------------------------------------------------------

MEDistanceMatrix::MEDistanceMatrix(int inDimension) : METaxonMatrix(inDimension)
{
	sourceAlignment = NULL;
	sourceMatrix = NULL;
	sourcePatterns = NULL;
}


// ---------------------------------------------------------------------------------
//		 ~MEDistanceMatrix
// ---------------------------------------------------------------------------------

MEDistanceMatrix::~MEDistanceMatrix()
{
}


// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix::setSize
// ---------------------------------------------------------------------------------

void MEDistanceMatrix::setSize(int inNumRows, int inNumCols)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();
	
	MEMatrix::setSize(inNumRows, inNumCols);
}

// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix::setMatrixFromTree
// ---------------------------------------------------------------------------------
// Construct a patristic matrix from a tree

void MEDistanceMatrix::setMatrixFromTree(METree *inTree)
{
	ThrowIfNull(inTree);
	
	int i, j, n = inTree->countTipsWithMask();
	
	int *tipNos;
	NewArray(tipNos, int, n);
	
	setSize(n, n);

	setTaxonList(inTree->getTaxonList(), false, false);
	MELink *tip = inTree->getFirstTip();
	for (i = 0; i < n; i++) {
		setTaxonIndex(i, tip->taxonNo);
		tipNos[i] = tip->tipNo;
		tip = inTree->getNextTip(tip);
	}
	
	for (i = 0; i < n-1; i++) {
		for (j = i+1; j < n; j++) {
			setElement(i, j, inTree->getTipToTipDistance(tipNos[i], tipNos[j]));
		}
	}
	
	delete [] tipNos;
	
	fillDiagonal(0.0);
	makeSymmetric(false /*kFromUpperTriangle*/);
}


// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix::calcPDistances
// ---------------------------------------------------------------------------------

/*void MEDistanceMatrix::calcPDistances(MEPatternList *inPatterns)
{
	ThrowIfNull(inPatterns);

	int numTaxa = inPatterns->getNumTaxa();
	int numPatterns = inPatterns->getNumPatterns();
	int numStates = inPatterns->getNumStates();

	double weight, sumWeight, sumMatch, sum1, sum2, dist;
	int site, i, j, k;
	char *pattern, *P, *Q;
	
	ThrowIf(numTaxa == 0);
	
	setSize(numTaxa, numTaxa);
	clearMatrix();
	
	double *freqs = sourcePatterns->getEqualFrequencies();
//	double *freqs = sourcePatterns->getEmpiricalFrequencies();
	
	sumWeight = 0.0;
	for (site=0; site<numPatterns; site++) {
		pattern = inPatterns->getPattern(site);
		weight = inPatterns->getPatternWeight(site);
		sumWeight += weight;
		
		dist = 0.0;
		P = pattern;
		for (i=0; i<numTaxa; i++) {
		
			Q = P + numStates;
			for (j=i+1; j<numTaxa; j++) {
			
				sumMatch = 0.0;
				sum1 = 0.0;
				sum2 = 0.0;
				for (k=0; k<numStates; k++) {
					if (P[k] > 0) {
						sum1 += freqs[k];
						if (P[k] == Q[k])
							sumMatch += freqs[k];
					}
					if (Q[k] > 0)
						sum2 += freqs[k];
				}
				
				dist = 1.0 - (sumMatch / (sum1 * sum2));
				plus(i, j, dist * weight);
				
				Q += numStates;
 			}
			P += numStates;
 		}
 	}
	
	divideBy(sumWeight);
	
	makeSymmetric(false);
}*/

void MEDistanceMatrix::setDataSource(MEAlignment *inAlignment)
{
	sourceAlignment = inAlignment;
	sourcePatterns = NULL;
	sourceMatrix = NULL;
	_matrixIsKnown = false;
}

void MEDistanceMatrix::setDataSource(MEPatternList *inList)
{
	sourceAlignment = NULL;
	sourcePatterns = inList;
	sourceMatrix = NULL;
	_matrixIsKnown = false;
}

void MEDistanceMatrix::setDataSource(MEDistanceMatrix *inMatrix)
{
	sourceAlignment = NULL;
	sourcePatterns = NULL;
	sourceMatrix = inMatrix;
	_matrixIsKnown = false;
}

void MEDistanceMatrix::calcDistanceMatrix()
{
	if (sourceAlignment != NULL) {
		// get the distance matrix from the alignment
	} else if (sourcePatterns != NULL) {
		int numTaxa = sourcePatterns->getNumTaxa();
		int numPatterns = sourcePatterns->getNumPatterns();
		int numStates = sourcePatterns->getNumStates();

		double weight, sumWeight, sumMatch, sum1, sum2, pDist, cDist;
		int site, i, j, k;
		char *pattern;
		
		ThrowIf(numTaxa == 0);
		
		setSize(numTaxa, numTaxa);
		clearMatrix();
		
		double *freqs = sourcePatterns->getEqualFrequencies();
	//	double *freqs = sourcePatterns->getEmpiricalFrequencies();
		
		sumWeight = sourcePatterns->getTotalPatternWeight();	

		int u = 0, v = 0;

		for (i=0; i<numTaxa; i++) {
			v = u + numStates;
			for (j=i+1; j<numTaxa; j++) {
				pDist = 0.0;
				for (site=0; site<numPatterns; site++) {
					pattern = sourcePatterns->getPattern(site);
					weight = sourcePatterns->getPatternWeight(site);
	
					sumMatch = 0.0;
					sum1 = 0.0;
					sum2 = 0.0;
					for (k=0; k<numStates; k++) {
						if (pattern[u + k] != '0' && pattern[u + k] != '\0') {
							sum1 += 1.0;
							if (pattern[u + k] == pattern[v + k])
								sumMatch += 1.0;
						}
						if (pattern[v + k] != '0' && pattern[v + k] != '\0')
							sum2 += 1.0;
					}
					
					pDist += (1.0 - (sumMatch / (sum1 * sum2))) * weight;
				}
				pDist /= sumWeight;
				if (pDist < kSmallReal)
					pDist = 0.0;
				
				// Now do the distance matrix stuff:
				switch (dMethod) {
					case kUncorrectedDistance:
					case kPDistance:
						cDist = pDist;
					break;
					case kPoissonDistance:
						double fract = 1.0 - (1.0 / numStates);
						cDist = -fract * log(1.0 - pDist/fract);
					break;
				}
				
				setElement(i, j, cDist);
									
				v += numStates;
 			}
			u += numStates;
 		}
		
		makeSymmetric(false /*kFromUpperTriangle*/);
	} else if (sourceMatrix != NULL) {
		// get the distance matrix from another matrix (may add correction)
	}
}


/*void MEDistanceMatrix::calcDistanceMatrix()
{
	char statei, statej;
	double r;
	double 
		gA, gC, gG, gT, gR, gY, // relative frequencies of A, C, G, T, R (purine) and Y (pyrimidine)
		H, // Hamming distance
		P1, P2, Q, 
		p, // proportional distance
		d, 
		w, // weight of each pattern
		W;	// total weight of all patterns in the list
	int i, j, k, nTax = getNumRows(), x, y;
	MEMatrix Fxx(4,4);
	MEMatrix Fyy(4,4);
	MEMatrix m(4,4);
	MEMatrix D(nTax);
	
	// counts of each of the pairs:
	int mat[4][4];
	
	// map the nucleotides to the numbers 0..3: don't deal with other states at the moment?
	short nucMap[256];
	for (i = 0; i < 256; i++)
		nucMap[i] = 0;
	nucMap['A'] = nucMap['a'] = 0;
	nucMap['C'] = nucMap['c'] = 1;
	nucMap['G'] = nucMap['g'] = 2;
	nucMap['T'] = nucMap['t'] = 3;
	nucMap['U'] = nucMap['u'] = 3;
	// + other, ambiguity codes here?
	// raises the problem of what to do with the ambiguous sites: weight
	// by relative frequency of nucleotides?  Best option I think.
	
	if (sourcePatterns != NULL) {
		char *pattern;
		NewArray(pattern, char, nTax);
		int nPat = sourcePatterns->getNumPatterns();
		// calculate the distance matrix from the source matrix:
		setSize(nTax);
		sourcePatterns->getNucleotideFrequencies(gA, gC, gG, gT);
		W = sourcePatterns->getTotalPatternWeight();	
			// for (int i = 0; i < numpatterns; i++) W += sourcePatterns->getPatternWeight(i);
			
		gR = gA + gG;
		gY = gC + gT;
		for (i = 1; i < nTax; i++) {
			for (j = 0; j < i; j++) {
				P1 = P2 = Q = 0.0;
				r = 0.0;
				for (x = 0; x < 4; x++)
					for (y = 0; y < 4; y++)
						mat[x][y] = 0;
				
				Fxx = 0.0;
				Fyy = 0.0;
				for (k = 0; k < nPat; k++) {
					sourcePatterns->getPattern(k, pattern);
					w = sourcePatterns->getPatternWeight(k);
					statei = pattern[i];
					statej = pattern[j];
					if (strchr("ACGTU", statei) && (strchr("ACGTU", statej))) {
						r += w;
						x = nucMap[(short)statei];
						y = nucMap[(short)statej];
						mat[x][y] += w;
						
						// for LogDet & GC content:
						Fxx.setElement(x, x, Fxx.getElement(x, x) + w);
						Fyy.setElement(y, y, Fyy.getElement(y, y) + w);
					}
				}
				
//				P1 = (mat[nucMap['A']][nucMap['G']] + mat[nucMap[
				P1 = (mat[0][2] + mat[2][0]) / W;	// rel. freq. of changes A<->G
				P2 = (mat[1][3] + mat[3][1]) / W;	// rel. freq. of changes C<->T
				H = W - (mat[0][0] + mat[1][1] + mat[2][2] + mat[3][3]);
					// total mismatches = Hamming Distance
				p = H / W;	// proportional distance = Hamming Distance over total sequence length
				Q = p - P1 - P2;	// rel. freq. of changes A<->C, A<->T, C<->G, G<->T
				
				// Now do the distance matrix stuff:
				switch (dMethod) {
					case kDMUncorrected:
						// default to Hamming Distance
						dMethod = kDMHamming;
					case kDMHamming:
						d = H;
						break;
					case kDMP:
						d = p;
						break;
					case kDMJukesCantor:
						d = -0.75 * log(1.0 - p/0.75);
						break;
					case kDMK2P:
						d = -0.5 * log(1.0 - 2.0 * (P1 + P2) - Q) - 0.25 * log(1.0 - 2 * Q);
						break;
					case kDMTamNei:
						d = -((2.0*gA*gG) / gR) * log(1.0 - gR*P1/(2.0*gA*gG) - 0.5*Q/gR) 
							-((2.0*gT*gC) / gY) * log(1.0 - gY*P2/(2.0*gT*gC) - 0.5*Q/gY)
							-2.0*(gR*gY - (gA*gG*gY)/gR - (gT*gC*gR)/gY) * log(1.0-0.5*Q/(gR*gY));
						break;
					case kDMLogDet:
						for (x = 0; x < 4; x++)
							for (y = 0; y < 4; y++)
								m(x,y) = mat[x][y]/W;
						d = -log(m.det());
						Fxx /= W;
						Fyy /= W;
						Fxx *= Fyy;
						d = (d + log(Fxx.det())/2.0) / 4.0;
						break;
					case kDMGCContent:
						Fxx /= W;
						Fyy /= W;
						d = 0.0;
						for (x = 0; x < 4; x++) {
							w = Fxx.getElement(x,x) - Fyy.getElement(x,x);
							d += w*w;
						}
						d = sqrt(d);
						break;
					default:
						break;
				}
				D(j,i) = d;
			}
		}
		for (i = 1; i < nTax; i++)
			for (j = 0; j < i; i++)
				setElement(j, i, D(j,i));
				
		D.makeSymmetric(true);	// here the "true" means copy the lower triangle into the upper

		delete [] pattern;
	} else if (sourceMatrix != NULL) {
		// get the distance matrix from the pattern list
		setSize(sourceMatrix->getNumRows());
		for (i  = 1; i < nTax; i++) {
			for (j = 0; j < i; j++) {
				w = sourceMatrix->getElement(j,i);
				switch (dMethod) {
					case kDMUncorrected:
						dMethod = kDMHamming;
					case kDMHamming:
						d = w * nTax * (nTax - 1) / 2.0;
						break;
					case kDMP:
						d = w;
					case kDMJukesCantor:
						d = -0.75 * log(1.0 - w / 0.75);
						break;
					case kDMK2P:
					case kDMTamNei:
					case kDMLogDet:
					case kDMGCContent:
						// what to do here? Exception? Default to J-C?
					default:
						break;
				}
			}
		}
	} else {
		throw MEException("No data source in MEDistanceMatrix::calcDistanceMatrix()");
	}
}
*/

void MEDistanceMatrix::setDistanceMethod(DistanceMethod dm)
{
	if (dm == dMethod)
		return;
	dMethod = dm;
	calcDistanceMatrix();
}

#pragma mark -
#pragma mark --- Object I/O ---

// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix::readObject
// ---------------------------------------------------------------------------------

void MEDistanceMatrix::readObject(MEStream *stream)
{
	METaxonMatrix::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		 MEDistanceMatrix::writeObject
// ---------------------------------------------------------------------------------

void MEDistanceMatrix::writeObject(MEStream *stream)
{
	METaxonMatrix::writeObject(stream);
}

