/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEDoubleArray.h

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of doubles

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MEObject.h"
#include "MEDoubleArray.h"

#pragma mark === MEDoubleArray ===

// ---------------------------------------------------------------------------------
//		MEDoubleArray::MEDoubleArray
// ---------------------------------------------------------------------------------

//: Constructs an empty MEDoubleArray
//  The array will only be allocated when a double is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
MEDoubleArray::MEDoubleArray(int inInitial, int inIncrement) : MEObject()
{
	capacity=0;
	numValues=0;
	values=NULL;
	iterator=0;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::~MEDoubleArray
// ---------------------------------------------------------------------------------

//: The destructor
MEDoubleArray::~MEDoubleArray()
{
	if (values!=NULL)
		delete[] values;
}

// ---------------------------------------------------------------------------------
//		MEObject::cloneList
// ---------------------------------------------------------------------------------

//: Clones this list from an existing one.

void MEDoubleArray::clone(MEObject *src)
{
	clone(src);

	setInitial(((MEDoubleArray*)src)->getCapacity(), ((MEDoubleArray*)src)->getIncrement());
	
	removeAll();
	int i, n=((MEDoubleArray*)src)->getNumValues();
	for (i=0; i<n; i++)
		add(((MEDoubleArray*)src)->getValue(i));
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void MEDoubleArray::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void MEDoubleArray::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void MEDoubleArray::setIncrement(int inIncrement)
{
	increment=inIncrement;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::getValue
// ---------------------------------------------------------------------------------

//: Returns the double that is at an index.
double MEDoubleArray::getValue(int index)
{
	TestBounds(index, numValues);
	
	return values[index];
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::getSum
// ---------------------------------------------------------------------------------

//: Returns the sum of all the doubles
double MEDoubleArray::getSum()
{
	int i;
	double sum = 0.0;
	for (i = 0; i < numValues; i++)
		sum += values[i];
		
	return sum;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::getMaxValue
// ---------------------------------------------------------------------------------

//: Returns the maximum double in the list
double MEDoubleArray::getMaxValue()
{
	if (numValues < 1)
		return 0.0;
		
	int i;
	double max = values[0];
	for (i = 1; i < numValues; i++)
		if ( values[i] > max)
			max = values[i];
		
	return max;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::setObject
// ---------------------------------------------------------------------------------

//: Sets the double at an index
//  If index>numValues then numValues becomes index+1 with memory 
//  allocated if needed.
void MEDoubleArray::setValue(int index, double inValue)
{
	ThrowIf(index<0);

	if (index>=numValues) {
		numValues=index+1;
		checkCapacity(numValues);
	} 
		
	values[index]=inValue;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::add
// ---------------------------------------------------------------------------------

//: Adds an double to the end of the list
//  This expands the list's capacity if necessary.
int MEDoubleArray::add(double inValue)
{
	checkCapacity(numValues+1);

	values[numValues]=inValue;
	numValues++;
	
	return numValues-1;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::insert
// ---------------------------------------------------------------------------------

//: Inserts an double into the list at a position
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new double. If the double's new position equals
//  numObjects then this adds it to the end of the list.
void MEDoubleArray::insert(int index, double inValue)
{
	ThrowIf(index<0 || index>numValues+1);

	checkCapacity(numValues+1);
		
	for (int i=numValues; i>index; i--)
		values[i]=values[i-1];
	
	values[index]=inValue;
	numValues++;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::remove
// ---------------------------------------------------------------------------------

//: Removes the double at index and moves others up 
//  This function returns the double being removed.
double MEDoubleArray::remove(int index)
{
	TestBounds(index, numValues);

	double tmp=values[index];

	for (int i=index; i<numValues-1; i++)
		values[i]=values[i+1];
		
	numValues--;
	
	return tmp;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the doubles in the list
void MEDoubleArray::removeAll()
{
	numValues=0;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::swap
// ---------------------------------------------------------------------------------

//: Swaps a value at index1 with that at index2
void MEDoubleArray::swap(int index1, int index2)
{
	TestBounds(index1, numValues);
	TestBounds(index2, numValues);

	double val = values[index1];
	values[index1] = values[index2];
	values[index2] = val;
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::sort
// ---------------------------------------------------------------------------------

//: Sorts the values in the list 
//  By default this is in ascending order unless inAscending is false.
void MEDoubleArray::sort(bool inAscending)
{
	if (numValues > 1) {
	
			// Shellsort Algorithm
		
		int	hh = 1;						// Find starting "h" value
		int	stopH = numValues / 9;
		while (hh < stopH) {
			hh = 3 * hh + 1;
		}
		
		double tmp;
		
		int direction = (inAscending ? 1 : -1);
		
		for ( ; hh > 0; hh /= 3) {
				
			for (int step = hh + 1; step <= numValues; step++) {
				tmp=getValue(step - 1);
				
				long i;
				for (i = step - hh; i > 0; i -= hh) {
					if ((tmp > getValue(i-1) ? 1 : -1) * direction > 0) {
						break;
					}
					setValue(i + hh - 1, getValue(i-1));
				}
					
				setValue(i + hh - 1, tmp);
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the double list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new double array of this
//  size, copies all the old doubles into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void MEDoubleArray::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity>capacity) {
		double *newValues;
		int i;
		
		// Increase capacity
		while (capacity<requiredCapacity)
			capacity+=increment;
		
		// Allocate the new list.
		NewArray(newValues, double, capacity);
		
		if (values!=NULL) {
			// Copy the existing list into the new list.
			for (i=0; i<numValues; i++)
				newValues[i]=values[i];
	
			// Free the old list.
			delete[] values;
		}
		
		// Point the old list to the new one.
		values=newValues;
	}
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::readObject
// ---------------------------------------------------------------------------------

//: Reads the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list reading mechanism called by readData. This allows descendant classes
//  to read in all their fields before the list contents is read. This doesn't
//  bother reading capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MEDoubleArray::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::readData
// ---------------------------------------------------------------------------------

//: Read the data of this object (i.e. the elements of the list)
//  This calls MEStream's getList function which reads the number of objects
//  and all the objects in the list.
void MEDoubleArray::readData(MEStream *stream)
{
	stream->getDoubleArray(this);
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::writeObject
// ---------------------------------------------------------------------------------

//: Writes the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list writing mechanism called by readData. This allows descendant classes
//  to write out all their fields before the list contents is written. This doesn't
//  bother writing capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MEDoubleArray::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}

// ---------------------------------------------------------------------------------
//		MEDoubleArray::writeData
// ---------------------------------------------------------------------------------

//: Write the data of this object (i.e. the elements of the list)
//  This calls MEStream's putList function which writes the number of objects
//  and all the objects in the list.
void MEDoubleArray::writeData(MEStream *stream)
{
	stream->putDoubleArray(this);
}

