/******************************************************************************

Phylogeny Generator (PhyloGen) v1.0
Copyright (C) 2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of the PhyloGen program.

    PhyloGen is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PhyloGen is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEEpisodicSimulator.cpp

				MEEpisodicSimulator Class

 ******************************************************************************/

#include <stdio.h>

#include "CProgress.h"

#include "MEEpisodicSimulator.h"

#include "MEMaths.h"

#pragma mark === MEEpisodicSimulator ===

// =========================================================================
//	MEEpisodicSimulator::MEEpisodicSimulator
// =========================================================================

MEEpisodicSimulator::MEEpisodicSimulator() :
	MEBirthDeathSimulator()
{
	mNumEpisodes = 0;
}

// =========================================================================
//	MEEpisodicSimulator::~MEEpisodicSimulator
// =========================================================================

MEEpisodicSimulator::~MEEpisodicSimulator()
{
	for (int i = 0; i < mNumEpisodes; i++)
		delete mEpisode[i];
}
		
// =========================================================================
//	MEEpisodicSimulator::addEpisode
// =========================================================================

void MEEpisodicSimulator::addEpisode(MEEpisode *inEpisode)
{
	ThrowIf(mNumEpisodes == mMaxEpisodes);
	
	mEpisode[mNumEpisodes] = inEpisode;
	mNumEpisodes++;
}

// =========================================================================
//	MEEpisodicSimulator::clearEpisodes
// =========================================================================

void MEEpisodicSimulator::clearEpisodes()
{
	for (int i = 0; i < mNumEpisodes; i++)
		delete mEpisode[i];
	mNumEpisodes = 0;
}

// =========================================================================
//	MEEpisodicSimulator::simulate
// =========================================================================

void MEEpisodicSimulator::simulate()
{
	initialize();
	
	mPhylogeny->setBirthTime(0, mCurrentTime);

	mLastExtantLineages = 0;

	mLastEpisodeTime = 0.0;
	mLastCurrentTime = 0.0;
	
	if (mProgress != NULL) {
		mProgress->description("Generating phylogeny...");
		mProgress->reset();
		mProgress->progress(0, mNumEpisodes);
		mProgress->setSkip(1);
	}
		
	for (int i = 0; i < mNumEpisodes; i++) {
		switch (mEpisode[i]->getEpisodeType()) {
			case MEEpisode::kGrowthType:
				simulateGrowthEpisode(i, static_cast<MEGrowthEpisode *>(mEpisode[i]));
				mLastEpisodeTime = mCurrentTime - mLastCurrentTime;
			break;
			case MEEpisode::kConstantType:
				simulateConstantEpisode(i, static_cast<MEConstantEpisode *>(mEpisode[i]));
				mLastEpisodeTime = mCurrentTime - mLastCurrentTime;
			break;
			case MEEpisode::kExtinctionType:
				simulateExtinctionEpisode(i, static_cast<MEExtinctionEpisode *>(mEpisode[i]));
				// don't set mLastEpisodeTime because an extinction is instantaneous.
			break;
		}
		
		mLastExtantLineages = getNumExtantLineages();
		mLastCurrentTime = mCurrentTime;
		
		if (mProgress != NULL)
			mProgress->progress(i+1, mNumEpisodes);
	}

	int i, n = getNumExtantLineages();
	for (i = 0; i < n; i++)
		mPhylogeny->setDeathTime(mPhylogeny->getExtantLineage(i), mCurrentTime);
}

// =========================================================================
//	MEEpisodicSimulator::simulateGrowthEpisode
// =========================================================================

void MEEpisodicSimulator::simulateGrowthEpisode(int inEpisodeNo, MEGrowthEpisode *inEpisode)
{
	double dt;
	long extant;
	
	setBirthDeathRates(inEpisode->mBirthRate, inEpisode->mDeathRate);
	
	while (checkStopConditions(inEpisodeNo, inEpisode)) {
		dt = timeToNextEvent();
		
		mCurrentTime += dt;
		
		extant = mPhylogeny->getRandomExtantIndex();
		
		if (eventIsBirth())
			lineageBirth(extant, 2);
		else
			lineageDeath(extant);
			
		if (getNumExtantLineages() < 1)
			throw MEAllExtinctTreeSimError();
	}
}

// =========================================================================
//	MEEpisodicSimulator::simulateConstantEpisode
// =========================================================================

void MEEpisodicSimulator::simulateConstantEpisode(int inEpisodeNo, MEConstantEpisode *inEpisode)
{
	double dt;
	long extant;
	
	if (inEpisode->mStochastic)
		setBirthDeathRates(inEpisode->mBirthDeathRate, inEpisode->mBirthDeathRate);
	else
		setPureBirth(inEpisode->mBirthDeathRate);
	
	while (checkStopConditions(inEpisodeNo, inEpisode)) {
		dt = timeToNextEvent();
		
		mCurrentTime += dt;
		
		extant = mPhylogeny->getRandomExtantIndex();
		
		if (inEpisode->mStochastic) {
			if (eventIsBirth())
				lineageBirth(extant, 2);
			else
				lineageDeath(extant);
		} else {
			lineageBirth(extant, 2);
			
			extant = mPhylogeny->getRandomExtantIndex();
			lineageDeath(extant);
		}
					
		if (getNumExtantLineages() < 1)
			throw MEAllExtinctTreeSimError();
	}
}

// =========================================================================
//	MEEpisodicSimulator::simulateExtinctionEpisode
// =========================================================================

void MEEpisodicSimulator::simulateExtinctionEpisode(int inEpisodeNo, MEExtinctionEpisode *inEpisode)
{	
	if (inEpisode->mType == MEExtinctionEpisode::kProbabilityType) {
		int i, n = getNumExtantLineages();
		for (i = 0; i < n; i++) {
			if (MERandom::bRan(inEpisode->mProbability))
				lineageDeath(i);
		}
	} else {
		long target, extant;
		
		if (inEpisode->mType == MEExtinctionEpisode::kFractionType)
			target = (1.0 - inEpisode->mFraction) * getNumExtantLineages();
		else
			target = getNumExtantLineages() - inEpisode->mNumber;

		while (getNumExtantLineages() > target) {
			extant = mPhylogeny->getRandomExtantIndex();
			lineageDeath(extant);
		}
	}
}

// =========================================================================
//	MEEpisodicSimulator::checkStopConditions
// =========================================================================

bool MEEpisodicSimulator::checkStopConditions(int inEpisodeNo, MEDurationEpisode *inEpisode)
{
	bool go = true;
	
	switch (inEpisode->mStopType) {
		case MEDurationEpisode::kExtantStopType: 
			if (mPhylogeny->getNumExtantLineages() >= inEpisode->mExtantStop) {
				if (inEpisodeNo == mNumEpisodes - 1) 
					mCurrentTime += timeToNextEvent();
				
				go = false;
			}
		break;
		case MEDurationEpisode::kByExtantStopType: {
			long change = mPhylogeny->getNumExtantLineages() - mLastExtantLineages;
			if (change < 0.0)
				change *= -1.0;
			if (change >= inEpisode->mByExtantStop) {
				if (inEpisodeNo == mNumEpisodes - 1) 
					mCurrentTime += timeToNextEvent();
				
				go = false;
			}
		} break;
		case MEDurationEpisode::kTimeStopType:
			if (mCurrentTime >= inEpisode->mTimeStop) {
				mCurrentTime = inEpisode->mTimeStop;
				go = false;
			}
		break;
		case MEDurationEpisode::kByTimeStopType: {
			double change = fabs(mCurrentTime - mLastCurrentTime);
			if (change >= inEpisode->mByTimeStop) {
				go = false;
				mCurrentTime = mLastCurrentTime + inEpisode->mByTimeStop;
			}
		} break;
		case MEDurationEpisode::kMultipleStopType: {
			double change = fabs(mCurrentTime - mLastCurrentTime);
			if (change >= inEpisode->mMultipleTimeStop * mLastEpisodeTime) {
				go = false;
				mCurrentTime = mLastCurrentTime + (inEpisode->mMultipleTimeStop * mLastEpisodeTime);
			}
		} break;
	}

	return go;
}

// =========================================================================
//	MEEpisodicSimulator::describe
// =========================================================================

void MEEpisodicSimulator::describe(CReportRecord *ioReport)
{
	CParagraphElement *heading = new CParagraphElement(kHeadingLevel2);
	heading->add("Episodic Birth-Death model");
	ioReport->add(heading);
	
	CParagraphElement *para;
	CListElement *list;
	
	if (mNumEpisodes == 0) {
		para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
		para->add("No episodes defined.");
		ioReport->add(para);
	} else {
		for (int i = 0; i < mNumEpisodes; i++) {
			switch (mEpisode[i]->getEpisodeType()) {
				case MEEpisode::kGrowthType: {
					para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
					para->add("Growth Episode:");
					ioReport->add(para);

					MEGrowthEpisode *episode = static_cast<MEGrowthEpisode *>(mEpisode[i]);
					list = new CListElement(kBullettedList, 2);
					para = new CParagraphElement();
					para->add("Birth rate = ");
					para->add(episode->mBirthRate);
					list->add(para);

					para = new CParagraphElement();
					para->add("Death rate = ");
					para->add(episode->mDeathRate);
					list->add(para);

					switch (episode->mStopType) {
						case MEDurationEpisode::kExtantStopType: 
							para = new CParagraphElement();
							para->add("Stops when extant lineages = ");
							para->add((int)episode->mExtantStop);
							list->add(para);
						break;
						case MEDurationEpisode::kByExtantStopType:
							para = new CParagraphElement();
							para->add("Stops when extant lineages changes by = ");
							para->add((int)episode->mByExtantStop);
							list->add(para);
						break;
						case MEDurationEpisode::kTimeStopType:
						para = new CParagraphElement();
							para->add("Stop when time reaches = ");
							para->add(episode->mTimeStop);
							list->add(para);
						break;
						case MEDurationEpisode::kByTimeStopType:
							para = new CParagraphElement();
							para->add("Stop when time changes by = ");
							para->add(episode->mByTimeStop);
							list->add(para);
						break;
						case MEDurationEpisode::kMultipleStopType:
							para = new CParagraphElement();
							para->add("Stop when time changes by a multiple of previous episode = ");
							para->add(episode->mMultipleTimeStop);
							list->add(para);
						break;
					}

					ioReport->add(list);

				} break;
				case MEEpisode::kConstantType: {
					MEConstantEpisode *episode = static_cast<MEConstantEpisode *>(mEpisode[i]);

					para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
					para->add("Constant Size Episode ");
					if (episode->mStochastic) {
						para->add("(Stochastic):");
					} else {
						para->add("(Deterministic):");
					}
					ioReport->add(para);

					list = new CListElement(kBullettedList, 2);
					para = new CParagraphElement();
					para->add("Birth/Death rate = ");
					para->add(episode->mBirthDeathRate);
					list->add(para);

					switch (episode->mStopType) {
						case MEDurationEpisode::kExtantStopType: 
							para = new CParagraphElement();
							para->add("Stops when extant lineages = ");
							para->add((int)episode->mExtantStop);
							list->add(para);
						break;
						case MEDurationEpisode::kByExtantStopType:
							para = new CParagraphElement();
							para->add("Stops when extant lineages changes by = ");
							para->add((int)episode->mByExtantStop);
							list->add(para);
						break;
						case MEDurationEpisode::kTimeStopType:
						para = new CParagraphElement();
							para->add("Stop when time reaches = ");
							para->add(episode->mTimeStop);
							list->add(para);
						break;
						case MEDurationEpisode::kByTimeStopType:
							para = new CParagraphElement();
							para->add("Stop when time changes by = ");
							para->add(episode->mByTimeStop);
							list->add(para);
						break;
						case MEDurationEpisode::kMultipleStopType:
							para = new CParagraphElement();
							para->add("Stop when time changes by a multiple of previous episode = ");
							para->add(episode->mMultipleTimeStop);
							list->add(para);
						break;
					}
					
					ioReport->add(list);

				} break;
				case MEEpisode::kExtinctionType: {
					MEExtinctionEpisode *episode = static_cast<MEExtinctionEpisode *>(mEpisode[i]);

					para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
					para->add("Extinction Episode:");
					ioReport->add(para);

					list = new CListElement(kBullettedList, 2);
					if (episode->mType == MEExtinctionEpisode::kFractionType) {
					
						para = new CParagraphElement();
						para->add("With ");
						para->add(episode->mFraction);
						para->add(" of extant lineages becoming extinct");
						list->add(para);
						
					} else if (episode->mType == MEExtinctionEpisode::kNumberType) {
					
						para = new CParagraphElement();
						para->add("With ");
						para->add((int)episode->mNumber);
						para->add(" extant lineages becoming extinct");
						list->add(para);
							
					} else if (episode->mType == MEExtinctionEpisode::kProbabilityType) {
					
						para = new CParagraphElement();
						para->add("With a ");
						para->add(episode->mProbability);
						para->add(" probability of extant lineages becoming extinct");
						list->add(para);
					}
					
					ioReport->add(list);

				} break;
			}
		}
	}
}
