/******************************************************************************

Phylogeny Generator (PhyloGen) v1.0
Copyright (C) 2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of the PhyloGen program.

    PhyloGen is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PhyloGen is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEEpisodicSimulator.h

		Interface for the MEEpisodicSimulator Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Trees"
 
#ifndef _MEEPISODIC_SIMULATOR_H_
#define _MEEPISODIC_SIMULATOR_H_

#include "CExceptions.h"
#include "MEPhylogeny.h"
#include "MEBirthDeathSimulator.h"


// =========================================================================
//	class MEEpisode
// =========================================================================

class MEEpisode {
	public:
		enum EpisodeType {
			kGrowthType,
			kConstantType,
			kExtinctionType
		};	
		
		MEEpisode() { }
		
		virtual EpisodeType	getEpisodeType() = 0;
};

class MEDurationEpisode : public MEEpisode {
	public:
		enum StopType {
			kExtantStopType,
			kByExtantStopType,
			kTimeStopType,
			kByTimeStopType,
			kMultipleStopType
		};	
		
		StopType mStopType;
		int mExtantStop;
		int mByExtantStop;
		double mTimeStop;
		double mByTimeStop;
		double mMultipleTimeStop;
		
		MEDurationEpisode() : MEEpisode() {
		}
			
		virtual EpisodeType	getEpisodeType() { return kGrowthType; }
};

class MEGrowthEpisode : public MEDurationEpisode {
	public:
		double mBirthRate;
		double mDeathRate;
		
		MEGrowthEpisode() : MEDurationEpisode() {
		}
			
		virtual EpisodeType	getEpisodeType() { return kGrowthType; }
};

class MEConstantEpisode : public MEDurationEpisode {
	public:
		bool mStochastic;
		double mBirthDeathRate;
				
		MEConstantEpisode() : MEDurationEpisode()
			{ 
			}
			
		virtual EpisodeType	getEpisodeType() { return kConstantType; }
};

class MEExtinctionEpisode : public MEEpisode {
	public:
		enum ExtinctionType {
			kFractionType,
			kNumberType,
			kProbabilityType
		};	

		ExtinctionType mType;
		double mFraction;
		long mNumber;
		double mProbability;
						
		MEExtinctionEpisode() : MEEpisode()
			{ 
			}
			
		virtual EpisodeType	getEpisodeType() { return kExtinctionType; }
};


// =========================================================================
//	class MEEpisodicSimulator
// =========================================================================

class MEEpisodicSimulator : public MEBirthDeathSimulator {
	public:
		static const int	mMaxEpisodes = 100;
		
	protected:
		MEEpisode*			mEpisode[mMaxEpisodes];
		int					mNumEpisodes;
		
		long				mLastExtantLineages;

		double				mLastEpisodeTime;
		double				mLastCurrentTime;
		
	public:
		
							MEEpisodicSimulator();
		virtual				~MEEpisodicSimulator();
			
		virtual void		addEpisode(MEEpisode *inEpisode);
		virtual int			getNumEpisodes() { return mNumEpisodes; }
		virtual MEEpisode*	getEpisode(int index) { return mEpisode[index]; }
		virtual void		clearEpisodes();
		
		virtual void		simulate();

	protected:
		virtual	void 		simulateGrowthEpisode(int inEpisodeNo, MEGrowthEpisode *inEpisode);
		virtual	void 		simulateConstantEpisode(int inEpisodeNo, MEConstantEpisode *inEpisode);
		virtual	void 		simulateExtinctionEpisode(int inEpisodeNo, MEExtinctionEpisode *inEpisode);
		
		virtual bool		checkStopConditions(int inEpisodeNo, MEDurationEpisode *inEpisode);
		
	public:
		virtual void		describe(CReportRecord *ioReport);
};

#endif /* _MEEPISODIC_SIMULATOR_H_ */
