/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  MEExceptions.h

				Exception classes

 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = Exceptions
 
#ifndef _MEEXCEPTIONS_H_
#define _MEEXCEPTIONS_H_

#include "CExceptions.h"

#pragma mark === Exception Classes ===

#pragma mark === Class MEException ===

//: A base general-purpose exception class
//  This class provides a base for all MolEvol library exceptions.
//  It provides various constructors that allow the setting of 
//  various forms of messages, and numbers.
class MEException : public CException {
	public:
		MEException();
			//: Default constructor - no message
			
		MEException(const char *inMsg) : CException(inMsg) { }
			//: Constructor with simple message
					
		MEException(int inData) : CException(inData) { }
			//: Constructor with numerical data
					
		MEException(const char *inMsg, int inData) : CException(inMsg, inData) { }
			//: Constructor with message and number
					
					
		MEException(const char *inMsg1, const char *inMsg2) : CException(inMsg1, inMsg2) { }
			//: Constructor with two message strings
					
		MEException(const char *inMsg1, const char *inMsg2, int inData) : CException(inMsg1, inMsg2, inData) { }
			//: Constructor with two message strings and a number
};

#pragma mark === Bitset Exceptions ===

//: A class of exceptions for the MEBitset functions.
class MEBitsetError : public MEException {
	public:
		MEBitsetError(const char *msg) : MEException(msg) { };
		MEBitsetError(const char *msg1, const char *msg2) : MEException(msg1, msg2) { };
};

class MEBadBitsetElement : public MEBitsetError {
	public:
		MEBadBitsetElement(const char *msg) : MEBitsetError(msg) { };
};

class MEBadBitsetCell : public MEBitsetError {
	public:
		MEBadBitsetCell(const char *msg) : MEBitsetError(msg) { };
};

//: A base exception for the graph related functions, thrown
// by MEGraph and MEAbstractGraph.
class MEGraphError : public MEException {
	public:
		MEGraphError(const char *aMsg) : MEException(aMsg) { };
		MEGraphError(const char *aMsg1, const char *aMsg2) : MEException(aMsg1, aMsg2) { };
};

#pragma mark === Utility Exceptions ===

//: Thrown when a pointer is unexpectedly NULL
//  This is generally a bug in the program.
class MENullPointer : public CNullPointer { 
	public:
		MENullPointer(const char *inMsg, int inData) : CNullPointer(inMsg, inData) { };
};

//: Thrown when a class allocation failed
//  Probably due to lack of memory.
class MENewClassFailed : public CNewClassFailed {  
	public:
		MENewClassFailed(const char *inMsg, int inData) : CNewClassFailed(inMsg, inData) { };
};

//: Thrown when an array allocation failed
//  Probably due to lack of memory.
class MENewArrayFailed : public CNewArrayFailed {  
	public:
		MENewArrayFailed(const char *inMsg, int inData) : CNewArrayFailed(inMsg, inData) { };
};

//: Thrown when the assert macros (ThrowIf, ThrowIfNot) fail
//  Usually used to test for conditions which should never happen (i.e. bugs).
class MEAssertFailed : public CAssertFailed {  
	public:
		MEAssertFailed(const char *inMsg, int inData) : CAssertFailed(inMsg, inData) { };
};

//: Thrown by the TestBounds macro
class MEIndexOutOfBounds : public CIndexOutOfBounds {  
	public:
		MEIndexOutOfBounds() : CIndexOutOfBounds() { };
		MEIndexOutOfBounds(const char *inMsg, int inData) : CIndexOutOfBounds(inMsg, inData) { };
};

//: Thrown by the MEUndead class when it is asked to resurrect an unknown class
class MEUnknownClass : public MEException {  
	public:
		MEUnknownClass() : MEException("Unknown class") { };
};

//: Can be thrown in cases where the user has interrupted a long process.
class MEInterrupted : public CInterrupted {  
	public:
		MEInterrupted() : CInterrupted() { };
		MEInterrupted(const char *aMsg) : CInterrupted(aMsg) { };
};

#pragma mark === Input Parsing Exceptions ===

//: The base to a number of parser related exceptions
class MEParserError : public MEException { 
	public:
		MEParserError(const char *aMsg) : MEException(aMsg) { };
		MEParserError(const char *aMsg1, const char *aMsg2) : MEException(aMsg1, aMsg2) { };
};

//: Throw when the input file format is unknown
class MEUnknownFormat : public MEParserError { 
	public:
		MEUnknownFormat() : MEParserError("Unknown File Format") { };
};

//: Thrown when the input file format is not yet supported
class MEUnsupportedFormat : public MEParserError { 
	public:
		MEUnsupportedFormat(const char *aMsg) : MEParserError("Unsupported File Format: ", aMsg) { };
};

//: Thrown when the data type (e.g. alignment, tree, etc.) is not supported in the current file format
class MEUnsupportedData : public MEParserError { 
	public:
		MEUnsupportedData(const char *aMsg) : MEParserError("Unsupported Data Type: ", aMsg) { };
};

//: Thrown when a required field in the file format is missing
class MEMissingField : public MEParserError { 
	public:
		MEMissingField(const char *aMsg) : MEParserError("Missing field: ", aMsg) { };
};

//: Thrown when a field in the file format is duplicated
class MEDuplicateField : public MEParserError { 
	public:
		MEDuplicateField(const char *aMsg) : MEParserError("Duplicated field: ", aMsg) { };
};

//: Thrown when a field in the file format is duplicated
class MEAmbiguousField : public MEParserError { 
	public:
		MEAmbiguousField(const char *aMsg) : MEParserError("Ambiguous field: ", aMsg) { };
};

//: Thrown when a field in the file format is unknown
class MEUnknownField : public MEParserError { 
	public:
		MEUnknownField(const char *aMsg) : MEParserError("Unknown field: ", aMsg) { };
};

//: A general problem in the file format
class MEBadFormat : public MEParserError { 
	public:
		MEBadFormat(const char *aMsg) : MEParserError("Error in input format: ", aMsg) { };
};

//: A taxon name in the current file is unknown
class MEUnknownTaxon : public MEParserError { 
	public:
		MEUnknownTaxon(const char *aMsg) : MEParserError("Unknown taxon: ", aMsg) { };
};

//: A sequence in the current file is unexpectedly short
class MEShortSequence : public MEParserError { 
	public:
		MEShortSequence(const char *aMsg) : MEParserError("Sequence too short: ", aMsg) { };
};

//: A sequence in the current file is unexpectedly long
class MELongSequence : public MEParserError { 
	public:
		MELongSequence(const char *aMsg) : MEParserError("Sequence too long: ", aMsg) { };
};

//: An alignment in the current file has unexpectedly few taxa
class METooFewTaxa : public MEParserError { 
	public:
		METooFewTaxa() : MEParserError("Too few taxa") { };
};

//: A general problem with an interleaved alignment
class MEInterleaveError : public MEParserError { 
	public:
		MEInterleaveError() : MEParserError("Error reading as interleaved") { };
};


#pragma mark === Tree Parsing Exceptions ===

//: The base to a number of tree parsing related exceptions
class METreeParserError : public MEParserError { 
	public:
		METreeParserError(const char *aMsg) : MEParserError("Error reading tree: ", aMsg) { };
};

//: A tree is missing from the file
class METreeMissing : public METreeParserError { 
	public:
		METreeMissing() : METreeParserError("Tree missing") { };
};

//: A branch is missing from the tree
//  This is usually thrown due to an error in the tree's format.
class MEBranchMissing : public METreeParserError { 
	public:
		MEBranchMissing() : METreeParserError("Branch missing") { };
};

//: A matching bracket is missing from the tree
//  This is usually thrown due to an error in the tree's format.
class MEBracketMissing : public METreeParserError { 
	public:
		MEBracketMissing() : METreeParserError("Bracket missing") { };
};

//: A branch length is missing from the tree
//  This will be thrown if previous branches had branch lengths.
class MELengthsMissing : public METreeParserError { 
	public:
		MELengthsMissing() : METreeParserError("Branch lengths missing") { };
};

//: A a branch label of the tree is not numerical
class MENonNumericalLabel : public METreeParserError { 
	public:
		MENonNumericalLabel() : METreeParserError("Non numerical branch label") { };
};

#pragma mark === Stream Exceptions ===

//: A base class that provides exceptions for MEStream mechanism
class MEStreamError : public MEException { 
	public:
		MEStreamError(const char *inMsg) : MEException(inMsg) { };
		MEStreamError(const char *inMsg, int inData) : MEException(inMsg, inData) { };
		MEStreamError(const char *inMsg1, const char *inMsg2) : MEException(inMsg1, inMsg2) { };
};

//: Thrown when an object reference in the input MEStream is unknown 
//  At present the streaming mechanism cannot forward reference an object.
class MEUnknownObject : public MEStreamError { 
	public:
		MEUnknownObject(const char *inMsg) : MEStreamError("Unknown object reference, ", inMsg) { };
};

#pragma mark === File I/O Exceptions ===
/*
//: A base class that provides exceptions for the file in/out
//  These are throw by the file using classes (MEStream, MEReader and CWriter) 
//  and their descendants.
class MEFileError : public MEException { 
	public:
		MEFileError(const char *inMsg) : MEException(inMsg) { };
		MEFileError(const char *inMsg, int inData) : MEException(inMsg, inData) { };
		MEFileError(const char *inMsg1, const char *inMsg2) : MEException(inMsg1, inMsg2) { };
};

//: A file was not opened due to an error
//  If opening a file for reading, then this will most likely be due to it not 
//  existing. Otherwise it will be some sort of operating system error.
class MEOpenFailed : public MEFileError { 
	public:
		MEOpenFailed(const char *inMsg) : MEFileError("Failed to open file, ", inMsg) { };
};

//: Thrown when the end of file is reached during a read operation
//  This can be caught in order to detect the end of file and finish the reading process.
class MEEndOfFile : public MEFileError {
	public:
		MEEndOfFile() : MEFileError("Unexpected end of file") { };
};

//: Thrown when attempting to read an integer but failed
class MEIllegalInt : public MEFileError {
	public:
		MEIllegalInt() : MEFileError("Unable to read integer") { };
};

//: Thrown when attempting to read a double but failed
class MEIllegalDbl : public MEFileError {
	public:
		MEIllegalDbl() : MEFileError("Unable to read double") { };
};

//: Thrown when some sort of write error occurred
//  These will probably be operating system level errors.
class MEWriteError : public MEFileError {
	public:
		MEWriteError() : MEFileError("Error writing file") { };
		MEWriteError(const char *inMsg) : MEFileError("Error writing file: ", inMsg) { };
		MEWriteError(int inData) : MEFileError("Error writing file: ", inData) { };
};

//: Thrown when some sort of read error occurred
//  These will probably be operating system level errors.
class MEReadError : public MEFileError {
	public:
		MEReadError() : MEFileError("Error writing file") { };
		MEReadError(const char *inMsg) : MEFileError("Error reading file: ", inMsg) { };
		MEReadError(int inData) : MEFileError("Error reading file: ", inData) { };
};
*/
#pragma mark === Taxon Exceptions ===

//: A base exception for playing around with taxa
//  Thrown by METaxon, METaxonList and their descendants.
class METaxonError : public MEException { 
};

//: Thown when applying a taxon list to another object and the number of taxa do not match
//  Thrown by METree, METreeList, METaxonMatix, etc.
class MEBadNumTaxa : public METaxonError { 
};

#pragma mark === Tree Exceptions ===

//: A base exception for errors in the METree class
class METreeError : public MEException { 
	public:
		METreeError(const char *aMsg) : MEException(aMsg) { };
};

//: Thrown if taxon referenced by name or number was not present in the tree
class METaxonMissing : public METreeError { 
	public:
		METaxonMissing() : METreeError("Taxon missing in tree") { };
};

//: Thrown if taxon name is used twice in the same tree
class MEDuplicateTaxon : public METreeError { 
	public:
		MEDuplicateTaxon() : METreeError("Duplicate taxon in tree") { };
};

//: Thrown if an internal node referenced by number was not present in the tree
class MENodeNotFound : public METreeError { 
	public:
		MENodeNotFound() : METreeError("Node not found in tree") { };
};

//: Thrown if an MELink referenced by number was not present in the tree
class MELinkNotFound : public METreeError { 
	public:
		MELinkNotFound() : METreeError("Link not found in tree") { };
};

//: Thrown if the tree was required to be rooted but it wasn't
class METreeNotRooted : public METreeError { 
	public:
		METreeNotRooted() : METreeError("Tree not rooted") { };
};

//: Thrown if the tree contained no selected tips (i.e. with flag>0)
class MENoSelectedTips : public METreeError { 
	public:
		MENoSelectedTips() : METreeError("Tree contained no selected tips") { };
};

//: Thrown if the tree not fully connected
class METreeNotConnected : public METreeError { 
	public:
		METreeNotConnected() : METreeError("Tree not fully connected") { };
};

//: Thrown if a bad node height is attempted to be set
class MEBadNodeHeight : public METreeError { 
	public:
		MEBadNodeHeight() : METreeError("Bad node height") { };
};



#pragma mark === Matrix Exceptions ===

//: A base exception for errors in the MEMatrix class
class MEMatrixError : public MEException { 
	public:
		MEMatrixError(const char *aMsg) : MEException(aMsg) { };
};

//: Thrown if the matrix was required to be square but it wasn't
class MENotSquareMatrix : public MEMatrixError { 
	public:
		MENotSquareMatrix() : MEMatrixError("Matrix is not square") { };
};

//: Thrown if the matrix was required to be symmetrical but it wasn't
class MENotSymmetricMatrix : public MEMatrixError { 
	public:
		MENotSymmetricMatrix() : MEMatrixError("Matrix is not symmetrical") { };
};

//: Thrown when the number of rows in an METaxonMatrix doesn't match the number of taxa
class MEBadNumTaxaMatrix : public MEMatrixError { 
	public:
		MEBadNumTaxaMatrix() : MEMatrixError("Matrix has different number of rows to number of taxa") { };
};

//: Thrown by operations on a pair of matrices when they don't have compatible dimensions
class MEWrongDimensions : public MEMatrixError { 
	public:
		MEWrongDimensions() : MEMatrixError("Matrix has wrong dimensions") { };
};

//: Thrown when a logged column contains a negative or zero value
class MENegativeLog : public MEMatrixError { 
	public:
		MENegativeLog() : MEMatrixError("A negative or zero value in a logged column") { };
};

//: A class of exceptions for the MEBitMatrix functions.
class MEBitMatrixError : public MEException {
	public:
		MEBitMatrixError(const char *msg) : MEException(msg) { };
};

#pragma mark === Sequence Exceptions ===

//: A base exception for errors in the sequence classes
//  Thrown by MEBasicSequence, MESequence and their descendants.
class MESequenceError : public MEException {
	public:
		MESequenceError(const char *aMsg) : MEException(aMsg) { };
};

//: Not clear where this one is used, but hey
class MEBadSequenceLength : public MESequenceError { 
	public:
		MEBadSequenceLength() : MESequenceError("Bad sequence length") { };
};

//: Thrown when a sequence contains a state character that is not appropriate
class MEBadState : public MESequenceError { 
	public:
		MEBadState() : MESequenceError("Bad state character in sequence") { };
};

//: Thrown when a sequence has an inappropriate sequence type
class MEBadSequenceType : public MESequenceError { 
	public:
		MEBadSequenceType() : MESequenceError("Bad sequence type") { };
};


#pragma mark === Alignment Exceptions ===

//: A base exception for errors in the alignment classes 
//  Thrown by MEBasicAlignment, MEAlignment and their descendants.
class MEAlignmentError : public MEException { 
	public:
		MEAlignmentError(const char *aMsg) : MEException(aMsg) { };
};

//: Thrown when trying to add an inappropriate sequence to the alignment
class MEIncompatibleSequence : public MEAlignmentError { 
	public:
		MEIncompatibleSequence() : MEAlignmentError("Incompatible sequence type in alignment") { };
};

//: Thrown when trying to add an overlarge sub-alignment to the alignment
class MEBadInsertion : public MEAlignmentError { 
	public:
		MEBadInsertion() : MEAlignmentError("Alignment to be inserted is too big") { };
};

#pragma mark === Aligner Exceptions ===

//: A base exception for errors in the aligner classes 
//  Thrown by MEAligner and its descendants.
class MEAlignerError : public MEException { 
	public:
		MEAlignerError(const char *aMsg) : MEException(aMsg) { };
};

#endif /* _MEEXCEPTIONS_H_ */
