/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEGenbankParser.cpp

				MEGenbankParser Class
	
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"
#include "CReader.h"
#include "MEParser.h"
#include "MEGenbankParser.h"
#include "MEBasicAlignment.h"
#include "MEBasicSequence.h"
#include "MESequence.h"
#include "MEFeature.h"


/******************************************************************************
 MEGenbankParser
 
 ******************************************************************************/

MEGenbankParser::MEGenbankParser(CReader *inReader) : MEParser(inReader)
{
	parseFeatures=false;
	exonRanges=NULL;
	numExons=0;
	capacity=0;
	reader->setCommentDelimiters('\0', '\0', '\0');

	strcpy(missing, "?.");
	strcpy(match, "");
}

/******************************************************************************
 ~MEGenbankParser
 
 ******************************************************************************/

MEGenbankParser::~MEGenbankParser()
{
	if (exonRanges!=NULL)
		delete [] exonRanges;
}



/******************************************************************************
 checkExonCapacity
 
 ******************************************************************************/

void MEGenbankParser::checkExonCapacity(int newCapacity)
{
	if (capacity<newCapacity) {
		while (capacity<newCapacity)
			capacity+=10;
			
		int* newExonRanges=new int[capacity];
		if (exonRanges!=NULL) {
			for (int i=0; i<numExons*2; i++)
				newExonRanges[i]=exonRanges[i];
			delete [] exonRanges;
		}
		
		exonRanges=newExonRanges;
	}	
}

/******************************************************************************
 getNextTag
 
 ******************************************************************************/

int MEGenbankParser::getNextTag(char *token)
{
	int i;
	char ch;

	i=reader->skipSpace();
	ch=reader->readToken(token, 255, "=\r\n", true);
	if (ch=='\r' || ch=='\n')
		reader->unreadChar(ch);
		
	return i;
}


/******************************************************************************
 findTag
 
 ******************************************************************************/

int MEGenbankParser::findTag(char *token, int indent)
{
	int i;
	char ch;
	char st[256];
	bool found;
	
	do {
		i=reader->skipSpace();
		if (i == indent) {
			ch=reader->readToken(st, 255, "=\r\n", true);
			if (ch=='\r' || ch=='\n')
				reader->unreadChar(ch);
				
			found = (strcmp(st, token)==0);
		}
		if (!found)
			reader->skipToEndOfLine();

	} while (!found);
	
	return i;
}


/******************************************************************************
 readFeatureRange
 ******************************************************************************/

void MEGenbankParser::readFeatureRange()
{	
	char ch, st[256];
	int start, stop;
	
	reader->skipSpace("<>");
	reader->readToken(st, 255, "(.", true);
	if (strcmp(st, "JOIN")==0 || strcmp(st, "COMPLEMENT")==0) { 
				// Should actually do something about complementation
		numExons=0;
		do {
			reader->skipSpace("<>");
			ch=reader->readToken(st, 255, ".", true);
			if (sscanf(st, "%d", &start)!=1)
				throw CIllegalInt();
			reader->skipSpace("<.>");
			ch=reader->readToken(st, 255, ",)", true);
			if (sscanf(st, "%d", &stop)!=1)
				throw CIllegalInt();
				
			checkExonCapacity(numExons*2 + 2);
			exonRanges[numExons*2]=start;
			exonRanges[numExons*2 + 1]=stop;
			numExons++;
		} while ( ch!=')' );
	} else {
		if (sscanf(st, "%d", &start)!=1)
			throw CIllegalInt();
			
		reader->skipSpace("<.>");
		ch=reader->readToken(st, 255, "\r\n", true);
		reader->unreadChar(ch);
			
		if (sscanf(st, "%d", &stop)!=1)
			throw CIllegalInt();
			
		checkExonCapacity(2);
		numExons=1;
		exonRanges[0]=start;
		exonRanges[1]=stop;
	}
}


/******************************************************************************
 readFeatures
 ******************************************************************************/

void MEGenbankParser::readFeatures(MESequence *sequence)
{	
	int i, frame;
	MEFeature *feature, *feature1;
	FeatureType featureType;
	char name[MAX_NAME_LEN+1];
	char st[256];
	reader->skipToEndOfLine();
	int indent=getNextTag(st);
	while (indent>0) {
		featureType=kUnknownFeature;
		if (indent>=5) {
			if (strcmp(st, "CDS")==0) {
				featureType=kCDSFeature;
			} else if (strcmp(st, "RRNA")==0) {
				featureType=kRRNAFeature;
			} else if (strcmp(st, "TRNA")==0) {
				featureType=kTRNAFeature;
			} else if (strcmp(st, "MISC_FEATURE")==0) {
				featureType=kOtherFeature;
			}

			if (featureType!=kUnknownFeature) {
				readFeatureRange();
				feature=sequence->newFeature();
				feature->setFeatureType(featureType);
				
				feature->setStart(exonRanges[0]);
				feature->setLength(exonRanges[1]-exonRanges[0]+1);
				indent=getNextTag(st);
				while (indent>=21) {
				// We could get the codon_start and transl_table out of this too.
					if (strcmp(st, "/PRODUCT")==0) {
						reader->skipSpace("=");
						reader->readToken(st, 255, "\r\n");
						reader->unreadChar('\r');
						feature->setName(st);
					} else if (strcmp(st, "/NOTE")==0) { // probably should append remaining to desc
						reader->skipSpace("=");
						reader->readToken(st, 255, "\r\n");
						reader->unreadChar('\r');
						feature->setName(st);
					} else if (strcmp(st, "/GENE")==0) {
						reader->skipSpace("=");
						reader->readToken(st, 255, "\r\n");
						reader->unreadChar('\r');
						feature->setDescription(st);
					} else if (strcmp(st, "/CODON_START")==0) {
						reader->skipSpace("=");
						reader->readInteger(&frame, "\r\n");
						reader->unreadChar('\r');
						feature->setFrame(frame);
					}
					indent=getNextTag(st);
				}
				sequence->addFeature(feature);
				
				if (numExons>1) {
					for (i=1; i<numExons; i++) {
						feature1=sequence->newFeature();
						feature1->clone(feature);
						feature1->setFeatureType(kExonFeature);
						if (feature1->getName()!=NULL)
							sprintf(name, "%s - exon %d", feature1->getName(), i+1);
						else
							sprintf(name, "exon %d", i+1);
						feature1->setName(name);
							
						feature1->setStart(exonRanges[i*2]);
						feature1->setLength(exonRanges[1*2+1]-exonRanges[i*2]+1);
					}
					
					if (feature->getName()!=NULL)
						sprintf(name, "%s - exon 1", feature1->getName());
					else
						strcat(name, "exon 1");
					feature->setName(name);
					feature->setFeatureType(kExonFeature);
				}
			} else {
				reader->skipToEndOfLine();
				indent=getNextTag(st);
			}
		} else {	
			reader->skipToEndOfLine();
			indent=getNextTag(st);
		}
	}
	if (strcmp(st, "ORIGIN")==0)
		originFound=true;
}


/******************************************************************************
 readSequences
 
 	Read all the sequences in. The inTreeList is always ignored - there are
 	no trees in genbank format files.
 ******************************************************************************/

void MEGenbankParser::readAlignment(MEBasicAlignment *inAlignment)
{	
	int seqLen;
	bool done;
	char name[MAX_NAME_LEN+1];
	char desc[MAX_DESC_LEN+1];
	char acc[MAX_ACC_LEN+1];

	reader->skipSpace();
	
	try {
		findTag("LOCUS", 0);
	} catch (CEndOfFile) {
		throw MEMissingField("LOCUS");
	}

	MEBasicSequence *sequence;
	seqType=kDNASequence;

/*	done=false;
	while (!done) {
		sequence=inAlignment->newSequence(seqType);
		reader->readToken(name, MAX_NAME_LEN);
		sequence->setName(name);
		
		seqLen=reader->readInteger();
		sequence->setLength(seqLen);		

		try {
			findTag("DEFINITION", 0);
			reader->readLine(desc, 255);
			sequence->setDescription(desc);
		} catch (CEndOfFile) {
		}
		
		try {
			findTag("ACCESSION", 0);
			reader->readToken(acc, 255);
			sequence->setAccession(acc);
		} catch (CEndOfFile) {
		}
		
		originFound=false;
		if (parseFeatures) {
			try {
				findTag("FEATURES", 0);
				readFeatures((MESequence *)sequence);
			} catch (CEndOfFile) {
			}
		}
				
		try {
			if (!originFound)
				findTag("ORIGIN", 0);
		} catch (CEndOfFile) {
			throw MEMissingField("ORIGIN");
		}
		reader->skipToEndOfLine();
		
		readSequence(sequence->getSequence(), seqLen, "0123456789");
		inAlignment->addSequence(sequence);
		
		try {
			reader->findToken("//");
			reader->findToken("LOCUS");
		} catch (CEndOfFile) {
			done=true;
		}
	}
*/	
	char st[256];
	int indent;

	done=false;
	while (!done) {
		sequence=inAlignment->newSequence(seqType);
		reader->readToken(name, MAX_NAME_LEN);
		sequence->setName(name);
		
		seqLen=reader->readInteger();
		sequence->setLength(seqLen);		

		try {
			reader->skipToEndOfLine();
			originFound=false;
			do {
				indent = getNextTag(st);
				
				if (indent == 0) {
					if (strcmp(st, "DEFINITION")==0) {
						reader->readLine(desc, 255, false, false);
						sequence->setDescription(desc);
					} else if (strcmp(st, "ACCESSION")==0) {
						reader->readToken(acc, 255, false, false);
						sequence->setAccession(acc);
					} else if (strcmp(st, "DEFINITION")==0) {
						reader->readLine(desc, 255, false, false);
						sequence->setDescription(desc);
					} else if (strcmp(st, "FEATURES")==0) {
						originFound=false;
						if (parseFeatures) {
							try {
								readFeatures((MESequence *)sequence);
							} catch (CEndOfFile) {
							}
						}
						
					} else if (strcmp(st, "ORIGIN")==0) {
						originFound = true;
					} else
						reader->skipToEndOfLine();

				} else
					reader->skipToEndOfLine();
			} while (!originFound);
			
		} catch (CEndOfFile) {
			throw MEMissingField("ORIGIN");
		}
				
		readSequence(sequence->getSequence(), seqLen, "0123456789");
		inAlignment->addSequence(sequence);
		
		try {
			reader->findToken("//");
			reader->findToken("LOCUS");
		} catch (CEndOfFile) {
			done=true;
		}
	}
}


