/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEGlobal.cpp

	Global constants
				
	Constants and tables for the molecular evolution library.
	
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = Global

#include "MEGlobal.h"

//********************
#pragma mark -- Constants --

//: The names of the sequence types
const char* kFlagTypeNames[numFlagTypes] =
	{ "hidden", "disabled", "hidden & disabled", "missing" };

//: The names of the sequence types
const char* kSequenceTypeNames[numSequenceTypes] =
	{ "unknown", "DNA", "RNA", "AA", "Discrete", "Continuous" };

//: The names of the alignment types
const char* kAlignmentTypeNames[numAlignmentTypes] =
	{ "unknown", "Nucleotide", "Amino Acid", "Codon", "Discrete" };

//: The names of the sequence types - many more to be added
const char* kFeatureTypeNames[numFeatureTypes] =
	{ "unknown", "CDS", "Exon", "Intron", "rRNA", "tRNA", "Motif", "Other" };

//********************
#pragma mark -- Nucleotide Tables --

//: A table to translate state numbers (0-17) into character codes
const char kNucleotides[]="ACGTURYMWSKBDHVN?-";

//: A table to map state numbers (0-17) to their ambiguities
const char* kNucleotideMaps[kNumNucleotides]={
//	 A    C	   G    T    U    R     Y     M     W     S     K
	"A", "C", "G", "T", "T", "AG", "CT", "AC", "AT", "CG", "GT",
//   B      D      H      V      N       ?       -
	"CGT", "AGT", "ACT", "ACG", "ACGT", "ACGT", "ACGT"
};

//: A table to translate state numbers (0-17) into bit codes (0-15)
//  Bit codes have one bit for each nucleotide (ACGT). Multiple
//  bits can be set for ambiguities.
const long kNucleotideBits[]={
// 	A	C	G	T	U	R	Y	M	W	S	K	B	D	H	V	N	?	-
	1,	2,	4,	8,	8,	5,	10,	3,	9,	6,	12,	14,	13,	11,	7,	15,	15,	0
};

// A table to give the consensus ambiguities of a pair of bases
const char *kNucleotideConsensus[kNumNucleotides]={
//   ACGTURYMWSKBDHVN?-
    "AMRWWRHMWVDNDHVN??",   //A
    "MCSYYVYMHSBBNHVN??",   //C
    "RSGKKRBVDSKBDNVN??",   //G
    "WYKTTDYHWBKBDHNN??",   //T
    "WYKTUDYHWBKBDHNN??",   //U
    "RVRDDRNVDVDNDNVN??",   //R
    "HYBYYNYHHBBBNHNN??",   //Y
    "MMVHHVHMHVNNNHVN??",   //M
    "WHDWWDHHWNDNDHNN??",   //W
    "VSSBBVBVNSBBNNVN??",   //S
    "DBKKKDBNDBKBDNNN??",   //K
    "NBBBBNBNNBBBNNNN??",   //B
    "DNDDDDNNDNDNDNNN??",   //D
    "HHNHHNHHHNNNNHNN??",   //H
    "VVVNNVNVNVNNNNVN??",   //V
    "NNNNNNNNNNNNNNNN??",   //N
    "??????????????????",   //?
    "?????????????????-"    //-
};

//: This table maps nucleotide characters into state codes (0-17)
// Nucleotides go ACGTURYMWSKBDHVN?-"
// Other letters are mapped to ?
// ? and - are mapped to themselves
// All other chars are mapped to -
const int kNucleotideNumbers[128]={
	17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,	// 0-15
	17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,	// 16-31
//                                          -
	17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,	// 32-47
//                                                ?
	17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,16,	// 48-63
//	    A  B  C  D  e  f  G  H  i  j  K  l  M  N  o
	17, 0,11, 1,12,16,16, 2,13,16,16,10,16, 7,15,16,	// 64-79
//	 p  q  R  S  T  U  V  W  x  Y  z
	16,16, 5, 9, 3, 4,14, 8,16, 6,16,17,17,17,17,17,	// 80-95
//	    A  B  C  D  e  f  G  H  i  j  K  l  M  N  o
	17, 0,11, 1,12,16,16, 2,13,16,16,10,16, 7,15,16,	// 96-111
//	 p  q  R  S  T  U  V  W  x  Y  z
	16,16, 5, 9, 3, 4,14, 8,16, 6,16,17,17,17,17,17		// 112-127
};

//: These tables are used to convert DNA to RNA and vice versa
/*												  ACGTURYMWSKBDHVN?-*/
const char kRNATranslateTable[kNumNucleotides+1]="ACGUURYMWSKBDHVN?-";

/*												  ACGTURYMWSKBDHVN?-*/
const char kDNATranslateTable[kNumNucleotides+1]="ACGTTRYMWSKBDHVN?-";

//: These tables are used to complement DNA and RNA
/*												   ACGTURYMWSKBDHVN?-*/
const char kDNAComplementTable[kNumNucleotides+1]="TGCAAYRKSWMVHDBN?-";
const char kRNAComplementTable[kNumNucleotides+1]="UGCAAYRKSWMVHDBN?-";

//: These tables are used to convert DNA and RNA to just Rs and Ys
/*												 ACGTURYMWSKBDHVN?-*/
const char kRYTranslateTable[kNumNucleotides+1]="RYRYYRYNNNNNNNNN?-";

//********************
#pragma mark -- Amino Acid Tables --

//: A table to translate state numbers (0-25) into one letter codes
const char kAminoAcids[]="ACDEFGHIKLMNPQRSTVWYBZX*?-";

//: A table to map state numbers (0-25) to their three letter codes
const char* kAminoAcidsWide[kNumAminoAcidCodes]={
	  " A ",  " C ",  " D ",  " E ",  " F ",  " G ",  " H ",  " I ",  " K ",  
	  " L ",  " M ",  " N ",  " P ",  " Q ",  " R ",  " S ",  " T ",  " V ",
	  " W ",  " Y ",  " B ",  " Z ",  " X ",  " * ",  " ? ",  " - "
};

//: A table to map state numbers (0-25) to their three letter codes
const char* kAminoAcidTriplets[kNumAminoAcidCodes]={
//		A		C		D		E		F		G		H		I		K
	  "Ala",  "Cys",  "Asp",  "Glu",  "Phe",  "Gly",  "His",  "Ile",  "Lys",  
//		L		M		N		P		Q		R		S		T		V
	  "Leu",  "Met",  "Asn",  "Pro",  "Gln",  "Arg",  "Ser",  "Thr",  "Val",
//		W		Y		B		Z		X		*		?		-
	  "Trp",  "Tyr",  "Asx",  "Glx",  " X ",  " * ",  " ? ",  " - "
};

// A table to map state numbers (0-25) to their ambiguities
const char* kAminoAcidMaps[kNumAminoAcidCodes]={
//	   A	C	 D	  E	   F	G	 H	  I	   K
	  "A", "C", "D", "E", "F", "G", "H", "I", "L",  
//	   L	M	 N	  P	   Q	R	 S	  T	   V
	  "L", "M", "N", "P", "Q", "R", "S", "T", "V",
//	   W	Y	 B	   Z
	  "W", "Y", "DN", "EQ", 
//	   X					   *	?						-
	  "ACDEFGHIKLMNPQRSTVWY", "*", "ACDEFGHIKLMNPQRSTVWY", "ACDEFGHIKLMNPQRSTVWY"
};

//: A table to translate state numbers (0-25) into bit codes (20 bits)
//  Bit codes have one bit for each amino acid (ACDEFGHIKLMNPQRSTVWY). 
//  Multiple bits can be set for ambiguities (BZX?). Stop and gap set
//  no bits.
const long kAminoAcidBits[kNumAminoAcidCodes]={
// 	A			C			D			E
	0x000001,	0x000002,	0x000004,	0x000008,
//	F			G			H			I
	0x000010,	0x000020,	0x000040,	0x000080,	
//	K			L			M			N			
	0x000100,	0x000200,	0x000400,	0x000800,	
//	P			Q			R			S
	0x001000,	0x002000,	0x004000,	0x008000,	
//	T			V			W			Y
	0x010000,	0x020000,	0x040000,	0x080000,
//	B			Z			X			*
	0x000804,	0x002008,	0x0FFFFF,	0x000000,	
//	?			-
	0x0FFFFF,	0x000000	
};

//: This table maps amino acid characters into state codes (0-25)
// Amino Acids go ACDEFGHIKLMNPQRSTVWYBZX*?-
// Other letters; j, o, and u are mapped to ?
// *, ? and - are mapped to themselves
// All other chars are mapped to -
const int kAminoAcidNumbers[128]={
	25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,	// 0-15
	25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,	// 16-31
//                                 *        -
	25,25,25,25,25,25,25,25,25,25,23,25,25,25,25,25,	// 32-47
//                                                ?
	25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,24,	// 48-63
//		A  B  C  D  E  F  G  H  I  j  K  L  M  N  o
	25, 0,20, 1, 2, 3, 4, 5, 6, 7,24, 8, 9,10,11,24,	// 64-79
//	 P  Q  R  S  T  u  V  W  X  Y  Z
	12,13,14,15,16,24,17,18,22,19,21,25,25,25,25,25,	// 80-95
//		A  B  C  D  E  F  G  H  I  j  K  L  M  N  o
	25, 0,20, 1, 2, 3, 4, 5, 6, 7,24, 8, 9,10,11,24,	// 96-111
//	 P  Q  R  S  T  u  V  W  X  Y  Z
	12,13,14,15,16,24,17,18,22,19,21,25,25,25,25,25		// 112-127
};

//********************
#pragma mark -- Genetic Codes --

//: The names of the built in code tables
const char* kGeneticCodeNames[numGeneticCodes]={
	"Universal", "Vertebrate Mitochondrial", "Yeast", "Mold Protozoan Mitochondrial",
	"Mycoplasma", "Invertebrate Mitochondrial", "Ciliate", "Echinoderm Mitochondrial",
	"Euplotid Nuclear", "Bacterial", "Alternative Yeast", "Ascidian Mitochondrial",
	"Flatworm Mitochondrial", "Blepharisma Nuclear"
};

//: The built-in code tables.
// Nucleotides go T, C, A, G - the order given by the Genbank web site
// With the first codon position most significant (i.e. TTT, TTC, TTA, TTG, TCT, etc.).
const char kCodeTable[numGeneticCodes][65]={
	// Universal
	"FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Vertebrate Mitochondrial
	"FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSS**VVVVAAAADDEEGGGG",
	// Yeast
	"FFLLSSSSYY**CCWWTTTTPPPPHHQQRRRRIIMMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Mold Protozoan Mitochondrial
	"FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Mycoplasma
	"FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Invertebrate Mitochondrial
	"FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSSSSVVVVAAAADDEEGGGG",
	// Ciliate
	"FFLLSSSSYYQQCC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Echinoderm Mitochondrial
	"FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNNKSSSSVVVVAAAADDEEGGGG",
	// Euplotid Nuclear
	"FFLLSSSSYY**CCCWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Bacterial
	"FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Alternative Yeast
	"FFLLSSSSYY**CC*WLLLSPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
	// Ascidian Mitochondrial
	"FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSSGGVVVVAAAADDEEGGGG",
	// Flatworm Mitochondrial
	"FFLLSSSSYYY*CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNNKSSSSVVVVAAAADDEEGGGG",
	// Blepharisma Nuclear
	"FFLLSSSSYY*QCC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG"
};

//: The number of stop codons in each genetic code.
const int kNumStopCodons[numGeneticCodes] = {
	// Universal
	3,
	// Vertebrate Mitochondrial
	4,
	// Yeast
	2,
	// Mold Protozoan Mitochondrial
	2,
	// Mycoplasma
	2,
	// Invertebrate Mitochondrial
	2,
	// Ciliate
	1,
	// Echinoderm Mitochondrial
	2,
	// Euplotid Nuclear
	2,
	// Bacterial
	3,
	// Alternative Yeast
	3,
	// Ascidian Mitochondrial
	2,
	// Flatworm Mitochondrial
	1,
	// Blepharisma Nuclear
	2
};



