/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEGlobal.h

	Global constants

	Constants and tables for the molecular evolution library.
	
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = Global

#ifndef _MEGLOBAL_H_
#define _MEGLOBAL_H_

// Generally the enums all have a corresponding array of name strings for user
// output. The last item in the enum will be the count of the number of items.
// The first item will usually be a default or unknown state. If the value of
// this is -1 then there is no corresponding name string.

//********************
#pragma mark -- Constants --

//: a small number - always useful
#define kSmallReal 			1.0E-20
//: a big integer - always useful
#define kBigInt				32767
//: a big number - always useful	
#define kBigReal 			1.0E20
//: a tiny number - always useful	
#define kTinyReal			1e-32
//: tolerance for searching
#define kTolerance 			1.0E-6
//: final tolerance for searching		
#define kFinalTolerance 	1.0E-8
//: tolerence for smoothing
#define kSmoothTolerance	1.0E-6
//: maximum number of smoothing iterations
#define kMaxIterations		30

#ifdef _BOOL_
	#define bool short
	enum {
		false,
		true
	};
#endif

//: Used to set the flags of MEObjects
enum {
	kHidden=1,
	kDisabled=2,
	kHiddenDisabled=3,
	kMissing=4,
	kSelected=8,
	kIncluded=16,
	numFlagTypes
};

// The names of the status types
extern const char* kFlagTypeNames[numFlagTypes];

//: Maximum length of a name string - may be removed
#define MAX_NAME_LEN 255
//: Maximum length of a description string - may be removed
#define MAX_DESC_LEN 1024
//: Maximum length of a accession string - may be removed
#define MAX_ACC_LEN 255

//: Maximum length of a set of characters in a NEXUS file should be removed
#define MAX_CHARS 16

//: Maximum number of states for sequences
#define MAX_NUM_STATES 64

//: The initial capacity of a list
#define INITIAL_SIZE 100

//: The amount a list will increment its capacity by
#define INCREMENT_SIZE 100

//: The amount a sequence will increment its capacity by
#define SEQ_INCREMENT_SIZE 500

//: Used to specify the sequence type.
enum SequenceType {
	kUnknownSequence,
	kDNASequence, 
	kRNASequence, 
	kAminoAcidSequence, 
	kDiscreteSequence,		// Not supported at present
	kContinuousSequence,	// Not supported at present
	numSequenceTypes
};

// The names of the sequence types
extern const char* kSequenceTypeNames[numSequenceTypes];

//: Used to specify the alignment type.
enum AlignmentType {
	kUnknownAlignment,
	kNucleotideAlignment, 
	kAminoAcidAlignment, 
	kCodonAlignment,
	kDiscreteAlignment,
	numAlignmentTypes
};

//The names of the alignment types
extern const char* kAlignmentTypeNames[numAlignmentTypes];

//: Used to specify the different types of sequence features
enum FeatureType { // Many more to be added
	kUnknownFeature,
	kCDSFeature,
	kExonFeature,
	kIntronFeature,
	kRRNAFeature,
	kTRNAFeature,
	kMotifFeature,
	kOtherFeature,
	numFeatureTypes
};

// The names of the sequence types
extern const char* kFeatureTypeNames[numFeatureTypes];


//********************
#pragma mark -- Nucleotide Tables --

//: The number of nucleotide states (including ? and -)
const int kNumNucleotides=18;

// A table to translate state numbers (0-17) into character codes
extern const char kNucleotides[];

// A table to map state numbers (0-17) to their ambiguities
extern const char* kNucleotideMaps[kNumNucleotides];

//: A table to translate state numbers (0-17) into bit codes (0-15)
extern const long kNucleotideBits[kNumNucleotides];

//: The index of the unknown nucleotide state
const int kNucleotideUnknown=16;

//: The index of the gap nucleotide state
const int kNucleotideGap=17;

// A table to give the consensus ambiguities of a pair of bases
extern const char *kNucleotideConsensus[kNumNucleotides];

// This table maps nucleotide characters into state numbers (0-17)
// Nucleotides go ACGTURYMWSKBDHVN?-";
// Other letters are mapped to ?
// ? and - are mapped to themselves
// All other chars are mapped to -
extern const int kNucleotideNumbers[128];

// These tables are used to convert DNA to RNA and vice versa
extern const char kRNATranslateTable[kNumNucleotides+1];
extern const char kDNATranslateTable[kNumNucleotides+1];

// These tables are used to complement DNA and RNA
extern const char kDNAComplementTable[kNumNucleotides+1];
extern const char kRNAComplementTable[kNumNucleotides+1];

// These tables are used to convert DNA and RNA to just Rs and Ys
extern const char kRYTranslateTable[kNumNucleotides+1];

//********************
#pragma mark -- Codon Constants --

//: A union used to convert codon codes to the nucleotide components
union CodonCode {
	short code;
	struct {
		unsigned int pos1: 5; 
		unsigned int pos2: 5; 
		unsigned int pos3: 5; 
		unsigned int padding: 1; 
	} nucs;
};
				
//: The index of the unknown codon state (unknown in all 3 codon positions)
const int kCodonUnknown=(kNucleotideUnknown << 10) + (kNucleotideUnknown << 5) + kNucleotideUnknown;

//: The index of the gap codon state (gap in all 3 codon positions)
const int kCodonGap=(kNucleotideGap << 10) + (kNucleotideGap << 5) + kNucleotideGap;

//********************
#pragma mark -- Amino Acid Tables --

//: The number of amino acid states (including *, ? and -)
const int kNumAminoAcidCodes=26;
const int kNumAminoAcids=26; // Should be 20?

//: A handy enum of amino acid triplet codes
enum AminoAcidTriplet { 
	ALA, CYS, ASP, GLU, PHE, GLY, HIS, ILE, LYS, 
	LEU, MET, ASN, PRO, GLN, ARG, SER, THR, VAL,
	TRP, TYR, ASX, GLX, XXX, TER, UNK, GAP 
};

// A table to translate state numbers (0-25) into one letter codes
extern const char kAminoAcids[];

// A table to map state numbers (0-25) to their one letter codes but 3 characters wide
// This is useful for aligning against nucleotides.
extern const char* kAminoAcidsWide[kNumAminoAcidCodes];

// A table to map state numbers (0-25) to their three letter codes
extern const char* kAminoAcidTriplets[kNumAminoAcidCodes];

// A table to map state numbers (0-25) to their ambiguities
extern const char* kAminoAcidMaps[kNumAminoAcidCodes];

//: A table to translate state numbers (0-25) into bit codes (20 bits)
extern const long kAminoAcidBits[kNumAminoAcidCodes];

//: The index of the stop codon amino acid state
const int kAminoAcidStop=23;

//: The index of the unknown amino acid state
const int kAminoAcidUnknown=24;

//: The index of the gap amino acid state
const int kAminoAcidGap=25;

// This table maps amino acid characters into state numbers (0-25)
// Amino Acids go ACDEFGHIKLMNPQRSTVWYBZX*?-
// Other letters; j, o, and u are mapped to ?
// *, ? and - are mapped to themselves
// All other chars are mapped to -
extern const int kAminoAcidNumbers[128];


//********************
#pragma mark -- Genetic Codes --

//: Constants used to refer to the built in code tables
enum GeneticCode { // More should be added
	kNonCoding=-1, 
	kUniversalCode,
	kVertebrateMtCode,
	kYeastCode,
	kMoldProtozoanMtCode,
	kMycoplasmaCode,
	kInvertebrateMtCode,
	kCiliateCode,
	kEchinodermMtCode,
	kEuplotidNucCode,
	kBacterialCode,
	kAltYeastCode,
	kAscidianMtCode,
	kFlatwormMtCode,
	kBlepharismaNucCode,
	numGeneticCodes
};

//: The names of the built in code tables
extern const char* kGeneticCodeNames[numGeneticCodes];

//: The built-in code tables.
// Nucleotides go T, C, A, G - the order given by the Genbank web site
// With the first codon position most significant (i.e. TTT, TTC, TTA, TTG, TCT, etc.).
extern const char kCodeTable[numGeneticCodes][65];

//: The number of stop codons in each genetic code.
extern const int kNumStopCodons[numGeneticCodes];


#endif /* _MEGLOBAL_H_ */
