/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEIntArray.cpp

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of integers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MEObject.h"
#include "MEIntArray.h"

#pragma mark === MEIntArray ===

// ---------------------------------------------------------------------------------
//		MEIntArray::MEIntArray
// ---------------------------------------------------------------------------------

//: Constructs an empty MEIntArray
//  The array will only be allocated when an integer is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
MEIntArray::MEIntArray(int inInitial, int inIncrement) : MEObject()
{
	capacity=0;
	numValues=0;
	values=NULL;
	iterator=0;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		MEIntArray::~MEIntArray
// ---------------------------------------------------------------------------------

//: The destructor
MEIntArray::~MEIntArray()
{
	if (values!=NULL)
		delete[] values;
}

// ---------------------------------------------------------------------------------
//		MEObject::cloneList
// ---------------------------------------------------------------------------------

//: Clones this list from an existing one.

void MEIntArray::clone(MEObject *src)
{
	clone(src);

	setInitial(((MEIntArray*)src)->getCapacity(), ((MEIntArray*)src)->getIncrement());
	
	removeAll();
	int i, n=((MEIntArray*)src)->getNumValues();
	for (i=0; i<n; i++)
		add(((MEIntArray*)src)->getValue(i));
}

// ---------------------------------------------------------------------------------
//		MEIntArray::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void MEIntArray::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		MEIntArray::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void MEIntArray::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		MEIntArray::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void MEIntArray::setIncrement(int inIncrement)
{
	increment=inIncrement;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::getValue
// ---------------------------------------------------------------------------------

//: Returns the integer that is at an index.
int MEIntArray::getValue(int index)
{
	TestBounds(index, numValues);
	
	return values[index];
}

// ---------------------------------------------------------------------------------
//		MEIntArray::getSum
// ---------------------------------------------------------------------------------

//: Returns the sum of all the integers
//  This function has no protection against overflow.
int MEIntArray::getSum()
{
	int i;
	int sum = 0;
	for (i = 0; i < numValues; i++)
		sum += values[i];
		
	return sum;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::setObject
// ---------------------------------------------------------------------------------

//: Sets the integer at an index
//  If index>numValues then numValues becomes index+1 with memory 
//  allocated if needed.
void MEIntArray::setValue(int index, int inValue)
{
	ThrowIf(index<0);

	if (index>=numValues) {
		numValues=index+1;
		checkCapacity(numValues);
	} 
		
	values[index]=inValue;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::add
// ---------------------------------------------------------------------------------

//: Adds an integer to the end of the list
//  This expands the list's capacity if necessary.
int MEIntArray::add(int inValue)
{
	checkCapacity(numValues+1);

	values[numValues]=inValue;
	numValues++;
	
	return numValues-1;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::insert
// ---------------------------------------------------------------------------------

//: Inserts an integer into the list at a position
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new integer. If the integer's new position equals
//  numObjects then this adds it to the end of the list.
void MEIntArray::insert(int index, int inValue)
{
	ThrowIf(index<0 || index>numValues+1);

	checkCapacity(numValues+1);
		
	for (int i=numValues; i>index; i--)
		values[i]=values[i-1];
	
	values[index]=inValue;
	numValues++;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::remove
// ---------------------------------------------------------------------------------

//: Removes the integer at index and moves others up 
//  This function returns the integer being removed.
int MEIntArray::remove(int index)
{
	TestBounds(index, numValues);

	int tmp=values[index];

	for (int i=index; i<numValues-1; i++)
		values[i]=values[i+1];
		
	numValues--;
	
	return tmp;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the integers in the list
void MEIntArray::removeAll()
{
	numValues=0;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::findValue
// ---------------------------------------------------------------------------------

//: Finds the index of an integer in the list
int MEIntArray::findValue(int inValue)
{
	int i = 0;
	while (i < numValues && values[i]!=inValue) i++;
	
	if (i == numValues)
		return -1;
			
	return i;
}

// ---------------------------------------------------------------------------------
//		MEIntArray::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the integer list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new integer array of this
//  size, copies all the old integers into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void MEIntArray::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity>capacity) {
		int *newValues;
		int i;
		
		// Increase capacity
		while (capacity<requiredCapacity)
			capacity+=increment;
		
		// Allocate the new list.
		NewArray(newValues, int, capacity);
		
		if (values!=NULL) {
			// Copy the existing list into the new list.
			for (i=0; i<numValues; i++)
				newValues[i]=values[i];
	
			// Free the old list.
			delete[] values;
		}
		
		// Point the old list to the new one.
		values=newValues;
	}
}

// ---------------------------------------------------------------------------------
//		MEIntArray::readObject
// ---------------------------------------------------------------------------------

//: Reads the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list reading mechanism called by readData. This allows descendant classes
//  to read in all their fields before the list contents is read. This doesn't
//  bother reading capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MEIntArray::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		MEIntArray::readData
// ---------------------------------------------------------------------------------

//: Read the data of this object (i.e. the elements of the list)
//  This calls MEStream's getList function which reads the number of objects
//  and all the objects in the list.
void MEIntArray::readData(MEStream *stream)
{
	stream->getIntArray(this);
}

// ---------------------------------------------------------------------------------
//		MEIntArray::writeObject
// ---------------------------------------------------------------------------------

//: Writes the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list writing mechanism called by readData. This allows descendant classes
//  to write out all their fields before the list contents is written. This doesn't
//  bother writing capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MEIntArray::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}

// ---------------------------------------------------------------------------------
//		MEIntArray::writeData
// ---------------------------------------------------------------------------------

//: Write the data of this object (i.e. the elements of the list)
//  This calls MEStream's putList function which writes the number of objects
//  and all the objects in the list.
void MEIntArray::writeData(MEStream *stream)
{
	stream->putIntArray(this);
}

