/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MELabelArray.cpp

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of integers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MEObject.h"
#include "MELabelArray.h"

#pragma mark === MELabelArray ===

// ---------------------------------------------------------------------------------
//		MELabelArray::MELabelArray
// ---------------------------------------------------------------------------------

//: Constructs an empty MELabelArray
//  The array will only be allocated when an integer is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
MELabelArray::MELabelArray(LabelType inLabelType, int inInitial, int inIncrement) : MEObject()
{
	labelType = inLabelType;
	capacity=0;
	numLabels=0;
	iterator=0;
	
	intLabels=NULL;
	realLabels=NULL;
	stringLabels=NULL;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::~MELabelArray
// ---------------------------------------------------------------------------------

//: The destructor
MELabelArray::~MELabelArray()
{
	if (intLabels!=NULL)
		delete[] intLabels;
	if (realLabels!=NULL)
		delete[] realLabels;
	if (stringLabels!=NULL) {
		for (int i=0; i<capacity; i++) {
			if (stringLabels[i] != NULL)
				delete [] stringLabels[i];
		}
		delete[] stringLabels;
	}
}

// ---------------------------------------------------------------------------------
//		MELabelArray::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void MELabelArray::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void MELabelArray::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void MELabelArray::setIncrement(int inIncrement)
{
	increment=inIncrement;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::getIntLabel
// ---------------------------------------------------------------------------------

//: Returns the label as an integer (converted if necessary).
int MELabelArray::getIntLabel(int index)
{
	TestBounds(index, numLabels);
	int label;
	
	switch (labelType) {
		case kIntegerLabel:
			label = intLabels[index];
		break;
		case kRealLabel:
			label = (int)realLabels[index];
		break;
		case kStringLabel: // no error checking
			label = atoi(stringLabels[index]);
		break;
	}
	
	return label;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::getRealLabel
// ---------------------------------------------------------------------------------

//: Returns the label as an real number (converted if necessary).
double MELabelArray::getRealLabel(int index)
{
	TestBounds(index, numLabels);
	double label;
	
	switch (labelType) {
		case kIntegerLabel:
			label = (double)intLabels[index];
		break;
		case kRealLabel:
			label = realLabels[index];
		break;
		case kStringLabel: // no error checking
			label = atof(stringLabels[index]);
		break;
	}
	
	return label;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::getStringLabel
// ---------------------------------------------------------------------------------

//: Returns the label as an string (converted if necessary). If converting, then
//  an outString of suitable size must be provided. If an outString is provided
//  then the label is put in this otherwise the pointer to the string in the list
//  is returned.
char* MELabelArray::getStringLabel(int index, char* outString)
{
	TestBounds(index, numLabels);
	char *label;
	
	if (outString == NULL) {
		ThrowIf(labelType != kStringLabel);
		label = stringLabels[index];
	} else {
		label = outString;
		switch (labelType) {
			case kIntegerLabel:
				sprintf(label, "%d", intLabels[index]);
			break;
			case kRealLabel:
				sprintf(label, "%lf", realLabels[index]);
			break;
			case kStringLabel:
				if (stringLabels[index] != NULL)
					strcpy(label, stringLabels[index]);
				else
					label[0] = '\0';
			break;
		}
	}
		
	return label;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::getStringLabel
// ---------------------------------------------------------------------------------

//: Returns the length of the string label at index (or the string produced by
//  converting an integer or real label).
int MELabelArray::getStringLabelLength(int index)
{
	TestBounds(index, numLabels);
	int length = 0;
	
	if (labelType == kStringLabel) {
		if (stringLabels[index] != NULL)
			length = strlen(stringLabels[index]);
	} else {
		char str[256];
		if (labelType == kStringLabel)
			sprintf(str, "%d", intLabels[index]);
		else
			sprintf(str, "%lf", realLabels[index]);
		length = strlen(str);
	}
		
	return length;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::setIntLabel
// ---------------------------------------------------------------------------------

//: Sets the label at index as an integer (converts if necessary).
//  If index>numLabels then numLabels becomes index+1 with memory 
//  allocated if needed.
void MELabelArray::setIntLabel(int index, int inLabel)
{
	ThrowIf(index<0);
	
	if (index>=numLabels) {
		numLabels=index+1;
		checkCapacity(numLabels);
	} 
		
	switch (labelType) {
		case kIntegerLabel:
			intLabels[index] = inLabel;
		break;
		case kRealLabel:
			realLabels[index] = (double)inLabel;
		break;
		case kStringLabel:
			if (stringLabels[index] != NULL)
				delete [] stringLabels[index];
				
			char str[256];
			sprintf(str, "%d", inLabel);
			NewArray(stringLabels[index], char, strlen(str));
			strcpy(stringLabels[index], str);
		break;
	}
}

// ---------------------------------------------------------------------------------
//		MELabelArray::setRealLabel
// ---------------------------------------------------------------------------------

//: Sets the label at index as a real (converts if necessary).
//  If index>numLabels then numLabels becomes index+1 with memory 
//  allocated if needed.
void MELabelArray::setRealLabel(int index, double inLabel)
{
	ThrowIf(index<0);
	
	if (index>=numLabels) {
		numLabels=index+1;
		checkCapacity(numLabels);
	} 
		
	switch (labelType) {
		case kIntegerLabel:
			intLabels[index] = (int)inLabel;
		break;
		case kRealLabel:
			realLabels[index] = inLabel;
		break;
		case kStringLabel:
			if (stringLabels[index] != NULL)
				delete [] stringLabels[index];
				
			char str[256];
			sprintf(str, "%lf", inLabel);
			NewArray(stringLabels[index], char, strlen(str));
			strcpy(stringLabels[index], str);
		break;
	}
}

// ---------------------------------------------------------------------------------
//		MELabelArray::setStringLabel
// ---------------------------------------------------------------------------------

//: Sets the label at index as a string (converts if necessary).
//  If index>numLabels then numLabels becomes index+1 with memory 
//  allocated if needed.
void MELabelArray::setStringLabel(int index, char *inLabel)
{
	ThrowIf(index<0);
	ThrowIfNull(inLabel);
	
	if (index>=numLabels) {
		numLabels=index+1;
		checkCapacity(numLabels);
	} 
		
	switch (labelType) {
		case kIntegerLabel: // no error checking
			intLabels[index] = atoi(inLabel);
		break;
		case kRealLabel: // no error checking
			realLabels[index] = atof(inLabel);
		break;
		case kStringLabel:
			NewArray(stringLabels[index], char, strlen(inLabel));
			strcpy(stringLabels[index], inLabel);
		break;
	}
}

// ---------------------------------------------------------------------------------
//		MELabelArray::addIntLabel
// ---------------------------------------------------------------------------------

//: Adds the label to the end of the list as an integer (converts if necessary).
//  This expands the list's capacity if necessary.
int MELabelArray::addIntLabel(int inLabel)
{
	setIntLabel(numLabels, inLabel);
	return numLabels-1;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::addRealLabel
// ---------------------------------------------------------------------------------

//: Adds the label to the end of the list as a real (converts if necessary).
//  This expands the list's capacity if necessary.
int MELabelArray::addRealLabel(double inLabel)
{
	setRealLabel(numLabels, inLabel);
	return numLabels-1;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::addStringLabel
// ---------------------------------------------------------------------------------

//: Adds the label to the end of the list as a string (converts if necessary).
//  This expands the list's capacity if necessary.
int MELabelArray::addStringLabel(char *inLabel)
{
	setStringLabel(numLabels, inLabel);
	return numLabels-1;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::insertIntLabel
// ---------------------------------------------------------------------------------

//: Inserts into the list at a position as an integer (converts if necessary).
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new integer. If the integer's new position equals
//  numObjects then this adds it to the end of the list.
void MELabelArray::insertIntLabel(int index, int inLabel)
{
	ThrowIf(index<0 || index>numLabels+1);

	checkCapacity(numLabels+1);
	
	int i;
	switch (labelType) {
		case kIntegerLabel:
			for (i=numLabels; i>index; i--)
				intLabels[i] = intLabels[i-1];
		break;
		case kRealLabel: // no error checking
			for (i=numLabels; i>index; i--)
				realLabels[i] = realLabels[i-1];
		break;
		case kStringLabel:
			for (i=numLabels; i>index; i--)
				stringLabels[i] = stringLabels[i-1];
			stringLabels[index] = NULL;
		break;
	}
	setIntLabel(index, inLabel);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::insertRealLabel
// ---------------------------------------------------------------------------------

//: Inserts into the list at a position as a real (converts if necessary).
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new integer. If the integer's new position equals
//  numObjects then this adds it to the end of the list.
void MELabelArray::insertRealLabel(int index, double inLabel)
{
	ThrowIf(index<0 || index>numLabels+1);

	checkCapacity(numLabels+1);
	
	int i;
	switch (labelType) {
		case kIntegerLabel:
			for (i=numLabels; i>index; i--)
				intLabels[i] = intLabels[i-1];
		break;
		case kRealLabel: // no error checking
			for (i=numLabels; i>index; i--)
				realLabels[i] = realLabels[i-1];
		break;
		case kStringLabel:
			for (i=numLabels; i>index; i--)
				stringLabels[i] = stringLabels[i-1];
			stringLabels[index] = NULL;
		break;
	}
	setRealLabel(index, inLabel);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::insertStringLabel
// ---------------------------------------------------------------------------------

//: Inserts into the list at a position as a string (converts if necessary).
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new integer. If the integer's new position equals
//  numObjects then this adds it to the end of the list.
void MELabelArray::insertStringLabel(int index, char *inLabel)
{
	ThrowIf(index<0 || index>numLabels+1);

	checkCapacity(numLabels+1);
	
	int i;
	switch (labelType) {
		case kIntegerLabel:
			for (i=numLabels; i>index; i--)
				intLabels[i] = intLabels[i-1];
		break;
		case kRealLabel: // no error checking
			for (i=numLabels; i>index; i--)
				realLabels[i] = realLabels[i-1];
		break;
		case kStringLabel:
			for (i=numLabels; i>index; i--)
				stringLabels[i] = stringLabels[i-1];
			stringLabels[index] = NULL;
		break;
	}
	setStringLabel(index, inLabel);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::removeLabel
// ---------------------------------------------------------------------------------

//: Removes the label at index and moves others up 
//  This function returns the integer being removed.
void MELabelArray::removeLabel(int index)
{
	TestBounds(index, numLabels);

	switch (labelType) {
		case kIntegerLabel:
			for (int i=index; i<numLabels-1; i++)
				intLabels[i] = intLabels[i+1];
		break;
		case kRealLabel:
			for (int i=index; i<numLabels-1; i++)
				realLabels[i] = realLabels[i+1];
		break;
		case kStringLabel:
			char* tmp = stringLabels[index];

			for (int i=index; i<numLabels-1; i++)
				stringLabels[i] = stringLabels[i+1];
				
			stringLabels[numLabels]	= NULL;
			if (tmp != NULL)
				delete [] tmp;
		break;
	}
				
	numLabels--;
}

// ---------------------------------------------------------------------------------
//		MELabelArray::removeAllLabels
// ---------------------------------------------------------------------------------

//: Removes all the labels in the list
void MELabelArray::removeAllLabels()
{
	numLabels=0;
	if (labelType == kStringLabel) {
		for (int i=0; i<numLabels; i++) {
			if (stringLabels[i] != NULL) {
				delete [] stringLabels[i];
				stringLabels[i] = NULL;
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		MELabelArray::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the label list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new integer array of this
//  size, copies all the old integers into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void MELabelArray::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity>capacity) {
		int i;
		
		// Increase capacity
		while (capacity<requiredCapacity)
			capacity+=increment;
		
		switch (labelType) {
			case kIntegerLabel:
				int *newIntLabels;
				// Allocate the new list.
				NewArray(newIntLabels, int, capacity);
				
				if (intLabels!=NULL) {
					// Copy the existing list into the new list.
					for (i=0; i<numLabels; i++)
						newIntLabels[i]=intLabels[i];
			
					// Free the old list.
					delete[] intLabels;
				}
		
				// Point the old list to the new one.
				intLabels=newIntLabels;
			break;
			case kRealLabel:
				double *newRealLabels;
				// Allocate the new list.
				NewArray(newRealLabels, double, capacity);
				
				if (realLabels!=NULL) {
					// Copy the existing list into the new list.
					for (i=0; i<numLabels; i++)
						newRealLabels[i]=realLabels[i];
			
					// Free the old list.
					delete[] realLabels;
				}
		
				// Point the old list to the new one.
				realLabels=newRealLabels;
			break;
			case kStringLabel:
				char **newStringLabels;
				// Allocate the new list.
				NewArray(newStringLabels, char*, capacity);
				for (i=0; i<capacity; i++)
					newStringLabels[i] = NULL;
				
				if (stringLabels!=NULL) {
					// Copy the existing list into the new list.
					for (i=0; i<numLabels; i++)
						newStringLabels[i]=stringLabels[i];
			
					// Free the old list.
					delete[] stringLabels;
				}
		
				// Point the old list to the new one.
				stringLabels=newStringLabels;
			break;
		}
	}
}

// ---------------------------------------------------------------------------------
//		MELabelArray::readObject
// ---------------------------------------------------------------------------------

//: Reads the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list reading mechanism called by readData. This allows descendant classes
//  to read in all their fields before the list contents is read. This doesn't
//  bother reading capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MELabelArray::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::readData
// ---------------------------------------------------------------------------------

//: Read the data of this object (i.e. the elements of the list)
//  This calls MEStream's getList function which reads the number of objects
//  and all the objects in the list.
void MELabelArray::readData(MEStream *stream)
{
	// Not yet implemented...
}

// ---------------------------------------------------------------------------------
//		MELabelArray::writeObject
// ---------------------------------------------------------------------------------

//: Writes the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list writing mechanism called by readData. This allows descendant classes
//  to write out all their fields before the list contents is written. This doesn't
//  bother writing capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MELabelArray::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}

// ---------------------------------------------------------------------------------
//		MELabelArray::writeData
// ---------------------------------------------------------------------------------

//: Write the data of this object (i.e. the elements of the list)
//  This calls MEStream's putList function which writes the number of objects
//  and all the objects in the list.
void MELabelArray::writeData(MEStream *stream)
{
	// Not yet implemented...
}

