/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MELabelArray.h

		Interface for the MELabelArray Class
					
 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global
 
 
#ifndef _MELABEL_ARRAY_H_
#define _MELABEL_ARRAY_H_

#include "MEObject.h"

enum LabelType {
	kIntegerLabel,
	kRealLabel,
	kStringLabel
};

class MEStream;

/******************************************************************************
	class MELabelArray
 ******************************************************************************/
#pragma mark MELabelArray class

//:	Implements a dynamically resizing list of integers

class MELabelArray : public MEObject {
	public:
		virtual long		getID() const { return 'MLaA'; }
		//: The ID of this object - 'MLaA'.

		virtual char*		getClassName() const { return "LabelArray"; }
		//: Returns the name of this class - "LabelArray"
		
	protected:
		LabelType			labelType; 	//: The type of labels in the list
		int 				numLabels; 	//: The number of labels in the list
		int					capacity;	//: The capacity of the list
		int					increment;	//: The amount by which capacity increments
		int					iterator;	//: Used to iterate through the list

		int		 			*intLabels;		//: An array of integers
		double		 		*realLabels;	//: An array of reals
		char		 		**stringLabels;	//: An array of strings

	public:
							MELabelArray(LabelType inLabelType,
									int inInitial=INCREMENT_SIZE, 
									int inIncrement=INCREMENT_SIZE);	//: The constructor
	 	virtual				~MELabelArray();	//: The destructor		
	 	
		virtual LabelType	getType() { return labelType; }
		//: Returns the type of the label array

		virtual void		setInitial(int inInitial, int inIncrement=INCREMENT_SIZE);
		//: Set the initial capacity and (optionally) the increment

		int					getNumLabels() { return numLabels; };
		//: Returns the number of labels in the list
		
		virtual void		setCapacity(int inCapacity);
		//: Set the capacity
		
		virtual int			getCapacity() { return capacity; };
		//: Returns the capacity

		virtual int			getIncrement() { return increment; };
		//: Sets the amount that capacity increments by
		virtual void		setIncrement(int inIncrement);
		//: Returns the increment
		
		virtual int			getIntLabel(int index);
		//: Returns the label as an integer (converted if necessary).
		
		virtual double		getRealLabel(int index);
		//: Returns the label as an real number (converted if necessary).
		
		virtual char*		getStringLabel(int index, char* outString = NULL);
		//: Returns the label as an string (converted if necessary).
		
		virtual int			getStringLabelLength(int index);
		//: Returns the length of the string label at index (or the string produced by converting an integer or real label)
		
		virtual void		setIntLabel(int index, int inLabel);
		//: Sets the label at index as an integer (converts if necessary).

		virtual void		setRealLabel(int index, double inLabel);
		//: Sets the label at index as a real (converts if necessary).

		virtual void		setStringLabel(int index, char *inLabel);
		//: Sets the label at index as a string (converts if necessary).

		virtual int			addIntLabel(int inLabel);
		//: Adds the label to the end of the list as an integer (converts if necessary).

		virtual int			addRealLabel(double inLabel);
		//: Adds the label to the end of the list as a real (converts if necessary).

		virtual int			addStringLabel(char *inLabel);
		//: Adds the label to the end of the list as a string (converts if necessary).

		virtual void		insertIntLabel(int index, int inLabel);
		//: Inserts into the list at a position as an integer (converts if necessary).

		virtual void		insertRealLabel(int index, double inLabel);
		//: Inserts into the list at a position as a real (converts if necessary).

		virtual void		insertStringLabel(int index, char *inLabel);
		//: Inserts into the list at a position as a string (converts if necessary).

		virtual void		removeLabel(int index);
		//: Removes the label at index and moves others up 
		
		virtual void		removeAllLabels();
		//: Removes all the labels in the list

	protected:
		virtual void		checkCapacity(int requiredCapacity);
		//: Checks that the integer list has enough space and resizes it as necessary
	
	public:
		virtual void		readObject(MEStream *stream);
		//: Reads the fields of this object

		virtual void		readData(MEStream *stream);
		//: Read the data of this object (i.e. the elements of the list)
		
		virtual void		writeObject(MEStream *stream);
		//: Writes the fields of this object

		virtual void		writeData(MEStream *stream);
		//: Write the data of this object (i.e. the elements of the list)
};

#endif /* _MELABEL_ARRAY_H_ */
