/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MELink.cpp

				MELink Class
					
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEUtilities.h"
#include "MEBitset.h"
#include "MELink.h"


#pragma mark === MELink Class ===

// ---------------------------------------------------------------------------------
//		 MELink
// ---------------------------------------------------------------------------------

//: Constructor
MELink::MELink()
{
	tipNo=-1;
	taxonNo=-1;
	nodeNo=-1;
	branchNo=-1;
	
	isStem=false;
	
	next=NULL;
	branch=NULL;
	
	navBits=NULL;
}

// ---------------------------------------------------------------------------------
//		 ~MELink
// ---------------------------------------------------------------------------------

//: Destructor
MELink::~MELink()
{
	if (navBits!=NULL)
		delete navBits;
}

// ---------------------------------------------------------------------------------
//		 MELink::clone
// ---------------------------------------------------------------------------------

//: This function creates a clone of inLink
//  This function cannot clone the links (next, branch) because these are pointers 
//  to uncloned nodes. These links have to be cloned by the METree cloning function.

void MELink::clone(MELink *inLink)
{
	linkNo=inLink->getLinkNo();
	tipNo=inLink->getTipNo();
	taxonNo=inLink->getTaxonNo();
	nodeNo=inLink->getNodeNo();
	branchNo=inLink->getBranchNo();
	
	isStem=inLink->getIsStem();
	
	next=NULL;
	branch=NULL;
	
	navBits=NULL;
}

// ---------------------------------------------------------------------------------
//		 MELink::getNumBranches
// ---------------------------------------------------------------------------------

//: returns the number of branches in the polytomy, including the stem branch.
int MELink::getNumBranches()
{
	int i=1;
	
	MELink *link=next;
	if (link!=NULL) {
		while (link!=this) {
			i++;
			link=link->next;
		}
	}
		
	return i;
}

// ---------------------------------------------------------------------------------
//		 MELink::getStemLink
// ---------------------------------------------------------------------------------

//: returns the link in the node that is the stem.
MELink* MELink::getStemLink()
{
	MELink *link=next;
	while (link!=this && !link->isStem) {
		link=link->next;
	}
	
	return link;
}

// ---------------------------------------------------------------------------------
//		 MELink::insertBranch
// ---------------------------------------------------------------------------------

//:  Adds a branch link to this node
void MELink::insertBranch(MELink *inLink)
{
// Insert the new branch link into the link loop
// To keep the branch order that is given, we must
// insert the new branch before the current one.
	MELink *link2=next;
	inLink->next=this;
	while (link2->next!=this)
		link2=link2->next;
	link2->next=inLink;
}


// ---------------------------------------------------------------------------------
//		 MELink::removeBranch
// ---------------------------------------------------------------------------------

//: Removes a branch link from this node
void MELink::removeBranch(MELink *inLink)
{
	ThrowIf(inLink==this);
	
	MELink *link=this;
	while (link->next!=inLink && link->next!=this)
		link=link->next;
	
	ThrowIf(link->next==this);
	link->next=inLink->next;
}


// ---------------------------------------------------------------------------------
//		 MELink::getPrevious
// ---------------------------------------------------------------------------------

//: Gets the link that points to this one
//  Simply follows the links round the list until it finds the one with the
//  next field pointing to this link.
MELink *MELink::getPrevious()
{
	MELink *link=this;
	while (link->next!=this)
		link=link->next;
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 MELink::contains
// ---------------------------------------------------------------------------------

//: returns true if the twig branches of the node contain inNode (a tip or a node)
bool MELink::contains(MELink *inLink)
{	
	MELink *link=next;
	while (link!=this) {
		if (link==inLink || link->branch==inLink)
			return true;
		link=link->next;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 MELink::cladeContains
// ---------------------------------------------------------------------------------

//: returns true if the clade of the node contains inNode (a tip or a node)
bool MELink::cladeContains(MELink *inLink)
{
	if (inLink==this)
		return true;
	
	if (getIsTip())
		return false;
		
	MELink *link=next;
	while (link!=this) {
		if (link->branch->cladeContains(inLink))
			return true;
		link=link->next;
	}
	
	return false;
}


// ---------------------------------------------------------------------------------
//		 MELink::getNavBits
// ---------------------------------------------------------------------------------

//: returns the navBits, creating them if need be.
MEBitset* MELink::getNavBits()
{
	if (navBits==NULL && tipNo==-1)
		NewClass(navBits, MEBitset());
		
	return navBits;
}
