/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEList.cpp

 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global

//	Implements a dynamically resizing list of MEObjects
//  MEList is derived itself from MEObject so it is possible to have lists of lists.
//  METaxonList and METreeList are derived from MEList. The list keeps track
//  of the number of objects and its current capacity. When adding items
//  to the list, it checks to see if it has the capacity and, if not, reallocates
//  its array, incrementing the capacity by increment. The object pointers
//  are copied into the new array and the old array is deleted. This means that
//  the capacity of MEList can increase but not decrease.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MEObject.h"
#include "MEList.h"

#pragma mark === MEList ===

// ---------------------------------------------------------------------------------
//		MEList::MEList
// ---------------------------------------------------------------------------------

//: Constructs an empty MEList
//  The array will only be allocated when an object is added. The increment 
//  amount will be set to a default value, and the initial capacity will be zero.
MEList::MEList(int inInitial, int inIncrement) : MEObject()
{
	capacity = 0;
	numObjects = 0;
	objects = NULL;
	iterator = 0;
	
	setInitial(inInitial, inIncrement);
}

// ---------------------------------------------------------------------------------
//		MEList::~MEList
// ---------------------------------------------------------------------------------

//: The destructor
//  Delete the list of objects and call removeReference for each object
MEList::~MEList()
{
	int i;

	if (objects != NULL) {
		for (i = 0; i < numObjects; i++) {
			if (objects[i] != NULL)
				objects[i]->removeReference(true);
		}
		delete[] objects;
	}
}

// ---------------------------------------------------------------------------------
//		MEList::cloneList
// ---------------------------------------------------------------------------------

//: Clones this list from an existing one. It also clones all the objects in
//  the list.

void MEList::cloneList(MEList *src)
{
	clone(src);
	setInitial(src->getCapacity(), src->getIncrement());
	
	MEObject *obj;
	int i, n = src->getNumObjects();
	for (i = 0; i < n; i++) {
		obj = createObject();
		obj->clone(src->getObject(i));
		add(obj, true);
	}
}

// ---------------------------------------------------------------------------------
//		MEList::setInitial
// ---------------------------------------------------------------------------------

//: Set the initial capacity and (optionally) the increment
//  If an array hasn't been created then this creates it.
void MEList::setInitial(int inInitial, int inIncrement)
{
	setIncrement(inIncrement);
	checkCapacity(inInitial);
}

// ---------------------------------------------------------------------------------
//		MEList::setCapacity
// ---------------------------------------------------------------------------------

//: Set the capacity
// If the capacity is not already greater than inCapacity, then this increases it.
void MEList::setCapacity(int inCapacity)
{
	checkCapacity(inCapacity);
}

// ---------------------------------------------------------------------------------
//		MEList::setIncrement
// ---------------------------------------------------------------------------------

//: Sets the amount that capacity increments by
//  If capacity needs to increase then it will do so by multiples of increment.
void MEList::setIncrement(int inIncrement)
{
	increment = inIncrement;
}

// ---------------------------------------------------------------------------------
//		MEList::createObject
// ---------------------------------------------------------------------------------

//: Creates a new object of a type appropriate to the list
//  By default this just returns a new instance of MEObject. By overriding
//  this, descendents (e.g. METreeList) can create instances appropriate
//  to the list (e.g. METree).
MEObject* MEList::createObject()
{
	MEObject *obj;
	NewClass(obj, MEObject());
	return obj;
}

// ---------------------------------------------------------------------------------
//		MEList::getObject
// ---------------------------------------------------------------------------------

//: Returns the MEObject that is at an index.
MEObject* MEList::getObject(int index)
{
	TestBounds(index, numObjects);
	
	return objects[index];
}

// ---------------------------------------------------------------------------------
//		MEList::setObject
// ---------------------------------------------------------------------------------

//: Sets the object at an index
//  If an object already exists then the reference to it is lost. 
//  If index>numObjects then numObjects becomes index+1 with memory 
//  allocated if needed.
void MEList::setObject(int index, MEObject *inObject, bool isOwner)
{
	ThrowIf(index < 0);

	if (index >= numObjects) {
		numObjects = index+1;
		checkCapacity(numObjects);
	} 
		
	if (isOwner)
		inObject->setOwner(this);
	inObject->addReference();
	objects[index] = inObject;
}

// ---------------------------------------------------------------------------------
//		MEList::add
// ---------------------------------------------------------------------------------

//: Adds an object to the end of the list
//  This expands the list's capacity if necessary.
int MEList::add(MEObject *inObject, bool isOwner)
{
	checkCapacity(numObjects+1);

	if (isOwner)
		inObject->setOwner(this);
	inObject->addReference();
	objects[numObjects]=inObject;
	numObjects++;
	
	return numObjects-1;
}

// ---------------------------------------------------------------------------------
//		MEList::insert
// ---------------------------------------------------------------------------------

//: Inserts an object into the list at a position
//  This expands the list's capacity if necessary and moves others downwards
//  to make space for the new object. If the object's new position equals
//  numObjects then this adds it to the end of the list.
void MEList::insert(int index, MEObject *inObject, bool isOwner)
{
	ThrowIf(index<0 || index>numObjects+1);

	checkCapacity(numObjects+1);
		
	for (int i = numObjects; i > index; i--)
		objects[i] = objects[i-1];
	
	if (isOwner)
		inObject->setOwner(this);
	inObject->addReference();
	objects[index] = inObject;
	numObjects++;
}

// ---------------------------------------------------------------------------------
//		MEList::swap
// ---------------------------------------------------------------------------------

//: Swaps an object at index1 with that at index2
//  Both objects must already be in the list.
void MEList::swap(int index1, int index2)
{
	TestBounds(index1, numObjects);
	TestBounds(index2, numObjects);

	MEObject *obj = objects[index1];
	objects[index1] = objects[index2];
	objects[index2] = obj;
}

// ---------------------------------------------------------------------------------
//		MEList::remove
// ---------------------------------------------------------------------------------

//: Removes the object at index and moves others up 
//  This does not delete the removed object so if it is not referenced 
//  elsewhere then it is lost. This function returns the object so it can 
//  be deleted or kept by the calling function.
MEObject* MEList::remove(int index)
{
	MEObject* object;
	
	TestBounds(index, numObjects);

	object = objects[index];

	for (int i = index; i < numObjects-1; i++)
		objects[i] = objects[i+1];
		
	numObjects--;
	objects[numObjects] = NULL;
	if (object->getOwner()==this)
		object->setOwner(NULL);
	object->removeReference();
	
	return object;
}

// ---------------------------------------------------------------------------------
//		MEList::removeAll
// ---------------------------------------------------------------------------------

//: Removes all the objects in the list
//  This does not delete them so if they are not referenced elsewhere then 
//  they are lost.
void MEList::removeAll()
{
	for (int i = 0; i < numObjects; i++) {
		if (objects[i]->getOwner() == this)
			objects[i]->setOwner(NULL);
		objects[i]->removeReference();
		objects[i] = NULL;
	}
	numObjects = 0;
}

// ---------------------------------------------------------------------------------
//		MEList::deleteAll
// ---------------------------------------------------------------------------------

//: Removes and deletes all the objects in the list. 
//  Actually it calls removeReference which will delete the object if
//  nothing else is referring to it.
void MEList::deleteAll()
{
	MEObject *obj;
	for (int i = 0; i < numObjects; i++) {
		obj = objects[i];
		objects[i] = NULL;
		obj->removeReference(true);
	}
	numObjects = 0;
}

// ---------------------------------------------------------------------------------
//		MEList::getFirst
// ---------------------------------------------------------------------------------

//: Gets the first object in the list
//  Sets the iterator variable to 0 and returns the first object.
MEObject* MEList::getFirst()
{
	iterator = 0;
	return getNext();
}

// ---------------------------------------------------------------------------------
//		MEList::getNext
// ---------------------------------------------------------------------------------

//: Gets the next object in the list
//  Increments the iterator variable and returns the object at that position.
MEObject* MEList::getNext()
{
	MEObject *obj = NULL;
	if (iterator < numObjects)
		obj = objects[iterator];
	iterator++;
	
	return obj;
}

// ---------------------------------------------------------------------------------
//		MEList::compare
// ---------------------------------------------------------------------------------

//: Compares two objects for sorting
//  By default this is an alphabetical comparison of the names.
int MEList::compare(MEObject *inObject1, MEObject *inObject2)
{
	char *name1 = inObject1->getName();
	char *name2 = inObject2->getName();
	
	if (name1 == NULL || name2 == NULL)
		return 0;
		
	return strcmp(name1, name2);
}

// ---------------------------------------------------------------------------------
//		MEList::compare
// ---------------------------------------------------------------------------------

//: Sorts the objects in the list 
//  By default this is in ascending order unless inAscending is false.
void MEList::sort(bool inAscending)
{
	if (numObjects > 1) {
	
			// Shellsort Algorithm
		
		int	hh = 1;						// Find starting "h" value
		int	stopH = numObjects / 9;
		while (hh < stopH) {
			hh = 3 * hh + 1;
		}
		
		MEObject *tmp;
		
		int direction = (inAscending ? 1 : -1);
		
		for ( ; hh > 0; hh /= 3) {
				
			for (int step = hh + 1; step <= numObjects; step++) {
				tmp=getObject(step - 1);
				
				long i;
				for (i = step - hh; i > 0; i -= hh) {
					if ((compare(tmp, getObject(i-1)) * direction) >= 0) {
						break;
					}
					setObject(i + hh - 1, getObject(i-1));
				}
					
				setObject(i + hh - 1, tmp);
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		MEList::getObjectName
// ---------------------------------------------------------------------------------

//: Returns the name string from the object at an index
char* MEList::getObjectName(int index)
{
	TestBounds(index, numObjects);
		 
	ThrowIfNull(objects[index]);

	return objects[index]->getName();
}

// ---------------------------------------------------------------------------------
//		MEList::findName
// ---------------------------------------------------------------------------------

//: Searches the objects for a name
//  Returns the index of the object or -1 if not found.
int	MEList::findName(char *inName, bool inIgnoreCase)
{
	int i = 0;
	
	if (inIgnoreCase) {
		bool found = false;
		char *P, *Q;
		
		while (i < numObjects && !found) {
			P = objects[i]->getName();
			Q = inName;
			while (*P != '\0' && *Q != '\0' && toupper(*P) == toupper(*Q)) {
				P++;
				Q++;
			}
			if (*P == '\0' && *Q == '\0')
				found = true;
			else
				i++;
		}
	} else
		while (i < numObjects && strcmp(objects[i]->getName(), inName) != 0)
			i++;
	
	if (i == numObjects)
		return -1;
	else
		return i;
}
		
// ---------------------------------------------------------------------------------
//		MEList::findIndex
// ---------------------------------------------------------------------------------

//: Searches the list for object referenced by pointer
//  Returns index of the object or -1 if not found.
int	MEList::findIndex(MEObject *object)
{
	int i = 0;
	
	while (i < numObjects && objects[i] != object) i++;
	
	if (i == numObjects)
		return -1;
	else
		return i;
}
		
// ---------------------------------------------------------------------------------
//		MEList::findByContents
// ---------------------------------------------------------------------------------

//: Searches the list for an item whos contents equals object
//  Returns index of the object or -1 if not found.
int	MEList::findByContents(MEObject *object)
{
	int i = 0;
	
	while (i < numObjects && compare(objects[i], object) != 0) i++;
	
	if (i == numObjects)
		return -1;
	else
		return i;
}
		
// ---------------------------------------------------------------------------------
//		MEList::getNumWithFlag
// ---------------------------------------------------------------------------------

//: Gets the number of objects with a particular flag set
//  Uses inFlag to check the state of one or more flag bits in each object and
//  counts the number which match inState.
int MEList::getNumWithFlag(short inFlag, bool inState)
{
	int n=0;
	MEObject *obj = getFirst();
	while (obj != NULL) {
		if (obj->getFlag(inFlag) == inState)
			n++;
		obj = getNext();
	}
	
	return n;
}

// ---------------------------------------------------------------------------------
//		MEList::getFirstWithFlag
// ---------------------------------------------------------------------------------

//: Gets the first object with a particular flag set
//  Uses inFlag to check the state of one or more flag bits in each object and
//  returns the pointer to the first that matches inState. If none match then it
//  returns NULL.
MEObject *MEList::getFirstWithFlag(short inFlag, bool inState)
{
	MEObject *obj = getFirst();
	while (obj != NULL && obj->getFlag(inFlag) != inState)
		obj = getNext();
	
	return obj;
}

// ---------------------------------------------------------------------------------
//		MEList::getNextWithFlag
// ---------------------------------------------------------------------------------

//: Gets the next object with a particular flag set
//  Uses inFlag to check the state of one or more flag bits in each object using
//  getNext and returns the pointer to the first that matches inState. If none 
//  match then it returns NULL.
MEObject *MEList::getNextWithFlag(short inFlag, bool inState)
{
	MEObject *obj = getNext();
	while (obj != NULL && obj->getFlag(inFlag) != inState)
		obj = getNext();
	
	return obj;
}

// ---------------------------------------------------------------------------------
//		MEList::getFlag
// ---------------------------------------------------------------------------------

//: Gets the flag state of the object at a particular index
//  Calls the equivalent function in MEObject.
bool MEList::getFlag(int index, short inFlag)
{
	MEObject *obj = getObject(index);
	return obj->getFlag(inFlag);
}

// ---------------------------------------------------------------------------------
//		MEList::setFlag
// ---------------------------------------------------------------------------------

//: Sets the flag state of the object at a particular index
//  Calls the equivalent function in MEObject.
void MEList::setFlag(int index, short inFlag, bool inState)
{
	MEObject *obj = getObject(index);
	obj->setFlag(inFlag, inState);
}

// ---------------------------------------------------------------------------------
//		MEList::setFlagForAll
// ---------------------------------------------------------------------------------

//: Sets the flag state for all the objects in the list
//  Calls the equivalent function in MEObject for each.
void MEList::setFlagForAll(short inFlag, bool inState)
{
	MEObject *obj = getFirst();
	while (obj != NULL) {
		obj->setFlag(inFlag, inState);
		obj = getNext();
	}
}

// ---------------------------------------------------------------------------------
//		MEList::checkCapacity
// ---------------------------------------------------------------------------------

//: Checks that the objects list has enough space and resizes it as necessary
//  Keeps increasing the capacity by increment until it is equal to or greater
//  than requiredCapacity. It then creates a new MEObject pointer list of this
//  size, copies all the old object pointers into it and sets the rest to NULL.
//  Finally it deletes the old list, and replaces it with the new.
void MEList::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity > capacity) {
		MEObject **newObjects;
		int i;
		
		// Increase capacity
		while (capacity < requiredCapacity)
			capacity += increment;
		
		// Allocate the new list.
		NewArray(newObjects, MEObject*, capacity);
		
		if (objects != NULL) {
			// Copy the existing list into the new list.
			for (i = 0; i < numObjects; i++)
				newObjects[i] = objects[i];
	
			// Free the old list.
			delete[] objects;

			// Set the remaining pointers to NULL.
			for (i = numObjects; i < capacity; i++)
				newObjects[i] = NULL;
		} else {
			// Set the all the pointers to NULL.
			for (i = 0; i < capacity; i++)
				newObjects[i] = NULL;
		}
		
		// Point the old list to the new one.
		objects = newObjects;
	}
}

// ---------------------------------------------------------------------------------
//		MEList::readObject
// ---------------------------------------------------------------------------------

//: Reads the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list reading mechanism called by readData. This allows descendant classes
//  to read in all their fields before the list contents is read. This doesn't
//  bother reading capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MEList::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		MEList::readData
// ---------------------------------------------------------------------------------

//: Read the data of this object (i.e. the elements of the list)
//  This calls MEStream's getList function which reads the number of objects
//  and all the objects in the list.
void MEList::readData(MEStream *stream)
{
	stream->getList(this);
}

// ---------------------------------------------------------------------------------
//		MEList::writeObject
// ---------------------------------------------------------------------------------

//: Writes the fields of this object
//  By default this just calls MEObject's function. This is because the MEStream
//  has a list writing mechanism called by readData. This allows descendant classes
//  to write out all their fields before the list contents is written. This doesn't
//  bother writing capacity, increment or iterator so these will be set to their 
//  defaults. This may be changed in future.
void MEList::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}

// ---------------------------------------------------------------------------------
//		MEList::writeData
// ---------------------------------------------------------------------------------

//: Write the data of this object (i.e. the elements of the list)
//  This calls MEStream's putList function which writes the number of objects
//  and all the objects in the list.
void MEList::writeData(MEStream *stream)
{
	stream->putList(this);
}

