/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEList.h

		Interface for the MEList Class
					
 ******************************************************************************/
//! author="Andrew Rambaut"
//! group=Global
 
 
#ifndef _MELIST_H_
#define _MELIST_H_

#include "MEObject.h"


#define INITIAL_SIZE 100
#define INCREMENT_SIZE 100

class MEStream;

/******************************************************************************
	class MEList
 ******************************************************************************/
#pragma mark MEList class

//:	Implements a dynamically resizing list of MEObjects
//  MEList is derived itself from MEObject so it is possible to have lists of lists.
//  METaxonList and METreeList are derived from MEList. The list keeps track
//  of the number of objects and its current capacity. When adding items
//  to the list, it checks to see if it has the capacity and, if not, reallocates
//  its array, incrementing the capacity by increment. The object pointers
//  are copied into the new array and the old array is deleted. This means that
//  the capacity of MEList can increase but not decrease.

class MEList : public MEObject {
	public:
		virtual long		getID() const { return 'MLst'; }
		//: The ID of this object - 'MLst'.

		virtual char*		getClassName() const { return "List"; }
		//: Returns the name of this class - "List"
		
	protected:
		int 				numObjects; //: The number of objects in the list
		int					capacity;	//: The capacity of the list
		int					increment;	//: The amount by which capacity increments
		int					iterator;	//: Used to iterate through the list

		MEObject 			**objects;	//: An array of MEObject pointers

	public:
							MEList(int inInitial=INCREMENT_SIZE, 
									int inIncrement=INCREMENT_SIZE);	//: The constructor
	 	virtual				~MEList();	//: The destructor		
	 	
		virtual void 		cloneList(MEList *src);
		//: Clones this list from an existing one, cloning all the contents too.

		virtual void		setInitial(int inInitial, int inIncrement=INCREMENT_SIZE);
		//: Set the initial capacity and (optionally) the increment

		int					getNumObjects() { return numObjects; }
		//: Returns the number of objects in the list
		
		virtual MEObject**	getObjects() { return objects; }
		//: Returns the list of MEObject pointers
		
		virtual void		setCapacity(int inCapacity);
		//: Set the capacity
		
		virtual int			getCapacity() { return capacity; }
		//: Returns the capacity

		virtual int			getIncrement() { return increment; }
		//: Returns the increment

		virtual void		setIncrement(int inIncrement);
		//: Sets the amount that capacity increments by
		
		virtual MEObject*	createObject();
		//: Creates a new object of a type appropriate to the list
		
		virtual MEObject*	getObject(int index);
		//: Returns the MEObject that is at an index
		
		virtual void		setObject(int index, MEObject *inObject, bool isOwner=false);
		//: Sets the object at an index

		virtual int			add(MEObject *inObject, bool isOwner=false);
		//: Adds an object to the end of the list
		
		virtual void		insert(int index, MEObject *inObject, bool isOwner=false);
		//: Inserts an object into the list at a position
		
		virtual void		swap(int index1, int index2);
		//: Swaps an object at index1 with that at index2
		
		virtual MEObject*	remove(int index);
		//: Removes the object at index and moves others up 
		
		virtual void		removeAll();
		//: Removes all the objects in the list

		virtual void		deleteAll();
		//: Removes and deletes all the objects in the list

		virtual MEObject*	getFirst();
		//: Gets the first object in the list

		virtual MEObject*	getNext();
		//: Gets the next object in the list
		
		virtual int			compare(MEObject *inObject1, MEObject *inObject2);
		//: Compares two objects for sorting
		
		virtual void		sort(bool inAscending=true);
		//: Sorts the objects in the list 
		
		virtual char* 		getObjectName(int index);
		//: Returns the name string from the object at an index

		virtual int			findName(char *inName, bool inIgnoreCase = false);
		//: Searches the objects for a name

		virtual int			findIndex(MEObject *object);		
		//: Searches the list for object referenced by pointer
		
		virtual int			findByContents(MEObject *object);		
		//: Searches the list for an item whos contents equals object
		
		virtual int			getNumWithFlag(short inFlag, bool inState);
		//: Gets the number of objects with a particular flag set
		
		virtual MEObject*	getFirstWithFlag(short inFlag, bool inState);
		//: Gets the first object with a particular flag set
		 
		virtual MEObject*	getNextWithFlag(short inFlag, bool inState);
		//: Gets the next object with a particular flag set
		
		virtual bool		getFlag(int index, short inFlag);
		//: Gets the flag state of the object at a particular index

		virtual void		setFlag(int index, short inFlag, bool inState);
		//: Sets the flag state of the object at a particular index

		virtual void		setFlagForAll(short inFlag, bool inState);
		//: Sets the flag state for all the objects in the list
		
//		virtual MEObject*	operator[](int idx) { return getObject(idx); }

	protected:
		virtual void		checkCapacity(int requiredCapacity);
		//: Checks that the objects list has enough space and resizes it as necessary
		virtual void		accommodate(int extra) { checkCapacity(extra+capacity); }
		
	public:
		virtual void		readObject(MEStream *stream);
		//: Reads the fields of this object

		virtual void		readData(MEStream *stream);
		//: Read the data of this object (i.e. the elements of the list)
		
		virtual void		writeObject(MEStream *stream);
		//: Writes the fields of this object

		virtual void		writeData(MEStream *stream);
		//: Write the data of this object (i.e. the elements of the list)
};

#endif /* _MELIST_H_ */
