/******************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEMaths.cpp
 	
	Some basic mathematical and miscellaneous functions.

 ******************************************************************/


//  includes:
// C/C++:
#include <stdlib.h>
#include <stdio.h>		// for 'sprintf' command
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream.h>

// Mol Evol:
#include "MEGlobal.h"
#include "MEUtilities.h"
#include "MEExceptions.h"
#include "MEMaths.h"

using namespace std;

// methods:

#pragma mark __________ MEMath __________

// ===========================================================================================
// 								MEMath static class 
// ===========================================================================================

int MEMath::Hamming(char *str1, char *str2, bool _caseSensitive)
//: return the Hamming distance between the two strings.
{
	int h = 0, i, k = min(strlen(str1), strlen(str2));
	if (_caseSensitive) {
		for (i = 0; i < k; i++) {
			if (str1[i] != str2[i])
				h++;
		}
		return h;
	}
	for (i = 0; i < k; i++) {
		if (toupper(str1[i]) != toupper(str2[i]))
			h++;
	}
	return h;
}

double MEMath::mod(double d, double m)
{
	if (m <= 0)
		return 0.0;
	return d - m * floor(d / m);
}


int MEMath::ones(long int x)
{
	int
		a = 0,
		nb = sizeof(long int) * 8;	// number of bits in a long int
	for (int i = 0; i < nb; i++)
		if (x & (long int) (1 << i))
			a++;
	return a;
}


char* MEMath::intToBinaryString(int x, char *str)
{
	int i, si = sizeof(int)*8;
	if (str == NULL)
		NewArray(str, char, si);
	for (i = 0; i < si; i++) {
		if (x & (1 << (si-i-1)))
			str[i] = '1';
		else
			str[i] = '0';
	}
	str[i] = '\0';
	return str;
}
char* MEMath::longToBinaryString(long int x, char *str)
{
	int i, si = sizeof(long int)*8;
	if (str == NULL)
		NewArray(str, char, si);
	for (i = 0; i < si; i++) {
		if (x & (1 << (si-i-1)))
			str[i] = '1';
		else
			str[i] = '0';
	}
	str[i] = '\0';
	return str;
}


void MEMath::intToIntPair(int pairNum, int &x, int &y)
{
	y = 1;
	int i = 0;
	while (i < pairNum) {
		y++;
		i += y;
	}
//	for (int i = 0; i <= pairNum; i += py+1)
//		py++;
	x = pairNum - y * (y - 1)/2;
}


void MEMath::intPairToInt(int x, int y, int &pairNum)
{
	ThrowIf(x >= y);
	pairNum = y * (y - 1)/2 + x;
}


#pragma mark _______ MECombinatoric _______

// ===========================================================================================
// 								MECombinatoric static class 
// ===========================================================================================


// -------------------------------------------------------------------------------------------
//   MECombinatoric::factorial
// -------------------------------------------------------------------------------------------
long int MECombinatoric::factorial(int x)
{
	long int f = 1;
	for (int i = x; i > 1; i--)
		f *= i;
	return f;
}


// -------------------------------------------------------------------------------------------
//   MECombinatoric::facfac
// -------------------------------------------------------------------------------------------
long int MECombinatoric::facfac(int x)
{
	long int ff = 1;
	for (int i = x; i > 1; i -= 2)
		ff *= i;
	return ff;
}


// -------------------------------------------------------------------------------------------
//   MECombinatoric::dfactorial
// -------------------------------------------------------------------------------------------
double MECombinatoric::dfactorial(int x)
{
	double d = 1.0;
	for (int i = x; i > 1; i--)
		d *= (double) i;
	return d;
}


// -------------------------------------------------------------------------------------------
//   MECombinatoric::nCr
// -------------------------------------------------------------------------------------------
double MECombinatoric::nCr(long int n, long int r)
{
	if (n < r)
		return 0.0;
	long int R = r;
	if (n/2 < r)
		R = n-r;		// this is just to make it a bit quicker..
	double d = 1.0;
	int i;
	for (i = 1; i <= R; i++) {
		d /= (double) i;			// 1, 2, 3, ..., R
		d *= (double) (n+1-i);	// n, n-1, n-2, ..., n+1-R
	}
	return d;
}


// -------------------------------------------------------------------------------------------
//   MECombinatoric::Stirling
// -------------------------------------------------------------------------------------------
double MECombinatoric::Stirling(long int x)
//: calculate the Stirling approximation to the factorial, for large numbers
{
	double y = log(sqrt(2*Pi*x)) + x*log(x/exp(1.0));
	return (exp(y));
}



#pragma mark _________ MERandom _________

// ===========================================================================================
// 								MERandom static class 
// ===========================================================================================


//  variables:

long int MERandom::seed = -1;


//  defines:

#define RANDOM_H
#define RAN_IA		16807
#define RAN_IM		2147483647
#define RAN_AM		(1.0/RAN_IM)
#define RAN_IQ		127773
#define RAN_IR		2836
#define RAN_NTAB	32
#define RAN_NDIV	(1+(RAN_IM-1)/RAN_NTAB)
#define RAN_EPS		1.2e-7
#define RAN_RNMX	(1.0-RAN_EPS)


// functions:

// -------------------------------------------------------------------------------------------
//   MERandom::fRan
// -------------------------------------------------------------------------------------------
float MERandom::fRan(double d)
{
	float
		temp;
	int
		j;
	long
		k;
	static long
		iy = 0,
		iv[RAN_NTAB];
	
	if (seed <= 0 || !iy) {
		if (-(seed) < 1)
			seed = 1;
		else
			seed = -(seed);
		for (j = RAN_NTAB+7; j >= 0; j--) {
			k = (seed)/RAN_IQ;
			seed = RAN_IA*(seed-k*RAN_IQ)-RAN_IR*k;
			if (seed < 0)
				seed += RAN_IM;
			if (j < RAN_NTAB)
				iv[j] = seed;
		}
		iy = iv[0];
	}
	k = (seed)/RAN_IQ;
	seed = RAN_IA*(seed-k*RAN_IQ)-RAN_IR*k;
	if (seed < 0)
		seed += RAN_IM;
	j = iy/RAN_NDIV;
	iy = iv[j];
	iv[j] = seed;
	if ((temp=RAN_AM*iy) > RAN_RNMX)
		return d*RAN_RNMX;
	else
		return d*temp;
}


// -------------------------------------------------------------------------------------------
//   MERandom::gammaRan
// -------------------------------------------------------------------------------------------
double MERandom::gammaRan(int shape)
{
	int j;
	double
		am,
		e,
		s,
		v1, v2,
		x, y;
	
	if (shape < 1) {
		cerr << "Internal error: gammaran(shape): Can't have non-positive integer for shape\n";
		shape = 1;
	}
	if (shape < 6) {	// use direct method, adding wait times
		x = 1.0;
		for (j = 0; j < shape; j++)
			x *= MERandom::fRan();
		x = -log(x);
	}
	else {	// use rejection method
		do {
			do {
				do {	// generate the tangent of a random angle
					v1 = 2.0*MERandom::fRan() - 1.0;
					v2 = 2.0*MERandom::fRan() - 1.0;
				} while (v1*v1 + v2*v2 > 1.0);
				y = v2/v1;
				am = shape-1;
				s = sqrt(2.0*am + 1.0);
				x = s*y + am;
			} while (x <= 0.0);
			e = (1.0 + y*y) * exp(am*log(x/am) - s*y);
		} while (MERandom::fRan() > e);
	}
	return x;
}


// -------------------------------------------------------------------------------------------
//   MERandom::iRan
// -------------------------------------------------------------------------------------------
long int MERandom::iRan(long int x)
//: return a randomly chosen number from the uniform discrete distribution on [0,x-1].
{
	return (long int) (MERandom::fRan() * (float) x );
}


// -------------------------------------------------------------------------------------------
//   MERandom::ulRan
// -------------------------------------------------------------------------------------------
unsigned long MERandom::ulRan(unsigned long x)
//: return a randomly chosen number from the uniform discrete distribution on [0,x-1].
{
	return (unsigned long) MERandom::fRan((float) x);
}


 // -------------------------------------------------------------------------------------------
//   MERandom::logRan
// -------------------------------------------------------------------------------------------
double MERandom::logRan(double x)
{
	double y;
	do {
		y = (double) MERandom::fRan();
	} while (y == 0.0);
	return (-x*log(y));

}

// -------------------------------------------------------------------------------------------
//   MERandom::nRan
// -------------------------------------------------------------------------------------------
long int MERandom::nRan(int x, int d)
//: return a random multinomially (d-) distributed variate from the range [0, d(x-1)]
{
	long int s = 0;
	for (int i = 0; i < x; i++)
		s += MERandom::iRan(d);
	return s;
}


// -------------------------------------------------------------------------------------------
//   MERandom::bRan
// -------------------------------------------------------------------------------------------
bool MERandom::bRan(double p)
//: return a true with probability p.
{
	return ( MERandom::fRan() < p ? true : false );
}


// -------------------------------------------------------------------------------------------
//   MERandom::SampleBinomial
// -------------------------------------------------------------------------------------------
int MERandom::SampleBinomial(int n, double p)
//: Sample from n trials each of which has probability p of success and return the number
// of successes.
{
	int i, j = 0;
	for (i = 0; i < n; i++)
		j += MERandom::bRan(p);
	return j;
}


// -------------------------------------------------------------------------------------------
//   MERandom::SampleGeometric
// -------------------------------------------------------------------------------------------
int MERandom::SampleGeometric(double p)
//: Sample from the geometric distribution with probabilities given by
//   f_k = (1-q)q^(k-1)
{
	if (p < 0.0 || p >= 1.0) {
		cerr << "p must be in the range (0.0, 1.0)\n";
		return -1.0;
	}
	return 1 + (int) (log(MERandom::fRan(1.0)) / log(1.0 - p));
}



// -------------------------------------------------------------------------------------------
//   MERandom::SampleLogNormal
// -------------------------------------------------------------------------------------------
double MERandom::SampleLogNormal(double mean, double stdDev)
{
	double temp = MERandom::SampleNormal(mean, stdDev);
	return exp(temp);
}


// -------------------------------------------------------------------------------------------
//   MERandom::SampleNormal
// -------------------------------------------------------------------------------------------
double MERandom::SampleNormal(double mean, double stdDev)
{
	double
		x1, x2, fac, rsq, bm;
	static double
		gset;
	static int
		iset = 0;
	
	if (iset == 0) {
		do {
			x1 = 2*MERandom::fRan() - 1.0;
			x2 = 2*MERandom::fRan() - 1.0;
			rsq = x1*x1+x2*x2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2*log(rsq)/rsq);
		gset = x1*fac;
		iset = 1;
		bm = x2*fac;
	}
	else {
		iset = 0;
		bm = gset;
	}
	bm = bm*stdDev + mean;
	return bm;
}


// -------------------------------------------------------------------------------------------
//   MERandom::SamplePowerLaw
// -------------------------------------------------------------------------------------------
int MERandom::SamplePowerLaw(double b)
{
	if (b <= 1.0) {
		cerr << "Must have b > 1.0 for this function.\n";
		return 0;
	}
	static double 
		C = 0.0,
		*cP = NULL;
	double
		d,
		e,
		tolerance = 1e-10;
	int 
		i = 1,
		j, 
		numPs;
	static double
		storedB = 0.0;
	if (b != storedB) {
		e = pow(1.0/tolerance, 1.0/b);
		if (e > (double) 32768) {
			tolerance = pow((double)32768, -b);
			numPs = 32768;
		}
		else
			numPs = (int) e;
		C = 0.0;
		delete [] cP;
		do {
			d = pow((double) i++, -b);
			C += d;
		} while (d > tolerance);
		C = 1.0/C;
		NewArray(cP, double, numPs);
		cP[0] = C;
		for (j = 1; j < numPs; j++) {
			d = C*pow((double) j+1, -b);
			cP[j] = cP[j-1] + d;
		}
		// C <- {\sum_{k>=1}{k^{-b}}}^{-1}
		storedB = b;
	}
	
	double r = MERandom::fRan(1.0);
	i = 0;
	do {
		i++;
	} while (r > cP[i] && i < numPs);
	return i;
}


// -------------------------------------------------------------------------------------------
//   MERandom::SampleUniform
// -------------------------------------------------------------------------------------------
double MERandom::SampleUniform(double minVal, double maxVal)
//: return a random variate drawn from the uniform continuous distribution 
// on the interval (minVal, maxVal)
{
	return (MERandom::fRan(maxVal-minVal) + minVal);
}


#pragma mark -


// -------------------------------------------------------------------------------------------
//   MERandom::selectRandom
// -------------------------------------------------------------------------------------------
void MERandom::SelectRandom(int x, int *in, int *out, bool _withReplacement)
//: Select 'x' items from the list in *in and put them into *out.
{
	// not sure whether this is necessary
}
