/******************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEMaths.h
 	
	Some basic mathematical and miscellaneous functions.

 ******************************************************************/

#ifndef MEMATHS_H
#define MEMATHS_H

#pragma mark // includes
#include <math.h>

#pragma mark // macros

#ifndef max
	#define max(A,B)	((A) > (B) ? (A) : (B))
#endif
#ifndef min
	#define min(A,B)	((A) < (B) ? (A) : (B))
#endif
#define sqr(A)		((A) * (A))

#pragma mark // constants
#define Pi			3.14159265358979323846
#define Sqrt2		1.41421356237309504880
#define kPrecision	7
#define ODD			% 2 == 1
#define EVEN		% 2 == 0

enum DistributionType {
	kDUniform,
	kDLogUniform,
	kDNormal,
	kDLogNormal,
	kDGamma,
	kDGeometric,
	kDBinomial,
	kDPower,
	kNumDistributionTypes
};


#pragma mark // MEMath
class MEMath {
public:
	static double		exp10(double d) { return exp(d * log(10)); }
	static int			Hamming(char *str1, char *str2, bool _caseSensitive = false);
	static bool			inRange(double d, double x, double y)
							{ return ((d >= min(x,y)) && (d <= max(x,y))); }
	static bool			inRange(double d, int x, int y)
							{ return ((d >= min(x,y)) && (d <= max(x,y))); }
	static bool			inRange(int i, int x, int y)
							{ return ((i >= min(x,y)) && (i <= max(x,y))); }
	static double		mod(double d, double m);
	static int			ones(long int x);
	static int			ones(int x) { return MEMath::ones((long int) x); }
	static char*		intToBinaryString(int i, char *str = NULL);
	static char*		longToBinaryString(long int i, char *str = NULL);
	static void			intToIntPair(int pairNum, int &x, int &y);
	static void			intPairToInt(int x, int y, int &pairNum);
};

#pragma mark // MECombinatoric
class MECombinatoric : public MEMath {
public:
	static long int		factorial(int x);
	static long int		facfac(int x);
	static double		dfactorial(int x);
	static double		nCr(long int n, long int r);
	static double		Stirling(int x)
							{ return MECombinatoric::Stirling((long int) x); }
	static double		Stirling(long int x);

};

#pragma mark // MERandom
class MERandom : public MEMath {
protected:
	static long int		seed;

public:
	static void			setSeed(long int inSeed) { seed=inSeed; }
	static long			getSeed() { return seed; }
	
	static float			fRan(double d = 1.0);
	static double		gammaRan(int shape);
	static long int		iRan(long i);
	static long int		iRan(int i) { return MERandom::iRan((long int) i); }
	static unsigned long	ulRan(unsigned long i);
	static double		logRan(double x);
	static long int		nRan(int x, int d);
	static bool			bRan(double p = 0.5);
	static int			SampleBinomial(int n, double p);
	static int			SampleGeometric(double p);
	static double		SampleLogNormal(double mean = 0.0, double stdDev = 1.0);
	static double		SampleNormal(double mean = 0.0, double stdDev = 1.0);
	static double		SampleUniform(double minVal, double maxVal);
	static int			SamplePowerLaw(double b);
	static void			SelectRandom(int x, int *in, int *out, bool _withReplacement);
};

#endif
