/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEMatrix.cpp

				MEMatrix Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEList.h"
#include "MEMatrix.h"
#include "MEExceptions.h"

#include "MEStream.h"

#pragma mark === MEMatrix Class ===

// ---------------------------------------------------------------------------------
//		 MEMatrix
// ---------------------------------------------------------------------------------

MEMatrix::MEMatrix() : MEObject()
{
	numRows=numCols=numElements=0;
	matrix=NULL;
		
	isSymmetric=false;
	isSymmetricKnown=false;
	
	hasLowerTriangle = false;
	hasUpperTriangle = false;
	hasDiagonal = false;

	rowList=NULL;
	colList=NULL;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix
// ---------------------------------------------------------------------------------

MEMatrix::MEMatrix(int inDimension) : MEObject()
{
	numRows=numCols=numElements=0;
	matrix=NULL;
	setSize(inDimension);

	hasLowerTriangle = false;
	hasUpperTriangle = false;
	hasDiagonal = false;

	rowList=NULL;
	colList=NULL;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix
// ---------------------------------------------------------------------------------

MEMatrix::MEMatrix(int inNumRows, int inNumCols) : MEObject()
{
	numRows=numCols=numElements=0;
	matrix=NULL;
	setSize(inNumRows, inNumCols);

	hasLowerTriangle = false;
	hasUpperTriangle = false;
	hasDiagonal = false;

	rowList=NULL;
	colList=NULL;
}


// ---------------------------------------------------------------------------------
//		 ~MEMatrix
// ---------------------------------------------------------------------------------

MEMatrix::~MEMatrix()
{
	if (matrix!=NULL)
		delete[] matrix;
	if (rowList!=NULL && rowList->getOwner()==this)
		delete rowList;
	if (colList!=NULL && colList->getOwner()==this)
		delete colList;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setSize
// ---------------------------------------------------------------------------------

// This resizes the matrix if neccessary. If it does then the contents are lost.

void MEMatrix::setSize(int inDimension)
{
	setSize(inDimension, inDimension);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setSize
// ---------------------------------------------------------------------------------

// This resizes the matrix if neccessary. If it does then the contents are lost.

void MEMatrix::setSize(int inNumRows, int inNumCols)
{
	ThrowIf(inNumRows<0 || inNumCols<0);
	
	if (inNumRows==numRows && inNumCols==numCols)
		return;
		
	numRows=inNumRows;
	numCols=inNumCols;
	
	if ((numRows*numCols)!=numElements) {
		numElements=numRows*numCols;
		
		if (matrix!=NULL)
			delete[] matrix;
		
		if (numElements>0)
			NewArray(matrix, double, numElements);
	}
	
	isSymmetric=false;
	isSymmetricKnown=false;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getIsSymmetric
// ---------------------------------------------------------------------------------

bool MEMatrix::getIsSymmetric()
{
	if (!isSymmetricKnown) {
		if (numRows != numCols) {
			isSymmetric=false;
		} else {
			int r, c;
			isSymmetric=true;
			r=0;
			while (r < numRows && isSymmetric) {
				c=r+1;
				while (c < numCols && isSymmetric) {
					isSymmetric=(getElement(r,c) == getElement(c,r));
					c++;
				}
				r++;
			}
			
		}
	}
	
	return isSymmetric;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::makeSymmetric
// ---------------------------------------------------------------------------------

void MEMatrix::makeSymmetric(bool fromLowerTriangle)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	int i, j;
	if (fromLowerTriangle) {
		for (i = 1; i < numRows; i++)
			for (j = 0; j < i; j++)
				setElement(j, i, getElement(i, j));
	} else {
		for (i = 1; i < numRows; i++)
			for (j = 0; j < i; j++)
				setElement(i, j, getElement(j, i));
	}
		
	hasLowerTriangle = true;
	hasUpperTriangle = true;
}

#pragma mark -
#pragma mark === Setting and Getting ===

// ---------------------------------------------------------------------------------
//		 MEMatrix::getRowList
// ---------------------------------------------------------------------------------

MEList *MEMatrix::getRowList()
{
	return rowList;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getRowObject
// ---------------------------------------------------------------------------------

// Gets the object representing row. If the object at row does not exist then 
// this is created. If the rowList does not exist then it is created.

MEObject *MEMatrix::getRowObject(int row)
{
	if (rowList==NULL) {
		NewClass(rowList, MEList());
		rowList->setOwner(this);	
	}
	
	MEObject *obj=NULL;
	if (row<rowList->getNumObjects())
		obj=rowList->getObject(row);
		
	if (obj==NULL) {
		NewClass(obj, MEObject());
		rowList->setObject(row, obj, true);
	}
	return obj;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setRowName
// ---------------------------------------------------------------------------------

void MEMatrix::setRowName(int row, char *inName)
{
	MEObject *obj=getRowObject(row);
	obj->setName(inName);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getRowName
// ---------------------------------------------------------------------------------

void MEMatrix::getRowName(int row, char *inName, int inMaxLen)
{
	if (getRowList()!=NULL) {
		MEObject *obj=getRowObject(row);
		if (obj!=NULL) {
			char *str=obj->getName();
			if (str!=NULL) {
				strncpy(inName, str, inMaxLen-1);
				inName[inMaxLen]=0;
				
				return;
			}
		}
	}

	sprintf(inName, "%d", row+1);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getRowName
// ---------------------------------------------------------------------------------

char *MEMatrix::getRowName(int row)
{
	if (getRowList()!=NULL)
		return getRowObject(row)->getName();
	else	
		return NULL;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::findRowName
// ---------------------------------------------------------------------------------

int MEMatrix::findRowName(char *inName)
{
	if (getRowList()!=NULL)
		return getRowList()->findName(inName);
	else	
		return -1;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumnList
// ---------------------------------------------------------------------------------

MEList *MEMatrix::getColumnList()
{
	return colList;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumnObject
// ---------------------------------------------------------------------------------

// Gets the object representing col. If the object at col does not exist 
// then this is created. If the colList does not exist then it is created.

MEObject *MEMatrix::getColumnObject(int col)
{
	if (colList==NULL) {
		NewClass(colList, MEList());
		colList->setOwner(this);	
	}
	
	MEObject *obj=NULL;
	if (col<colList->getNumObjects())
		obj=colList->getObject(col);
		
	if (obj==NULL) {
		NewClass(obj, MEObject());
		colList->setObject(col, obj, true);
	}
	return obj;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setColumnName
// ---------------------------------------------------------------------------------

void MEMatrix::setColumnName(int col, char *inName)
{
	MEObject *obj=getColumnObject(col);
	obj->setName(inName);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumnName
// ---------------------------------------------------------------------------------

void MEMatrix::getColumnName(int col, char *inName, int inMaxLen, char *inDefault)
{
	if (getColumnList()!=NULL) {
		MEObject *obj=getColumnObject(col);
		if (obj!=NULL) {
			char *str=obj->getName();
			if (str!=NULL) {
				strncpy(inName, str, inMaxLen-1);
				inName[inMaxLen]=0;
				
				return;
			}
		}
	}

	if (inDefault != NULL)
		sprintf(inName, "%s %d", inDefault, col+1);
	else
		sprintf(inName, "%d", col+1);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumnName
// ---------------------------------------------------------------------------------

char *MEMatrix::getColumnName(int col)
{
	if (getColumnList()!=NULL)
		return getColumnObject(col)->getName();
	else	
		return NULL;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::findColumnName
// ---------------------------------------------------------------------------------

int MEMatrix::findColumnName(char *inName)
{
	if (getColumnList()!=NULL)
		return getColumnList()->findName(inName);
	else	
		return -1;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getNumRowsWithFlag
// ---------------------------------------------------------------------------------

int MEMatrix::getNumRowsWithFlag(short inFlag, bool inState)
{
	MEList *list=getRowList();
	if (list!=NULL)
		return list->getNumWithFlag(inFlag, inState);
	else if (inState)
		return 0;
	else
		return numRows;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getRowFlag
// ---------------------------------------------------------------------------------

bool MEMatrix::getRowFlag(int row, short inFlag)
{
	MEList *list=getRowList();
	if (list!=NULL)
		return list->getFlag(row, inFlag);
	else
		return false;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setRowFlag
// ---------------------------------------------------------------------------------

void MEMatrix::setRowFlag(int row, short inFlag, bool inState)
{
	MEObject *obj=getRowObject(row);
	obj->setFlag(inFlag, inState);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setRowFlagForAll
// ---------------------------------------------------------------------------------

void MEMatrix::setRowFlagForAll(short inFlag, bool inState)
{
	for (int i=0; i<numRows; i++)
		setRowFlag(i, inFlag, inState);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getNumColumnsWithFlag
// ---------------------------------------------------------------------------------

int MEMatrix::getNumColumnsWithFlag(short inFlag, bool inState)
{
	MEList *list=getColumnList();
	if (list!=NULL)
		return list->getNumWithFlag(inFlag, inState);
	else if (inState)
		return 0;
	else
		return numCols;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumnFlag
// ---------------------------------------------------------------------------------

bool MEMatrix::getColumnFlag(int col, short inFlag)
{
	MEList *list=getColumnList();
	if (list!=NULL)
		return list->getFlag(col, inFlag);
	else
		return false;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setColumnFlag
// ---------------------------------------------------------------------------------

void MEMatrix::setColumnFlag(int col, short inFlag, bool inState)
{
	MEObject *obj=getColumnObject(col);
	obj->setFlag(inFlag, inState);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setColumnFlagForAll
// ---------------------------------------------------------------------------------

void MEMatrix::setColumnFlagForAll(short inFlag, bool inState)
{
	for (int i=0; i<numCols; i++)
		setColumnFlag(i, inFlag, inState);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::extractVisibleMatrix
// ---------------------------------------------------------------------------------

// Sets inMatrix with the elements of this matrix that are not in hidden
// rows or columns.

void MEMatrix::extractMatrixWithFlag(MEMatrix *inMatrix, short inFlag, bool inState)
{
	ThrowIfNull(inMatrix);
	
	inMatrix->setSize(
		getNumRowsWithFlag(inFlag, inState), 
		getNumColumnsWithFlag(inFlag, inState));
	
	int r, c, i, j;
	c=0;
	for (j=0; j<numCols; j++) {
		if (getColumnFlag(j, inFlag)==inState) {
			r=0;
			for (i=0; i<numRows; i++) {
				if (getRowFlag(i, inFlag)==inState) {
					inMatrix->setElement(r, c, getElement(i, j));
					r++;
				}
			}
			c++;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setMatrix
// ---------------------------------------------------------------------------------

void MEMatrix::setMatrix(MEMatrix *inMatrix)
{
	ThrowIfNull(inMatrix);
	setSize(inMatrix->getNumRows(), inMatrix->getNumCols());
	
	for (int i=0; i<numElements; i++)
		setElement(i, inMatrix->getElement(i));
		
	setHasLowerTriangle(inMatrix->getHasLowerTriangle());
	setHasUpperTriangle(inMatrix->getHasUpperTriangle());
	setHasDiagonal(inMatrix->getHasDiagonal());
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setMatrix
// ---------------------------------------------------------------------------------

void MEMatrix::setMatrix(double *inArray)
{
	ThrowIfNull(inArray);
	ThrowIfNull(matrix);
	
	for (int i=0; i<numElements; i++)
		setElement(i, inArray[i]);
		
	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getMatrix
// ---------------------------------------------------------------------------------

void MEMatrix::getMatrix(double *outArray)
{
	ThrowIfNull(outArray);
	ThrowIfNull(matrix);
	
	for (int i=0; i<numElements; i++)
		outArray[i]=getElement(i);
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getMatrix
// ---------------------------------------------------------------------------------

double* MEMatrix::getMatrix()
{
	double* d;
	NewArray(d, double, numElements);
	getMatrix(d);
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getNumTriangleElements
// ---------------------------------------------------------------------------------

int MEMatrix::getNumTriangleElements()
{
	if (!getIsSquare())
		throw MENotSquareMatrix();

	return (numElements-numCols)/2;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setLowerTriangle
// ---------------------------------------------------------------------------------

void MEMatrix::setLowerTriangle(double *inArray)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(inArray);
	ThrowIfNull(matrix);
	double *P=inArray;
	
	int i, j;
	for (i=1; i<numRows; i++) {
		for (j=0; j<i; j++) {
			setElement(i, j, *P);
			P++;
		}
	}
	hasLowerTriangle = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setLowerTriangleElement
// ---------------------------------------------------------------------------------

// Sets the index'th element in the lower triangle to the value f. 

void MEMatrix::setLowerTriangleElement(int index, double f)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	
	int i, j, n=0;
	for (i=1; i<numRows; i++) {
		for (j=0; j<i; j++) {
			if (n==index) {
				setElement(i, j, f);
				return;
			}
			n++;
		}
	}
	hasLowerTriangle = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getLowerTriangle
// ---------------------------------------------------------------------------------

void MEMatrix::getLowerTriangle(double *outArray)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(outArray);
	ThrowIfNull(matrix);
	double *P=outArray;
	
	int i, j;
	for (i=1; i<numRows; i++) {
		for (j=0; j<i; j++) {
			*P=getElement(i, j);
			P++;
		}
	}
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getLowerTriangle
// ---------------------------------------------------------------------------------

double* MEMatrix::getLowerTriangle()
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	double* d;
	NewArray(d, double, (numElements-numRows)/2);
	getLowerTriangle(d);
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getLowerTriangleElement
// ---------------------------------------------------------------------------------

double MEMatrix::getLowerTriangleElement(int index)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	
	int i, j, n=0;
	for (i=1; i<numRows; i++) {
		for (j=0; j<i; j++) {
			if (n==index)
				return 	getElement(i, j);
			n++;
		}
	}
	
	return 0.0;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setUpperTriangle
// ---------------------------------------------------------------------------------

void MEMatrix::setUpperTriangle(double *inArray)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(inArray);
	ThrowIfNull(matrix);		
	double *P=inArray;
	
	int i, j;
	for (i=0; i<numRows-1; i++) {
		for (j=i+1; j<numCols; j++) {
			setElement(i, j, *P);
			P++;
		}
	}
	hasUpperTriangle = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setUpperTriangleElement
// ---------------------------------------------------------------------------------

void MEMatrix::setUpperTriangleElement(int index, double f)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);		

	int i, j, n=0;	
	for (i=0; i<numRows-1; i++) {
		for (j=i+1; j<numCols; j++) {
			if (n==index) {
				setElement(i, j, f);
				return;
			}
			n++;
		}
	}
	hasUpperTriangle = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getUpperTriangle
// ---------------------------------------------------------------------------------

void MEMatrix::getUpperTriangle(double *outArray)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(outArray);
	ThrowIfNull(matrix);
	double *P=outArray;
	
	int i, j;
	for (i=0; i<numRows-1; i++) {
		for (j=i+1; j<numCols; j++) {
			*P=getElement(i, j);
			P++;
		}
	}
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getUpperTriangle
// ---------------------------------------------------------------------------------

double* MEMatrix::getUpperTriangle()
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	double* d;
	NewArray(d, double, (numElements-numRows)/2);
	getUpperTriangle(d);
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getUpperTriangleElement
// ---------------------------------------------------------------------------------

double MEMatrix::getUpperTriangleElement(int index)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	
	int i, j, n=0;
	for (i=0; i<numRows-1; i++) {
		for (j=i+1; j<numCols; j++) {
			if (n==index)
				return getElement(i, j);
			n++;
		}
	}
	
	return 0.0;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setDiagonal
// ---------------------------------------------------------------------------------

void MEMatrix::setDiagonal(double *inArray)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(inArray);
	ThrowIfNull(matrix);		
	double *P=inArray;
	
	for (int i=0; i<numRows; i++) {
		setElement(i, i, *P);
		P++;
	}
	hasDiagonal = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getDiagonal
// ---------------------------------------------------------------------------------

void MEMatrix::getDiagonal(double *outArray)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(outArray);
	ThrowIfNull(matrix);
	double *P=outArray;
	
	for (int i=0; i<numRows; i++) {
		*P=getElement(i, i);
		P++;
	}
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getDiagonal
// ---------------------------------------------------------------------------------

double* MEMatrix::getDiagonal()
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	double* d;
	NewArray(d, double, numRows);
	getDiagonal(d);
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::setRow
// ---------------------------------------------------------------------------------

void MEMatrix::setRow(int row, double *inArray)
{
	ThrowIfNull(inArray);
	ThrowIfNull(matrix);		
	double *P=inArray;
	
	for (int i=0; i<numCols; i++) {
		setElement(row, i, *P);
		P++;
	}
		
	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getRow
// ---------------------------------------------------------------------------------

void MEMatrix::getRow(int row, double *outArray)
{
	ThrowIfNull(outArray);
	ThrowIfNull(matrix);

	double *P=outArray;
	
	for (int i=0; i<numCols; i++) {
		*P=getElement(row, i);
		P++;
	}
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::getRow
// ---------------------------------------------------------------------------------

double* MEMatrix::getRow(int row)
{
	ThrowIfNull(matrix);

	double* d;
	NewArray(d, double, numCols);
	getRow(row, d);
	return d;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setColumn
// ---------------------------------------------------------------------------------

void MEMatrix::setColumn(int col, double *inArray)
{
	ThrowIfNull(inArray);
	ThrowIfNull(matrix);		

	double *P=inArray;
	
	for (int i=0; i<numRows; i++) {
		setElement(i, col, *P);
		P++;
	}
		
	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumn
// ---------------------------------------------------------------------------------

void MEMatrix::getColumn(int col, double *outArray)
{
	ThrowIfNull(outArray);
	ThrowIfNull(matrix);
	
	double *P=outArray;
	
	for (int i=0; i<numRows; i++) {
		*P=getElement(i, col);
		P++;
	}
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getColumn
// ---------------------------------------------------------------------------------

double* MEMatrix::getColumn(int col)
{
	double* d;
	NewArray(d, double, numRows);
	getColumn(col, d);
	return d;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setElement
// ---------------------------------------------------------------------------------

// Sets the element of the matrix at posistion (row, col) with value f.
void MEMatrix::setElement(int row, int col, double f)
{
	TestBounds(row, numRows);
	TestBounds(col, numCols);
	ThrowIfNull(matrix);		
	matrix[row*numCols+col]=f;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setElement
// ---------------------------------------------------------------------------------

// Sets the index'th element of the matrix with value f.

void MEMatrix::setElement(int index, double f)
{
	TestBounds(index, numElements);
	ThrowIfNull(matrix);		
	matrix[index]=f;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getElement
// ---------------------------------------------------------------------------------

// Returns the value of the element of the matrix at posistion (row, col).

double MEMatrix::getElement(int row, int col)
{
	TestBounds(row, numRows);
	TestBounds(col, numCols);
	ThrowIfNull(matrix);
	return matrix[row*numCols+col];
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getElement
// ---------------------------------------------------------------------------------

// Returns the value of the index'th element of the matrix.

double MEMatrix::getElement(int index)
{
	TestBounds(index, numElements);
	ThrowIfNull(matrix);
	return matrix[index];
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::setMissingElement
// ---------------------------------------------------------------------------------

void MEMatrix::setMissingElement(int row, int col)
{
//	setElement(row, col, NAN);
	setElement(row, col, HUGE_VAL);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getIsMissing
// ---------------------------------------------------------------------------------

bool MEMatrix::getIsMissing(int row, int col)
{
//	return isnan(getElement(row, col));
	return getElement(row, col) == HUGE_VAL;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getMinValue
// ---------------------------------------------------------------------------------

double MEMatrix::getMinValue()
{
	double min=getElement(0);
	for (int i=1; i<getNumElements(); i++) {
		if (getElement(i)<min) 
			min=getElement(i);
	}
	return min;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::getMaxValue
// ---------------------------------------------------------------------------------

double MEMatrix::getMaxValue()
{
	double max=getElement(0);
	for (int i=1; i<getNumElements(); i++) {
		if (getElement(i)>max) 
			max=getElement(i);
	}
	return max;
}

#pragma mark -
#pragma mark === Filling and Clearing ===

// ---------------------------------------------------------------------------------
//		 MEMatrix::clearMatrix
// ---------------------------------------------------------------------------------

void MEMatrix::clearMatrix()
{
	for (int i=0; i<numElements; i++)
		setElement(i, 0.0);

		
	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::fillMatrix
// ---------------------------------------------------------------------------------

void MEMatrix::fillMatrix(double f)
{
	ThrowIfNull(matrix);
	for (int i=0; i<numElements; i++)
		setElement(i, f);
		
	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::fillMatrix
// ---------------------------------------------------------------------------------

void MEMatrix::fillDiagonal(double f)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	for (int i=0; i<numRows; i++)
		setElement(i, i, f);

	hasDiagonal = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::fillOffDiagonal
// ---------------------------------------------------------------------------------

void MEMatrix::fillOffDiagonal(double f)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	int i, j;
	for (i=0; i<numRows; i++)
		for (j=0; j<numCols; j++)
			if (i!=j)
				setElement(i, j, f);

	hasLowerTriangle = true;
	hasUpperTriangle = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::fillLowerTriangle
// ---------------------------------------------------------------------------------

void MEMatrix::fillLowerTriangle(double f)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);
	
	int i, j;
	for (i=1; i<numRows; i++) {
		for (j=0; j<i; j++)
			setElement(i, j, f);
	}

	hasLowerTriangle = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::fillUpperTriangle
// ---------------------------------------------------------------------------------

void MEMatrix::fillUpperTriangle(double f)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	ThrowIfNull(matrix);		
	
	int i, j;
	for (i=0; i<numRows-1; i++) {
		for (j=i+1; j<numCols; j++)
			setElement(i, j, f);
	}

	hasUpperTriangle = true;
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::fillRow
// ---------------------------------------------------------------------------------

void MEMatrix::fillRow(int row, double f)
{
	ThrowIfNull(matrix);		
	
	int i;
	for (i=0; i<numCols; i++)
		setElement(row, i, f);

	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::fillColumn
// ---------------------------------------------------------------------------------

void MEMatrix::fillColumn(int col, double f)
{
	ThrowIfNull(matrix);		
	
	int i;
	for (i=0; i<numRows; i++)
		setElement(i, col, f);

	hasLowerTriangle = true;
	hasUpperTriangle = true;
	hasDiagonal = true;
}

//------------------------------------------------------------------------------
//    MEMatrix::setIdentity
//------------------------------------------------------------------------------
void MEMatrix::setIdentity()
{
	clearMatrix();
	fillDiagonal(1.0);
}



#pragma mark -
#pragma mark === Swapping ===

//------------------------------------------------------------------------------
//    MEMatrix::swapElements
//------------------------------------------------------------------------------
void MEMatrix::swapElements(int a, int b, int c, int d)
{
	double temp = getElement(a, b);
	setElement(a, b, getElement(c, d));
	setElement(c, d, temp);
}


//------------------------------------------------------------------------------
//    MEMatrix::swapCols
//------------------------------------------------------------------------------
void MEMatrix::swapCols(int c1, int c2)
{
	for (int r = 0; r < numRows; r++)
		swapElements(r, c1, r, c2);
}


//------------------------------------------------------------------------------
//    MEMatrix::swapRows
//------------------------------------------------------------------------------
void MEMatrix::swapRows(int r1, int r2)
{
	for (int c = 0; c < numCols; c++)
		swapElements(r1, c, r2, c);
}


#pragma mark -
#pragma mark === Arithmatic ===

//------------------------------------------------------------------------------
//    MEMatrix::multiplyBy
//------------------------------------------------------------------------------

// Multiplies this matrix by inMatrix - storing the results in this
// matrix - will change the dimensions.

void MEMatrix::multiplyBy(MEMatrix *inMatrix)
{
	double sum;
	if (numCols != inMatrix->getNumRows())
		throw MEWrongDimensions();
		
	int c1 = inMatrix->getNumCols();
	MEMatrix tM(numRows, c1);
	int i, c, r;
	for (r = 0; r < numRows; r++) {
		for (c = 0; c < c1; c++) {
			sum = 0.0;
			for (i = 0; i < numCols; i++)
				sum += getElement(r, i) * inMatrix->getElement(i, c);
			tM.setElement(r, c, sum);
		}
	}
	setMatrix(&tM);
}

//------------------------------------------------------------------------------
//    MEMatrix::multiplyBy
//------------------------------------------------------------------------------

void MEMatrix::multiplyBy(double f)
{
	int c, r;
	for (r = 0; r < numRows; r++) {
		for (c = 0; c < numCols; c++)
			setElement(r, c, getElement(r, c)*f);
	}
}

//------------------------------------------------------------------------------
//    MEMatrix::multiply
//------------------------------------------------------------------------------

// Multiplies this matrix by inMatrix storing the result in outMatrix - which
// must be the right dimensions.

void MEMatrix::multiply(MEMatrix *outMatrix, MEMatrix *inMatrix)
{
	double sum;
	if (numCols != inMatrix->getNumRows())
		throw MEWrongDimensions();
	
	int i, c, r;
	int c1 = inMatrix->getNumCols();
	for (r = 0; r < numRows; r++) {
		for (c = 0; c < c1; c++) {
			sum = 0.0;
			for (i = 0; i < numCols; i++)
				sum += getElement(r, i) * inMatrix->getElement(i, c);
			outMatrix->setElement(r, c, sum);
		}
	}
}

//------------------------------------------------------------------------------
//    MEMatrix::multiply
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::multiply(MEMatrix *inMatrix)
{
	if (numCols != inMatrix->getNumRows())
		throw MEWrongDimensions();
		
	MEMatrix *matrix;
	NewClass(matrix, MEMatrix(numRows, inMatrix->getNumCols()));
	multiply(matrix, inMatrix);
	return matrix;
}

//------------------------------------------------------------------------------
//    MEMatrix::plus
//------------------------------------------------------------------------------

// Adds inMatrix to this matrix, storing the result in this matrix.

void MEMatrix::plus(MEMatrix *inMatrix)
{
	int c, r;
	for (r = 0; r < numRows; r++)
		for (c = 0; c < numCols; c++)
			setElement(r, c, getElement(r, c) + inMatrix->getElement(r, c));
}

//------------------------------------------------------------------------------
//    MEMatrix::plus
//------------------------------------------------------------------------------

// Adds inMatrix to this matrix, storing the result in this matrix.

void MEMatrix::plus(double f)
{
//	AER -- the following isn't very fast:
//	int c, r;
//	for (r = 0; r < numRows; r++)
//		for (c = 0; c < numCols; c++)
//			setElement(r, c, getElement(r, c) + f);
//	*This* is fast:
	int i, k = getNumElements();
	for (i = 0; i < k; i++)
		matrix[i] += f;
}


//------------------------------------------------------------------------------
//    MEMatrix::plus
//------------------------------------------------------------------------------
void MEMatrix::plus(int r, int c, double d)
{
	setElement(r, c, getElement(r, c) + d);
}


//------------------------------------------------------------------------------
//    MEMatrix::add
//------------------------------------------------------------------------------

// Adds inMatrix to this matrix, storing the result in outMatrix - which
// must be the right dimensions.

void MEMatrix::add(MEMatrix *outMatrix, MEMatrix *inMatrix)
{
	int c, r;
	for (r = 0; r < numRows; r++)
		for (c = 0; c < numCols; c++)
			outMatrix->setElement(r, c, getElement(r, c) + inMatrix->getElement(r, c));
}

//------------------------------------------------------------------------------
//    MEMatrix::add
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::add(MEMatrix *inMatrix)
{
	MEMatrix *matrix;
	NewClass(matrix, MEMatrix(numRows, numCols));
	add(matrix, inMatrix);
	return matrix;
}

//------------------------------------------------------------------------------
//    MEMatrix::minus
//------------------------------------------------------------------------------

// Subtracts inMatrix from this matrix, storing the result in this matrix.

void MEMatrix::minus(MEMatrix *inMatrix)
{
	int c, r;
	for (r = 0; r < numRows; r++)
		for (c = 0; c < numCols; c++)
			setElement(r, c, getElement(r, c) - inMatrix->getElement(r, c));
}

//------------------------------------------------------------------------------
//    MEMatrix::minus
//------------------------------------------------------------------------------

// Subtracts f from this matrix, storing the result in this matrix.

void MEMatrix::minus(double f)
{
	int c, r;
	for (r = 0; r < numRows; r++)
		for (c = 0; c < numCols; c++)
			setElement(r, c, getElement(r, c) - f);
}

//------------------------------------------------------------------------------
//    MEMatrix::subtract
//------------------------------------------------------------------------------

// Subtracts inMatrix to this matrix, storing the result in outMatrix - which
// must be the right dimensions.

void MEMatrix::subtract(MEMatrix *outMatrix, MEMatrix *inMatrix)
{
	int c, r;
	for (r = 0; r < numRows; r++)
		for (c = 0; c < numCols; c++)
			outMatrix->setElement(r, c, getElement(r, c) - inMatrix->getElement(r, c));
}

//------------------------------------------------------------------------------
//    MEMatrix::subtract
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::subtract(MEMatrix *inMatrix)
{
	MEMatrix *matrix;
	NewClass(matrix, MEMatrix(numRows, numCols));
	subtract(matrix, inMatrix);
	return matrix;
}

//------------------------------------------------------------------------------
//    MEMatrix::divideBy
//------------------------------------------------------------------------------

// Divides this matrix by f, storing the result in this matrix.

void MEMatrix::divideBy(double f)
{
	int c, r;
	for (r = 0; r < numRows; r++)
		for (c = 0; c < numCols; c++)
			setElement(r, c, getElement(r, c) / f);
}

#pragma mark -
#pragma mark === Operators ===

//------------------------------------------------------------------------------
//    MEMatrix::operator*=
//------------------------------------------------------------------------------
MEMatrix& MEMatrix::operator*=(MEMatrix& M)
{
	multiplyBy(&M);
	
	return *this;
}

MEMatrix& MEMatrix::operator*=(double& d)
{
	for (int i = 0; i < numElements; i++)
		setElement(i, getElement(i) * d);
	return *this;
}


//------------------------------------------------------------------------------
//    MEMatrix::operator/=
//------------------------------------------------------------------------------
MEMatrix& MEMatrix::operator/=(double& d)
{
	for (int i = 0; i < numElements; i++)
		setElement(i, getElement(i) / d);
	return *this;
}


//------------------------------------------------------------------------------
//    MEMatrix::operator+=
//------------------------------------------------------------------------------
MEMatrix& MEMatrix::operator+=(MEMatrix& M)
{
	if (numRows != M.getNumRows())
		throw MEWrongDimensions();
	if (numCols != M.getNumCols())
		throw MEWrongDimensions();
		
	for (int i = 0; i < numElements; i++)
		setElement(i, getElement(i)+M.getElement(i));
	return *this;
}

MEMatrix& MEMatrix::operator+=(double d)
{
	for (int i = 0; i < numElements; i++)
		setElement(i, getElement(i)+d);
	return *this;
}


//------------------------------------------------------------------------------
//    MEMatrix::operator-=
//------------------------------------------------------------------------------
MEMatrix& MEMatrix::operator-=(MEMatrix& M)
{
	if (numRows != M.getNumRows())
		throw MEWrongDimensions();
	if (numCols != M.getNumCols())
		throw MEWrongDimensions();
		
	for (int i = 0; i < numElements; i++)
		setElement(i, getElement(i)-M.getElement(i));
	return *this;
}

MEMatrix& MEMatrix::operator-=(double d)
{
	for (int i = 0; i < numElements; i++)
		setElement(i, getElement(i)-d);
	return *this;
}

//------------------------------------------------------------------------------
//    MEMatrix::operator=
//------------------------------------------------------------------------------
MEMatrix& MEMatrix::operator=(MEMatrix& M)
{
	setMatrix(&M);
	return *this;
}

MEMatrix& MEMatrix::operator=(double *D)
{
	setMatrix(D);
	return *this;
}

MEMatrix& MEMatrix::operator=(double d)
{
	fillMatrix(d);
	return *this;
}

#pragma mark -
#pragma mark === Matrix Maths ===

//------------------------------------------------------------------------------
//    MEMatrix::jacobiRotate
//------------------------------------------------------------------------------
void MEMatrix::jacobiRotate(int i, int j, int k, int l, double s, double tau)
{
	double g = getElement(i, j);
	double h = getElement(k, l);
	setElement(i, j, g-s*(h+g*tau));
	setElement(k, l, h+s*(g-h*tau));
}


//------------------------------------------------------------------------------
//    MEMatrix::jacobi
//------------------------------------------------------------------------------

//	jacobi uses Jacobian rotations to diagonalise a REAL, SYMMETRIC MEMatrix and 
//	put the eigenvalues into the array d. The eigenvectors are put into MEMatrix
//	V.

bool MEMatrix::jacobi(double *d, MEMatrix *V, int *nrot)
{
	int j, iq, ip, i;
	double tresh, theta, tau, t, sum, s, h, g, c, *b, *z;
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	NewArray(b, double, numRows);
	NewArray(z, double, numRows);

	V->setIdentity();

	getDiagonal(b);					// initialize b
	for (ip = 0; ip < numRows; ip++) {	// ... and d to the diagonal of A
		d[ip] = b[ip];
		z[ip] = 0.0;
	}
	*nrot = 0;
	for (i = 1; i <= 50; i++) {
		sum = 0.0;
		for (ip = 0; ip < numRows-1; ip++)
			for (iq = ip+1; iq < numRows; iq++)
				sum += fabs(getElement(ip, iq));
		if (sum == 0.0)
			return true;
		if (i < 4)
			tresh = 0.2 * sum/((double) (numElements));
		else
			tresh = 0.0;

		for (ip = 0; ip < numRows-1; ip++) {
			for (iq = ip+1; iq < numRows; iq++) {
				g = 100.0 * fabs(getElement(ip, iq));
				if ((i > 4) && (fabs(d[ip])+g == fabs(d[ip])) && (fabs(d[iq])+g == fabs(d[iq])))
					setElement(ip, iq, 0.0);
				else if (fabs(getElement(ip, iq)) > tresh) {
					h = d[iq]-d[ip];

					if (fabs(h)+g == fabs(h))
						t = getElement(ip,iq)/h;
					else {
						theta = 0.5*h/getElement(ip,iq);
						t = 1.0/(fabs(theta) + sqrt(1.0 + theta*theta));
						if (theta < 0.0)
							t = -t;
					}

					c = 1.0/sqrt(1.0 + t*t);

					s = t*c;

					tau = s/(1.0+c);

					h = t*getElement(ip,iq);
					z[ip] -= h;
					z[iq] += h;
					d[ip] -= h;
					d[iq] += h;
					setElement(ip, iq, 0.0);

					for (j = 0; j < ip; j++) {
						jacobiRotate(j, ip, j, iq, s, tau);
					}
					for (j = ip+1; j < iq; j++) {
						jacobiRotate(ip, j, j, iq, s, tau);
					}
					for (j = iq+1; j < numRows; j++) {
						jacobiRotate(ip, j, iq, j, s, tau);
					}
					for (j = 0; j < numRows; j++) {
						V->jacobiRotate(j, ip, j, iq, s, tau);
					}
					*nrot++;
				}
			}
		}
		for (ip = 0; ip < numRows; ip++) {
			b[ip] += z[ip];
			d[ip] = b[ip];
			z[ip] = 0.0;
		}
	}
	return false;
}

//------------------------------------------------------------------------------
//    MEMatrix::eigenSort
//------------------------------------------------------------------------------
void MEMatrix::eigenSort(double *d, MEMatrix *E)
{
	double p;
	
	int dim = E->getNumRows();
	int i, j, k;
	
	for (i = 0; i < dim; i++) {
		p = d[k=i];
		for (j = i+1; j < dim; j++)
			if (d[j] >= p)
				p = d[k=j];
		if (k != i) {
			d[k] = d[i];
			d[i] = p;
			E->swapCols(i, k);
		}
	}
}


//------------------------------------------------------------------------------
//    MEMatrix::lubksb	Lower/Upper Back Substitution
//------------------------------------------------------------------------------
void MEMatrix::lubksb(int *indx, double *b)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();
		
	int i, ii=-1, ip, j;
	double sum;
	for (i = 0; i < numRows; i++) {
		ip = indx[i];
		sum = b[ip];
		b[ip] = b[i];
		if (ii>=0)
			for (j = ii; j < i; j++)
				sum -= getElement(i, j)*b[j];
		else if (sum)
			ii = i;
		b[i] = sum;
	}
	for (i = numRows-1; i >= 0; i--) {
		sum = b[i];
		for (j = i+1; j < numRows; j++)
			sum -= getElement(i, j) * b[j];
		b[i] = sum/getElement(i, i);
	}
}


//------------------------------------------------------------------------------
//    MEMatrix::ludcmp Lower/Upper Decomposition
//------------------------------------------------------------------------------
bool MEMatrix::ludcmp(int *indx, double *sign)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	double *vv;
	NewArray(vv, double, numRows);	// implicit scaling for each row

	int	 i, j, k, imax=0;
	double big=0, dum=0, sum=0, temp=0;
	*sign = 1.0;	// no row interchanges as yet
	
	for (i = 0; i < numRows; i++) {	// loop over rows to get implicit scaling information
		big = 0.0;
		for (j = 0; j < numRows; j++) {
			temp = fabs(getElement(i, j));
			if (temp > big)
				big = temp;
		}
		
		if (big == 0.0) {
			delete [] vv;
			return false;
		}
		
		vv[i] = 1.0/big;
	}
	
	for (j = 0; j < numRows; j++) {
		for (i = 0; i < j; i++) {
			sum = getElement(i, j);
			for (k = 0; k < i; k++)
				sum -= getElement(i, k) * getElement(k, j);
			setElement(i, j, sum);
		}
		
		big = 0.0;
		for (i = j; i < numRows; i++) {
			sum = getElement(i, j);
			for (k = 0; k < j; k++)
				sum -= getElement(i, k) * getElement(k, j);

			setElement(i, j, sum);
			dum = vv[i] * fabs(sum);
			if (dum >= big) {
				big = dum;
				imax = i;
			}
		}

		if (j != imax) {
			swapRows(imax, j);

			*sign = -(*sign);
			vv[imax] = vv[j];
		}
		if (indx != NULL)
			indx[j] = imax;

		if (getElement(j, j) == 0.0) { // Non-singular
			delete [] vv;
			return false;
		}

		if (j != numRows-1) {
			dum = 1.0/getElement(j, j);
			for (i = j+1; i < numRows; i++) {
				temp = getElement(i, j) * dum;
				setElement(i, j, temp);
			}
		}
	}
	
	delete [] vv;
	return true;
}


//------------------------------------------------------------------------------
//    MEMatrix::det
//------------------------------------------------------------------------------
double MEMatrix::det()
{
	double d = 0.0;
	MEMatrix tM(numRows);
	
	tM.setMatrix(this);

	if (!tM.ludcmp(NULL, &d))	// matrix is singular, i.e. has determinant zero
		return 0.0;
	
	for (int j = 0; j < numRows; j++)
		d *= tM.getElement(j, j);
	
	return d;
}


//------------------------------------------------------------------------------
//    MEMatrix::logdet
//------------------------------------------------------------------------------
double MEMatrix::logdet()
{
	double d = 0.0;
	MEMatrix tM(numRows);
	
	tM.setMatrix(this);

	if (!tM.ludcmp(NULL, &d))	// matrix is singular, i.e. has determinant zero
 		throw MEException("Matrix is singular, i.e. has zero determinant.");
	
	if (d < 0.0)
 		throw MEException("Negative determinant cannot be logged.");
 		
	double v;
	d = 0.0;
	for (int j = 0; j < numRows; j++) {
		v = tM.getElement(j, j);
		if (v <= 0.0) 
 			throw MEException("Negative determinant cannot be logged.");
			
		d += log(v);
	}
	
	return d;
}


//------------------------------------------------------------------------------
//    MEMatrix::diagonalise
//------------------------------------------------------------------------------
bool MEMatrix::diagonalise(double *d, MEMatrix *E, MEMatrix *iE)
{
	int ignored;
	if (!jacobi(d, E, &ignored))
		return false;
	eigenSort(d, E);
	E->invert(iE);
	return true;
}


//------------------------------------------------------------------------------
//    MEMatrix::invert
//------------------------------------------------------------------------------
bool MEMatrix::invert(MEMatrix *outMatrix)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	int i, j;
	double d;

/*	// First check to see if the off-diagonal elements are zero
	double sum=0.0;
	for (i = 0; i < numRows; i++) {
		for (j = 0; j < numCols; j++) {
			if (i!=j)
				sum+=fabs(getElement(i, j));
		}
	}
	
	if (sum<1.0E-18) {
		// the off-diagonal elements are pretty close to zero
		// just take the reciprocal of each diagonal element
		for (i = 0; i < numRows; i++) {
			d=getElement(i, i);
			if (fabs(d) < 1.0E-18) 
				return false;
			outMatrix->setElement(i, i, 1.0 / d);
		}
	} else {*/
		int *indx;
		NewArray(indx, int, numRows);
		double *col;
		NewArray(col, double, numRows);
		
		MEMatrix tmp(numRows);
		tmp.setMatrix(this);

		if (!tmp.ludcmp(indx, &d)) {
			delete [] indx;
			delete [] col;
			return false;
		}
		
		for (j = 0; j < numRows; j++) {
			for (i = 0; i < numRows; i++)
				col[i] = 0.0;
			col[j] = 1.0;
			tmp.lubksb(indx, col);
			outMatrix->setColumn(j, col);
		}
		
		delete [] indx;
		delete [] col;
//	}
	
	return true;
}


//------------------------------------------------------------------------------
//    MEMatrix::invert
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::invert()
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	MEMatrix *matrix;
	NewClass(matrix, MEMatrix(numRows));
	
	if (!invert(matrix)) {
		delete matrix;
		return NULL;
	}
	
	return matrix;
}


//------------------------------------------------------------------------------
//    MEMatrix::invertProduct
//------------------------------------------------------------------------------
bool MEMatrix::invertProduct(MEMatrix *outMatrix, MEMatrix *inMatrix)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	int *indx;
	NewArray(indx, int, numRows);
	double *col;
	NewArray(col, double, numRows);
	
	MEMatrix tmp(numRows);
	tmp.setMatrix(this);

	double d;
	if (!tmp.ludcmp(indx, &d)) {
		delete [] indx;
		delete [] col;
		return false;
	}
	
	int j;
	for (j = 0; j < numRows; j++) {
		inMatrix->getColumn(j, col);
		tmp.lubksb(indx, col);
		outMatrix->setColumn(j, col);
	}
	
	delete [] indx;
	delete [] col;
	return true;
}


//------------------------------------------------------------------------------
//    MEMatrix::invertProduct
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::invertProduct(MEMatrix *inMatrix)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	MEMatrix *matrix;
	NewClass(matrix, MEMatrix(numRows));
	
	if (!invertProduct(matrix, inMatrix)) {
		delete matrix;
		return NULL;
	}
	
	return matrix;
}


//------------------------------------------------------------------------------
//    MEMatrix::raise
//------------------------------------------------------------------------------
bool MEMatrix::raise(MEMatrix *outMatrix, double d)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();

	MEMatrix iE(numRows), M(numRows);
	double *v;
	NewArray(v, double, numRows);

	if (!diagonalise(v, outMatrix, &iE))
		return false;

	M.clearMatrix();
	for (int i = 0; i < numRows; i++)
		M.setElement(i, i, pow(v[i], d));
	M.multiplyBy(&iE);
	outMatrix->multiplyBy(&M);
	
	return true;
}


//------------------------------------------------------------------------------
//    MEMatrix::raise
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::raise(double d)
{
	if (numCols!=numRows)
		throw MENotSquareMatrix();
		
	MEMatrix *matrix;
	NewClass(matrix, MEMatrix(numRows));
	
	if (!raise(matrix, d)) {
		delete matrix;
		return NULL;
	}
	
	return matrix;
}

//------------------------------------------------------------------------------
//    MEMatrix::transpose
//------------------------------------------------------------------------------
void MEMatrix::transpose(MEMatrix *outMatrix)
{
	int i, j;
	
	for (i=0; i<numRows; i++)
		for (j=0; j<numCols; j++)
			outMatrix->setElement(j, i, getElement(i, j));
}


//------------------------------------------------------------------------------
//    MEMatrix::transpose
//------------------------------------------------------------------------------
MEMatrix *MEMatrix::transpose()
{
	MEMatrix *matrix=new MEMatrix(numCols, numRows);
	transpose(matrix);
	return matrix;
}


//------------------------------------------------------------------------------
//    MEMatrix::kroneckerProduct
//------------------------------------------------------------------------------
void MEMatrix::kroneckerProduct(MEMatrix *outMatrix, MEMatrix *inMatrix)
{
	int c1, r1, c2, r2;
	int nr = inMatrix->getNumRows();
	int nc = inMatrix->getNumCols();
	for (r1 = 0; r1 < nr; r1++) {
		for (c1 = 0; c1 < nc; c1++) {
			for (r2 = 0; r2 < numRows; r2++) {
				for (c2 = 0; c2 < numCols; c2++) {
					outMatrix->setElement(r1*numRows+r2, c1*numCols+c2, 
						getElement(r2, c2) * inMatrix->getElement(r1, c1));
				}
			}
					
		}
	}
}


#pragma mark -
#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 MEMatrix::readObject
// ---------------------------------------------------------------------------------

void MEMatrix::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
	stream->get("Rows", &numRows);
	stream->get("Cols", &numCols);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::readData
// ---------------------------------------------------------------------------------

void MEMatrix::readData(MEStream *stream)
{
//	stream->getMatrix(this);
}

// ---------------------------------------------------------------------------------
//		 MEMatrix::writeObject
// ---------------------------------------------------------------------------------

void MEMatrix::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
	stream->put("Rows", numRows);
	stream->put("Cols", numCols);
}


// ---------------------------------------------------------------------------------
//		 MEMatrix::writeData
// ---------------------------------------------------------------------------------

void MEMatrix::writeData(MEStream *stream)
{
//	stream->putMatrix(this);
}
