/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEMatrix.h

		Interface for the MEMatrix Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut & Mike Charleston"
//! group = "Matrices"
 
#ifndef _MEMATRIX_H_
#define _MEMATRIX_H_

#include "MEObject.h"

class MEList;
class MEStream;

class MEMatrix : public MEObject {
	public:
		virtual long		getID() const { return 'MMat'; }
		//: The ID of this object - 'MMat'.
	
		virtual char*		getClassName() const { return "Matrix"; }
		//: Returns the name of this class - "Matrix"
		
	protected:
		int					numRows;
		int					numCols;
		int					numElements;
		double*				matrix;
		
		bool				isSymmetric, isSymmetricKnown;
		bool				hasLowerTriangle, hasUpperTriangle, hasDiagonal;
		

		MEList				*rowList, *colList;
		
	public:
	 						MEMatrix();						
	 						MEMatrix(int inDimension);						
	 						MEMatrix(int inNumRows, int inNumCols);	
	 										
	 	virtual				~MEMatrix();

		virtual void		setSize(int inDimension);
		virtual void		setSize(int inNumRows, int inNumCols);
		void				setNumCols(int inNumCols) { numCols=inNumCols; };
		int					getNumCols() { return numCols; };
		int					getNumRows() { return numRows; };
		int					getNumElements() { return numElements; };

		virtual MEList		*getRowList();
		virtual MEObject	*getRowObject(int row);
		virtual void		setRowName(int row, char *inName);
		virtual char		*getRowName(int row);
		virtual void		getRowName(int row, char *inName, int inMaxLen);
		virtual int			findRowName(char *inName);

		virtual	MEList		*getColumnList();
		virtual MEObject	*getColumnObject(int col);
		virtual void		setColumnName(int col, char *inName);
		virtual char		*getColumnName(int col);
		virtual void		getColumnName(int col, char *inName, int inMaxLen, char *inDefault = NULL);
		virtual int			findColumnName(char *inName);

		virtual int			getNumRowsWithFlag(short inFlag, bool inState);
		virtual bool		getRowFlag(int row, short inFlag);
		virtual void		setRowFlag(int row, short inFlag, bool inState);
		virtual void		setRowFlagForAll(short inFlag, bool inState);

		virtual int			getNumColumnsWithFlag(short inFlag, bool inState);
		virtual bool		getColumnFlag(int col, short inFlag);
		virtual void		setColumnFlag(int col, short inFlag, bool inState);
		virtual void		setColumnFlagForAll(short inFlag, bool inState);

		virtual void		extractMatrixWithFlag(MEMatrix *inMatrix, short inFlag, bool inState);

		bool				getIsSquare() { return numRows==numCols; };

		bool				getIsSymmetric();
		void				makeSymmetric(bool fromLowerTriangle);

		virtual void 		setHasLowerTriangle(bool inHasLowerTriangle) { hasLowerTriangle = inHasLowerTriangle; }
		virtual bool 		getHasLowerTriangle() { return hasLowerTriangle; }
		virtual void 		setHasUpperTriangle(bool inHasUpperTriangle) { hasUpperTriangle = inHasUpperTriangle; }
		virtual bool 		getHasUpperTriangle() { return hasUpperTriangle; }
		virtual void 		setHasDiagonal(bool inHasDiagonal) { hasDiagonal = inHasDiagonal; }
		virtual bool 		getHasDiagonal() { return hasDiagonal; }
		
		void				setMatrix(MEMatrix *inMatrix);
		void				setMatrix(double *inArray);
		void				getMatrix(double *inArray);
		double*				getMatrix();
		
		int					getNumTriangleElements();

		void				setLowerTriangle(double *inArray);
		void				setLowerTriangleElement(int index, double f);
		void				getLowerTriangle(double *outArray);
		double*				getLowerTriangle();
		double				getLowerTriangleElement(int index);
		
		void				setUpperTriangle(double *inArray);
		void				setUpperTriangleElement(int index, double f);
		void				getUpperTriangle(double *outMatrix);
		double*				getUpperTriangle();
		double				getUpperTriangleElement(int index);
		
		void				setDiagonal(double *inArray);
		void				getDiagonal(double *outArray);
		double*				getDiagonal();
	
		void				setRow(int row, double *inArray);
		void				getRow(int row, double *outArray);
		double*				getRow(int row);
		
		void				setColumn(int col, double *inArray);
		void				getColumn(int col, double *outArray);
		double*				getColumn(int col);
		
		virtual void		setElement(int row, int col, double f);
		virtual void		setElement(int index, double f);
		virtual double		getElement(int row, int col);
		virtual double		getElement(int index);
		
		virtual double		getMinValue();
		virtual double		getMaxValue();

		void				setMissingElement(int row, int col);
		bool				getIsMissing(int row, int col);

		void				clearMatrix();
		void				fillMatrix(double f);
		void				fillDiagonal(double f);
		void				fillOffDiagonal(double f);
		void				fillLowerTriangle(double f);
		void				fillUpperTriangle(double f);
		void				fillRow(int row, double f);
		void				fillColumn(int row, double f);
		void				setIdentity();

		void				swapElements(int a, int b, int c, int d);
		void				swapCols(int c1, int c2);
		void				swapRows(int r1, int r2);
		
		void				multiplyBy(MEMatrix *inMatrix);
		void				multiplyBy(double f);
		
		void				multiply(MEMatrix *outMatrix, MEMatrix *inMatrix);
		MEMatrix			*multiply(MEMatrix *inMatrix);
		
		void				plus(MEMatrix *inMatrix);
		void				plus(double f);
		void				plus(int r, int c, double d);
		
		void				add(MEMatrix *outMatrix, MEMatrix *inMatrix);
		MEMatrix			*add(MEMatrix *inMatrix);
		
		void				minus(MEMatrix *inMatrix);
		void				minus(double f);
		
		void				subtract(MEMatrix *outMatrix, MEMatrix *inMatrix);
		MEMatrix			*subtract(MEMatrix *inMatrix);
		
		void				divideBy(double f);
		
		MEMatrix&			operator*=(MEMatrix& M);	// usage: B *= M
		MEMatrix&			operator*=(double& d);		// usage: B *= d
		MEMatrix&			operator/=(double& d);		// usage: B /= d
		MEMatrix&			operator+=(MEMatrix& M);	// usage: B += M
		MEMatrix&			operator+=(double d);		// usage: B += d
		MEMatrix&			operator-=(double d);		// usage: B -= d
		MEMatrix&			operator-=(MEMatrix& M);	// usage: B -= M
		MEMatrix&			operator=(MEMatrix& M);		// usage: B = M
		MEMatrix&			operator=(double *D);		// usage: B = d
		MEMatrix&			operator=(double d);

	protected:
		void				jacobiRotate(int i, int j, int k, int l, double s, double tau);
		bool				jacobi(double *d, MEMatrix *V, int *nrot);
		void				eigenSort(double *d, MEMatrix *E);
		void				lubksb(int *indx, double *b);
		bool				ludcmp(int *indx, double *sign);
		bool				diagonalise(double *d, MEMatrix *E, MEMatrix *iE);
		
	public:
		double				det();
		double				logdet();
		
		bool				invert(MEMatrix *outMatrix);
		MEMatrix			*invert();
		
		bool				invertProduct(MEMatrix *inMatrix, MEMatrix *outMatrix);
		MEMatrix			*invertProduct(MEMatrix *inMatrix);
		
		bool				raise(MEMatrix *outMatrix, double d);
		MEMatrix			*raise(double d);
		
		void	 			transpose(MEMatrix *outMatrix);
		MEMatrix 			*transpose();

		void 				kroneckerProduct(MEMatrix *outMatrix, MEMatrix *inMatrix);
				
		virtual void		readObject(MEStream *stream);
		virtual void		readData(MEStream *stream);
		virtual void		writeObject(MEStream *stream);
		virtual void		writeData(MEStream *stream);
};

#endif /* _MEMATRIX_H_ */
