/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MENexusCreator.cpp

				MENexusCreator Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEExceptions.h"

#include "CWriter.h"
#include "MECreator.h"
#include "MENexusCreator.h"
#include "MEBasicSequence.h"
#include "MEBasicAlignment.h"
#include "METreeList.h"
#include "METree.h"
#include "MEPhylogenyList.h"
#include "MEPhylogeny.h"
#include "MEDistanceMatrix.h"
#include "METaxon.h"
#include "METaxonList.h"
#include "MEMatrix.h"
#include "MEDistanceMatrix.h"

/******************************************************************************
 MENexusCreator
 
 ******************************************************************************/

MENexusCreator::MENexusCreator() : MECreator()
{
	hasHeader=false;
	useTaxaBlock=false;
	hasTaxaBlock = false;
	useTranslationList=true;
	hasTranslationList=false;
	treesBlockOpen=false;
	
	translationList = NULL;
}


MENexusCreator::MENexusCreator(CWriter *inWriter) : MECreator(inWriter)
{
	hasHeader=false;
	useTaxaBlock=false;
	hasTaxaBlock = false;
	useTranslationList=true;
	hasTranslationList=false;
	treesBlockOpen=false;
	
	translationList = NULL;
}

MENexusCreator::~MENexusCreator()
{
	if (translationList != NULL)
		delete translationList;
}

void MENexusCreator::writeTaxaBlock(METaxonList *inTaxonList)
{
	if (treesBlockOpen) {
		writer->writeString("End;\r\r");
		treesBlockOpen = false;
	}
	
	hasTaxaBlock = true;
	
	int i, n = inTaxonList->getNumTaxa();
	
	writer->writeString("Begin TAXA;\r");
	
	writer->writeString("\tDimensions");
	writer->writeString(" ntax=");
	writer->writeInteger(n);
	
	writer->writeString(";\r\tTaxLabels\r");
	
	char *name;
	
	for (i=0; i<n; i++) {
		name=inTaxonList->getTaxonName(i);
		
		writer->writeChar('\t');
		writer->writeChar('\t');
		writer->writeLabel(name);
		writer->writeChar('\r');
	}
	writer->writeString(";\r");

	writer->writeString("End;\r\r");
}

void MENexusCreator::writeDataBlockHeader(MEBasicAlignment *inAlignment, bool taxa, bool chars)
{
	writer->writeString("\tDimensions");
	if (taxa) {
		writer->writeString(" ntax=");
		writer->writeInteger(inAlignment->getNumSequences());
	}
	if (chars) {
		writer->writeString(" nchar=");
		writer->writeInteger(inAlignment->getNumSites());
	}
	writer->writeString(";\r\tFormat ");
	
	if (inAlignment->getAlignmentType()==kAminoAcidAlignment)
		writer->writeString("datatype=PROTEIN ");
	else
		writer->writeString("datatype=NUCLEOTIDE ");
	
	if (interleaved)
		writer->writeString("interleave ");
	
	writer->writeString("gap=-;\r");
}

void MENexusCreator::writeCharactersBlock(MEBasicAlignment *inAlignment)
{
	closeTreesBlock();

	writer->writeString("Begin CHARACTERS;\r");
	writeDataBlockHeader(inAlignment, false, true);
	
	writer->writeString("\tMatrix\r");
	writeSequenceData(inAlignment, -1, 0, true, true, true, true, true, false, 0);
	writer->writeString(";\r");
	
	writer->writeString("End;\r\r");
}

void MENexusCreator::writeDataBlock(MEBasicAlignment *inAlignment)
{
	closeTreesBlock();

	writer->writeString("Begin DATA;\r");
	writeDataBlockHeader(inAlignment, true, true);
	
	writer->writeString("\tMatrix\r");
	writeSequenceData(inAlignment, -1, 0, false, true, true, true, true, false, 0);
	writer->writeString(";\r");

	writer->writeString("End;\r\r");
}

void MENexusCreator::writeTranslationList(METree *inTree)
{
	writer->writeString("\r\tTranslate\r");

	if (translationList == NULL)
		NewClass(translationList, METaxonList());
	else
		translationList->deleteAll();

	char *name;
	int i = 0;
	MELink *tip = inTree->getFirstTip();
	METaxon *taxon;
	while (tip!=NULL) {
		name=inTree->getTipName(tip);
		NewClass(taxon, METaxon());
		taxon->setName(name);
		translationList->add(taxon, true);
		
		writer->writeChar('\t');
		writer->writeChar('\t');
		writer->writeInteger(i+1);
		writer->writeChar('\t');
		writer->writeLabel(name);
		
		tip = inTree->getNextTip(tip);
		i++;
		
		if (tip!=NULL)
			writer->writeChar(',');
		writer->writeChar('\r');
	}
	writer->writeString("\t\t;\r");
	
	hasTranslationList = true;
}

void MENexusCreator::writeTreesBlock(METree *inTree)
{
	char *name;
	
	if (!treesBlockOpen) {
		if (useTaxaBlock && !hasTaxaBlock) 
			writeTaxaBlock(inTree->getTaxonList());
			
		writer->writeString("Begin TREES;\r");
		treesBlockOpen = true;
	}

	bool useNumericalTips = inTree->getUseNumericalTips();
	if (useTranslationList && inTree->getTaxonList() != NULL) {
		if (!hasTranslationList || inTree->compareTaxa(translationList) < 1)
			writeTranslationList(inTree);
		inTree->setUseNumericalTips(true);
		inTree->setTranslationList(translationList);
	}		
	
	if (inTree->getIsRooted())
		writer->writeString("TREE\t");
	else
		writer->writeString("UTREE\t");
	
	name=inTree->getName();
	if (name==NULL || strlen(name)==0)
		writer->writeString("tree1");
	else
		writer->writeLabel(name);
	
	writer->writeString(" = ");
	writeTree(inTree);
	writer->writeChar('\r');

	inTree->setUseNumericalTips(useNumericalTips);
	inTree->setTranslationList(NULL);
}

void MENexusCreator::writeTreesBlock(METreeList *inTreeList)
{
	int i, n = inTreeList->getNumTrees();
	METree *tree;
	char *name;
	
	if (!treesBlockOpen) {
		if (useTaxaBlock && !hasTaxaBlock) {
			ThrowIfNull(inTreeList->getTaxonList());
			writeTaxaBlock(inTreeList->getTaxonList());
		}
		
		writer->writeString("Begin TREES;\r");
		treesBlockOpen = true;
	}

	bool useNumericalTips;
	for (i=0; i<n; i++) {
		tree=inTreeList->getTree(i);
		
		useNumericalTips = tree->getUseNumericalTips();
		
		if (useTranslationList && tree->getTaxonList() != NULL) {
			if (!hasTranslationList || tree->compareTaxa(translationList) < 1)
				writeTranslationList(tree);
			tree->setUseNumericalTips(true);
			tree->setTranslationList(translationList);
		}
		
		if (tree->getIsRooted())
			writer->writeString("TREE\t");
		else
			writer->writeString("UTREE\t");
		
		name=tree->getName();
		if (name==NULL || strlen(name)==0) {
			writer->writeString("tree");
			writer->writeInteger(i+1);
		} else
			writer->writeLabel(name);
		
		writer->writeString(" = ");
		
		writeTree(tree);
		writer->writeChar('\r');

		tree->setUseNumericalTips(useNumericalTips);
		tree->setTranslationList(NULL);
	}
}

void MENexusCreator::writeTreesBlock(MEPhylogeny *inPhylogeny)
{
	char *name;
	
	if (!treesBlockOpen) {
//		if (useTaxaBlock && !hasTaxaBlock) {
//			ThrowIfNull(inPhylogeny->getTaxonList());
//			writeTaxaBlock(inPhylogeny->getTaxonList());
//		}
			
		writer->writeString("Begin TREES;\r");
		treesBlockOpen = true;
	}

	writer->writeString("TREE\t");
	
	name=inPhylogeny->getName();
	if (name==NULL || strlen(name)==0)
		writer->writeString("phylogeny");
	else
		writer->writeLabel(name);
	
	writer->writeString(" = ");
	writePhylogeny(inPhylogeny);
	writer->writeChar('\r');
}

void MENexusCreator::writeTreesBlock(MEPhylogenyList *inPhylogenyList)
{
	int i, n = inPhylogenyList->getNumPhylogenies();
	MEPhylogeny *phylogeny;
	char *name;
	
	if (!treesBlockOpen) {
//		if (useTaxaBlock && !hasTaxaBlock) {
//			ThrowIfNull(inPhylogenyList->getTaxonList());
//			writeTaxaBlock(inPhylogenyList->getTaxonList());
//		}
		
		writer->writeString("Begin TREES;\r");
		treesBlockOpen = true;
	}

	for (i=0; i<n; i++) {
		phylogeny=inPhylogenyList->getPhylogeny(i);
		
		writer->writeString("TREE\t");
		
		name=phylogeny->getName();
		if (name==NULL || strlen(name)==0) {
			writer->writeString("phylogeny");
			writer->writeInteger(i+1);
		} else
			writer->writeLabel(name);
		
		writer->writeString(" = ");
		writePhylogeny(phylogeny);
		writer->writeChar('\r');
	}
}

void MENexusCreator::writeMatrixData(MEMatrix *inMatrix)
{
	int i, j, n, maxNameLen, nameLen, startCol, stopCol;
	char *name;
	
	writer->writeString("\tMatrix\r");

	n=inMatrix->getNumRows();
		
	maxNameLen=0;
	for (i=0; i<n; i++) {
		nameLen=strlen(inMatrix->getRowName(i));
		if (nameLen>maxNameLen)
			maxNameLen=nameLen;
	}
		
	for (i=0; i<n; i++) {
		name=inMatrix->getRowName(i);
		nameLen=strlen(name);
		writer->writeLabel(name);
		writer->writeSpaces(maxNameLen-nameLen+1);

		if (inMatrix->getHasLowerTriangle())
			startCol = 0;
		else if (inMatrix->getHasDiagonal())
			startCol = i;
		else // must have upper triangle
			startCol = i + 1;
			
		if (inMatrix->getHasUpperTriangle())
			stopCol = n;
		else if (inMatrix->getHasDiagonal())
			stopCol = i + 1;
		else // must have lower triangle
			stopCol = i;
			
		for (j = startCol; j < stopCol; j++) {
			writer->writeDouble(inMatrix->getElement(i, j));
			if (j < stopCol - 1)
				writer->writeChar(' ');
		}
		writer->writeChar('\r');
	}
	writer->writeString("\t;\r");
}

void MENexusCreator::writeDistancesBlock(MEDistanceMatrix *inMatrix)
{
	closeTreesBlock();

	writer->writeString("Begin DISTANCES;\r");
	
	writer->writeString("\tDimensions ntax=");
	writer->writeInteger(inMatrix->getNumRows());
	writer->writeString("\r\tFormat");
	
	if (inMatrix->getHasLowerTriangle() && inMatrix->getHasUpperTriangle())
		writer->writeString(" triangle=both");
	else if (inMatrix->getHasLowerTriangle())
		writer->writeString(" triangle=lower");
	else
		writer->writeString(" triangle=upper");
	
	writer->writeString(" labels");
	
	if (!inMatrix->getHasDiagonal())
		writer->writeString(" nodiagonal");
	
	writer->writeString(";\r");

	writeMatrixData(inMatrix);
	
	writer->writeString("End;\r\r");
}

void MENexusCreator::openHeader()
{
	if (!hasHeader) {
		writer->writeString("#NEXUS\r\r");
		hasHeader=true;
	}
}

void MENexusCreator::closeTreesBlock()
{
	if (treesBlockOpen) {
		writer->writeString("End;\r");
		treesBlockOpen = false;
		hasTranslationList=false;
	}
}

void MENexusCreator::createAlignment(MEBasicAlignment *inAlignment)
{
	ThrowIfNull(inAlignment);
	
	openHeader();
	if (useTaxaBlock) {
		writeTaxaBlock(inAlignment);
		writeCharactersBlock(inAlignment);
	} else 
		writeDataBlock(inAlignment);
}

//: This function does not actually close the trees block so subsequent
//  calls to "createTree" or "createTreeList" will add the trees to the
//  same block. Creating a different block will close the trees block.
void MENexusCreator::createTree(METree *inTree)
{
	ThrowIfNull(inTree);
	
	openHeader();
	
	if (useTaxaBlock && !hasTaxaBlock) {
		METaxonList *taxonList = inTree->getTaxonList();
		if (taxonList != NULL) 
			writeTaxaBlock(taxonList);
	}

	writeTreesBlock(inTree);
}

//: This function does not actually close the trees block so subsequent
//  calls to "createTree" or "createTreeList" will add the trees to the
//  same block. Creating a different block will close the trees block.
void MENexusCreator::createTreeList(METreeList *inTreeList)
{
	ThrowIfNull(inTreeList);
	ThrowIf(inTreeList->getNumTrees() < 1);
	
	openHeader();
	
	if (useTaxaBlock && !hasTaxaBlock) {
		METaxonList *taxonList = inTreeList->getTaxonList();
		if (taxonList != NULL) 
			writeTaxaBlock(taxonList);
	}

	writeTreesBlock(inTreeList);
}

//: This function does not actually close the trees block so subsequent
//  calls to "createPhylogeny" or "createPhylogenyList" will add the trees to the
//  same block. Creating a different block will close the trees block.
void MENexusCreator::createPhylogeny(MEPhylogeny *inPhylogeny)
{
	ThrowIfNull(inPhylogeny);
	
	openHeader();
	
//	if (useTaxaBlock && !hasTaxaBlock) {
//		METaxonList *taxonList = inPhylogeny->getTaxonList();
//		if (taxonList != NULL) 
//			writeTaxaBlock(taxonList);
//	}

	writeTreesBlock(inPhylogeny);
}

//: This function does not actually close the trees block so subsequent
//  calls to "createPhylogeny" or "createPhylogenyList" will add the trees to the
//  same block. Creating a different block will close the trees block.
void MENexusCreator::createPhylogenyList(MEPhylogenyList *inPhylogenyList)
{
	ThrowIfNull(inPhylogenyList);
	ThrowIf(inPhylogenyList->getNumPhylogenies() < 1);
	
	openHeader();
	
//	if (useTaxaBlock && !hasTaxaBlock) {
//		METaxonList *taxonList = inPhylogenyList->getTaxonList();
//		if (taxonList != NULL) 
//			writeTaxaBlock(taxonList);
//	}

	writeTreesBlock(inPhylogenyList);
}

void MENexusCreator::createDistanceMatrix(MEDistanceMatrix* inDistanceMatrix)
{
	ThrowIfNull(inDistanceMatrix);
	
	openHeader();
	if (useTaxaBlock) {
		writeTaxaBlock(inDistanceMatrix->getTaxonList());
		writeDistancesBlock(inDistanceMatrix);
	} else 
		writeDistancesBlock(inDistanceMatrix);
}

void MENexusCreator::finishCreation()
{
	hasTaxaBlock = false;
	closeTreesBlock();
}