/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MENexusParser.cpp

				MENexusParser Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "CReader.h"
#include "MEParser.h"
#include "MENexusParser.h"
#include "MEBasicSequence.h"
#include "MEBasicAlignment.h"
#include "MEDataMatrix.h"
#include "METaxonMatrixList.h"
#include "METreeList.h"
#include "METree.h"

// ---------------------------------------------------------------------------------
//		 MENexusParser
// ---------------------------------------------------------------------------------

MENexusParser::MENexusParser(CReader *inReader) : MEParser(inReader)
{
	reader->setCommentDelimiters('[', ']', 0xFF, '!');
	nextBlock=kUnknownBlock;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::findNextBlock
// ---------------------------------------------------------------------------------

// This function returns a value from the NexusBlocks enum to specify what the
// next block in the file is. The internal variable nextBlock is also set to this
// value. This should be overridden to provide support for other blocks. Once
// the block is read in, nextBlock is automatically set to kUnknownBlock by 
// findEndBlock. 

NexusBlocks MENexusParser::findNextBlock()
{
	char blockName[256];
	
	reader->findToken("BEGIN", 1);
	reader->readToken(blockName, 255, ";", true);
	
	return findBlockName(blockName);
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::findBlockName
// ---------------------------------------------------------------------------------

// This function returns a value from the NexusBlocks enum to specify what the
// block given by inBlockName is. 

NexusBlocks MENexusParser::findBlockName(char *inBlockName)
{
	if (strcmp(inBlockName, "TAXA")==0) {
		nextBlock=kTaxaBlock;
	} else if (strcmp(inBlockName, "CHARACTERS")==0) {
		nextBlock=kCharactersBlock;
	} else if (strcmp(inBlockName, "DATA")==0) {
		nextBlock=kDataBlock;
	} else if (strcmp(inBlockName, "UNALIGNED")==0) {
		nextBlock=kUnalignedBlock;
	} else if (strcmp(inBlockName, "DISTANCES")==0) {
		nextBlock=kDistancesBlock;
	} else if (strcmp(inBlockName, "SETS")==0) {
		nextBlock=kSetsBlock;
	} else if (strcmp(inBlockName, "CODONS")==0) {
		nextBlock=kCodonsBlock;
	} else if (strcmp(inBlockName, "TREES")==0) {
		nextBlock=kTreesBlock;
	}
	
	return nextBlock;
}



// ---------------------------------------------------------------------------------
//		 MENexusParser::findEndBlock
// ---------------------------------------------------------------------------------

void MENexusParser::findEndBlock()
{
	char token[256];
	
	try {
		do {
			reader->readToken(token, 255, ";", true);
		} while (strcmp(token, "END")!=0 && strcmp(token, "ENDBLOCK")!=0);
	} catch (CEndOfFile) { } // Doesn't matter if the End is missing

	nextBlock=kUnknownBlock;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readDataBlockHeader
// ---------------------------------------------------------------------------------

void MENexusParser::readDataBlockHeader(char *lookFor, int *nTaxa, int *nSites)
{
	char token[256], ch;
	bool dim, ttl, fmt;
	
	dim=ttl=fmt=false;
	
	do {
		reader->readToken(token, 255, true);
		if (strcmp(token, "TITLE")==0) {
			if (ttl) 
				throw MEDuplicateField("TITLE");
				
			ttl=true;
		} else if (strcmp(token, "DIMENSIONS")==0) {
			if (dim) 
				throw MEDuplicateField("DIMENSIONS");
				
			do {
				ch=reader->readToken(token, 255, "=;", true);
				if (strcmp(token, "NTAX")==0) {
					ch=reader->readInteger(nTaxa, ";");
				} else if (strcmp(token, "NCHAR")==0) {
					if (nSites==NULL)
						throw MEBadFormat("NCHAR field in TaxaBlock");
					ch=reader->readInteger(nSites, ";");
				}
			} while (ch!=';');
			dim=true;
		} else if (strcmp(token, "FORMAT")==0) {
			if (fmt) 
				throw MEDuplicateField("FORMAT");
				
			seqType=kUnknownSequence;
			do {
				ch=reader->readToken(token, 255, "=\"';", true);
				if (strcmp(token, "GAP")==0) {
					ch=reader->readToken(gap, MAX_CHARS, "\"';", false);
				} else if (strcmp(token, "MISSING")==0) {
					ch=reader->readToken(missing, MAX_CHARS, "\"';", false);
				} else if (strcmp(token, "MATCHCHAR")==0) {
					ch=reader->readToken(match, MAX_CHARS, "\"';", false);
				} else if (strcmp(token, "DATATYPE")==0) {
					ch=reader->readToken(token, 255, "\"';", true);
					if (strcmp(token, "NUCLEOTIDE")==0) {
						seqType=kDNASequence;
					} else if (strcmp(token, "DNA")==0) {
						seqType=kDNASequence;
					} else if (strcmp(token, "RNA")==0) {
						seqType=kRNASequence;
					} else if (strcmp(token, "PROTEIN")==0) {
						seqType=kAminoAcidSequence;
					} else if (strcmp(token, "CONTINUOUS")==0) {
						seqType=kContinuousSequence;
					}
				} else if (strcmp(token, "INTERLEAVE")==0) {
					isInterleaved=true;
				}
			} while (ch!=';');
			fmt=1;
		}
	} while (strcmp(token, lookFor)!=0);
	if (dim==0)
		throw MEMissingField("DIMENSIONS");
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readTaxaBlock
// ---------------------------------------------------------------------------------

void MENexusParser::readTaxaBlock(METaxonList *inTaxonList)
{
	int i, nTaxa; 
	char ch, name[MAX_NAME_LEN+1];
	METaxon *taxon;
	
	readDataBlockHeader("TAXLABELS", &nTaxa);

	if (nTaxa==0)
		throw MEMissingField("NTAXA");
		
	for (i=0; i<nTaxa; i++) {
		ch=reader->readToken(name, MAX_NAME_LEN, ";");
		NewClass(taxon, METaxon());
		taxon->setName(name);
		inTaxonList->addTaxon(taxon);
		
		if (ch==';' && i<nTaxa-1)
			throw METooFewTaxa();
	}
	
	findEndBlock();
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readSequenceData
// ---------------------------------------------------------------------------------

void MENexusParser::readSequenceData(MEBasicAlignment *inAlignment, METaxonList *inTaxonList, 
										int nTaxa, int nSites)
{
	int i, nRead, n, b, taxon;
	int numericalTaxa=-1;
	char token[256], *str, ch;
	MEBasicSequence* seq;
	
	if (isInterleaved) {
		bool firstRun=true;
		nRead=0;
		while (nRead<nSites) {
			n=-1;
			for (i=0; i<nTaxa; i++) {				
				taxon=reader->readTaxonLabel(token, 255, inAlignment->getNumTaxa());

				if (firstRun) {
					// Create new sequence using inAlignment's newSequence() function
					seq=inAlignment->newSequence(seqType);
					inAlignment->addSequence(seq);
					if (inTaxonList!=NULL) {
						if (taxon!=-1) {
							if (numericalTaxa==0)
								throw MEUnknownTaxon(token);
							numericalTaxa=1;
						} else {
							if (numericalTaxa==1)
								throw MEUnknownTaxon(token);
							numericalTaxa=0;
							if ((taxon=inTaxonList->findName(token))<0)
								throw MEUnknownTaxon(token);
						}
						seq->setName(inTaxonList->getTaxonName(i));
					} else
						seq->setName(token);
				} else {
					if (taxon!=-1) {
						if (numericalTaxa==0)
							throw MEUnknownTaxon(token);
					} else {
						if (numericalTaxa==1)
							throw MEUnknownTaxon(token);
						if ( (taxon=inAlignment->findName(token)) < 0)
							throw MEUnknownTaxon(token);
					}
					seq=inAlignment->getSequence(taxon);
				}
					
				ch=readSequenceUntil(token, 255, "", "\r\n;", &b);
				seq->appendSequence(token, strlen(token));
				
				if (ch==';' && i<nTaxa-1)
					METooFewTaxa();

				if (n==-1)
					n=b;
				if (n!=b)
					throw MEShortSequence(seq->getName());
			}
			firstRun=false;
			nRead+=n;
		}

	} else {
		for (i=0; i<nTaxa; i++) {
			taxon=reader->readTaxonLabel(token, 255, inAlignment->getNumTaxa());

			// Create new sequence using inAlignment's newSequence() function
			seq=inAlignment->newSequence(seqType);
			seq->setLength(nSites);
			inAlignment->addSequence(seq);
			if (inTaxonList!=NULL) {
				if (taxon!=-1) {
					if (numericalTaxa==0)
						throw MEUnknownTaxon(token);
					numericalTaxa=1;
				} else {
					if (numericalTaxa==1)
						throw MEUnknownTaxon(token);
					numericalTaxa=0;
					if ((taxon=inTaxonList->findName(token))<0)
						throw MEUnknownTaxon(token);
				}
				seq->setName(inTaxonList->getTaxonName(taxon));
			} else
				seq->setName(token);

			str=seq->getSequence();
			ch=readSequenceUntil(str, nSites, "", ";", &b);

			if (ch==';' && i<nTaxa-1)
				METooFewTaxa();
			if (b<nSites)
				throw MEShortSequence(seq->getName());
		}
	}
	
	if (inAlignment->getAlignmentType()==kUnknownAlignment) {
		if (seqType==kDNASequence || seqType==kRNASequence)
			inAlignment->setAlignmentType(kNucleotideAlignment);
		else if (seqType==kAminoAcidSequence)
			inAlignment->setAlignmentType(kAminoAcidAlignment);
	}
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readContinuousData
// ---------------------------------------------------------------------------------

void MENexusParser::readContinuousData(MEDataMatrix *inDataMatrix, METaxonList *inTaxonList, 
										int nTaxa, int nChars)
{
	int i, j, taxon;
	int numericalTaxa=-1;
	char token[256], ch;
	double value;
	
	inDataMatrix->setSize(nTaxa, nChars);
	
	for (i=0; i<nTaxa; i++) {
		taxon=reader->readTaxonLabel(token, 255, nTaxa);

		if (inTaxonList!=NULL) {
			if (taxon!=-1) {
				if (numericalTaxa==0)
					throw MEUnknownTaxon(token);
				numericalTaxa=1;
			} else {
				if (numericalTaxa==1)
					throw MEUnknownTaxon(token);
				numericalTaxa=0;
				if ((taxon=inTaxonList->findName(token))<0)
					throw MEUnknownTaxon(token);
			}
			inDataMatrix->setRowName(i, inTaxonList->getTaxonName(taxon));
		} else
			inDataMatrix->setRowName(i, token);

		for (j=0; j<nChars; j++) {
			try {
				ch=reader->readDouble(&value, ";");
				inDataMatrix->setElement(i, j, value);
			} catch (CIllegalDbl e) {
				inDataMatrix->setMissingElement(i, j);
			}
		}
		
		if (ch==';' && i<nTaxa-1)
			METooFewTaxa();
	}
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readCharactersBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readCharactersBlock(MEBasicAlignment *inAlignment, METaxonList *inTaxonList)
{
	int nSites, nTaxa;
	
	isInterleaved=false;
	nSites=0;
	nTaxa=inTaxonList->getNumTaxa();
	
	readDataBlockHeader("MATRIX", &nTaxa, &nSites);
		
	if (seqType!=kDNASequence && seqType!=kRNASequence && seqType!=kAminoAcidSequence)
		return false;
		
	if (nSites==0)
		throw MEMissingField("NCHARS");
	
	// The alignment should have the appropriate number of taxa in it.
	readSequenceData(inAlignment, inTaxonList, nTaxa, nSites);
	
	findEndBlock();
	
	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readCharactersBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readCharactersBlock(MEDataMatrix *inDataMatrix, METaxonList *inTaxonList)
{
	isInterleaved=false;
	int nChars=0;
	int nTaxa=inTaxonList->getNumTaxa();
	
	readDataBlockHeader("MATRIX", &nTaxa, &nChars);
		
	if (seqType!=kContinuousSequence && seqType!=kDiscreteSequence)
		return false;
		
	if (nChars==0)
		throw MEMissingField("NCHARS");
	
	readContinuousData(inDataMatrix, inTaxonList, nTaxa, nChars);
	
	findEndBlock();
	
	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readDataBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readDataBlock(MEBasicAlignment *inAlignment)
{
	int nTaxa, nSites;
	
	isInterleaved=false;
	nTaxa=0;
	nSites=0;
	readDataBlockHeader("MATRIX", &nTaxa, &nSites);

	if (seqType!=kDNASequence && seqType!=kRNASequence && seqType!=kAminoAcidSequence &&
		seqType!=kDiscreteSequence)
		return false;
		
	if (nTaxa==0)
		throw MEMissingField("NTAXA");
	if (nSites==0)
		throw MEMissingField("NCHARS");

	readSequenceData(inAlignment, NULL, nTaxa, nSites);

	findEndBlock();

	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readDataBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readDataBlock(MEDataMatrix *inDataMatrix)
{
	isInterleaved=false;
	int nTaxa=0;
	int nChars=0;
	readDataBlockHeader("MATRIX", &nTaxa, &nChars);

	if (seqType!=kContinuousSequence && seqType!=kDiscreteSequence)
		return false;
		
	if (nTaxa==0)
		throw MEMissingField("NTAXA");
	if (nChars==0)
		throw MEMissingField("NCHARS");

	readContinuousData(inDataMatrix, NULL, nTaxa, nChars);

	findEndBlock();

	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readUnalignedBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readUnalignedBlock(MEBasicAlignment* /*inAlignment*/, METaxonList* /*inTaxonList*/)
{
	findEndBlock();

	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readDistancesBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readDistancesBlock(METaxonMatrixList* /*inMatrixList*/, METaxonList* /*inTaxonList*/)
{
	findEndBlock();

	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readSetsBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readSetsBlock()
{
	findEndBlock();

	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readCodonsBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::readCodonsBlock(MECodonTable* /*inCodonTable*/)
{
	findEndBlock();

	return true;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readTreesBlock
// ---------------------------------------------------------------------------------

void MENexusParser::readTreesBlock(METree *inTree)
{
	METaxonList *translationList=NULL;
	METaxon *taxon;
	char token[256], ch;
	char name[MAX_NAME_LEN+1];
	int index;
	
	reader->readToken(token, 255, ";", true);
	if (strcmp(token, "TRANSLATE")==0) {
	// This reads in the translation table.
		NewClass(translationList, METaxonList());
		do {
			index=reader->readInteger();

			if (index<1)
				throw MEBadFormat("Translate index less than 1");
				
			ch=reader->readToken(token, 255, ",;", false);
			NewClass(taxon, METaxon());
			taxon->setName(token);
			translationList->setObject(index-1, taxon);
			if (ch!=',' && ch!=';')
				ch=reader->skipUntil(",;");
		} while (ch!=';');
		reader->readToken(token, 255, ";", true);
	}
	
	if (strcmp(token, "UTREE")!=0 && strcmp(token, "TREE")!=0)
		throw METreeMissing();
	
	reader->readToken(name, MAX_NAME_LEN, "=", false);

	if (*name=='*') // what is the star for? is it optional?
		reader->readToken(name, MAX_NAME_LEN, "=", false);
	
	inTree->setTranslationList(translationList);
	
	bool oldUseNumTips = inTree->getUseNumericalTips();
	if (translationList != NULL)
		inTree->setUseNumericalTips(true);
	else
		inTree->setUseNumericalTips(false);
	readNewickTree(inTree);
	inTree->setTranslationList(NULL);
	inTree->setName(name);

	inTree->setUseNumericalTips(oldUseNumTips);
	if (translationList!=NULL)
		delete translationList;

	findEndBlock();
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readTreesBlock
// ---------------------------------------------------------------------------------

void MENexusParser::readTreesBlock(METreeList *inTreeList)
{
	METaxonList *translationList=NULL;
	METree *tree;
	METaxon *taxon;
	char token[256], ch;
	char name[MAX_NAME_LEN+1];
	int index;
	
	reader->readToken(token, 255, ";", true);
	do {
		if (strcmp(token, "TRANSLATE")==0) {
		// This reads in the translation table.
			if (translationList==NULL)
				NewClass(translationList, METaxonList());
			else
				translationList->deleteAll();
				
			do {
				index=reader->readInteger();

				if (index<1)
					throw MEBadFormat("Translate index less than 1");
					
				ch=reader->readToken(token, 255, ",;", false);
				NewClass(taxon, METaxon());
				taxon->setName(token);
				translationList->setObject(index-1, taxon, true);
				if (ch!=',' && ch!=';')
					ch=reader->skipUntil(",;");
			} while (ch!=';');
			reader->readToken(token, 255, ";", true);
		}
	
		if (strcmp(token, "UTREE")!=0 && strcmp(token, "TREE")!=0)
			throw METreeMissing();
		
		reader->readToken(name, MAX_NAME_LEN, "=", false);

		if (*name=='*') // what is the star for? is it optional?
			reader->readToken(name, MAX_NAME_LEN, "=", false);
		
		tree=inTreeList->newTree();
		
		tree->setTranslationList(translationList);

		if (translationList != NULL)
			tree->setUseNumericalTips(true);
		else
			tree->setUseNumericalTips(false);

		readNewickTree(tree);
		tree->setTranslationList(NULL);
		tree->setName(name);
		inTreeList->addTree(tree);

		reader->readToken(token, 255, ";", true);
	} while (strcmp(token, "END")!=0 && strcmp(token, "ENDBLOCK")!=0);

	if (translationList!=NULL)
		delete translationList;
		
	nextBlock=kUnknownBlock;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readAlignment
// ---------------------------------------------------------------------------------

void MENexusParser::readAlignment(MEBasicAlignment* inAlignment)
{
	bool done=false, taxa=false;
	METaxonList *taxonList=NULL;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			if (done==false)
				throw MEMissingField("DATA block");
		}
		
		switch (nextBlock) {
			case kTaxaBlock:
				taxa=true;
				
				if (taxonList!=NULL)
					delete taxonList;
					
				NewClass(taxonList, METaxonList());
				readTaxaBlock(taxonList);
			break;
			case kCharactersBlock:
				if (!taxa)
					throw MEMissingField("TAXA block");
				
				if (readCharactersBlock(inAlignment, taxonList))
					done=true;
			break;
			case kDataBlock:
				if (readDataBlock(inAlignment))
					done=true;
			break;
		}
		if (taxonList!=NULL)
			delete taxonList;
	}
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::readTree
// ---------------------------------------------------------------------------------

void MENexusParser::readTree(METree *inTree)
{
	bool done=false;
	
	ThrowIfNull(inTree);
	
	done=false;
	while (!done) {
		try {
			nextBlock=findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("TREES block");
		}
		
		if (nextBlock==kTreesBlock) {
			readTreesBlock(inTree);
			done=true;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::readTreeList
// ---------------------------------------------------------------------------------

void MENexusParser::readTreeList(METreeList *inTreeList)
{
	bool done=false;
	
	ThrowIfNull(inTreeList);
	
	done=false;
	while (!done) {
		try {
			nextBlock=findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("TREES block");
		}
		
		if (nextBlock==kTreesBlock) {
			readTreesBlock(inTreeList);
			done=true;
		}
	}
}

// These functions can be used to read a NEXUS file, one block at a time. If 
// getNextBlock() has been called and is of the right type for this function,
// then the block is read and the function returns. Otherwise, the function
// searches for the next occurrance of the appropriate block. The passed data
// structure should be allocated and setup as desired. For example the MEBasicAlignment
// passed to parseCharactersBlock should already have the appropriate taxa previously
// read from a TAXA block. Likewise, the METreeList could have its taxon list set
// to a previously read MEBasicAlignment to provide taxon labels.

// ---------------------------------------------------------------------------------
//		 MENexusParser::getNextBlock
// ---------------------------------------------------------------------------------

// This function calls findNextBlock to get the next block type if nextBlock is
// set to kUnknownBlock. If not, then it just returns nextBlock. This means that
// getNextBlock can be called many times without reading further in the file unless
// findEndBlock has been called.

NexusBlocks MENexusParser::getNextBlock()
{
	if (nextBlock==kUnknownBlock) 
		return findNextBlock();
	else
		return nextBlock;
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::skipBlock
// ---------------------------------------------------------------------------------

// This function should be used after getNextBlock() to skip a block that is
// kUnknownBlock or is not supported in the application.

void MENexusParser::skipBlock()
{
	findEndBlock();
}


// ---------------------------------------------------------------------------------
//		 MENexusParser::parseTaxaBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseTaxaBlock(METaxonList* inTaxonList)
{
	if (nextBlock==kTaxaBlock) {
		readTaxaBlock(inTaxonList);
		return true;
	}

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("TAXA block");
		}
		
		if (nextBlock==kTaxaBlock) {
			readTaxaBlock(inTaxonList);
			done=true;
		}
	}
	
	inTaxonList->setTaxaDefined(true);
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseCharactersBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseCharactersBlock(MEBasicAlignment* inAlignment, METaxonList* inTaxonList)
{
	if (nextBlock==kCharactersBlock)
		return readCharactersBlock(inAlignment, inTaxonList);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("CHARACTERS block");
		}
		
		if (nextBlock==kCharactersBlock) {
			if (readCharactersBlock(inAlignment, inTaxonList))
				done=true;
		}
	}
	
	inAlignment->setTaxaDefined(true);
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseCharactersBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseCharactersBlock(MEDataMatrix* inDataMatrix, METaxonList* inTaxonList)
{
	if (nextBlock==kCharactersBlock)
		return readCharactersBlock(inDataMatrix, inTaxonList);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("CHARACTERS block");
		}
		
		if (nextBlock==kCharactersBlock) {
			if (readCharactersBlock(inDataMatrix, inTaxonList))
				done=true;
		}
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseDataBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseDataBlock(MEBasicAlignment* inAlignment)
{
	if (nextBlock==kDataBlock)
		return readDataBlock(inAlignment);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("DATA block");
		}
		
		if (nextBlock==kDataBlock) {
			if (readDataBlock(inAlignment))
				done=true;
		}
	}

	inAlignment->setTaxaDefined(true);

	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseDataBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseDataBlock(MEDataMatrix* inDataMatrix)
{
	if (nextBlock==kDataBlock)
		return readDataBlock(inDataMatrix);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("DATA block");
		}
		
		if (nextBlock==kDataBlock) {
			if (readDataBlock(inDataMatrix))
				done=true;
		}
	}

	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseUnalignedBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseUnalignedBlock(MEBasicAlignment* inAlignment, METaxonList* inTaxonList)
{
	if (nextBlock==kUnalignedBlock)
		return readUnalignedBlock(inAlignment, inTaxonList);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("UNALIGNED block");
		}
		
		if (nextBlock==kUnalignedBlock) {
			if (readUnalignedBlock(inAlignment, inTaxonList))
				done=true;
		}
	}
	
	inAlignment->setTaxaDefined(true);

	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseDistancesBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseDistancesBlock(METaxonMatrixList* inMatrixList, METaxonList* inTaxonList)
{
	if (nextBlock==kDistancesBlock)
		return readDistancesBlock(inMatrixList, inTaxonList);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("DISTANCES block");
		}
		
		if (nextBlock==kDistancesBlock) {
			if (readDistancesBlock(inMatrixList, inTaxonList))
				done=true;
		}
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseSetsBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseSetsBlock()
{
	if (nextBlock==kSetsBlock)
		return readSetsBlock();

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("SETS block");
		}
		
		if (nextBlock==kSetsBlock) {
			if (readSetsBlock())
				done=true;
		}
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseCodonsBlock
// ---------------------------------------------------------------------------------

bool MENexusParser::parseCodonsBlock(MECodonTable* inCodonTable)
{
	if (nextBlock==kCodonsBlock)
		return readCodonsBlock(inCodonTable);

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("CODONS block");
		}
		
		if (nextBlock==kCodonsBlock) {
			if (readCodonsBlock(inCodonTable))
				done=true;
		}
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseTreesBlock
// ---------------------------------------------------------------------------------

// This reads a single tree from the next tree block.

bool MENexusParser::parseTreesBlock(METree *inTree)
{
	if (nextBlock==kTreesBlock) {
		readTreesBlock(inTree);
		return true;
	}

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("TREES block");
		}
		
		if (nextBlock==kTreesBlock) {
			readTreesBlock(inTree);
			done=true;
		}
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MENexusParser::parseTreesBlock
// ---------------------------------------------------------------------------------

// This reads a set of trees from the next tree block.

bool MENexusParser::parseTreesBlock(METreeList *inTreeList)
{
	if (nextBlock==kTreesBlock) {
		readTreesBlock(inTreeList);
		return true;
	}

	bool done=false;
	
	while (!done) {
		try {
			findNextBlock();
		} catch (CEndOfFile) {
			throw MEMissingField("TREES block");
		}
		
		if (nextBlock==kTreesBlock) {
			readTreesBlock(inTreeList);
			done=true;
		}
	}
	
	return true;
}

