/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEObject.cpp

 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = Global

//	The base for all data objects
//	MEObject provides a common base to all the data containing objects in
//	the MolEvol Library: MEList, METaxonList, METaxon, MESequence etc.
//	This allows them all to be included in lists (MEList) and to share
//	a reading and writing mechanism. MEObject and thus all its derived
//	classes have the name and desc fields which can be null if
//	not required. The owner field points to the object that owns this
//	one. When an object is deleted, it will generally delete all the 
//	objects it owns.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEObject.h"

#include "MEStream.h"

#pragma mark === MEObject ===

// ---------------------------------------------------------------------------------
//		MEObject::MEObject
// ---------------------------------------------------------------------------------

//: the constructor
//  Constructs an MEObject with no name, description or owner
MEObject::MEObject()
{
	owner = NULL;
	name = NULL;
	desc = NULL;
	
	flags = 0;

	numReferences = 1;
}

// ---------------------------------------------------------------------------------
//		MEObject::~MEObject
// ---------------------------------------------------------------------------------

//: The destructor
//  Deletes the name and description if present
MEObject::~MEObject()
{
	if (name != NULL)
		delete[] name;
	if (desc != NULL)
		delete[] desc;
}

// ---------------------------------------------------------------------------------
//		MEObject::clone
// ---------------------------------------------------------------------------------

//: Clones this object from an existing one
//  Copies all the data members from the src MEObject to this one.

void MEObject::clone(MEObject *src)
{
	setOwner(src->getOwner());
	setName(src->getName());
	setDescription(src->getDescription());
	setFlags(src->getFlags());
}

// ---------------------------------------------------------------------------------
//		MEObject::addReference
// ---------------------------------------------------------------------------------

//: Adds a reference to this object
//  Increments the references variable by 1. This is a mechanism by which other
//  objects can register that they are using an object. When they are finished
//  they should call removeReference. If ther references variable hits 0 (when
//  removeReferences is called with inAutoDelete = true) the object deletes itself.

void MEObject::addReference()
{
	numReferences++;
}

// ---------------------------------------------------------------------------------
//		MEObject::removeReference
// ---------------------------------------------------------------------------------

//: Removes a reference to this object
//  Decrements the references variable by 1. This is a mechanism by which other
//  objects can register that they are using an object. When they are finished
//  they should call removeReference. If ther references variable hits 0 (when
//  removeReferences is called with inAutoDelete = true) the object deletes itself.

void MEObject::removeReference(bool inAutoDelete)
{
	numReferences--;
	if (inAutoDelete && numReferences < 1)
		delete this;
}

// ---------------------------------------------------------------------------------
//		MEObject::setName
// ---------------------------------------------------------------------------------

//: Sets the name field of this object
//  Allocates memory if needed. An inName which is null will set the name to 
//  null without deleting the array.
void MEObject::setName(char *inName)
{
	if (inName != NULL) {	
		if (name == NULL)
			NewArray(name, char, MAX_NAME_LEN+1);
			
		strncpy(name, inName, MAX_NAME_LEN);
		name[MAX_NAME_LEN]='\0';
	} else 
		name = NULL;
}

// ---------------------------------------------------------------------------------
//		MEObject::getName
// ---------------------------------------------------------------------------------
		
//: Returns the name of this object
//  If null, then this object has no name.
char* MEObject::getName()
{
	return name;
}

// ---------------------------------------------------------------------------------
//		MEObject::getName
// ---------------------------------------------------------------------------------
		
//: Copies the name of this object into a string
//  Writes the name of this object into inName which should contain enough space.
void MEObject::getName(char *inName)
{
	if (name != NULL)
		strcpy(inName, name);
	else
		inName[0] = '\0';
}

// ---------------------------------------------------------------------------------
//		MEObject::setDescription
// ---------------------------------------------------------------------------------

//: Sets the description field of this object
//  Allocates memory if needed. An inDescription which is null will set the 
//  description to null without deleting the array.
void MEObject::setDescription(char *inDescription)
{
	if (inDescription != NULL) {	
		if (desc == NULL)
			NewArray(desc, char, MAX_DESC_LEN+1);
			
		strncpy(desc, inDescription, MAX_DESC_LEN);
		desc[MAX_DESC_LEN]='\0';
	} else 
		desc=NULL;
}

// ---------------------------------------------------------------------------------
//		MEObject::getDescription
// ---------------------------------------------------------------------------------
		
//: Returns the description of this object
//  If null, then this object has no description.
char* MEObject::getDescription()
{
	return desc;
}

// ---------------------------------------------------------------------------------
//		MEObject::getDescription
// ---------------------------------------------------------------------------------
		
//: Copies the description of this object into a string
//  Writes the description of this object into inDescription which should contain 
//  enough space.
void MEObject::getDescription(char *inDescription)
{
	if (desc != NULL)
		strcpy(inDescription, desc);
	else
		inDescription[0] = '\0';
}

// ---------------------------------------------------------------------------------
//		MEObject::setFlag
// ---------------------------------------------------------------------------------

//: Sets one or more bits of the flag field
//  This uses inFlag to mask the bits and inState gives the state (true or false)
//  to which they should be set.
void MEObject::setFlag(short inFlag, bool inState)
{
	if (inState) 
		flags |= inFlag;
	else
		flags &= ~inFlag; 
}

// ---------------------------------------------------------------------------------
//		MEObject::getFlag
// ---------------------------------------------------------------------------------

//: Gets the boolean state of one or more bits of the flag
//  This uses inFlag to mask the bits and returns their state (true or false).
//  If more that one bit is specified then this function returns true if either
//  of the bits are set.
bool MEObject::getFlag(short inFlag)
{
	return flags & inFlag;
}

// ---------------------------------------------------------------------------------
//		MEObject::readObject
// ---------------------------------------------------------------------------------

//: Reads the fields of this object from an MEStream
void MEObject::readObject(MEStream *stream)
{
	char inName[MAX_NAME_LEN+1];
	char inDesc[MAX_DESC_LEN+1];
	
	owner=stream->get("Owner");
		
	if (stream->get("Name", inName, MAX_NAME_LEN))
		setName(inName);
	if (stream->get("Description", inDesc, MAX_DESC_LEN))
		setDescription(inDesc);
		
	int inFlags;
	if (stream->get("Flags", &inFlags))
		flags = inFlags;
}

// ---------------------------------------------------------------------------------
//		MEObject::readData
// ---------------------------------------------------------------------------------

//: Reads the data of this object from an MEStream
//  The data of an object differs from the fields of an object in that it can be 
//  written in an efficient or more flexible form. The object must provide the 
//  ability to read and write this data.
void MEObject::readData(MEStream * /*stream*/)
{ 
}

// ---------------------------------------------------------------------------------
//		MEObject::writeObject
// ---------------------------------------------------------------------------------

//: Writes the fields of this object to an MEStream
void MEObject::writeObject(MEStream *stream)
{
	stream->put("Owner", owner);
	stream->put("Name", name);
	stream->put("Description", desc);
	stream->put("Flags", flags);
}

// ---------------------------------------------------------------------------------
//		MEObject::writeData
// ---------------------------------------------------------------------------------

//: Writes the data of this object to an MEStream
//  The data of an object differs from the fields of an object in that it can be 
//  written in an efficient or more flexible form. The object must provide the 
//  ability to read and write this data.
void MEObject::writeData(MEStream * /*stream*/)
{ 
}


