/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEObject.h

		Interface for the MEObject Class

 ******************************************************************************/
//! author = "Andrew Rambaut"
//! library = MolEvol
//! group = Global
 
#ifndef _MEOBJECT_H_
#define _MEOBJECT_H_

#include "MEGlobal.h"

class MEStream;

/******************************************************************************
	class MEObject
 ******************************************************************************/
#pragma mark MEObject class

//: The base for all data objects
//	MEObject provides a common base to all the data containing objects in
//	the MolEvol Library: MEList, METaxonList, METaxon, MESequence etc.
//	This allows them all to be included in lists (MEList) and to share
//	a reading and writing mechanism. MEObject and thus all its derived
//	classes have the name and desc fields which can be null if
//	not required. The owner field points to the object that owns this
//	one. When an object is deleted, it will generally delete all the 
//	objects it owns.

class MEObject {
	public:
		virtual long		getID() const { return 'MObj'; }
		//: Returns a unique ID for this class - 'MObj'
		//  An ID is a four character code stored as a long integer.
		//	Each object descended from MEObject has an unique code.
	
		virtual char*		getClassName() const { return "Object"; }
		//: Returns the name of this class - "Object"
		//  A class name is character string which gives a unique name for
		//	MEObject and its descendants.
	
	protected:
		MEObject*			owner;			//: A pointer to the owner of this object
		char				*name; 			//: The name of this object
		char				*desc;			//: The description of this object
		
		short				flags;			//: Flags for this object (kHidden etc.)

		short				numReferences;	//: Number of objects referencing this one

	public:
	 						MEObject();		//: The constructor
	 	virtual				~MEObject();	//: The destructor

		virtual void 		addReference();
		//: Adds a reference to this object
		
		virtual void 		clone(MEObject *src);
		//: Clones this object from an existing one

		virtual void 		removeReference(bool inAutoDelete = false);
		//: Removes a reference to this object

		virtual short 		getNumReference() { return numReferences; }
		//: Returns the number of references to this object

		virtual void 		setOwner(MEObject *inOwner) { owner=inOwner; }
		//: Sets the owner of this object
		//  If inOwner is null, then this object has no owner.

		virtual MEObject*	getOwner() { return owner; }
		//: Returns the owner of this object
		//  If null, then this object has no owner.
		
		virtual void 		setName(char *inName);
		//: Sets the name field of this object

		virtual char*		getName();
		//: Returns a pointer to the name of this object

		virtual void		getName(char *inName);
		//: Copies the name of this object into a string

		virtual void 		setDescription(char *inDescription);
		//: Sets the description field of this object
		
		virtual char*		getDescription();
		//: Returns the description of this object

		virtual void		getDescription(char *inDescription);
		//: Copies the description of this object into a string

		virtual void 		setFlags(short inFlags) { flags = inFlags; }
		//: Sets the entire flag field
		
		virtual short		getFlags() { return flags; }
		//: Returns the entire flag field
		
		virtual void 		setFlag(short inFlag, bool inState);
		//: Sets one or more bits of the flag field
		
		virtual bool		getFlag(short inFlag);
		//: Gets the boolean state of one or more bits of the flag

		virtual void		readObject(MEStream *stream);
		//: Reads the fields of this object from an MEStream
		
		virtual void		readData(MEStream *stream);
		//: Reads the data of this object from an MEStream
		
		virtual void		writeObject(MEStream *stream);
		//: Writes the fields of this object to an MEStream

		virtual void		writeData(MEStream *stream);
		//: Writes the data of this object to an MEStream
};

#endif /* _MEOBJECT_H_ */
