/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEOptimiser.h

		Interface for the MEOptimiser Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Trees"

#pragma mark // includes:
// ANSI / C++
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
// Mol. Evol.
#include "MEExceptions.h"
#include "MEGlobal.h"
#include "MEStream.h"

#include "METree.h"
#include "MECriterion.h"

#include "CProgress.h"

#include "MEOptimiser.h"


#pragma mark // methods:


// ---------------------------------------------------------------------------------
//		 MEOptimiser::MEOptimiser
// ---------------------------------------------------------------------------------

MEOptimiser::MEOptimiser()
{
	mCriterion = NULL;
	
	progress = NULL;
	
	numParameters = 0;

	kMaxBrentIterations = 200;
	kBrentTolerance = 1.0E-6;
	kIterativeTolerance = 1.0E-6;
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::~MEOptimiser
// ---------------------------------------------------------------------------------

MEOptimiser::~MEOptimiser()
{
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::setCriterion
// ---------------------------------------------------------------------------------

void MEOptimiser::setCriterion(MECriterion *inCriterion)
{
	mCriterion = inCriterion;
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::optimise
// ---------------------------------------------------------------------------------

double MEOptimiser::optimise()
{
	ThrowIfNull(mCriterion);
	
	evaluateCount=0;
	lineCount=0;
	
	numParameters = mCriterion->getNumParameters();
	
	mOptimum = findOptimum();
		
	return mOptimum;
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::findOptimum
// ---------------------------------------------------------------------------------

double MEOptimiser::findOptimum()
{
	int i, j, estRem=99999, newRem;
	double score, lastScore, diff, oldDiff;
						
	if (progress != NULL) {
		progress->description("Optimising Tree...");
	}

	lastScore=kBigReal;
	bool stop = false;;
	int n=0;
	do {
		for (i=0; i<numParameters; i++) {
			score = brent(i, mCriterion->getValue(i), mCriterion->getMinValue(i), mCriterion->getMaxValue(i));
			lineCount++;
			
			if (progress != NULL)
				progress->score(score);
		}
				
		oldDiff=diff;	
					
		diff = lastScore - score;
		if (diff < kIterativeTolerance)
			stop = true;
		lastScore = score;
		
		n++;
	
		if (n>2) {
			if ((newRem=estimateRemaining(oldDiff, diff))!=-1)
				estRem=newRem;
				
			if (progress != NULL) {
				if (progress->progress(n, n+estRem))
					throw MEInterrupted();
			}
		}
	} while (!stop);
	
	if (progress!=NULL)
		progress->progress(n, n);
	
	return score;	
}
		
// ---------------------------------------------------------------------------------
//		 MEOptimiser::estimateRemaining
// ---------------------------------------------------------------------------------

int MEOptimiser::estimateRemaining(double oldDiff, double newDiff)
{
	double ratio, diff;
	int i=0;
	
	if (newDiff>=oldDiff)
		return -1;
		
	ratio=newDiff/oldDiff;
	diff=newDiff;
	
	while (diff>kIterativeTolerance) {
		diff*=ratio;
		i++;
	}
	
	return i;
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::brent
// ---------------------------------------------------------------------------------

const double CGOLD = 0.3819660;
const double ZEPS = 1.0e-10;

double MEOptimiser::brent(int inIndex, double inValue, double inMinValue, double inMaxValue)
{
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double e=0.0;

	a=inMinValue;
	b=inMaxValue;
	x=w=v=inValue;	
	fw=fv=fx=evaluate(inIndex, x);
	
	for (iter=1; iter<=kMaxBrentIterations; iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=kBrentTolerance*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			mCriterion->setValue(inIndex, x);
			return fx;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) 
				p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=sign(tol1, xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		
		u=(fabs(d) >= tol1 ? x+d : x+sign(tol1, d));
		fu=evaluate(inIndex, u);
		
		if (fu <= fx) {
			if (u >= x)
				a=x; 
			else 
				b=x;
				
			shift(v,w,x,u);
			shift(fv,fw,fx,fu);
		} else {
			if (u < x)
				a=u; 
			else
				b=u;
				
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	mCriterion->setValue(inIndex, x);

	throw MEException("Too many iterations in MEOptimiser::brent");
	
	return fx;
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::evaluate
// ---------------------------------------------------------------------------------

double MEOptimiser::evaluate(int inIndex, double inValue)
{
	if (progress != NULL)
		progress->yield();
		
	evaluateCount++;
	mCriterion->setValue(inIndex, inValue);
	
	return mCriterion->getScore();
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::readObject
// ---------------------------------------------------------------------------------

// Read the fields of this object from stream.
void MEOptimiser::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		 MEOptimiser::writeObject
// ---------------------------------------------------------------------------------

// Write the fields of this object to stream.
void MEOptimiser::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}


