/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPagParser.cpp

				MEPagParser Class
	
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "CReader.h"
#include "MEParser.h"
#include "MEPagParser.h"
#include "METaxon.h"
#include "METaxonList.h"
#include "MELink.h"
#include "METree.h"
#include "METreeList.h"
#include "MEDataMatrix.h"
#include "MEDataMatrixList.h"


// ---------------------------------------------------------------------------------
//		 MEPagParser
// ---------------------------------------------------------------------------------

MEPagParser::MEPagParser(CReader *inReader) : MEParser(inReader)
{
	reader->setCommentDelimiters('\0', '\0', '\0');
}

// ---------------------------------------------------------------------------------
//		 MEPagParser::readTree
// ---------------------------------------------------------------------------------

void MEPagParser::readTree(METree *inTree)
{
	readDataAndTree(NULL, inTree);
}

// ---------------------------------------------------------------------------------
//		 MEPagParser::readTreeList
// ---------------------------------------------------------------------------------

void MEPagParser::readTreeList(METreeList *inTreeList)
{
	METree *tree=inTreeList->newTree();
	readDataAndTree(NULL, tree);
	inTreeList->addTree(tree);
}

// ---------------------------------------------------------------------------------
//		 MEPagParser::readDataAndTree
// ---------------------------------------------------------------------------------

void MEPagParser::readDataAndTree(MEDataMatrix *inMatrix, METree *inTree)
{
	int i, j;
	char tok[16];
	reader->markPosition();
	reader->readToken(tok, 15);
	if (tok[0]!='#')
		reader->restorePosition();
		
	int nTaxa=reader->readInteger();
	int nTraits=reader->readInteger();
	
	if (inMatrix!=NULL) {
		inMatrix->setSize(0, 0);
		for (j=0; j<nTraits; j++)
			inMatrix->addColumn();

		if (inMatrix->getTaxonList()==NULL) {
			METaxonList *taxonList;
			NewClass(taxonList, METaxonList());		
			inMatrix->setTaxonList(taxonList, true);
		}		
	}
		
	if (inTree->getTaxonList()==NULL) {
		METaxonList *taxonList;
		NewClass(taxonList, METaxonList());		
		inTree->setTaxonList(taxonList, true);
	}
		
	int nodeNo;
	double f, branchLen;
	MELink *link0, *link1, *link2;
	char name[256], str[256];
	
	METaxonList *nodeLabels;
	METaxon *label;
	NewClass(nodeLabels, METaxonList());
	
	for (i=0; i<nTaxa; i++) {
		reader->readToken(name, 255, ",");
		
		reader->readToken(str, 255, ",");

		nodeNo = nodeLabels->findName(str);
		if (nodeNo < 0) {
			NewClass(label, METaxon());
			label->setName(str);
			label->setFlag(kSelected, true);
			nodeNo = nodeLabels->addTaxon(label);
		}
		
		reader->readDouble(&branchLen, ",");
		
		if (inMatrix!=NULL)
			inMatrix->addRow(name);
			
		for (j=0; j<nTraits; j++) {			
			reader->readToken(str, 255, ",");
				
			if (inMatrix!=NULL) {
				if (sscanf(str, "%lf", &f)==1)
					inMatrix->setElement(i, j, f);
				else
					inMatrix->setMissingElement(i, j);
			}
		}
		
		link0=inTree->newTip(name);

		try {
			link1=inTree->getInternalNode(nodeNo);
		} catch (MENodeNotFound e) {
			link1=inTree->newNode(nodeNo);
		}
			
		link2=inTree->newBranch(link1, branchLen);
		
		link0->branchNo=link2->branchNo;
		link0->branch=link2;
		link2->branch=link0;
	}

	int nodeNo1, nodeNo2;
	bool done=false;
	try {
		while (!done) {
			reader->readToken(str, 255, ",");

			nodeNo1 = nodeLabels->findName(str);
			if (nodeNo1 < 0) {
				NewClass(label, METaxon());
				label->setName(str);
				nodeNo1 = nodeLabels->addTaxon(label);
			}
			// can't be the root
			nodeLabels->setFlag(nodeNo1, kSelected, false);
			
			reader->readToken(str, 255, ",");

			nodeNo2 = nodeLabels->findName(str);
			if (nodeNo2 < 0) {
				NewClass(label, METaxon());
				label->setFlag(kSelected, true);
				label->setName(str);
				nodeNo2 = nodeLabels->addTaxon(label);
			}

			reader->readDouble(&branchLen, ",");
			
			try {
				link0=inTree->getInternalNode(nodeNo1);
			} catch (MENodeNotFound e) {
				link0=inTree->newNode(nodeNo1);
			}

			try {
				link1=inTree->getInternalNode(nodeNo2);
			} catch (MENodeNotFound e) {
				link1=inTree->newNode(nodeNo2);
			}
			
			link2=inTree->newBranch(link1, branchLen);
		
			link0->branchNo=link2->branchNo;
			link0->branch=link2;
			link2->branch=link0;
		}
	} catch (CEndOfFile e) {
	}
	
	inTree->setHasLengths(true);
	
	if (nodeLabels->getNumWithFlag(kSelected, true) != 1)
		throw MEParserError("Could not find a root node. Please check your tree");
		
	nodeNo = nodeLabels->findIndex(nodeLabels->getFirstWithFlag(kSelected, true));
	
	delete nodeLabels;
	
	MELink *link=inTree->getInternalNode(nodeNo);
	
	// prune out the temporary MELink that pointed to the root
	link1=link;
	while (link->next!=link1) 
		link=link->next;
		
	link->next=link1->next;
	inTree->removeLink(link1);
	
	// move on to the first branch in the node
	link=link->next;
	
	if (link->getNumBranches()<3) {
		// if the root node has less than 3 branches then this node is
		// removed. This makes the tree unrooted but we store the position
		// of the root by setting METree's root1 and root2 to point to the
		// nodes either side of the root.
		inTree->setIsRooted(true);
		
		// Set the tree's root pointers
		MELink* root1=link->branch;
		MELink* root2=link->next->branch;
		inTree->setRoot1(root1);
		inTree->setRoot2(root2);
		
		// Set the tree's root branch lengths
		inTree->setRootDistances(inTree->getBranchLength(root1->branchNo),
									inTree->getBranchLength(root2->branchNo));
		
		// Set the nodes either side of the root to point to each other
		root1->branch=root2;
		root2->branch=root1;
		
		// Remove root2's branch length from the tree's array
		inTree->removeBranchLength(root2->branchNo);
		
		// Set the new branches branchNo and the corresponding branchLength
		// This just uses root1's old branch length slot and sets it to the
		// sum of the two old branch lengths
		root2->branchNo=root1->branchNo;
		inTree->setBranchLength(root1->branchNo, 
								inTree->getRootDistance1() + inTree->getRootDistance2());
		
		inTree->removeLink(link->next);
		inTree->removeLink(link);
		
		inTree->setNumInternalNodes(inTree->getNumInternalNodes()-1);
		inTree->setNumNodes(inTree->getNumNodes()-1);
	} else {
		inTree->setIsRooted(false);
		inTree->setRoot1(link);
		inTree->setRoot2(link->branch);
	}
	
	// Check that all tips lead up to the root... If they don't
	// then the tree wasn't fully specified.
	int count;
	link0=inTree->getFirstTip();
		
	while (link0!=NULL) {
		link=link0->branch;
		link1=link0;
		while (link!=NULL && link1!=inTree->getRoot1() && link1!=inTree->getRoot2()) {
			link1=link->getStemLink();
			link=link1->branch;
		}
		if (link==NULL) 
			throw METreeNotConnected();
			
		link0=inTree->getNextTip(link0);	
	}
}

// ---------------------------------------------------------------------------------
//		 MEPagParser::readDataAndTreeList
// ---------------------------------------------------------------------------------

void MEPagParser::readDataAndTreeList(MEDataMatrixList *inDataMatrices, METreeList *inTreeList)
{
	MEDataMatrix *matrix=(MEDataMatrix *)inDataMatrices->newMatrix();
	METree *tree=inTreeList->newTree();

	readDataAndTree(matrix, tree);
	inDataMatrices->add(matrix, true);
	inTreeList->addTree(tree);
}

