/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEParser.cpp

				MEParser Class
	
 ******************************************************************************/

// MEParser provides a common base to all classes reading in sequences or
// trees from standard file formats.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "CReader.h"
#include "MEParser.h"
#include "MEPhylipParser.h"
#include "MENexusParser.h"
#include "MEFlatParser.h"
#include "MENBRFParser.h"
#include "MEFASTAParser.h"
//#include "MEClustalParser.h"
#include "MEGCGParser.h"
#include "MEGenbankParser.h"
#include "MEPagParser.h"
#include "MECAICParser.h"
#include "METabParser.h"

#include "METaxonList.h"
#include "MEBasicSequence.h"
#include "MEBasicAlignment.h"
#include "MELink.h"
#include "METree.h"
#include "METreeList.h"
#include "MEDataMatrix.h"
#include "MELabelArray.h"


// The names of the format types
const char* kFormatTypeNames[numFormatTypes] = 
	{ "Unknown", "Nexus", "XML", "Phylip", "MolPhy", "Mega", "GDE Flat",
	  "NBRF/PIR", "FASTA", "Clustal", "GCG", "GenBank", "Pag", "CAIC", 
	  "Compare", "Tab", "Tree", "PAML Tree", "Pretty" }; 

// Whether the format is currently supported for input
const bool kInputFormatSupported[numFormatTypes] = {
	false,	//	kFormatUnknown
	true,	//	kFormatNEXUS
	false, 	//	kFormatXML
	true, 	//	kFormatPHYLIP
	true, 	//	kFormatMOLPHY
	false, 	//	kFormatMEGA
	true, 	//	kFormatFLAT
	true,	//	kFormatNBRF
	true, 	//	kFormatFASTA
	false, 	//	kFormatCLUSTAL
	true, 	//	kFormatGCG
	true, 	//	kFormatGENBANK
	true, 	//	kFormatPAG
	true, 	//	kFormatCAIC
	true, 	//	kFormatCOMPARE
	true, 	//	kFormatTAB
	true, 	//	kFormatTree
	false, 	//	kFormatPAMLTree
	false 	//	kFormatPretty
}; 

enum {
	uncertain,
	sequential,
	interleaved
};

#pragma mark === MEParser ===

// ---------------------------------------------------------------------------------
//		 MEParser
// ---------------------------------------------------------------------------------

MEParser::MEParser(CReader *inReader)
{
	reader=inReader;
	ThrowIfNull(reader);
	
	seqType=kUnknownSequence;

	strcpy(gap, "-");
	strcpy(missing, "?");
	strcpy(match, ".");

	isInterleaved=false;
}

MEParser::~MEParser()
{
}


// ---------------------------------------------------------------------------------
//		 MEParser::detectFormat
// ---------------------------------------------------------------------------------

// A static function that will attempt to detect the file format of inStream.
// It then resets inStream back ready for reading. Should be called thus:
// 
// format=MEParser::DetectFormat(aStream);

FormatType MEParser::detectFormat(CReader *inReader)
{
	char st[256];
	int i, j;
	FormatType fileFormat;

	fileFormat=kFormatUnknown;	

	ThrowIfNull(inReader);
	
	inReader->markPosition();
	inReader->skipSpace();
	inReader->readLine(st, 255, true, false);

	if ( *st=='#' ) { 	// either MEGA, NEXUS or FLAT format
		if ( strncmp(st, "#MEGA", 5)==0 )
			fileFormat=kFormatMEGA;
		else if ( strncmp(st, "#NEXUS", 6)==0 )
			fileFormat=kFormatNEXUS;
		else if ( strncmp(st, "#PAG", 4)==0 )
			fileFormat=kFormatPAG;
		else if ( strncmp(st, "#CAIC", 5)==0 )
			fileFormat=kFormatCAIC;
		else if ( strncmp(st, "#COMPARE", 8)==0 )
			fileFormat=kFormatCOMPARE;
		else
			fileFormat=kFormatFLAT;
	} else if ( strncmp(st, "//", 2)==0 || strncmp(st, "LOCUS", 5)==0 ) { 	// GENBANK format
		fileFormat=kFormatGENBANK;
	} else if ( strncmp(st, "!!RICH_SEQUENCE", 15)==0 ) { 	// GCG rich sequence format
		fileFormat=kFormatGCG;
	} else if ( *st=='>' ) { 		// NBRF format or PIR?
		if (st[3]==';')
			fileFormat=kFormatNBRF;
		else
			fileFormat=kFormatFASTA;
	} else if ( sscanf(st, " %d %d ", &i, &j)==2 ) {
		if (strchr(st, '#')!=NULL)	// MOLPHY format
			fileFormat=kFormatMOLPHY;
		else 						// PHYLIP format
			fileFormat=kFormatPHYLIP;
	} else if ( sscanf(st, " %d ", &i)==1 || *st=='(' ) { 	// Simple tree file
		fileFormat=kFormatTree;
	} else if ( strchr(st, '\t') ) {// Tab-delineated format?
		fileFormat=kFormatTAB;
	}
	
	inReader->restorePosition();

	return fileFormat;
}

// ---------------------------------------------------------------------------------
//		 MEParser::createParserByType
// ---------------------------------------------------------------------------------

MEParser* MEParser::createParserByType(FormatType inType, CReader *inReader)
{
	MEParser* parser;
	
	switch (inType) {
		case kFormatPHYLIP: 
		case kFormatMOLPHY: 
		case kFormatTree: 
			parser=new MEPhylipParser(inReader);
		break;
		case kFormatNEXUS: 
			parser=new MENexusParser(inReader);
		break;
		case kFormatFLAT: 
			parser=new MEFlatParser(inReader);
		break;
		case kFormatNBRF: 
			parser=new MENBRFParser(inReader);
		break;
		case kFormatFASTA: 
			parser=new MEFASTAParser(inReader);
		break;
//		case kFormatCLUSTAL: 
//			parser=new MEClustalParser(inReader);
//		break;
		case kFormatGCG: 
			parser=new MEGCGParser(inReader);
		break;
		case kFormatGENBANK: 
			parser=new MEGenbankParser(inReader);
		break;
		case kFormatPAG: 
			parser=new MEPagParser(inReader);
		break;
		case kFormatCAIC: 
			parser=new MECAICParser(inReader);
		break;
		case kFormatTAB: 
			parser=new METabParser(inReader);
		break;
		case kFormatUnknown: 
			throw MEUnknownFormat();
		break;
		default:
			throw MEUnsupportedFormat(kFormatTypeNames[inType]);
	}
	return parser;
}

#pragma mark === Alignment Reading ===

// ---------------------------------------------------------------------------------
//		 MEParser::parseAlignment
// ---------------------------------------------------------------------------------

MEBasicAlignment *MEParser::parseAlignment()
{
	MEBasicAlignment *newAlignment;
	
	NewClass(newAlignment, MEBasicAlignment());
	parseAlignment(newAlignment);
	
	return newAlignment;
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseAlignment
// ---------------------------------------------------------------------------------

void MEParser::parseAlignment(MEBasicAlignment *inAlignment)
{
	int i, n;
	MEBasicSequence *sequence, *sequence1;
	
	readAlignment(inAlignment);
	
	n=inAlignment->getNumSequences();
	if (n==0)
		throw METooFewTaxa();
				
	// get the first sequence
	sequence1=(MEBasicSequence *)inAlignment->getSequence(0);
	
	// Filter the sequences to use the standard gap/missing/match characters
	if (strlen(gap)>1 || *gap!='-')
		sequence1->filterSequence(gap, '-');
	if (strlen(missing)>1 || *missing!='?')
		sequence1->filterSequence(missing, '?');

	// If the first sequence uses a match character then throw an error
	if (sequence1->countStates(match)!=0)
		throw MEBadState();
		
	if (sequence1->getSequenceType()==kUnknownSequence)
		sequence1->setSequenceType(seqType);

	char *seq=sequence1->getSequence();
	// Now for the rest of the taxa
	for (i=1; i<n; i++) {
		sequence=(MEBasicSequence *)inAlignment->getSequence(i);
		
		// Filter the sequences to use the standard gap/missing/match characters
		if (strlen(gap)>1 || *gap!='-')
			sequence->filterSequence(gap, '-');
		if (strlen(missing)>1 || *missing!='?')
			sequence->filterSequence(missing, '?');
			
		sequence->filterMatches(match, seq);
		
		if (sequence->getSequenceType()==kUnknownSequence)
			sequence->setSequenceType(seqType);
	}
	
	if (seqType==kDNASequence || seqType==kRNASequence)
		inAlignment->setAlignmentType(kNucleotideAlignment);
	else if (seqType==kAminoAcidSequence)
		inAlignment->setAlignmentType(kAminoAcidAlignment);

	inAlignment->trimGapsFromEnd();
	inAlignment->setTaxaDefined(true);
}


#pragma mark === Tree Reading ===

// ---------------------------------------------------------------------------------
//		 MEParser::parseTree
// ---------------------------------------------------------------------------------

METree *MEParser::parseTree()
{
	METree *newTree;
	
	NewClass(newTree, METree());
	parseTree(newTree);

	return newTree;
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseTree
// ---------------------------------------------------------------------------------

void MEParser::parseTree(METree *inTree)
{
	readTree(inTree);
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseTreeList
// ---------------------------------------------------------------------------------

METreeList *MEParser::parseTreeList()
{
	METreeList *newTreeList;
	
	NewClass(newTreeList, METreeList());
	parseTreeList(newTreeList);

	return newTreeList;
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseTreeList
// ---------------------------------------------------------------------------------

void MEParser::parseTreeList(METreeList *inTreeList)
{
	readTreeList(inTreeList);
}


#pragma mark === Matrix Reading ===

// ---------------------------------------------------------------------------------
//		 MEParser::parseDataMatrix
// ---------------------------------------------------------------------------------

MEDataMatrix *MEParser::parseDataMatrix()
{
	MEDataMatrix *newDataMatrix;
	
	NewClass(newDataMatrix, MEDataMatrix());
	parseDataMatrix(newDataMatrix);

	return newDataMatrix;
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseDataMatrix
// ---------------------------------------------------------------------------------

void MEParser::parseDataMatrix(MEDataMatrix *inDataMatrix)
{
	readDataMatrix(inDataMatrix);
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseDataAndTree
// ---------------------------------------------------------------------------------

// A number of formats (e.g., PAG, Compare) have data embedded within a tree so we
// need this special function. Other formats that support both data matrices 
// and trees (e.g., NEXUS) can also implement this.

void MEParser::parseDataAndTree(MEDataMatrix *inDataMatrix, METree *inTree)
{
	readDataAndTree(inDataMatrix, inTree);
}


// ---------------------------------------------------------------------------------
//		 MEParser::parseDataAndTreeList
// ---------------------------------------------------------------------------------

// A number of formats (e.g., PAG, Compare) have data embedded within a tree so we
// need this special function. Other formats that support both data matrices 
// and trees (e.g., NEXUS) can also implement this.

void MEParser::parseDataAndTreeList(MEDataMatrixList *inDataMatrices, METreeList *inTreeList)
{
	readDataAndTreeList(inDataMatrices, inTreeList);
}


#pragma mark === Read Functions ===

// ---------------------------------------------------------------------------------
//		 METabParser::readAlignment
// ---------------------------------------------------------------------------------

void MEParser::readAlignment(MEBasicAlignment * /*inAlignment*/)
{	
	throw MEUnsupportedData("Alignment");
}

// ---------------------------------------------------------------------------------
//		 MEParser::readTree
// ---------------------------------------------------------------------------------

void MEParser::readTree(METree * /*inTree*/)
{
	throw MEUnsupportedData("Tree");
}

// ---------------------------------------------------------------------------------
//		 MEParser::readTreeList
// ---------------------------------------------------------------------------------

void MEParser::readTreeList(METreeList * /*inTreeList*/)
{
	throw MEUnsupportedData("Trees");
}

// ---------------------------------------------------------------------------------
//		 MEParser::readDataMatrix
// ---------------------------------------------------------------------------------

void MEParser::readDataMatrix(MEDataMatrix * /*inDataMatrix*/)
{
	throw MEUnsupportedData("Data Matrix");
}

// ---------------------------------------------------------------------------------
//		 MEParser::readDataAndTree
// ---------------------------------------------------------------------------------

void MEParser::readDataAndTree(MEDataMatrix * /*inDataMatrix*/, METree * /*inTree*/)
{
	throw MEUnsupportedData("Data and Tree");
}

// ---------------------------------------------------------------------------------
//		 MEParser::readDataAndTreeList
// ---------------------------------------------------------------------------------

void MEParser::readDataAndTreeList(MEDataMatrixList * /*inDataMatrices*/, METreeList * /*inTreeList*/)
{
	throw MEUnsupportedData("Data and Tree");
}

#pragma mark === Low-level Sequence Reading ===

// ---------------------------------------------------------------------------------
//		 MEParser::readSequence
// ---------------------------------------------------------------------------------
 
// Reads sequence into seq, upto a maximum of maxLen characters. Converts to 
// upper case. Will skip any space, characters in 'ignore' or comments. Returns 
// the number of characters read.

int MEParser::readSequence(char *seq, int maxLen, char *ignore)
{
	int n=0;
	char ch, *P;
	
	P=seq;
	
	try {
		while (n<maxLen) {
			ch=toupper(reader->readChar());
			if (ch==reader->getStartComment())
				reader->skipComments();
			else if (ch==reader->getLineComment())
				reader->skipToEndOfLine();
			else if (!isspace(ch) && strchr(ignore, ch)==NULL) {
				*P=ch;
				P++;
				n++;
			}
		}
	} catch (CEndOfFile e) {
	}
	*P='\0';
	
	return n;
}


// ---------------------------------------------------------------------------------
//		 MEParser::readSequenceUntil
// ---------------------------------------------------------------------------------
 
// Reads sequence into 'seq', up to a maximum of maxLen characters or until a 
// character in 'delimiters' occurs. Converts to upper case. Will skip any 
// space, characters in 'ignore' or comments. Returns the number of 
// characters read in 'nRead' and the character that delimited the sequence.

char MEParser::readSequenceUntil(char *seq, int maxBases, char *ignore, char *delimiters, int *nRead)
{
	int n=0;
	bool done=false;
	char ch='\0', *P;
	
	P=seq;
	
	try {
		while (n<maxBases && !done) {
			ch=toupper(reader->readChar());
			if (ch==reader->getStartComment())
				reader->skipComments();
			else if (ch==reader->getLineComment())
				reader->skipToEndOfLine();
			else if (strchr(delimiters, ch)==NULL) {
				if (!isspace(ch) && strchr(ignore, ch)==NULL) {
					*P=ch;
					P++;
					n++;
				}
			} else
				done=true;
		}
	} catch(CEndOfFile) {
	}
	*P='\0';
	
	*nRead=n;
	
	return ch;
}


// ---------------------------------------------------------------------------------
//		 MEParser::readSequenceUntil
// ---------------------------------------------------------------------------------
 
// Reads sequence into 'seq', until a character in 'delimiters' occurs. Converts 
// to upper case. Will skip any space, characters in 'ignore' or comments. 
// Returns the character that delimited the sequence.

char MEParser::readSequenceUntil(MEBasicSequence *seq, char *ignore, char *delimiters)
{
	bool done=false;
	char ch='\0';
	
	try {
		while (!done) {
			ch=toupper(reader->readChar());
			if (ch==reader->getStartComment())
				reader->skipComments();
			else if (ch==reader->getLineComment())
				reader->skipToEndOfLine();
			else if (strchr(delimiters, ch)==NULL) {
				if (!isspace(ch) && strchr(ignore, ch)==NULL) {
					seq->appendSite(ch);
				}
			} else 
				done=true;
		}
	} catch(CEndOfFile) {
	}
	
	return ch;
}


#pragma mark === Low-level Tree Reading ===

// ---------------------------------------------------------------------------------
//		 MEParser::readTipLabel
// ---------------------------------------------------------------------------------
 
bool MEParser::readTipLabel(char *outName, long *outNumber)
{
	char ch;
	
	ch=reader->readToken(outName, MAX_NAME_LEN, ",):[");
	if ( strchr(",):[", ch)==NULL ) {
	
		ch = reader->skipUntil(",):[");
	}
	
	reader->unreadChar(ch);
	
	char *P=outName;
	bool isNumber=true;
	while (*P!='\0' && isNumber) {
		if (!isdigit(*P))
			isNumber=false;
		P++;
	}
	
	*outNumber = atoi(outName);
	
	return isNumber;
}

// ---------------------------------------------------------------------------------
//		 MEParser::readTip
// ---------------------------------------------------------------------------------
 
MELink *MEParser::readTip(METree *inTree)
{
	char name[MAX_NAME_LEN+1];
	MELink *link;
	long n;
	
	bool isNumber=readTipLabel(name, &n);
	
	try {
		if (inTree->getUseNumericalTips() && isNumber && n>0)
			link=inTree->newTip(n-1);
		else
			link=inTree->newTip(name);
	} catch (METaxonMissing e) {
		throw MEUnknownTaxon("Unknown taxon label in tree");
	}
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 MEParser::readNode
// ---------------------------------------------------------------------------------
 
//	First this function creates a new MELink that will point back up at
//	the ancestoral node. Then it works through each branch in a node and 
//	calls readBranch to read them. By passing the root MELink, each branch
//	can hook on to the linked list and form a polytomy.

MELink *MEParser::readNode(METree *inTree, MEDataMatrix *inData)
{
	MELink *link;
	char ch;
	
	// This MELink is the one that points back up to the root
	link=inTree->newNode();
	
	readBranch(inTree, link, inData);		
	
	ch=reader->skipUntil(",():[;");
	
	if (ch!=',')
		throw MEBranchMissing();
		
	do {
		readBranch(inTree, link, inData);

		ch=reader->skipUntil(",():[;");
	} while (ch==',');

	if (ch!=')')
		throw MEBracketMissing();
		
	return link;
}


// ---------------------------------------------------------------------------------
//		 MEParser::readBranch
// ---------------------------------------------------------------------------------
 
// Reads a branch in. This could be a node or a tip (calls readNode or readTip
// accordingly). It then reads the branch length and creates an MELink that will
// point at the new node or tip. It connects these two up and sets them to use
// the same branch number (an index to an array of branch lengths in the METree).
// Finally it inserts the new branch MELink into the polytomy given by inLink.

MELink *MEParser::readBranch(METree *inTree, MELink *inLink, MEDataMatrix *inData)
{
	char ch;
	double len;
	MELink *link, *branchLink;

	reader->skipSpace();
	if (reader->nextChar()=='(') {	/* is a node */
		reader->readChar();
		link=readNode(inTree, inData);
	} else {		/* is a tip */
		link=readTip(inTree);
	}

	ch=reader->skipUntil(",():[;");
	if (ch==':') {
		if (firstBranch) {
			inTree->setHasLengths(true);
			firstBranch=false;
		} else {
			if (!inTree->getHasLengths())
				throw MELengthsMissing();
		}
					
		ch=reader->readDouble(&len, ",():[;");
		reader->unreadChar(ch);
	} else {
		if (firstBranch) {
			inTree->setHasLengths(false);
			firstBranch=false;
		} else {
			if (inTree->getHasLengths())
				throw MELengthsMissing();
		}
					
		len=0.0;
	}
	
	branchLink=inTree->newBranch(inLink, len);
	link->branchNo=branchLink->branchNo;
	link->branch=branchLink;
	branchLink->branch=link;
	
	if (ch=='[') {
		if (inData!=NULL) {
			if (link->tipNo!=-1) {
				double param;
				
				if (inData->getNumCols()==0)
					inData->setSize(1,1);
				
				while (link->tipNo>=inData->getNumRows())
					inData->addRow();
					
				int i=0;
				do {
					ch=reader->readDouble(&param, ",]");
					if (i>=inData->getNumCols())
						inData->addColumn();
					inData->setElement(link->tipNo, i, param);
					i++;
				} while (ch==',');
				inData->setRowName(link->tipNo, inTree->getTipName(link->tipNo));
			}
		} else { // read it as a label...
			char name[MAX_NAME_LEN+1];
			char value[MAX_NAME_LEN+1];
			MELabelArray *labels;
			bool hasName;
			
			int i, j, n = 0;
			
			do {
				hasName = false;
				ch = reader->nextChar();
				if (!isdigit(ch) && ch != '.' && ch != '-') {
					// not numerical therefore probably a label name...
					ch = reader->readToken(name, MAX_NAME_LEN, "=];,");
					hasName = true;
					
					if (ch != '=')
					// ...then again, perhaps not.
						throw MENonNumericalLabel();
				}
				
				ch = reader->readToken(value, MAX_NAME_LEN, "];,");
				bool isNum = true;
				bool isInt = true;
				
				for (i = 0; i < strlen(value); i++) {
					if (!isdigit(value[i]) && value[i] != '-' && value[i] != '.')
						isNum = false;
					if (value[i] == '.')
						isInt = false;
				}
										
				if (hasName) {
					j = inTree->findBranchLabelArray(name);
					if (j < 0)
						j = n;
				} else
					j = n;
									
				if (inTree->getNumBranchLabelArrays() > j) {
					labels = inTree->getBranchLabelArray(j);
				} else {
					if (!isNum)
						NewClass(labels, MELabelArray(kStringLabel));
					else if (isInt)
						NewClass(labels, MELabelArray(kIntegerLabel));
					else
						NewClass(labels, MELabelArray(kRealLabel));

					inTree->addBranchLabelArray(labels);
					if (hasName)
						labels->setName(name);
				}
				
				if (!isNum)
					labels->setStringLabel(branchLink->branchNo, value);
				else if (isInt) {
					int intValue = atoi(value);
					labels->setIntLabel(branchLink->branchNo, intValue);
				} else {
					double realValue = atof(value);
					labels->setRealLabel(branchLink->branchNo, realValue);
				}
			} while (ch != ']');
		}
		if (ch!=']')
			ch=reader->skipUntil("]");
	}
		
	return branchLink;
}	


// ---------------------------------------------------------------------------------
//		 MEParser::readNewickTree
// ---------------------------------------------------------------------------------

void MEParser::readNewickTree(METree *inTree, MEDataMatrix *inData)
{
	MELink *link, *link1;
	
	// Determines whether the first branch has been read in order
	// to decide if the tree has branch lengths or not.
	firstBranch=true;
	
	try {
		reader->skipUntil("(");
		link=readNode(inTree, inData);
	} catch (CEndOfFile) {
		delete inTree;
		throw METreeMissing();
	} catch (MEException e) {
		delete inTree;
		throw e;
	}
			
	reader->skipUntil(";");

	// prune out the temporary MELink that pointed to the root
	link1=link;
	while (link->next!=link1) 
		link=link->next;
		
	link->next=link1->next;
	inTree->removeLink(link1);
	
	// move on to the first branch in the node
	link=link->next;
	link->setIsStem(true);
	
	if (link->getNumBranches()<3) {
		// if the root node has less than 3 branches then this node is
		// removed. This makes the tree unrooted but we store the position
		// of the root by setting METree's root1 and root2 to point to the
		// nodes either side of the root.
		inTree->setIsRooted(true);
		
		// Set the tree's root pointers
		MELink* root1=link->branch;
		MELink* root2=link->next->branch;
		inTree->setRoot1(root1);
		inTree->setRoot2(root2);
		
		// Set the tree's root branch lengths
		inTree->setRootDistances(inTree->getBranchLength(root1->branchNo),
									inTree->getBranchLength(root2->branchNo));
		
		// Set the nodes either side of the root to point to each other
		root1->branch=root2;
		root2->branch=root1;
		
		// Remove root2's branch length from the tree's array
		inTree->removeBranchLength(root2->branchNo);
		
		// Set the new branches branchNo and the corresponding branchLength
		// This just uses root1's old branch length slot and sets it to the
		// sum of the two old branch lengths
		root2->branchNo=root1->branchNo;
		inTree->setBranchLength(root1->branchNo, 
								inTree->getRootDistance1() + inTree->getRootDistance2());
		
		inTree->removeInternalNode(link);
		inTree->removeLink(link->next);
		inTree->removeLink(link);
	} else {
		inTree->setIsRooted(false);
		
		if (link->branch->getNumBranches() == 1) {
			inTree->setRoot1(link);
			inTree->setRoot2(link->branch);
		} else {
			inTree->setRoot2(link);
			inTree->setRoot1(link->branch);
		}
	}
}

