/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	MEParsimonyCriterion.cpp

		Class: MEParsimonyCriterion

	This class provides a simple parsimony criterion for pattern list given
	a tree.

 ******************************************************************************/

#include <string.h>

#include "MEExceptions.h"
#include "MEParsimonyCriterion.h"

#include "MEPatternList.h"
#include "MELink.h"
#include "METree.h"

#pragma mark === MEParsimonyCriterion ===

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion
// ---------------------------------------------------------------------------------

MEParsimonyCriterion::MEParsimonyCriterion() : METreeCriterion()
{
	patterns=NULL;
}

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion::setPatterns
// ---------------------------------------------------------------------------------

void MEParsimonyCriterion::setPatterns(MEPatternList *inPatterns)
{
	patterns=inPatterns;
	invalidate();
}

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion::calcScore
// ---------------------------------------------------------------------------------

void MEParsimonyCriterion::calcScore()
{
	ThrowIfNull(mTree);
	ThrowIfNull(patterns);
	
	int numPatterns=patterns->getNumPatterns();
	numStates=patterns->getNumStates();

	MELink *root1=mTree->getRoot1();
	MELink *root2=mTree->getRoot2();
	
	score=0;
	
	char set1[MAX_NUM_STATES];
	char set2[MAX_NUM_STATES];

	int i;
	for (i=0; i<numPatterns; i++) {
		calcScoreForNode(root1, i, set1);
		calcScoreForNode(root2, i, set2);
		
		intersectSets(set1, set2, set1);		
		if (isEmptySet(set1))
			score+=patterns->getPatternWeight(i);
	}
		
	scoreKnown=true;
}

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion::calcScore
// ---------------------------------------------------------------------------------

void MEParsimonyCriterion::calcScoreForNode(MELink *inNode, int site, char *outSet)
{
	char iSet[MAX_NUM_STATES];
	char uSet[MAX_NUM_STATES];
	char set[MAX_NUM_STATES];
	
	if (inNode->tipNo==-1) {
		MELink *link=inNode->next;
		calcScoreForNode(link->branch, site, iSet);
		memcpy(uSet, iSet, numStates);
		
		link=link->next;
		while (link!=inNode) {
			calcScoreForNode(link->branch, site, set);
			intersectSets(iSet, set, iSet);
			unionSets(uSet, set, uSet);
			
			link=link->next;
		}
		
		if (isEmptySet(iSet)) {
			score+=patterns->getPatternWeight(site);
			memcpy(outSet, uSet, numStates);
		} else
			memcpy(outSet, iSet, numStates);
	} else
		patterns->getPattern(site, inNode->tipNo, outSet);
}

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion::isEmptySet
// ---------------------------------------------------------------------------------

bool MEParsimonyCriterion::isEmptySet(char *inSet)
{
	int i;
	for (i=0; i<numStates; i++) {
		if ( inSet[i]==1 )
			return false;
	}
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion::intersectSets
// ---------------------------------------------------------------------------------

void MEParsimonyCriterion::intersectSets(char *inSet1, char *inSet2, char *outSet)
{
	int i;
	for (i=0; i<numStates; i++) {
		if ( inSet1[i] && inSet2[i] )
			outSet[i]=1;
		else
			outSet[i]=0;
	}
}

// ---------------------------------------------------------------------------------
//		 MEParsimonyCriterion::unionSets
// ---------------------------------------------------------------------------------

void MEParsimonyCriterion::unionSets(char *inSet1, char *inSet2, char *outSet)
{
	int i;
	for (i=0; i<numStates; i++) {
		if ( inSet1[i] || inSet2[i] )
			outSet[i]=1;
		else
			outSet[i]=0;
	}
}
