/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPatristicMatrix.cpp

				MEPatristicMatrix Class
					
 ******************************************************************************/

// includes:

// C/C++:
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// Mol Evol.:
#include "MEExceptions.h"
#include "MEUtilities.h"
#include "MEDistanceMatrix.h"
#include "MEPatristicMatrix.h"
#include "METree.h"
#include "MEBitset.h"
#include "MEBitsetList.h"
#include "MEPermutation.h"

// ---------------------------------------------------------------------------------
//   MEPatristicMatrix::MEPatristicMatrix
// ---------------------------------------------------------------------------------

MEPatristicMatrix::MEPatristicMatrix() : MEDistanceMatrix()
{
	pathsWithBranch = NULL;
	branchesInPath = NULL;
	tree = NULL;

}


// ---------------------------------------------------------------------------------
//   MEPatristicMatrix::~MEPatristicMatrix
// ---------------------------------------------------------------------------------
MEPatristicMatrix::~MEPatristicMatrix()
{
	if (pathsWithBranch)
		delete pathsWithBranch;
	if (branchesInPath)
		delete branchesInPath;
}


// ---------------------------------------------------------------------------------
//   MEPatristicMatrix::getDistancesFrom
// ---------------------------------------------------------------------------------
double MEPatristicMatrix::getDistancesFrom(MEBitset *A, MEDistanceMatrix *M)
{
	if (A->isEmpty())
		return 0.0;
	if (M == NULL)
		M = this;
	double d = 0.0;
	
	int i, j, n = tree->getNumTips();
	if (A->getCard() == n)
		return 0.0;
	MEBitset *B;
	NewClass(B, MEBitset(n));
	B->fill(n);
	B->del(*A);
	for (i = A->getFirstElement(); i >= 0; i = A->getNextElement()) {
		for (j = B->getFirstElement(); j >= 0; j = B->getNextElement()) {
			d += M->getElement(i, j);
		}
	}
	return d / (double) (A->getCard() * B->getCard());
}



// ---------------------------------------------------------------------------------
//   MEPatristicMatrix::getLeastSquaresFit
// ---------------------------------------------------------------------------------
double MEPatristicMatrix::getLeastSquaresFit(MEDistanceMatrix *M, 
		double inBranchLength, int inBranch)
//: Calculate the weighted sum of absolute differences between this and another matrix
// Calculate the sum of squared difference between path distances
// for this matrix and the matrix M.  Weight each term by 1/(z^r), where z is 
// the patristic distance in this matrix and r is a non-negative real number.
{
	if (!_fitKnown)
		calcLeastSquaresFit(M, inBranchLength, inBranch);
	return fit;
}

void MEPatristicMatrix::calcLeastSquaresFit(MEDistanceMatrix *M, 
		double inBranchLength, int inBranch)
{
	double w, x, y, z;
	int i, j, n = M->getNumCols(), pathNum, px, py;
	if (inBranch == -1) {
//	This is the naive way of doing it: just count up all over again....
		for (i = 0; i < n; i++) {
			for (j = i+1; j < n; j++) {
				z = tree->getNodeToNodeDistance(i, j);	// these are the tips
				w = M->getElement(i, j) - z;
				x = w * w;
				y = pow(z, r);
				fit += x/y;
			}
		}
	} else {
		MEBitset *A = pathsWithBranch->getBitset(inBranch);
		for (pathNum = A->getFirstElement(); pathNum >= 0; pathNum = A->getNextElement()) {
			// find the pair of tips for this path number:
			// (arranged (0,1), (0,2), (1,2), (0,3), (1,3), (2,3)...)
			py = 1;
			for (i = 0; i <= pathNum; i += py)
				py++;
			px = pathNum - (py-1)*(py-2)/2 - 1;
			
			// remove the old contribution from this path:
			z = tree->getNodeToNodeDistance(px, py);
			w = M->getElement(px, py) - z;
			x = w * w;
			y = pow(z, r);
			fit -= x/y;
			
			// add the contribution when the inBranch-th branch has length inBranchLength:
			z += inBranchLength - tree->getBranchLength(inBranch);
			w = M->getElement(px, py) - z;
			x = w * w;
			y = pow(z, r);
			fit += x/y;
		}
	}
	_fitKnown = true;
}


// ---------------------------------------------------------------------------------
//   MEPatristicMatrix
// ---------------------------------------------------------------------------------
//: optimize the branch lengths on T to minimize the sum of squares of differences
// between this patristic matrix
void MEPatristicMatrix::minimizeLeastSquares(MEDistanceMatrix *M)
// Calculate the path lengths for each pair of tips, and store the bitsets of
// branch indices as we go.  Then we can access all the distances affected by a 
// given branch.  In fact, store the indices in a list of bitsets, one set for 
// each branch: then we can alter all the distances depending on one branch really fast.
{
	// branchesInPath contains path sets (0,1), (0,2), (1,2), (0,3),...
	// 1. choose a branch -- the largest?
	// 2. do brent on that branch
	// 3. test for convergence
	
	bool _notDone = true;
	int b, i, j, k = tree->getNumBranches();
	double *delta;	// the array of amounts of change introduced to each of these branches.
	NewArray(delta, double, k);
	double 
		x,		// the branch length
		dz,	// the change in z, the minimization criterion 
		maxdz,
		newz,	// the new score of the tree w.r.t. the distance matrix M
		z;		// the old score of the tree w.r.t. the distance matrix M

	for (i = 0; i < k; i++)
		delta[i] = tree->getBranchLength(i);
	
	double tol = 0.01;	// quite big for the moment -- shall reduce tolerance later
	j = 0;
	int maxIterations = 100;
	permutation P(k);
	P.sort(descending, delta);
	
	while (_notDone) {
		// iterate around all the branches:
		maxdz = 0.0;
		for (i = 0; i < k; i++) {
			b = P[i];
	
			z = getLeastSquaresFit(M, b, x);
			
			// optimize the branch with Brent:
			Brent(b, 0.0, 10000.0, x, M);
			
			newz = getLeastSquaresFit(M, b, x);
			
			dz = z - newz;
			delta[b] = dz / z;
			maxdz = max(dz, maxdz);
		}
		// test for convergence:
		if (fabs(maxdz) < tol || j++ > maxIterations)
			break;

	}
	
/*
		// choose a branch:
		x = 0.0;
		for (i = 0; i < k; i++) {
			d = delta[i] * tree->getBranchLength(i);
				// this, and setting the delta's to all large numbers to begin with, means
				// that all the branches will be gone through once, and then we'll only 
				// go through those with big changes.  Neat, even if I do say so myself.
			if (d > x) {
				x = d;
				b = i;
			}
		}
*/

/*	int i, j, idx = 0, k = tree->getNumTips(), p;
	MEBitset *A;
	for (j = 1; j < k; j++) {
		for (i = 0; i < j; i++) {
			A = pathSet->getBitset(idx++);
			for (p = A->getFirstElement(); p >= 0; p = A->getNextElement()) {
				plus(i, j, tree->getBranchLength(p));	// add the branch length for this 
			}
		}
	}
*/

/*
	// do a couple of passes around all the branches:
	for (int pass = 0; pass < 10; pass++) {
		k = tree->getNumBranches();
		for (i = 0; i < k; i++) {
			// linear search on each branch length:
			// since this is a least squares approach we model the space with a 2nd order 
			// polynomial function in (n-2)-space:
			
			tree->putSplit(A, i);	// find the leaf set partition induced by branch i
			a = A->getCard();
			// find the sum of distances defined in M going through this branch:
			d = getDistancesFrom(A, M);
			cout << "\tg[" << i << "] = " << tree->getBranchLength(i) << " <- ";
			tree->setBranchLength(i, d/(double) a);
			cout << tree->getBranchLength(i) << "\n";
		}
		fit = getLeastSquaresFit(M);
		cout << "fit = " << fit << "\n";
	}	
*/

}


// ---------------------------------------------------------------------------------
//   MEPatristicMatrix
// ---------------------------------------------------------------------------------
void MEPatristicMatrix::setTree(METree *T)
{
	double d;
	int i, j, k, nBranches, nTips, nPaths;
	if (tree == NULL)
		nTips = -1;
	else
		nTips = tree->getNumTips();
	if (nTips != T->getNumTips()) {
		if (pathsWithBranch != NULL) {
			delete pathsWithBranch;
			pathsWithBranch = NULL;
		}
		if (branchesInPath != NULL) {
			delete branchesInPath;
			branchesInPath = NULL;
		}
	}
	if (branchesInPath == NULL) {
		NewClass(branchesInPath, MEBitsetList());
		branchesInPath->setOwningSets(true);	// so the bitsets will die at the end
	}
	if (pathsWithBranch == NULL) {
		NewClass(pathsWithBranch, MEBitsetList());
		pathsWithBranch->setOwningSets(true);	// so the bitsets will die at the end
	}
	
	nTips = T->getNumTips();
	nBranches = T->getNumBranches();
	nPaths = nTips * (nTips - 1)/2;
	tree = T;

	MEBitset *A;
	
	k = branchesInPath->getNumObjects();
	if (k < nPaths) {
		for (i = 0; i < k; i++)
			branchesInPath->getBitset(i)->clear();
		for (i = k; i < nPaths; i++) {
			NewClass(A, MEBitset());
			branchesInPath->add(A);
		}
	}
	// Now find the paths with each branch:
	k = pathsWithBranch->getNumObjects();
	if (k < nBranches) {
		for (i = 0; i < k; i++)
			pathsWithBranch->getBitset(i)->clear();
		for (i = k; i < nBranches; i++) {
			NewClass(A, MEBitset());
			pathsWithBranch->add(A);
		}
	}
	
	setSize(nTips, nTips);
	
	// paths are arranged (0,1), (0,2), (1,2), (0,3), (1,3), (2,3)...
	// THIS ORDERING IS CRITICAL!!!  
	// THE CONVERSION FROM PATH NUMBER TO TIP PAIR RELIES ON IT!!!
	int idx = 0;
	for (j = 1; j < nTips; j++) {
		for (i = 0; i < j; i++) {
			d = tree->getPath(tree->getTip(i), tree->getTip(j), branchesInPath->getBitset(idx));
				// returns the weighted path length (=distance) AND puts the set of 
				// branch indices into the idx-th bitset of branchesInPath
			setElement(i, j, d);
			setElement(j, i, d);
			idx++;
		}
	}
	
	// now fill in the bitset list for each branch:
	for (i = 0; i < nPaths; i++) {
		A = branchesInPath->getBitset(i);
		for (j = 0; j < nBranches; j++) {
			if (A->contains(j))
				(pathsWithBranch->getBitset(j))->add(i);
		}
	}
}



// ---------------------------------------------------------------------------------
//		 MLIterativeOptimiser::brent
// ---------------------------------------------------------------------------------
#define kMaxBrentIterations 100
#define kBrentTolerance 0.001
const double CGOLD = 0.3819660;
const double ZEPS = 1.0e-10;

double MEPatristicMatrix::Brent(int inBranch, double inMin, double inMax, 
	double &inCurrent, MEDistanceMatrix *D)
{
	int iter;
	double a, b, d, etemp, fu, fv, fw, fx, p, q, r, tol1, tol2, u, v, w, x, xm;
	double e = 0.0;

	a = inMin;
	b = inMax;
	x = w = v = inCurrent;	
	fw = fv = fx = getLeastSquaresFit(D, inCurrent, inBranch);
	
	for (iter = 1; iter <= kMaxBrentIterations; iter++) {
		xm = 0.5*(a + b);
		tol2 = 2.0*(tol1 = kBrentTolerance * fabs(x) + ZEPS);
		if (fabs(x-xm) <= (tol2 - 0.5*(b - a))) {
			inCurrent = x;
			return fx;
		}
		if (fabs(e) > tol1) {
			r = (x - w) * (fx - fv);
			q = (x - v) * (fx - fw);
			p = (x - v)*q - (x - w)*r;
			q = 2.0*(q - r);
			if (q > 0.0) 
				p = -p;
			q = fabs(q);
			etemp = e;
			e = d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d = CGOLD*(e = (x >= xm ? a-x : b-x));
			else {
				d = p/q;
				u = x + d;
				if (u-a < tol2 || b-u < tol2)
					d = sign(tol1, xm - x);
			}
		} else {
			d = CGOLD*(e = (x >= xm ? a-x : b-x));
		}
		
		u = (fabs(d) >= tol1 ? x + d : x + sign(tol1, d));
		inCurrent = u;
		fu = getLeastSquaresFit(D, inCurrent, inBranch);
		
		if (fu <= fx) {
			if (u >= x)
				a = x; 
			else 
				b = x;
				
			shift(v, w, x, u);
			shift(fv, fw, fx, fu);
		} else {
			if (u < x)
				a = u; 
			else
				b = u;
				
			if (fu <= fw || w == x) {
				v = w;
				w = u;
				fv = fw;
				fw = fu;
			} else if (fu <= fv || v == x || v == w) {
				v = u;
				fv = fu;
			}
		}
	}
	inCurrent = x;

	throw MEException("Too many iterations in MEPatristicMatrix::brent");
	
	return fx;
}

/*double MEPatristicMatrix::evaluate(MLParameter *inParam, double inValue)
{
	evaluateCount++;
	inParam->setValue(inValue);
	return function->evaluate();
}*/

