/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	MEPatternList.cpp

		Class: MEPatternList
				
	Stores and sorts patterns.
	    
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "MEGlobal.h"
#include "MEExceptions.h"
#include "MEUtilities.h"
#include "MEBitset.h"
#include "MEPatternList.h"
#include "METaxonList.h"
#include "MEBasicAlignment.h"

#pragma mark === MEPatternList ===

// ---------------------------------------------------------------------------------
//		 MEPatternList
// ---------------------------------------------------------------------------------

MEPatternList::MEPatternList() : MEObject()
{
	patternStrings=NULL;
	patternWeights=NULL;
	normWeights=NULL;
	normWeightsKnown=false;
	numPatterns=0;
	patternCapacity=0;
		
	numTaxa=0;
	numStates=0;
}

// ---------------------------------------------------------------------------------
//		 ~MEPatternList
// ---------------------------------------------------------------------------------
MEPatternList::~MEPatternList()
{
	int i;
	
	if (patternStrings!=NULL) {
		for (i=0; i<numPatterns; i++)
			if (patternStrings[i]!=NULL)
				delete[] patternStrings[i];
		delete[] patternStrings;
	}
	
	if (patternWeights!=NULL)
		delete[] patternWeights;
		
	if (normWeights!=NULL)
		delete[] normWeights;
}


#pragma mark -


// ---------------------------------------------------------------------------------
//		 MEPatternList::addPatterns
// ---------------------------------------------------------------------------------
// Adds all the patterns contained within inAlignment.

void MEPatternList::addPatterns(MEBasicAlignment *inAlignment)
{
	ThrowIfNull(inAlignment);

	int numTaxa = inAlignment->getNumTaxa();
	int numStates = inAlignment->getNumStates();
	int numSites = inAlignment->getNumSites();
	
	setDimensions(numTaxa, numStates);
	
	char *patternString;
	double d;
	int i;
	
	for (i = 0; i < numSites; i++) {
		NewArray(patternString, char, numTaxa*numStates);
	
		if (numStates == 1)
			inAlignment->getSiteString(i, patternString);
		else
			inAlignment->getPattern(i, patternString);
		d = inAlignment->getSiteWeight(i);
		addPattern(patternString, d, false);
	}
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::addPattern
// ---------------------------------------------------------------------------------
// Adds a new pattern to the list. This would normally have been previously
// removed from another MEPatternList object using removePattern. This will be
// compared with the current patterns and if it already exists the weight
// will be incremented instead of adding it.
int MEPatternList::addPattern(char *inPattern, double inWeight, bool makeCopy)
{
	// Check to see if this string already exists
	int j = 0;
	int patternLen = numStates*numTaxa;
	
	j = findPattern(inPattern);
	if (j == -1) {// i.e. pattern not found - add it
		checkCapacity(numPatterns+1);
	
		j = numPatterns;
		numPatterns++;

		// make the new entry in patternStrings equal to pattern...
		if (makeCopy) {
			NewArray(patternStrings[j], char, patternLen);
			memcpy(patternStrings[j], inPattern, patternLen);
		} else
			patternStrings[j] = inPattern;
		patternWeights[j] = inWeight;
	} else {
		patternWeights[j] += inWeight;
		
		if (!makeCopy)
			delete [] inPattern;

/*		This causes problems with MLSitePatterns because it would need
		to update its own list of pattern indexes.
		
		if (j > 0 && patternWeights[j] > patternWeights[j-1]) { // swap pattern's places
			char *ptr = patternStrings[j];
			patternStrings[j] = patternStrings[j-1];
			patternStrings[j-1] = ptr;
			
			double tmp = patternWeights[j];
			patternWeights[j] = patternWeights[j-1];
			patternWeights[j-1] = tmp;
			
			j--;
		} */
	}
	
	normWeightsKnown = false;

	return j;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::setPattern
// ---------------------------------------------------------------------------------
// Sets the pattern at index, and its weight. If index >= numPatterns then
// numPatterns becomes index + 1. This means that all the interviening patterns
// should have been set.
void MEPatternList::setPattern(int index, char *inPattern, double inWeight, bool makeCopy)
{
	// Check to see if this string already exists
	int patternLen = numStates*numTaxa;

	if (index >= numPatterns) {
		checkCapacity(index+1);
		numPatterns = index+1;
	}
	
	if (makeCopy) {
		if (patternStrings[index]==NULL)
			NewArray(patternStrings[index], char, patternLen);
			
		memcpy(patternStrings[index], inPattern, patternLen);
	} else {
		if (patternStrings[index]!=NULL)
			delete [] patternStrings[index];
			
		patternStrings[index] = inPattern;
	}
	patternWeights[index] = inWeight;
	
	normWeightsKnown = false;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::setPatterns
// ---------------------------------------------------------------------------------
// Sets the entire pattern list. The pattern strings are co-opted (i.e. the patternString
// array point to the strings in inPatterns. The pattern weights array is copied.
void MEPatternList::setPatterns(char **inPatterns, double *inWeights, 
									int inNumPatterns, int inNumTaxa, int inNumStates)
{
	int i;
	
	resetPatternStrings();
	
	numTaxa=inNumTaxa;
	numStates=inNumStates;

	checkCapacity(inNumPatterns);
	
	numPatterns=inNumPatterns;
	for (i=0; i<numPatterns; i++)
		patternStrings[i]=inPatterns[i];
		
	memcpy(patternWeights, inWeights, sizeof(double)*numPatterns);

	normWeightsKnown=false;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::findPattern
// ---------------------------------------------------------------------------------
// Return the index in the list of the input pattern, if it occurs in the list;
// otherwise, return -1.
int MEPatternList::findPattern(char *searchPattern)
{
	int j=0;
	while (j < numPatterns && !isSamePatterns(patternStrings[j], searchPattern))
		j++;
	
	if (j==numPatterns)
		return -1;

	return j;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::isSamePatterns
// ---------------------------------------------------------------------------------
// Return whether the two patterns are the same.
bool MEPatternList::isSamePatterns(char *pattern1, char *pattern2)
{
	return memcmp(pattern1, pattern2, numStates*numTaxa) == 0;
}


#pragma mark -


// ---------------------------------------------------------------------------------
//		 MEPatternList::checkCapacity
// ---------------------------------------------------------------------------------
// Checks that the patternList has enough space and resizes it as necessary
void MEPatternList::checkCapacity(int requiredCapacity)
{
	if (requiredCapacity>patternCapacity) {
		char **newPatternStrings;
		double *newWeights;
		int i;
		
		// Increase capacity
		while (patternCapacity<requiredCapacity)
			patternCapacity+=100;
		
		// NewArray the new arrays.
		NewArray(newPatternStrings, char*, patternCapacity);
		NewArray(newWeights, double, patternCapacity);
		
		if (patternStrings!=NULL) {
			// Copy the existing arrays into the new arrays.
			for (i=0; i<numPatterns; i++)
				newPatternStrings[i]=patternStrings[i];
			delete[] patternStrings;
		}
		if (patternWeights!=NULL) {
			// Copy the existing arrays into the new arrays.
			memcpy(newWeights, patternWeights, sizeof(double)*numPatterns);
			delete[] patternWeights;
		}
		
		if (normWeights!=NULL)
			delete[] normWeights;
		normWeights=NULL;
			
		// Set the remaining pointers to NULL.
		// Probably unnecessary but makes for easier error checking.
		for (i=numPatterns; i<patternCapacity; i++)
			newPatternStrings[i]=NULL;
		
		// Free the old arrays and copy over the new one.
		patternStrings=newPatternStrings;
		patternWeights=newWeights;
	}
}


#pragma mark -


// ---------------------------------------------------------------------------------
//		 MEPatternList::getPattern
// ---------------------------------------------------------------------------------
char* MEPatternList::getPattern(int index)
{
	TestBounds(index, numPatterns);
	ThrowIfNull(patternStrings);
	
	return patternStrings[index];
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::getPattern
// ---------------------------------------------------------------------------------
void MEPatternList::getPattern(int index, char *inPattern)
{
	TestBounds(index, numPatterns);
	ThrowIfNull(patternStrings);
	ThrowIfNull(inPattern);
	
	memcpy(inPattern, patternStrings[index], numTaxa*numStates);
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::getPattern
// ---------------------------------------------------------------------------------
void MEPatternList::getPattern(int index, int taxon, char *inPattern)
{
	TestBounds(index, numPatterns);
	TestBounds(taxon, numTaxa);
	ThrowIfNull(patternStrings);
	ThrowIfNull(inPattern);
	
	memcpy(inPattern, patternStrings[index]+taxon*numStates, numStates);
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::getPatternWeight
// ---------------------------------------------------------------------------------
double MEPatternList::getPatternWeight(int index)
{
	TestBounds(index, numPatterns);
	ThrowIfNull(patternWeights);
	
	return patternWeights[index];
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::getNormPatternWeight
// ---------------------------------------------------------------------------------
double MEPatternList::getNormPatternWeight(int index)
{
	TestBounds(index, numPatterns);
	ThrowIfNull(patternWeights);
	
	if (!normWeightsKnown)
		normaliseWeights();
		
	return normWeights[index];
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::getTotalPatternWeight
// ---------------------------------------------------------------------------------
double MEPatternList::getTotalPatternWeight()
{
	double d = 0.0;
	for (int i = 0; i < numPatterns; i++)
		d += getPatternWeight(i);
	return d;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::getSequence
// ---------------------------------------------------------------------------------
char* MEPatternList::getSequence(int index, char *str)
{
	if (str == NULL)
		NewArray(str, char, getNumPatterns());
	int j = 0, n = getNumTaxa(), p = getNumPatterns();
	while (j < p) {
		str[j] = getPattern(j)[index];	// get the index-th character at the j-th pattern
		j++;
	}
	str[p] = '\0';
	return str;
}


char MEPatternList::getState(int n, int s)
{
	char *str = patternStrings[s];
	return str[n];
}

// ---------------------------------------------------------------------------------
//		 MEPatternList::getEmpiricalFrequencies
// ---------------------------------------------------------------------------------
// This function estimates the state frequencies for the patternList. This consists of
// one frequency for each state returned in the provided array (which should
// be allocated enough space).
double *MEPatternList::getEmpiricalFrequencies(double *inFrequencies)
{
	int i, j, k;
	double total, sum, x, w, difference;
	double *tempFreq;
	char *pattern;

	int nStates = getNumStates();
	int nTaxa = getNumTaxa();
	int nPatterns = getNumPatterns();
	
	double *freqs = getEqualFrequencies(inFrequencies);
		
	NewArray(tempFreq, double, nStates);
	NewArray(pattern, char, nStates*nTaxa);
	
	do {
		for (i = 0; i < nStates; i++)
			tempFreq[i] = 0.0;
			
		total=0.0;
		for (i = 0; i < nTaxa; i++) {
			for (k = 0; k < nPatterns; k++) {
				getPattern(k, i, pattern);
				w = getPatternWeight(k);
				sum=0.0;
				for (j = 0; j < nStates; j++)
					sum += freqs[j] * pattern[j];
					
				for (j = 0; j < nStates; j++) {
					x = (freqs[j] * pattern[j] * w) / sum;
					tempFreq[j] += x;
					total += x;
				}
			}
		}
		
		difference = 0.0;
		for (i = 0; i < nStates; i++) {
			difference += fabs((tempFreq[i] / total) - freqs[i]);
			freqs[i] = tempFreq[i] / total;
		}
	} while (difference > kTolerance);
	
	delete[] tempFreq;
	delete[] pattern;
	
	return freqs;
}

// ---------------------------------------------------------------------------------
//		 MEPatternList::getEqualFrequencies
// ---------------------------------------------------------------------------------
// This function returns the equal state frequencies for the alignment. 
double *MEPatternList::getEqualFrequencies(double *inFrequencies)
{
	int nStates=getNumStates();
	
	double *freqs = inFrequencies;
	if (freqs == NULL)
		NewArray(freqs, double, nStates);
	
	for (int i=0; i<nStates; i++)
		freqs[i]=1.0/nStates;
		
	return freqs;
}
		
#pragma mark -


// ---------------------------------------------------------------------------------
//		 MEPatternList::removePattern
// ---------------------------------------------------------------------------------
int MEPatternList::removePattern(int index, double inWeight)
{
	TestBounds(index, numPatterns);
	
	normWeightsKnown=false;
	
	patternWeights[index]-=inWeight;
	if (patternWeights[index]<=0.0) { // remove the pattern totally
		delete [] patternStrings[index];
		numPatterns--;
		
		if (index!=numPatterns) { // Now move down the pattern at the end of the list
			patternStrings[index]=patternStrings[numPatterns];
			patternWeights[index]=patternWeights[numPatterns];
		}
		return numPatterns;
	}

	return -1;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::removePattern
// ---------------------------------------------------------------------------------
int MEPatternList::removePattern(char *inPattern, double inWeight)
{
	int index=findPattern(inPattern);
	ThrowIf(index<0);

	return removePattern(index, inWeight);
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::resetPatternStrings
// ---------------------------------------------------------------------------------
void MEPatternList::resetPatternStrings()
{
	int i;

	if (patternStrings!=NULL) {
		for (i=0; i<numPatterns; i++) {
			if (patternStrings[i]!=NULL)
				delete [] patternStrings[i];
		}
	}
	numPatterns=0;
	normWeightsKnown=false;
}


#pragma mark -


// ---------------------------------------------------------------------------------
//		 MEPatternList::setDimensions
// ---------------------------------------------------------------------------------
void MEPatternList::setDimensions(int inNumTaxa, int inNumStates)
{
	if (numTaxa!=inNumTaxa || numStates!=inNumStates)
		resetPatternStrings();
		
	numTaxa=inNumTaxa;
	numStates=inNumStates;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::setPatternWeight
// ---------------------------------------------------------------------------------
// Sets an individual pattern weight
void MEPatternList::setPatternWeight(int index, double w)
{
	TestBounds(index, numPatterns);
	ThrowIfNull(patternWeights);
	patternWeights[index] = w;
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::sortPatterns
// ---------------------------------------------------------------------------------

// Sort the patterns into descending order of frequency

void MEPatternList::sort(bool _ascending)
//: Sort the patterns in order by their weights.  Default order is descending.
{
	if (numPatterns > 1) {

		int	hh = 1;						// Find starting "h" value
		int	stopH = numPatterns / 9;
		while (hh < stopH) {
			hh = 3 * hh + 1;
		}
		
		char *str, *str2;
		
		int direction = (_ascending ? 1 : -1);
		double w, w2;
		
		for ( ; hh > 0; hh /= 3) {
				
			for (int step = hh + 1; step <= numPatterns; step++) {
				w = getPatternWeight(step - 1);
				str = patternStrings[step-1];
				long i;
				for (i = step - hh; i > 0; i -= hh) {
					if ((int) (w > getPatternWeight(i-1)) * direction >= 0) {
						break;
					}
					str2 = patternStrings[i - 1];
					w2 = getPatternWeight(i-1);
					patternStrings[i + hh - 1] = str2;
					patternWeights[i + hh - 1] = w2;
				}
					
				patternStrings[i + hh - 1] = str;
				patternWeights[i + hh - 1] = w;
			}
		}
	}
}


// ---------------------------------------------------------------------------------
//		 MEPatternList::normaliseWeights
// ---------------------------------------------------------------------------------
void MEPatternList::normaliseWeights()
{
	if (numPatterns==0)
		return;
		
	if (normWeights==NULL)
		NewArray(normWeights, double, patternCapacity);
		
	double sum=0.0;
	int i;
	for (i=0; i<numPatterns; i++)
		sum+=patternWeights[i];
		
	for (i=0; i<numPatterns; i++)
		normWeights[i]=patternWeights[i]/sum;

	normWeightsKnown=true;
}


