/******************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPermutation.h
 
 ******************************************************************/


//  includes:
// C/C++:
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
// Project specific:
#include "MEGlobal.h"
#include "MEExceptions.h"
// my includes:
#include "MEUtilities.h"
#include "MEMaths.h"
#include "MEBitset.h"
#include "MEPermutation.h"


//  macros:
#define ThrowBadPermPos(x)	if ((x < 0) || (x > dim)) { \
	cerr << "permutation position x=" << x << " is out of range\n"; \
	exit(669); }

// variables:
char
	pmsg[kShortStringLength];

//  accessed classes:
class MEBitset;


// methods:

// ------------------------------------------------------------------------------------
//    permutation::permutation
// ------------------------------------------------------------------------------------
permutation::permutation(int d)
{
	if ((d < 0) || (d > kMaxPermutationSize)) {
		sprintf(pmsg, "permutation::permutation(): int d=%d is"
			" out of range [0, %d]", d, kMaxPermutationSize);
		cerr << pmsg;
		exit(668);
	}
	dim = size = d;
//	perm = new int[size];	// deleted in ~permutation()
	NewArray(perm, int, size);
//	if (perm == NULL) {
//		cerr << "permutation::permutation(): cannot create perm array\n";
//		exit(668);
//	}
	for (int i = 0; i < size; i++)
		perm[i] = i;
}


// ------------------------------------------------------------------------------------
//    permutation::~permutation
// ------------------------------------------------------------------------------------
permutation::~permutation()
{
	delete [] perm;
	perm = NULL;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::_contains
// ------------------------------------------------------------------------------------
bool permutation::_contains(MEBitset &bs)
{
	if (&bs == NULL) {
		cerr << "permutation::_contains(bs): MEBitset bs is NULL\n";
		exit(668);
	}
	MEBitset A(size);
	for (int i = 0; i < dim; i++)
		A.add(perm[i]);
	bool _ok = false;
	if (A.contains(bs))
		_ok = true;
	return _ok;
}
bool permutation::_contains(permutation &p)
{
	if (&p == NULL) {
		cerr << "permutation::_contains(p): permutation p is NULL\n";
		exit(668);
	}
	int i,j=0;
	for (i = 0; i < dim; i++) {
		if (perm[i] == p.perm[j])
			j++;
	}
	if ((i >= dim) || (j != p.dim-1))
		return false;
	return true;
}
bool permutation::_contains(int x)
{
	for (int i = 0; i < dim; i++)
		if (perm[i] == x)
			return true;
	return false;
}


// ------------------------------------------------------------------------------------
//    permutation::_same
// ------------------------------------------------------------------------------------
bool permutation::_same(const permutation &p)
{
	if (&p == NULL) {
		cerr << "permutation::_same(p): permutation p is NULL\n";
		exit(668);
	}
	if (dim != p.dim)
		return false;
	for (int i = 0; i < dim; i++)
		if (perm[i] != p.perm[i])
			return false;
	return true;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::choose
// ------------------------------------------------------------------------------------
permutation& permutation::choose(int n, permutation &p)
{
	// choose n elements from this and put them into p
	p.copyFrom(*this);
	if (n > dim) {
		p.resize(n);
		p.randomize();
	}
	else {
		p.randomize();
		p.resize(n);
	}
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::clear
// ------------------------------------------------------------------------------------
permutation& permutation::clear()
{
	for (int i = 0; i < size; i++)
		perm[i] = 0;
	dim = 0;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::copyFrom
// ------------------------------------------------------------------------------------
permutation& permutation::copyFrom(const permutation &p)
{
	// copy p into this
	if (&p == NULL) {
		cerr << "permutation::copyFrom(p): permutation p is NULL\n";
		exit(668);
	}
	resize(p.size);
	int i;
	for (i = 0; i < size; i++)
		perm[i] = p.perm[i];
	return *this;
}
permutation& permutation::copyFrom(const MEBitset &b)
{
	// copy p into this
	if (&b == NULL) {
		cerr << "permutation::copyFrom(p): permutation p is NULL\n";
		exit(668);
	}
	resize(b.getCard());
	int i;
	b.resetCounter();
	for (i = 0; i < size; i++)
		perm[i] = b.getNextElement();
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::del
// ------------------------------------------------------------------------------------
permutation& permutation::del(int pos, int x)
{
	// delete element x which should be at position pos
	ThrowBadPermPos(pos);
	if (perm[pos] != x) {
		cerr << "permutation::del(pos, x): perm[pos=" << pos 
				<< "] is not x=" << x << "\n";
		exit(668);
	}
	dim--;
	for (int i = pos; i < dim; i++)
		perm[i] = perm[i+1];
	perm[dim] = 0;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::delElement
// ------------------------------------------------------------------------------------
permutation& permutation::delElement(int x)
{
	for (int pos = 0; pos < dim; pos++)
		if (perm[pos] == x)
			del(pos, x);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::delIthElement
// ------------------------------------------------------------------------------------
permutation& permutation::delIthElement(int i)
{
	ThrowBadPermPos(i);
	del(i, perm[i]);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::delLast
// ------------------------------------------------------------------------------------
permutation& permutation::delLast(int y)
{
	// cut the last y elements from this permutation
	ThrowBadPermPos(y);
	for (int i = dim-y-1; i < dim; i++)
		perm[i] = 0;
	dim -= y;
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::fillFrom
// ------------------------------------------------------------------------------------
permutation& permutation::fillFrom(int x)
{
	for (int i = 0; i < dim; i++)
		perm[i] = i + x;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::fillRandom
// ------------------------------------------------------------------------------------
permutation& permutation::fillRandom(const MEBitset &b)
{
	int i, k = b.getCard();
	for (i = 0; i < dim; i++)
		perm[i] = b.getRandomElement();
	size = dim;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::fillRandomMap
// ------------------------------------------------------------------------------------
permutation& permutation::fillRandomMap(int x, int y, int type)
{
	// fill in the elements of this so that i is mapped to perm[i], with image space [x,...,y]
	int i, r;
	if (x < y)
		r = y - x + 1;
	else
		r = x - y + 1;
	permutation *pi;
	switch (type) {
		case 0:
			for (i = 0; i < dim; i++)
				perm[i] = MERandom::iRan(r) + min(x, y);
			break;
		case one2one:
			// one2one: f(a) = f(b) => a = b
			if (r < dim)
				resize(r);
			NewClass(pi, permutation(r));
//			pi = new permutation(r);	// deleted in this method
//			if (pi == NULL) {
//				cerr << "permutation::fillRandomMap(x, y, type): cannot create interim"
//					" permutation pi\n";
//				exit(668);
//			}
			pi->fillFrom(x);	// make permutation of [x,...,y]
			pi->choose(dim, *this);	// choose (dim) of these at random and put them into this
			delete pi;
			break;
		case onto:
			// onto: z in f([]) => there exists a u in perm s.t. f(u) = z
			if (dim < r)
				resize(r);
			fillFrom(x);	// to guarantee that each element in [x,...,y] has a pre-image in this
			for (i = r; i < dim; i++)
				perm[i] = MERandom::iRan(r) + x;	// fill up the others at random
			randomize();	// rearrange to give random mapping
			break;
		case one2one+onto:
			// one2one: f(a) = f(b) => a = b
			// onto: z in f([]) => there exists a u in perm s.t. f(u) = z
			resize(r);	// to get correct number of elements in image set
			fillFrom(x);
			randomize();	// simple bijection			
			break;
		default:
			break;
	}
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::fillRepeat
// ------------------------------------------------------------------------------------
permutation& permutation::fillRepeat(const MEBitset &b)
{
	int i, k = b.getCard();
	for (i = 0; i < dim; i += k)
		suffix(b);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::fillStep
// ------------------------------------------------------------------------------------
permutation& permutation::fillStep(int maxVal)
{
	int 
		i, j = 0, 
		k = div(size, maxVal).quot;
	for (i = 0; i < size; ) {
		perm[i++] = j;
		if (i % k == 0)
			j++;
	}
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::getIthElement
// ------------------------------------------------------------------------------------
int permutation::getIthElement(int i)
{
	ThrowBadPermPos(i);
	return perm[i];
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::insert
// ------------------------------------------------------------------------------------
permutation& permutation::insert(int pos, int x)
{
	ThrowBadPermPos(pos);
	dim++;
	if (dim >= size) {
		cerr << "permutation::insert(pos, x): no more room to insert\n";
		exit(668);
	}
	for (int i = dim-1; i > pos; i--)
		perm[i] = perm[i-1];
	perm[pos] = x;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::isFirstLex
// ------------------------------------------------------------------------------------
bool permutation::isFirstLex()
{
	for (int i = 0; i < dim-1; i++) {
		if (perm[i] > perm[i+1])
			return false;
	}
	return true;
}


// ------------------------------------------------------------------------------------
//    permutation::isFirstLexSection
// ------------------------------------------------------------------------------------
bool permutation::isFirstLexSection(int start, int end)
{
	ThrowBadPermPos(start);
	ThrowBadPermPos(end);
	for (int i = start; i < end; i++) {
		if (perm[i] > perm[i+1])
			return false;
	}
	return true;
}


// ------------------------------------------------------------------------------------
//    permutation::isLastLex
// ------------------------------------------------------------------------------------
bool permutation::isLastLex()
{
	for (int i = 0; i < dim-1; i++) {
		if (perm[i] < perm[i+1])
			return false;
	}
	return true;
}


// ------------------------------------------------------------------------------------
//    permutation::isLastLexSection
// ------------------------------------------------------------------------------------
bool permutation::isLastLexSection(int start, int end)
{
	ThrowBadPermPos(start);
	ThrowBadPermPos(end);
	for (int i = start; i < end; i++) {
		if (perm[i] < perm[i+1])
			return false;
	}
	return true;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::operator[]
// ------------------------------------------------------------------------------------
int permutation::operator[](int i) const
{
	ThrowBadPermPos(i);
	return perm[i];
}


// ------------------------------------------------------------------------------------
//    permutation::operator=
// ------------------------------------------------------------------------------------
permutation& permutation::operator=(const permutation &pi)
{
	dim = pi.dim;
	resize(pi.getSize());
	for (int i = 0; i < dim; i++)
		perm[i] = pi.perm[i];
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::prefix
// ------------------------------------------------------------------------------------
permutation& permutation::prefix(int x)
{
	insert(0, x);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::putNextJT
// ------------------------------------------------------------------------------------
permutation& permutation::putNextJT(permutation &pi)
{
	if (&pi == NULL) {
		cerr << "permutation::putNextJT(pi): pi is NULL\n";
		exit(668);
	}
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::putNextLex
// ------------------------------------------------------------------------------------
permutation& permutation::putNextLex(permutation &pi)
{
	if (&pi == NULL) {
		cerr << "permutation::getnextlex(pi): pi is NULL\n";
		exit(668);
	}
	pi.dim = dim;
	for (int i = 0; i < dim; i++)
		pi.perm[i] = perm[i];
	pi.setNextLex();
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::putPrevLex
// ------------------------------------------------------------------------------------
permutation& permutation::putPrevLex(permutation &pi)
{
	cerr << "permutation::getprevlex(pi): cannot do this bit yet.\n";
	if (&pi == NULL) {
		cerr << "permutation::getprevlex(pi): permutation pi is NULL\n";
		exit(668);
	}
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::putSection
// ------------------------------------------------------------------------------------
permutation& permutation::putSection(int start, int end, permutation &pi)
{
	// put perm[start]...perm[end] into permutation pi
	if ((start < 0) || (start > end) || (end > dim)) {
		cerr << "permutation::section(start, end): invalid range (" << start
				<< ", " << end << "\n";
		exit(668);
	}
	if (&pi == NULL) {
		cerr << "permutation::putSection(start, end, pi): pi is NULL\n";
		exit(668);
	}
	for (int i = 0; i < end - start + 1; i++)
		pi.perm[i] = perm[i+start];
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::randomize
// ------------------------------------------------------------------------------------
permutation& permutation::randomize()
{
	int i, j, tempi;
	setFirstLex();
	for (i = 0; i < dim; i++) {
		j = MERandom::iRan(dim-i) + i;
		if ((j < i) || (j >= dim)) {
			cerr << "permutation::randomize(): random int j=" << j
					<< " is out of range [" << i << ", " << dim-1 << "\n";
			exit(668);
		}
		tempi = perm[j];
		perm[j] = perm[i];
		perm[i] = tempi;
	}
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::renew
// ------------------------------------------------------------------------------------
permutation& permutation::renew(int d)
{
	if ((d < 0) || (d >= size)) {
		cerr << "permutation::renew(d): int d=" << d 
				<< " is out of range [0, " << size-1 << "]\n";
		exit(668);
	}
	wipe();
	dim = d;
	setFirstLex();
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::resize
// ------------------------------------------------------------------------------------
permutation& permutation::resize(int d)
{
	delete perm;
//	perm = new int[d];	// deleted in ~permutation
//	if (perm == NULL) {
//		cerr << "permutation::resize(d): cannot recreate perm array\n";
//		exit(668);
//	}
	NewArray(perm, int, d);
	for (int i = 0; i < d; i++)
		perm[i] = 0;
	dim = d;
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::setBnBNextLex
// ------------------------------------------------------------------------------------
permutation& permutation::setBnBNextLex(int pos)
{
	ThrowBadPermPos(pos);
	while (isLastLexSection(0, pos)) {
		pos++;
		if (pos == dim)
			return *this;
	}
	setSectionNextLex(0, pos);
	if (pos < dim-1)
		setSectionFirstLex(pos+1, dim-1);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setIthElement
// ------------------------------------------------------------------------------------
permutation& permutation::setIthElement(int pos, int x)
{
	ThrowBadPermPos(pos);
	perm[pos] = x;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setNextLex
// ------------------------------------------------------------------------------------
permutation& permutation::setNextLex()
{
	int i, j, k, min;
	permutation *pi;// = new permutation(dim);	// deleted in this method
	NewClass(pi, permutation(dim));
	if (pi == NULL) {
		cerr << "permutation::setNextLex(): cannot create interim permutation pi\n";
		exit(668);
	}
	for (i = dim-1; i > 0; i--) {
		if (perm[i] > perm[i-1]) {
			min = perm[i];
			k = i;
			for (j = dim-1; j > i; j--) {
				if ((perm[j] < min) && (perm[j] > perm[i-1])) {
					k = j;
					min = perm[k];
				}
			}
			putSection(i-1, dim-1, *pi);
			pi->truncate(dim-i+1);
			truncate(i-1);
			suffix(min);
			pi->delElement(min);
			pi->sort(ascending);
			suffix(pi);
			return *this;
		}
	}
	if (i == 0)
		sort(ascending);
	delete pi;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setPrevLex
// ------------------------------------------------------------------------------------
permutation& permutation::setPrevLex()
{
	cerr << "permutation::setPrevLex:cannot do this bit yet.\n";
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setSection
// ------------------------------------------------------------------------------------
permutation& permutation::setSection(int start, int end, permutation &pi)
{
	ThrowBadPermPos(start);
	ThrowBadPermPos(end);
	for (int i = 0; i <= end-start; i++)
		perm[i+start] = pi.perm[i];
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setSectionFirstLex
// ------------------------------------------------------------------------------------
permutation& permutation::setSectionFirstLex(int start, int end)
{
	permutation tp(end-start);
	putSection(start, end, tp);
	tp.setFirstLex();
	setSection(start, end, tp);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setSectionLastLex
// ------------------------------------------------------------------------------------
permutation& permutation::setSectionLastLex(int start, int end)
{
	permutation tp(end-start);
	putSection(start, end, tp);
	tp.setLastLex();
	setSection(start, end, tp);
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::setSectionNextLex
// ------------------------------------------------------------------------------------
permutation& permutation::setSectionNextLex(int start, int end)
{
	permutation tp(end-start+1);
	putSection(start, end, tp);
	tp.setNextLex();
	setSection(start, end, tp);
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::
// ------------------------------------------------------------------------------------
permutation& permutation::sort(SortOrder direction, double *d)
{
	// set up the ShellSort: (see Algorithms, by Robert Sedgewick, Addison Wesley, pg 108)
	int 
		h = 1, 
		i, j, 
		jh,
		k = getSize(),
		tempi;
	do {
		h = 3*h + 1;
	} while (h <= k);
	if (d == NULL) {
		// just sort by the elements of the permutation:
		switch (direction) {
			case descending:
				do {
					h = (div(h, 3)).quot;	// see THINK Reference for definition of div
					for (i = h; i < k; i++) {
						for (j = i-h; perm[j+h] < perm[j]; j -= h) {
							jh = j + h;
							tempi = perm[j];
							perm[j] = perm[jh];
							perm[jh] = tempi;
							if (j < h)
								break;
						}
					}
				} while (h > 1);
				break;
			case ascending:
				do {
					h = (div(h, 3)).quot;	// see THINK Reference for definition of div
					for (i = h; i < k; i++) {
						for (j = i-h; perm[j+h] > perm[j]; j -= h) {
							jh = j + h;
							tempi = perm[j];
							perm[j] = perm[jh];
							perm[jh] = tempi;
							if (j < h)
								break;
						}
					}
				} while (h > 1);
				break;
			
			default:
				break;
		}
		return *this;
	}
	// else sort by the value of d[], indexed by this permutation.
	switch (direction) {
		case descending:
			do {
				h = (div(h, 3)).quot;	// see THINK Reference for definition of div
				for (i = h; i < k; i++) {
					for (j = i-h; d[perm[j+h]] > d[perm[j]]; j -= h) {
						jh = j + h;
						tempi = perm[j];
						perm[j] = perm[jh];
						perm[jh] = tempi;
						if (j < h)
							break;
					}
				}
			} while (h > 1);
			break;
		case ascending:
			do {
				h = (div(h, 3)).quot;	// see THINK Reference for definition of div
				for (i = h; i < k; i++) {
					for (j = i-h; d[perm[j+h]] < d[perm[j]]; j -= h) {
						jh = j + h;
						tempi = perm[j];
						perm[j] = perm[jh];
						perm[jh] = tempi;
						if (j < h)
							break;
					}
				}
			} while (h > 1);
			break;
		
		default:
			break;
	}
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::strmSelect
// ------------------------------------------------------------------------------------
void permutation::strmSelect(int start, int end, ostream &os)
{
	writeSelection(pmsg, start, end);
	os << pmsg;
}
void permutation::strmSelect(int pos, ostream &os)
{
	strmSelect(pos, pos, os);
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::suffix
// ------------------------------------------------------------------------------------
permutation& permutation::suffix(int x)
{
	if (dim > size-1) {
		cerr << "permutation::suffix(x): no more room in permutation\n";
		exit(668);
	}
	perm[dim++] = x;
	return *this;
}
permutation& permutation::suffix(permutation *addp)
{
	if (addp == NULL) {
		cerr << "permutation::suffix(addp): permutation addp is NULL\n";
		exit(668);
	}
	if (dim + addp->dim > size) {
		cerr << "permutation::suffix(permutation *addp): total permutation is too long\n";
		exit(668);
	}
	for (int i = 0; i < addp->dim; i++)
		perm[dim+i] = addp->perm[i];
	dim += addp->dim;
	return *this;
}
permutation& permutation::suffix(const MEBitset &b)
{
	int 
		i, 
		k = min(b.getCard(), dim-size);	
		// can't add more objects than there are in the MEBitset, 
		// or go off the end of the permutation.
	for (i = 0; i < k; i++)
		perm[size+i] = b[i];
	size += k;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::swap
// ------------------------------------------------------------------------------------
permutation& permutation::swap(int pos1, int pos2)
{
	if ((pos1 < 0) || (pos1 >= dim)){
		cerr << "permutation::swap(pos1, pos2): int pos1=" << pos1 
				<< " is out of range [0, " << dim-1 << "]\n";
		exit(668);
	}
	if ((pos2 < 0) || (pos2 >= dim)){
		cerr << "permutation::swap(pos1, pos2): int pos2=" << pos2
				<< " is out of range [0, " << dim-1 << "]\n";
		exit(668);
	}
	int tempi = perm[pos1];
	perm[pos1] = perm[pos2];
	perm[pos2] = tempi;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::swapRandomPair
// ------------------------------------------------------------------------------------
permutation& permutation::swapRandomPair()
{
	if (dim <= 1)
		return *this;
	int x, y;
	x = MERandom::iRan(dim);
	do {
		y = MERandom::iRan(dim);
	} while (x == y);
	swap(x, y);
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::truncate
// ------------------------------------------------------------------------------------
permutation& permutation::truncate(int y)
{	// truncate to y elements
	if ((y < 0) || (y > dim)) {
		cerr << "permutation::truncate(y): int y=" << y
				<< " is out of range [0, " << dim << "]\n";
		exit(668);
	}
	for (int i = y; i < dim; i++)
		perm[i] = 0;
	dim = y;
	return *this;
}


#pragma mark -


// ------------------------------------------------------------------------------------
//    permutation::wipe
// ------------------------------------------------------------------------------------
permutation& permutation::wipe()
{
	for (int i = 0; i < size; i++)
		perm[i] = 0;
	return *this;
}


// ------------------------------------------------------------------------------------
//    permutation::write
// ------------------------------------------------------------------------------------
char* permutation::write(char *str)
{
	char tempstr[8];
	sprintf(pmsg, "( ");
	for (int i = 0; i < dim; i++) {
		sprintf(tempstr, "%d ", perm[i]);
		if (strlen(pmsg) + 9 > 2048)
			break;
		strcat(pmsg, tempstr);
	}
	strcat(pmsg, ")");
	if (str)
		strcpy(str, pmsg);
	return pmsg;
}


// ------------------------------------------------------------------------------------
//    permutation::writeSelection
// ------------------------------------------------------------------------------------
char* permutation::writeSelection(char *str, int start, int end)
{
	ThrowBadPermPos(start);
	ThrowBadPermPos(end);
	char tempstr[8];
	sprintf(str, "( ");
	int i;
	for (i = 0; i < start; i++) {
		sprintf(tempstr, "%d ", perm[i]);
		strcat(str, tempstr);
	}
	strcat(str, "*");
	for (i = start; i < end; i++) {
		sprintf(tempstr, "%d ", perm[i]);
		strcat(str, tempstr);
	}
	strcat(str, "* ");
	for (i = end+1; i < dim; i++) {
		sprintf(tempstr, "%d ", perm[i]);
		strcat(str, tempstr);
	}
	strcat(str, ")");
	return str;
}


