/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPhylipParser.cpp

				MEPhylipParser Class
	
 ******************************************************************************/

// This class parses PHYLIP format sequence and tree files. Sequences come
// with the first line containing the number of taxa and the number of sites.
// The sequence names are contained in the first 10 characters. 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "CReader.h"
#include "MEParser.h"
#include "MEPhylipParser.h"
#include "MEBasicSequence.h"
#include "MEBasicAlignment.h"
#include "METreeList.h"
#include "METree.h"
#include "MEDataMatrix.h"
#include "MEDataMatrixList.h"


// ---------------------------------------------------------------------------------
//		 MEPhylipParser
// ---------------------------------------------------------------------------------

MEPhylipParser::MEPhylipParser(CReader *inReader) : MEParser(inReader)
{
	interleavedKnown=false;
	reader->setCommentDelimiters('\0', '\0', '#');
}

void MEPhylipParser::setIsInterleaved(bool inIsInterleaved)
{
	isInterleaved=inIsInterleaved;
	interleavedKnown=true;
}

void MEPhylipParser::setIsStrict(bool inIsStrict)
{
	isStrict  =inIsStrict;
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::detectInterleaved
// ---------------------------------------------------------------------------------

// Detects whether the phylip format is sequential or interleaved. Unfortunately
// PHYLIP format is not exact and as a result this is not fool proof and the user 
// should be allowed to override the detected interleaved/sequential option.

bool MEPhylipParser::detectInterleaved(int nTaxa)
{
	int i, j, k, len;
	char str[256];
	
	reader->markPosition();
	
	isInterleaved = true;
	isStrict = false;
	
	try {
		// Check first nTaxa lines are the same length
		for (i=0; i<nTaxa; i++) {
			do {
				reader->readLine(str, 255, false);
			} while (strlen(str)==0);
		
			if (i==0)
				len=strlen(str);
			else if (strlen(str)!=len || isspace(str[0]))
				isInterleaved = false;
			
			k = -1;
			for (j = 0; j < strlen(str); j++) {
				if (isspace(str[j]) && k < 0)
					k = j;
			}
			if (k < 0 && strlen(str)>=10)
				isStrict = true;
		}
		
		if (!isInterleaved)
			isStrict = false;

		if (len==255) // long lines - must be sequential
			isInterleaved = false;
		
		if (isInterleaved) {
			// Check that the next nTaxa lines are the same as the last
			for (i=0; i<nTaxa; i++) {
				do {
					reader->readLine(str, 255, false);
				} while (strlen(str)==0);
				 
				if (strlen(str)!=len)
					isInterleaved = false;
			}
		}
	} catch (CEndOfFile e) {
		isInterleaved = false;
	}
	
	reader->restorePosition();
	return isInterleaved;
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::readAlignment
// ---------------------------------------------------------------------------------

void MEPhylipParser::readAlignment(MEBasicAlignment *inAlignment)
{	
	int i, j, n, nTaxa, nSites;
	char name[MAX_NAME_LEN+1], *str;
	MEBasicSequence *seq;

	nTaxa=reader->readInteger();
	nSites=reader->readInteger();

	if (!interleavedKnown) // If the user has not set isInterleaved
		detectInterleaved(nTaxa);
	
	int maxNameLen = getMaxNameLen();
		
	if (isInterleaved) {
		int len, nRead=0;
		char line[256];

		int nameLen = 0;
		do {
			for (i=0; i<nTaxa; i++) {
				if (nRead == 0 || reader->skipSpace() <= nameLen) {
					if (isStrict) {
						reader->skipSpace();
						for (j = 0; j < 10; j++)
							name[j] = reader->readChar();
						j = 9;
						while (j > 0 && isspace(name[j]))
							j--;
						name[j+1] = '\0';
					} else
						reader->readToken(name, maxNameLen);
				}
					
				if (nRead == 0) {
					seq=inAlignment->newSequence(kUnknownSequence);
					seq->setName(name);
					if (strlen(name) > nameLen)
						nameLen = strlen(name);
					inAlignment->addSequence(seq);
				} else {
					seq=inAlignment->getSequence(i);
				}
				
				reader->skipSpace();
				readSequenceUntil(line, 255, "", "\r\n", &len);
				seq->appendSequence(line, strlen(line));
			}
			nRead+=len;
		} while (nRead<nSites);
	} else {
		for (i=0; i<nTaxa; i++) {
			if (isStrict) {
				reader->skipSpace();
				for (j = 0; j < 10; j++)
					name[j] = reader->readChar();
				j = 9;
				while (j > 0 && isspace(name[j]))
					j--;
				name[j+1] = '\0';
			} else
				reader->readToken(name, maxNameLen);
			seq=inAlignment->newSequence(kUnknownSequence);
			seq->setName(name);
			seq->setLength(nSites);
			inAlignment->addSequence(seq);
			
			str=seq->getSequence();
			n = readSequence(str, nSites, "");
			if (n != nSites)
				throw MEShortSequence("Bad sequence in imported file");
		}
	}
	
	seqType=inAlignment->inferSequenceType();
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::readTree
// ---------------------------------------------------------------------------------

void MEPhylipParser::readTree(METree *inTree)
{
	int numTrees;
	
	// Is there a count of the number of trees?
	reader->skipSpace();
	if (isdigit(reader->nextChar())) {
		char ch=reader->readInteger(&numTrees, "(");
		reader->unreadChar(ch);
	}
	
	readNewickTree(inTree);
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::readTreeList
// ---------------------------------------------------------------------------------

void MEPhylipParser::readTreeList(METreeList *inTreeList)
{
	METree *tree;
	int numTrees=0;
	char ch;
	
	// Is there a count of the number of trees?
	reader->skipSpace();
	if (isdigit(reader->nextChar())) {
		ch=reader->readInteger(&numTrees, " \t(");
		
		// Is there a second integer?
		if (ch == ' ' || ch == '\t') {
			reader->skipSpace();
			if (isdigit(reader->nextChar())) {
			// if so then this will be the number of tree and
			// the previous one would be the number of taxa (which
			// we will ignore
				ch=reader->readInteger(&numTrees, "(");
			}
		}
		
		reader->unreadChar(ch);
	}
	
	// If there are, read that many trees.
	if (numTrees>0) {
		try {
			for (int i=0; i<numTrees; i++) {
				tree=inTreeList->newTree();
				readNewickTree(tree);
				inTreeList->addTree(tree);
			}
		} catch (CFileError) {
			throw METreeMissing();
		}
	} else {
	// Otherwise read as many trees as possible.
		bool done=false;
		do {
			try {
				tree=inTreeList->newTree();
				readNewickTree(tree);
				inTreeList->addTree(tree);
			} catch (METreeMissing e) {
				done=true;
			}
		} while (!done);
	}
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::readDataMatrix
// ---------------------------------------------------------------------------------

void MEPhylipParser::readDataMatrix(MEDataMatrix *inDataMatrix)
{
	int i, j, nTaxa, nColumns;
	char name[MAX_NAME_LEN+1];
	double f;
	
	nTaxa=reader->readInteger();
	nColumns=reader->readInteger();
	
	inDataMatrix->setSize(nTaxa, nColumns);

	for (i=0; i<nTaxa; i++) {
		reader->readToken(name, getMaxNameLen());

		inDataMatrix->setRowName(i, name);
		
		for (j=0; j<nColumns; j++) {
			f=reader->readDouble();
			inDataMatrix->setElement(i, j, f);
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::readDataAndTree
// ---------------------------------------------------------------------------------

void MEPhylipParser::readDataAndTree(MEDataMatrix *inMatrix, METree *inTree)
{
	int numTrees;
	
	// Is there a count of the number of trees?
	reader->skipSpace();
	if (isdigit(reader->nextChar())) {
		char ch=reader->readInteger(&numTrees, "(");
		reader->unreadChar(ch);
	}
	
	readNewickTree(inTree, inMatrix);
}

// ---------------------------------------------------------------------------------
//		 MEPhylipParser::readDataAndTreeList
// ---------------------------------------------------------------------------------

void MEPhylipParser::readDataAndTreeList(MEDataMatrixList *inDataMatrices, METreeList *inTreeList)
{
	MEDataMatrix *matrix=NULL;
	METree *tree;
	int numTrees=0;
	
	// Is there a count of the number of trees?
	reader->skipSpace();
	if (isdigit(reader->nextChar())) {
		char ch=reader->readInteger(&numTrees, "(");
		reader->unreadChar(ch);
	}
	
	// If there are, read that many trees.
	if (numTrees>0) {
		try {
			for (int i=0; i<numTrees; i++) {
				tree=inTreeList->newTree();
				matrix=(MEDataMatrix *)inDataMatrices->newMatrix();
				readNewickTree(tree, matrix);
				inTreeList->addTree(tree);
				inDataMatrices->add(matrix, true);
			}
		} catch (CFileError) {
			throw METreeMissing();
		}
	} else {
	// Otherwise read as many trees as possible.
		bool done=false;
		do {
			try {
				tree=inTreeList->newTree();
				matrix=(MEDataMatrix *)inDataMatrices->newMatrix();
				readNewickTree(tree, matrix);
				inTreeList->addTree(tree);
				inDataMatrices->add(matrix, true);
			} catch (METreeMissing e) {
				done=true;
			}
		} while (!done);
	}
}

