/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPhylogeny.cpp

				MEPhylogeny Class
				
 ******************************************************************************/

#include <stdio.h>

#include "MEExceptions.h"

#include "MEPhylogeny.h"

#include "MEGlobal.h"
#include "MEMaths.h"

#include "METaxon.h"
#include "METaxonList.h"
#include "METree.h"

#pragma mark === MELineage ===

// =========================================================================
//	MELineage::MELineage
// =========================================================================

MELineage::MELineage()
{
	mAncestor = -1;
	mDescendant = -1;
	mSibling = -1;
	
	mBirthTime = -1.0;
	mDeathTime = -1.0;
	
	mFlag = 0;
}

#pragma mark === MEPhylogeny ===

// =========================================================================
//	MEPhylogeny::mCapacityIncrement
// =========================================================================

int MEPhylogeny::mCapacityIncrement = 1000;

// =========================================================================
//	MEPhylogeny::MEPhylogeny
// =========================================================================

MEPhylogeny::MEPhylogeny() : MEObject()
{
	mLineages = NULL;
	mNumLineages = 0;
	mNumAllocatedLineages = 0;
	mLineageCapacity = 0;
	
	mExtantLineages = NULL;
	mNumExtantLineages = 0;
	mExtantCapacity = 0;
	
	mTipLineages = NULL;
	mNumTipLineages = 0;
	mTipCapacity = 0;
	
	mAvailableLineages = NULL;
	mNumAvailableLineages = 0;
	mAvailableCapacity = 0;
}

// =========================================================================
//	MEPhylogeny::~MEPhylogeny
// =========================================================================

MEPhylogeny::~MEPhylogeny()
{
	if (mLineages != NULL) {
		for (int i = 0; i < mNumAllocatedLineages; i++) 
			if (mLineages[i] != NULL)
				delete mLineages[i];
		delete [] mLineages;
	}
	
	if (mExtantLineages != NULL)
		delete [] mExtantLineages;
		
	if (mTipLineages != NULL)
		delete [] mTipLineages;
		
	if (mAvailableLineages != NULL)
		delete [] mAvailableLineages;
}

// =========================================================================
//	MEPhylogeny::initialize
// =========================================================================

void MEPhylogeny::initialize()
{
	if (mNumAllocatedLineages > 0) {
	// If lineages exist, put them on the available list.
	
		checkAvailableCapacity(mNumAllocatedLineages);
		
		// put them in reverse order so that lineage 0  comes off first
		for (int i = 0; i < mNumAllocatedLineages; i++)
			mAvailableLineages[i] = mNumAllocatedLineages - i - 1;
			
		mNumAvailableLineages = mNumAllocatedLineages;
		mNumLineages = 0;
	}
	
	// Create the first lineage - always lineage 0, the root
	long d = createLineage();
	ThrowIf(d != 0);

	checkExtantCapacity(1);
	mExtantLineages[0] = d;
	mNumExtantLineages = 1;

	checkTipCapacity(1);
	mTipLineages[0] = d;
	mNumTipLineages = 1;
}

// =========================================================================
//	MEPhylogeny::getMemoryUsage
// =========================================================================

int MEPhylogeny::getMemoryUsage()
{
	int sum = mNumAllocatedLineages * sizeof(MELineage);
	sum += mLineageCapacity * sizeof(MELineage*);
	sum += mExtantCapacity * sizeof(long);
	sum += mTipCapacity * sizeof(long);
	sum += mAvailableCapacity * sizeof(long);
	
	return sum;
}

#pragma mark === Phylogeny Sampling ===

// =========================================================================
//	MEPhylogeny::getSampledPhylogeny
// =========================================================================

void MEPhylogeny::getSampledPhylogeny(MEPhylogeny* ioPhylogeny)
{
	ThrowIfNull(ioPhylogeny);
		
	// Traverse from the root, recreating the flagged lineages...
	double time = getBirthTime(0);
	
	ioPhylogeny->initialize();
	ioPhylogeny->setBirthTime(0, time);

	ioPhylogeny->removeExtantLineage(0);
	
	getSampledNode(ioPhylogeny, 0, 0, time);
}

// =========================================================================
//	MEPhylogeny::getSampledNode
// =========================================================================

void MEPhylogeny::getSampledNode(MEPhylogeny* ioPhylogeny, long inNewNode, long inNode, double inTime)
{
	ThrowIf(inNode == -1);
	
	long lineage;
	long descendant = getDescendant(inNode);
	double time = getDeathTime(inNode);		
	
	if (descendant == -1) {
		// Node is a tip
		
		ioPhylogeny->setDescendant(inNewNode, -1);
		ioPhylogeny->setSibling(inNewNode, -1);
		ioPhylogeny->setBirthTime(inNewNode, inTime);
		ioPhylogeny->setDeathTime(inNewNode, time);
		
		if (getIsExtant(inNode))
			ioPhylogeny->addExtantLineage(inNewNode);
		
	} else {
		short flag = getFlag(inNode);
		
		ThrowIf( flag == 0 );
		
		if (flag > 1) {
			// Node has more than 1 extant descendant
			
			long s1 = -1;
			long s2 = -1;
			
			if (getFlag(descendant) > 0) {
				lineage = ioPhylogeny->createLineage();
			
				getSampledNode(ioPhylogeny, lineage, descendant, time);
				
				ioPhylogeny->setDescendant(inNewNode, lineage);
				ioPhylogeny->setBirthTime(inNewNode, inTime);
				ioPhylogeny->setDeathTime(inNewNode, time);
				
				ioPhylogeny->setAncestor(lineage, inNewNode);

				s1 = lineage;
			} else
				lineage = -1;
			
			long sibling = getSibling(descendant);
			
			while (sibling != descendant) {
				if (getFlag(sibling) > 0) {
				
					s2 = ioPhylogeny->createLineage();
					ioPhylogeny->setBirthTime(s2, inTime);
					ioPhylogeny->setDeathTime(s2, time);
			
					getSampledNode(ioPhylogeny, s2, sibling, time);
				
					if (s1 != -1)
						ioPhylogeny->setSibling(s1, s2);
					
					ioPhylogeny->setAncestor(s2, inNewNode);

					s1 = s2;

					if (lineage == -1) {
						lineage = s2;
						ioPhylogeny->setDescendant(inNewNode, lineage);
						ioPhylogeny->setBirthTime(inNewNode, inTime);
						ioPhylogeny->setDeathTime(inNewNode, time);
					}
				}
				
				sibling = getSibling(sibling);
			}
			
			ioPhylogeny->setSibling(s1, lineage);
			
		} else {
			// Node has only 1 extant descendant
						
			if (getFlag(descendant) == 0) {
			
				long sibling = getSibling(descendant);
				
				while (sibling != descendant && getFlag(sibling) == 0) {

					sibling = getSibling(sibling);
					
				}
				ThrowIf(sibling == descendant);
				
				getSampledNode(ioPhylogeny, inNewNode, sibling, inTime);
			} else {
				getSampledNode(ioPhylogeny, inNewNode, descendant, inTime);
			}
		}	
	}
}

#pragma mark === Lineage Access ===

// =========================================================================
//	MEPhylogeny::getLineage
// =========================================================================

MELineage* MEPhylogeny::getLineage(long inLineageIndex)
{
	return mLineages[inLineageIndex];
}

// =========================================================================
//	MEPhylogeny::setLineage
// =========================================================================

void MEPhylogeny::setLineage(	long inLineageIndex,
								long inAncestor,
								long inDescendant,
								long inSibling,
								double inBirthTime,
								double inDeathTime)
{
	mLineages[inLineageIndex]->mAncestor = inAncestor;
	mLineages[inLineageIndex]->mDescendant = inDescendant;
	mLineages[inLineageIndex]->mSibling = inSibling;
	mLineages[inLineageIndex]->mBirthTime = inBirthTime;
	mLineages[inLineageIndex]->mDeathTime = inDeathTime;
}

// =========================================================================
//	MEPhylogeny::setAncestor
// =========================================================================

void MEPhylogeny::setAncestor(long inLineageIndex, long inAncestor)
{
	mLineages[inLineageIndex]->mAncestor = inAncestor;
}

// =========================================================================
//	MEPhylogeny::getAncestor
// =========================================================================

long MEPhylogeny::getAncestor(long inLineageIndex)
{
	return mLineages[inLineageIndex]->mAncestor;
}

// =========================================================================
//	MEPhylogeny::setDescendant
// =========================================================================

void MEPhylogeny::setDescendant(long inLineageIndex, long inDescendant)
{
	mLineages[inLineageIndex]->mDescendant = inDescendant;
}

// =========================================================================
//	MEPhylogeny::getDescendant
// =========================================================================

long MEPhylogeny::getDescendant(long inLineageIndex)
{
	return mLineages[inLineageIndex]->mDescendant;
}

// =========================================================================
//	MEPhylogeny::setSibling
// =========================================================================

void MEPhylogeny::setSibling(long inLineageIndex, long inSibling)
{
	mLineages[inLineageIndex]->mSibling = inSibling;
}

// =========================================================================
//	MEPhylogeny::getSibling
// =========================================================================

long MEPhylogeny::getSibling(long inLineageIndex)
{
	return mLineages[inLineageIndex]->mSibling;
}

// =========================================================================
//	MEPhylogeny::setBirthTime
// =========================================================================

void MEPhylogeny::setBirthTime(long inLineageIndex, double inBirthTime)
{
	mLineages[inLineageIndex]->mBirthTime = inBirthTime;
}

// =========================================================================
//	MEPhylogeny::getBirthTime
// =========================================================================

double MEPhylogeny::getBirthTime(long inLineageIndex)
{
	return mLineages[inLineageIndex]->mBirthTime;
}

// =========================================================================
//	MEPhylogeny::setDeathTime
// =========================================================================

void MEPhylogeny::setDeathTime(long inLineageIndex, double inDeathTime)
{
	mLineages[inLineageIndex]->mDeathTime = inDeathTime;
}

// =========================================================================
//	MEPhylogeny::getDeathTime
// =========================================================================

double MEPhylogeny::getDeathTime(long inLineageIndex)
{
	return mLineages[inLineageIndex]->mDeathTime;
}

// =========================================================================
//	MEPhylogeny::setFlag
// =========================================================================

void MEPhylogeny::setFlag(long inLineageIndex, short inFlag)
{
	mLineages[inLineageIndex]->mFlag = inFlag;
}

// =========================================================================
//	MEPhylogeny::getFlag
// =========================================================================

short MEPhylogeny::getFlag(long inLineageIndex)
{
	return mLineages[inLineageIndex]->mFlag;
}

// =========================================================================
//	MEPhylogeny::setAllFlags
// =========================================================================

void MEPhylogeny::setAllFlags(short inFlag)
{
	for (int i = 0; i < mNumLineages; i++) {
		mLineages[i]->mFlag = inFlag;
	}
}

// =========================================================================
//	MEPhylogeny::getExtantLineage
// =========================================================================

long MEPhylogeny::getExtantLineage(long inExtantIndex)
{
	return mExtantLineages[inExtantIndex];
}

// =========================================================================
//	MEPhylogeny::getIsExtant
// =========================================================================
// This is brute force - if used alot it may be worth having a flag in the 
// lineage.

bool MEPhylogeny::getIsExtant(long inLineageIndex)
{
	bool found = false;
	int i = 0;
	
	while (i < mNumExtantLineages && !found) {
		found = mExtantLineages[i] == inLineageIndex;
		i++;
	}
	
	return found;
}

// =========================================================================
//	MEPhylogeny::addExtantLineage
// =========================================================================

void MEPhylogeny::addExtantLineage(long inLineageIndex)
{
	checkExtantCapacity(mNumExtantLineages + 1);
	mExtantLineages[mNumExtantLineages] = inLineageIndex;
	mNumExtantLineages++;
}

// =========================================================================
//	MEPhylogeny::setExtantLineage
// =========================================================================

void MEPhylogeny::setExtantLineage(long inExtantIndex, long inLineageIndex)
{
	mExtantLineages[inExtantIndex] = inLineageIndex;
}

// =========================================================================
//	MEPhylogeny::removeExtantLineage
// =========================================================================

void MEPhylogeny::removeExtantLineage(long inExtantIndex)
{
	mExtantLineages[inExtantIndex] = mExtantLineages[mNumExtantLineages - 1];
	mNumExtantLineages--;
}

// =========================================================================
//	MEPhylogeny::getRandomExtantIndex
// =========================================================================

long MEPhylogeny::getRandomExtantIndex()
{
	return MERandom::iRan(mNumExtantLineages);
}


// =========================================================================
//	MEPhylogeny::getTipLineage
// =========================================================================

long MEPhylogeny::getTipLineage(long inTipIndex)
{
	return mTipLineages[inTipIndex];
}

// =========================================================================
//	MEPhylogeny::getIsTip
// =========================================================================
// This is brute force - if used alot it may be worth having a flag in the 
// lineage.

bool MEPhylogeny::getIsTip(long inLineageIndex)
{
	bool found = false;
	int i = 0;
	
	while (i < mNumTipLineages && !found) {
		found = mTipLineages[i] == inLineageIndex;
		i++;
	}
	
	return found;
}

// =========================================================================
//	MEPhylogeny::addTipLineage
// =========================================================================

void MEPhylogeny::addTipLineage(long inLineageIndex)
{
	checkTipCapacity(mNumTipLineages + 1);
	mTipLineages[mNumTipLineages] = inLineageIndex;
	mNumTipLineages++;
}

// =========================================================================
//	MEPhylogeny::setTipLineage
// =========================================================================

void MEPhylogeny::setTipLineage(long inTipIndex, long inLineageIndex)
{
	mTipLineages[inTipIndex] = inLineageIndex;
}

// =========================================================================
//	MEPhylogeny::removeTipLineage
// =========================================================================

void MEPhylogeny::removeTipLineage(long inTipIndex)
{
	mTipLineages[inTipIndex] = mTipLineages[mNumTipLineages - 1];
	mNumTipLineages--;
}

// =========================================================================
//	MEPhylogeny::getRandomTipIndex
// =========================================================================

long MEPhylogeny::getRandomTipIndex()
{
	return MERandom::iRan(mNumTipLineages);
}


// =========================================================================
//	MEPhylogeny::createLineage
// =========================================================================

long MEPhylogeny::createLineage()
{
	long index;
	
	if (mNumAvailableLineages > 0) {
		index = mAvailableLineages[mNumAvailableLineages - 1];
		mNumAvailableLineages--;
	} else {
		index = mNumAllocatedLineages;
		
		checkLineageCapacity(mNumAllocatedLineages+1);
		mNumAllocatedLineages++;
		
		mLineages[index] = newLineage();
	}
	
	mNumLineages++;
	
	initializeLineage(index);

	return index;
}

// =========================================================================
//	MEPhylogeny::initializeLineage
// =========================================================================

void MEPhylogeny::initializeLineage(long inLineageIndex)
{
	setAncestor(inLineageIndex, -1);
	setDescendant(inLineageIndex, -1);
	setSibling(inLineageIndex, -1);
	setBirthTime(inLineageIndex, -1.0);
	setDeathTime(inLineageIndex, -1.0);
	setFlag(inLineageIndex, 0);
}

// =========================================================================
//	MEPhylogeny::pruneLineage
// =========================================================================
// Prune a lineage out removing all redundant nodes.

void MEPhylogeny::pruneLineage(long inLineageIndex)
{
	// **** THIS NEEDS REVISION for mDescendant ****

	
	long ancestor = mLineages[inLineageIndex]->mAncestor;

	// Find the sibling that points to the extinct lineage...
	long sibling = mLineages[inLineageIndex]->mSibling;
	
	while (mLineages[sibling]->mSibling != inLineageIndex) {
		sibling = mLineages[sibling]->mSibling;
	}

	// Remove extinct lineage from sibling chain...
	mLineages[sibling]->mSibling = mLineages[inLineageIndex]->mSibling;
	
	if (mLineages[sibling]->mSibling == sibling) { 
		// sibling is the only one left in the chain so we must
		// add it into the ancestoral sibling chain. 
		
		// Find the sibling that points to the ancestor...
		long sibling2 = mLineages[ancestor]->mSibling;
		
		while (mLineages[sibling2]->mSibling != ancestor) {
			sibling2 = mLineages[sibling2]->mSibling;
		}
		
		// Put the sibling into the ancestor's sibling chain...
		mLineages[sibling]->mSibling = mLineages[ancestor]->mSibling;
		mLineages[sibling]->mAncestor = mLineages[ancestor]->mAncestor;
		mLineages[sibling2]->mSibling = sibling;
		
		// Dispose of the ancestor lineage - makes the lineage available
		// for future use.
		disposeLineage(ancestor);
	} 
	
	// Dispose of the extinct lineage - makes the lineage available
	// for future use.
	disposeLineage(inLineageIndex);
}
		
// =========================================================================
//	MEPhylogeny::disposeLineage
// =========================================================================

void MEPhylogeny::disposeLineage(long inLineageIndex)
{
	checkAvailableCapacity(mNumAvailableLineages + 1);
	
	mAvailableLineages[mNumAvailableLineages] = inLineageIndex;
	mNumAvailableLineages++;

	// Set birth time to -1.0 to flag that it is not being used
	// This is not necessary but helps debugging...
	mLineages[inLineageIndex]->mBirthTime = -1.0;

	mNumLineages--;
}

// =========================================================================
//	MEPhylogeny::getNumTips
// =========================================================================

long MEPhylogeny::getNumTips(long inLineageIndex)
{
	long numTips = 1;
	long descendant, sibling;
	
	descendant = getDescendant(inLineageIndex);
	if (descendant != -1) { // not a tip
		numTips = getNumTips(descendant);

		sibling = getSibling(descendant);
		while (sibling != descendant) {
			numTips += getNumTips(sibling);
			sibling = getSibling(sibling);
		}
	}
	
	return numTips;
}

// =========================================================================
//	MEPhylogeny::newLineage
// =========================================================================

MELineage* MEPhylogeny::newLineage()
{
	return new MELineage();
}

// =========================================================================
//	MEPhylogeny::checkLineageCapacity
// =========================================================================

void MEPhylogeny::checkLineageCapacity(long inCapacity)
{
	if (inCapacity > mLineageCapacity) {
		while (inCapacity > mLineageCapacity)
			mLineageCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		MELineage** newArray = new MELineage*[mLineageCapacity];
		
		// Copy existing trees into it
		for (i = 0; i < mNumAllocatedLineages; i++)
			newArray[i] = mLineages[i];
			
		// Delete the old array
		delete [] mLineages;
		
		// And copy in the new one
		mLineages = newArray;
	}
}

// =========================================================================
//	MEPhylogeny::checkExtantCapacity
// =========================================================================

void MEPhylogeny::checkExtantCapacity(long inCapacity)
{
	if (inCapacity > mExtantCapacity) {
		while (inCapacity > mExtantCapacity)
			mExtantCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		long *newArray = new long[mExtantCapacity];
		
		// Copy existing trees into it
		for (i = 0; i < mNumExtantLineages; i++)
			newArray[i] = mExtantLineages[i];
			
		// Delete the old array
		delete [] mExtantLineages;
		
		// And copy in the new one
		mExtantLineages = newArray;
	}
}

// =========================================================================
//	MEPhylogeny::checkTipCapacity
// =========================================================================

void MEPhylogeny::checkTipCapacity(long inCapacity)
{
	if (inCapacity > mTipCapacity) {
		while (inCapacity > mTipCapacity)
			mTipCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		long *newArray = new long[mTipCapacity];
		
		// Copy existing lineages into it
		for (i = 0; i < mNumTipLineages; i++)
			newArray[i] = mTipLineages[i];
			
		// Delete the old array
		delete [] mTipLineages;
		
		// And copy in the new one
		mTipLineages = newArray;
	}
}

// =========================================================================
//	MEPhylogeny::checkAvailableCapacity
// =========================================================================

void MEPhylogeny::checkAvailableCapacity(long inCapacity)
{
	if (inCapacity > mAvailableCapacity) {
		while (inCapacity > mAvailableCapacity)
			mAvailableCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		long *newArray = new long[mAvailableCapacity];
		
		// Copy existing trees into it
		for (i = 0; i < mNumAvailableLineages; i++)
			newArray[i] = mAvailableLineages[i];
			
		// Delete the old array
		delete [] mAvailableLineages;
		
		// And copy in the new one
		mAvailableLineages = newArray;
	}
}

// =========================================================================
//	MEPhylogeny::getHeight
// =========================================================================

double MEPhylogeny::getHeight()
{
	return getDeathTime(getExtantLineage(0)) - getDeathTime(0);
}

// =========================================================================
//	MEPhylogeny::getHeight
// =========================================================================

double MEPhylogeny::getHeight(long inLineageIndex)
{
	return getDeathTime(getExtantLineage(0)) - getDeathTime(inLineageIndex);
}

// =========================================================================
//	MEPhylogeny::getBalance
// =========================================================================

double MEPhylogeny::getBalance()
{	
	int i, n = getNumLineages();
	double sum = 0.0;
	for (i = 0; i < n; i++)
		sum += calcBalanceForLineage(i);
		
	return sum;
}

// =========================================================================
//	MEPhylogeny::calcBalanceForLineage
// =========================================================================

double MEPhylogeny::calcBalanceForLineage(long inLineageIndex)
{
	long descendant = getDescendant(inLineageIndex);
	long l1, l2, tmp, sibling;
	double balance;
	
	if (descendant != -1) { // not a tip
		l1 = getNumTips(descendant);
		
		sibling = getSibling(descendant);
		l2 = getNumTips(sibling);
		
		if (l1 < l2) {
			tmp = l2;
			l2 = l1;
			l1 = tmp;
		}

// Slowinski
//		balance = (2.0 * l2) / (l1 + l2 - 1);

// Fusco & Cronk
		long sum = l1 + l2;
		long m1 = (sum - 1);
		long m2 = (sum / 2) + (sum % 2);
		balance = ((double)l1 - m2) / (m1 - m2);
	}
	
	return balance;
}	

#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 MEPhylogeny::readObject
// ---------------------------------------------------------------------------------

// Read the fields of this object from stream.
void MEPhylogeny::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}

// ---------------------------------------------------------------------------------
//		 MEPhylogeny::readData
// ---------------------------------------------------------------------------------

// Read the data of this object (i.e. the tree).
void MEPhylogeny::readData(MEStream *stream)
{
//	stream->getTree(this);
}

// ---------------------------------------------------------------------------------
//		 MEPhylogeny::writeObject
// ---------------------------------------------------------------------------------

// Write the fields of this object to stream.
void MEPhylogeny::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}


// ---------------------------------------------------------------------------------
//		 MEPhylogeny::writeData
// ---------------------------------------------------------------------------------

// Write the data of this object (i.e. the tree).
void MEPhylogeny::writeData(MEStream *stream)
{
//	stream->putTree(this);
}
