/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPhylogeny.h

		Interface for the MEPhylogeny Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Trees"
 
#ifndef _MEPHYLOGENY_H_
#define _MEPHYLOGENY_H_

#include "MEObject.h"

// =========================================================================
//	class MELineage
// =========================================================================

class MELineage {
	friend class MEPhylogeny;

	protected:
		long				mAncestor;
		long				mDescendant;
		long				mSibling;
		
		float				mBirthTime;
		float				mDeathTime;
		
		short				mFlag;
			
	public:
		
							MELineage();
};

// =========================================================================
//	class MEPhylogeny
// =========================================================================

class MEPhylogeny : public MEObject {
	public:
		virtual long		getID() const { return 'MPhy'; }
		//: The ID of this object - 'MPhy'.
		
		virtual char*		getClassName() const { return "Phylogeny"; }
		//: Returns the name of this class - "Phylogeny"
		
		static int			mCapacityIncrement;

	protected:
		MELineage			**mLineages;
		long				mNumLineages;
		long				mNumAllocatedLineages;
		long				mLineageCapacity;
		
		long				*mExtantLineages;
		long				mNumExtantLineages;
		long				mExtantCapacity;
		
		long				*mTipLineages;
		long				mNumTipLineages;
		long				mTipCapacity;
		
		long				*mAvailableLineages;
		long				mNumAvailableLineages;
		long				mAvailableCapacity;			
			
	public:
		
							MEPhylogeny();
		virtual				~MEPhylogeny();
			
		virtual void		initialize();
		
		virtual int			getMemoryUsage();

		long				getNumLineages() { return mNumLineages; }
		long				getNumExtantLineages() { return mNumExtantLineages; }
		long				getNumTipLineages() { return mNumTipLineages; }
		
		virtual void		getSampledPhylogeny(MEPhylogeny* ioPhylogeny);

	protected:	
		virtual void		getSampledNode(MEPhylogeny* ioPhylogeny, long inNewNode, long inNode, double inTime);

	public:
		MELineage*			getLineage(	long inLineageIndex);
		void				setLineage(	long inLineageIndex,
										long inAncestor,
										long inDescendant,
										long inSibling,
										double inBirthTime,
										double inDeathTime);
										
		void				setAncestor(long inLineageIndex, long inAncestor);
		long				getAncestor(long inLineageIndex);

		void				setDescendant(long inLineageIndex, long inDescendant);
		long				getDescendant(long inLineageIndex);
		
		void				setSibling(long inLineageIndex, long inSibling);
		long				getSibling(long inLineageIndex);
		
		void				setBirthTime(long inLineageIndex, double inBirthTime);
		double				getBirthTime(long inLineageIndex);

		void				setDeathTime(long inLineageIndex, double inDeathTime);
		double				getDeathTime(long inLineageIndex);

		void				setFlag(long inLineageIndex, short inFlag);
		short				getFlag(long inLineageIndex);
		void 				setAllFlags(short inFlag);
		
		long				getExtantLineage(long inExtantIndex);
		bool				getIsExtant(long inLineageIndex);
		void				addExtantLineage(long inLineageIndex);
		void				setExtantLineage(long inExtantIndex, long inLineageIndex);
		void				removeExtantLineage(long inExtantIndex);
		
		long				getRandomExtantIndex();

		long				getTipLineage(long inTipIndex);
		bool				getIsTip(long inLineageIndex);
		void				addTipLineage(long inLineageIndex);
		void				setTipLineage(long inTipIndex, long inLineageIndex);
		void				removeTipLineage(long inTipIndex);
		
		long				getRandomTipIndex();

		long				createLineage();
		virtual void		initializeLineage(long inLineageIndex);
		void				pruneLineage(long inLineageIndex);
		void				disposeLineage(long inLineageIndex);
		
		long				getNumTips(long inLineageIndex);
		
		double				getHeight();
		double				getHeight(long inLineageIndex);
		double				getBalance();
		
		virtual double		calcBalanceForLineage(long inLineageIndex);

	protected:	
		virtual MELineage*	newLineage();
		
		void				checkLineageCapacity(long inCapacity);
		void				checkExtantCapacity(long inCapacity);
		void				checkTipCapacity(long inCapacity);
		void				checkAvailableCapacity(long inCapacity);

#pragma mark === Streaming ===
		virtual void		readObject(MEStream *stream);
		virtual void		readData(MEStream *stream);
		virtual void		writeObject(MEStream *stream);
		virtual void		writeData(MEStream *stream);
};

#endif /* _MEPHYLOGENY_H_ */
