/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEPhylogenySimulator.cpp

				MEPhylogenySimulator Class
				
 ******************************************************************************/

#include <stdio.h>

#include "CExceptions.h"
#include "CIntArray.h"
#include "CDoubleArray.h"

#include "MEPhylogeny.h"
#include "MEPhylogenySimulator.h"

#include "MEGlobal.h"
#include "MEMaths.h"

#include "METaxon.h"
#include "METaxonList.h"
#include "METree.h"

#pragma mark === MEPhylogenySimulator ===

// =========================================================================
//	MEPhylogenySimulator::mCapacityIncrement
// =========================================================================

int MEPhylogenySimulator::mCapacityIncrement = 100;

// =========================================================================
//	MEPhylogenySimulator::MEPhylogenySimulator
// =========================================================================

MEPhylogenySimulator::MEPhylogenySimulator()
{
	mPhylogeny = NULL;

	mGenotypes = NULL;
	mNumGenotypes = 0;
	mGenotypeCapacity = 0;

	mAvailableGenotypes = NULL;
	mNumAvailableGenotypes = 0;
	mAvailableGenotypeCapacity = 0;
	
	mPruneExtinctLineages = false;
	
	mBirthOrder = new CIntArray(1000, 1000);
	
	mCurrentTime = 0.0;
}

// =========================================================================
//	MEPhylogenySimulator::~MEPhylogenySimulator
// =========================================================================

MEPhylogenySimulator::~MEPhylogenySimulator()
{
	if (mPhylogeny != NULL)
		delete mPhylogeny;
		
	if (mGenotypes != NULL) {
		for (int i = 0; i < mNumGenotypes; i++) 
			if (mGenotypes[i] != NULL)
				delete mGenotypes[i];
		delete [] mGenotypes;
	}
	
	if (mAvailableGenotypes != NULL) {
		delete [] mAvailableGenotypes;
	}
	
	delete mBirthOrder;
}
		
// =========================================================================
//	MEPhylogenySimulator::initialize
// =========================================================================

void MEPhylogenySimulator::initialize()
{
	if (mPhylogeny == NULL)
		mPhylogeny = newPhylogeny();
		
	mPhylogeny->initialize();
	
	mBirthOrder->removeAll();
	
	mCurrentTime = 0.0;
}

// =========================================================================
//	MEPhylogenySimulator::setPhylogeny
// =========================================================================

void MEPhylogenySimulator::setPhylogeny(MEPhylogeny* inPhylogeny)
{
	if (mPhylogeny != NULL)
		delete mPhylogeny;
		
	mPhylogeny = inPhylogeny;
}

// =========================================================================
//	MEPhylogenySimulator::getWholePhylogeny
// =========================================================================

MEPhylogeny* MEPhylogenySimulator::getWholePhylogeny(bool inExtantOnly)
{
	MEPhylogeny *phylogeny = new MEPhylogeny();
	getWholePhylogeny(phylogeny, inExtantOnly);
	
	return phylogeny;
}

// =========================================================================
//	MEPhylogenySimulator::getWholePhylogeny
// =========================================================================

void MEPhylogenySimulator::getWholePhylogeny(MEPhylogeny* ioPhylogeny, bool inExtantOnly)
{
	ThrowIfNull(mPhylogeny);
	ThrowIfNull(ioPhylogeny);
	
	mPhylogeny->setAllFlags(0);

	long i, lineage;
	long n = mPhylogeny->getNumLineages();
	
	for (i = 0; i < n; i++) {
		if (mPhylogeny->getDescendant(i) == -1)
			mPhylogeny->setFlag(i, 1);
	}

	setSampleFromFlags();
	
	mPhylogeny->getSampledPhylogeny(ioPhylogeny);
}

// =========================================================================
//	MEPhylogenySimulator::getSamplePhylogeny
// =========================================================================

MEPhylogeny* MEPhylogenySimulator::getSamplePhylogeny(long inNumLineages, bool inExtantOnly)
{
	MEPhylogeny *phylogeny = new MEPhylogeny();
	getSamplePhylogeny(phylogeny, inNumLineages, inExtantOnly);
	
	return phylogeny;
}

// =========================================================================
//	MEPhylogenySimulator::getSamplePhylogeny
// =========================================================================

void MEPhylogenySimulator::getSamplePhylogeny(MEPhylogeny* ioPhylogeny, long inNumLineages, bool inExtantOnly)
{
	ThrowIfNull(mPhylogeny);
	ThrowIfNull(ioPhylogeny);
	
	long i, extant;
	long sampleSize = inNumLineages;
	long n = mPhylogeny->getNumExtantLineages();
	
	if (sampleSize > n)
		sampleSize = n;
	
	mPhylogeny->setAllFlags(0);

	// First flag the tips we want
	if (sampleSize == n) {
		for (i = 0; i < sampleSize; i++) {
			extant = mPhylogeny->getExtantLineage(i);
			mPhylogeny->setFlag(extant, 1);
		}
	} else {
		for (i = 0; i < sampleSize; i++) {
			extant = mPhylogeny->getExtantLineage(mPhylogeny->getRandomExtantIndex());
			
			while (mPhylogeny->getFlag(extant) != 0) {
				extant = mPhylogeny->getExtantLineage(mPhylogeny->getRandomExtantIndex());
			}
				
			mPhylogeny->setFlag(extant, 1);
		}
	}
	
	setSampleFromFlags();
	
	mPhylogeny->getSampledPhylogeny(ioPhylogeny);
}

// =========================================================================
//	MEPhylogenySimulator::getSamplePhylogeny
// =========================================================================

MEPhylogeny* MEPhylogenySimulator::getSamplePhylogeny(double inProbability, bool inExtantOnly)
{
	MEPhylogeny *phylogeny = new MEPhylogeny();
	getSamplePhylogeny(phylogeny, inProbability, inExtantOnly);
	
	return phylogeny;
}

// =========================================================================
//	MEPhylogenySimulator::getSamplePhylogeny
// =========================================================================

void MEPhylogenySimulator::getSamplePhylogeny(MEPhylogeny* ioPhylogeny, double inProbability, bool inExtantOnly)
{
	ThrowIfNull(mPhylogeny);
	ThrowIfNull(ioPhylogeny);
	
	long i, extant;
	long n = mPhylogeny->getNumExtantLineages();
		
	mPhylogeny->setAllFlags(0);

	for (i = 0; i < n; i++) {
		if (MERandom::bRan(inProbability)) {
			extant = mPhylogeny->getExtantLineage(i);
					
			mPhylogeny->setFlag(extant, 1);
		}
	}
	
	setSampleFromFlags();
	
	mPhylogeny->getSampledPhylogeny(ioPhylogeny);
}

// =========================================================================
//	MEPhylogenySimulator::setSampleFromFlags
// =========================================================================

void MEPhylogenySimulator::setSampleFromFlags()
{
	ThrowIfNull(mPhylogeny);
	
	long i, lineage;
	long n = mPhylogeny->getNumLineages();
	
	// Then work up the tree incrementing flags. We stop
	// when a flag is > 1 which means that at least two
	// selected tips as descended from that node...
	short flag;
	for (i = 0; i < n; i++) {
		if (mPhylogeny->getDescendant(i)==-1) {
			lineage = i;
			flag = mPhylogeny->getFlag(lineage);
			
			if (flag == 1) {
				do {
					lineage = mPhylogeny->getAncestor(lineage);
					if (lineage != -1) {
						flag = mPhylogeny->getFlag(lineage);
						flag++;
						mPhylogeny->setFlag(lineage, flag);
					}
				} while (lineage != -1 && flag < 2);
			}
		}
	}
}

// =========================================================================
//	MEPhylogenySimulator::lineageBirth
// =========================================================================

bool MEPhylogenySimulator::lineageBirth(long inExtantIndex,
                                   int inNumOffspring,
                                   bool inParentSurvives )
{
	int i; 
	long d0, d1, d2;
	long parent = mPhylogeny->getExtantLineage(inExtantIndex);
	
	mBirthOrder->add(parent);

	d0 = mPhylogeny->createLineage();
	
	mPhylogeny->setLineage(d0, parent, -1, -1, mCurrentTime, -1.0);

	if (inParentSurvives) {
		cloneLineage(parent, d0);
	} else {
		mutateLineage(parent, d0);
	}

	// replace the parent in the extant lineages list with the first offspring
	mPhylogeny->setExtantLineage(inExtantIndex, d0);
	mPhylogeny->setTipLineage(inExtantIndex, d0);

	// parent is made extinct
	mPhylogeny->setDeathTime(parent, mCurrentTime);
	mPhylogeny->setDescendant(parent, d0);

	d1 = d0;
	
	for (i = 1; i < inNumOffspring; i++) {
		d2 = mPhylogeny->createLineage();
		mPhylogeny->setLineage(d2, parent, -1, d1, mCurrentTime, -1.0);
		
		mutateLineage(parent, d2);

		// Add this offspring to the extant lineages list
		mPhylogeny->addExtantLineage(d2);
		mPhylogeny->addTipLineage(d2);
		
		d1 = d2;
	}
	mPhylogeny->setSibling(d0, d2);
        
	return true;
}

// =========================================================================
//	MEPhylogenySimulator::lineageDeath
// =========================================================================

bool MEPhylogenySimulator::lineageDeath(long inExtantIndex)
{
	long extinct = mPhylogeny->getExtantLineage(inExtantIndex);

	mPhylogeny->removeExtantLineage(inExtantIndex);

	if (mPruneExtinctLineages) {
		mPhylogeny->pruneLineage(extinct);
	} else {
		// Just set the time of death
		mPhylogeny->setDeathTime(extinct, mCurrentTime);
	}
		
	return true;
}

// =========================================================================
//	MEPhylogenySimulator::createGenotype
// =========================================================================

long MEPhylogenySimulator::createGenotype()
{
	long index;
	
	if (mNumAvailableGenotypes > 0) {
		index = mAvailableGenotypes[mNumAvailableGenotypes - 1];
		mNumAvailableGenotypes--;
	} else {
		index = mNumGenotypes;
		
		mNumGenotypes++;
		checkGenotypeCapacity(mNumGenotypes);
		
		mGenotypes[index] = newGenotype();
	}
	
	return index;
}


// =========================================================================
//	MEPhylogenySimulator::disposeGenotype	
// =========================================================================

void MEPhylogenySimulator::disposeGenotype(long inGenotype)
{
	checkAvailableGenotypeCapacity(mNumAvailableGenotypes + 1);
	
	mAvailableGenotypes[mNumAvailableGenotypes] = inGenotype;
	mNumAvailableGenotypes++;
}
// =========================================================================
//	MEPhylogenySimulator::checkGenotypeCapacity
// =========================================================================

void MEPhylogenySimulator::checkGenotypeCapacity(long inCapacity)
{
	if (inCapacity > mGenotypeCapacity) {
		while (inCapacity > mGenotypeCapacity)
			mGenotypeCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		MEGenotype **newGenotypes = new MEGenotype*[mGenotypeCapacity];
		
		// Copy existing trees into it
		for (i = 0; i < mNumGenotypes; i++)
			newGenotypes[i] = mGenotypes[i];
		for (i = mNumGenotypes; i < mGenotypeCapacity; i++)
			newGenotypes[i] = NULL;
			
		// Delete the old array
		delete [] mGenotypes;
		
		// And copy in the new one
		mGenotypes = newGenotypes;
	}
}

// =========================================================================
//	MEPhylogenySimulator::checkAvailableGenotypeCapacity
// =========================================================================

void MEPhylogenySimulator::checkAvailableGenotypeCapacity(long inCapacity)
{
	if (inCapacity > mAvailableGenotypeCapacity) {
		while (inCapacity > mAvailableGenotypeCapacity)
			mAvailableGenotypeCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		long *newArray = new long[mAvailableGenotypeCapacity];
		
		// Copy existing trees into it
		for (i = 0; i < mNumAvailableGenotypes; i++)
			newArray[i] = mAvailableGenotypes[i];
			
		// Delete the old array
		delete [] mAvailableGenotypes;
		
		// And copy in the new one
		mAvailableGenotypes = newArray;
	}
}

