/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	MESandersonTreeCriterion.cpp

		Class: MESandersonTreeCriterion

	This class provides a criterion a tree branch lengths which will
	minimise the amount of change of rate across a tree. This is 
	described by Sanderson, 1997.

 ******************************************************************************/

#include <string.h>

#include "MEExceptions.h"
#include "MESandersonTreeCriterion.h"

#include "MEPatternList.h"
#include "MELink.h"
#include "METree.h"

#pragma mark === MESandersonTreeCriterion ===

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion
// ---------------------------------------------------------------------------------

MESandersonTreeCriterion::MESandersonTreeCriterion() : METreeCriterion()
{
	numParameters = 0;
	parameters = NULL;
	
	mSandersonModel = kRateDifferenceAtEachNode;
	
	mRate = 1.0;
}

// ---------------------------------------------------------------------------------
//		 ~MESandersonTreeCriterion
// ---------------------------------------------------------------------------------

MESandersonTreeCriterion::~MESandersonTreeCriterion()
{
	if (parameters != NULL)
		delete [] parameters;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::setTree
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::setTree(METree *inTree)
{
	ThrowIf(!inTree->getIsRooted());

	METreeCriterion::setTree(inTree);
	
	int n = mTree->getNumInternalNodes();
	
	if (n != numParameters) {
		if (parameters != NULL)
			delete [] parameters;
		numParameters = n;
		
		NewArray(parameters, double, numParameters);
	}
	
	mTree->lengthsToHeights(mRate);
	heightsToValues();

	invalidate();
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::setSandersonModel
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::setSandersonModel(SandersonModel inSandersonModel)
{
	mSandersonModel = inSandersonModel;

	invalidate();
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getNumParameters
// ---------------------------------------------------------------------------------

int MESandersonTreeCriterion::getNumParameters()
{
	ThrowIfNull(mTree);
	
	return numParameters;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getValue
// ---------------------------------------------------------------------------------

double MESandersonTreeCriterion::getValue(int inIndex)
{
	TestBounds(inIndex, numParameters);
	return parameters[inIndex];
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getMinValue
// ---------------------------------------------------------------------------------

double MESandersonTreeCriterion::getMinValue(int inIndex)
{
	return 0.0;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getMaxValue
// ---------------------------------------------------------------------------------

double MESandersonTreeCriterion::getMaxValue(int inIndex)
{
	return 1.0;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::setValue
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::setValue(int inIndex, double inValue)
{
	TestBounds(inIndex, getNumParameters());
	parameters[inIndex] = inValue;
	invalidate();
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getRate
// ---------------------------------------------------------------------------------

double MESandersonTreeCriterion::getRate()
{
	return mRate;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getMinRate
// ---------------------------------------------------------------------------------

double MESandersonTreeCriterion::getMinRate()
{
	return kSmallReal;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::getMaxRate
// ---------------------------------------------------------------------------------

double MESandersonTreeCriterion::getMaxRate()
{
	return 10.0;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::setRate
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::setRate(double inRate)
{
	mRate = inRate;
	invalidate();
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::calcScore
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::calcScore()
{
	ThrowIfNull(mTree);
	
	valuesToHeights();
	score = 0.0;
	
	MELink *root1 = mTree->getRoot1();
	MELink *root2 = mTree->getRoot2();
	
	double dist1 = mTree->getRootDistance1();
	double dist2 = mTree->getRootDistance2();
	
	double height1 = mTree->getNodeHeight(root1->nodeNo);
	double height2 = mTree->getNodeHeight(root2->nodeNo);
			
	sumRates = 0.0;
	numRates = 0;
	
	double time1 = mTree->getRootHeight() - height1;
	double time2 = mTree->getRootHeight() - height2;
	
	if (time1 < kSmallReal)
		time1 = kSmallReal;

	if (time2 < kSmallReal)
		time2 = kSmallReal;

	double rate1 = mRate * (dist1 / time1);
	double rate2 = mRate * (dist2 / time2);

	if (mSandersonModel == kRateDifferenceAtEachNode) {
		// This method is suggested by Sanderson.
		calcScoreForNode(root1, rate1, height1);			
		calcScoreForNode(root2, rate2, height2);
	
		double diff = rate1 - rate2;
		score += diff * diff;	
	} else if (mSandersonModel == kDifferenceWithMeanAtRoot) {
		// This is the score for the root used by Sanderson.
		calcScoreForNode(root1, rate1, height1);			
		calcScoreForNode(root2, rate2, height2);
		
		double meanRate = sumRates / numRates;
	
		double diff = meanRate - rate1;
		score += diff * diff;	
				
		diff = meanRate - rate2;
		score += diff * diff;			
	} else if (mSandersonModel == kDifferenceWithMeanAtEachNode) {
		// This is a method suggested in discussion with Mike.
		calcSumRateForNode(root1, rate1, height1);			
		calcSumRateForNode(root2, rate2, height2);

		double meanRate = sumRates / numRates;
	
		calcScoreForNode2(root1, meanRate, height1);			
		calcScoreForNode2(root2, meanRate, height2);
		
		double diff = meanRate - rate1;
		score += diff * diff;	
				
		diff = meanRate - rate2;
		score += diff * diff;			
/*	} else { // mRootWeighting == kNoRootWeighting
		// No weighting for the node - this is if the position of the root on the
		// root branch is unknown
		double rate = (dist1 + dist2) / 
						((mTree->getRootHeight() - height1) + 
						(mTree->getRootHeight() - height2));

		calcScoreForNode(root1, rate, height1);			
		calcScoreForNode(root2, rate, height2);
*/		
 	}
	
	scoreKnown=true;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::calcSumRateForNode
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::calcSumRateForNode(MELink *inNode, double inRate, double inHeight)
{
	MELink *link;
	
	if (inNode->tipNo==-1) {
		double rate, diff, height;
		
		link=inNode->next;
		while (link!=inNode) {
			height = mTree->getNodeHeight(link->branch->nodeNo);
			diff = inHeight - height;
			if (diff < kSmallReal)
				diff = kSmallReal;
			rate = mRate * (mTree->getBranchLength(link->branchNo) / diff);
			calcScoreForNode(link->branch, rate, height);
			
			link=link->next;
		}
	}
	
	sumRates += inRate;
	numRates++;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::calcScoreForNode
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::calcScoreForNode(MELink *inNode, double inRate, double inHeight)
{
	MELink *link;
	
	if (inNode->tipNo==-1) {
		double rate, height, diff;
		
		link=inNode->next;
		while (link!=inNode) {
			height = mTree->getNodeHeight(link->branch->nodeNo);
			diff = inHeight - height;
			if (diff < kSmallReal)
				diff = kSmallReal;
			rate = mRate * (mTree->getBranchLength(link->branchNo) / diff);
			calcScoreForNode(link->branch, rate, height);
			
			diff = rate - inRate;
			score += diff * diff;			
			
			link=link->next;
		}
	}
	
	sumRates += inRate;
	numRates++;
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::calcScoreForNode2
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::calcScoreForNode2(MELink *inNode, double inMeanRate, double inHeight)
{
	MELink *link;
	
	if (inNode->tipNo==-1) {
		double rate, height, diff;
		
		link=inNode->next;
		while (link!=inNode) {
			height = mTree->getNodeHeight(link->branch->nodeNo);
			diff = inHeight - height;
			if (diff < kSmallReal)
				diff = kSmallReal;
			rate = mRate * (mTree->getBranchLength(link->branchNo) / diff);
			calcScoreForNode2(link->branch, inMeanRate, height);
			
			diff = rate - inMeanRate;
			score += diff * diff;			
			
			link=link->next;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::calcScoreForNode
// ---------------------------------------------------------------------------------

/*double MESandersonTreeCriterion::calcScoreForNode(MELink *inNode, double inLength, double inHeight)
{
	MELink *link;
	
	double height = mTree->getNodeHeight(inNode->nodeNo);
	double rate = inLength / (inHeight - height);
	
	if (inNode->tipNo==-1) {
		double rate1, diff;
		
		link=inNode->next;
		while (link!=inNode) {
			rate1 = calcScoreForNode(link->branch, mTree->getBranchLength(link->branchNo), height);
			
			diff = rate - rate1;
			score += diff * diff;			
			
			link=link->next;
		}
	}
	
	sumRates += rate;
	numRates++;
	
	return rate;
}*/

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::nodeHeightsToValues
// ---------------------------------------------------------------------------------

// Recursive function called by heightsToValues.

void MESandersonTreeCriterion::nodeHeightsToValues(MELink *inNode, double inHeight)
{
	if (inNode->tipNo==-1) {
		MELink *link, *branch;
		double height;
		
		height = mTree->getNodeHeight(inNode->nodeNo);

		link=inNode->next;
		while (link!=inNode) {
			branch=link->branch;
			nodeHeightsToValues(branch, height);	
			link=link->next;
		}
		
		double value = height / inHeight;
		ThrowIf(value<0.0);

		setValue(inNode->nodeNo, value);
	}
}


// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::heightsToValues
// ---------------------------------------------------------------------------------

// Calculate node values from node heights. This is mainly use to convert
// user branch lengths into starting node values.

void MESandersonTreeCriterion::heightsToValues()
{
	MELink *root1 = mTree->getRoot1();
	MELink *root2 = mTree->getRoot2();
	
	double height = mTree->getRootHeight();
	nodeHeightsToValues(root1, height);			
	nodeHeightsToValues(root2, height);			
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::nodeValuesToHeightsBelow
// ---------------------------------------------------------------------------------

// Recursive function called by valuesToHeights when below a calibration.

void MESandersonTreeCriterion::nodeValuesToHeightsBelow(MELink *inNode, double inHeight)
{
	if (inNode->tipNo==-1) {
		double height = inHeight * getValue(inNode->nodeNo);
		
		MELink *link, *branch;
		link=inNode->next;
		while (link!=inNode) {
			branch=link->branch;
			nodeValuesToHeightsBelow(branch, height);
			link=link->next;
		}
		
		mTree->setNodeHeight(inNode->nodeNo, height);
	}
}

// ---------------------------------------------------------------------------------
//		 MESandersonTreeCriterion::valuesToHeights
// ---------------------------------------------------------------------------------

void MESandersonTreeCriterion::valuesToHeights()
{
	MELink *root1 = mTree->getRoot1();
	MELink *root2 = mTree->getRoot2();
	
	double height = mTree->getRootHeight();
	nodeValuesToHeightsBelow(root1, height);			
	nodeValuesToHeightsBelow(root2, height);			
}

