/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MESequence.cpp

				MESequence Class
				
 ******************************************************************************/

// MESequence is derived from MEbasicSequence, and acts as a base for 
// MENucleotideSequence, MEAminoAcidSequence These derived classes limit the 
// character states wereas MESequence allows any letter and the standard gap, 
// missing characters.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MESequence.h"

#include "MECodeTableList.h"
#include "MECodeTable.h"
#include "MEFeature.h"
#include "MEFeatureList.h"

#pragma mark === MESequence ===


// ---------------------------------------------------------------------------------
//		 MESequence
// ---------------------------------------------------------------------------------

MESequence::MESequence(	) : MEBasicSequence()
{
	// All states required for amino acid and nucleotides (kUnknownSequence)
	stateCodes = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-?*";
	numStates = 39;
	numStateCodes = 39;
	
	geneticCode = kNonCoding;
	codeTable = NULL;
	readingFrame = 1;
	frameOffset = 0;
	
	featureList=NULL;
}

// ---------------------------------------------------------------------------------
//		 ~MESequence
// ---------------------------------------------------------------------------------

MESequence::~MESequence()
{
	if (codeTable!=NULL && codeTable->getOwner()==this)
		delete codeTable;

	if (featureList!=NULL && featureList->getOwner()==this)
		delete featureList;
}

// ---------------------------------------------------------------------------------
//		 MESequence::clone
// ---------------------------------------------------------------------------------

void MESequence::clone(MEObject *src)
{
	MEBasicSequence::clone(src);
	
	MESequence *seq=(MESequence *)src;
	setGeneticCode(seq->getGeneticCode());
	if (geneticCode==kNonCoding)
		setCodeTable(seq->getCodeTable());
	setReadingFrame(seq->getReadingFrame());
	
	// Must clone indels
}

// ---------------------------------------------------------------------------------
//		 MESequence::cloneExceptSequence
// ---------------------------------------------------------------------------------

void MESequence::cloneExceptSequence(MESequence *src)
{
	MEBasicSequence::cloneExceptSequence(src);
	
	setGeneticCode(src->getGeneticCode());
	if (geneticCode==kNonCoding)
		setCodeTable(src->getCodeTable());
	setReadingFrame(src->getReadingFrame());
}

// ---------------------------------------------------------------------------------
//		 MESequence::setSequenceType
// ---------------------------------------------------------------------------------
void MESequence::setSequenceType(SequenceType inSequenceType)
{ 
	switch (inSequenceType) {
		case kDNASequence:
		case kRNASequence:
			stateCodes = (char *)kNucleotides;
			numStateCodes = kNumNucleotides;
			numStates = 4; 
		break;
		case kAminoAcidSequence:
			stateCodes = (char *)kAminoAcids;
			numStateCodes = kNumAminoAcids;
			numStates = 20; 
		break;
		case kDiscreteSequence:
		case kUnknownSequence:
			// Allow all letters and digits
			stateCodes = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-?*";
			numStateCodes = 65;
			numStates = 65;
		break;
		default:
			throw MEIncompatibleSequence();
	}
	seqType=inSequenceType; 
}

// ---------------------------------------------------------------------------------
//		 MESequence::isSequenceCompatible
// ---------------------------------------------------------------------------------
// Returns true if sequence is compatible with inType. By default this is the case
// if inType equals kUnknownSequence

bool MESequence::isSequenceCompatible(AlignmentType inType)
{	
	if (inType==kUnknownAlignment)
		return true;
		
	switch (seqType) {
		case kDNASequence:
		case kRNASequence:
			if (codeTable!=NULL && (inType==kAminoAcidAlignment || inType==kCodonAlignment)) 
				return true;
			if (inType==kNucleotideAlignment)
				return true;
		break;
		case kAminoAcidSequence:
			if (inType==kAminoAcidAlignment)
				return true;
		break;
		case kUnknownSequence:
			return true;
		break;
	}

	return false;
}


// ---------------------------------------------------------------------------------
//		 MESequence::checkSequence
// ---------------------------------------------------------------------------------

// Checks that a sequence only contains recognised state codes
void MESequence::checkSequence(char *inSequence, int inLength)
{
	char *P=inSequence;
	
	for (int i=0; i<inLength; i++) {
		if (strchr(stateCodes, *P)==NULL)
			throw MEBadState();
			
		P++;
	}
}


// ---------------------------------------------------------------------------------
//		 MESequence::setSequence
// ---------------------------------------------------------------------------------

void MESequence::setSequence(char *inSequence, int inLength)
{
	MEBasicSequence::setSequence(inSequence, inLength);
}

// ---------------------------------------------------------------------------------
//		 MESequence::setGeneticCode
// ---------------------------------------------------------------------------------
void MESequence::setGeneticCode(GeneticCode inCode)
{
	geneticCode=inCode;
	if (geneticCode!=kNonCoding)
		codeTable=MECodeTableList::getStandardCodeTable(inCode);
	else
		codeTable=NULL;
	if (alignmentType!=kNucleotideAlignment)
		seqChanged=true;
}

// ---------------------------------------------------------------------------------
//		 MESequence::setCodeTable
// ---------------------------------------------------------------------------------
void MESequence::setCodeTable(MECodeTable *inCodeTable)
{
	geneticCode=kNonCoding;
	codeTable=inCodeTable;
	if (alignmentType!=kNucleotideAlignment)
		seqChanged=true;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNumStates()
// ---------------------------------------------------------------------------------

// Returns the number of states in the sequence which will be set by setSequenceType
int MESequence::getNumStates()
{
	return numStates;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNumStateCodes()
// ---------------------------------------------------------------------------------

int MESequence::getNumStateCodes()
{
	return numStateCodes;
}

// ---------------------------------------------------------------------------------
//		 MESequence::setReadingFrame
// ---------------------------------------------------------------------------------

void MESequence::setReadingFrame(int inReadingFrame)
{
	if (readingFrame!=inReadingFrame) {
		numSitesKnown = false;
		seqChanged = true;
	}
	
	readingFrame=inReadingFrame; 
	switch (readingFrame) {
		case 1: frameOffset=0; break;
		case 2: frameOffset=2; break;
		case 3: frameOffset=1; break;
		default:
			throw MEException("Bad reading frame");
	}
}

// ---------------------------------------------------------------------------------
//		 MESequence::convertToDNA
// ---------------------------------------------------------------------------------
// Converts the sequence to DNA (mapping all Us to Ts).

void MESequence::convertToDNA()
{
	if (seqType!=kDNASequence && seqType!=kRNASequence)
		throw MEIncompatibleSequence();
		
	int i;
	char *P=seq;
	
	ThrowIfNull(seq);
	
	for (i=0; i<length; i++) {
		if (*P=='U')
			*P='T';
		P++;
	}

	seqChanged=true;
	setSequenceType(kDNASequence);
}

// ---------------------------------------------------------------------------------
//		 MESequence::convertToRNA
// ---------------------------------------------------------------------------------

// Converts the sequence to RNA (mapping all Ts to Us).
void MESequence::convertToRNA()
{
	if (seqType!=kDNASequence && seqType!=kRNASequence)
		throw MEIncompatibleSequence();
		
	int i;
	char *P=seq;
	
	ThrowIfNull(seq);
	
	for (i=0; i<length; i++) {
		if (*P=='T')
			*P='U';
		P++;
	}
	
	seqChanged=true;
	setSequenceType(kRNASequence);
}

// ---------------------------------------------------------------------------------
//		 MESequence::translateToAminoAcid
// ---------------------------------------------------------------------------------

// Translates the sequence to Amino Acids. This is irreversible. It is usually
// better to use the Amino Acid access functions to obtain a translation.

void MESequence::translateToAminoAcid()
{
	if (codeTable==NULL || (seqType!=kDNASequence && seqType!=kRNASequence))
		throw MEIncompatibleSequence();
		
	// To be done...
	
	seqChanged=true;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getFirstORF
// ---------------------------------------------------------------------------------

// Finds the first open reading frame (ORF) in the current reading frame
// returning the coordinates in terms of the amino acid sequence. The first 
// and last site of the ORF are put into outStart and outFinish.

void MESequence::getFirstORF(int &outStart, int &outFinish)
{
	outStart=0;
	outFinish=0;
	getNextORF(outStart, outFinish);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNextORF
// ---------------------------------------------------------------------------------

// Finds the next open reading frame (ORF) in the current reading frame.
// The parameters outStart and outFinish should give the start and end
// of the previously found ORF.

bool MESequence::getNextORF(int &outStart, int &outFinish)
{
	int total=getAminoAcidLength();
	
	outStart=outFinish;
	while (outStart<total && getAminoAcidSite(outStart)=='*')
		outStart++;
	if (outStart==total)
		return false;
		
	outFinish=outStart;
	while (outFinish<total && getAminoAcidSite(outFinish)!='*')
		outFinish++;
	
	return true;
}

#pragma mark === Sequence access ===

// ---------------------------------------------------------------------------------
//		 MESequence::getNumSites
// ---------------------------------------------------------------------------------

int MESequence::getNumSites()
{
	if (!numSitesKnown) {
		switch (alignmentType) {
			case kCodonAlignment:
				numSites=getNumCodonSites();
			break;
			case kNucleotideAlignment: 
				numSites=getNumNucleotideSites();
			break;
			case kAminoAcidAlignment: 
				numSites=getNumAminoAcidSites();
			break;
			default: 
				numSites=getLength();
		}
		numSitesKnown=true;
	}
		
	return numSites;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNumSites
// ---------------------------------------------------------------------------------
// This gets the number of sites for the sequence if it were of type inAlignmentType
// which can be different from the sequence's own type.

int MESequence::getNumSites(AlignmentType inAlignmentType)
{
	int n;
	switch (inAlignmentType) {
		case kCodonAlignment:
			n=getNumCodonSites();
		break;
		case kNucleotideAlignment: 
			n=getNumNucleotideSites();
		break;
		case kAminoAcidAlignment: 
			n=getNumAminoAcidSites();
		break;
		default: 
			n=getLength();
	}
	return n;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getSiteCode
// ---------------------------------------------------------------------------------

short MESequence::getSiteCode(int inSite)
{
	short code;
	switch (alignmentType) {
		case kNucleotideAlignment: 
			code = getNucleotideCode(inSite);
		break;
		case kAminoAcidAlignment: 
			code = getAminoAcidCode(inSite);
		break;
		case kCodonAlignment:
			code = getCodonCode(inSite);
		break;
		default: 
			throw MEIncompatibleSequence();
	}
	return code;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getSiteIsGap
// ---------------------------------------------------------------------------------

bool MESequence::getSiteIsGap(int inSite)
{
	bool isGap;
	switch (alignmentType) {
		case kNucleotideAlignment: 
			isGap = getNucleotideCode(inSite) == kNucleotideGap;
		break;
		case kAminoAcidAlignment: 
			isGap = getAminoAcidCode(inSite) == kAminoAcidGap;
		break;
		case kCodonAlignment:
			isGap = getCodonCode(inSite) == kCodonGap;
		break;
		default: 
			throw MEIncompatibleSequence();
	}
	return isGap;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getSiteCodes
// ---------------------------------------------------------------------------------

// Copies a portion of the site codes into outStr. An inNumSites of 0 means copy
// all the sites from inStartSite. Thus if inStartSite=0 and inNumSites=0
// then the whole sequence is copied. If any of the required portion lies
// outside the actual sequence then the copy is padded with gaps. The
// sequence is first put into buffer (with appropriate transformations).
// The output string is null terminated.

void MESequence::getSiteCodes(int inStartSite, int inNumSites, short *outCodes)
{
	short *Q=outCodes;
	int i, j = inStartSite;
	switch (alignmentType) {
		case kNucleotideAlignment: 
			for (i = 0; i < inNumSites; i++) {
				*Q = getNucleotideCode(j);
				Q++;
				j++;
			}
		break;
		case kAminoAcidAlignment: 
			for (i = 0; i < inNumSites; i++) {
				*Q = getAminoAcidCode(j);
				Q++;
				j++;
			}
		case kCodonAlignment:
			for (i = 0; i < inNumSites; i++) {
				*Q = getCodonCode(j);
				Q++;
				j++;
			}
		default: 
			throw MEIncompatibleSequence();
	}
}

// ---------------------------------------------------------------------------------
//		 MESequence::getSeq2GapSite
// ---------------------------------------------------------------------------------

int MESequence::getSeq2GapSite(int inSequenceSite)
{
	int n = 0, i = 0;
	while (i < length && n <= inSequenceSite) {
		if (seq[i] != '-')
			n++;
		i++;
	}
	
	return i-1;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getGap2AlignSite
// ---------------------------------------------------------------------------------

int MESequence::getGap2AlignSite(int inGappedSite)
{
	return getGap2AlignSite(inGappedSite, alignmentType);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getGap2AlignSite
// ---------------------------------------------------------------------------------

int MESequence::getGap2AlignSite(int inGappedSite, AlignmentType inAlignmentType)
{
	switch (inAlignmentType) {
		case kNucleotideAlignment: 
			if (seqType==kAminoAcidSequence)
				return (inGappedSite*3) + frameOffset;
		break;
		case kCodonAlignment:
		case kAminoAcidAlignment: 
			if (seqType!=kAminoAcidSequence)
				return (inGappedSite + frameOffset)/3;
		break;
	}
	
	return inGappedSite + frameOffset;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getGap2AlignLength
// ---------------------------------------------------------------------------------

int MESequence::getGap2AlignLength(int inGappedLength)
{
	return getGap2AlignLength(inGappedLength, alignmentType);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getGap2AlignLength
// ---------------------------------------------------------------------------------

int MESequence::getGap2AlignLength(int inGappedLength, AlignmentType inAlignmentType)
{
	switch (inAlignmentType) {
		case kNucleotideAlignment: 
			if (seqType==kAminoAcidSequence)
				return (inGappedLength*3);
		break;
		case kCodonAlignment:
		case kAminoAcidAlignment: 
			if (seqType!=kAminoAcidSequence)
				return (inGappedLength)/3;
		break;
	}
	
	return inGappedLength;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAlign2GapLength
// ---------------------------------------------------------------------------------

int MESequence::getAlign2GapLength(int inAlignmentLength)
{
	return getAlign2GapLength(inAlignmentLength, alignmentType);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAlign2GapLength
// ---------------------------------------------------------------------------------

int MESequence::getAlign2GapLength(int inAlignmentLength, AlignmentType inAlignmentType)
{
	switch (inAlignmentType) {
		case kNucleotideAlignment: 
			if (seqType==kAminoAcidSequence)
				return (inAlignmentLength/3);
		break;
		case kCodonAlignment:
		case kAminoAcidAlignment: 
			if (seqType!=kAminoAcidSequence)
				return (inAlignmentLength*3);
		break;
	}
	
	return inAlignmentLength;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAlign2GapSite
// ---------------------------------------------------------------------------------

int MESequence::getAlign2GapSite(int inAlignmentSite)
{
	return getAlign2GapSite(inAlignmentSite, alignmentType);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAlign2GapSite
// ---------------------------------------------------------------------------------

int MESequence::getAlign2GapSite(int inAlignmentSite, AlignmentType inAlignmentType)
{
	switch (inAlignmentType) {
		case kNucleotideAlignment: 
			if (seqType==kAminoAcidSequence)
				return (inAlignmentSite  -frameOffset)/3;
		break;
		case kCodonAlignment:
		case kAminoAcidAlignment: 
			if (seqType!=kAminoAcidSequence)
				return (inAlignmentSite*3) - frameOffset;
		break;
	}
	
	return inAlignmentSite - frameOffset;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getGap2SeqSite
// ---------------------------------------------------------------------------------

int MESequence::getGap2SeqSite(int inGappedSite)
{
	int n=0;
	for (int i=0; i<inGappedSite; i++) {
		if (seq[i]!='-')
			n++;
	}
	
	return n;
}


// ---------------------------------------------------------------------------------
//		 MESequence::getStringLength
// ---------------------------------------------------------------------------------

int MESequence::getStringLength()
{
	int len;
	
	switch (alignmentType) {
		case kCodonAlignment:
			len=getCodonLength();
		break;
		case kNucleotideAlignment: 
			len=getNucleotideLength();
		break;
		case kAminoAcidAlignment: 
			len=getAminoAcidLength();
		break;
		default: 
			len=getLength();
	}
	
	return len;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getStringLength
// ---------------------------------------------------------------------------------

int MESequence::getStringLength(int inNumSites)
{
	int len;
	
	switch (alignmentType) {
		case kCodonAlignment:
			len=inNumSites*3;
		break;
		default: 
			len=inNumSites;
	}
	
	return len;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getSequenceString
// ---------------------------------------------------------------------------------

void MESequence::getSequenceString(char *outStr)
{
	switch (alignmentType) {
		case kCodonAlignment:
			getCodonString(outStr);
		break;
		case kNucleotideAlignment: 
			getNucleotideString(outStr);
		break;
		case kAminoAcidAlignment: 
			getAminoAcidString(outStr);
		break;
		default: 
			getNucleotideString(outStr);
	}
}


// ---------------------------------------------------------------------------------
//		 MESequence::getSequenceString
// ---------------------------------------------------------------------------------
// Copies a portion of the sequence into a string. Not very efficient because
// it calls getSite for each site.

void MESequence::getSequenceString(char *outStr, int inStartSite, int inNumSites)
{
	int i, site = inStartSite;
	int code;
	char *P = outStr;
	CodonCode codon;
	
	switch (alignmentType) {
		case kCodonAlignment:
			for (i = 0; i < inNumSites; i++) {
				codon.code = getCodonCode(site);
				
				*P = kNucleotides[codon.nucs.pos1];
				P++;
				*P = kNucleotides[codon.nucs.pos2];
				P++;
				*P = kNucleotides[codon.nucs.pos3];
				P++;
				site++;
			}
		break;
		case kAminoAcidAlignment: 
			for (i = 0; i < inNumSites; i++) {
				*P = getAminoAcidSite(site);
				site++;
				P++;
			}
		break;
		case kNucleotideAlignment: 
		default: 
			for (i = 0; i < inNumSites; i++) {
				*P = getNucleotideSite(site);
				site++;
				P++;
			}
		break;
	}
	*P = '\0';
}


// ---------------------------------------------------------------------------------
//		 MESequence::getGapString
// ---------------------------------------------------------------------------------

void MESequence::getGapString(char *outStr, int inNumSites)
{
	char *P=outStr;
	int i, n;
	if (alignmentType==kCodonAlignment ||
		(alignmentType==kNucleotideAlignment && seqType==kAminoAcidSequence))
		n=inNumSites*3;
	else
		n=inNumSites;
		
	for (i=0; i<n; i++) {
		*P='-';
		P++;
	}
	
	*P='\0';		
}


// ---------------------------------------------------------------------------------
//		 MESequence::getSequenceCodes
// ---------------------------------------------------------------------------------

void MESequence::getSequenceCodes(short *outCodes)
{
	switch (alignmentType) {
		case kCodonAlignment:
			getCodonCodes(outCodes);
		break;
		case kNucleotideAlignment: 
			getNucleotideCodes(outCodes);
		break;
		case kAminoAcidAlignment: 
			getAminoAcidCodes(outCodes);
		break;
		default: 
			throw MEIncompatibleSequence();
	}
}


#pragma mark === Nucleotide access ===

// ---------------------------------------------------------------------------------
//		 MESequence::getNumNucleotideSites
// ---------------------------------------------------------------------------------

int MESequence::getNumNucleotideSites()
{
	if (seqType==kAminoAcidSequence)
		return length*3 + frameOffset;
	
	return length + frameOffset;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideLength
// ---------------------------------------------------------------------------------

int MESequence::getNucleotideLength()
{
	if (seqType==kAminoAcidSequence) // Write out as triplet code
		return length*3 + frameOffset;
	
	return length + frameOffset;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideString
// ---------------------------------------------------------------------------------

// Copies the sequence as nucleotides to str. This should be allocated with
// getNucleotideLength() plus 1 for null termination.
void MESequence::getNucleotideString(char *outStr)
{
	if (seqType==kAminoAcidSequence) { // Write out as triplet code
		getAminoAcidTripletString(outStr);
		return;
	}
		
	int pos=-frameOffset;
	char *P=outStr;
	while (pos<0) {
		*P='-';
		P++;
		pos++;
	}
		
	if (length>0) {		
		memcpy(P, seq, length);
		P[length]='\0';
	} else
		P[0]='\0';
}


// ---------------------------------------------------------------------------------
//		 MESequence::copyNucleotideString
// ---------------------------------------------------------------------------------

// Writes the nucleotides to outStr using the 1 letter code. 

void MESequence::copyNucleotideString(int inStartSite, int inNumSites, char *outStr)
{
	if (seqType==kDNASequence || seqType==kRNASequence) {
		char *P=outStr;
		if (inStartSite < 0) {
			memset(P, '-', -inStartSite);
			P -= inStartSite;
			inNumSites += inStartSite;
			inStartSite = 0;
		}
		
		int endSite = inStartSite + inNumSites;
		if (endSite > length) {
			memcpy(P, seq + inStartSite, length - inStartSite);
			P += length - inStartSite;
			memset(P, '-', endSite - length);
			P += endSite - length;
		} else {
			memcpy(P, seq + inStartSite, inNumSites);
			P += inNumSites;
		}
		*P='\0';
	} else
		throw MEIncompatibleSequence();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideCodes
// ---------------------------------------------------------------------------------

// Copies the sequence as nucleotides number codes (0-17) to outCodes. This should be 
// allocated with the approapriate number of sites.
void MESequence::getNucleotideCodes(short *outCodes)
{
	if (seqType!=kDNASequence && seqType!=kRNASequence)
		throw MEIncompatibleSequence();
		
	int pos=-frameOffset;
	short *P=outCodes;
	while (pos<0) {
		*P=kNucleotideGap;
		P++;
		pos++;
	}
	
	char *Q = seq;
	while (pos < length) {
		*P=kNucleotideNumbers[*Q];
		P++;
		Q++;
		pos++;
	}
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideStateString
// ---------------------------------------------------------------------------------
// This is a static function used to translate nucleotide character, outSite, into
// the inSite char array. This array has one char ('\0' or '\1') for each state but 
// more than one may be set if the state is ambiguous.

void MESequence::getNucleotideStateString(char inState, char *outSiteStr)
{
	memset(outSiteStr, 0, 4);
		
	switch (inState) {
		case 'A': outSiteStr[0]=1; break; 
		case 'C': outSiteStr[1]=1; break; 
		case 'G': outSiteStr[2]=1; break; 
		case 'T':
		case 'U': outSiteStr[3]=1; break;
		
		case 'M': outSiteStr[0]=1; outSiteStr[1]=1; break;
		case 'R': outSiteStr[0]=1; outSiteStr[2]=1; break;
		case 'W': outSiteStr[0]=1; outSiteStr[3]=1; break;
		case 'S': outSiteStr[1]=1; outSiteStr[2]=1; break;
		case 'Y': outSiteStr[1]=1; outSiteStr[3]=1; break;
		case 'K': outSiteStr[2]=1; outSiteStr[3]=1; break;
		case 'B': outSiteStr[1]=1; outSiteStr[2]=1; outSiteStr[3]=1; break;
		case 'D': outSiteStr[0]=1; outSiteStr[2]=1; outSiteStr[3]=1; break;
		case 'H': outSiteStr[0]=1; outSiteStr[1]=1; outSiteStr[3]=1; break;
		case 'V': outSiteStr[0]=1; outSiteStr[1]=1; outSiteStr[2]=1; break;
		case 'N':
		case '?':
		case '-': 
			outSiteStr[0]=1; 
			outSiteStr[1]=1; 
			outSiteStr[2]=1; 
			outSiteStr[3]=1; 
		break;
	}
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideSite
// ---------------------------------------------------------------------------------
// Returns the site union for a given site.

void MESequence::getNucleotideSite(int inSite, char *outSiteStr)
{
	getNucleotideStateString(getNucleotideSite(inSite), outSiteStr);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideSite
// ---------------------------------------------------------------------------------
// Returns the state of a given site.

char MESequence::getNucleotideSite(int inSite)
{
	inSite-=frameOffset;
	if (inSite>=0 && inSite<length)
		return seq[inSite];
	else
		return '-';
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNucleotideCode
// ---------------------------------------------------------------------------------
// Returns the code of a given site.

short MESequence::getNucleotideCode(int inSite)
{
	inSite-=frameOffset;
	if (inSite>=0 && inSite<length)
		return kNucleotideNumbers[seq[inSite]];
	else
		return kNucleotideGap;
}

#pragma mark === Amino Acid access ===

// ---------------------------------------------------------------------------------
//		 MESequence::getNumAminoAcidSites
// ---------------------------------------------------------------------------------

int MESequence::getNumAminoAcidSites()
{
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL)
		return getNumCodonSites();

	return length;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidLength
// ---------------------------------------------------------------------------------

int MESequence::getAminoAcidLength()
{
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL)
		return getNumCodonSites();

	return length;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidString
// ---------------------------------------------------------------------------------

// Writes the amino acid to outStr using the 1 letter code. This should be allocated with
// getAminoAcidLength() plus 1 for null termination.
void MESequence::getAminoAcidString(char *outStr)
{
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL) { // can translate
		char *P=outStr;
		char *Q=seq;
		
		int seqLen=length;

		if (frameOffset==1) { // Add some padding if not in frame 0
			*P=kAminoAcids[codeTable->translateCodon('?', *Q, *(Q+1))];
			P++;
			Q+=2;
			seqLen-=2;
		} else if (frameOffset==2) { // Add some padding if not in frame 0
			*P=kAminoAcids[codeTable->translateCodon('?', '?', *Q)];
			P++;
			Q++;
			seqLen-=1;
		}
				
		int mod=seqLen%3;
		if (mod>0)
			seqLen-=mod;
		seqLen/=3;
		
		for (int i=0; i<seqLen; i++) {
			*P=kAminoAcids[codeTable->translateCodon(*(Q), *(Q+1), *(Q+2))];
			P++;
			Q+=3;
		}
		
		if (mod==1) {
			*P=kAminoAcids[codeTable->translateCodon(*(Q), '?', '?')];
			P++;
		} else if (mod==2) {
			*P=kAminoAcids[codeTable->translateCodon(*(Q), *(Q+1), '?')];
			P++;
		}	
			
		*P='\0';
	} else if (seqType==kAminoAcidSequence) {
		memcpy(outStr, seq, length);
		outStr[length]='\0';
	} else
		throw MEIncompatibleSequence();
}

// ---------------------------------------------------------------------------------
//		 MESequence::copyAminoAcidString
// ---------------------------------------------------------------------------------

// Writes the amino acid to outStr using the 1 letter code. 

void MESequence::copyAminoAcidString(int inStartSite, int inNumSites, char *outStr)
{
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL) { // can translate
		int seqStart=(inStartSite*3) - frameOffset;		

		char *P=outStr;
		
		if (seqStart<0) {
			if (seqStart==-1) { // Add some padding if not in frame 0
				*P=kAminoAcids[codeTable->translateCodon('?', seq[0], seq[1])];
				P++;
				seqStart=2;
			} else if (seqStart==-2) { // Add some padding if not in frame 0
				*P=kAminoAcids[codeTable->translateCodon('?', '?', seq[0])];
				P++;
				seqStart=1;
			} else
				throw MEIndexOutOfBounds();
			inNumSites--;
		}
				
		char *Q=seq+seqStart;

		int seqLen=length-seqStart;
		if (inNumSites>0) {
			int len3=inNumSites*3;
			if (len3<seqLen)
				seqLen=len3;
		}
			
		int mod=seqLen%3;
		if (mod>0)
			seqLen-=mod;
		seqLen/=3;
		
		Q=seq+seqStart;
		for (int i=0; i<seqLen; i++) {
			*P=kAminoAcids[codeTable->translateCodon(*(Q), *(Q+1), *(Q+2))];
			P++;
			Q+=3;
		}
		
		if (mod==1) {
			*P=kAminoAcids[codeTable->translateCodon(*(Q), '?', '?')];
			P++;
		} else if (mod==2) {
			*P=kAminoAcids[codeTable->translateCodon(*(Q), *(Q+1), '?')];
			P++;
		}	
			
		*P='\0';
	} else if (seqType==kAminoAcidSequence) {
		int seqLen=length-inStartSite;
		if (inNumSites>0 && inNumSites<seqLen)
			seqLen=inNumSites;
			
		strncpy(outStr, seq, seqLen);
		outStr[seqLen]='\0';
	} else
		throw MEIncompatibleSequence();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidCodes
// ---------------------------------------------------------------------------------

// Copies the sequence as amino acid number codes to outCodes. This should be 
// allocated with the approapriate number of sites.
void MESequence::getAminoAcidCodes(short *outCodes)
{
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL) { // can translate
		short *P=outCodes;
		char *Q=seq;
		
		int seqLen=length;

		if (frameOffset==1) { // Add some padding if not in frame 0
			*P=codeTable->translateCodon('?', *Q, *(Q+1));
			P++;
			Q+=2;
			seqLen-=2;
		} else if (frameOffset==2) { // Add some padding if not in frame 0
			*P=codeTable->translateCodon('?', '?', *Q);
			P++;
			Q++;
			seqLen-=1;
		}
				
		int mod=seqLen%3;
		if (mod>0)
			seqLen-=mod;
		seqLen/=3;
		
		for (int i=0; i<seqLen; i++) {
			*P=codeTable->translateCodon(*(Q), *(Q+1), *(Q+2));
			P++;
			Q+=3;
		}
		
		if (mod==1) {
			*P=codeTable->translateCodon(*(Q), '?', '?');
			P++;
		} else if (mod==2) {
			*P=codeTable->translateCodon(*(Q), *(Q+1), '?');
			P++;
		}	
			
		*P='\0';
	} else if (seqType==kAminoAcidSequence) {
		short *P=outCodes;
		char *Q = seq;
		for (int i=0; i<length; i++) {
			*P=kAminoAcidNumbers[*Q];
			P++;
			Q++;
		}
	} else
		throw MEIncompatibleSequence();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidTripletLength
// ---------------------------------------------------------------------------------
// Unlike one-letter amino acid codes, triplets can be frame shifted.

int MESequence::getAminoAcidTripletLength()
{
	return getCodonLength();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidTripletString
// ---------------------------------------------------------------------------------

// Writes the amino acid to outStr using the 3 letter code
void MESequence::getAminoAcidTripletString(char *outStr)
{
	int i, n;
		
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL) { // can translate
		char *P=outStr;
		char *Q=seq;
		
		if (frameOffset>0) { // Add some padding if not in frame 0
			if (frameOffset==1) {
				n=codeTable->translateCodon('?', *Q, *(Q+1));
				Q+=2;
			} else if (frameOffset==2) {
				n=codeTable->translateCodon('?', '?', *Q);
				Q++;
			} else
				throw MEIndexOutOfBounds();
			*P=kAminoAcidTriplets[n][0];	P++;
			*P=kAminoAcidTriplets[n][1];	P++;
			*P=kAminoAcidTriplets[n][2];	P++;
		}
				
		int seqLen=length-frameOffset;
			
		int mod=seqLen%3;
		if (mod>0)
			seqLen-=mod;
		seqLen/=3;
		
		for (i=0; i<seqLen; i++) {
			n=codeTable->translateCodon(*(Q), *(Q+1), *(Q+2));
			*P=kAminoAcidTriplets[n][0];	P++;
			*P=kAminoAcidTriplets[n][1];	P++;
			*P=kAminoAcidTriplets[n][2];	P++;
			Q+=3;
		}
		
		if (mod>0) { // Add a padded codon at the end
			if (mod==1)
				n=codeTable->translateCodon(*(Q), '?', '?');
			else if (mod==2)
				n=codeTable->translateCodon(*(Q), *(Q+1), '?');
			*P=kAminoAcidTriplets[n][0];	P++;
			*P=kAminoAcidTriplets[n][1];	P++;
			*P=kAminoAcidTriplets[n][2];	P++;
		}
					
		*P='\0';
	} else if (seqType==kAminoAcidSequence) {
		int seqStart= - frameOffset;		

		char *P=outStr;
		int seqLen=length*3;
		
		if (frameOffset==1) { // Add some padding if not in frame 0
			*P='-'; P++;
			seqLen+1;
		} else if (frameOffset==2) { // Add some padding if not in frame 0
			*P='-'; P++;
			*P='-'; P++;
			seqLen+=2;
		}
				
		char *Q=seq;
		for (i=0; i<length; i++) {
			n=kAminoAcidNumbers[*Q];
			*P=kAminoAcidTriplets[n][0];		P++;
			*(P+1)=kAminoAcidTriplets[n][1];	P++;
			*(P+2)=kAminoAcidTriplets[n][2];	P++;
			Q++;
		}
		
		int mod=seqLen%3; // How many short last complete codon
		if (mod>0) {
			*P='-'; P++;
			if (mod<2)
				*P='-'; P++;
		}
			
		*P='\0';
	} else
		throw MEIncompatibleSequence();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidStateString
// ---------------------------------------------------------------------------------
// This is a static function used to translate amino acid character, inState, into
// a Site code string.

void MESequence::getAminoAcidStateString(char inState, char *outSiteStr)
{
	memset(outSiteStr, 0, 20);

	int index=kAminoAcidNumbers[inState];
	if (index>=20) {
		switch (inState) {
			// Asx		Asp=D		Asn=N
			case 'B': outSiteStr[2]=1; outSiteStr[11]=1; break;
			
			// Glx		Gly=G		Gln=Q
			case 'Z': outSiteStr[5]=1; outSiteStr[13]=1; break;
			
			// X, ?, -, *
			default:
				memset(outSiteStr, 1, 20);
		}
	} else
		outSiteStr[index]=1;
}


// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidSite
// ---------------------------------------------------------------------------------
// Returns the state code string for a given site.

void MESequence::getAminoAcidSite(int inSite, char *outSiteStr)
{
	getAminoAcidStateString(getAminoAcidSite(inSite), outSiteStr);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidSite
// ---------------------------------------------------------------------------------
// Returns the state of a given site.

char MESequence::getAminoAcidSite(int inSite)
{
	ThrowIfNull(seq);
		
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL) { // can translate
		int seqSite=(inSite*3) - frameOffset;
		
		if (seqSite<0) {
			if (seqSite==-1) { // Add some padding if not in frame 0
				return kAminoAcids[codeTable->translateCodon('?', seq[0], seq[1])];
			} else if (seqSite==-2) { // Add some padding if not in frame 0
				return kAminoAcids[codeTable->translateCodon('?', '?', seq[0])];
			} else
				return kAminoAcidGap;
		} else if (seqSite<length) {
			if (seqSite==length-2)
				return kAminoAcids[codeTable->translateCodon(seq[seqSite], seq[seqSite+1], '?')];
			else if (seqSite==length-1)
				return kAminoAcids[codeTable->translateCodon(seq[seqSite], '?', '?')];
			else			
				return kAminoAcids[codeTable->translateCodon(seq[seqSite], seq[seqSite+1], seq[seqSite+2])];
		} else
			return '-';
	} else if (seqType==kAminoAcidSequence) {
		if (inSite>=0 && inSite<length)
			return seq[inSite];
	}
		
	return '-';
}

// ---------------------------------------------------------------------------------
//		 MESequence::getAminoAcidCode
// ---------------------------------------------------------------------------------
// Returns the state code of a given site.

short MESequence::getAminoAcidCode(int inSite)
{
	ThrowIfNull(seq);
		
	if ((seqType==kDNASequence || seqType==kRNASequence) && codeTable!=NULL) { // can translate
		int seqSite=(inSite*3) - frameOffset;
		
		if (seqSite<0) {
			if (seqSite==-1) { // Add some padding if not in frame 0
				return codeTable->translateCodon('?', seq[0], seq[1]);
			} else if (seqSite==-2) { // Add some padding if not in frame 0
				return codeTable->translateCodon('?', '?', seq[0]);
			} else
				throw MEIndexOutOfBounds();
		} else if (seqSite<length) {
			if (seqSite==length-2)
				return codeTable->translateCodon(seq[seqSite], seq[seqSite+1], '?');
			else if (seqSite==length-1)
				return codeTable->translateCodon(seq[seqSite], '?', '?');
			else			
				return codeTable->translateCodon(seq[seqSite], seq[seqSite+1], seq[seqSite+2]);
		}
	} else if (seqType==kAminoAcidSequence) {
		if (inSite>=0 && inSite<length)
			return kAminoAcidNumbers[seq[inSite]];
	}
		
	return kAminoAcidGap;
}

#pragma mark === Codon access ===

// ---------------------------------------------------------------------------------
//		 MESequence::getNumCodonSites
// ---------------------------------------------------------------------------------

// This returns the number of codons in the sequence (rounded up if the number
// of nucleotides is not a multiple of 3).

int MESequence::getNumCodonSites()
{
	int len;
	
	if (seqType==kAminoAcidSequence)	// Write out as triplet code
		len=length*3;
	else
		len=length;

	len+=frameOffset;
	
	int mod=len%3;
	if (mod>0)
		len+=3-mod;
	
	return len/3;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getCodonLength
// ---------------------------------------------------------------------------------

// This returns the length of a string required to take the sequence of codons. 

int MESequence::getCodonLength()
{
	int len;
	
	if (seqType==kAminoAcidSequence)	// Write out as triplet code
		len=length*3;
	else
		len=length;

	len+=frameOffset;
	
	int mod=len%3;
	if (mod>0)
		len+=3-mod;
	
	return len;
}

// ---------------------------------------------------------------------------------
//		 MESequence::getCodonString
// ---------------------------------------------------------------------------------

// Copies the sequence as codons to outStr.
void MESequence::getCodonString(char *outStr)
{
	if (seqType==kAminoAcidSequence) { // write as amino acid triplets
		getAminoAcidTripletString(outStr);
	} else {
		int seqStart= - frameOffset;		

		char *P=outStr;
		int seqLen=length;
		
		if (frameOffset==1) { // Add some padding if not in frame 0
			*P='-'; P++;
			seqLen+=1;
		} else if (frameOffset==2) { // Add some padding if not in frame 0
			*P='-'; P++;
			*P='-'; P++;
			seqLen+=2;
		}
				
		strncpy(P, seq, length);
		P+=length;
		
		int mod=seqLen%3; // How many short last complete codon
		if (mod>0) {
			*P='-'; P++;
			if (mod<2) {
				*P='-'; P++;
			}
		}
			
		*P='\0';
	}
}

// ---------------------------------------------------------------------------------
//		 MESequence::getCodonCodes
// ---------------------------------------------------------------------------------

// Copies the sequence as codon number codes (12 bits) to outCodes. This should be 
// allocated with the approapriate number of sites.
void MESequence::getCodonCodes(short *outCodes)
{
	if (seqType!=kDNASequence && seqType!=kRNASequence)
		throw MEIncompatibleSequence();

	int i, n = getNumSites();
	short *P = outCodes;
	for (i = 0; i < n; i++) {
		*P = getCodonCode(i);
		P++;
	}
}

// ---------------------------------------------------------------------------------
//		 MESequence::getCodonSite
// ---------------------------------------------------------------------------------

void MESequence::getCodonSite(int inSite, char* outSiteStr)
{
	int seqStart=inSite*3-frameOffset;
	if (seqStart<length) {	
		char cp1[4], cp2[4], cp3[4];
		
		if (seqStart == -2) { // Add some padding at front
			getNucleotideStateString('?', cp1);
			getNucleotideStateString('?', cp2);
			getNucleotideStateString(seq[0], cp3);
		} else if (seqStart == -1) { // Add some padding at front
			getNucleotideStateString('?', cp1);
			getNucleotideStateString(seq[0], cp2);
			getNucleotideStateString(seq[1], cp3);
		} else if (seqStart==length-2) { // Add some padding at end
			getNucleotideStateString(seq[seqStart], cp1);
			getNucleotideStateString(seq[seqStart+1], cp2);
			getNucleotideStateString('?', cp3);
		} else if (seqStart==length-1) { // Add some padding at end
			getNucleotideStateString(seq[seqStart], cp1);
			getNucleotideStateString('?', cp2);
			getNucleotideStateString('?', cp3);
		} else  {
			getNucleotideStateString(seq[seqStart], cp1);
			getNucleotideStateString(seq[seqStart+1], cp2);
			getNucleotideStateString(seq[seqStart+2], cp3);
		}
		
		memset(outSiteStr, 0, 64);
		
		int i, j, k, codon;
		for (i=0; i<4; i++) {
			for (j=0; j<4; j++) {
				for (k=0; k<4; k++) {
					if (cp1[i]=='\1' && cp2[j]=='\1' && cp3[k]=='\1') {
						codon=i*16 + j*4 + k;
						outSiteStr[codon]='\1';
					}
				}
			}
		
		}
	} else 
		memset(outSiteStr, 1, 64);
}


// ---------------------------------------------------------------------------------
//		 MESequence::getCodonCode
// ---------------------------------------------------------------------------------
// A 15 bit code for the codon at inSite. This is three nucleotide codes (5 bits each)
// compressed into the first 15 bits of a short.
short MESequence::getCodonCode(int inSite)
{
	ThrowIfNull(seq);
	
	int seqSite=(inSite*3) - frameOffset;
	CodonCode codon;
	
	if (seqSite<0) {
		if (seqSite==-1) { // Add some padding if not in frame 0
			codon.nucs.pos1 = kNucleotideGap;
			codon.nucs.pos2 = kNucleotideNumbers[seq[0]];
			codon.nucs.pos3 = kNucleotideNumbers[seq[1]];
		} else if (seqSite==-2) { // Add some padding if not in frame 0
			codon.nucs.pos1 = kNucleotideGap;
			codon.nucs.pos2 = kNucleotideGap;
			codon.nucs.pos3 = kNucleotideNumbers[seq[0]];
		} else {
			codon.code = kCodonGap;
		}
	} else if (seqSite<length) {
		if (seqSite==length-2) {
			codon.nucs.pos1 = kNucleotideNumbers[seq[seqSite]];
			codon.nucs.pos2 = kNucleotideNumbers[seq[seqSite+1]];
			codon.nucs.pos3 = kNucleotideGap;
		} else if (seqSite==length-1) {
			codon.nucs.pos1 = kNucleotideNumbers[seq[seqSite]];
			codon.nucs.pos2 = kNucleotideGap;
			codon.nucs.pos3 = kNucleotideGap;
		} else {		
			codon.nucs.pos1 = kNucleotideNumbers[seq[seqSite]];
			codon.nucs.pos2 = kNucleotideNumbers[seq[seqSite+1]];
			codon.nucs.pos3 = kNucleotideNumbers[seq[seqSite+2]];
		}
	} else {
		codon.code = kCodonGap;
	}
	
	return codon.code;
}


#pragma mark === Features ===

// ---------------------------------------------------------------------------------
//		 MESequence::setFeatureList
// ---------------------------------------------------------------------------------

void MESequence::setFeatureList(MEFeatureList *inFeatureList, bool isOwner)
{
	featureList=inFeatureList;
	if (featureList==NULL || isOwner)
		featureList->setOwner(this);
}


// ---------------------------------------------------------------------------------
//		 MESequence::cloneFeatureList
// ---------------------------------------------------------------------------------

void MESequence::cloneFeatureList(MEFeatureList *outFeatureList)
{
	outFeatureList->cloneList(featureList);
}


// ---------------------------------------------------------------------------------
//		 MESequence::newFeatureList
// ---------------------------------------------------------------------------------

MEFeatureList *MESequence::newFeatureList()
{
	MEFeatureList *list;
	NewClass(list, MEFeatureList());
	return list;
}


// ---------------------------------------------------------------------------------
//		 MESequence::newFeature
// ---------------------------------------------------------------------------------

MEFeature *MESequence::newFeature()
{
	if (featureList==NULL)
		setFeatureList(newFeatureList(), true);

	return (MEFeature *)featureList->createObject();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNumFeatures
// ---------------------------------------------------------------------------------

int MESequence::getNumFeatures()
{
	ThrowIfNull(featureList);

	return featureList->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getFeature
// ---------------------------------------------------------------------------------

MEFeature* MESequence::getFeature(int index)
{
	ThrowIfNull(featureList);
	
	return (MEFeature *)featureList->getObject(index);
}

// ---------------------------------------------------------------------------------
//		 MESequence::getFirstFeature
// ---------------------------------------------------------------------------------

MEFeature* MESequence::getFirstFeature()
{
	if (featureList == NULL)
		return NULL;
	
	return (MEFeature *)featureList->getFirst();
}

// ---------------------------------------------------------------------------------
//		 MESequence::getNextFeature
// ---------------------------------------------------------------------------------

MEFeature* MESequence::getNextFeature()
{
	ThrowIfNull(featureList);
	
	return (MEFeature *)featureList->getNext();
}

// ---------------------------------------------------------------------------------
//		 MESequence::addFeature
// ---------------------------------------------------------------------------------

void MESequence::addFeature(MEFeature *inFeature)
{
	if (featureList==NULL)
		setFeatureList(newFeatureList(), true);

	featureList->add(inFeature);
}

// ---------------------------------------------------------------------------------
//		 MESequence::insertFeature
// ---------------------------------------------------------------------------------

void MESequence::insertFeature(int index, MEFeature *inFeature)
{
	if (featureList==NULL)
		setFeatureList(newFeatureList(), true);

	featureList->insert(index, inFeature);
}

// ---------------------------------------------------------------------------------
//		 MESequence::removeFeature
// ---------------------------------------------------------------------------------

MEFeature* MESequence::removeFeature(int index)
{
	return (MEFeature*)featureList->remove(index);
}

// ---------------------------------------------------------------------------------
//		 MESequence::removeAllFeatures
// ---------------------------------------------------------------------------------

void MESequence::removeAllFeatures()
{
	featureList->removeAll();
}

#pragma mark === Read / Write ===

// ---------------------------------------------------------------------------------
//		 MESequence::readObject
// ---------------------------------------------------------------------------------

void MESequence::readObject(MEStream *stream)
{
	MEBasicSequence::readObject(stream);
	
	int code;
	if (stream->get("GeneticCode", &code))
		setGeneticCode((GeneticCode)code);
	
	MECodeTable* newCodeTable=(MECodeTable*)stream->get("CodeTable");
	if (newCodeTable!=NULL)
		setCodeTable(newCodeTable);
		
	int frame;
	if (stream->get("Frame", &frame))
		setReadingFrame(frame);
	
	featureList=(MEFeatureList*)stream->get("Features");
}

// ---------------------------------------------------------------------------------
//		 MESequence::writeObject
// ---------------------------------------------------------------------------------

void MESequence::writeObject(MEStream *stream)
{
	MEBasicSequence::writeObject(stream);

	stream->put("GeneticCode", (int)geneticCode);
	if (geneticCode==kNonCoding)
		stream->put("CodeTable", codeTable);
	else
		stream->put("CodeTable", (MEObject *)NULL);

	stream->put("Frame", readingFrame);
	
	stream->put("Features", featureList);	
}


