/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MESequence.h

		Interface for the MESequence Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Advanced Sequences"
 
#ifndef _MESEQUENCE_H_
#define _MESEQUENCE_H_

#include "MEBasicSequence.h"

class MEStream;
class MECodeTable;
class MEFeatureList;
class MEFeature;

/******************************************************************************
	class MESequence
 ******************************************************************************/
#pragma mark MESequence class

class MESequence : public MEBasicSequence {
	public:
		virtual long		getID() const { return 'MSeq'; }
		//: The ID of this object - 'MSeq'.
		
		virtual char*		getClassName() const { return "Sequence"; }
		//: Returns the name of this class - "Sequence"
		
	protected:
		char*			stateCodes;		// String of state codes (including ambiguity, gap, etc.)
		int				numStateCodes;	// Number of state codes
		
		GeneticCode		geneticCode;	// Current genetic code type
		MECodeTable*	codeTable;		// Current genetic code
		
		int				readingFrame, 	// Current reading frame (1-3)
						frameOffset, 	// Offset of sequence in current reading frame
						oldFrame;		// Place to store reading frame whilst it is suspended
		
		MEFeatureList*	featureList;	// List of features for this sequence
		
	public:
	 					MESequence();
	 	virtual			~MESequence();

		virtual void 	clone(MEObject *src);
		
		virtual void 	cloneExceptSequence(MESequence *src);	

		virtual void 	setSequenceType(SequenceType inSequenceType);
		virtual bool	isSequenceCompatible(AlignmentType inType);
		virtual void	checkSequence(char *inSequence, int inLength);
		
		virtual void 	setSequence(char *inSequence, int inLength);
		
		virtual void			setGeneticCode(GeneticCode inCode);
		virtual GeneticCode		getGeneticCode() { return geneticCode; }
		virtual void			setCodeTable(MECodeTable *inCodeTable);
		virtual MECodeTable*	getCodeTable() { return codeTable; }

		virtual void	setReadingFrame(int inReadingFrame);
		virtual int		getReadingFrame() { return readingFrame; }
		virtual void	suspendReadingFrame() { oldFrame=readingFrame; setReadingFrame(1); }
		virtual void	restoreReadingFrame() { setReadingFrame(oldFrame); }

		virtual int		getNumStates();
		virtual int		getNumStateCodes();

		virtual void	convertToDNA();
		virtual void	convertToRNA();
		virtual void	translateToAminoAcid();
				
		virtual void	getFirstORF(int &outStart, int &outFinish);
		virtual bool	getNextORF(int &outStart, int &outFinish);
				
		virtual int		getNumSites();
		virtual int		getNumSites(AlignmentType inAlignmentType);

		virtual short	getSiteCode(int inSite);
		virtual bool	getSiteIsGap(int inSite);
		virtual void	getSiteCodes(int inStartSite, int inNumSites, short *outCodes);

		virtual int		getSeq2GapSite(int inSequenceSite);
		virtual int		getGap2AlignSite(int inGappedSite);
		virtual int		getGap2AlignSite(int inGappedSite, AlignmentType inAlignmentType);
		virtual int		getGap2AlignLength(int inGappedLength);
		virtual int		getGap2AlignLength(int inGappedLength, AlignmentType inAlignmentType);
		
		virtual int		getAlign2GapSite(int inAlignmentSite);
		virtual int		getAlign2GapSite(int inAlignmentSite, AlignmentType inAlignmentType);
		virtual int		getAlign2GapLength(int inAlignmentLength);
		virtual int		getAlign2GapLength(int inAlignmentLength, AlignmentType inAlignmentType);
		virtual int		getGap2SeqSite(int inGappedSite);
		
		virtual int		getStringLength();
		virtual int		getStringLength(int inNumSites);
		virtual void	getSequenceString(char *outStr);
		virtual void	getSequenceString(char *outStr, int inStartSite, int inNumSites);
		virtual void	getGapString(char *outStr, int inNumSites);
		virtual void	getSequenceCodes(short *outCodes);
		
		bool			getIsTranslatable() { return getIsNucleotide() && codeTable!=NULL; }

		virtual int		getNumNucleotideSites();
		virtual int		getNucleotideLength();
		virtual void	getNucleotideString(char *outStr);
		virtual void	copyNucleotideString(int inStartSite, int inNumSites, char *outStr);
		virtual void	getNucleotideCodes(short *outCodes);
		
		static void		getNucleotideStateString(char inState, char *outSiteStr);
		virtual void	getNucleotideSite(int inSite, char *outSiteStr);
		virtual char	getNucleotideSite(int inSite);
		virtual short	getNucleotideCode(int inSite);

		virtual int		getNumAminoAcidSites();
		virtual int		getAminoAcidLength();
		virtual void	getAminoAcidString(char *outStr);
		virtual void	copyAminoAcidString(int inStartSite, int inNumSites, char *outStr);
		virtual void	getAminoAcidCodes(short *outCodes);
		
		virtual int		getAminoAcidTripletLength();
		virtual void	getAminoAcidTripletString(char *outStr);

		virtual void	getAminoAcidStateString(char inState, char *outSiteStr);
		virtual void	getAminoAcidSite(int inSite, char *outSiteStr);
		virtual char	getAminoAcidSite(int inSite);
		virtual short	getAminoAcidCode(int inSite);

		virtual int		getNumCodonSites();
		virtual int		getCodonLength();
		virtual void	getCodonString(char *str);
		virtual void	getCodonCodes(short *outCodes);
				
		virtual void	getCodonSite(int inSite, char* outSiteStr);
		virtual short 	getCodonCode(int inSite);
			
		virtual void			setFeatureList(MEFeatureList *inFeatureList, bool isOwner);
		virtual void			cloneFeatureList(MEFeatureList *outFeatureList);
		virtual MEFeatureList*	getFeatureList() { return featureList; };
		virtual MEFeatureList*	newFeatureList();
		
		virtual MEFeature*		newFeature();
		virtual int				getNumFeatures();
		virtual MEFeature*		getFeature(int index);
		virtual MEFeature*		getFirstFeature();
		virtual MEFeature*		getNextFeature();
		virtual void			addFeature(MEFeature *inFeature);
		virtual void			insertFeature(int index, MEFeature *inFeature);
		virtual MEFeature*		removeFeature(int index);
		virtual void			removeAllFeatures();

		virtual void	readObject(MEStream *stream);
		virtual void	writeObject(MEStream *stream);
};

#endif /* _MESEQUENCE_H_ */

