/******************************************************************************
 MESpatialSimulator.cpp

				MESpatialSimulator Class
				
	Copyright  2001 Andrew Rambaut, University of Oxford. All rights reserved.

 ******************************************************************************/

#include <stdio.h>
#include <math.h>

#include "MESpatialSimulator.h"

#include "MEMaths.h"

#include "CReportRecord.h"
#include "CGraphicsRecord.h"

#pragma mark === MESpatialLineage ===

// =========================================================================
//	MESpatialLineage::MESpatialLineage
// =========================================================================

MESpatialLineage::MESpatialLineage() : MELineage()
{
    mX = mY = 0;
    mOffspringDirection = 0;
    for (int i = 0; i < 8; i++)
    	mLinks[i] = -1;
    	
    mOrder = 0;
    mBalance = -1.0;
}

// =========================================================================
//	MESpatialPhylogeny::initializeLineage
// =========================================================================

void MESpatialPhylogeny::initializeLineage(long inLineageIndex)
{
	MEPhylogeny::initializeLineage(inLineageIndex);
	
	MESpatialLineage *lineage = static_cast<MESpatialLineage *>(getLineage(inLineageIndex));
    lineage->mX = 0;
    lineage->mY = 0;
    lineage->mOffspringDirection = 0;
    for (int i = 0; i < 8; i++)
    	lineage->mLinks[i] = -1;
    	
    lineage->mOrder = 0;
    lineage->mBalance = -1.0;
}

#pragma mark === MESpatialSimulator ===

// =========================================================================
//	MESpatialSimulator::MESpatialSimulator
// =========================================================================

MESpatialSimulator::MESpatialSimulator() :
	MEBirthDeathSimulator()
{
    // this can be 2, 4, 6 or 8
    mNumAdjacencies = 6;
    
    mMode = 0;

    mOrder = 1;
}

// =========================================================================
//	MESpatialSimulator::~MESpatialSimulator
// =========================================================================

MESpatialSimulator::~MESpatialSimulator()
{
}
		
// =========================================================================
//	MESpatialSimulator::setNumAdjacencies
// =========================================================================

void MESpatialSimulator::setNumAdjacencies(int inNumAdjacencies)
{
	if (inNumAdjacencies != 2 &&
		inNumAdjacencies != 3 &&
		inNumAdjacencies != 4 &&
		inNumAdjacencies != 6 &&
		inNumAdjacencies != 8)
		throw CException("Number of adjacencies must be 2, 3, 4, 6 or 8");
		
	mNumAdjacencies = inNumAdjacencies;
}

// =========================================================================
//	MESpatialSimulator::initialize
// =========================================================================

void MESpatialSimulator::initialize()
{
    mOrder = 1;
	MEBirthDeathSimulator::initialize();
}

// =========================================================================
//	MESpatialSimulator::event
// =========================================================================

bool MESpatialSimulator::event( )
{
	bool result = true;
	MESpatialLineage *lineage;
	long extant, parent;
	long i, j, k, m, r;
	
	if (mPureBirth || eventIsBirth()) {
		if (mMode == 0) {
			// Pick randomly from the boundaries between the extant
			// lineages and their empty neighbors

			// pick a neighbour randomly by:
			// first finding its extant lineage
			r = MERandom::iRan(mNeighbourCount);
			i = 0;
			k = 0;
			bool found = false;
			while (!found) {
		    	parent = mPhylogeny->getExtantLineage(i);
				lineage = static_cast<MESpatialLineage *>(
								mPhylogeny->getLineage(parent));
				extant = i;
				
				if (lineage->mCount > 0) {
					
					if ((k + lineage->mCount) > r)
						found = true;
					else
						k += lineage->mCount;
				}
				i++;
			}
			
			// then finding the neighbour
			k = r - k;
			
			j = 0;
			m = 0;
			found = false;
			while (!found) {
				if (lineage->mLinks[j] == -1) {
					if (m == k)
						found = true;
					m++;
				}
				if (!found)
					j++;
			}

			
			if (lineage->mLinks[j] != -1)
				throw CException("Error - lineage being born into an occupied niche");
				
		    lineage->mOffspringDirection = j;
		    result = lineageBirth(extant, 2, true);
			
		} else {

			extant = mPhylogeny->getRandomExtantIndex();
				
		    // Override to allow parent survival and check that the lineage can divide.
		    parent = mPhylogeny->getExtantLineage(extant);
		    lineage = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(parent));
		    
		    // Look for occupied cells in adjacent positions. This is a rather
		    // crude process. It may be possible to keep the adjacencies up to
		    // date on the fly.
		    long x, y;
		    for (i = 0; i < mNumAdjacencies; i++) {
				offsetCoordinates(i, lineage->mX, lineage->mY, x, y);
				lineage->mLinks[i] = findLineageAtPosition(x, y);
		    }
		    
		    // Pick a neighbour randomly and if it is blocked, discard the birth
		    r = MERandom::iRan(mNumAdjacencies);
		    if (lineage->mLinks[r] == -1) {
		        lineage->mOffspringDirection = r;
		        result = lineageBirth(extant, 2, true);
		    } else
		    	result = false;// else birth fails - do nothing

		}    
	} else {
		extant = mPhylogeny->getRandomExtantIndex();
		result = lineageDeath(extant);
	}
	
	return result;
}

// =========================================================================
//	MESpatialSimulator::timeToNextEvent
// =========================================================================

double MESpatialSimulator::timeToNextEvent()
{
	if (mMode == 0) {
		// count the number of empty neighbours for each extant lineage
		MESpatialLineage *lineage;
		int i, j, n = mPhylogeny->getNumExtantLineages();
		mNeighbourCount = 0;
		
		for (i = 0; i < n; i++) {
			lineage = static_cast<MESpatialLineage *>(
							mPhylogeny->getLineage(
									mPhylogeny->getExtantLineage(i)));
			
			lineage->mCount = 0;

		    // Look for occupied cells in adjacent positions. This is a rather
		    // crude process. It may be possible to keep the adjacencies up to
		    // date on the fly.
		    long x, y;
		    for (j = 0; j < mNumAdjacencies; j++) {
				offsetCoordinates(j, lineage->mX, lineage->mY, x, y);
				lineage->mLinks[j] = findLineageAtPosition(x, y);
				
				if (lineage->mLinks[j] == -1)
					lineage->mCount++;
		    }
		    
			mNeighbourCount +=  lineage->mCount;
		}
		
		mSumRate = ((double(mNeighbourCount) * mBirthRate) +
					(double(mPhylogeny->getNumExtantLineages()) * mDeathRate));
		
		return -log((double)MERandom::fRan()) / mSumRate;
			
    } else 
    	return MEBirthDeathSimulator::timeToNextEvent();
}

// =========================================================================
//	MESpatialSimulator::eventIsBirth
// =========================================================================

bool MESpatialSimulator::eventIsBirth()
{
	if (mMode == 0) {
		mRelativeBirthRate = (double(mNeighbourCount) * mBirthRate) / mSumRate;
		return MERandom::bRan(mRelativeBirthRate);
	} else
		return MEBirthDeathSimulator::eventIsBirth();
}

// =========================================================================
//	MESpatialSimulator::mutateLineage
// =========================================================================

void MESpatialSimulator::mutateLineage(long inParent, long inOffspring)
{
	MESpatialLineage *p = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(inParent));
	MESpatialLineage *o = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(inOffspring));
    
    offsetCoordinates(p->mOffspringDirection, p->mX, p->mY, o->mX, o->mY);
    
    int ir = getOppositeDirection(p->mOffspringDirection);
    p->mLinks[p->mOffspringDirection] = inOffspring; 
    o->mLinks[ir] = inParent;
    
    o->mOrder = mOrder;
    mOrder++;
}

// =========================================================================
//	MESpatialSimulator::cloneLineage
// =========================================================================

void MESpatialSimulator::cloneLineage(long inParent, long inOffspring)
{
	MESpatialLineage *p = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(inParent));
	MESpatialLineage *o = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(inOffspring));
    
    o->mX = p->mX;
    o->mY = p->mY;
    o->mOffspringDirection = p->mOffspringDirection;
    
    for (int i = 0; i < mNumAdjacencies; i++)
		o->mLinks[i] = p->mLinks[i];
		
	o->mOrder = -1;
}

// =========================================================================
//	MESpatialSimulator::newPhylogeny
// =========================================================================

MEPhylogeny* MESpatialSimulator::newPhylogeny()
{
    return new MESpatialPhylogeny();
}

// =========================================================================
//	MESpatialSimulator::offsetCoordinates
// =========================================================================

void MESpatialSimulator::offsetCoordinates(int inDirection, long inX, long inY, long &outX, long &outY)
{
    switch (mNumAdjacencies) {
        case 2: // 1-dimensional
            outY = inY;
            if (inDirection == 0) // E
                outX = inX + 1;
            else	// W
                outX = inX - 1;
        break;

        case 3: {// 2-dimensional triangular grid
        	bool xEven = inX % 2 == 0; // x is even
        	bool yEven = inY % 2 == 0; // y is even
        	
            if ((xEven && yEven) || (!xEven && !yEven)) {
            	// triangle has point upwards        
	            switch (inDirection) {
	                case 0: outX = inX + 1; outY = inY; 	break; // NE
	                case 1: outX = inX;		outY = inY - 1; break; // S
	                case 2: outX = inX - 1; outY = inY; 	break; // NW
	            }
			} else {
	            // triangle has point downward        
	            switch (inDirection) {
	                case 0: outX = inX; 	outY = inY + 1; break; // N
	                case 1: outX = inX + 1;	outY = inY; 	break; // SE
	                case 2: outX = inX - 1; outY = inY; 	break; // SW
	            }
			}
		} break;

        case 4: // 2-dimensional square grid (no diagonals)
            if (inDirection == 0) {	// N
                outX = inX;		outY = inY + 1;
            } else if (inDirection == 1) { // E
                outX = inX + 1;	outY = inY;	
            } else if (inDirection == 2) { // S
                outX = inX;		outY = inY - 1;
            } else if (inDirection == 3) { // W
                outX = inX - 1;	outY = inY;
            }
        break;

        case 6: // 2-dimensional hexagonal grid
            // This defines moves on a hexagonal board. The columns of hexagons
            // are lined up but the rows zig-zag.
            if (inX % 2 == 0) { // x is even
	            switch (inDirection) {
	                case 0: outX = inX; 	outY = inY + 1; break; // N
	                case 1: outX = inX + 1; outY = inY + 1; break; // NE
	                case 2: outX = inX + 1; outY = inY; 	break; // SE
	                case 3: outX = inX; 	outY = inY - 1; break; // S
	                case 4: outX = inX - 1; outY = inY; 	break; // SW
	                case 5: outX = inX - 1; outY = inY + 1; break; // NW
	            }
			} else {
	            switch (inDirection) {
	                case 0: outX = inX; 	outY = inY + 1; break; // N
	                case 1: outX = inX + 1; outY = inY; 	break; // NE
	                case 2: outX = inX + 1; outY = inY - 1; break; // SE
	                case 3: outX = inX; 	outY = inY - 1; break; // S
	                case 4: outX = inX - 1; outY = inY - 1; break; // SW
	                case 5: outX = inX - 1; outY = inY; 	break; // NW
	            }
			}
        break;

        case 8: // 2-dimensional square grid (with diagonals)
        	outX = inX;	outY = inY;
        	
            if (inDirection == 7 || inDirection == 0 || inDirection == 1) // NW, N, NE
                outY = inY + 1;
            if (inDirection == 1 || inDirection == 2 || inDirection == 3) // NE, E, SE
                outX = inX + 1;
            if (inDirection == 3 || inDirection == 4 || inDirection == 5) // SE, S, SW
                outY = inY - 1;
            if (inDirection == 5 || inDirection == 6 || inDirection == 7) // SW, W, NW
                outX = inX - 1;
        break;
    }
}

// =========================================================================
//	MESpatialSimulator::getOppositeDirection
// =========================================================================

int MESpatialSimulator::getOppositeDirection(int inDirection)
{
    int offset = mNumAdjacencies / 2;
    return (inDirection + offset) % mNumAdjacencies;
}

// =========================================================================
//	MESpatialSimulator::findLineageAtPosition
// =========================================================================

long MESpatialSimulator::findLineageAtPosition(long inX, long inY)
{
    // Brute force - eghh.
    
    long i = 0;
    bool found = false;
    MESpatialLineage *p;
     
    while (!found && i < mPhylogeny->getNumExtantLineages()) {
    	p = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(mPhylogeny->getExtantLineage(i)));
        if (p->mX == inX && p->mY == inY)
            found = true;
        else
            i++;
    }
    
    if (found)
        return i;
    else
        return -1;
}	

// =========================================================================
//	MESpatialSimulator::calcBalance
// =========================================================================

void MESpatialSimulator::calcBalance()
{
	int i, n = mPhylogeny->getNumLineages();
	MESpatialLineage *lineage;
	
	long descendant;
	
	for (i = 0; i < n; i++) {
		descendant = mPhylogeny->getDescendant(i);
		if (descendant != -1) { // not a tip
	    	lineage = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(i));
			lineage->mBalance = mPhylogeny->calcBalanceForLineage(i);
		}
	}
}	

#define PI_BY_3 1.047197551
#define PI_BY_6 PI_BY_3 / 2.0

// =========================================================================
//	MESpatialSimulator::drawSpatial
// =========================================================================

void MESpatialSimulator::drawSpatial(CReportRecord* ioReport)
{
	double gridSize = 30.0;
	double cX, cY;

    mU30 = gridSize * cos(PI_BY_6);
    mV30 = gridSize * sin(PI_BY_6);
    mU60 = gridSize * cos(PI_BY_3);
    mV60 = gridSize * sin(PI_BY_3);

	calcBalance();
	
	if (mNumAdjacencies == 2)
		throw CException("Unable to draw figure when number of adjacencies is 2. "
							"Let's face it, it would be a line.");
		
	CFigureElement *f = new CFigureElement("Skyline");
	f->setSize(kPixelPageUnits, 560.0, 560.0);

	CGroupElement *gridGroup = new CGroupElement();
	CGroupElement *branchGroup = new CGroupElement();
	CGroupElement *blobGroup = new CGroupElement();

	drawLineage(0, gridSize, gridGroup, branchGroup, blobGroup, cX, cY);

	f->add(gridGroup);
	f->add(branchGroup);
	f->add(blobGroup);
	f->setPreserveAspectRatio(kXMidYMidAspect);
	f->fitFigureInBounds();
	ioReport->add(f);
}	

// =========================================================================
//	MESpatialSimulator::drawLineage
// =========================================================================

void MESpatialSimulator::drawLineage(long inLineage, double inGridSize, 
											CComplexGraphicElement *ioGridGraphic,
											CComplexGraphicElement *ioBranchGraphic,
											CComplexGraphicElement *ioBlobGraphic,
											double &outX, double &outY)
{
	MESpatialLineage *lineage = static_cast<MESpatialLineage *>(mPhylogeny->getLineage(inLineage));
	
	double cX, cY;
	
	getGridCentreCoordinates(lineage->mX, lineage->mY, inGridSize, cX, cY);
	
	int i, n = 4;
	bool isUpsideDown;
	
	if (mNumAdjacencies == 6)
		n = 6;
	else if (mNumAdjacencies == 3) {
		n = 3;
    	bool xEven = lineage->mX % 2 == 0; // x is even
    	bool yEven = lineage->mY % 2 == 0; // y is even
    	
        isUpsideDown = !((xEven && yEven) || (!xEven && !yEven));
	}

	double vX, vY;
	
	if (lineage->mOrder != -1) {
		// Only need to draw the grid cell once
		CPathElement *path = new CPathElement(cX, cY);
/*		
		if (lineage->mBalance == -1)
			path->setStyle(1.0, "grey", kRoundLineCap, kRoundLineJoin, "none");
		else if (lineage->mBalance > 1.0)
			path->setStyle(1.0, "grey", kRoundLineCap, kRoundLineJoin, "none");
		else
			path->setStyle(1.0, "grey", kRoundLineCap, kRoundLineJoin, "red", 1.0 - lineage->mBalance);
*/
		if (lineage->mBalance == -1) {
			path->setStrokeStyle(1.0, kRoundLineCap, kRoundLineJoin);
			path->setStrokeColour(kGrayColour);
			path->setFillColour(kNoColour);
		} else if (lineage->mBalance > 0.5) {
			path->setStrokeStyle(1.0, kRoundLineCap, kRoundLineJoin);
			path->setStrokeColour(kGrayColour);
			path->setFillColour(kYellowColour, 2 * (lineage->mBalance - 0.5));
		} else {
			path->setStrokeStyle(1.0, kRoundLineCap, kRoundLineJoin);
			path->setStrokeColour(kGrayColour);
			path->setFillColour(kRedColour, 2 * (0.5 - lineage->mBalance));
		}
		
		for (i = 0; i < n; i++) {
			if (mNumAdjacencies == 3) {
				getTriangleVertexCoordinates(i, inGridSize, isUpsideDown, vX, vY);
			} else if (mNumAdjacencies == 6)
				getHexagonVertexCoordinates(i, inGridSize, vX, vY);
			else
				getSquareVertexCoordinates(i, inGridSize, vX, vY);
			
			if (i == 0 )
				path->moveTo(cX + vX, cY + vY);
			else
				path->lineTo(cX + vX, cY + vY);
		}
		
		path->setPathClosed(true);
		ioGridGraphic->add(path);
	
		if (mNumAdjacencies == 3) {
			if (isUpsideDown) {
				getTriangleVertexCoordinates(2, inGridSize, isUpsideDown, vX, vY);
				vX *= 0.85;	vY *= 0.85; vY += (inGridSize * 0.16);
			} else {
				getTriangleVertexCoordinates(1, inGridSize, isUpsideDown, vX, vY);
				vX *= 0.85;	vY *= 0.85;
			}
		} else if (mNumAdjacencies == 6) {
			getHexagonVertexCoordinates(4, inGridSize, vX, vY);
			vX *= 0.75;	vY *= 0.75;
		} else {
			getSquareVertexCoordinates(2, inGridSize, vX, vY);
			vX *= 0.75;	vY *= 0.75;
		}
		
		char str[64];
		sprintf(str, "%d", lineage->mOrder);
		
		CTextGraphicElement *text = new CTextGraphicElement(cX + vX, cY + vY, str);
		text->setFontStyle(inGridSize * 0.4, "Helvetica");
		ioBranchGraphic->add(text);
	}
		
	long descendant = mPhylogeny->getDescendant(inLineage);
	
	if (descendant == -1) {
		CCircleElement *circle = new CCircleElement(cX, cY, inGridSize / 5);
		
		if (lineage->mX == 0 && lineage->mY == 0) {
			circle->setStrokeColour(kNoColour);
			circle->setFillColour(kGreenColour);
		} else {
			circle->setStrokeColour(kNoColour);
			circle->setFillColour(kBlueColour);
		}
		
		ioBlobGraphic->add(circle);
	} else {
		double cX2, cY2;
		
		drawLineage(descendant, inGridSize, ioGridGraphic, ioBranchGraphic, ioBlobGraphic, 
					cX2, cY2);

		CLineElement *line = new CLineElement(cX, cY, cX2, cY2);
		line->setStrokeStyle(0.5, kRoundLineCap);
		line->setStrokeColour(kBlackColour);
		ioBranchGraphic->add(line);

		long sibling = mPhylogeny->getSibling(descendant);
		long d1;
		
		while (sibling != descendant) {
			drawLineage(sibling, inGridSize, ioGridGraphic, ioBranchGraphic, ioBlobGraphic,
						cX2, cY2);

			line = new CLineElement(cX, cY, cX2, cY2);
			line->setStrokeStyle(0.5, kRoundLineCap);
			line->setStrokeColour(kBlackColour);
			ioBranchGraphic->add(line);

			sibling = mPhylogeny->getSibling(sibling);
		}
	}
	
	outX = cX;
	outY = cY;
}	

// =========================================================================
//	MESpatialSimulator::getGridCentreCoordinates
// =========================================================================

void MESpatialSimulator::getGridCentreCoordinates(long inX, long inY, double inGridSize, 
													double &outX, double &outY)
{
	// Drawing axis has y increasing down the page so negate the y grid cordinate
	inY = -inY;

    switch (mNumAdjacencies) {
        case 2: // 1-dimensional
            outX = (double)inX *inGridSize*2.0;
            outY = 0.0;
        break;

        case 3: // 2-dimensional triangular grid
            outX = (double)inX * mU30;
            outY = (double)inY * (inGridSize + mV30);
        	if ((inX % 2 != 0 && inY % 2 ==0) ||
        		(inX % 2 == 0 && inY % 2 !=0))
        		outY -= inGridSize - mV30;
		break;

        case 4: // 2-dimensional square grid (no diagonals)
            outX = (double)inX * inGridSize * 2.0;
            outY = (double)inY * inGridSize * 2.0;
        break;

        case 6:	// 2-dimensional hexagonal grid
       		outX = (double)inX * 3.0 * mU60;
         	outY = (double)inY * 2.0 * mV60;
       		if (inX % 2 != 0)
        		outY += mV60;
		break;

        case 8: // 2-dimensional square grid (with diagonals)
            outX = (double)inX * inGridSize * 2.0;
            outY = (double)inY * inGridSize * 2.0;
        break;
    }
}

// =========================================================================
//	MESpatialSimulator::getTriangleVertexCoordinates
// =========================================================================

void MESpatialSimulator::getTriangleVertexCoordinates(int inVertex, double inGridSize, 
															bool inUpsideDown,
															double &outX, double &outY)
{
	if (inUpsideDown) {
		switch (inVertex) {
			case 0:
				outX = mU30;
				outY = -mV30;
			break;

			case 1:
				outX = 0.0;
				outY = inGridSize;
			break;

			case 2:
				outX = -mU30;
				outY = -mV30;
			break;
		}
	} else {
		switch (inVertex) {
			case 0:
				outX = mU30;
				outY = mV30;
			break;

			case 1:
				outX = -mU30;
				outY = mV30;
			break;

			case 2:
				outX = 0.0;
				outY = -inGridSize;
			break;
		}
	}
}

// =========================================================================
//	MESpatialSimulator::getHexagonVertexCoordinates
// =========================================================================

void MESpatialSimulator::getHexagonVertexCoordinates(int inVertex, double inGridSize, 
															double &outX, double &outY)
{
    switch (inVertex) {
        case 0:
            outX = inGridSize;
            outY = 0.0;
        break;

        case 1:
            outX = mU60;
            outY = mV60;
        break;

        case 2:
            outX = -mU60;
            outY = mV60;
        break;

        case 3:
            outX = -inGridSize;
            outY = 0.0;
        break;

        case 4:
            outX = -mU60;
            outY = -mV60;
        break;

        case 5:
            outX = mU60;
            outY = -mV60;
        break;
    }
}

// =========================================================================
//	MESpatialSimulator::getSquareVertexCoordinates
// =========================================================================

void MESpatialSimulator::getSquareVertexCoordinates(int inVertex, double inGridSize, 
														double &outX, double &outY)
{
    switch (inVertex) {
        case 0:
            outX = inGridSize;
            outY = inGridSize;
        break;

        case 1:
            outX = -inGridSize;
            outY = inGridSize;
        break;

        case 2:
            outX = -inGridSize;
            outY = -inGridSize;
        break;

        case 3:
            outX = inGridSize;
            outY = -inGridSize;
        break;
    }
}

// ---------------------------------------------------------------------------------
//		 MESpatialSimulator::describe
// ---------------------------------------------------------------------------------

void MESpatialSimulator::describe(CReportRecord *ioReport)
{
	CParagraphElement *heading = new CParagraphElement(kHeadingLevel2);
	heading->add("Spatial model");
	ioReport->add(heading);
	
	CParagraphElement *para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
	para->add("Settings:");
	ioReport->add(para);

	CListElement *list = new CListElement(kBullettedList, 2);
	para = new CParagraphElement();
	para->add("Number adjacencies = ");
	para->add(mNumAdjacencies);
	list->add(para);

	para = new CParagraphElement();
	para->add("Birth rate = ");
	para->add(mBirthRate);
	list->add(para);
	if (!mPureBirth) {
		para = new CParagraphElement();
		para->add("Death rate = ");
		para->add(mDeathRate);
		list->add(para);
	}
	if (mExtantStop > 0) {
		para = new CParagraphElement();
		para->add("Extant stop = ");
		para->add((int)mExtantStop);
		list->add(para);
	}
	if (mTimeStop > 0.0) {
		para = new CParagraphElement();
		para->add("Time stop = ");
		para->add(mTimeStop);
		list->add(para);
	}
	if (mExtantStop <= 0 && mTimeStop <= 0.0) {
		list->add("No stopping conditions specified");
	}	
	
	ioReport->add(list);
}
