/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEStream.cpp

				MEStream Class
				
 ******************************************************************************/

// MEStream provides a mechanism for streaming descendents of MEObject. It is 
// an abstract class and must be overridden to implement the low-level reading
// and writing.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"
#include "MEUndead.h"
#include "MEObject.h"
#include "MEList.h"
#include "MEStream.h"

#pragma mark === MEStream ===

// ---------------------------------------------------------------------------------
//		 MEStream
// ---------------------------------------------------------------------------------

MEStream::MEStream()
{
	initStream();
}

// ---------------------------------------------------------------------------------
//		 MEStream::initStream
// ---------------------------------------------------------------------------------

void MEStream::initStream()
{
	NewClass(cache, MEList());
	undead=NULL;
	lastLabel[0]='\0';
}

// ---------------------------------------------------------------------------------
//		 ~MEStream
// ---------------------------------------------------------------------------------

MEStream::~MEStream()
{
	if (cache!=NULL)
		delete cache;
}

// ---------------------------------------------------------------------------------
//		 MEStream::setUndead
// ---------------------------------------------------------------------------------

// The undead class takes an ObjectID and resurrects an appropriate (news)
// class. By default it can resurrect all the data classes in this library
// (data classes are derived from MEObject).
		
void MEStream::setUndead(MEUndead *inUndead)
{
	undead=inUndead;
}


// ---------------------------------------------------------------------------------
//		 MEStream::clearCache
// ---------------------------------------------------------------------------------

void MEStream::clearCache()
{
	if (cache!=NULL)
		cache->removeAll();
}


#pragma mark === Data Object reading ===
// These functions are used for the streaming of data objects.

// ---------------------------------------------------------------------------------
//		 MEStream::getObject
// ---------------------------------------------------------------------------------

// Gets an object from the stream

MEObject *MEStream::getObject()
{
	MEObject *obj;
	ObjectID id;
	
	if (cache==NULL)
		NewClass(cache, MEList());

	get("ID", &id);
	
	// If we have been given an undead object, use it...
	if (undead!=NULL)
		obj=undead->newByID(id.num);
	else // ...otherwise use the default static version
		obj=MEUndead::defaultNewByID(id.num);
	
	// add this object to the cache
	cache->add(obj, false);

	obj->readObject(this);
	obj->readData(this);
		
	return obj;
}

// ---------------------------------------------------------------------------------
//		 MEStream::getObjectRef
// ---------------------------------------------------------------------------------

// Finds an object in the cache by reference

MEObject *MEStream::getObjectRef(int ref)
{
	if (ref>=cache->getNumObjects())
		throw MEUnknownObject("Object reference in stream not found in cache");
		
	return cache->getObject(ref);
}

#pragma mark === Data Object writing ===

// These functions are used to put data objects onto a stream

// ---------------------------------------------------------------------------------
//		 MEStream::putObject
// ---------------------------------------------------------------------------------

void MEStream::putObject(MEObject *inObject)
{
	ThrowIfNull(inObject);
	
	cache->add(inObject, false);

	ObjectID id;
	id.num=inObject->getID();
	put("ID", id);
		
	inObject->writeObject(this);
	inObject->writeData(this);
}

