/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METaxonMatrix.cpp

				METaxonMatrix Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "METaxon.h"
#include "METaxonList.h"
#include "METaxonMatrix.h"
#include "MEIntArray.h"

#include "MEStream.h"

#pragma mark === METaxonMatrix Class ===

// ---------------------------------------------------------------------------------
//		 METaxonMatrix
// ---------------------------------------------------------------------------------

METaxonMatrix::METaxonMatrix() : MEMatrix()
{
	taxonList=NULL;
	taxonIndices=NULL;
}


// ---------------------------------------------------------------------------------
//		 METaxonMatrix
// ---------------------------------------------------------------------------------

METaxonMatrix::METaxonMatrix(int inDimension) : MEMatrix(inDimension)
{
	taxonList=NULL;
	taxonIndices=NULL;
}


// ---------------------------------------------------------------------------------
//		 METaxonMatrix
// ---------------------------------------------------------------------------------

METaxonMatrix::METaxonMatrix(int inNumRows, int inNumCols) : 
	MEMatrix(inNumRows, inNumCols)
{
	taxonList=NULL;
	taxonIndices=NULL;
}


// ---------------------------------------------------------------------------------
//		 ~METaxonMatrix
// ---------------------------------------------------------------------------------

METaxonMatrix::~METaxonMatrix()
{
	if (taxonList!=NULL && taxonList->getOwner()==this)
		delete taxonList;
		
	if (taxonIndices!=NULL)
		delete taxonIndices;
}


// ---------------------------------------------------------------------------------
//		 METaxonMatrix::setTaxonList
// ---------------------------------------------------------------------------------

void METaxonMatrix::setTaxonList(METaxonList *inTaxonList, bool isOwner, bool merge)
{
	char *name;
	int index, i;
	
	if ( inTaxonList!=NULL ) {	
		// If a taxon list already exists for this matrix, then check to see 
		// if the new list contains the same taxa.
		if (taxonList!=NULL && inTaxonList!=taxonList) {
			MEIntArray* newTaxonIndices;
			NewClass(newTaxonIndices, MEIntArray());
			
			for (i=0; i<numRows; i++) {
				name=getTaxonName(i); 
				if ((index=inTaxonList->findName(name))==-1) { // taxon missing
					if (merge) {
						// Get the taxon object from the old list
						METaxon *taxon=getTaxon(i);
						
						// Add it to the new list. Because the taxon objects
						// owner field now points to the new taxon list, it
						// won't be deleted by the old list's destructor.
						index=inTaxonList->add(taxon, true);
					} else
						throw MEBadNumTaxaMatrix();
				}
					
				newTaxonIndices->setValue(i, index);
			}
			
			if (taxonIndices!=NULL)
				delete taxonIndices;
			taxonIndices=newTaxonIndices;
			
		} else {
			// Given that we have no other information, we assume that the
			// rows are in the same order as the taxa in the new list.
			for (i=0; i<numRows; i++)
				setTaxonIndex(i, i);
		}
	}

	if (taxonList!=NULL && taxonList->getOwner()==this)
		delete taxonList;

	taxonList=inTaxonList;
	
	if (taxonList!=NULL && isOwner)
		taxonList->setOwner(this);
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::setTaxonIndex
// ---------------------------------------------------------------------------------

void METaxonMatrix::setTaxonIndex(int row, int index)
{
	if (taxonIndices==NULL)
		NewClass(taxonIndices, MEIntArray());
		
	taxonIndices->setValue(row, index);
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::getTaxonName
// ---------------------------------------------------------------------------------

char* METaxonMatrix::getTaxonName(int row)
{
	TestBounds(row, numRows);
	ThrowIfNull(taxonList);
	
	if (taxonIndices==NULL)
		return taxonList->getTaxonName(row);
	else
		return taxonList->getTaxonName(taxonIndices->getValue(row));
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::getTaxon
// ---------------------------------------------------------------------------------

METaxon* METaxonMatrix::getTaxon(int row)
{
	TestBounds(row, numRows);
	ThrowIfNull(taxonList);
	
	if (taxonIndices==NULL)
		return taxonList->getTaxon(row);
	else
		return taxonList->getTaxon(taxonIndices->getValue(row));
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::createLocalTaxonList
// ---------------------------------------------------------------------------------

// Replaces the existing taxonList with a local one, matching all the tips to
// the same named taxa in the new list. The matrix should already have a taxonlist
// (usually a global one) from which to create the local one.

void METaxonMatrix::createLocalTaxonList()
{
	char *name;
	int index;
	
	ThrowIfNull(taxonList);
	
	METaxonList *newTaxonList = new METaxonList();
	METaxon *taxon0, *taxon;

	int i, n = getNumRows();
	for (i = 0; i < n; i++) {
		taxon0 = getTaxon(i);
		taxon = new METaxon();
		taxon->clone(taxon0);
		taxon->setOwner(NULL);
		setTaxonIndex( i, newTaxonList->addTaxon(taxon));
	}
	
	if (taxonList->getOwner()==this)
		delete taxonList;

	taxonList=newTaxonList;
	taxonList->setOwner(this);
}


// ---------------------------------------------------------------------------------
//		 METaxonMatrix::findRow
// ---------------------------------------------------------------------------------

int METaxonMatrix::findRow(char *inName)
{
	ThrowIfNull(taxonList);
	int index=taxonList->findName(inName);
	
	if (index<0)
		return -1;
		
	return findRow(index);
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::findRow
// ---------------------------------------------------------------------------------

int METaxonMatrix::findRow(int inTaxonNo)
{
	ThrowIfNull(taxonList);
	
	if (taxonIndices==NULL)
		return inTaxonNo;
	
	int i=0;
	while (i<taxonIndices->getNumValues() && taxonIndices->getValue(i)!=inTaxonNo)
		i++;

	if (i==numRows)
		return -1;
		
	return i;
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::getRowList
// ---------------------------------------------------------------------------------

MEList *METaxonMatrix::getRowList()
{
	return taxonList;
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::getRowObject
// ---------------------------------------------------------------------------------

// Gets the object representing row. If the object at row does not exist then 
// this is created. If the rowList does not exist then it is created.

MEObject *METaxonMatrix::getRowObject(int row)
{
	if (taxonList==NULL) {
		METaxonList* newTaxonList;
		NewClass(newTaxonList, METaxonList());
		setTaxonList(newTaxonList, true);
	}
	
	MEObject *obj=NULL;
	int index=row;
	if (taxonIndices!=NULL)
		index = taxonIndices->getValue(row);
	if (index<taxonList->getNumObjects())
		obj=taxonList->getObject(index);
		
	if (obj==NULL) {
		NewClass(obj, METaxon());
		taxonList->setObject(index, obj, true);
	}
	
	return obj;
}

#pragma mark -
#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::readObject
// ---------------------------------------------------------------------------------

void METaxonMatrix::readObject(MEStream *stream)
{
	MEMatrix::readObject(stream);
	taxonList=(METaxonList *)stream->get("Taxa");
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrix::writeObject
// ---------------------------------------------------------------------------------

void METaxonMatrix::writeObject(MEStream *stream)
{
	MEMatrix::writeObject(stream);
	stream->put("Taxa", taxonList);
}
