/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METaxonMatrixList.cpp

				METaxonMatrixList Class
				
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "METaxonList.h"
#include "METaxonMatrixList.h"
#include "METaxonMatrix.h"

#include "MEStream.h"


#pragma mark === METaxonMatrixList ===

// ---------------------------------------------------------------------------------
//		 METaxonMatrixList
// ---------------------------------------------------------------------------------

METaxonMatrixList::METaxonMatrixList() : 
	MEMatrixList()
{
	taxonList=NULL;
}

// ---------------------------------------------------------------------------------
//		 ~METaxonMatrixList
// ---------------------------------------------------------------------------------

METaxonMatrixList::~METaxonMatrixList()
{
	if (taxonList!=NULL && taxonList->getOwner()==this)
		delete taxonList;
}

// ---------------------------------------------------------------------------------
//		METaxonMatrixList::createObject
// ---------------------------------------------------------------------------------

MEObject* METaxonMatrixList::createObject()
{
	MEObject *obj;
	NewClass(obj, METaxonMatrix());
	return obj;
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrixList::setTaxonList
// ---------------------------------------------------------------------------------

// The taxonList options are:
//   1) Each matrix owns its own taxonList, with matrixList's taxonList set to NULL
//   2) The matrixList owns a taxonList, and each matrix points to this
//   3) The matrices use an external taxonList (such as an Alignment), i.e. the
//      matrices point to it and the matrixList points to it but none of them owns it.
//   4) Matrices point to any manner of other taxonLists.

// Replaces the existing taxonList with a new one, calling setTaxonList for each matrix.
// If merge is true then taxa in a matrix that are missing from the new taxon list
// are added otherwise an exception is thrown.
void METaxonMatrixList::setTaxonList(METaxonList *inTaxonList, bool isOwner, bool merge)
{
	if (inTaxonList != NULL) {
		for (int i=0; i<numObjects; i++)
			((METaxonMatrix*)objects[i])->setTaxonList(inTaxonList, false, merge);

		if (taxonList!=NULL && taxonList->getOwner()==this)
			delete taxonList;

		taxonList=inTaxonList;
		if (isOwner)
			taxonList->setOwner(this);
	} else {
		METaxonMatrix *matrix;
		for (int i=0; i<numObjects; i++) {
			matrix = (METaxonMatrix *)getMatrix(i);
			matrix->createLocalTaxonList();
		}
		
		if (taxonList!=NULL && taxonList->getOwner()==this)
			delete taxonList;

		taxonList=NULL;
	}
}


// ---------------------------------------------------------------------------------
//		 METaxonMatrixList::newMatrix
// ---------------------------------------------------------------------------------

MEMatrix *METaxonMatrixList::newMatrix()
{
	METaxonMatrix *matrix=(METaxonMatrix*)createObject();

	matrix->setOwner(this);
	
	if (taxonList!=NULL)
		matrix->setTaxonList(taxonList, false);
	
	return matrix;
}

#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 METaxonMatrixList::readObject
// ---------------------------------------------------------------------------------

void METaxonMatrixList::readObject(MEStream *stream)
{
	MEMatrixList::readObject(stream);
	taxonList=(METaxonList *)stream->get("Taxa");
}

// ---------------------------------------------------------------------------------
//		 METaxonMatrixList::writeObject
// ---------------------------------------------------------------------------------

void METaxonMatrixList::writeObject(MEStream *stream)
{
	MEMatrixList::writeObject(stream);
	stream->put("Taxa", taxonList);
}


