/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METextStream.cpp

				METextStream Class

 ******************************************************************************/

// METextStream overrides MEStream to provide a text based format for
// streaming objects. This format looks somewhat like C source code.
// Note that floating point numbers may loose some precision during
// the streaming process. This format has the advantage that it can be
// inspected by humans, it is robust to editing and it can be served
// by web servers.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"
#include "MEUndead.h"
#include "MEObject.h"
#include "MEList.h"
#include "MEIntArray.h"
#include "MEDoubleArray.h"
#include "MELink.h"
#include "METree.h"
#include "MEMatrix.h"
#include "METextStream.h"

#pragma mark === METextStream ===

// ---------------------------------------------------------------------------------
//		 METextStream
// ---------------------------------------------------------------------------------

METextStream::METextStream() : MEStream()
{
	tabLevel=0;
	newLine[0]='\n';
	newLine[1]='\0';
}

// ---------------------------------------------------------------------------------
//		 ~METextStream
// ---------------------------------------------------------------------------------

METextStream::~METextStream()
{
}


// ---------------------------------------------------------------------------------
//		 METextStream::setNewLine
// --------------------------------------------------------------------------------
 
void METextStream::setNewLine(char *inNewLine)
{
	strncpy(newLine, inNewLine, 2);
	newLine[2]='\0';
}


#pragma mark === Low Level Reading ===

// ---------------------------------------------------------------------------------
//		 METextStream::skipSpace
// --------------------------------------------------------------------------------
 
// Skips over any space (plus tabs and returns) in the file. 

int METextStream::skipSpace()
{
	int i=-1;
	
	while (isspace(nextChar())) {
		readChar();
		i++;
	}
	
	return i;
}

// ---------------------------------------------------------------------------------
//		 METextStream::skipUntil
// ---------------------------------------------------------------------------------
 
// Skips over the file until the character delimiter is found.

void METextStream::skipUntil(char delimiter)
{
	char ch;
	
	do {
		ch=readChar();
	} while (ch!=delimiter);
}


// ---------------------------------------------------------------------------------
//		 METextStream::skipUntil
// ---------------------------------------------------------------------------------

// Skips over the file until a character from delimiters is found. Returns 
// the delimiter found.

char METextStream::skipUntil(char *delimiters)
{
	char ch;
	
	do {
		ch=readChar();
	} while (strchr(delimiters, ch)==NULL);
	
	return ch;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readInteger
// ---------------------------------------------------------------------------------

int METextStream::readInteger()
{
	char str[256];
	int d;
	
	readToken(str, 255);
	if (sscanf(str, "%d", &d)!=1)
		throw CIllegalInt();
		
	return d;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readInteger
// ---------------------------------------------------------------------------------

int METextStream::readInteger(char *delimiters)
{
	char str[256];
	int d;
	
	readToken(str, 255, delimiters);
	if (sscanf(str, "%d", &d)!=1)
		throw CIllegalInt();
		
	return d;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readDouble
// ---------------------------------------------------------------------------------

double METextStream::readDouble()
{
	char str[256];
	double f;
	
	readToken(str, 255);
	if (sscanf(str, "%lf", &f)!=1)
		throw CIllegalDbl();
		
	return f;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readDouble
// ---------------------------------------------------------------------------------

double METextStream::readDouble(char *delimiters)
{
	char str[256];
	double f;
	
	readToken(str, 255, delimiters);
	if (sscanf(str, "%lf", &f)!=1)
		throw CIllegalDbl();
		
	return f;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readToken
// ---------------------------------------------------------------------------------
 
// Reads a token into 'token' up to a maximum of maxLen characters. Will stop 
// when any character in delimiters is found. If upper is true, then the token 
// will be forced to upper case. Returns the character that delimited the token. 
// If the token begins with a quote char then all characters will be included 
// in token until a matching quote is found.

char METextStream::readToken(char *token, int maxLen, char *delimiters, bool upper)
{
	int n=0, space=0;
	char ch, ch2, *P, quoteChar='\0';
	bool done=false, quoted=false;
	
	skipSpace();
	
	P=token;	
	while (!done) {
		ch=nextChar();
		
		if (quoted && ch==quoteChar) { // Found the closing quote
			readChar();
			ch2=nextChar();
				
			if (ch==ch2) {	// A repeated quote character so add this to the token
				if (n<maxLen) {
					*P=ch;
					P++;
					n++;
				}
				readChar();
			} else {		// otherwise it terminates the token
				done=true;
				quoted=false;
			}
		} else if (n==0 && (ch=='\'' || ch=='"')) { 	// if the opening character is a quote
			readChar();
			quoted=true;						// read everything upto the closing quote
			quoteChar=ch;
			space=0;
		} else {
			if (quoted) { // compress multiple spaces into one
				readChar();
				if (isspace(ch)) {
					space++;
					ch=' ';
				} else
					space=0;
				
				if (ch=='\'' || ch=='"') {
					ch2=nextChar();
						
					if (ch==ch2) {	// A repeated quote character so add this to the token
						if (n<maxLen) {
							*P=ch;
							P++;
							n++;
						}
						readChar();
					}
				}
				
				// Even if we run out of space
				// we must read until the close quote
				if (space<2 && n<maxLen) { 
					if (upper)
						*P=toupper(ch);
					else
						*P=ch;
					P++;
					n++;
				}
			} else if (!done && !isspace(ch) && strchr(delimiters, ch)==NULL) {
				readChar();
				// Even if we run out of space
				// we must read until a delimiter is found
				if (n<maxLen) {
					if (upper)
						*P=toupper(ch);
					else
						*P=ch;
					P++;
					n++;
				}
			} else
				done=true;
		}
	}
	*P='\0';
		
	return ch;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readToken
// ---------------------------------------------------------------------------------
 
// Reads a token into 'token' up to a maximum of maxLen characters. Will stop 
// when any space is found. Will then skip any further space.If upper is true, 
// then the token will be forced to upper case. Returns character that delimited 
// the token.

char METextStream::readToken(char *token, int maxLen, bool upper)
{
	return readToken(token, maxLen, "", upper);
}


// ---------------------------------------------------------------------------------
//		 METextStream::readText
// ---------------------------------------------------------------------------------
 
// Similar to readToken but if the text is quoted, it reads the spaces as found.

char METextStream::readText(char *text, int maxLen, char *delimiters, bool upper)
{
	int n=0;
	char ch, ch2, *P, quoteChar='\0';
	bool done=false, quoted=false;
	
	skipSpace();
	
	P=text;	
	while (!done) {
		ch=nextChar();
		
		if (quoted && ch==quoteChar) { // Found the closing quote
			readChar();
			ch2=nextChar();
				
			if (ch==ch2) {	// A repeated quote character so add this to the token
				if (n<maxLen) {
					*P=ch;
					P++;
					n++;
				}
				readChar();
			} else {		// otherwise it terminates the token
				done=true;
				quoted=false;
			}
		} else if (n==0 && (ch=='\'' || ch=='"')) { 	// if the opening character is a quote
			readChar();
			quoted=true;						// read everything upto the closing quote
			quoteChar=ch;
		} else {
			if (quoted) { // compress multiple spaces into one
				readChar();
				
				if (ch=='\'' || ch=='"') {
					ch2=nextChar();
						
					if (ch==ch2) {	// A repeated quote character so add this to the token
						if (n<maxLen) {
							*P=ch;
							P++;
							n++;
						}
						readChar();
					}
				}
				
				// Even if we run out of space
				// we must read until the close quote
				if (n<maxLen) { 
					if (upper)
						*P=toupper(ch);
					else
						*P=ch;
					P++;
					n++;
				}
			} else if (!done && !isspace(ch) && strchr(delimiters, ch)==NULL) {
				readChar();
				// Even if we run out of space
				// we must read until a delimiter is found
				if (n<maxLen) {
					if (upper)
						*P=toupper(ch);
					else
						*P=ch;
					P++;
					n++;
				}
			} else
				done=true;
		}
	}
	*P='\0';
		
	return ch;
}


// ---------------------------------------------------------------------------------
//		 METextStream::readText
// ---------------------------------------------------------------------------------
 
char METextStream::readText(char *text, int maxLen, bool upper)
{
	return readText(text, maxLen, "", upper);
}


#pragma mark === Data Object Reading ===
// These functions are used for the streaming of data objects.
// If the object or item is not proceeded by label then the 
// passed parameter is not touched and the function returns false.

// ---------------------------------------------------------------------------------
//		 METextStream::getLabel
// ---------------------------------------------------------------------------------
// Looks to see if the next field matches 'label'. If not then the read label
// is stored and the function returns false meaning that this field is missing
// from the input. This allows backward compatibility with older files - if you
// add a field but it is missing in an older file, it will just return a default
// value.

bool METextStream::getLabel(char *label)
{	
	// if we don't have a label stored then read it
	if (lastLabel[0]=='\0') {
		skipSpace();
		char ch = nextChar();
		if (ch  != '}' && isalpha(ch)) { // not end of object
			readToken(lastLabel, 255, "=");
			skipUntil("=");
			skipSpace();
		} else
			return false;
	} 
	
	// is it the label we want?
	if (strcmp(label, lastLabel)!=0)
		return false;
	
	// clear the stored label
	lastLabel[0]='\0';
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

bool METextStream::get(char *label, int *i)
{
	if (getLabel(label)) {
		*i=readInteger(";");
		skipUntil(";");
			
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

bool METextStream::get(char *label, double *d)
{
	if (getLabel(label)) {
		*d=readDouble(";");
		skipUntil(';');
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

bool METextStream::get(char *label, bool *b)
{
	char token[256];
	
	if (getLabel(label)) {
		readToken(token, 255, ";", true);
		skipUntil(';');
		if (strcmp(token, "TRUE")==0)
			*b=true;
		else
			*b=false;
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

bool METextStream::get(char *label, char *ch)
{
	if (getLabel(label)) {
		*ch=readChar();
		skipUntil(';');
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

bool METextStream::get(char *label, char *str, int maxLen)
{
	if (getLabel(label)) {
		if (maxLen < 4) {
			char tmp[5];
			readText(tmp, 4, ";");
			skipUntil(';');
			if (strcmp(tmp, "null")==0)
				return false;
			
			strncpy(str, tmp, maxLen);
		} else {
			readText(str, maxLen, ";");
			skipUntil(';');
			if (strcmp(str, "null")==0)
				return false;
		}
		
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

bool METextStream::get(char *label, ObjectID *id)
{
	if (getLabel(label)) {
		char str[5];
		readToken(str, 5, ";");
		skipUntil(';');
		id->chars[0]=str[0];
		id->chars[1]=str[1];
		id->chars[2]=str[2];
		id->chars[3]=str[3];
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

// Find the index of the parameter in stringList and returns it in d.

bool METextStream::get(char *label, int *d, const char **stringList, const int numStrings)
{
	if (getLabel(label)) {
		char token[256];
		readToken(token, 255, ";");
		skipUntil(';');
		
		int n=-1;
		for (int i=0; i<numStrings; i++)
			if (strcmp(token, stringList[i])==0)
				n=i;
				
		if (n>-1)
			*d=n;
			
		return true;
	}
	
	return false;
}

// ---------------------------------------------------------------------------------
//		 METextStream::get
// ---------------------------------------------------------------------------------

MEObject * METextStream::get(char *label)
{
	MEObject *obj=NULL;
	
	if (getLabel(label)) {
		char ch=nextChar();
		if (ch=='{') { // This object is given in full
			readChar();
			obj=getObject();
			skipUntil('}');		
		} else if (isdigit(ch)) {
			int ref=readInteger(";");
			obj=getObjectRef(ref-1);
		} // otherwise assume obj is null
		
		skipUntil(';');
	}
	
	return obj;
}
		
// ---------------------------------------------------------------------------------
//		 METextStream::getList
// ---------------------------------------------------------------------------------

void METextStream::getList(MEList *inList)
{
	int count=0;
	get("Count", &count);
	
	if (count>0) {
		skipUntil('{');
		MEObject *obj;
		for (int i=0; i<count; i++) {
			skipSpace();
			char ch=readChar();
			if (ch=='{') {	// This object is given in full
				obj=getObject();
				skipUntil('}');		
			} else {
				int ref=readInteger(";}");
				obj=getObjectRef(ref-1);
			}
			inList->add(obj);	
			if (i<count-1)
				skipUntil(',');
		}
		skipUntil('}');
		skipUntil(';');
	}
}


// ---------------------------------------------------------------------------------
//		 METextStream::getIntArray
// ---------------------------------------------------------------------------------

void METextStream::getIntArray(MEIntArray *inList)
{
	int count=0;
	get("Count", &count);
	
	if (count>0) {
		skipUntil('{');
		int val;
		for (int i=0; i<count; i++) {
			skipSpace();
			val=readInteger(",}");
			inList->add(val);	
			if (i<count-1)
				skipUntil(',');
		}
		skipUntil('}');
		skipUntil(';');
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::getDoubleArray
// ---------------------------------------------------------------------------------

void METextStream::getDoubleArray(MEDoubleArray *inList)
{
	int count=0;
	get("Count", &count);
	
	if (count>0) {
		skipUntil('{');
		double val;
		for (int i=0; i<count; i++) {
			skipSpace();
			val=readDouble(",}");
			inList->add(val);	
			if (i<count-1)
				skipUntil(',');
		}
		skipUntil('}');
		skipUntil(';');
	}
}


#pragma mark === Tree Reading ===

// ---------------------------------------------------------------------------------
//		 METextStream::getTree
// ---------------------------------------------------------------------------------

// This reads in a tree. It can cope with white-space but otherwise conform
// to a strict format. Tips should be numbers, and branch lengths should be
// present (even if just 0.0).

void METextStream::getTree(METree *inTree)
{
	MELink *link, *link1;
	
	skipUntil("(");
	link=getNode(inTree);
	skipUntil(";");

	// prune out the temporary MELink that pointed to the root
	link1=link;
	while (link->next!=link1) 
		link=link->next;
		
	link->next=link1->next;
	inTree->removeLink(link1);
	
	// move on to the first branch in the node
	link=link->next;
	link->setIsStem(true);
	
	if (link->getNumBranches()<3) {
		// if the root node has less than 3 branches then this node is
		// removed. This makes the tree unrooted but we store the position
		// of the root by setting METree's root1 and root2 to point to the
		// nodes either side of the root.
		inTree->setIsRooted(true);
		
		// Set the tree's root pointers
		MELink* root1=link->branch;
		MELink* root2=link->next->branch;
		inTree->setRoot1(root1);
		inTree->setRoot2(root2);
		
		// Set the tree's root branch lengths
		inTree->setRootDistances(inTree->getBranchLength(root1->branchNo),
									inTree->getBranchLength(root2->branchNo));
		
		// Set the nodes either side of the root to point to each other
		root1->branch=root2;
		root2->branch=root1;
		
		// Remove root2's branch length from the tree's array
		inTree->removeBranchLength(root2->branchNo);
		
		// Set the new branches branchNo and the corresponding branchLength
		// This just uses root1's old branch length slot and sets it to the
		// sum of the two old branch lengths
		root2->branchNo=root1->branchNo;
		inTree->setBranchLength(root1->branchNo, 
								inTree->getRootDistance1() + inTree->getRootDistance2());
		
		inTree->removeInternalNode(link);
		inTree->removeLink(link->next);
		inTree->removeLink(link);
	} else {
		inTree->setIsRooted(false);
		inTree->setRoot1(link);
		inTree->setRoot2(link->branch);
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::getTip
// ---------------------------------------------------------------------------------
 
MELink *METextStream::getTip(METree *inTree)
{
	int tipNo=readInteger(":");
	return inTree->newTip(tipNo-1);
}


// ---------------------------------------------------------------------------------
//		 METextStream::getNode
// ---------------------------------------------------------------------------------
 
//	First this function creates a new MELink that will point back up at
//	the ancestoral node. Then it works through each branch in a node and 
//	calls readBranch to read them. By passing the root MELink, each branch
//	can hook on to the linked list and form a polytomy.

MELink *METextStream::getNode(METree *inTree)
{
	MELink *link;
	char ch;
	
	// This MELink is the one that points back up to the root
	link=inTree->newNode();
	
	getBranch(inTree, link);				
	ch=skipUntil(",");
	
	do {
		getBranch(inTree, link);
		ch=skipUntil(",)");
	} while (ch==',');

	if (ch!=')')
		throw MEBracketMissing();
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 METextStream::getBranch
// ---------------------------------------------------------------------------------
 
// Reads a branch in. This could be a node or a tip (calls getNode or getTip
// accordingly). It then reads the branch length and creates an MELink that will
// point at the new node or tip. It connects these two up and sets them to use
// the same branch number (an index to an array of branch lengths in the METree).
// Finally it inserts the new branch MELink into the polytomy given by inLink.

MELink *METextStream::getBranch(METree *inTree, MELink *inLink)
{
	double len, param=0.0;
	MELink *link, *branchLink;

	skipSpace();
	if (nextChar()=='(') {	/* is a node */
		readChar();
		link=getNode(inTree);
	} else {		/* is a tip */
		link=getTip(inTree);
	}
	
	if (readChar()!=':')
		skipUntil(":");

	len=readDouble(",);");
	// readChar();
	
	branchLink=inTree->newBranch(inLink, len);
	link->branchNo=branchLink->branchNo;
	link->branch=branchLink;
	branchLink->branch=link;
	
	return branchLink;
}	


#pragma mark === Matrix Reading ===

// ---------------------------------------------------------------------------------
//		 METextStream::getMatrix
// ---------------------------------------------------------------------------------

void METextStream::getMatrix(MEMatrix *inMatrix)
{
	ThrowIfNull(inMatrix);
	
	int nRows=inMatrix->getNumRows();
	int nCols=inMatrix->getNumCols();
	int r, c;
	double e;
	
	skipUntil("(");
	
	for (r=0; r<nRows; r++) {
		skipUntil("(");
		for (c=0; c<nCols; c++) {
			e=readDouble(",)");
			readChar();
			inMatrix->setElement(r, c, e);
		}
	}
	
	skipUntil(")");
	skipUntil(";");
}

#pragma mark === Low Level Writing ===

// ---------------------------------------------------------------------------------
//		 METextStream::writeString
// ---------------------------------------------------------------------------------

void METextStream::writeString(const char *st)
{
	writeText(st, strlen(st));
}


// ---------------------------------------------------------------------------------
//		 METextStream::writePString
// ---------------------------------------------------------------------------------

void METextStream::writePString(const char *st)
{
	writeText(st+1, st[0]);
}


// ---------------------------------------------------------------------------------
//		 METextStream::writeLabel
// ---------------------------------------------------------------------------------

// This differs from writeString because if st contains spaces (or optionally,
// punctuation (except for underscore '-')) then it is written in quotes.

void METextStream::writeLabel(const char *st, bool allowPunctuation)
{
	bool quote = false;
	const char *P=st;
	
	while (*P!='\0' && !quote) {
		if ( isspace(*P) || (!allowPunctuation && ispunct(*P) && *P != '_') )
			quote = true;
		P++;
	}
	
	if (quote) {
		writeChar('\'');
		writeText(st, strlen(st));
		writeChar('\'');
	} else {
		writeText(st, strlen(st));
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::writeSpaces
// ---------------------------------------------------------------------------------

void METextStream::writeSpaces(int n)
{
	int i;
	
	for (i=0; i<n; i++)
		writeChar(' ');
}


// ---------------------------------------------------------------------------------
//		 METextStream::writeTabs
// ---------------------------------------------------------------------------------

void METextStream::writeTabs(int n)
{
	int i;
	
	for (i=0; i<n; i++)
		writeChar('\t');
}


// ---------------------------------------------------------------------------------
//		 METextStream::writeNewLine
// ---------------------------------------------------------------------------------

void METextStream::writeNewLine()
{
	writeString(newLine);
}


#pragma mark === Data Object Writing ===

// These functions are used to put data objects onto a stream

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, int i)
{
	writeTabs(tabLevel);
	writeString(label);
	writeChar('=');
	writeInteger(i);
	writeChar(';');
	writeNewLine();
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, double d)
{
	writeTabs(tabLevel);
	writeString(label);
	writeChar('=');
	writeDouble(d);
	writeChar(';');
	writeNewLine();
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, bool b)
{
	writeTabs(tabLevel);
	writeString(label);
	writeChar('=');
	if (b)
		writeString("true");
	else
		writeString("false");
	writeChar(';');
	writeNewLine();
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, char ch)
{
	writeTabs(tabLevel);
	writeString(label);
	writeString("=\'");
	writeChar(ch);
	writeString("\';");
	writeNewLine();
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, const char *str)
{
	writeTabs(tabLevel);
	writeString(label);
	if (str!=NULL) {
		writeString("=\"");
		writeString(str);
		writeString("\";");
		writeNewLine();
	} else {
		writeString("=null;");
		writeNewLine();
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, const char *str, int length)
{
	writeTabs(tabLevel);
	writeString(label);
	if (str!=NULL && length > 0) {
		writeString("=\"");
		writeText(str, length);
		writeString("\";");
		writeNewLine();
	} else {
		writeString("=null;");
		writeNewLine();
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, const ObjectID id)
{
	writeTabs(tabLevel);
	writeString(label);
	writeString("=\'");
	writeChar(id.chars[0]);
	writeChar(id.chars[1]);
	writeChar(id.chars[2]);
	writeChar(id.chars[3]);
	writeString("\';");
	writeNewLine();
}

// ---------------------------------------------------------------------------------
//		 METextStream::put
// ---------------------------------------------------------------------------------

void METextStream::put(char *label, MEObject *inObject)
{
	writeTabs(tabLevel);
	writeString(label);
	
	if (inObject!=NULL) {
		int index=cache->findIndex(inObject);
			
		if (index<0) { // This object is not in the cache...
			writeString("={");
			writeNewLine();
			tabLevel++;
			putObject(inObject);
			tabLevel--;
			writeTabs(tabLevel);
			writeString("};");
			writeNewLine();
		} else { // This object is in the cache...
			// ...so just refer to it by its index
			writeString("=");
			writeInteger(index+1);
			writeChar(';');
			writeNewLine();
		}
	} else {
		writeString("=null;");
		writeNewLine();
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::putList
// ---------------------------------------------------------------------------------

void METextStream::putList(MEList *inList)
{
	int i, n;
	
	ThrowIfNull(inList);
	
	n=inList->getNumObjects();
	put("Count", n);
	if (n>0) {
		writeTabs(tabLevel);
		writeChar('{');
		writeNewLine();
		tabLevel++;
		for (i=0; i<n; i++) {
			writeTabs(tabLevel);

			int index=cache->findIndex(inList->getObject(i));
			
			if (index<0) { // This object is not in the cache...
				// ...and write it out in full.
				writeChar('{');
				writeNewLine();
				tabLevel++;
				putObject(inList->getObject(i));
				tabLevel--;
				writeTabs(tabLevel);
				writeChar('}');
			} else { // This object is in the cache...
				// ...so just refer to it by its index
				writeInteger(index+1);
			}

			if (i<n-1) {
				writeChar(',');
				writeNewLine();
			}
		}
		writeNewLine();
		tabLevel--;
		writeTabs(tabLevel);
		writeString("};");
		writeNewLine();
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::putIntArray
// ---------------------------------------------------------------------------------

void METextStream::putIntArray(MEIntArray *inList)
{
	int i, n;
	
	ThrowIfNull(inList);
	
	n=inList->getNumValues();
	put("Count", n);
	if (n>0) {
		writeTabs(tabLevel);
		writeChar('{');
		writeNewLine();
		tabLevel++;
		for (i=0; i<n; i++) {
			writeTabs(tabLevel);
			writeInteger(inList->getValue(i));
			if (i<n-1) {
				writeChar(',');
				writeNewLine();
			}
		}
		writeNewLine();
		tabLevel--;
		writeTabs(tabLevel);
		writeString("};");
		writeNewLine();
	}
}

// ---------------------------------------------------------------------------------
//		 METextStream::putDoubleArray
// ---------------------------------------------------------------------------------

void METextStream::putDoubleArray(MEDoubleArray *inList)
{
	int i, n;
	
	ThrowIfNull(inList);
	
	n=inList->getNumValues();
	put("Count", n);
	if (n>0) {
		writeTabs(tabLevel);
		writeChar('{');
		writeNewLine();
		tabLevel++;
		for (i=0; i<n; i++) {
			writeTabs(tabLevel);
			writeDouble(inList->getValue(i));
			if (i<n-1) {
				writeChar(',');
				writeNewLine();
			}
		}
		writeNewLine();
		tabLevel--;
		writeTabs(tabLevel);
		writeString("};");
		writeNewLine();
	}
}

#pragma mark === Tree Writing ===

// ---------------------------------------------------------------------------------
//		 METextStream::putTree
// ---------------------------------------------------------------------------------

void METextStream::putTree(METree *inTree)
{
	ThrowIfNull(inTree);
	
	writeTabs(tabLevel);
	
	writeChar('(');	
	if (inTree->getIsRooted()) {
		putNode(inTree, inTree->getRoot1());
		if (inTree->getHasLengths()) {
			writeChar(':');	
			writeDouble(inTree->getRootDistance1());
		}	
		writeChar(',');	
		
		putNode(inTree, inTree->getRoot2());
		if (inTree->getHasLengths()) {
			writeChar(':');	
			writeDouble(inTree->getRootDistance2());
		}
	} else {
		MELink *root=inTree->getRoot1();
		MELink *link=root->next;
		while (link!=root) {
			putBranch(inTree, link->branch);
			link=link->next;
			writeChar(',');	
		}
		putBranch(inTree, link->branch);
	}
	writeChar(')');	
	writeChar(';');
		
	writeNewLine();
}


// ---------------------------------------------------------------------------------
//		 METextStream::putBranch
// ---------------------------------------------------------------------------------

void METextStream::putBranch(METree *inTree, MELink *inLink)
{
	putNode(inTree, inLink);
		
	writeChar(':');	
	writeDouble(inTree->getBranchLength(inLink->branchNo));
}


// ---------------------------------------------------------------------------------
//		 METextStream::putNode
// ---------------------------------------------------------------------------------

void METextStream::putNode(METree *inTree, MELink *inLink)
{
	MELink *link=inLink->next;
	
	if (inLink->tipNo==-1) {
		writeChar('(');	
		
		while (link!=inLink) {
			putBranch(inTree, link->branch);
			link=link->next;
			if (link!=inLink)
				writeChar(',');	
		}
		
		writeChar(')');	
	} else
		writeInteger(inLink->taxonNo+1);
}

#pragma mark === Matrix Writing ===

// ---------------------------------------------------------------------------------
//		 METextStream::putMatrix
// ---------------------------------------------------------------------------------

void METextStream::putMatrix(MEMatrix *inMatrix)
{
	ThrowIfNull(inMatrix);
	
	int nRows=inMatrix->getNumRows();
	int nCols=inMatrix->getNumCols();
	int r, c;
	double e;
	
	writeTabs(tabLevel);
	writeChar('(');
	writeNewLine();
	tabLevel++;
	
	for (r=0; r<nRows; r++) {
		writeTabs(tabLevel);
		writeChar('(');
		for (c=0; c<nCols; c++) {
			e=inMatrix->getElement(r, c);
			writeDouble(e);
			if (c<nCols-1)
				writeChar(',');
		}
		writeChar(')');	
		writeNewLine();
	}
	
	tabLevel--;
	writeTabs(tabLevel);
	writeChar(')');	
	writeChar(';');
		
	writeNewLine();
}



