/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	METipDateTreeCriterion.cpp

		Class: METipDateTreeCriterion

	This class provides a criterion a tree branch lengths which will
	minimise the amount of change of rate across a tree. This is 
	described by Sanderson, 1997. This is a variant by me that allows
	dated tips

 ******************************************************************************/

#include <string.h>

#include "MEExceptions.h"
#include "METipDateTreeCriterion.h"

#include "MEPatternList.h"
#include "MELink.h"
#include "METree.h"

#pragma mark === METipDateTreeCriterion ===

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion
// ---------------------------------------------------------------------------------

METipDateTreeCriterion::METipDateTreeCriterion() : MESandersonTreeCriterion()
{
	mRate = 1.0;
}

// ---------------------------------------------------------------------------------
//		 ~METipDateTreeCriterion
// ---------------------------------------------------------------------------------

METipDateTreeCriterion::~METipDateTreeCriterion()
{
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::setTree
// ---------------------------------------------------------------------------------

void METipDateTreeCriterion::setTree(METree *inTree)
{
	ThrowIf(!inTree->getIsRooted());

	METreeCriterion::setTree(inTree);
	
	int n = mTree->getNumInternalNodes() + 1;
	
	if (n != numParameters) {
		if (parameters != NULL)
			delete [] parameters;
		numParameters = n;
		
		// one extra parameter for the rate, in case we have dated tips
		NewArray(parameters, double, numParameters);
	}
	
	mTree->lengthsToHeights(mRate);
	heightsToValues();

	invalidate();
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::getNumParameters
// ---------------------------------------------------------------------------------

int METipDateTreeCriterion::getNumParameters()
{
	ThrowIfNull(mTree);
	
	// an extra one for the rate
	return numParameters + 1;
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::getValue
// ---------------------------------------------------------------------------------

double METipDateTreeCriterion::getValue(int inIndex)
{
	if (inIndex == numParameters)
		return getRate();
		
	TestBounds(inIndex, numParameters);
	return parameters[inIndex];
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::getMinValue
// ---------------------------------------------------------------------------------

double METipDateTreeCriterion::getMinValue(int inIndex)
{
	if (inIndex == numParameters)
		return getMinRate();
		
	return 0.0;
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::getMaxValue
// ---------------------------------------------------------------------------------

double METipDateTreeCriterion::getMaxValue(int inIndex)
{
	if (inIndex == numParameters)
		return getMaxRate();
		
	return 1.0;
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::setValue
// ---------------------------------------------------------------------------------

void METipDateTreeCriterion::setValue(int inIndex, double inValue)
{
	if (inIndex == numParameters) {
		setRate(inValue);
		return;
	}
		
	TestBounds(inIndex, getNumParameters());
	parameters[inIndex] = inValue;
	invalidate();
}


// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::nodeHeightsToValues
// ---------------------------------------------------------------------------------

// Recursive function called by heightsToValues.

double METipDateTreeCriterion::nodeHeightsToValues(MELink *inNode)
{
	double height;
	
	if (inNode->tipNo==-1) {
		double maxHeight = 0.0;
		MELink *link, *branch;
		double height0;
		
		height0 = mTree->getNodeHeight(inNode->nodeNo);

		link=inNode->next;
		while (link!=inNode) {
			branch=link->branch;
			height = nodeHeightsToValues(branch);	
			if (height > maxHeight)
				maxHeight = height;
			link=link->next;
		}
		
		double value = height0 - maxHeight;
		ThrowIf(value<0.0);

		setValue(inNode->nodeNo + 1, value);
	} else
		height = mTree->getTipDate(inNode->tipNo);

	return height;
}


// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::heightsToValues
// ---------------------------------------------------------------------------------

// Calculate node values from node heights. This is mainly use to convert
// user branch lengths into starting node values.

void METipDateTreeCriterion::heightsToValues()
{
	MELink *root1 = mTree->getRoot1();
	MELink *root2 = mTree->getRoot2();
	
	double height = mTree->getRootHeight();
	double height1 = nodeHeightsToValues(root1);			
	double height2 = nodeHeightsToValues(root2);			
	if (height1 > height2)
		setValue(0, height - height1);
	else
		setValue(0, height - height2);
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::nodeValuesToHeightsBelow
// ---------------------------------------------------------------------------------

// Recursive function called by valuesToHeights when below a calibration.

double METipDateTreeCriterion::nodeValuesToHeights(MELink *inNode)
{
	double height;
	
	if (inNode->tipNo==-1) {
		double maxHeight = 0.0;
		MELink *link, *branch;
		
		link=inNode->next;
		
		while (link!=inNode) {
			branch=link->branch;
			height = nodeValuesToHeights(branch);
			if (height > maxHeight)
				maxHeight = height;
			link=link->next;
		}
		
		height = maxHeight + getValue(inNode->nodeNo + 1);
		mTree->setNodeHeight(inNode->nodeNo, height);
	} else
		height = mTree->getTipDate(inNode->tipNo);
		
	return height;
}

// ---------------------------------------------------------------------------------
//		 METipDateTreeCriterion::valuesToHeights
// ---------------------------------------------------------------------------------

void METipDateTreeCriterion::valuesToHeights()
{
	MELink *root1 = mTree->getRoot1();
	MELink *root2 = mTree->getRoot2();
	
	double height1 = nodeValuesToHeights(root1);			
	double height2 = nodeValuesToHeights(root2);
	
	if (height1 > height2)
		mTree->setRootHeight(height1 + getValue(0));
	else
		mTree->setRootHeight(height2 + getValue(0));
}

