/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METree.cpp

				METree Class
					
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "MEStream.h"
#include "MEObject.h"

#include "METaxon.h"
#include "METaxonList.h"
#include "MELink.h"
#include "METree.h"

#include "MEUtilities.h"
#include "MEMaths.h"
#include "MEBitset.h"
#include "MEIntArray.h"
#include "MEDoubleArray.h"
#include "MELabelArray.h"

#pragma mark === METree Class ===

// ---------------------------------------------------------------------------------
//		 METree
// ---------------------------------------------------------------------------------

METree::METree() : MEObject()
{
	taxonList=NULL;
	translationList=NULL;
	
	hasLengths=false;
	isRooted=false;
	
	numTips=0;
	numIntNodes=0;
	numNodes=0;
	numBranches=0;
	numLinks=0;
	
	root1=root2=NULL;
	dist1=dist2=0.0;
	rootDistsKnown=false;
	
	useNumericalTips=false;
	
	branchCapacity=INITIAL_SIZE;
	linkCapacity=INITIAL_SIZE;
	
	nodeHeights = NULL;
	heightCapacity=0;
	
	tipDates = NULL;
	tipDateCapacity=0;

	NewArray(links, MELink *, linkCapacity);
	for (int i=0; i<linkCapacity; i++)
		links[i]=NULL;
		
	branchCapacity = 0;
	branchLengths = NULL;

	nodeHeights = NULL;
	rootHeight = 0.0;
	
	navBitsKnown = false;
	flagMask = 0;
	
	nodeLabels = NULL;
	branchLabels = NULL;
}

// ---------------------------------------------------------------------------------
//		 METree::cloneTree
// ---------------------------------------------------------------------------------

// This function creates a clone of inTree. This does not make a copy of
// the taxonList or the translationList, rather just points to the same one
// as inTree.

void METree::cloneTree(METree *inTree)
{
	taxonList=inTree->getTaxonList();
	translationList=inTree->getTranslationList();
	
	hasLengths=inTree->getHasLengths();
	isRooted=inTree->getIsRooted();
	
	numTips=inTree->getNumTips();
	numIntNodes=inTree->getNumInternalNodes();
	numNodes=inTree->getNumNodes();
	numBranches=inTree->getNumBranches();
	numLinks=0;

	useNumericalTips=inTree->getUseNumericalTips();
	
	checkLinkCapacity(inTree->getLinkCapacity());
	checkBranchCapacity(inTree->getBranchCapacity());
	
	int i;
	
	for (i=0; i<numBranches; i++)
		branchLengths[i]=inTree->getBranchLength(i);
	
	// Create and clone all the links...
	MELink *link;
	for (i=0; i<inTree->getNumLinks(); i++) {
		link=createLink();
		link->clone(inTree->getLink(i));
	}
	
	MELink *link1;
	// ...and then connect them up
	for (i=0; i<numLinks; i++) {
		link=inTree->getLink(i);
		
		link1=link->getNext();
		if (link1!=NULL)
			links[i]->setNext(links[link1->getLinkNo()]);
		else
			links[i]->setNext(NULL);

		link1=link->getBranch();
		if (link1!=NULL)
			links[i]->setBranch(links[link1->getLinkNo()]);
		else
			links[i]->setBranch(NULL);
	}
	
	if (inTree->getRoot1()!=NULL)
		setRoot1(links[inTree->getRoot1()->getLinkNo()]);
	else
		root1=NULL;
	dist1=inTree->getRootDistance1();
	if (inTree->getRoot2()!=NULL)
		setRoot2(links[inTree->getRoot2()->getLinkNo()]);
	else
		root2=NULL;
	dist2=inTree->getRootDistance2();
	rootDistsKnown=inTree->getRootDistancesKnown();

	navBitsKnown = false;
	flagMask = inTree->getFlagMask();
}

// ---------------------------------------------------------------------------------
//		 ~METree
// ---------------------------------------------------------------------------------

METree::~METree()
{
	if (branchLengths!=NULL)
		delete[] branchLengths;
		
	if (nodeHeights!=NULL)
		delete[] nodeHeights;

	if (tipDates!=NULL)
		delete[] tipDates;

	if (links!=NULL) {
		for (int i=0; i<linkCapacity; i++)
			if (links[i]!=NULL)
				delete links[i];
		delete[] links;
	}
	
	if (taxonList!=NULL && taxonList->getOwner()==this)
		delete taxonList;		

	if (nodeLabels!=NULL)
		delete[] nodeLabels;

	if (branchLabels!=NULL)
		delete[] branchLabels;
}

// ---------------------------------------------------------------------------------
//		 METree::setTaxonList
// ---------------------------------------------------------------------------------

// Replaces the existing taxonList with a new one, matching all the tips to
// the same named taxa in the new list. If merge is true, then taxa missing 
// from the new taxonList are added otherwise an exception is thrown 
// (METaxonMissing). If isOwner is true then the tree will delete the old 
// taxonList. The new taxon list then replaces the old.
void METree::setTaxonList(METaxonList *inTaxonList, bool isOwner, bool merge)
{
	char *name;
	int index;
	
	if ( inTaxonList!=NULL ) {
		// If a taxon list already exists for this tree, then check to see 
		// if the new list contains the same taxa.
		if (taxonList!=NULL) {
			MELink *link=getFirstTip();
			while (link!=NULL) {
				name=taxonList->getTaxonName(link->taxonNo); 
				if ((index=inTaxonList->findName(name))==-1) { // taxon missing
					if (merge) {
						// Get the taxon object from the old list
						METaxon *taxon=taxonList->getTaxon(link->taxonNo);
						
						// Add it to the new list. Because the taxon objects
						// owner field now points to the new taxon list, it
						// won't be deleted by the old list's destructor.
						index=inTaxonList->add(taxon, true);
					} else
						throw METaxonMissing();
				}
					
				link->setTaxonNo(index);
				
				link=getNextTip(link);
			}
		} else {
			// Given that we have no other information, we assume that the
			// tips are in the same order as the taxa in the new list.
			MELink *link=getFirstTip();
			while (link!=NULL) {
				link->setTaxonNo(link->getTipNo());
				link=getNextTip(link);
			}
		}
	}
	
	if (taxonList!=NULL && taxonList->getOwner()==this)
		delete taxonList;

	taxonList=inTaxonList;
	if (taxonList!=NULL && isOwner)
		taxonList->setOwner(this);

	if (taxonList==NULL)
		useNumericalTips=true;
	else
		useNumericalTips=false;
}


// ---------------------------------------------------------------------------------
//		 METree::createLocalTaxonList
// ---------------------------------------------------------------------------------

// Replaces the existing taxonList with a local one, matching all the tips to
// the same named taxa in the new list. The tree should already have a taxonlist
// (usually a global one) from which to create the local one.

void METree::createLocalTaxonList()
{
	char *name;
	int index;
	
	ThrowIfNull(taxonList);
	
	METaxonList *newTaxonList = new METaxonList();
	METaxon *taxon0, *taxon;

	MELink *tip = getFirstTip();
	while (tip!=NULL) {
		taxon0 = taxonList->getTaxon(tip->taxonNo);
		taxon = new METaxon();
		taxon->clone(taxon0);
		taxon->setOwner(NULL);
		tip->taxonNo = newTaxonList->addTaxon(taxon);
		
		tip = getNextTip(tip);
	}
	
	if (taxonList->getOwner()==this)
		delete taxonList;

	taxonList=newTaxonList;
	taxonList->setOwner(this);
}


// ---------------------------------------------------------------------------------
//		 METree::setTranslationList
// ---------------------------------------------------------------------------------

// Sets the translationList. This is used to look up taxa numbers to get taxa
// names which in turn are looked up in the taxonList. This is used in the NEXUS
// file format.

void METree::setTranslationList(METaxonList *inTranslationList)
{
	translationList=inTranslationList;
}



// ---------------------------------------------------------------------------------
//		 METree::compareTaxa
// ---------------------------------------------------------------------------------

// Compares the taxa of the tree with those in inTaxonList. Returns 1 if they have 
// exactly the same taxa. Returns 0 if the inTaxonList contains all taxa that are 
// present in the tree (but with extras). Returns -1 in all other situations.

int METree::compareTaxa(METaxonList *inTaxonList)
{
	ThrowIfNull(inTaxonList);
	
	int n = getNumTips();
	for (int i = 0; i < n; i++) {
		if (inTaxonList->findName(getTipName(i))==-1)
			return -1;
	} 
	
	if (n < inTaxonList->getNumTaxa()) // n1 is a subset of n2
		return 0;
		
	return 1;	
}

#pragma mark === Node access ===

// ---------------------------------------------------------------------------------
//		 METree::getTaxon
// ---------------------------------------------------------------------------------

METaxon *METree::getTaxon(MELink* inTip)
{
	ThrowIf(!inTip->getIsTip());
	if (taxonList==NULL)
		return NULL;
	else
		return taxonList->getTaxon(inTip->taxonNo);
}


// ---------------------------------------------------------------------------------
//		 METree::getTipName
// ---------------------------------------------------------------------------------

char *METree::getTipName(int index)
{
	if (taxonList==NULL)
		return NULL;
	else
		return taxonList->getTaxonName(getTip(index)->taxonNo);
}


// ---------------------------------------------------------------------------------
//		 METree::getTipName
// ---------------------------------------------------------------------------------

char *METree::getTipName(MELink* inTip)
{
	ThrowIfNull(inTip);
	ThrowIf(!inTip->getIsTip());
	if (taxonList==NULL)
		return NULL;
	else
		return taxonList->getTaxonName(inTip->taxonNo);
}


// ---------------------------------------------------------------------------------
//		 METree::getTipNumber
// ---------------------------------------------------------------------------------

int METree::getTipNumber(MELink* inTip)
{
	ThrowIfNull(inTip);
	ThrowIf(!inTip->getIsTip());
	
	int tipNo = inTip->taxonNo;
	
	if (translationList!=NULL) {
		tipNo=translationList->findName(taxonList->getTaxonName(inTip->taxonNo));
		if (tipNo<0)
			throw METaxonMissing();
	}
	
	return tipNo;
}


// ---------------------------------------------------------------------------------
//		 METree::getTip
// ---------------------------------------------------------------------------------

MELink *METree::getTip(int index)
{
	int i=0;
	
	while (i<numLinks && links[i]->tipNo!=index) i++;
	
	if (i==numLinks)
		throw MENodeNotFound();
		
	return links[i];
}

// ---------------------------------------------------------------------------------
//		 METree::findTip
// ---------------------------------------------------------------------------------

//: Find the tip with a given name. Returns NULL if there is no taxon list.

MELink *METree::findTip(char *inName)
{
	if (taxonList==NULL)
		return NULL;
		
	MELink *link=getFirstTip();
	while (link!=NULL && strcmp(getTipName(link), inName)!=0)
		link=getNextTip(link);
		
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::findTip
// ---------------------------------------------------------------------------------

//: Find the tip with a given taxon number. This is the index into the taxon
//  list. Returns NULL if there is no taxon list.

MELink *METree::findTip(int inTaxonNo)
{
	if (taxonList==NULL)
		return NULL;
		
	MELink *link=getFirstTip();
	while (link!=NULL && link->taxonNo!=inTaxonNo)
		link=getNextTip(link);
		
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::countTipsWithMask
// ---------------------------------------------------------------------------------
// Counts the number of tips that don't have flags set according to flagMask

int METree::countTipsWithMask()
{
	int n = 0;
	MELink *link=getFirstTip();
	while (link!=NULL) {
		n++;
		link=getNextTip(link);
	}
	return n;	
}

// ---------------------------------------------------------------------------------
//		 METree::getFirstTip
// ---------------------------------------------------------------------------------

MELink *METree::getFirstTip()
{
	return getNextTip(NULL);
}

// ---------------------------------------------------------------------------------
//		 METree::getNextTip
// ---------------------------------------------------------------------------------

MELink *METree::getNextTip(MELink* previous)
{
	int counter;
	if (previous == NULL)
		counter = 0;
	else
		counter = previous->linkNo + 1;
		
	while ( counter<numLinks && (!links[counter]->getIsTip() ||
			(flagMask != 0 && !links[counter]->getFlag(flagMask))) ) 
		counter++;
	
	if (counter>=numLinks)
		return NULL;
	
	return links[counter];
}

// ---------------------------------------------------------------------------------
//		 METree::getInternalNode
// ---------------------------------------------------------------------------------

MELink *METree::getInternalNode(int index)
{
	int i=0;
	MELink *link=getFirstInternalNode();
	
	while ( link!=NULL && link->nodeNo!=index) {
		link=getNextInternalNode(link);
		i++;
	}
	
	if (link==NULL)
		throw MENodeNotFound();
		
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::getFirstInternalNode
// ---------------------------------------------------------------------------------

MELink *METree::getFirstInternalNode()
{
	return getNextInternalNode(NULL);
}

// ---------------------------------------------------------------------------------
//		 METree::getNextInternalNode
// ---------------------------------------------------------------------------------

MELink *METree::getNextInternalNode(MELink* previous)
{
	int counter;
	if (previous == NULL)
		counter = 0;
	else
		counter = previous->linkNo + 1;

	while (counter<numLinks && (!links[counter]->getIsStem() || links[counter]->getIsTip())) 
		counter++;
	
	if (counter>=numLinks)
		return NULL;
	
	return links[counter];
}

// ---------------------------------------------------------------------------------
//		 METree::getNode
// ---------------------------------------------------------------------------------

MELink *METree::getNode(int index)
{
	if (index<getNumTips())
		return getTip(index);
	else
		return getInternalNode(index-getNumTips());
}

// ---------------------------------------------------------------------------------
//		 METree::getBranch
// ---------------------------------------------------------------------------------

//: Gets a link at one end or the other of branch number 'index'.
MELink *METree::getBranch(int index)
{
	int i=0;
	
	while (i<numLinks && links[i]->branchNo!=index) i++;
	
	if (i==numLinks)
		throw MENodeNotFound();
		
	return links[i];
}

// ---------------------------------------------------------------------------------
//		 METree::getLink
// ---------------------------------------------------------------------------------

MELink *METree::getLink(int index)
{
	TestBounds(index, numLinks);
		
	return links[index];
}

// ---------------------------------------------------------------------------------
//    METree::putSplit
// ---------------------------------------------------------------------------------
void METree::putSplit(MEBitset *split, int idx)
//: put the taxon set to the "left" of branch 'idx' into the bitset 'split'.
{
	TestBounds(idx, numBranches);
	/*
		In MELink there is bool MELink::cladeContains(MELink *inLink).  
		This can be used to obtain which links are to the left(right) of a given
		branch.
		In MELink there is int tipNum.
		This can be used to identify the taxa for the split.
	*/
	MELink *lin, *lon;
	int 
		i, 
		n = getNumTips();
	// clear the bitset:
	split->clear();
	lin = getLink(idx);
	for (i = 0; i < n; i++) {
		lon = getTip(i);
		if (lin->cladeContains(lon))
			split->add(i);
	}
	// now check to see whether the n-th tip is in the split 
	// (for recoding according to Hendy's scheme):
	if (lin->cladeContains(lon)) {
		split->setComplement(n);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::getNodeDegree
// ---------------------------------------------------------------------------------

int METree::getNodeDegree(MELink *inNode)
{
	return inNode->getNumBranches();
}


// ---------------------------------------------------------------------------------
//		 METree::getMaxNodeDegree
// ---------------------------------------------------------------------------------

int METree::getMaxNodeDegree()
{
	int degree, maxDegree=0;
	
	MELink *link=getFirstInternalNode();
	while (link!=NULL) {
		degree=link->getNumBranches();
		if (degree>maxDegree)
			maxDegree=degree;
		link=getNextInternalNode(link);
	}
	
	return maxDegree;
}


// ---------------------------------------------------------------------------------
//		 METree::setRoot1
// ---------------------------------------------------------------------------------

void METree::setRoot1(MELink *inNode)
{
	ThrowIfNull(inNode);
	root1=inNode;
	rootDistsKnown=false;
	setStemLink(root1);
}

// ---------------------------------------------------------------------------------
//		 METree::getRootDistance1
// ---------------------------------------------------------------------------------

double METree::getRootDistance1()
{
	if (isRooted) {
		if (!rootDistsKnown) {
			// For visual reasons we don't want a zero branch length at the root
			adjustRootDistances(0.1);
		}

		if (hasLengths)
			return dist1;
		else
			return 1.0;
	} else
		return 0.0;
}

// ---------------------------------------------------------------------------------
//		 METree::setRoot2
// ---------------------------------------------------------------------------------

void METree::setRoot2(MELink *inNode)
{
	ThrowIfNull(inNode);
	root2=inNode;
	rootDistsKnown=false;
	setStemLink(root2);
}

// ---------------------------------------------------------------------------------
//		 METree::getRootDistance2
// ---------------------------------------------------------------------------------

double METree::getRootDistance2()
{
	if (isRooted) {
		if (!rootDistsKnown) {
			adjustRootDistances(0.1);
		}

		if (hasLengths)
			return dist2;
		else
			return 1.0;
	} else
		return getBranchLength(root2->branchNo);
}


// ---------------------------------------------------------------------------------
//		 METree::setRootDistances
// ---------------------------------------------------------------------------------

void METree::setRootDistances(double inDist1, double inDist2)
{
	dist1=inDist1;
	dist2=inDist2;
	rootDistsKnown=true;
}


#pragma mark === Node allocation ===

// ---------------------------------------------------------------------------------
//		 METree::newLinkObject
//	Override to allocate subclass
// ---------------------------------------------------------------------------------

MELink *METree::newLinkObject()
{
	MELink *link;
	
	NewClass(link, MELink());
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 METree::createLink
// ---------------------------------------------------------------------------------

// Creates a new link object. If one is available in the link array then
// use it otherwise call newLinkObject().

MELink *METree::createLink()
{
	checkLinkCapacity(numLinks+1);

	MELink *link = links[numLinks];
	if (link == NULL) {
		link=newLinkObject();
		
		link->setLinkNo(numLinks);
		
		links[numLinks]=link;
	}
	numLinks++;
	
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::newNode
// ---------------------------------------------------------------------------------

MELink *METree::newNode()
{
	MELink *link = createLink();
	
	link->setNodeNo(numIntNodes);
	link->setIsStem(true);

	numIntNodes++;
	numNodes++;
	
// Initially point the next field to itself - this will allow additional
// branches to be added into this loop.
	link->setNext(link); 
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 METree::newNode
// ---------------------------------------------------------------------------------

// Creates a new node with a particular node number. Care must be taken to
// keep these node numbers unique. Also, the full complement of node numbers
// should be present before any node access is attempted.

MELink *METree::newNode(int index)
{
	MELink *link = createLink();
	
	link->setNodeNo(index);
	link->setIsStem(true);
	
	if (numIntNodes<=index)
		numIntNodes=index+1;
		
	numNodes++;
	
// Initially point to the next field to itself - this will allow additional
// branches to be added into this loop.
	link->setNext(link); 
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 METree::newBranch
//  Adds branch link to an existing node.
// ---------------------------------------------------------------------------------

MELink *METree::newBranch(MELink *inLink, double len)
{
	MELink *link = createLink();

	link->setNodeNo(inLink->nodeNo);

// Insert the new branch link into the ancestor loop
	inLink->insertBranch(link);

	link->setBranchNo(addBranchLength(len));
	
	return link;
}


// ---------------------------------------------------------------------------------
//		 METree::newTip
// ---------------------------------------------------------------------------------

MELink *METree::newTip(char *inName)
{
	if (taxonList==NULL) {
		NewClass(taxonList, METaxonList());
		taxonList->setOwner(this);
	}
	
	int taxonNo = taxonList->findName(inName);

	if (taxonNo==-1) {
		if (taxonList->getOwner()!=this && taxonList->getTaxaDefined()) {
			throw METaxonMissing();
		} else {
			METaxon* taxon;
			NewClass(taxon, METaxon());
			taxon->setName(inName);
			taxonNo = taxonList->addTaxon(taxon);
		}
	} else {
		// see if another tip of this tree is using this taxon...
		int i = 0; 
		while (i < numTips && getTip(i)->taxonNo != taxonNo)
			i++;
		
		if (i < numTips) {
			// duplicate tip name
			if (taxonList->getOwner()!=this && taxonList->getTaxaDefined()) {
				// Taxa are fixed so throw an exception 
				throw MEDuplicateTaxon();
			} else {
				// Add a new taxon (with the same name). May cause complications
				// later if the application assumes that all taxa are unique.
				METaxon* taxon;
				NewClass(taxon, METaxon());
				taxon->setName(inName);
				taxonNo = taxonList->addTaxon(taxon);
			}
		}
	}
	
	MELink *tip = createLink();
	
	tip->setTaxonNo(taxonNo);
	tip->setTipNo(numTips);

	numTips++;
	numNodes++;
		
	return tip;
}


// ---------------------------------------------------------------------------------
//		 METree::newTip
// ---------------------------------------------------------------------------------

MELink *METree::newTip(int inTaxonNo)
{
	checkLinkCapacity(numLinks+1);

	ThrowIf(inTaxonNo<0);
	
	int taxonNo = inTaxonNo;
	
	if (translationList!=NULL) {
		if (taxonList==NULL) {
			NewClass(taxonList, METaxonList());
			taxonList->cloneList(translationList);
		}
	
		taxonNo=taxonList->findName(translationList->getTaxonName(inTaxonNo));
		if (taxonNo==-1) {
			if (taxonList->getOwner()!=this && taxonList->getTaxaDefined()) {
				throw METaxonMissing();
			} else {
				METaxon* taxon;
				NewClass(taxon, METaxon());
				taxon->setName(translationList->getTaxonName(inTaxonNo));
				taxonNo = taxonList->addTaxon(taxon);
			}
		}
	} else {
		if (taxonList==NULL) {
			NewClass(taxonList, METaxonList());
		}
	
		if (taxonNo >= taxonList->getNumTaxa() || taxonList->getTaxon(taxonNo)==NULL) {
			if (taxonList->getOwner()!=this && taxonList->getTaxaDefined()) {
				throw METaxonMissing();
			} else {
				METaxon* taxon;
				NewClass(taxon, METaxon());
				char name[32];
				sprintf(name, "%d", inTaxonNo+1);
				taxon->setName(name);
				taxonList->setObject(taxonNo, taxon);
			}
		}
	}
	
	MELink *tip = createLink();
		
	tip->setTaxonNo(taxonNo);
	tip->setTipNo(numTips);

	numTips++;
	numNodes++;
		
	return tip;
}


// ---------------------------------------------------------------------------------
//		 METree::removeLink
// ---------------------------------------------------------------------------------

// This function removes a link from the link list but doesn't delete it. 
// A call to createNode will resurrect it.

void METree::removeLink(MELink *inLink)
{
	// Find the node
//	int i=0;
//	while (i<numLinks && links[i]!=inLink) i++;
	
	int i = inLink->linkNo;
	if (i>=numLinks)
		throw MELinkNotFound();

	numLinks--;

	if (i!=numLinks) {
		// Swap the one at the end of the list with the vacant spot
		MELink *tmp = links[i];
		links[i] = links[numLinks];
		links[i]->setLinkNo(i);
		links[numLinks] = inLink;
		links[numLinks]->setLinkNo(numLinks);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::removeTip
// ---------------------------------------------------------------------------------

// This function renumbers the tips to eliminate the tip to which inLink
// belongs. The links for the tip must still be removed using removeLink and deleted.

void METree::removeTip(MELink *inLink)
{
	ThrowIf(!inLink->getIsTip());
	
	numTips--;
	numNodes--;
		
	int oldTipNo=inLink->getTipNo();
	if (oldTipNo==numTips) // Nothing to do
		return;
		
	MELink *link;
	for (int i=0; i < numLinks; i++) {
		link = getLink(i);
		if (link->getTipNo()==numTips) {
			link->setTipNo(oldTipNo);
			return;
		}
	}
}


// ---------------------------------------------------------------------------------
//		 METree::removeInternalNode
// ---------------------------------------------------------------------------------

// This function renumbers the internal node to eliminate the node to which inLink
// belongs. The links in the node must still be removed using removeLink and deleted.

void METree::removeInternalNode(MELink *inLink)
{
	ThrowIf(inLink->getIsTip());
	ThrowIf(numIntNodes<2);
	
	numIntNodes--;
	numNodes--;
		
	int oldNodeNo=inLink->getNodeNo();
	if (oldNodeNo==numIntNodes) // Nothing to do
		return;
		
	MELink *link;
	for (int i=0; i < numLinks; i++) {
		link = getLink(i);
		if (link->getNodeNo()==numIntNodes)
			link->setNodeNo(oldNodeNo);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::removeAllInternalNodes()
// ---------------------------------------------------------------------------------

// This function removes all the internal nodes, leaving only the tip nodes.
// All the redundant link objects are left in the array ready for reuse. This
// function is used before rebuilding a tree's shape from scratch.

void METree::removeAllInternalNodes()
{
	// First we must sort the tips so that they make up the first numTips links.
	int i = 0;
	MELink *tmp, *tip = getFirstTip();
	
	while (tip!=NULL) {
		if (tip->linkNo!=tip->tipNo) {
			tmp = links[tip->tipNo];
			links[tip->tipNo] = links[tip->linkNo];
			links[tip->linkNo] = tmp;
			
			tip->linkNo = tip->tipNo;
		}
		tip = getNextTip(tip);
	}
		
	numIntNodes = 0;
	numBranches = 0;
	numNodes = numTips;
	numLinks = numTips;
}


// ---------------------------------------------------------------------------------
//		 METree::removeAllNodes
// ---------------------------------------------------------------------------------

// This function removes all the nodes and tips (and their links).

void METree::removeAllNodes()
{
	numLinks=0;
	numTips=0;
	numIntNodes=0;
	numNodes=0;
	numBranches=0;
}


// ---------------------------------------------------------------------------------
//		 METree::checkLinkCapacity
// ---------------------------------------------------------------------------------

// Checks that the link array has enough space and resizes it as necessary
void METree::checkLinkCapacity(int requiredCapacity)
{
	if (requiredCapacity>linkCapacity) {
		MELink **newLinks;
		int i;
		
		// Increase capacity
		while (linkCapacity<requiredCapacity)
			linkCapacity+=INCREMENT_SIZE;
		
		// NewArray the new list.
		NewArray(newLinks, MELink*, linkCapacity);
		
		// Copy the existing list into the new list.
		for (i=0; i<numLinks; i++)
			newLinks[i]=links[i];
			
		// Set the remaining pointers to NULL.
		for (i=numLinks; i<linkCapacity; i++)
			newLinks[i]=NULL;
			
		// Free the old taxon list and copy over the new one.
		delete[] links;
		links=newLinks;
	}
}

#pragma mark === Branch lengths ===

// ---------------------------------------------------------------------------------
//		 METree::getBranchLength
// ---------------------------------------------------------------------------------

double METree::getBranchLength(int index)
{
	TestBounds(index, numBranches);
	
	if (hasLengths)
		return branchLengths[index];
	else
		return 1.0;
}


// ---------------------------------------------------------------------------------
//		 METree::getBranchLength
// ---------------------------------------------------------------------------------

double METree::getBranchLength(MELink *inLink)
{
	ThrowIfNull(inLink);
	
	return getBranchLength(inLink->branchNo);
}


// ---------------------------------------------------------------------------------
//		 METree::getBranchLengthSum
// ---------------------------------------------------------------------------------

double METree::getBranchLengthSum()
{
	int i, n = getNumBranches();
	double sum = 0.0;
	for (i=0; i<n; i++)
		sum += getBranchLength(i);
		
	return sum;
}


// ---------------------------------------------------------------------------------
//		 METree::getMeanBranchLength
// ---------------------------------------------------------------------------------

double METree::getMeanBranchLength()
{
	int n = getNumBranches();
	if (n==0)
		return 0.0;
	return getBranchLengthSum() / n;
}


// ---------------------------------------------------------------------------------
//		 METree::addBranchLength
// ---------------------------------------------------------------------------------

int METree::addBranchLength(double inLength)
{
	checkBranchCapacity(numBranches+1);
	branchLengths[numBranches]=inLength;
	numBranches++;
	
	return numBranches-1;
}


// ---------------------------------------------------------------------------------
//		 METree::setBranchLength
// ---------------------------------------------------------------------------------

void METree::setBranchLength(int index, double inLength)
{
	TestBounds(index, numBranches);
	
	branchLengths[index]=inLength;
}


// ---------------------------------------------------------------------------------
//		 METree::setBranchLengths
// ---------------------------------------------------------------------------------

void METree::setBranchLengths(double inLength)
{
	for (int i = 0; i < numBranches; i++)
		branchLengths[i] = inLength;
	rootDistsKnown = false;
}


// ---------------------------------------------------------------------------------
//		 METree::scaleBranchLengths
// ---------------------------------------------------------------------------------

void METree::scaleBranchLengths(double inFactor)
{
	for (int i = 0; i < numBranches; i++)
		branchLengths[i] *= inFactor;
		
	dist1 *= inFactor;
	dist2 *= inFactor;
}

// ---------------------------------------------------------------------------------
//		 METree::scaleBranchLengthsOfClade
// ---------------------------------------------------------------------------------

void METree::scaleBranchLengthsOfClade(MELink *inLink, double inFactor)
{
	ThrowIf(inLink->getIsTip());
	
	MELink *link = inLink->next;
	while (link != inLink) {
		branchLengths[link->branchNo] *= inFactor;
		if (!link->branch->getIsTip())
			scaleBranchLengthsOfClade(link->branch, inFactor);
		link = link->next;
	}
}

// ---------------------------------------------------------------------------------
//		 METree::raiseBranchLengths
// ---------------------------------------------------------------------------------

void METree::raiseBranchLengths(double inPower)
{
	for (int i = 0; i < numBranches; i++)
		branchLengths[i] = pow(branchLengths[i], inPower);

	rootDistsKnown = false;
}

// ---------------------------------------------------------------------------------
//		 METree::raiseRootDistances
// ---------------------------------------------------------------------------------

void METree::raiseRootDistances(double inPower)
{
	dist1 = pow(dist1, inPower);
	dist2 = pow(dist2, inPower);

	rootDistsKnown = true;
}

// ---------------------------------------------------------------------------------
//		 METree::removeBranchLength
// ---------------------------------------------------------------------------------

void METree::removeBranchLength(int index)
{
	TestBounds(index, numBranches);

	// Copy the one at the end of the list into the vacant spot
	branchLengths[index]=branchLengths[numBranches-1];
	numBranches--;
	
	// Change the branchNo's of any affected nodes in the list
	for (int i=0; i<numLinks; i++) {
		if (links[i]->branchNo==numBranches)
			links[i]->branchNo=index;
	}
}


// ---------------------------------------------------------------------------------
//		 METree::checkBranchCapacity
// ---------------------------------------------------------------------------------

// Checks that the branch list has enough space and resizes it as necessary
void METree::checkBranchCapacity(int requiredCapacity)
{
	if (requiredCapacity>branchCapacity) {
		double *newBranches;
		int i;
		
		// Increase capacity
		while (branchCapacity<requiredCapacity)
			branchCapacity+=INCREMENT_SIZE;
		
		// NewArray the new array.
		NewArray(newBranches, double, branchCapacity);
		
		// Copy the existing array into the new array.
		for (i=0; i<numBranches; i++)
			newBranches[i]=branchLengths[i];
			
		// Free the old branch list and copy over the new one.
		delete[] branchLengths;
		branchLengths=newBranches;
	}
}

// ---------------------------------------------------------------------------------
//		 METree::getNodeHeight
// ---------------------------------------------------------------------------------

double METree::getNodeHeight(int index)
{
	if (index < 0)
		return 0.0;
	
	if (nodeHeights != NULL)
		return nodeHeights[index];
	else
		return 1.0;
}


// ---------------------------------------------------------------------------------
//		 METree::setNodeHeight
// ---------------------------------------------------------------------------------

void METree::setNodeHeight(int index, double inHeight)
{
	TestBounds(index, numIntNodes);
	
	checkHeightCapacity(numIntNodes);
	nodeHeights[index]=inHeight;
}


// ---------------------------------------------------------------------------------
//		 METree::nodeLengthsToHeights
// ---------------------------------------------------------------------------------

// Recursive function called by lengthsToHeights.

double METree::nodeLengthsToHeights(MELink *inNode, double inRate)
{
	double maxHeight=0.0;
	
	if (inNode->tipNo==-1) {
		double height;
		MELink *link, *branch;
		
		link=inNode->next;
		
		while (link!=inNode) {
			branch = link->branch;
			
			height = getBranchLength(branch->branchNo) / inRate;
				
			height += nodeLengthsToHeights(branch, inRate);
			if (height>maxHeight)
				maxHeight = height;
			
			link=link->next;
		}
		
		setNodeHeight(inNode->nodeNo, maxHeight);
	} else
		maxHeight = getTipDate(inNode->tipNo);
	
	return maxHeight;
}


// ---------------------------------------------------------------------------------
//		 METree::lengthsToHeights
// ---------------------------------------------------------------------------------

// Set the heights of a rooted tree from the branch lengths.
// Doesn't require contemporaneous tips, however if the tree is not
// ultrametric (or has dated tips), then the heights will be approximate.

void METree::lengthsToHeights(double inRate)
{
	if (!getIsRooted())
		throw METreeNotRooted();
		
	double h1=getRootDistance1() / inRate;
	h1+=nodeLengthsToHeights(getRoot1(), inRate);

	double h2=getRootDistance2() / inRate;
	h2+=nodeLengthsToHeights(getRoot2(), inRate);
	
	if (h1>h2)
		rootHeight=h1;
	else
		rootHeight=h2;
}


// ---------------------------------------------------------------------------------
//		 METree::nodeHeightsToLengths
// ---------------------------------------------------------------------------------

// Recursive function called by heightsToLengths.

void METree::nodeHeightsToLengths(MELink* inNode, double inRate, bool isTesting)
{
	double dist;

	if (inNode->tipNo==-1) {
		MELink *link=inNode->next;
		MELink *branch;
		while (link!=inNode) {
			branch=link->branch;
			
			if (branch->tipNo==-1)
				dist=getNodeHeight(inNode->nodeNo) - getNodeHeight(branch->nodeNo);
			else
				dist=getNodeHeight(inNode->nodeNo) - getTipDate(branch->tipNo);
			
			if (dist < 0.0)
				throw MEBadNodeHeight();
			
			if (!isTesting)
				setBranchLength(branch->branchNo, dist * inRate);
			
			nodeHeightsToLengths(branch, inRate, isTesting);
			link=link->next;
		}
	} 
}


// ---------------------------------------------------------------------------------
//		 METree::heightsToLengths
// ---------------------------------------------------------------------------------

// Calculate branch lengths from node heights. This expects the tree to be
// rooted and the nodes to have heights.

void METree::heightsToLengths(double inRate)
{
	nodeHeightsToLengths(root1, inRate, true);
	nodeHeightsToLengths(root2, inRate, true);
	
	if (rootHeight < getNodeHeight(root1->nodeNo) || rootHeight < getNodeHeight(root2->nodeNo))
		throw MEBadNodeHeight();
	
	nodeHeightsToLengths(root1, inRate, false);
	nodeHeightsToLengths(root2, inRate, false);
	
	setRootDistances((rootHeight - getNodeHeight(root1->nodeNo)) * inRate, 
						(rootHeight - getNodeHeight(root2->nodeNo)) * inRate);	
	setBranchLength(root1->branchNo, dist1+dist2);
}


// ---------------------------------------------------------------------------------
//		 METree::checkBranchCapacity
// ---------------------------------------------------------------------------------

// Checks that the height list has enough space and resizes it as necessary
void METree::checkHeightCapacity(int requiredCapacity)
{
	if (requiredCapacity>heightCapacity) {
		double *newHeights;
		int i;
		
		// Increase capacity
		while (heightCapacity<requiredCapacity)
			heightCapacity+=INCREMENT_SIZE;
		
		// NewArray the new array.
		NewArray(newHeights, double, heightCapacity);
		
		// Free the old branch list and copy over the new one.
		delete[] nodeHeights;
		nodeHeights=newHeights;
	}
}

// ---------------------------------------------------------------------------------
//		 METree::getNodeHeight
// ---------------------------------------------------------------------------------

double METree::getTipDate(int index)
{
	if (index < 0)
		return 0.0;
	
	if (tipDates != NULL)
		return tipDates[index];
	else
		return 0.0;
}


// ---------------------------------------------------------------------------------
//		 METree::setTipDate
// ---------------------------------------------------------------------------------

void METree::setTipDate(int index, double inDate)
{
	TestBounds(index, numTips);
	
	checkTipDateCapacity(numTips);
	tipDates[index]=inDate;
}

// ---------------------------------------------------------------------------------
//		 METree::checkTipDateCapacity
// ---------------------------------------------------------------------------------

// Checks that the tip date list has enough space and resizes it as necessary
void METree::checkTipDateCapacity(int requiredCapacity)
{
	if (requiredCapacity>tipDateCapacity) {
		double *newTipDates;
		int i;
		
		// Increase capacity
		while (tipDateCapacity<requiredCapacity)
			tipDateCapacity+=INCREMENT_SIZE;
		
		// NewArray the new array.
		NewArray(newTipDates, double, tipDateCapacity);
		
		// Free the old branch list and copy over the new one.
		delete[] tipDates;
		tipDates=newTipDates;
	}
}

// ---------------------------------------------------------------------------------
//		 METree::reorderTaxa
// ---------------------------------------------------------------------------------

// Attempts to reorder the tips into something close to alphabetical order
void METree::reorderTaxa(TaxonOrder inTaxonOrder)
{
	reorderTaxaForNode(inTaxonOrder, root1);
	reorderTaxaForNode(inTaxonOrder, root2);
	
	// The root nodes point to each other so swap their order
	// when calling reorderTaxaCompare
	if (reorderTaxaCompare(inTaxonOrder, root2, root1) < 0) {
		MELink *link0 = root1;
		root1 = root2;
		root2 = link0;
		setRootDistances(dist2, dist1);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::reorderTaxaForNode
// ---------------------------------------------------------------------------------

void METree::reorderTaxaForNode(TaxonOrder inTaxonOrder, MELink *inNode)
{	
	if (!inNode->getIsTip()) {
		int i, j, n = getNodeDegree(inNode) - 1;
		MELink *link0 = inNode;
		MELink *link1 = link0->next;
		MELink *link2 = link1->next;
		
		for (i = 0; i < n; i++) {				
			reorderTaxaForNode(inTaxonOrder, link1->branch);
			link1 = link1->next;
		}

		link1 = link0->next;
		for (j = n - 1; j > 0; j--) {
			for (i = 0; i < j; i++) {				
				if (reorderTaxaCompare(inTaxonOrder, link1, link2) < 0) {
					link0->next = link2;
					link1->next = link2->next;
					link2->next = link1;
					
					link0 = link2;
					link2 = link1->next;
				} else {
					link0 = link1;
					link1 = link2;
					link2 = link2->next;
				}
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		 METree::reorderTaxaCompare
// ---------------------------------------------------------------------------------

int METree::reorderTaxaCompare(TaxonOrder inTaxonOrder, MELink *inNode1, MELink *inNode2)
{
	int result = 0;
	
	if (inTaxonOrder == kAlphabeticalOrder || inTaxonOrder == kRevAlphabeticalOrder) {
//		char *name1, *name2;
	
//		name1 = reorderTaxaForNode(inTaxonOrder, link1);
//		name2 = reorderTaxaForNode(inTaxonOrder, link2);
	} else if (inTaxonOrder == kIncNodeDensityOrder || inTaxonOrder == kDecNodeDensityOrder) {
		int density1 = calcNodeDensity(inNode1->branch);	
		int density2 = calcNodeDensity(inNode2->branch);	
		
		if (density1 < density2)
			result = 1;
		else if (density1 > density2)
			result = -1;
			
		if (inTaxonOrder == kDecNodeDensityOrder)
			result *= -1;
	} else { // kRandomOrder
		if (MERandom::bRan() < 0.5)
			result = 1;
		else
			result = -1;
	}
			
	return result;
}

#pragma mark === Node and Branch Labels ===

// ---------------------------------------------------------------------------------
//		 METree::getNumNodeLabelArrays
// ---------------------------------------------------------------------------------

int METree::getNumNodeLabelArrays()
{
	if (nodeLabels == NULL)
		return 0;
	else	
		return nodeLabels->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 METree::getNodeLabelName
// ---------------------------------------------------------------------------------

char *METree::getNodeLabelName(int index)
{
	if (nodeLabels == NULL)
		return NULL;
	else	
		return nodeLabels->getObjectName(index);
}

// ---------------------------------------------------------------------------------
//		 METree::getNodeLabelArray
// ---------------------------------------------------------------------------------

MELabelArray *METree::getNodeLabelArray(int index)
{
	if (nodeLabels == NULL)
		return NULL;
	else	
		return (MELabelArray *)nodeLabels->getObject(index);
}

// ---------------------------------------------------------------------------------
//		 METree::findNodeLabelArray
// ---------------------------------------------------------------------------------

int METree::findNodeLabelArray(char *inName)
{
	if (nodeLabels == NULL)
		return -1;
	else	
		return nodeLabels->findName(inName);
}

// ---------------------------------------------------------------------------------
//		 METree::addNodeLabelArray
// ---------------------------------------------------------------------------------

void METree::addNodeLabelArray(MELabelArray *inLabelArray)
{
	if (nodeLabels == NULL)	
		NewClass(nodeLabels, MEList());
	
	nodeLabels->add((MEObject *)inLabelArray);
}

// ---------------------------------------------------------------------------------
//		 METree::setNodeLabelArray
// ---------------------------------------------------------------------------------

void METree::setNodeLabelArray(int index, MELabelArray *inLabelArray)
{
	if (nodeLabels == NULL)	
		NewClass(nodeLabels, MEList());
		
	nodeLabels->setObject(index, (MEObject *)inLabelArray);
}

// ---------------------------------------------------------------------------------
//		 METree::getNumBranchLabelArrays
// ---------------------------------------------------------------------------------

int METree::getNumBranchLabelArrays()
{
	if (branchLabels == NULL)
		return 0;
	else	
		return branchLabels->getNumObjects();
}

// ---------------------------------------------------------------------------------
//		 METree::getBranchLabelName
// ---------------------------------------------------------------------------------

char *METree::getBranchLabelName(int index)
{
	if (branchLabels == NULL)
		return NULL;
	else	
		return branchLabels->getObjectName(index);
}

// ---------------------------------------------------------------------------------
//		 METree::getBranchLabelArray
// ---------------------------------------------------------------------------------

MELabelArray *METree::getBranchLabelArray(int index)
{
	if (branchLabels == NULL)
		return NULL;
	else	
		return (MELabelArray *)branchLabels->getObject(index);
}

// ---------------------------------------------------------------------------------
//		 METree::findBranchLabelArray
// ---------------------------------------------------------------------------------

int METree::findBranchLabelArray(char *inName)
{
	if (branchLabels == NULL)
		return -1;
	else	
		return branchLabels->findName(inName);
}

// ---------------------------------------------------------------------------------
//		 METree::addBranchLabelArray
// ---------------------------------------------------------------------------------

void METree::addBranchLabelArray(MELabelArray *inLabelArray)
{
	if (branchLabels == NULL)	
		NewClass(branchLabels, MEList());
	
	branchLabels->add((MEObject *)inLabelArray);
}

// ---------------------------------------------------------------------------------
//		 METree::setBranchLabelArray
// ---------------------------------------------------------------------------------

void METree::setBranchLabelArray(int index, MELabelArray *inLabelArray)
{
	if (branchLabels == NULL)	
		NewClass(branchLabels, MEList());
	
	branchLabels->setObject(index, (MEObject *)inLabelArray);
}

#pragma mark === Tree Rearrangement ===

// ---------------------------------------------------------------------------------
//		 METree::coaleseNodes
// ---------------------------------------------------------------------------------

// coalesces two nodes (passed as MELinks) and returns the new link which is 
// ancestral to both of them.
MELink* METree::coalesceNodes(MELink *l1, MELink *l2, double bLen1, double bLen2)
{
	// Create an internal node
	MELink *link=newNode();
	
	// Create a branch for l1 and insert it into the internal node
	MELink *branchLink=newBranch(link, bLen1);
	l1->branchNo=branchLink->branchNo;
	l1->branch=branchLink;
	branchLink->branch=l1;
	
	// Create a branch for l2 and insert it into the internal node
	branchLink=newBranch(link, bLen2);
	l2->branchNo=branchLink->branchNo;
	l2->branch=branchLink;
	branchLink->branch=l2;
	
	return link;
}

MELink* METree::coalesceNodes(int nodeNum1, int nodeNum2, double bLen1, double bLen2)
{
	return coalesceNodes(getNode(nodeNum1), getNode(nodeNum2), bLen1, bLen2);
}


// ---------------------------------------------------------------------------------
//		 METree::connectNodes
// ---------------------------------------------------------------------------------

// connects two nodes (passed as MELinks) creating a branch between them.
void METree::connectNodes(MELink *l1, MELink *l2, double bLen)
{

	l1->branchNo=addBranchLength(bLen);
	
	l2->branchNo=l1->branchNo;
	l2->branch=l1;
	l1->branch=l2;
}

void METree::connectNodes(int nodeNum1, int nodeNum2, double bLen)
{
	connectNodes(getNode(nodeNum1), getNode(nodeNum2), bLen);
}


// ---------------------------------------------------------------------------------
//		 METree::pruneSubtree
// ---------------------------------------------------------------------------------

// Used to prune out a subtree that is pointed to by inLink. 
// Will remove any redundant (2 link) internal nodes too. This function
// leaves the tree not fully connected. Either all the nodes in the
// subtree must be removed (removeSubtree) or it must be rejoined 
// somewhere else.
 
void METree::pruneSubtree(MELink *inLink)
{
	ThrowIfNull(inLink);
	ThrowIf(inLink->getIsTip());
	
	MELink *node=inLink->next;
	MELink *node1=node->branch;
	MELink *node2=node->next->branch;
	
	double sum=getBranchLength(node1->branchNo) + getBranchLength(node2->branchNo);
		
	node->removeBranch(inLink);
	if (getNodeDegree(node)==2) { // remove the node
		node1->branch=node2;
		node2->branch=node1;
		setBranchLength(node1->branchNo, sum);
		removeBranchLength(node2->branchNo);
		node2->branchNo=node1->branchNo;
		
		removeInternalNode(node);
		removeLink(node->next);
		removeLink(node);
	}
	
	navBitsKnown = false;
}

// ---------------------------------------------------------------------------------
//		 METree::graftSubtreeToBranch
// ---------------------------------------------------------------------------------
// Inserts a subtree on the branch given by inDest. inSource should be a hanging
// link pointing to the first node of the subtree. It is inserted with a zero
// branch length at the far end of the destination branch.

void METree::graftSubtreeToBranch(MELink *inSource, MELink *inDest, double inDist)
{
	ThrowIfNull(inSource);
	ThrowIfNull(inDest);
	ThrowIf(inSource->getIsTip());
	
	MELink *node=inDest->branch;
	
	inSource->next=inSource;
	inSource->setNodeNo(numIntNodes);
	numIntNodes++;
	numNodes++;
	
	MELink *link1 = createLink();
	inSource->insertBranch(link1);
	
	link1->branch = inDest;
	inDest->branch = link1;
	link1->branchNo = inDest->branchNo;
	
	MELink *link2 = createLink();
	inSource->insertBranch(link2);
	connectNodes(link2, node, 0.0);
	
	if (hasLengths) {
		double dist1, dist2;
		
		dist1 = getBranchLength(link1->branchNo);
		if (inDist < dist1) {
			dist2 = dist1 - inDist;
			dist1 = inDist;
		} else
			dist2 = 0.0;
			
		setBranchLength(link1->branchNo, dist1);
		setBranchLength(link2->branchNo, dist2);
	}
	
	navBitsKnown = false;
}

// ---------------------------------------------------------------------------------
//		 METree::graftSubtreeToNode
// ---------------------------------------------------------------------------------
// Inserts a subtree into the node given by inDest. inSource should be a hanging
// link pointing to the first node of the subtree.

void METree::graftSubtreeToNode(MELink *inSource, MELink *inDest)
{
	ThrowIfNull(inSource);
	ThrowIfNull(inDest);
	ThrowIf(inSource->getIsTip());
	
	inDest->insertBranch(inSource);
	
	navBitsKnown = false;
}

// ---------------------------------------------------------------------------------
//		 METree::disposeSubtree
// ---------------------------------------------------------------------------------
// Removes and disposes of a subtree given by inSource. This function will
// remove all the nodes, tips, links and branches from the tree. The subtree
// must have been pruned from the tree prior to this function being called.


void METree::disposeSubtree(MELink *inSource)
{
	ThrowIfNull(inSource);
	ThrowIf(inSource->getIsTip());
	
	disposeSubtreeNode(inSource->branch);
	removeBranchLength(inSource->branchNo);
	removeLink(inSource);
	
	navBitsKnown = false;
}

void METree::disposeSubtreeNode(MELink *inSource)
{
	if (inSource->getIsTip()) {
		removeTip(inSource);
		removeLink(inSource);
	} else {
		MELink *link = inSource->next;
		MELink *link0;
		
		while (link!=inSource) {
			disposeSubtreeNode(link->branch);
			removeBranchLength(link->branchNo);
			link0 = link;
			link = link->next;
			
			removeLink(link0);
		}
		
		removeLink(inSource);
	}
}
	
// ---------------------------------------------------------------------------------
//		 METree::cutAndPasteToBranch
// ---------------------------------------------------------------------------------
void METree::cutAndPasteToBranch(MELink *inSource, MELink *inDest, double inDist)
{
	ThrowIfNull(inSource);
	ThrowIfNull(inDest);
	ThrowIf(inSource->getIsTip());
	ThrowIf(inDest->getIsTip());
	
	if (!navBitsKnown)
		calcNavBits();
		
	// Check to see that the subtree given by source does not contain inDest.
	MEBitset *navBits = inSource->getNavBits();
	if (navBits->contains(inDest->getNodeNo() + getNumTips()))
		throw MENodeNotFound();

	pruneSubtree(inSource);
	graftSubtreeToBranch(inSource, inDest, inDist);
	
	MELink *link1 = getRoot1();
	MELink *link2 = getRoot2();
	ThrowIf(link1->linkNo >= numLinks && link2->linkNo >= numLinks);
	if ( link1->linkNo >= numLinks ) {
		link1 = link2;
		link2 = link2->branch;
	} else if ( link2->linkNo >= numLinks ) {
		link2 = link1->branch;
	} else if ( link1->branch != link2 ) {
		link2 = link1->branch;
	}
	setRoot1(link1);
	setRoot2(link2);
}

// ---------------------------------------------------------------------------------
//		 METree::cutAndPasteToBranch
// ---------------------------------------------------------------------------------
void METree::cutAndPasteToNode(MELink *inSource, MELink *inDest)
{
	ThrowIf(inSource->getIsTip());
	ThrowIf(inDest->getIsTip());
	
	if (!navBitsKnown)
		calcNavBits();
		
	// Check to see that the subtree given by source does not contain inDest.
	MEBitset *navBits = inSource->getNavBits();
	if (navBits->contains(inDest->getNodeNo() + getNumTips()))
		throw MENodeNotFound();

	pruneSubtree(inSource);
	graftSubtreeToNode(inSource, inDest);
	
	MELink *link1 = getRoot1();
	MELink *link2 = getRoot2();
	ThrowIf(link1->linkNo >= numLinks && link2->linkNo >= numLinks);
	if ( link1->linkNo >= numLinks ) {
		link1 = link2;
		link2 = link2->branch;
	} else if ( link2->linkNo >= numLinks ) {
		link2 = link1->branch;
	} else if ( link1->branch != link2 ) {
		link2 = link1->branch;
	}
	setRoot1(link1);
	setRoot2(link2);
}

// ---------------------------------------------------------------------------------
//		 METree::cutSubtree
// ---------------------------------------------------------------------------------
void METree::cutSubtree(MELink *inSource)
{
	ThrowIf(inSource->getIsTip());
	
	pruneSubtree(inSource);
	disposeSubtree(inSource);
	
	MELink *link1 = getRoot1();
	MELink *link2 = getRoot2();
	ThrowIf(link1->linkNo >= numLinks && link2->linkNo >= numLinks);
	if ( link1->linkNo >= numLinks ) {
		link1 = link2;
		link2 = link2->branch;
	} else if ( link2->linkNo >= numLinks ) {
		link2 = link1->branch;
	} else if ( link1->branch != link2 ) {
		link2 = link1->branch;
	}
	setRoot1(link1);
	setRoot2(link2);
}

// ---------------------------------------------------------------------------------
//		 METree::collapseSubtree
// ---------------------------------------------------------------------------------

//: Collapses an entire subtree down to one polytomy.
void METree::collapseSubtree(MELink *inLink) 
{
	MELink *link0 = inLink->branch;
	ThrowIf(link0->getIsTip());
	
	MELink *link = link0->next;
	MELink *link1;
	
	while (link!=link0) {
		link1 = link;
		link = link->next;
		
		if (!link1->branch->getIsTip())
			collapseSubtreeNode(link1);
	}
}

void METree::collapseSubtreeNode(MELink *inLink) 
{
	MELink *link0 = inLink->branch;
	
	MELink *link = link0->next;
	MELink *link1;
	
	while (link!=link0) {
		link1 = link;
		link = link->next;
		
		if (!link1->branch->getIsTip())
			collapseSubtreeNode(link1);
	}
	
	collapseBranch(inLink);
}

// ---------------------------------------------------------------------------------
//		 METree::collapseBranch
// ---------------------------------------------------------------------------------

//: Removes an internal branch, merging the nodes at each end
void METree::collapseBranch(MELink *inLink) 
{
	ThrowIf(inLink->tipNo!=-1);
	
	MELink *link0=inLink->branch;
	ThrowIf(link0->tipNo!=-1);
	
	// Prune out the two links making up the branch.
	MELink *link1=inLink->getPrevious();	
	MELink *link2=link0->getPrevious();
	
	link1->setNext(link0->next);
	link2->setNext(inLink->next);
	
	// Set nodeNo of the merged links of the second node to that of the first.
	MELink *link3=link1->next;
	do {
		link3->setNodeNo(link1->nodeNo);
		link3=link3->next;
	} while (link3!=link2);
	
	// This function removes the internal node of link0 from the tree
	// by reallocating its node number to another node and reducing
	// numInternalNodes by 1. 
	removeInternalNode(link0);

	if (root1==inLink || root1==link0)
		outgroupTree(link1);	

	if (hasLengths)
		removeBranchLength(inLink->branchNo);

	// Now remove the two extraneous links from the link list and dispose of them.
	removeLink(inLink);
	removeLink(link0);
}

// ---------------------------------------------------------------------------------
//		 METree::collapsePolytomies
// ---------------------------------------------------------------------------------

//: Collapses near zero branch lengths into polytomies 
//  Any internal branch lengths <= inThreshold (which defaults to 0.0)
//  will be collapsed to give a polytomy. This function looses phylogenetic
//  structure and cannot be reversed. The resolvePolytomy function is a sort
//  of opposite function but it resolves the polytomies arbitrarily so it
//  will not recover the previous structure.
void METree::collapsePolytomies(double inThreshold) 
{
	bool found;
	MELink *link;
	
	do {
		found=false;
		link=getFirstInternalNode();
		while (link!=NULL && !found) {
			if (getBranchLength(link->branchNo) <= inThreshold) {
				collapseBranch(link);
				found=true;
			}
			if (!found)
				link=getNextInternalNode(link);
		}	
	} while (found);
}

// ---------------------------------------------------------------------------------
//		 METree::resolveTree
// ---------------------------------------------------------------------------------

//: Resolves all polytomies in this tree randomly.
void METree::resolveTree(double inLength) 
{
	if (getIsRooted()) {
		resolveSubtreeNode(root1, inLength);
		resolveSubtreeNode(root2, inLength);
	} else {
		resolveSubtreeNode(root1, inLength);
		resolveSubtreeNode(root1->branch, inLength);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::resolveSubtree
// ---------------------------------------------------------------------------------

//: Resolves all polytomies in a subtree randomly.
void METree::resolveSubtree(MELink *inLink, double inLength) 
{
	ThrowIf(inLink->branch->getIsTip());
	resolveSubtreeNode(inLink, inLength);
}

void METree::resolveSubtreeNode(MELink *inLink, double inLength) 
{
	MELink *link0 = inLink->branch;

	if (link0->getIsTip())
		return;
			
	MELink *link = link0->next;
	
	while (link!=link0) {
		if (!link->branch->getIsTip())
			resolveSubtreeNode(link, inLength);
			
		link = link->next;
	}
	
	resolveNode(link0, inLength);
}

// ---------------------------------------------------------------------------------
//		 METree::resolveSubtree
// ---------------------------------------------------------------------------------

//: Resolves a polytomous node.
void METree::resolveNode(MELink *inNode, double inLength) 
{
	int degree = getNodeDegree(inNode);
	if (degree<4)
		return; // nothing to do
		
	int i, card, n = degree-1;
	bool *leftSet = new bool [n];
	
	MELink *link0 = inNode->next;
	MELink *link, *link1;
	
	// randomly assign branches to the left and right side of the node
	do {
		card = 0;
		link = link0;
		for (i=0; i<n; i++) {
			leftSet[i] = MERandom::bRan();
			if (leftSet[i])
				card++;
			
			link = link->next;
		}
	} while (card==0 || card==n);
	
	inNode->next = inNode;
	
	MELink *leftLink, *rightLink;
	MELink *leftNode, *rightNode;
	
	if (card > 1) { // There is more than 1 branch on the left
		leftLink = createLink();
		leftNode = newNode();
		connectNodes(leftLink, leftNode, inLength);
	}
	
	if (card < n - 1) { // There is more than 1 branch on the right
		rightLink = createLink();
		rightNode = newNode();
		connectNodes(rightLink, rightNode, inLength);
	}
			
	link = link0;
	for (i=0; i<n; i++) {
		link1 = link;
		link = link->next;
		
		if (leftSet[i]) {
			if (card > 1) { // There is more than 1 branch on the left
				leftNode->insertBranch(link1);
				link1->setNodeNo(leftNode->nodeNo);
			} else
				leftLink = link1;
		} else {
			if (card < n - 1) { // There is more than 1 branch on the right
				rightNode->insertBranch(link1);
				link1->setNodeNo(rightNode->nodeNo);
			} else
				rightLink = link1;
		}			
	}
		
	if (card > 2) // There is more than 2 branches on the left
		resolveNode(leftNode, inLength);
	if (card < n - 2) // There is more than 2 branches on the right
		resolveNode(rightNode, inLength);
		
	inNode->insertBranch(leftLink);
	leftLink->setNodeNo(inNode->nodeNo);
	inNode->insertBranch(rightLink);
	rightLink->setNodeNo(inNode->nodeNo);
	
	setStemLink(inNode);
	
	delete [] leftSet;
}

// ---------------------------------------------------------------------------------
//		 METree::setStemLink
// ---------------------------------------------------------------------------------

void METree::setStemLink(MELink *inLink) 
{
	inLink->isStem=true;
	
	if (inLink->tipNo==-1) {
		MELink *link=inLink->next;
		while (link!=inLink) {
			link->isStem=false;
			setStemLink(link->branch);
			link=link->next;
		}
	}	
}

// ---------------------------------------------------------------------------------
//		 METree::rotateNode
// ---------------------------------------------------------------------------------

void METree::rotateNode(MELink *inLink) 
{
	if (inLink->getIsTip())
		return;
	
	MELink *link = inLink;
	
	if (inLink == NULL) { // Rotate the root
		if (isRooted) {
			MELink *link = root2; // simply swap the root pointers
			root2 = root1;
			root1 = link;
			
			link = NULL; // nothing further is needed
		} else
			link = root1; // set the root node to rotate
	}
	
	if (link != NULL) { // Rotate the node
		MELink *link1 = link->next;
		MELink *link0 = link->getPrevious();
		
		link->next = link1->next;
		link0->next = link1;
		link1->next = link;
	}
}

#pragma mark === Tree Rooting ===

// ---------------------------------------------------------------------------------
//		 METree::outgroupTree
// ---------------------------------------------------------------------------------

//: Used to outgroup root an unrooted or rooted tree 
void METree::outgroupTree(MELink *inLink) 
{
	
	setRoot2(inLink);
	setRoot1(inLink->branch);
	rootDistsKnown=false;
		
	swapRootNodes();
}


// ---------------------------------------------------------------------------------
//		 METree::outgroupTree
// ---------------------------------------------------------------------------------

//: Used to outgroup root an unrooted or rooted tree 
void METree::outgroupTree(MELink *inLink, double inPosition) 
{	
	setRoot2(inLink);
	setRoot1(inLink->branch);
	rootDistsKnown=false;
	
	if (inPosition < 0.0)
		inPosition = 0.0;
	if (inPosition > 1.0)
		inPosition = 1.0;
		
	double bLen = getBranchLength(root1->branchNo);
	
	setRootDistances(bLen * (1.0 - inPosition), bLen * inPosition);	

	swapRootNodes();
}


// ---------------------------------------------------------------------------------
//		 METree::findMidPoint
// ---------------------------------------------------------------------------------

//: Finds the link that is a certain length from tip1 to tip2
MELink *METree::findMidPoint(int inTip1, int inTip2, double inMidPoint) 
{
	double len;
	MELink *tip1, *tip2, *link, *link1;
	
	tip1=getTip(inTip1);
	tip2=getTip(inTip2);
	
	link=tip1->branch;
	if (hasLengths)
		len=branchLengths[link->branchNo];
	else
		len=1.0;
		
	while (link!=tip2 && len<inMidPoint) {
		link1=link->next;
		while (link1!=link && !link1->branch->cladeContains(tip2))
			link1=link1->next;
			
		if (link1==link)
			throw MELinkNotFound();
		
		link=link1->branch;
		if (hasLengths)
			len+=branchLengths[link1->branchNo];
		else
			len++;
	}
	
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::midPointTree
// ---------------------------------------------------------------------------------

//: Used to midpoint root the tree. 
// If it is unrooted then this finds the node that best represents the midpoint. 
// If it is rooted then it sets the root along the midpoint branch.
void METree::midPointTree() 
{
	int i, j, maxTip1, maxTip2;
	double len, maxLen=0.0;
	
	for (i=0; i<numTips-1; i++) {
		for (j=i+1; j<numTips; j++) {
			if ((len=getTipToTipDistance(i, j))>maxLen) {
				maxLen=len;
				maxTip1=i;
				maxTip2=j;
			}
		}
	}
	
	MELink *link=findMidPoint(maxTip1, maxTip2, maxLen/2);
	outgroupTree(link);
	
	adjustRootDistances();

	swapRootNodes();
}


// ---------------------------------------------------------------------------------
//		 METree::adjustRootDistances
// ---------------------------------------------------------------------------------

//: Used to adjust the root distances to the midpoint. 
void METree::adjustRootDistances(double inMinFraction) 
{
	double height1 = calcNodeHeight(root1);
	double height2 = calcNodeHeight(root2);
	double bLen = getBranchLength(root1->branchNo);
	
	double rootHeight = (height1 + height2 + bLen) / 2;
	
	double d1 = rootHeight - height1;
	double d2 = rootHeight - height2;
	
	if (d1 < 0.0) {
		d2 += d1;
		double fraction = d2 * inMinFraction;
		d1 = fraction;
		d2 -= fraction;
	} else if (d2 < 0.0) {
		d1 += d2;
		double fraction = d1 * inMinFraction;
		d2 = fraction;
		d1 -= fraction;
	}
	setRootDistances(d1, d2);	
}


// ---------------------------------------------------------------------------------
//		 METree::swapRootNodes
// ---------------------------------------------------------------------------------

//: Swaps the root nodes so that the smaller clade is below the larger. 
void METree::swapRootNodes() 
{

	if (!navBitsKnown)
		calcNavBits();

	MEBitset *navBits1 = root1->getNavBits();
	MEBitset *navBits2 = root2->getNavBits();
	
	int n1, n2;
	n1 = n2 = getNumNodes() - 1;
	
	if (navBits1 != NULL)
		n1 = navBits1->getCard();
	if (navBits2 != NULL)
		n2 = navBits2->getCard();
		
	if (n1 > n2) {
		MELink *tmp;
		tmp = root1;
		root1 = root2;
		root2 = tmp;
		
		setRootDistances(dist2, dist1);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::rootTree
// ---------------------------------------------------------------------------------

//: Used to root an unrooted tree 
void METree::rootTree() 
{
	if (!isRooted)
		isRooted=true;
}

// ---------------------------------------------------------------------------------
//		 METree::unrootTree
// ---------------------------------------------------------------------------------

//: Used to unroot a rooted tree 
void METree::unrootTree() 
{
	if (isRooted)
		isRooted=false;
}

#pragma mark === Tree Navigation ===

// ---------------------------------------------------------------------------------
//		 METree::clearAllNodeFlags
// ---------------------------------------------------------------------------------

void METree::clearAllNodeFlags()
{
	for (int i=0; i<numLinks; i++)
		links[i]->setFlags(0);
}

// ---------------------------------------------------------------------------------
//		 METree::setFlagForAllNodes
// ---------------------------------------------------------------------------------

void METree::setFlagForAllNodes(short inFlag)
{
	for (int i=0; i<numLinks; i++)
		links[i]->setFlag(inFlag);
}

// ---------------------------------------------------------------------------------
//		 METree::toggleFlagForAllNodes
// ---------------------------------------------------------------------------------

void METree::toggleFlagForAllNodes(short inFlag)
{
	for (int i=0; i<numLinks; i++)
		links[i]->toggleFlag(inFlag);
}

// ---------------------------------------------------------------------------------
//		 METree::clearFlagForAllNodes
// ---------------------------------------------------------------------------------

void METree::clearFlagForAllNodes(short inFlag)
{
	for (int i=0; i<numLinks; i++)
		links[i]->clearFlag(inFlag);
}

// ---------------------------------------------------------------------------------
//		 METree::copyFlagForAllNodes
// ---------------------------------------------------------------------------------

void METree::copyFlagForAllNodes(short inFlag, METree *inTree)
{
	for (int i=0; i<numLinks; i++) {
		if (inTree->getLink(i)->getFlag(inFlag))
			links[i]->setFlag(inFlag);
		else
			links[i]->clearFlag(inFlag);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::copyFlagForAllTips
// ---------------------------------------------------------------------------------

void METree::copyFlagForAllTips(short inFlag, METree *inTree)
{
	for (int i=0; i<numTips; i++) {
		if (inTree->getTip(i)->getFlag(inFlag))
			getTip(i)->setFlag(inFlag);
		else
			getTip(i)->clearFlag(inFlag);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::setNodeFlagIfFlagSet
// ---------------------------------------------------------------------------------

void METree::setNodeFlagIfFlagSet(short inFlag, short ifFlag)
{
	for (int i=0; i<numLinks; i++) {
		if (links[i]->getFlag(ifFlag))
			links[i]->setFlag(inFlag);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::clearNodeFlagIfFlagSet
// ---------------------------------------------------------------------------------

void METree::clearNodeFlagIfFlagSet(short inFlag, short ifFlag)
{
	for (int i=0; i<numLinks; i++) {
		if (links[i]->getFlag(ifFlag))
			links[i]->clearFlag(inFlag);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::setNodeFlagIfFlagClear
// ---------------------------------------------------------------------------------

void METree::setNodeFlagIfFlagClear(short inFlag, short ifFlag)
{
	for (int i=0; i<numLinks; i++) {
		if (!links[i]->getFlag(ifFlag))
			links[i]->setFlag(inFlag);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::clearNodeFlagIfFlagClear
// ---------------------------------------------------------------------------------

void METree::clearNodeFlagIfFlagClear(short inFlag, short ifFlag)
{
	for (int i=0; i<numLinks; i++) {
		if (!links[i]->getFlag(ifFlag))
			links[i]->clearFlag(inFlag);
	}
}

// ---------------------------------------------------------------------------------
//		 METree::getNumTipsWithFlag
// ---------------------------------------------------------------------------------

int METree::getNumTipsWithFlag(short inFlag)
{
	int count = 0;
	for (int i=0; i<numLinks; i++) {
		if (links[i]->getIsTip() && links[i]->getFlag(inFlag))
			count++;
	}
	
	return count;
}

// ---------------------------------------------------------------------------------
//		 METree::findNodeWithFlagSet
// ---------------------------------------------------------------------------------

MELink* METree::findNodeWithFlagSet(MELink *inLink, short inFlag)
{
	MELink *link, *link0 = inLink;
	
	if (inLink->getIsTip()) {
		if (!inLink->getFlag(inFlag))
			link0 = NULL;
	} else {
		link = inLink->next;
		bool found = false;
		while (link != inLink && !found) {
			if (link->getFlag(inFlag))
				found = true;
			else 
				link = link->next;
		}
		
		if (!found) {
			link = inLink->next;
			bool found = false;
			while (link != inLink && !found) {
				link0 = findNodeWithFlagSet(link->branch, inFlag);
				if (link0 != NULL)
					found = true;
				link = link->next;
			}
		}
	}
	
	return link0;
}

// ---------------------------------------------------------------------------------
//		 METree::setFlagForClade
// ---------------------------------------------------------------------------------

//: Sets flags for clade defined by inLink.
void METree::setFlagForClade(MELink *inLink, short inFlag)
{
	inLink->setFlag(inFlag);
	
	MELink *link0 = inLink->branch;
	link0->setFlag(inFlag);

	if (!link0->getIsTip()) {
		MELink *link = link0->next;
		while (link != link0) {
			setFlagForClade(link, inFlag);
			
			link = link->next;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 METree::toggleFlagForClade
// ---------------------------------------------------------------------------------

//: Toggles (negates) flags for clade defined by inLink.
void METree::toggleFlagForClade(MELink *inLink, short inFlag)
{
	if (!inLink->getIsTip())
		inLink->toggleFlag(inFlag);
	
	MELink *link0 = inLink->branch;
	link0->toggleFlag(inFlag);

	if (!link0->getIsTip()) {
		MELink *link = link0->next;
		while (link != link0) {
			toggleFlagForClade(link, inFlag);
			
			link = link->next;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 METree::clearFlagForClade
// ---------------------------------------------------------------------------------

//: Clears flags for clade defined by inLink.
void METree::clearFlagForClade(MELink *inLink, short inFlag)
{
	inLink->clearFlag(inFlag);
	
	MELink *link0 = inLink->branch;
	link0->clearFlag(inFlag);

	if (!link0->getIsTip()) {
		MELink *link = link0->next;
		while (link != link0) {
			clearFlagForClade(link, inFlag);
			
			link = link->next;
		}
	}
}

 
// ---------------------------------------------------------------------------------
//		 METree::setTipFlagsFromTaxa
// ---------------------------------------------------------------------------------

//: Sets flags of the tips to match the flags of the taxa.
void METree::setTipFlagsFromTaxa()
{
	MELink *tip = getFirstTip();
	METaxon *taxon;
	
	while (tip != NULL) {
		taxon = getTaxon(tip);
		tip->setFlags(taxon->getFlags());
	
		tip = getNextTip(tip);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::setTreeFlagsFromTips
// ---------------------------------------------------------------------------------

//: Sets flags for the whole tree dependant on the flags of the tips.
void METree::setNodeFlagsFromTips()
{
	if (!navBitsKnown)
		calcNavBits();

	MELink *link, *node = getFirstInternalNode();
	short flags1, flags2;
	
	while (node != NULL) {
		link = node;
		
		do {
			getFlagsForTips(link->getNavBits(), flags1, flags2);
			link->setFlags(flags1 & flags2);
			link = link->next;
		} while (link != node);
	
		node = getNextInternalNode(node);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::setTaxonFlagsFromTips
// ---------------------------------------------------------------------------------

//: Sets flags of the taxa to match the flags of the tips.
void METree::setTaxonFlagsFromTips()
{
	MELink *tip = getFirstTip();
	METaxon *taxon;
	
	while (tip != NULL) {
		taxon = getTaxon(tip);
		taxon->setFlags(tip->getFlags());
	
		tip = getNextTip(tip);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::getFlagsForTips
// ---------------------------------------------------------------------------------

//: Gets the union of the flags for the tips given by a navBits in both directions.
void METree::getFlagsForTips(MEBitset *inNavBits, short &outFlags1, short &outFlags2)
{
	outFlags1 = outFlags2 = 0;
	
	MELink *tip = getFirstTip();
	while (tip != NULL) {
		if (inNavBits->contains(tip->tipNo)) 
			outFlags1 |= tip->getFlags();
		else
			outFlags2 |= tip->getFlags();
		
		tip = getNextTip(tip);
	}
}


// ---------------------------------------------------------------------------------
//		 METree::getNavBits
// ---------------------------------------------------------------------------------

MEBitset* METree::getNavBits(MELink *inLink) 
{
	if (!navBitsKnown)
		calcNavBits();

	if (inLink->getIsTip()) 
		return inLink->branch->getNavBits();
	
	return inLink->getNavBits();
}
	
// ---------------------------------------------------------------------------------
//		 METree::findNavBits
// ---------------------------------------------------------------------------------

MELink* METree::findNavBits(MEBitset* inNavBits) 
{
	if (!navBitsKnown)
		calcNavBits();

	MELink *link;
	for (int i = 0; i < numLinks; i++) {
		link = getLink(i);
		
		if (!link->getIsTip() && inNavBits->equals(*link->getNavBits())) 
			return link;
	}
	
	return NULL;
}
	
// ---------------------------------------------------------------------------------
//		 METree::calcNavBits
// ---------------------------------------------------------------------------------

void METree::calcNavBits() 
{
	MELink *tip=getFirstTip();

	MEBitset *navBits;
	NewClass(navBits, MEBitset());
	
	calcNavBitsAtNode(tip->branch, navBits);
	
	MEBitset *navBits2=tip->branch->getNavBits();
	*navBits2=*navBits;
	navBits2->setComplement(getNumNodes());
	
	navBitsKnown = true;
	delete navBits;
	
	// This should just have the bit for this tip set.
}

// ---------------------------------------------------------------------------------
//		 METree::calcNavBitsAtNode
// ---------------------------------------------------------------------------------

void METree::calcNavBitsAtNode(MELink *inLink, MEBitset *outBits) 
{
	if (inLink->getIsTip()) {
		outBits->add(inLink->tipNo);
	} else {
		MEBitset *navBits, *navBits2;
	
		MELink *link=inLink->next;
		while (link!=inLink) {
			navBits=link->getNavBits();
			navBits->clear();
			
			calcNavBitsAtNode(link->branch, navBits);
			
			if (!link->branch->getIsTip()) {
				navBits2=link->branch->getNavBits();
				*navBits2=*navBits;
				navBits2->setComplement(getNumNodes());
			}
			outBits->add(*navBits);
			
			link=link->next;
		}

		outBits->add(inLink->nodeNo+getNumTips());
	}	
}

// ---------------------------------------------------------------------------------
//		 METree::calcNodeHeight
// ---------------------------------------------------------------------------------

double METree::calcNodeHeight(MELink *inNode) 
{
	double maxHeight = 0.0;
	
	if (inNode==NULL || !inNode->getIsTip()) {
		MELink *nextBranch, *descBranch;
		nextBranch = getFirstBranch(inNode);
		double length, height;

		while (nextBranch!=NULL) {	
			descBranch = getDescendantNode(nextBranch, &length);
			height = calcNodeHeight(descBranch) + length;
			if (height > maxHeight)
				maxHeight = height;
				
	    	nextBranch = getNextBranch(inNode, nextBranch);
	    }
	}
	
	return maxHeight;
}

// ---------------------------------------------------------------------------------
//		 METree::calcTreeHeight
// ---------------------------------------------------------------------------------
// Finds the total height of the tree from root to tip. If the tree is not 
// ultrametric then it returns the longest length from root to tip. If the tree
// does not have branch lengths then this returns the maximum number of branches
// travelled from root to tip.

double METree::calcTreeHeight() 
{
	return calcNodeHeight(NULL);
}

// ---------------------------------------------------------------------------------
//		 METree::calcNodeHeights
// ---------------------------------------------------------------------------------

double METree::calcNodeHeights(MELink *inNode, MEDoubleArray *outHeights) 
{
	double maxHeight = 0.0;
	
	if (inNode==NULL || !inNode->getIsTip()) {
		MELink *nextBranch, *descBranch;
		nextBranch = getFirstBranch(inNode);
		double length, height;

		while (nextBranch!=NULL) {	
			descBranch = getDescendantNode(nextBranch, &length);
			height = calcNodeHeights(descBranch, outHeights) + length;
			if (height > maxHeight)
				maxHeight = height;
				
	    	nextBranch = getNextBranch(inNode, nextBranch);
	    }
	    
		outHeights->add(maxHeight);
	}
	
	return maxHeight;
}

// ---------------------------------------------------------------------------------
//		 METree::calcTreeHeights
// ---------------------------------------------------------------------------------

double METree::calcTreeHeights(MEDoubleArray *outHeights) 
{
	ThrowIfNull(outHeights);
	
	return calcNodeHeights(NULL, outHeights);
}

// ---------------------------------------------------------------------------------
//		 METree::calcNodeDensity
// ---------------------------------------------------------------------------------
// Returns the maximum number of nodes between (and including) inNode an its
// descendent tips

int METree::calcNodeDensity(MELink *inNode) 
{
	int maxDensity = 0.0;
	
	if (inNode==NULL || !inNode->getIsTip()) {
		MELink *nextBranch, *descBranch;
		nextBranch = getFirstBranch(inNode);
		int density;

		while (nextBranch!=NULL) {	
			descBranch = getDescendantNode(nextBranch);
			density = calcNodeDensity(descBranch) + 1;
			if (density > maxDensity)
				maxDensity = density;
				
	    	nextBranch = getNextBranch(inNode, nextBranch);
	    }
	}
	
	return maxDensity;
}

// ---------------------------------------------------------------------------------
//		 METree::calcTipDensity
// ---------------------------------------------------------------------------------
// Returns the number of descendent tips for inNode

int METree::calcTipDensity(MELink *inNode) 
{
	if (inNode==NULL || !inNode->getIsTip()) {
		MELink *nextBranch, *descBranch;
		nextBranch = getFirstBranch(inNode);
		int density = 0;

		while (nextBranch!=NULL) {	
			descBranch = getDescendantNode(nextBranch);
			density += calcTipDensity(descBranch);
				
	    	nextBranch = getNextBranch(inNode, nextBranch);
	    }
		return density;
	}
	
	return 1;
}

// ---------------------------------------------------------------------------------
//		 METree::getCommonAncestor
// ---------------------------------------------------------------------------------
// Finds the node in a rooted tree that is the most recent common ancestor of inTip1
// and inTip2.

MELink* METree::getCommonAncestor(MELink *inTip1, MELink *inTip2)
{
	ThrowIf(inTip1 == inTip2);
	
	if (!navBitsKnown)
		calcNavBits();
	
	MELink *link0;
	MEBitset *navBits1, *navBits2;
	
	if (root1->getIsTip()) {
		if (root1 == inTip1 || root1 == inTip2)
			// if either side of the root is inTip1 or inTip2 then the root must
			// be the MRCA (assumming the other tip is in the other side of the root)
			link0 = NULL;
		else
			// Otherwise the MRCA must be on the other side of the root
			link0 = root2;
	} else if (root2->getIsTip()) {
		if (root2 == inTip1 || root2 == inTip2)
			// if either side of the root is inTip1 or inTip2 then the root must
			// be the MRCA (assumming the other tip is in the other side of the root)
			link0 = NULL;
		else
			// Otherwise the MRCA must be on the other side of the root
			link0 = root1;
	} else {
		navBits1=root2->getNavBits(); // the navbits at the root nodes point back to each other
		navBits2=root1->getNavBits(); // so reverse them.
		
		if (navBits1->contains(inTip1->getTipNo())) {
			if (navBits1->contains(inTip2->getTipNo())) 
				link0 = root1;
			else if (navBits2->contains(inTip2->getTipNo())) 
				link0 = NULL; // the root is the MRCA
			else
				throw MENodeNotFound();
			
		} else if (navBits2->contains(inTip1->getTipNo())) {
			if (navBits2->contains(inTip2->getTipNo())) 
				link0 = root2;
			else if (navBits1->contains(inTip2->getTipNo())) 
				link0 = NULL; // the root is the MRCA
			else
				throw MENodeNotFound();
		} else
			throw MENodeNotFound();
	}
				
	if (link0 != NULL) {
		bool found = false;
		MELink *link, *link1, *link2;
		MEBitset *navBits;
		
		while (!found) {
			link = link0->next;
			link1 = link2 = NULL;
			
			while (link != link0) {
				navBits = link->getNavBits();
				if (navBits->contains(inTip1->getTipNo()))
					link1 = link;

				if (navBits->contains(inTip2->getTipNo()))
					link2 = link;
					
				link = link->next;
			}
			ThrowIf(link1 == NULL || link2 == NULL);
			
			if (link1 == link2)
				link0 = link1->branch;
			else
				found = true;
		}
	} else if (!getIsRooted())
		link0 = root1;
	
	return link0;
}

// ---------------------------------------------------------------------------------
//		 METree::getPathMeet
// ---------------------------------------------------------------------------------
// Finds the node on the path between inNode1 and inNode2 which meets the path
// from inNode0. It returns the link in this node that points back to inNode0
// (or is inNode0 if it lies on the path itself).

MELink* METree::getPathMeet(MELink *inNode0, MELink *inNode1, MELink *inNode2)
{
	ThrowIf(inNode0 == inNode1);
	ThrowIf(inNode0 == inNode2);
	ThrowIf(inNode1 == inNode2);
	
	if (!navBitsKnown)
		calcNavBits();
		
	MELink *link, *link1, *link2;
	MEBitset *navBits;
	
	if (inNode0->getIsTip())
		link = inNode0->branch;
	else
		link = inNode0;
			
	do {
		link2 = NULL;
		link1 = link->next;
		do {
			navBits=link1->getNavBits();
			if (navBits->contains(inNode1->getNodeNo() + getNumTips()) && 
				navBits->contains(inNode2->getNodeNo() + getNumTips()) )
				link2=link1->branch;
		} while (link1!=link); 
		
		if (link2!=NULL) // There is a branch that leads to both so follow it...
			link = link2;
			
	} while (link2!=NULL); // ...otherwise we have found the meeting
	
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::getPath
// ---------------------------------------------------------------------------------

double METree::getPath(MELink *inNode1, MELink *inNode2, MEBitset *outPath)
{
	ThrowIfNull(outPath);
	
	int len;
	double dist;
	
	calcPath(inNode1, inNode2, dist, len, NULL, outPath);

	return dist;
}

// ---------------------------------------------------------------------------------
//		 METree::getPath
// ---------------------------------------------------------------------------------

double METree::getPath(MELink *inNode1, MELink *inNode2, MEIntArray *outPath)
{
	ThrowIfNull(outPath);
	
	int len;
	double dist;
	
	calcPath(inNode1, inNode2, dist, len, outPath, NULL);

	return dist;
}

// ---------------------------------------------------------------------------------
//		 METree::getPathLength
// ---------------------------------------------------------------------------------
// Returns the path length, in terms of number of branch segment, between 
// inNode1 and inNode2. If inNode1 is NULL then this represents the root.

int METree::getPathLength(MELink *inNode1, MELink *inNode2)
{
	int len;
	double dist;
	
	calcPath(inNode1, inNode2, dist, len, NULL, NULL);
	
	return len;
}

// ---------------------------------------------------------------------------------
//		 METree::getPathDistance
// ---------------------------------------------------------------------------------
// Returns the path length, in terms of the sum of the branch lengths, between 
// inNode1 and inNode2. If inNode1 is NULL then this represents the root.

double METree::getPathDistance(MELink *inNode1, MELink *inNode2)
{
	int len;
	double dist;
	
	calcPath(inNode1, inNode2, dist, len, NULL, NULL);

	return dist;
}

// ---------------------------------------------------------------------------------
//		 METree::calcPath
// ---------------------------------------------------------------------------------
// If inNode1 is NULL then this represents the root.

void METree::calcPath(MELink *inNode1, MELink *inNode2,
										double &outDist,
										int &outLength,
										MEIntArray *outIntArray,
										MEBitset *outBitset)
{
	if (!navBitsKnown)
		calcNavBits();
		
	MELink *link1, *link2, *link;
	MEBitset *navBits;
	bool found = false;
	
	int targetNo;
	if (inNode2->getIsTip())
		targetNo = inNode2->tipNo;
	else
		targetNo = getNumTips() + inNode2->nodeNo;
		
	outDist=0.0;
	outLength=0;
	
	if (inNode1 == NULL) { // start at the root
		navBits = getRoot1()->getNavBits();
		if (navBits->contains(targetNo)) {
			link1 = getRoot2();
			outDist += getRootDistance2();
			outLength++;
		} else {
			link1 = getRoot1();
			outDist += getRootDistance1();
		}
		outLength++;
		if (link1 == inNode2)
			found = true;
	} else if (inNode1->getIsTip()) {
		link1 = inNode1->branch;
		outDist += getBranchLength(inNode1->branchNo);
		outLength++;
		if (outIntArray != NULL)
			outIntArray->add(inNode1->branchNo);
		if (outBitset != NULL)
			outBitset->add(inNode1->branchNo);
	} else
		link1 = inNode1;
	
	link2 = inNode2;

	link=link1->next;
	while (!found) {
		navBits=link->getNavBits();
		if (navBits->contains(targetNo)) {
			link1 = link->branch;
			outDist += getBranchLength(link->branchNo);
			outLength++;
			if (outIntArray != NULL)
				outIntArray->add(link->branchNo);
			if (outBitset != NULL)
				outBitset->add(link->branchNo);
			link = link1;
			if (link1 == link2)
				found = true;
		}
		if (!found) {
			link=link->next;
			if (link == link1)
				throw MENodeNotFound();
		}
	} 
}

// ---------------------------------------------------------------------------------
//		 METree::getMaxPathDistance
// ---------------------------------------------------------------------------------
// Returns the longest path length, in terms of the sum of the branch lengths, between 
// any two tips.

double METree::getMaxPathDistance()
{
	int i, j;
	double len, maxLen=0.0;
	
	for (i=0; i<numTips-1; i++) {
		for (j=i+1; j<numTips; j++) {
			if ((len=getTipToTipDistance(i, j))>maxLen) {
				maxLen=len;
			}
		}
	}
	return maxLen;
}

// ---------------------------------------------------------------------------------
//		 METree::getTipToTipDistance
// ---------------------------------------------------------------------------------

// Calculates the length from inTip1 to inTip2. If the tree has no branch lengths then 
// this returns the number of branch segments between the nodes. 
// Simply calls getPathDistance.

double METree::getTipToTipDistance(int inTip1, int inTip2) 
{
	MELink *tip1, *tip2;

	tip1 = getTip(inTip1);
	tip2 = getTip(inTip2);	
	
	return getPathDistance(tip1, tip2);
}

// ---------------------------------------------------------------------------------
//		 METree::getRootToTipDistance
// ---------------------------------------------------------------------------------

// Calculates the length from the root to inTip. If the tree has no branch lengths then 
// this returns the number of branch segments between the nodes. 
// Simply calls getPathDistance.

double METree::getRootToTipDistance(int inTip) 
{
	MELink *tip = getTip(inTip);	
	return getPathDistance(NULL, tip);;
}

// ---------------------------------------------------------------------------------
//		 METree::getNodeToNodeDistance
// ---------------------------------------------------------------------------------

// Calculates the length from inNode1 to inNode2 (tips or internal nodes) by 
// navigating the tree. If the tree has no branch lengths then this returns 
// the number of branch segments between the nodes. Simply calls getPathDistance.

double METree::getNodeToNodeDistance(int inNode1, int inNode2) 
{
	MELink *node1, *node2;

	node1 = getNode(inNode1);
	node2 = getNode(inNode2);	
	
	return getPathDistance(node1, node2);
}

// ---------------------------------------------------------------------------------
//		 METree::getPointOnPath
// ---------------------------------------------------------------------------------
// 
MELink* METree::getPointOnPath(double inPoint, MELink *inNode1, MELink *inNode2, double *outDist)
{
	ThrowIf(inNode1 == inNode2);
	
	if (!navBitsKnown)
		calcNavBits();
		
	MELink *link, *link1, *link2;
	MEBitset *navBits;
	double dist = 0.0;
	
	if (inNode1->getIsTip()) {
		dist += getBranchLength(inNode1->branchNo);
		link1 = inNode1->branch;
	} else
		link1 = inNode1;
	
	link2 = inNode2;
	
	int targetNo;
	if (inNode2->getIsTip())
		targetNo = inNode2->tipNo;
	else
		targetNo = getNumTips() + inNode2->nodeNo;
		
	while (link1 != link2 && dist < inPoint) {
		link=link1;
		do {
			navBits = link->getNavBits();
			if (navBits->contains(targetNo)) {
				link1 = link->branch;
				dist += getBranchLength(link->branchNo);
			}
		} while (link != link1 && link1 != link2); 
	}
	
	if (outDist != NULL) 
		*outDist = dist - inPoint;
	
	return link1;
}

// ---------------------------------------------------------------------------------
//		 METree::getPointOnPath
// ---------------------------------------------------------------------------------
MELink* METree::getPointOnPath(int inPoint, MELink *inNode1, MELink *inNode2)
{
	ThrowIf(inNode1 == inNode2);
	
	if (!navBitsKnown)
		calcNavBits();
		
	MELink *link, *link1, *link2;
	MEBitset *navBits;
	int len = 0;
	
	if (inNode1->getIsTip()) {
		len ++;
		link1 = inNode1->branch;
	} else
		link1 = inNode1;
	
	link2 = inNode2;
	
	int targetNo;
	if (inNode2->getIsTip())
		targetNo = inNode2->tipNo;
	else
		targetNo = getNumTips() + inNode2->nodeNo;
		
	while (link1 != link2 && len < inPoint) {
		link=link1;
		do {
			navBits = link->getNavBits();
			if (navBits->contains(targetNo)) {
				link1 = link->branch;
				len++;
			}
		} while (link != link1 && link1 != link2); 
	}
	
	return link1;
}

// ---------------------------------------------------------------------------------
//		 METree::getFirstBranch
// ---------------------------------------------------------------------------------

//: Gets the first branch and branch length for 'inNode'. If 'inNode' is NULL
//  then this returns the first branch in the root. If inFlag is not zero
//  then this only returns branches that have inFlag set.

MELink* METree::getFirstBranch(MELink *inNode)
{
	if (inNode == NULL) { // return the root...
		MELink *link = getRoot1();
		if (getIsRooted()) {
			if (flagMask != 0 && !link->getFlag(flagMask)) {
				MELink *link1 = findNodeWithFlagSet(link, flagMask);
				MELink *link2 = findNodeWithFlagSet(getRoot2(), flagMask);
				
				ThrowIf(link1==NULL && link2==NULL);
				if (link1!=NULL)
					return getNextBranch(link1, link1);
				
				return getNextBranch(link2, link2);
			}
				
			return link;
		}
			
		return getNextBranch(NULL, NULL);
	} else
		ThrowIf(inNode->getIsTip());
			
	return getNextBranch(inNode, inNode);
}

// ---------------------------------------------------------------------------------
//		 METree::getNextBranch
// ---------------------------------------------------------------------------------

//: Gets the next branch and branch length for 'inNode' (after 'inLastBranch' which
//  was returned by getFirstBranch or getNextBranch). If 'inNode' is NULL
//  then this returns the next branch in the root. If inFlag is not zero
//  then this only returns branches that have inFlag set.

MELink* METree::getNextBranch(MELink *inNode, MELink *inLastBranch)
{
	MELink *link, *link0;
	
	if (inNode == NULL) { // return the root...
		link = getRoot1();
		if (getIsRooted()) {
			if (flagMask != 0 && !link->getFlag(flagMask)) {
				MELink *link1 = findNodeWithFlagSet(link, flagMask);
				MELink *link2 = findNodeWithFlagSet(getRoot2(), flagMask);
					
				ThrowIf(link1==NULL && link2==NULL);
				if (link1!=NULL)
					link = link1;
				else if (link2!=NULL)
					link = link2;
				link0 = inLastBranch->next;
			} else {
				if (inLastBranch == link)
					return getRoot2();
				else
					return NULL;
			}
		} else {
			if (inLastBranch==NULL) {
				// called by getFirstBranch... 
				// start with first link after root1:
				link0 = link->next;
			} else if (inLastBranch->next==link) {
				// done all the branches in root1... 
				// except the one that points back to root2:
				link0 = inLastBranch->next;
				// we must set link forward to otherwise
				// it will return NULL below.
				link = link->next;
			} else if (inLastBranch==link) {
				// Now we are really done so return a NULL
				return NULL;
			} else
				link0 = inLastBranch->next;
		}
	} else {
		link = inNode;
		link0 = inLastBranch->next;
	}

	if (link0 == link)
		return NULL;

	if (flagMask != 0) {
		while (link0 != link && !link0->getFlag(flagMask))
			link0 = link0->next;
			
		if (link0 == link)
			return NULL;
	}

	return link0;
}

// ---------------------------------------------------------------------------------
//		 METree::getDescendantNode
// ---------------------------------------------------------------------------------
MELink* METree::getDescendantNode(MELink *inBranch, double *outDist)
{
	double length;
	MELink *link;
	
	if (getIsRooted()) {
		if (inBranch == getRoot1()) {
			length = getRootDistance1();
			link = inBranch;
		} else if (inBranch == getRoot2()) {
			length = getRootDistance2();
			link = inBranch;
		} else {
			length = getBranchLength(inBranch);
			link = inBranch->branch;
		}
	} else {
	 	length = getBranchLength(inBranch);
		link = inBranch->branch;
	}
	
	if (flagMask != 0 && !link->getIsTip()) {
		// We must now skip over nodes that are degree 2 with
		// respect to the flagMask.
		
		int i=0;
		MELink *link1 = link->next;
		MELink *link2 = NULL;
		while (link1 != link) {
			if (link1->getFlag(flagMask)) {
				i++;
				link2 = link1;
			}
			link1=link1->next;
		}
			
		ThrowIfNull(link2);

		if (i==1) {
			double len2;
			link = getDescendantNode(link2, &len2);
			length += len2;
		}
	}

	*outDist = length;
	return link;
}

// ---------------------------------------------------------------------------------
//		 METree::getDescendantNode
// ---------------------------------------------------------------------------------
MELink* METree::getDescendantNode(MELink *inBranch)
{
	MELink *link;
	
	if (getIsRooted() && (inBranch == getRoot1() || inBranch == getRoot2())) {
		link = inBranch;
	} else {
		link = inBranch->branch;
	}
	
	if (flagMask != 0 && !link->getIsTip()) {
		// We must now skip over nodes that are degree 2 with
		// respect to the flagMask.
		
		int i=0;
		MELink *link1 = link->next;
		MELink *link2 = NULL;
		while (link1 != link) {
			if (link1->getFlag(flagMask)) {
				i++;
				link2 = link1;
			}
			link1=link1->next;
		}
			
		ThrowIfNull(link2);

		if (i==1) {
			link = getDescendantNode(link2);
		}
	}

	return link;
}

#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 METree::readObject
// ---------------------------------------------------------------------------------

// Read the fields of this object from stream.
void METree::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
	taxonList=(METaxonList *)stream->get("Taxa");
	stream->get("HasLengths", &hasLengths);
}

// ---------------------------------------------------------------------------------
//		 METree::readData
// ---------------------------------------------------------------------------------

// Read the data of this object (i.e. the tree).
void METree::readData(MEStream *stream)
{
	stream->getTree(this);
}

// ---------------------------------------------------------------------------------
//		 METree::writeObject
// ---------------------------------------------------------------------------------

// Write the fields of this object to stream.
void METree::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
	stream->put("Taxa", taxonList);
	stream->put("HasLengths", hasLengths);
}


// ---------------------------------------------------------------------------------
//		 METree::writeData
// ---------------------------------------------------------------------------------

// Write the data of this object (i.e. the tree).
void METree::writeData(MEStream *stream)
{
	stream->putTree(this);
}
