/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METree.h

		Interface for the METree Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Trees"

#ifndef _METREE_H_
#define _METREE_H_

#include "MEObject.h"

class MEBitset;
class MEIntArray;
class MELink;
class MEStream;
class METaxon;
class METaxonList;
class MEDoubleArray;
class MEList;
class MELabelArray;

enum TaxonOrder {
	kAlphabeticalOrder,
	kRevAlphabeticalOrder,
	kIncNodeDensityOrder,
	kDecNodeDensityOrder,
	kRandomOrder
};
	
/******************************************************************************
	class METree
 ******************************************************************************/
#pragma mark METree Class

class METree : public MEObject {
	public:
		virtual long		getID() const { return 'MTre'; }
		//: The ID of this object - 'MTre'.
		
		virtual char*		getClassName() const { return "Tree"; }
		//: Returns the name of this class - "Tree"
		
		bool				isRooted, hasLengths;

	protected:
		int 				numTips, 			//: the number of tips
							numIntNodes, 		//: the number of internal nodes
							numNodes, 			//: the number of tips + internal nodes
							numBranches,		//: the number of branches
							numLinks;			//: the total number of links 

		MELink				**links;			//: array of pointers to links 
												//:	(first numTips are the tips)

		double				*branchLengths;		//: array of branch lengths

		double				*nodeHeights;		//: array of node heights
		double 				rootHeight;			//: height of the root node
		
		double				*tipDates;			//: array of tip dates
		
		MELink				*root1, *root2;		//: pointers to the links either side of the root
		double				dist1, dist2;		//: distance to nodes either side of the root
		bool				rootDistsKnown;		//: Are the root distances known?
		
		int					linkCapacity,		//: the capacity of the node array
							branchCapacity,		//: the capacity of the branch array	
							heightCapacity,		//: the capacity of the node height array	
							tipDateCapacity;	//: the capacity of the tip date array	
		
		bool				useNumericalTips;	//: Whether should write using tip numbers
							
		METaxonList 		*taxonList;
		
		METaxonList 		*translationList;	//: This is used when reading in NEXUS
		
		int					position;			//: Used in the getFirst/getNext functions

		bool				navBitsKnown;		//: Are the navigation bitsets known?
		short				flagMask;			//: Flags act as a mask when navigating
		
		MEList				*nodeLabels;		//: User defined labels for nodes
		MEList				*branchLabels;		//: User defined labels for branches
		
	public:
	 						METree();
	 	virtual				~METree();

		virtual void		cloneTree(METree *inTree);

		virtual void		setTaxonList(METaxonList *inTaxonList, bool isOwner, bool merge=false);
		virtual METaxonList *getTaxonList() { return taxonList; };
		virtual void		createLocalTaxonList();

		virtual void		setTranslationList(METaxonList *inTranslationList);
		virtual METaxonList *getTranslationList() { return translationList; };
		
		virtual int			compareTaxa(METaxonList *inTaxonList);

#pragma mark === Node access ===

		virtual METaxon*	getTaxon(MELink* inTip);
		
		virtual char*		getTipName(int index);
		virtual char*		getTipName(MELink* inTip);
		virtual int 		getTipNumber(MELink* inTip);
		
		virtual MELink*		getTip(int index);
		virtual MELink*		findTip(char *inName);
		virtual MELink*		findTip(int inTaxonNo);

		virtual int 		countTipsWithMask();
		virtual MELink*		getFirstTip();
		virtual MELink*		getNextTip(MELink* previous);

		virtual MELink*		getInternalNode(int index);

		virtual MELink*		getFirstInternalNode();
		virtual MELink*		getNextInternalNode(MELink* previous);

		virtual MELink*		getNode(int index);

		virtual MELink*		getBranch(int index);

		void				setUseNumericalTips(bool inUseNumericalTips) 
								{ useNumericalTips=inUseNumericalTips; };
		bool				getUseNumericalTips() { return useNumericalTips; };
		

		void				setNumTips(int inNumTips) { numTips=inNumTips; };
		int					getNumTips() { return numTips; };
		void				setNumInternalNodes(int inNumIntNodes) { numIntNodes=inNumIntNodes; };
		int					getNumInternalNodes() { return numIntNodes; };
		void				setNumNodes(int inNumNodes) { numNodes=inNumNodes; };
		int					getNumNodes() { return numNodes; };
		void				setNumBranches(int inNumBranches) { numBranches=inNumBranches; };
		int					getNumBranches() { return numBranches; };
		
		int					getNodeDegree(MELink *inNode);
		int					getMaxNodeDegree();
		int					getMaxPolytomy() { return getMaxNodeDegree(); }
		
		bool				getIsRooted() { return isRooted; };
		void				setIsRooted(bool inIsRooted) { isRooted=inIsRooted; };
		bool				getHasLengths() { return hasLengths; };
		void				setHasLengths(bool inHasLengths) { hasLengths=inHasLengths; };

		int					getNumLinks() { return numLinks; };
		virtual MELink*		getLink(int index);
		
		virtual int			getLinkCapacity() { return linkCapacity; };
		virtual int			getBranchCapacity() { return branchCapacity; };

		virtual void		putSplit(MEBitset *split, int idx);	// Mike's new bit

		virtual MELink*		getRoot1() { return root1; };
		virtual void		setRoot1(MELink *inNode);
		virtual double		getRootDistance1();
		virtual MELink*		getRoot2() { return root2; };
		virtual void		setRoot2(MELink *inNode);
		virtual double		getRootDistance2();
		virtual void		setRootDistances(double inDist1, double inDist2);
		virtual bool		getRootDistancesKnown() { return rootDistsKnown; };
		virtual void		setRootDistancesKnown(bool inFlag) { rootDistsKnown = inFlag; };
		
#pragma mark === Node allocation ===
	protected:
		virtual MELink*		newLinkObject();
		virtual MELink*		createLink();
		
		virtual void		checkLinkCapacity(int requiredCapacity);
		virtual void		checkBranchCapacity(int requiredCapacity);

	public:
		virtual MELink*		newNode();
		virtual MELink*		newNode(int index);
		virtual MELink*		newBranch(MELink *inLink, double len);
		virtual MELink*		newTip(char *inName);
		virtual MELink*		newTip(int inTaxonNo);
		
		virtual void		removeLink(MELink *inLink);
		virtual void		removeTip(MELink *inLink);
		virtual void		removeInternalNode(MELink *inLink);
		virtual void		removeAllInternalNodes();
		virtual void		removeAllNodes();
		
#pragma mark === Branch lengths ===
		virtual double*		getBranchLengths() { return branchLengths; };
		virtual double		getBranchLength(int index);
		virtual double		getBranchLength(MELink *inLink);
		virtual double		getBranchLengthSum();
		virtual double		getMeanBranchLength();
		
		
		virtual int 		addBranchLength(double inLength);
		virtual void 		setBranchLength(int index, double inLength);
		virtual void 		setBranchLengths(double inLength);
		virtual void 		scaleBranchLengths(double inFactor);
		virtual void 		scaleBranchLengthsOfClade(MELink *inLink, double inFactor);
		virtual void 		raiseBranchLengths(double inPower);
		virtual void 		raiseRootDistances(double inPower);
		virtual void		removeBranchLength(int index);

		virtual double		getNodeHeight(int index);
		virtual void 		setNodeHeight(int index, double inHeight);
		virtual double		getRootHeight() { return rootHeight; }
		virtual void		setRootHeight(double inHeight) { rootHeight = inHeight; }

		virtual void		lengthsToHeights(double inRate = 1.0);
		virtual void		heightsToLengths(double inRate = 1.0);
	protected:
		virtual double		nodeLengthsToHeights(MELink *inNode, double inRate);
		virtual void		nodeHeightsToLengths(MELink *inNode, double inRate, bool isTesting = false);

		virtual void 		checkHeightCapacity(int requiredCapacity);

	public:
		virtual double		getTipDate(int index);
		virtual void 		setTipDate(int index, double inDate);
	protected:
		virtual void 		checkTipDateCapacity(int requiredCapacity);

	public:
		virtual void		reorderTaxa(TaxonOrder inTaxonOrder);
	protected:
		virtual void		reorderTaxaForNode(TaxonOrder inTaxonOrder, MELink *inNode);
		virtual int			reorderTaxaCompare(TaxonOrder inTaxonOrder, MELink *inNode1, MELink *inNode2);
		
#pragma mark === Node and Branch Labels ===
	public:
		virtual MEList*		getNodeLabelArrays() { return nodeLabels; }
		virtual int			getNumNodeLabelArrays();
		virtual char*		getNodeLabelName(int index);
		virtual MELabelArray*	getNodeLabelArray(int index);
		virtual int			findNodeLabelArray(char *inName);
		virtual void		addNodeLabelArray(MELabelArray *inLabelArray);
		virtual void		setNodeLabelArray(int index, MELabelArray *inLabelArray);

		virtual MEList*		getBranchLabelArrays() { return branchLabels; }
		virtual int			getNumBranchLabelArrays();
		virtual char*		getBranchLabelName(int index);
		virtual MELabelArray*	getBranchLabelArray(int index);
		virtual int			findBranchLabelArray(char *inName);
		virtual void		addBranchLabelArray(MELabelArray *inLabelArray);
		virtual void		setBranchLabelArray(int index, MELabelArray *inLabelArray);

#pragma mark === Tree Rearrangement ===
	public:
		virtual MELink*		coalesceNodes(MELink *l1, MELink *l2, double bLen1=1.0, double bLen2=1.0);
		virtual MELink*		coalesceNodes(int nodeNum1, int nodeNum2, double bLen1=1.0, double bLen2=1.0);
		
		virtual void		connectNodes(MELink *l1, MELink *l2, double bLen=1.0);
		virtual void		connectNodes(int nodeNum1, int nodeNum2, double bLen=1.0);
		
	protected:
		virtual void		pruneSubtree(MELink *inLink);
		virtual void		graftSubtreeToBranch(MELink *inSource, MELink *inDest, double inDist=1.0);
		virtual void		graftSubtreeToNode(MELink *inSource, MELink *inDest);
		virtual void		disposeSubtree(MELink *inLink);
		virtual void		disposeSubtreeNode(MELink *inLink);

	public:
		virtual void		cutAndPasteToBranch(MELink *inSource, MELink *inDest, double inDist=1.0);
		virtual void		cutAndPasteToNode(MELink *inSource, MELink *inDest);
		virtual void		cutSubtree(MELink *inSource);

		virtual void		resolveTree(double inLength = 0.0);
		virtual void		resolveSubtree(MELink *inLink, double inLength = 0.0);
		virtual void		resolveNode(MELink *inLink, double inLength = 0.0);
		
		virtual void		collapseSubtree(MELink *inLink);
		virtual void		collapseBranch(MELink *inLink);
		virtual void 		collapsePolytomies(double inThreshold = 0.0);
		
	protected:
		virtual void		collapseSubtreeNode(MELink *inLink);
		virtual void		resolveSubtreeNode(MELink *inLink, double inLength);

	public:
		virtual void		setStemLink(MELink *inLink);
		virtual void		rotateNode(MELink *inLink);
		
#pragma mark === Tree Rooting ===
		virtual void		outgroupTree(MELink *inLink);
		virtual void		outgroupTree(MELink *inLink, double inPosition);

		virtual MELink*		findMidPoint(int inTip1, int inTip2, double inMidPoint);
		virtual void		midPointTree();
		
		virtual void		swapRootNodes();
		virtual void		adjustRootDistances(double inMinFraction = 0.0);
		
		virtual void		rootTree();
		virtual void		unrootTree();

#pragma mark === Tree Navigation ===
		virtual void		clearAllNodeFlags();
		
		virtual void		setFlagForAllNodes(short inFlag);
		virtual void		toggleFlagForAllNodes(short inFlag);
		virtual void		clearFlagForAllNodes(short inFlag);
		virtual void		copyFlagForAllNodes(short inFlag, METree *inTree);
		virtual void		copyFlagForAllTips(short inFlag, METree *inTree);
		
		virtual void		setNodeFlagIfFlagSet(short inFlag, short ifFlag);
		virtual void		clearNodeFlagIfFlagSet(short inFlag, short ifFlag);
		virtual void		setNodeFlagIfFlagClear(short inFlag, short ifFlag);
		virtual void		clearNodeFlagIfFlagClear(short inFlag, short ifFlag);
		virtual int			getNumTipsWithFlag(short inFlag);
		virtual MELink*		findNodeWithFlagSet(MELink *inLink, short inFlag);
		
		virtual void		setFlagForClade(MELink *inLink, short inFlag);
		virtual void		toggleFlagForClade(MELink *inLink, short inFlag);
		virtual void		clearFlagForClade(MELink *inLink, short inFlag);
		
		virtual void		setTipFlagsFromTaxa();
		virtual void		setNodeFlagsFromTips();
		virtual void		setTaxonFlagsFromTips();

	protected:
		virtual void		getFlagsForTips(MEBitset *inNavBits, short &outFlags1, short &outFlags2);

//		virtual void 		setSubtreeFlagsAtNode1(MELink *inLink, short inFlags);
//		virtual void 		setSubtreeFlagsAtNode2(MELink *inLink, short inFlags);
		
	public:
		virtual MEBitset*	getNavBits(MELink *inLink);
		virtual MELink*		findNavBits(MEBitset* inNavBits);
		
	protected:
		virtual void 		calcNavBits();
		virtual void 		calcNavBitsAtNode(MELink *inLink, MEBitset *outBits);
		
	public:
		virtual double 		calcNodeHeight(MELink *inNode);		
		virtual double 		calcTreeHeight();
		
		virtual double 		calcNodeHeights(MELink *inNode, MEDoubleArray *outHeights);		
		virtual double 		calcTreeHeights(MEDoubleArray *outHeights);

		virtual int	 		calcNodeDensity(MELink *inNode);		
		virtual int	 		calcTipDensity(MELink *inNode);		

		virtual MELink*		getCommonAncestor(MELink *inNode1, MELink *inNode2);
		virtual MELink*		getPathMeet(MELink *inNode0, MELink *inNode1, MELink *inNode2);
		
		virtual double		getPath(MELink *inNode1, MELink *inNode2, MEIntArray *outPath);
		virtual double		getPath(MELink *inNode1, MELink *inNode2, MEBitset *outPath);
		virtual int			getPathLength(MELink *inNode1, MELink *inNode2);		
		virtual double 		getPathDistance(MELink *inNode1, MELink *inNode2);
		
	protected:
		virtual void 		calcPath(MELink *inNode1, MELink *inNode2,
										double &outDist,
										int &outLength,
										MEIntArray *outIntArray,
										MEBitset *outBitset);
		
	public:
		virtual double 		getMaxPathDistance();
		
		virtual double 		getTipToTipDistance(int inTip1, int inTip2);
		virtual double 		getRootToTipDistance(int inTip);
		virtual double 		getNodeToNodeDistance(int inNode1, int inNode2);
		
		virtual MELink*		getPointOnPath(double inPoint, MELink *inNode1, MELink *inNode2, double *outDist = NULL);
		virtual MELink*		getPointOnPath(int inPoint, MELink *inNode1, MELink *inNode2);

		virtual void 		setFlagMask(short inFlagMask) { flagMask = inFlagMask; }
		virtual short 		getFlagMask() { return flagMask; }
		virtual MELink*		getFirstBranch(MELink *inNode);
		virtual MELink*		getNextBranch(MELink *inNode, MELink *inLastBranch);
		virtual MELink*		getDescendantNode(MELink *inBranch, double *outDist);
		virtual MELink*		getDescendantNode(MELink *inBranch);

#pragma mark === Streaming ===
		virtual void		readObject(MEStream *stream);
		virtual void		readData(MEStream *stream);
		virtual void		writeObject(MEStream *stream);
		virtual void		writeData(MEStream *stream);
};

#endif /* _METREE_H_ */
