/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METreeConstructor.cpp

		METreeConstructor Class
					
 ******************************************************************************/
//! author = "Michael Charleston"
//! group = "Trees"

#pragma mark // includes:
// ANSI / C++
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
// Mol. Evol.
#include "MEExceptions.h"
#include "MEGlobal.h"
#include "MEBitset.h"
#include "METree.h"
#include "METaxonList.h"
#include "MEDistanceMatrix.h"
#include "METreeConstructor.h"


#pragma mark // methods:


void METreeConstructor::NJTree(MEDistanceMatrix *D, METree *T)
{
	// create a list of tip nodes:
	double
		rmin;
	int
		i, j,
		internals = 1,
		newnode,
		numCols = D->getNumCols(),
		ntax,
		x, y,
		numTips;
	
	METaxonList *taxonList = D->getTaxonList();
	if (taxonList) {
		ntax = taxonList->getNumTaxa();
		T->setTaxonList(taxonList, false, false);
	}
	else
		ntax = numCols;
	
	newnode = ntax;
	numTips = ntax;
	
	for (i = 0; i < ntax; i++) {
		T->newTip(i);
	}
	
	MEBitset *activeSet;
	NewClass(activeSet, MEBitset());
	activeSet->fill(ntax);
	
	double *netDiv;
	NewArray(netDiv, double, 2*ntax);
	
	// M is the matrix of (distances-net divergences), including those between internal nodes:
//	MEMatrix *M;
//	NewClass(M, MEMatrix(2*ntax));
//	M->clearMatrix();
	
	// Create a local copy of this matrix with 2x the dimensions 
	// (to accommodate internal nodes):
	MEMatrix *lD;
	NewClass(lD, MEMatrix(2*ntax));
	lD->clearMatrix();
	// Fill in the first quarter of this matrix with the current matrix values:
	for (i = 0; i < ntax; i++) {
		for (j = i+1; j < ntax; j++) {
			lD->setElement(i, j, D->getElement(i, j));
			lD->setElement(j, i, D->getElement(i, j));
		}
	}
	
	double d, lamda, xLen = 0.0, yLen = 0.0; 
	
	for (i = 0; i < 2*ntax; i++)
		netDiv[i] = 0.0;	// (initially all net divergences are zero)

	while (numTips >= 2) { // While we still have two branches
	
		// choose first two elements in activeSet:
		x = activeSet->getFirstElement();
		y = activeSet->getNextElement();
			
		// calculate the net divergences:
		
		for (i = x; i < newnode; i++) {
			if (activeSet->contains(i)) {
				netDiv[i] = 0.0;
				for (j = x; j < newnode; j++) {
					if (activeSet->contains(j)) {
						netDiv[i] += lD->getElement(i,j);
						// netDiv[i] is the sum of the distances between i and all
						// the other taxa in the active set.
					}
				}
			}
		}
//		for (i = activeSet->getFirstElement(); i >= 0; i = activeSet->getNextElement()) {
//			cerr << "netDiv[" << i << "]\t=\t" << netDiv[i] << "\n";
//		}
		
		// find min_(i,j in activeSet) of { M_(i,j) - (netDiv_i + netDiv_j)/(num tips - 2) } :
		rmin = kBigReal;
		
		// set up the new entries in the matrix M from which we need the minimal 
		// non-diagonal entry:
		for (i = x; i < newnode-1; i++) {
			if (activeSet->contains(i)) {
				for (j = i+1; j < newnode; j++) {
					if (activeSet->contains(j)) {
						d = lD->getElement(i, j) - (netDiv[i] + netDiv[j])/((double) numTips - 2.0);
						if (d < rmin) {
							rmin = d;
							x = i;
							y = j;
						}
					}
				}
			}
		}
		
//		cerr << "x = " << x << "; y = " << y << "\n";
		activeSet->del(x);
		activeSet->del(y);
//		cerr << "Active set <- " << *activeSet << "\n";
		
		
		d = lD->getElement(x, y);
				
		// we shall now coalesce these tips indexed x & y.
		if (numTips > 2) {
			// estimate the distances within the tree:
//			cerr << "d(x,y) = " << d << "\n";
			lamda = (netDiv[x] - netDiv[y]) / (2.0*((double)numTips - 2.0));
//			cerr << "lamda = " << lamda << "\n";
			
			xLen = d/2.0 + lamda;
			yLen = d - xLen;//lD->getElement(x, y) - d;
//			cerr << "xLen = " << xLen << "; yLen = " << yLen << "\n";
			if (xLen < 0.0)
				xLen = 0.0;
			if (yLen < 0.0)
				yLen = 0.0;
	
			lD->setElement(x, newnode, xLen);
			lD->setElement(y, newnode, yLen);
								
			// put in the internal distances:
			for (i = activeSet->getFirstElement(); i >= 0; i = activeSet->getNextElement()) {
				d = lD->getElement(i,x);
				d += lD->getElement(i,y);
				d -= lD->getElement(x,y);
				d /= 2.0;
				lD->setElement(i, newnode, d);
				lD->setElement(newnode, i, d);
//				cerr << "new d(" << newnode << "," << i << ") = " << d << "\n";
			}
			
//			for (i = 0; i < newnode; i++) {
//				for (j = i+1; j <= newnode; j++) {
//					cerr << "d(" << i << "," << j << ") = " << lD->getElement(i,j) << "\n";
//				}
//			}

			// coalesce the nodes with minimal M value:
			T->coalesceNodes(x, y, xLen, yLen);
		}
		else {	// join the last two links
			T->connectNodes(x, y, d);
		}
		
		// update active set:
		activeSet->add(newnode);
		
		// bookkeeping:
		numTips--;
		internals++;
		newnode++;
		
	}
	
	T->setIsRooted(false);
	T->setHasLengths(true);
	T->midPointTree();
	
	// clear up:
	delete activeSet;
	delete [] netDiv;
//	delete M;
	delete lD;
}



void METreeConstructor::UPGMATree(MEDistanceMatrix *D, METree *T)
{
	ThrowIfNull(D);
	ThrowIfNull(T);
	
	double
		d, lamda, rmin, 
		xLen, yLen;	// edge lengths for the coalesced tips
	int
		i, j,	// convenient counters
		ntax = D->getNumCols(),	// number of taxa, of course
		x, y;	// the two tips to coalesce
	
	MEMatrix *M;
	NewClass(M, MEMatrix(2*ntax, 2*ntax));
	for (i = 0; i < ntax; i++) {
		for (j = 0; j < ntax; j++) {
			M->setElement(i, j, D->getElement(i, j)); 
		}
	}
	
	// make an active set of tips:
	MEBitset activeSet(2*ntax);
	activeSet.fill(ntax);	// to begin with all the tips are available
	
	int *cladeSize;
	NewArray(cladeSize, int, 2*ntax);
	for (i = 0; i < ntax; i++)
		cladeSize[i] = 1;
		
	int newnode = ntax;
	int numTips = ntax;
	for (int i = 0; i < ntax; i++)
		T->newTip(i);
	
	while (numTips > 2) {
		// choose the first two elements in $active:
		x = activeSet.getFirstElement();
		y = activeSet.getNextElement();
		
		// get a first value for the min. pairwise distance:
		rmin = M->getElement(x, y);
		
		// now find the actual minimum and the pair (x,y) to join:
		for (i = x; i < newnode-1; i++) {
			if (activeSet.contains(i)) {
				for (j = 0; j < newnode; j++) {
					if (activeSet.contains(j)) {
						d = M->getElement(i, j);
						if (d < rmin) {
							rmin = d;
							x = i;
							y = j;
						}
					}
				}
			}
		}
		
		numTips--;
		
		if (numTips >= 2) {
			// keep track of the number of tips descendant from the new node:
			cladeSize[newnode] = cladeSize[x] + cladeSize[y];
			
			// take out x and y from the active set:
			activeSet -= x;
			activeSet -= y;
			
			// add in the new node to the active set:
			activeSet += newnode;
			
			// note the estimated distance between this tip and the remainder in $activeSet:
			for (i = activeSet.getFirstElement(); i >= 0; i = activeSet.getNextElement()) {
				d = (cladeSize[x] * M->getElement(i, newnode) + cladeSize[y] * M->getElement(i, newnode))
						/ cladeSize[newnode];
				M->setElement(i, newnode, d);
				M->setElement(newnode, i, d);
			}
			
			// estimate the distances within the tree:
			// lamda is the mean difference between D_i,x and D_i,y for all active nodes i.
			lamda = 0.0;
			for (i = activeSet.getFirstElement(); i >= 0; i = activeSet.getNextElement())
				lamda += M->getElement(i, x) - M->getElement(i, y);
			lamda /= ((double) numTips - 2.0);
			
			// put in the internal distances:
			for (i = activeSet.getFirstElement(); i >= 0; i = activeSet.getNextElement()) {
				d = M->getElement(i,x) - (M->getElement(x,y) + lamda)/2.0;
				M->setElement(i, newnode, d);
				M->setElement(newnode, i, d);
			}
			
			// get the distances from x and y to newnode:
			xLen = M->getElement(x, newnode);
			yLen = M->getElement(y, newnode);
			
			// coalesce the nodes with minimal M value:
			T->coalesceNodes(x, y, xLen, yLen);
			
		}
		else {
			T->connectNodes(x, y, M->getElement(x, y));
		}
	}
	
	delete [] cladeSize;
	delete M;
}
