/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METreeGenerator.h

		Interface for the METreeGenerator Class

 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Trees"

#pragma mark // includes:
// ANSI / C++
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
// Mol. Evol.
#include "MEExceptions.h"
#include "MEGlobal.h"
#include "MEStream.h"

#include "METree.h"
#include "METaxonList.h"

#include "METreeGenerator.h"


#pragma mark // methods:


// ---------------------------------------------------------------------------------
//		 METreeGenerator::METreeGenerator
// ---------------------------------------------------------------------------------

METreeGenerator::METreeGenerator()
{
	mSourceTree = NULL;
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::~METreeGenerator
// ---------------------------------------------------------------------------------

METreeGenerator::~METreeGenerator()
{
	if (mSourceTree != NULL && mSourceTree->getOwner() == this) 
		delete mSourceTree;
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::setSourceTree
// ---------------------------------------------------------------------------------

void METreeGenerator::setSourceTree(METree *inTree)
{
	mSourceTree = inTree;
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::getNumTrees
// ---------------------------------------------------------------------------------
// Returns the number of trees. If there are just too many trees to put in
// a long int then this returns -1. The getApproxNumTrees should be called
// instead. The maximum number of trees that can be return exactly is
// 2,000,000,000 and this is probably far more than can be generated anyway.

long METreeGenerator::getNumTrees()
{
	ThrowIfNull(mSourceTree);
	
	MELink *node = mSourceTree->getFirstInternalNode();
	int degree;
	long Nr, prod = 1;
	while (node != NULL) {
		degree = mSourceTree->getNodeDegree(node);

		Nr = countResolutions(degree);
		
		if (Nr==-1)
			return -1;
			
		prod *= Nr;
		
		node = mSourceTree->getNextInternalNode(node);
	}
	
	return prod;
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::countResolutions
// ---------------------------------------------------------------------------------

long METreeGenerator::countResolutions(int degree)
{
	if (degree > 11)
		return -1;
		
	if (degree < 4)
		return 1;
		
	long prod1 = 1.0;
	int i, x = 2 * degree - 5;
	
	for (i=1; i<=x; i++)	
		prod1 *= i;
	
	long prod2 = 1.0;
	x = degree - 3;
	
	for (i=1; i<=x; i++)
		prod2 *= i;

	long Nr = prod1 / (prod2 * pow(2, degree - 3));
	
	return Nr;
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::getApproxNumTrees
// ---------------------------------------------------------------------------------

// Returns the approximate number of trees as a floating point number.
double METreeGenerator::getApproxNumTrees()
{
	ThrowIfNull(mSourceTree);
	
	MELink *node = mSourceTree->getFirstInternalNode();
	double sum = 0.0;
	int degree;
	while (node != NULL) {
		degree = mSourceTree->getNodeDegree(node);

		sum += countApproxResolutions(degree);
		
		node = mSourceTree->getNextInternalNode(node);
	}
	
	return pow(10.0, sum);
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::countApproxResolutions
// ---------------------------------------------------------------------------------

double METreeGenerator::countApproxResolutions(int degree)
{
	double sum1=0;
	int i, x=2*degree-5;
	for (i=1; i<=x; i++)	
		sum1+=log10(i);
	
	double sum2=0;
	x=degree-3;
	for (i=1; i<=x; i++)
		sum2+=log10(i);

	double logNr=sum1-((degree-3)*log10(2))-sum2;
	
	return logNr;
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::generate
// ---------------------------------------------------------------------------------

void METreeGenerator::generate(int index, METree *inTree)
{
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::generateRandom
// ---------------------------------------------------------------------------------

void METreeGenerator::generateRandom(METree *inTree)
{
	ThrowIfNull(mSourceTree);
	
	inTree->clone(mSourceTree);
	inTree->resolveTree();
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::generateFirst
// ---------------------------------------------------------------------------------

void METreeGenerator::generateFirst(METree *inTree)
{
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::generateNext
// ---------------------------------------------------------------------------------

void METreeGenerator::generateNext(METree *inTree)
{
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::readObject
// ---------------------------------------------------------------------------------

// Read the fields of this object from stream.
void METreeGenerator::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
	mSourceTree=(METree *)stream->get("Source");
}

// ---------------------------------------------------------------------------------
//		 METreeGenerator::writeObject
// ---------------------------------------------------------------------------------

// Write the fields of this object to stream.
void METreeGenerator::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
	stream->put("Source", mSourceTree);
}


