/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METreeList.cpp

				METreeList Class
			
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "METaxonList.h"
#include "METreeList.h"
#include "METree.h"
#include "MEStream.h"


#pragma mark === METreeList ===

// ---------------------------------------------------------------------------------
//		 METreeList
// ---------------------------------------------------------------------------------

METreeList::METreeList() : 
	MEList()
{
	taxonList=NULL;
}

// ---------------------------------------------------------------------------------
//		 ~METreeList
// ---------------------------------------------------------------------------------

METreeList::~METreeList()
{
	if (taxonList!=NULL && taxonList->getOwner()==this) {
		int i, n = getNumTrees();
		METree *tree;
		for (i = 0; i < n; i++) {
			tree = getTree(i);
			if (tree->getTaxonList() == taxonList)
				tree->setTaxonList(NULL, false, false);
		}

		delete taxonList;
	}
}

// ---------------------------------------------------------------------------------
//		METreeList::createObject
// ---------------------------------------------------------------------------------

MEObject* METreeList::createObject()
{
	MEObject *obj;
	NewClass(obj, METree());
	return obj;
}

// ---------------------------------------------------------------------------------
//		 METreeList::getTree
// ---------------------------------------------------------------------------------

METree* METreeList::getTree(int index)
{
	return (METree *)getObject(index);
}

// ---------------------------------------------------------------------------------
//		 METreeList::addTree
// ---------------------------------------------------------------------------------

int METreeList::addTree(METree *inTree)
{
	return add(inTree, inTree->getOwner()==NULL);
}

// ---------------------------------------------------------------------------------
//		 METreeList::add
// ---------------------------------------------------------------------------------

int METreeList::add(MEObject *inObject, bool isOwner)
{
	// If we have and own the taxonList, then when we
	// add the first tree, the taxa are then set as
	// defined so that all subsequent trees using this taxonList
	// must use these taxa.
	if (taxonList!=NULL && taxonList->getOwner()==this)
		taxonList->setTaxaDefined(true);
		
	return MEList::add(inObject, isOwner);
}

// ---------------------------------------------------------------------------------
//		 METreeList::insertTree
// ---------------------------------------------------------------------------------

void METreeList::insertTree(int index, METree *inTree)
{
	insert(index, inTree, inTree->getOwner()==NULL);
}

// ---------------------------------------------------------------------------------
//		 METreeList::insert
// ---------------------------------------------------------------------------------

void METreeList::insert(int index, MEObject *inObject, bool isOwner)
{
	// If we have and own the taxonList, then when we
	// add the first tree, the taxa are then set as
	// defined so that all subsequent trees using this taxonList
	// must use these taxa.
	if (taxonList!=NULL && taxonList->getOwner()==this)
		taxonList->setTaxaDefined(true);
		
	MEList::insert(index, inObject, isOwner);
}

// ---------------------------------------------------------------------------------
//		 METreeList::removeTree
// ---------------------------------------------------------------------------------

METree* METreeList::removeTree(int index)
{
	return (METree *)remove(index);
}


// ---------------------------------------------------------------------------------
//		 METreeList::setTaxonList
// ---------------------------------------------------------------------------------

// The taxonList options are:
//   1) Each tree owns its own taxonList, with treeList's taxonList set to NULL
//   2) The treeList owns a taxonList, and each tree points to this
//   3) The trees use an external taxonList (such as an Alignment), i.e. the
//      trees point to it and the treeList points to it but none of them owns it.
//   4) Trees point to any manner of other taxonLists.

// Replaces the existing taxonList with a new one, calling setTaxonList for each tree.
// If merge is true then taxa in a tree that are missing from the new taxon list
// are added otherwise an exception is thrown.
void METreeList::setTaxonList(METaxonList *inTaxonList, bool isOwner, bool merge)
{
	if (inTaxonList != NULL) {
		for (int i=0; i<numObjects; i++)
			((METree*)objects[i])->setTaxonList(inTaxonList, false, merge);

		if (taxonList!=NULL && taxonList->getOwner()==this)
			delete taxonList;

		taxonList=inTaxonList;
		if (isOwner)
			taxonList->setOwner(this);
	} else {
		METree *tree;
		for (int i=0; i<numObjects; i++) {
			tree = getTree(i);
			tree->createLocalTaxonList();
		}
		
		if (taxonList!=NULL && taxonList->getOwner()==this)
			delete taxonList;

		taxonList=NULL;
	}
}


// ---------------------------------------------------------------------------------
//		 METreeList::newTree
// ---------------------------------------------------------------------------------

// NewArrays a new tree passing the taxonList and other flags
METree *METreeList::newTree()
{
	METree *tree;
	
	NewClass(tree, METree());
	tree->setOwner(this);
	
	if (taxonList!=NULL)
		tree->setTaxonList(taxonList, false);
	
	return tree;
}

#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 METreeList::readObject
// ---------------------------------------------------------------------------------

void METreeList::readObject(MEStream *stream)
{
	MEList::readObject(stream);
	
	taxonList=(METaxonList *)stream->get("Taxa");
}

// ---------------------------------------------------------------------------------
//		 METreeList::writeObject
// ---------------------------------------------------------------------------------

void METreeList::writeObject(MEStream *stream)
{
	MEList::writeObject(stream);
	
	stream->put("Taxa", taxonList);
}


