/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METreeSet.cpp

				METreeSet Class
				
 ******************************************************************************/

#include <stdio.h>

#include "METreeSet.h"

#include "MEGlobal.h"
#include "MEExceptions.h"

#include "METaxon.h"
#include "METaxonList.h"
#include "METree.h"
#include "MELink.h"
#include "MEStream.h"

// Trees are stored using an array of unsigned longs. mTrees is an array
// of pointers to these arrays. They are of a fixed size (mNumTips * 2)
// so mNumTips must be set first (or set by adding the first tree in). All
// subsequent trees must have the same number of tips and the same taxon 
// names. The trees are difficult to navigate because they are stored as
// indices to the node above (you can go up but not down). This is very
// compact but inconvenient. The intended use is to expand the tree into
// an METree using 'getTree' before use. Not storing branch lengths
// will save memory considerably.

// =========================================================================
//	union LongUnion
// =========================================================================

union LongUnion {
	unsigned long value;
	unsigned char bytes[4];
};

#pragma mark === METreeList ===

// =========================================================================
//	METreeSet::mCapacityIncrement
// =========================================================================

int METreeSet::mCapacityIncrement = 100;

// =========================================================================
//	METreeSet::METreeSet
// =========================================================================

METreeSet::METreeSet() : MEObject()
{
	mTaxonList = NULL;
		
	mNumTrees = 0;
	mNumTips = 0;
	mStoreLengths = true;
		
	mTrees = NULL;
	mLengths = NULL;
		
	mCapacity = 0;
}

// =========================================================================
//	METreeSet::~METreeSet
// =========================================================================

METreeSet::~METreeSet()
{
	if (mTaxonList != NULL)
		mTaxonList->removeReference(true);
		
	if (mTrees != NULL) {
		removeAll();
		delete [] mTrees;
	}
	
	if (mLengths != NULL)
		delete [] mLengths;
}
		
// =========================================================================
//	METreeSet::setNumTips
// =========================================================================

void METreeSet::setNumTips(long inNumTips)
{
	if (mNumTrees > 0 && inNumTips != mNumTips) {
		removeAll();
	}
	
	mNumTips = inNumTips;
	mNumInternalNodes = mNumTips - 1;
	// The number of indices required to store the tree.
	// First mNumTips for the tips and then mNumInternalNodes for
	// the internal nodes. This is for a rooted, bifucating tree,
	// so represents the maximum for any tree shape.
	mNumIndices = mNumTips + mNumInternalNodes;
	
	// Set the number of bytes required to store indices up to mNumIndices.
	// The internal node indices are given from 0 so mNumTips is added to
	// find the node in the list.
	if (mNumInternalNodes <= 256) 			// 2^8
		mNumBytes = 1;
	else if (mNumInternalNodes <= 65536)	// 2^16
		mNumBytes = 2;
	else if (mNumInternalNodes <= 16777216)	// 2^24
		mNumBytes = 3;
	else
		mNumBytes = 4;
}

// =========================================================================
//	METreeSet::setStoreLengths
// =========================================================================

void METreeSet::setStoreLengths(bool inStoreLengths)
{
	if (inStoreLengths != mStoreLengths) {
		long i, j, n = mNumTips*2;
		LengthReal *P;
		
		if (inStoreLengths) {
			// mLengths should be NULL or something has screwed up.
			ThrowIf(mLengths != NULL);
			
			if (mCapacity > 0) {
			// If capacity is already non zero then allocate some
			// branch length arrays.
			
				mLengths = new LengthReal*[mCapacity];

				// If mNumTrees is non zero then allocate some
				// actual branch lengths for each tree and set them
				// to a default length of 1.0.
				for (i=0; i < mNumTrees; i++) {
					mLengths[i] = new LengthReal[n];
					P = mLengths[i];
					for (j=0; i < n; i++) {
						*P = 1.0;
						P++;
					}
				}
				for (i=mNumTrees; i < mCapacity; i++)
					mLengths[i] = NULL;
			}
		} else {
			if (mLengths != NULL) {
				// Dispose of any existing branch lengths
				for (i=0; i < mNumTrees; i++)
					delete [] mLengths[i];
						
				delete [] mLengths;
				mLengths = NULL;
			}
		}
		
		mStoreLengths = inStoreLengths;
	}
}

// =========================================================================
//	METreeSet::setTaxonList
// =========================================================================

void METreeSet::setTaxonList(METaxonList *inTaxonList)
{
	if (mTaxonList != NULL)
		mTaxonList->removeReference();
		
	mTaxonList = inTaxonList;
}

// ---------------------------------------------------------------------------------
//		 METreeSet::newTip
// ---------------------------------------------------------------------------------

long METreeSet::newTip(char *inName)
{
	if (mTaxonList==NULL) {
		NewClass(mTaxonList, METaxonList());
		mTaxonList->setOwner(this);
	}
	
	int taxonNo = mTaxonList->findName(inName);

	if (taxonNo==-1) {
		if (mTaxonList->getOwner()!=this && mTaxonList->getTaxaDefined()) {
			throw METaxonMissing();
		} else {
			METaxon* taxon;
			NewClass(taxon, METaxon());
			taxon->setName(inName);
			taxonNo = mTaxonList->addTaxon(taxon);
		}
	}
	
	return taxonNo;
}


// ---------------------------------------------------------------------------------
//		 METreeSet::newTip
// ---------------------------------------------------------------------------------

long METreeSet::newTip(long inTaxonNo)
{
	ThrowIf(inTaxonNo<0);
	
	int taxonNo = inTaxonNo;
	
	if (mTaxonList==NULL) {
		NewClass(mTaxonList, METaxonList());
	}

	if (taxonNo >= mTaxonList->getNumTaxa() || mTaxonList->getTaxon(taxonNo)==NULL) {
		if (mTaxonList->getOwner()!=this && mTaxonList->getTaxaDefined()) {
			throw METaxonMissing();
		} else {
			METaxon* taxon;
			NewClass(taxon, METaxon());
			char name[32];
			sprintf(name, "%d", inTaxonNo+1);
			taxon->setName(name);
			mTaxonList->setObject(taxonNo, taxon);
		}
	}
	
	return taxonNo;
}


// =========================================================================
//	METreeSet::addTree
// =========================================================================

// Adds a tree without any structure which must then be specified.
long METreeSet::addTree()
{
	setTree(mNumTrees, NULL);
	return mNumTrees - 1;
}

// =========================================================================
//	METreeSet::addTree
// =========================================================================

long METreeSet::addTree(unsigned char *inTree, LengthReal *inLengths, long inNumTips)
{
	setTree(mNumTrees, inTree, inLengths, inNumTips);
	return mNumTrees - 1;
}

// =========================================================================
//	METreeSet::addTree
// =========================================================================

long METreeSet::addTree(METree *inTree)
{
	setTree(mNumTrees, inTree);
	return mNumTrees - 1;
}

// =========================================================================
//	METreeSet::insertTree
// =========================================================================

void METreeSet::insertTree(long index, unsigned char *inTree, LengthReal *inLengths, long inNumTips)
{
	TestBounds(index, mNumTrees + 1);
	
	checkCapacity(index + 1);
	for (long i = mNumTrees; i > index; i--)
		mTrees[i+1] = mTrees[i];
	setTree(index, inTree, inLengths, inNumTips);
}

// =========================================================================
//	METreeSet::insertTree
// =========================================================================

void METreeSet::insertTree(long index, METree *inTree)
{
	TestBounds(index, mNumTrees + 1);
	
	checkCapacity(index + 1);
	for (long i = mNumTrees; i > index; i--)
		mTrees[i+1] = mTrees[i];
	setTree(index, inTree);
}

// =========================================================================
//	METreeSet::removeTree
// =========================================================================

void METreeSet::removeTree(long index, unsigned char **inTree, LengthReal **inLengths)
{
	TestBounds(index, mNumTrees);
	
	if (inTree != NULL)
		*inTree = mTrees[index];
	else if (mTrees[index] != NULL)
		delete [] mTrees[index];
	
	if (mStoreLengths) {
		if (inLengths != NULL)
			*inLengths = mLengths[index];
		else if (mLengths[index] != NULL)
			delete [] mLengths[index];
	}
			
	for (long i = index; i < mNumTrees; i++)
		mTrees[i] = mTrees[i+1];
		
	mNumTrees--;
}

// =========================================================================
//	METreeSet::removeTree
// =========================================================================

void METreeSet::removeTree(long index, METree *inTree)
{
	TestBounds(index, mNumTrees);
	
	if (inTree != NULL)
		getTree(index, inTree);
		
	if (mTrees[index] != NULL)
		delete [] mTrees[index];
		
	if (mStoreLengths && mLengths[index] != NULL)
		delete [] mLengths[index];
		
	for (long i = index; i < mNumTrees; i++)
		mTrees[i] = mTrees[i+1];
		
	mNumTrees--;
}

// =========================================================================
//	METreeSet::removeAll
// =========================================================================

void METreeSet::removeAll()
{
	long i;
	
	if (mTrees != NULL) {
		for (i = 0; i < mNumTrees; i++) {
			if (mTrees[i] != NULL) 
				delete [] mTrees[i];
		}
		delete [] mTrees;
	}
	if (mStoreLengths && mLengths != NULL) {
		for (i = 0; i < mNumTrees; i++) {
			if (mLengths[i] != NULL) 
				delete [] mLengths[i];
		}
		delete [] mLengths;
	}
	
	mNumTrees = 0;
}
	
// =========================================================================
//	METreeSet::getTreeArray
// =========================================================================

unsigned char *METreeSet::getTreeArray(long inTreeIndex)
{
	TestBounds(inTreeIndex, mNumTrees);
	
	return mTrees[inTreeIndex];
}

// =========================================================================
//	METreeSet::getLengthsArray
// =========================================================================

LengthReal *METreeSet::getLengthsArray(long inTreeIndex)
{
	TestBounds(inTreeIndex, mNumTrees);
	
	if (mStoreLengths)
		return mLengths[inTreeIndex];
	else
		return NULL;
}

// =========================================================================
//	METreeSet::getTreeElement
// =========================================================================

long METreeSet::getTreeElement(long inTreeIndex, long inNodeIndex)
{
	TestBounds(inTreeIndex, mNumTrees);
	TestBounds(inNodeIndex, mNumIndices);
	
	return getNodeIndex(mTrees[inTreeIndex], inNodeIndex);
}

// =========================================================================
//	METreeSet::setTreeElement
// =========================================================================

void METreeSet::setTreeElement(long inTreeIndex, long inNodeIndex, long inAncestorIndex)
{
	TestBounds(inTreeIndex, mNumTrees);
	TestBounds(inNodeIndex, mNumIndices);
	
	setNodeIndex(mTrees[inTreeIndex], inNodeIndex, inAncestorIndex);
}

// =========================================================================
//	METreeSet::getBranchLength
// =========================================================================

double METreeSet::getBranchLength(long inTreeIndex, long inNodeIndex)
{
	TestBounds(inTreeIndex, mNumTrees);
	TestBounds(inNodeIndex, mNumIndices);
	
	return (double)mLengths[inTreeIndex][inNodeIndex];
}

// =========================================================================
//	METreeSet::setBranchLength
// =========================================================================

void METreeSet::setBranchLength(long inTreeIndex, long inNodeIndex, double inBranchLength)
{
	TestBounds(inTreeIndex, mNumTrees);
	TestBounds(inNodeIndex, mNumIndices);
	
	mLengths[inTreeIndex][inNodeIndex] = (LengthReal)inBranchLength;
}

// =========================================================================
//	METreeSet::getTree
// =========================================================================

METree *METreeSet::getTree(long index)
{
	TestBounds(index, mNumTrees);
	
	METree *tree;
	NewClass(tree, METree());
	getTree(index, tree);

	return tree;
}

// =========================================================================
//	METreeSet::getTree
// =========================================================================

void METreeSet::getTree(long index, METree *inTree)
{
	TestBounds(index, mNumTrees);

	MELink **links = new MELink*[mNumTips];
	long *indices = new long[mNumTips];
	int i, u, v;
	bool found;
	int numLinks = mNumTips;
	
	for (i = 0; i < mNumTips; i++) {
		indices[i] = getNodeIndex(mTrees[index], i);
		links[i] = inTree->newTip(i);
	}
	
	while (numLinks > 1) {
		
		u = 0;
		found = false;
		while (!found && u < numLinks - 1) {
			
			v = u + 1;
			while (!found && v < numLinks) {
				if (indices[u] == indices[v])
					found = true;
				else
					v++;
			}
			
			if (!found)	
				u++;
		}
		
		ThrowIf(!found); // there should be a match!
		
		
	}	
}

// =========================================================================
//	METreeSet::setTree
// =========================================================================

void METreeSet::setTree(long index, unsigned char *inTree, LengthReal *inLengths, long inNumTips)
{
	if (mNumTips == 0) {
		setNumTips(inNumTips);
	} else {
		if (inNumTips != mNumTips)
			throw MEBadNumTaxa();
	}
	
	checkCapacity(index + 1);
	if (index >= mNumTrees)
		mNumTrees = index + 1;
	
	long i, n = mNumIndices * mNumBytes;
	unsigned char *P, *Q;
	if (mTrees[index] == NULL) 
		mTrees[index] = new unsigned char[n];
		
	P = mTrees[index];
	Q = inTree;
	for (i = 0; i < n; i++) {
		*P = *Q;
		P++;
		Q++;
	}
	
	if (mStoreLengths) {
		LengthReal *R, *S;
		if (mLengths[index] == NULL) 
			mLengths[index] = new LengthReal[mNumIndices];
			
		ThrowIfNull(inLengths);
		
		R = mLengths[index];
		S = inLengths;
		for (i = 0; i < n; i++) {
			*R = *S;
			R++;
			S++;
		}
	}
}

// =========================================================================
//	METreeSet::setTree
// =========================================================================

void METreeSet::setTree(long index, METree *inTree)
{
	if (mNumTips == 0) {
		setNumTips(inTree->getNumTips());
	} else {
		if (inTree->getNumTips() != mNumTips)
			throw MEBadNumTaxa();
	}
	
	checkCapacity(index + 1);
	if (index >= mNumTrees)
		mNumTrees = index + 1;
	
	long i, n = mNumIndices * mNumBytes;
	unsigned char *P;
	if (mTrees[index] == NULL) 
		mTrees[index] = new unsigned char[n];
		
	// Clear the tree with zeros
	P = mTrees[index];
	for (i = 0; i < n; i++) {
		*P = 0;
		P++;
	}
	
	if (mStoreLengths) {
		LengthReal *R;
		if (mLengths[index] == NULL) 
			mLengths[index] = new LengthReal[mNumIndices];
			
		// Clear the tree with zeros
		R = mLengths[index];
		for (i = 0; i < mNumIndices; i++) {
			*R = 0.0;
			R++;
		}
	}

	if (inTree != NULL) {
		long current = 1;
		setTreeNode(index, &current, 0, inTree, inTree->getRoot1());
		setTreeNode(index, &current, 0, inTree, inTree->getRoot2());
		
		// node index of zero represents the root (which will always be at mNumTips).
		setNodeIndex(mTrees[index], mNumTips, 0);
		if (mStoreLengths)
			mLengths[index][mNumTips] = 0.0;
	}
}

// =========================================================================
//	METreeSet::setTreeNode
// =========================================================================

void METreeSet::setTreeNode(int inTreeIndex, long *ioCurrent, long inAncestor, METree *inTree, MELink *inNode)
{
	if (inNode->getIsTip()) {
		// Ancestor of the tip is inAncestor
		setNodeIndex(mTrees[inTreeIndex], inNode->getTipNo(), inAncestor);
		
		if (mStoreLengths)
			mLengths[inTreeIndex][inNode->getTipNo()] = (LengthReal)inTree->getBranchLength(inNode->getBranchNo());
			
	} else {
		// Ancestor of the node is inAncestor
		long nodeIndex = *ioCurrent;
		(*ioCurrent)++;
		setNodeIndex(mTrees[inTreeIndex], nodeIndex + mNumTips, inAncestor);
		
		if (mStoreLengths)
			mLengths[inTreeIndex][nodeIndex + mNumTips] = (LengthReal)inTree->getBranchLength(inNode->getBranchNo());
			
		
		MELink *link = inNode->next;
		while (link != inNode) {
			setTreeNode(inTreeIndex, ioCurrent, nodeIndex, inTree, link->branch);
			link = link->next;
		}
	}
}

// =========================================================================
//	METreeSet::getNodeIndex
// =========================================================================

long METreeSet::getNodeIndex(unsigned char *inTree, long inIndex)
{
	unsigned char *P = inTree + (inIndex * mNumBytes);
	LongUnion value;
	value.value = 0;
	
	for (int i = 3 - mNumBytes; i < 4; i++) {
		value.bytes[i] = *P;
		P++;
	}
	
	return value.value;
}

// =========================================================================
//	METreeSet::setNodeIndex
// =========================================================================

void METreeSet::setNodeIndex(unsigned char *inTree, long inIndex, long inValue)
{
	unsigned char *P = inTree + (inIndex * mNumBytes);
	LongUnion value;
	value.value = inValue;
	
	for (int i = 3 - mNumBytes; i < 4; i++) {
		*P = value.bytes[i];
		P++;
	}
}

// =========================================================================
//	METreeSet::checkCapacity
// =========================================================================

void METreeSet::checkCapacity(long inCapacity)
{
	if (inCapacity > mCapacity) {
		while (inCapacity > mCapacity)
			mCapacity += mCapacityIncrement;
			
		long i;
		
		// Allocate a new array
		unsigned char **newTrees = new unsigned char *[mCapacity];
		// Copy existing trees into it
		for (i = 0; i < mNumTrees; i++)
			newTrees[i] = mTrees[i];
		// Set the remaining capacity to NULL (this is for ease of debugging)
		for (i = mNumTrees; i < mCapacity; i++)
			newTrees[i] = NULL;
		// Delete the old array
		delete [] mTrees;
		// And copy in the new one
		mTrees = newTrees;

		if (mStoreLengths) {
			// Repeat the process for the branch lengths
			LengthReal **newLengths = new LengthReal*[mCapacity];
		
			for (i = 0; i < mNumTrees; i++)
				newLengths[i] = mLengths[i];
			for (i = mNumTrees; i < mCapacity; i++)
				newLengths[i] = NULL;
				
			delete [] mTrees;
			mLengths = newLengths;
		}
		
	}
}

#pragma mark === Object I/O ===

// ---------------------------------------------------------------------------------
//		 METreeSet::readObject
// ---------------------------------------------------------------------------------

void METreeSet::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
	
	mTaxonList=(METaxonList *)stream->get("Taxa");
}

// ---------------------------------------------------------------------------------
//		 METreeSet::writeObject
// ---------------------------------------------------------------------------------

void METreeSet::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
	
	stream->put("Taxa", mTaxonList);
}


