/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 METreeSet.h

		Interface for the METreeSet Class
					
 ******************************************************************************/
//! author = "Andrew Rambaut"
//! group = "Trees"
 
#ifndef _METREESET_H_
#define _METREESET_H_

#include "MEObject.h"

// Trees are stored using an array of unsigned integers. mTrees is an array
// of pointers to these arrays. They are of a fixed size (mNumTips * 2)
// so mNumTips must be set first (or set by adding the first tree in). All
// subsequent trees must have the same number of tips and the same taxon 
// names. The trees are difficult to navigate because they are stored as
// indices to the node above (you can go up but not down). This is very
// compact but inconvenient. The intended use is to expand the tree into
// an METree using 'getTree' before use. Not storing branch lengths
// will save memory considerably. #defining TREESET_FLOATS would reduce
// the memory used by branch lengths at the expense of precision.

class METaxonList;
class METree;
class METreeList;
class MELink;

#define TREESET_FLOATS

#ifdef TREESET_FLOATS
	typedef float LengthReal;
#else
	typedef double LengthReal;
#endif

// =========================================================================
//	class METreeSet
// =========================================================================

class METreeSet : public MEObject {
	public:
		virtual long		getID() const { return 'MTrS'; }
		//: The ID of this object - 'MTrS'.
		
		virtual char*		getClassName() const { return "TreeSet"; }
		//: Returns the name of this class - "Tree Set"
		
		static int			mCapacityIncrement;

	protected:
		METaxonList			*mTaxonList;
		
		long				mNumTrees;
		long				mNumTips;
		long				mNumInternalNodes;
		long				mNumIndices;
		long				mNumBytes;
		
		bool				mStoreLengths;
		
		unsigned char		**mTrees;
		LengthReal			**mLengths;
		
		long 				mCapacity;
			
	public:
		
							METreeSet();
		virtual				~METreeSet();
			
		virtual void		setNumTips(long inNumTips);
		virtual long		getNumTips() { return mNumTips; }
		
		virtual void		setStoreLengths(bool inStoreLengths);
		virtual bool		getStoreLengths() { return mStoreLengths; }

		virtual void		setTaxonList(METaxonList *inTaxonList);
		virtual METaxonList	*getTaxonList() { return mTaxonList; }
		
		virtual long		newTip(char *inName);
		virtual long		newTip(long inTaxonNo);
		
		virtual long		addTree();
		virtual long		addTree(unsigned char *inTree, LengthReal *inLengths, long inNumTips);
		virtual long		addTree(METree *inTree);
		
		virtual void		insertTree(long index, unsigned char *inTree, LengthReal *inLengths, long inNumTips);
		virtual void		insertTree(long index, METree *inTree);
		
		virtual void		removeTree(long index, unsigned char **inTree = NULL, LengthReal **inLengths = NULL);
		virtual void		removeTree(long index, METree *inTree = NULL);
		
		virtual void		removeAll();
		
//		virtual void		resetTreeSet();

		virtual unsigned char	*getTreeArray(long inTreeIndex);
		virtual LengthReal		*getLengthsArray(long inTreeIndex);

		virtual long		getTreeElement(long inTreeIndex, long inNodeIndex);
		virtual void		setTreeElement(long inTreeIndex, long inNodeIndex, long inAncestorIndex);
		virtual double		getBranchLength(long inTreeIndex, long inNodeIndex);
		virtual void		setBranchLength(long inTreeIndex, long inNodeIndex, double inBranchLength);

		virtual METree 		*getTree(long index);
		virtual void		getTree(long index, METree *inTree);
		
	protected:
//		virtual MELink		*getTreeNode(int inTreeIndex, METree *ioTree);

	public:
		virtual void		setTree(long index, unsigned char *inTree, LengthReal *inLengths, long inNumTips);
		virtual void		setTree(long index, METree *inTree);

	protected:
		virtual void		setTreeNode(int inTreeIndex, long *ioCurrent, long inAncestor, METree *inTree, MELink *inNode);

		long				getNodeIndex(unsigned char *inTree, long inIndex);
		void				setNodeIndex(unsigned char *inTree, long inIndex, long inValue);
		
		virtual void		checkCapacity(long inCapacity);

	public:
		virtual void		readObject(MEStream *stream);
		virtual void		writeObject(MEStream *stream);
};

#endif /* _METREESET_H_ */
