/***********************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


	MEVariate.cpp
	
	Basic MEVariate class incorporating simple statistics.

 ***********************************************************************************/

//  includes:
// ANSI C/C++:
#include <iostream.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "MEGlobal.h"
// my stuff:
#include "MEUtilities.h"
#include "MEExceptions.h"
#include "MEMaths.h"
#include "MEVariate.h"


//  variables:
char
	MEVariateMsg[kShortStringLength];


// --------------------------------------------------------------------------------
//    MEVariate::MEVariate
// --------------------------------------------------------------------------------
MEVariate::MEVariate(char *str, bool keepvals, int ss, bool _discr)
{
	_discrete = _discr;
	_have_discretized = false;
	_have_kurtosis = false;
	_have_max = false;
	_have_mean = false;
	_have_median = false;
	_have_min = false;
	_have_mode = false;
	_have_skewness = false;
	_have_sum[kMaxOrder] = false;
	_storingValues = keepvals;
	_useBins = false;

	setName(str);

	binSize = 0.0;
	freq = NULL;
	kurtosis = 0.0;
	maximum = 0.0;
	minimum = 0.0;
	mean = 0.0;
	median = 0.0;
	mode = 0.0;
	rDisMax = 0.0;
	rDisMin = 0.0;
	rVal = NULL;
	skewness = 0.0;
	for (int i = 0; i < kMaxOrder; i++)
		sum[kMaxOrder] = 0.0;
	
	iDisMax = 0.0;
	iDisMin = 0.0;
	iVal = NULL;
	maxVals = ss;
	n = 0;
	numBins = 0;
	sorted = unsorted;

	if (keepvals) {
		if (_discrete)
			NewArray(iVal, int, ss);
		else
			NewArray(rVal, double, ss);
	}
	
	clear();
}


// --------------------------------------------------------------------------------
//    MEVariate::~MEVariate
// --------------------------------------------------------------------------------
MEVariate::~MEVariate()
{
	if (rVal)
		delete [] rVal;
	if (iVal)
		delete [] iVal;
	if (freq)
		delete [] freq;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::accommodate
// --------------------------------------------------------------------------------
void MEVariate::accommodate(int extra)
{
	if (_storingValues) {
		if (n + extra > maxVals) {
			int i, k = maxVals;
			while (k < n+extra)
				k += kVariateResizeChunk;
			if (_discrete) {
				int *tempI;
				NewArray(tempI, int, k);
				for (i = 0; i < maxVals; i++)
					tempI[i] = iVal[i];
				for ( ; i < k; i++) 
					tempI[i] = 0;
				// delete the old array of data pointers:
				delete [] iVal;
				// set data to point to the new list:
				iVal = tempI;
			}
			else {
				double *tempR;
				NewArray(tempR, double, k);
				for (i = 0; i < maxVals; i++)
					tempR[i] = rVal[i];
				for ( ; i < k; i++) 
					tempR[i] = 0.0;
				// delete the old array of data pointers:
				delete [] rVal;
				// set data to point to the new list:
				rVal = tempR;
			}
			maxVals = k;
		}
	}
}


// --------------------------------------------------------------------------------
//    MEVariate::add
// --------------------------------------------------------------------------------
/*
	Adding value x to MEVariate V:
		if storing values then
			if (using bins)
				get b, the bin-number of x, an integer.
				if (b < 0) then have to shift the stored values up
				if (b > maxIndex) then have to extend the stored values
			if V is discrete then
				add 1 to iVal[b]
			else
				make sure there is room 
*/
MEVariate& MEVariate::add(double d)
{
	int b, i;
	if (d != NAN) {
		for (i = 0; i < kMaxOrder; i++)
			sum[i] += pow(d, i+1);
	}
	
	if (_storingDistribution) {
		b = getBinNumber(d);		// e.g. get b = -43;
		if (b < 0) {
			// have to extend the distribution downwards
			extendDistributionDown(-b);		// so the index we need is 100-43 = 57
			b = 0;
		}
		else if (b >= numBins)
			extendDistributionUp(b + 1 - numBins);
		freq[b] += 1.0;
	}
	
	if (_storingValues) {
		accommodate(1);
		if (_discrete)
			iVal[n] = (int) d;
		else
			rVal[n] = d;
	}
	
	if (n == 0 || maximum == NAN)
		maximum = d;
	else
		maximum = max(maximum, d);

	if (n == 0 || minimum == NAN)
		minimum = d;
	else
		minimum = min(minimum, d);

	n++;
	_have_mean = false;
	_have_median = false;
	_have_mode = false;

	return *this;
}



// --------------------------------------------------------------------------------
//    MEVariate::addInRange
// --------------------------------------------------------------------------------
MEVariate& MEVariate::addInRange(double d, double x, double y)
{
	if ((d <= y) && (d >= x))
		add(d);
	return *this;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::clear
// --------------------------------------------------------------------------------
void MEVariate::clear()
{
	int i, j;
	_have_discretized = false;
	_have_kurtosis = false;
	_have_max = false;
	_have_mean = false;
	_have_median = false;
	_have_min = false;
	_have_mode = false;
	_have_skewness = false;
	for (i = 0; i < kMaxOrder; i++)
		_have_sum[i] = false;
		
	kurtosis = 0.0;
	maximum = 0.0;
	mean = 0.0;
	median = 0.0;
	minimum = 0.0;
	mode = 0.0;
	skewness = 0.0;
	n = 0;
	for (i = 0; i < kMaxOrder; i++)
		sum[i] = 0.0;
	if (rVal) {
		for (j = 0; j < maxVals; j++)
			rVal[j] = 0.0;
//		maxVals = 0;
	}
	if (freq)
		for (i = 0; i < numBins; i++)
			freq[i] = 0.0;
	_have_discretized = false;
	
	sorted = unsorted;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::extendDistributionDown
// --------------------------------------------------------------------------------
void MEVariate::extendDistributionDown(int down)
{
	if (down <= 0)
		return;
	double *tempF;
	int i, k = 0;
	// find out how much to add (in chunks):
//	while (k <= down)
//		k += kVariateResizeChunk;
	k = down;
	// make a new array to store the extended distribution:
	NewArray(tempF, double, k + numBins);
	for (i = 0; i < k; i++)
		tempF[i] = 0;
	for ( ; i < k + numBins; i++)
		tempF[i] = freq[i-k];
	
	delete [] freq;
	freq = tempF;
	numBins += k;
/*
	if (_discrete) {
		int i, *tempI, k = 0;
		// find out how much to add (in chunks):
		while (k <= down)
			k += kVariateResizeChunk;
		// make a new array to store the extended distribution:
		NewArray(tempI, int, k + maxVals);
		for (i = 0; i < k; i++)
			tempI[i] = 0;
		for ( ; i < k+maxVals; i++)
			tempI[i] = iVal[i-k];
		
		delete [] iVal;
		iVal = tempI;
	}
	else {
		double *tempD;
		int i, k = 0;
		// find out how much to add (in chunks):
		while (k <= down)
			k += kVariateResizeChunk;
		// make a new array to store the extended distribution:
		NewArray(tempD, double, k + maxVals);
		for (i = 0; i < k; i++)
			tempD[i] = 0.0;
		for ( ; i < k+maxVals; i++)
			tempD[i] = rVal[i-k];
		
		delete [] rVal;
		rVal = tempD;
	}
*/
}


// --------------------------------------------------------------------------------
//    MEVariate::extendDistributionUp
// --------------------------------------------------------------------------------
void MEVariate::extendDistributionUp(int up)
{
	if (up <= 0)
		return;
	double *tempF;
	int i, k = 0;
	// find out how much to add (in chunks):
	while (k <= up)
		k += kVariateResizeChunk;
	// make a new array to store the extended distribution:
	NewArray(tempF, double, k + numBins);
	for (i = 0; i < numBins; i++)
		tempF[i] = freq[i];
	for ( ; i < k + numBins; i++)
		tempF[i] = 0;
	
	delete [] freq;
	freq = tempF;
	numBins += k;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::getBinNumber
// --------------------------------------------------------------------------------
int MEVariate::getBinNumber(double d)
{
	// we take the maximum here so there is no overflow
	// -- any extras will just get thrown into the same first bin.
	if (n <= 0)
		return 0;
	return (int) floor((d - getMin())/binSize);
}
int MEVariate::getBinNumber(int i)
{
	return getBinNumber((double) i);
}


// --------------------------------------------------------------------------------
//    MEVariate::getMax
// --------------------------------------------------------------------------------
double MEVariate::getMax()
{
	if (!_have_max) {
		int i;
		double d;
		i=0;
		do {
			maximum=getVal(i);
			i++;
		} while (i<n && maximum == NAN);
		
		while (i<n) {
			d=getVal(i);
			if (d>maximum)
				maximum=d;
			i++;
		}
		_have_max = true;
	}
	return maximum;
}


// --------------------------------------------------------------------------------
//    MEVariate::getMean
// --------------------------------------------------------------------------------
double MEVariate::getMean()
{
	if (_have_mean)
		return mean;
	mean = getSum(1)/(double)n;
	_have_mean = true;
	return mean;
}


// --------------------------------------------------------------------------------
//    MEVariate::getMedian
// --------------------------------------------------------------------------------
double MEVariate::getMedian()
{
	if (_have_median)
		return median;
	if (_storingValues) {
		sort(ascending);
		_have_median = true;
		if (n % 2 == 1)		// n is odd
			return rVal[(int)n/2];
		else				// n is even
			return (rVal[(int)(n/2-1)] + rVal[(int)(n/2)]) / 2.0;
	}
	throw MEException("MEVariate:: Call for median with no stored values.");
	return 0.0;		// just to please the compiler
}


// --------------------------------------------------------------------------------
//    MEVariate::getMin
// --------------------------------------------------------------------------------
double MEVariate::getMin()
{
/*
	if (!_have_min) {
		if (n <= 0)
			return 0.0;	// or maybe throw an exception.
		double d;
		int i;
		d = getVal(0);
		for (int i = 0; 
	}
*/
	if (!_have_min) {
		int i = 0;
		double d;
		do {
			minimum = getVal(i);
			i++;
		} while (i < n && minimum == NAN);
		
		while (i < n) {
			d = getVal(i);
			if (d < minimum)
				minimum = d;
			i++;
		}
		_have_min = true;
	}
	return minimum;
}


// --------------------------------------------------------------------------------
//    MEVariate::getMode
// --------------------------------------------------------------------------------
double MEVariate::getMode()
{
	if (_have_mode)
		return mode;
	double d;
	int i, k = 0;
	if (_useBins) {
		d = 0.0;	// the frequency of values in the first bin
		for (i = 0; i < numBins; i++) {
			if (getFreq((int) i) > d) {
				k = i;
				d = getFreq((int) i);
			}
		}
		_have_mode = true;
		return rVal[k];
	}
	return 0.0;
	// (not defined if there isn't a frequency distribution)
}


// --------------------------------------------------------------------------------
//    MEVariate::getSampleVariance
// --------------------------------------------------------------------------------
double MEVariate::getSampleVariance()
{
	double EX2 = getSum(2);
	double EX1 = n*sqr(getMean());
	double TEX1 = getSum(1) * getMean();
	double sVar = (EX2 - EX1) / (double) (n-1);
	return ( getSum(2) - getSampleSize()*sqr(getMean()) ) / (double) (n-1);
}


// --------------------------------------------------------------------------------
//    MEVariate::getSum
// --------------------------------------------------------------------------------
double MEVariate::getSum(int order)
{
	if (!_have_sum[order-1]) {
		int i;
		double d;
		sum[order-1]=0.0;
		for (i=0; i<n; i++) {
			d=getVal(i);
			if (d != NAN)
				sum[order-1] += pow(d, order);
		}
		_have_sum[order-1] = true;
	}
	return sum[order-1];
}


// --------------------------------------------------------------------------------
//    MEVariate::getVal
// --------------------------------------------------------------------------------
double MEVariate::getVal(int idx)
{
	static int bin;
	if (_storingValues) {
		if (_discrete)
			return (double) iVal[idx];
		else
			return rVal[idx];
/*		if (_useBins)
			return rDisMin + (double) (idx + 0.5) * binSize;
		else
			return rVal[idx];
*/	}
	return 0.0;
}


// --------------------------------------------------------------------------------
//    MEVariate::getVariance
// --------------------------------------------------------------------------------
double MEVariate::getVariance()
{
	return ( getSum(2) - n * sqr(getMean()) ) / (double) (n - 1);
}


// --------------------------------------------------------------------------------
//    MEVariate::getPercentile
// --------------------------------------------------------------------------------
double MEVariate::getPercentile(double inPercentile)
{
// modified by AER to interpolate between percentiles.
// ugly and verbose for debugging purposes...
	sort(ascending);
	
	int numDivs = getN() - 1;
	double division = 100.0 / numDivs;
	int below = (int)floor(inPercentile / division);
	int above = below + 1;

	double belowPer = below * division;
	double abovePer = above * division;

	double belowVal = getVal(below);
	double aboveVal = getVal(above);

	double fraction = (inPercentile - belowPer) / division;

	return belowVal + ((aboveVal - belowVal) * fraction);
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::multipleAdd
// --------------------------------------------------------------------------------
MEVariate& MEVariate::multipleAdd(double d, int num)
{
	for (int j = 0; j < num; j++)
		add(d);
	return *this;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::operator=
// --------------------------------------------------------------------------------
MEVariate& MEVariate::operator=(const MEVariate &v)
{
	_have_kurtosis = v._have_kurtosis;
	_have_max = v._have_max;
	_have_mean = v._have_mean;
	_have_median = v._have_median;
	_have_min = v._have_min;
	_have_mode = v._have_mode;
	_have_skewness = v._have_skewness;
	int i;
	for (i = 0; i < kMaxOrder; i++) {
		_have_sum[i] = v._have_sum[i];
		sum[i] = v.sum[i];
	}
	_storingValues = v._storingValues;
	
	strcpy(name, v.name);
	
	kurtosis = v.kurtosis;
	maximum = v.maximum;
	minimum = v.minimum;
	mean = v.mean;
	median = v.median;
	mode = v.mode;
	skewness = v.skewness;
	resizeVals(v.maxVals);
	for (i = 0; i < n; i++)
		rVal[i] = v.rVal[i];
	
	maxVals = v.maxVals;
	n = v.n;
		
	sorted = v.sorted;
	return *this;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::putInBins
// --------------------------------------------------------------------------------
void MEVariate::putInBins()
{
	// Must set the rDisMin and rDisMax first with setBinRange(rDisMin, rDisMax) !
	numBins = min(kMaxNumBins, (int) ((getMax() - getMin()) / binSize));
	if (freq)
		delete [] freq;
	NewArray(freq, double, numBins);
	double x;
	int i, k;
	for (i = 0; i < numBins; i++)
		freq[i] = 0.0;
	if (_discrete) {
		for (i = 0; i < n; i++) {
			k = getBinNumber(iVal[i]);
			if (k >= numBins)
				k = numBins-1;
			freq[k] += 1.0;
		}
	}
	else {
		for (i = 0; i < n; i++) {
			k = getBinNumber(rVal[i]);
			if (k >= numBins)
				k = numBins-1;
			// we take the minimum here, so there is no overflow:
			// any extras get thrown into the same last bin:
			x = min((double) (k + 0.5) * binSize, rDisMax-0.5*binSize);
			freq[k] += 1.0;
		}
	}
	for (i = 0; i < numBins; i++)
		freq[i] /= (double) n;
	_have_discretized = true;
	_useBins = true;
}


// --------------------------------------------------------------------------------
//    MEVariate::putRange
// --------------------------------------------------------------------------------
void MEVariate::putRange(double &x, double &y)
{
	x = getMin();
	y = getMax();
}


#pragma mark -


// ---------------------------------------------------------------------------------
//		 MEVariate::readData
// ---------------------------------------------------------------------------------
// Read the data of this object (i.e. the data of the variate).
void MEVariate::readData(MEStream */*stream*/)
{
}


// ---------------------------------------------------------------------------------
//		 MEVariate::readObject
// ---------------------------------------------------------------------------------
// Read the fields of this object.
void MEVariate::readObject(MEStream *stream)
{
	MEObject::readObject(stream);
}


// --------------------------------------------------------------------------------
//    MEVariate::resizeVals
// --------------------------------------------------------------------------------
void MEVariate::resizeVals(int newsize)
{
	if (!_storingValues)
		return;
	if (newsize <= maxVals)
		return;
	if (newsize > maxVals) {
		double *D;
		NewArray(D, double, newsize);
		for (int i = 0; i < maxVals; i++)
			D[i] = rVal[i];
		delete [] rVal;
		rVal = D;
		
		maxVals = newsize;
	}
}


/*
void MEVariate::resizeDistribution(int minIdx, int maxIdx)
{
	
}
*/

#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::setBins
// --------------------------------------------------------------------------------
void MEVariate::setBins(int i, double x, double y)
{
	numBins = i;
	rDisMin = min(x, y);
	rDisMax = max(x, y);
	binSize = (rDisMax-rDisMin)/(double)numBins;
}


// --------------------------------------------------------------------------------
//    MEVariate::setStoreDistribution
// --------------------------------------------------------------------------------
void MEVariate::setStoreDistribution(bool b, double d)
{
	_storingDistribution = b;
	if (b) {
		if (binSize <= 0.0)
			binSize = d;
	}
}


// --------------------------------------------------------------------------------
//    MEVariate::setStoreValues
// --------------------------------------------------------------------------------
void MEVariate::setStoreValues(bool b)
{
	_storingValues = b;
	if (b) {
		if (_discrete) {
			if (iVal == NULL)
				NewArray(iVal, int, maxVals);
		}
		else {
			if (rVal == NULL)
				NewArray(rVal, double, maxVals);
		}
	}
}


/*
// --------------------------------------------------------------------------------
//    MEVariate::setStoreValues
// --------------------------------------------------------------------------------
void MEVariate::setStoreValues(bool b)
{
	_storingValues = b;
	if (b) {
		if (rVal == NULL)
			NewArray(rVal, double, kVariateResizeChunk);
	}
}
*/

// --------------------------------------------------------------------------------
//    MEVariate::setVal
// --------------------------------------------------------------------------------
void MEVariate::setVal(int idx, double d)
{
	if (_storingValues) {
		if (idx >= maxVals)
			resizeVals(idx+1);
		if (_discrete)
			iVal[idx] = (int) d;
		else
			rVal[idx] = d;
	}
	
	_have_discretized = false;
	_have_kurtosis = false;
	_have_max = false;
	_have_mean = false;
	_have_median = false;
	_have_min = false;
	_have_mode = false;
	_have_skewness = false;
	for (int i = 0; i < kMaxOrder; i++)
		_have_sum[i] = false;
	if (idx >= n)
		n = idx + 1;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::sort
// --------------------------------------------------------------------------------
void MEVariate::sort(SortOrder d)
{
	if (!_storingValues)
		return;
	if (sorted == d)
		return;
	int gap, i, j, k;
	if (d == ascending) {
		for (gap = n/2; gap > 0; gap /= 2) {
			for (i = gap; i < n; i++) {
				for (j = i-gap; (j >= 0) && (rVal[j] > rVal[j+gap]); j -= gap) {
					k = j + gap;
					swapVals(j, k);
				}
			}
		}
	}
	else {	// sort descending
		for (gap = n/2; gap > 0; gap /= 2) {
			for (i = gap; i < n; i++) {
				for (j = i-gap; (j >= 0) && (rVal[j] < rVal[j+gap]); j -= gap) {
					k = j + gap;
					swapVals(j, k);
				}
			}
		}
	}
	sorted = d;
}


// --------------------------------------------------------------------------------
//    MEVariate::strmDistribution
// --------------------------------------------------------------------------------
void MEVariate::strmDistribution(ostream &os)
{
	if (!_storingValues) {
		os << "No distribution present.\n";
		return;
	}
	double
		d = getMin(), 
		e = getMax();
	int i, k;
	k = (int) (e - d)/binSize + 1;
	os << "Distribution of " << getName() << ":\n";
	if (_discrete) {
		do {
			os << d << "\t" << freq[getBinNumber(d)] << "\n";
			d += binSize;
		} while (d <= e);
	}
	else {
		os << d << " -\t" << d + binSize << "\t" << freq[getBinNumber(d)] << "\n";
		d += 2.0*binSize;
		for (i = getBinNumber(d); i < k; i++) {
			os << "\t" << d << "\t" << freq[i] << "\n";
			d += binSize;
		}
	}
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::swapVals
// --------------------------------------------------------------------------------
void MEVariate::swapVals(int i, int j)
{
	if (!_storingValues)
		return;
	double d = rVal[i];
	rVal[i] = rVal[j];
	rVal[j] = d;
}


#pragma mark -


// --------------------------------------------------------------------------------
//    MEVariate::write
// --------------------------------------------------------------------------------
char* MEVariate::write(char *str)
{
	sprintf(MEVariateMsg, "[MEVariate] %s:\n\tn = %d\n\trange = %g - %g\n\tmean = %g\n"
		"\tvar = %g\n", getName(), n, getMin(), getMax(), getMean(), getVariance());
	if (_storingValues) {
		char tempStr[128];
//		sprintf(tempStr, "\tmedian = %g\n", getMedian());
//		strcat(MEVariateMsg, tempStr);
		if (_have_discretized) {
			sprintf(tempStr, "\tmode = %g\n", getMode());
			strcat(MEVariateMsg, tempStr);
		}
	}
	if (str)
		strcpy(str, MEVariateMsg);
	return MEVariateMsg;
}


// ---------------------------------------------------------------------------------
//		 MEVariate::writeData
// ---------------------------------------------------------------------------------
// Write the data of this object (i.e. the data of the variate).
void MEVariate::writeData(MEStream */*stream*/)
{
}


// ---------------------------------------------------------------------------------
//		 MEVariate::writeObject
// ---------------------------------------------------------------------------------
// Write the fields of this object.
void MEVariate::writeObject(MEStream *stream)
{
	MEObject::writeObject(stream);
}

