/***********************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	MEVariate.h
	
	Basic variate class incorporating simple statistics.

 ***********************************************************************************/

#ifndef _MEVARIATE_H_
#define _MEVARIATE_H_

#include <math.h>

#include "MEGlobal.h"
#include "MEObject.h"
#include "MEUtilities.h"

class MEStream;

//  defines:
#define kMaxOrder	4
#define kVariateNameLength	kShortStringLength
#define kVariateResizeChunk	25
#define kMaxNumBins	1000

// class definition:
class MEVariate : public MEObject {

public:
// methods:
	virtual long			getID() const { return 'MVar'; }
	virtual char*			getClassName() const { return "MEVariate"; }
		
	MEVariate(char* name = "unnamed MEVariate",
		bool keepvals = false, 
		int ss = 100, 
		bool _discr = false);
	~MEVariate();
	
	virtual void			accommodate(int extra);
	
	virtual MEVariate&		add(double d);
	virtual MEVariate&		add(int i) { return add((double) i); }
	virtual MEVariate&		addInRange(double d, double x, double y);
	
	virtual void			clear();
	
	virtual void			extendDistributionDown(int up);
	virtual void			extendDistributionUp(int up);
		
	virtual int				getBinNumber(double d);
	virtual int				getBinNumber(int i);
	
	virtual double			getCoefficientOfVariation() { return getSSD() * 100.0 / getMean(); }
	virtual double			getCoV() { return getCoefficientOfVariation(); }
	
	virtual double			getFreq(double d) { return freq[getBinNumber(d)]; }
	virtual double			getFreq(int i) { return freq[i]; }
	
	virtual double			getMax();
	virtual double			getMean();
	virtual double			getMedian();
	virtual double			getMin();
	virtual double			getMode();
	
	virtual int				getN() { return n; }
	virtual char*			getName() { return name; }
	
	virtual double			getRange() { return getMax() - getMin(); }
	
	virtual double			getPercentile(double inPercentile);
	
	virtual int				getSampleSize() { return n; }
	virtual double			getSampleStandardDeviation() { return getSSD(); }
	virtual double			getSampleVariance();
	virtual double			getSD() { return sqrt(getVariance()); }
	virtual double			getSSD() { return sqrt(getSampleVariance()); }
	virtual double			getStandardDeviation() { return getSD(); }
	virtual double			getStandardError() { return getSD() / sqrt((double)getSampleSize()); }
	virtual double			getSE() { return getStandardError(); }

	virtual double			getSum(int order = 1);
	virtual double			getSumOfSquares() { return getSum(2); }
	
	virtual double			getVal(int idx);
	virtual void			setVal(int idx, double d);
	
	virtual double			getVariance();
	
	virtual MEVariate&		multipleAdd(double d, int num);

	virtual MEVariate&		operator+=(double d) { return add(d); }
	virtual MEVariate&		operator=(const MEVariate &v);
	friend ostream&			operator<<(ostream &o, MEVariate &v) { return o << v.write(); }
	
	virtual void			putInBins();
	virtual void			putRange(double &x, double &y);
	
//	virtual void			resizeDistribution(int minIdx, int maxIdx);
	virtual void			resizeVals(int newsize);

	virtual void			setBins(int n, double x, double y);
	virtual void			setBinSize(double d) { binSize = d; }
	virtual void			setContinuous(bool b = true) { _discrete = !b; }
	virtual void			setDiscrete(bool b = true) { _discrete = b; }
	virtual void			setBinSize(int i) { setBinSize((double) i); }
	virtual void			setStoreDistribution(bool b = true, double d = 1.0);
	virtual void			setStoreValues(bool b = true); //{ _storingValues = b; }
	
	virtual void			sort(SortOrder = ascending);
	
	virtual void			strm(ostream &os) { os << write(); }
	virtual void			strmDistribution(ostream &os);
	
	virtual void			swapVals(int i, int j);
	
	virtual void			useBins(bool b) { _useBins = b; }
	
	virtual char*			write(char *str = NULL);

	virtual void			readObject(MEStream *stream);
	virtual void			readData(MEStream *stream);
	virtual void			writeObject(MEStream *stream);
	virtual void			writeData(MEStream *stream);

protected:
	// variables:
	bool
		_discrete,				// ? variate is integer : variate is real
		_have_discretized,		// ? have put values into bins : have not
		_have_kurtosis,			// ? have the fourth moment : not
		_have_max,				// ? have the maximum value : not
		_have_mean,				// ? have the mean (1st moment) : not
		_have_median,			// ? have the median value : not
		_have_min,				// ? have the minimum value : not
		_have_mode,				// ? have the modal value : not
		_have_skewness,			// ? have the third moment : not
		_have_sum[kMaxOrder],	// ? have \sum{x^i} : not
		_storingDistribution,		// ? storing the distribution as we go : not
		_storingValues,			// ? storing the values as we go : not
		_useBins;				// ? putting the values into bins as we go
	
	double
		binSize,				// each bin takes values (b, b+binSize)
		*freq,					// distribution of relative frequencies in distribution of
								// real numbers; null if _discrete is true.
		kurtosis,				// value of kurtosis
		maximum,				// value of max.
		minimum,				// value of min.
		mean,					// value of mean
		median,					// value of median (middle value in distribution)
		mode,					// value of mode (most frequent value in distribution)
		rDisMax,				// max. permitted real value of discretized distribution
		rDisMin,				// min. permitted real value of discretized distribution
		*rVal,					// if !_discrete, list of all the values stored;
								// else they are in *iVal
		skewness,				// value of third moment
		sum[kMaxOrder];			// accumulated sums of powers \sum{x^i}
	
	int
		iDisMax,				// max. permitted integer value of discretized distribution
		iDisMin,				// min. permitted integer value of discretized distribution
		*iVal,					// if _discrete, the integral values stored; 
								// else they are in *rVal
		maxVals,				// max. permitted number of values to be stored
		n,						// number of values in sample
		numBins;				// number of bins used in discretization
		
	SortOrder
		sorted;					// ascending, descending, or unsorted
	
};

#endif