/******************************************************************************

Molecular Evolution C++ Class Library (MolEvolLib) v1.0
Copyright (C) 1997-2002 Andrew Rambaut and Michael Charleston, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of MolEvolLib C++ Class Library.

    MolEvolLib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolEvolLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 MEXMLCreator.cpp

				MEXMLCreator Class

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "MEExceptions.h"

#include "CWriter.h"
#include "MECreator.h"
#include "MEXMLCreator.h"
#include "MEBasicSequence.h"
#include "MEBasicAlignment.h"
#include "MELink.h"
#include "METreeList.h"
#include "METree.h"
#include "MEPhylogenyList.h"
#include "MEPhylogeny.h"
#include "MEDistanceMatrix.h"
#include "METaxon.h"
#include "METaxonList.h"
#include "MEMatrix.h"
#include "MEDistanceMatrix.h"

/******************************************************************************
 MEXMLCreator
 
 ******************************************************************************/

MEXMLCreator::MEXMLCreator() : MECreator()
{
	hasHeader=false;
	useTaxaElement=false;
	hasTaxaElement = false;
	indent = 0;
}


MEXMLCreator::MEXMLCreator(CWriter *inWriter) : MECreator(inWriter)
{
	hasHeader=false;
	useTaxaElement=false;
	hasTaxaElement = false;
	indent = 0;
}

MEXMLCreator::~MEXMLCreator()
{
}

void MEXMLCreator::writeIndent()
{
	for (int i = 0; i < indent; i++)
		writer->writeChar('\t');
}

void MEXMLCreator::writeOpenElement(MEObject *inObject, char *inTag, char *inObjectType)
{
	writeIndent();
	writer->writeChar('<');
	writer->writeString(inTag);
	if (inObject->getName() != NULL) {
		writer->writeString(" id=\"");
		writer->writeString(inObject->getName());
		if (inObjectType != NULL) {
			writer->writeString("_");
			writer->writeString(inObjectType);
		}
		writer->writeString("\"");
	}
	writer->writeString(">\r");
	 
	indent++;
}

void MEXMLCreator::writeCloseElement(char *inTag)
{
	indent--;
	writeIndent();
	writer->writeChar('<');
	writer->writeChar('/');
	writer->writeString(inTag);
	writer->writeString(">\r");
}

void MEXMLCreator::writeTaxonElement(METaxon *inTaxon)
{
	writeIndent();
	writer->writeString("\t<taxon id=\"");
	writer->writeString(inTaxon->getName());
	writer->writeString("\"/>\r");
}

void MEXMLCreator::writeTaxaElement(METaxonList *inTaxonList)
{
	int i, n = inTaxonList->getNumTaxa();
	
	writeOpenElement(inTaxonList, "taxa", "taxa");
		
	char *name;
	
	for (i=0; i<n; i++) {
		writeTaxonElement(inTaxonList->getTaxon(i));
	}
	
	writeCloseElement("taxa");
	
	hasTaxaElement = true;
}

void MEXMLCreator::writeSequenceElement(MEBasicSequence *inSequence, char *inSeqString, int length)
{
	writeIndent();
	writer->writeString("<sequence>\r");
	indent++;
	if (useTaxaElement) {
		writeIndent();
		writer->writeString("<taxon idref=\"");
		writer->writeString(inSequence->getName());
		writer->writeString("\"/>\r");
	} else
		writeTaxonElement(inSequence);

	writeIndent();
	writeSequence(inSeqString, length, -1, 0, -1);
	writer->writeChar('\r');
	indent--;
	writeIndent();
	writer->writeString("</sequence>\r");
}

void MEXMLCreator::writeAlignmentElement(MEBasicAlignment *inAlignment)
{
	int i, n = inAlignment->getNumSequences();
	
	writeOpenElement(inAlignment, "alignment", NULL);
		
	char *name;
	int length = inAlignment->getNumSites();
	char *sequence = new char[length + 1];
	
	for (i=0; i<n; i++) {
		inAlignment->getSequenceString(i, sequence, 0, length);
		writeSequenceElement(inAlignment->getSequence(i), sequence, length);
	}
	
	writeCloseElement("alignment");
}

void MEXMLCreator::writeNodeElement(METree *inTree, MELink *inNode, double inHeight, double inLength)
{
	double height = inHeight - inLength;
	
	writeIndent();
	writer->writeString("<node height=\"");
	writer->writeDouble(height);
	writer->writeString("\">\r");
	indent++;
	
	if (inNode != NULL && inNode->getIsTip()) {
	
		METaxonList *taxonList = inTree->getTaxonList();
		if (useTaxaElement || taxonList == NULL) {
			writeIndent();
			writer->writeString("<taxon idref=\"");
			writer->writeString(inTree->getTipName(inNode));
			writer->writeString("\"/>\r");
		} else {
			writeTaxonElement(taxonList->getTaxon(inNode->getTaxonNo()));
		}
		
	} else {
		
		MELink *branch = inTree->getFirstBranch(inNode);
		MELink *node;
		double length;
		while (branch != NULL) {
			node = inTree->getDescendantNode(branch, &length);
			writeNodeElement(inTree, node, height, length);
			branch = inTree->getNextBranch(inNode, branch);
		}
	}

	indent--;
	writeIndent();
	writer->writeString("</node>\r");
}

void MEXMLCreator::writeTreeElement(METree *inTree)
{
	writeOpenElement(inTree, "tree", NULL);
	
	double height, maxHeight = 0.0;
	int i, n = inTree->getNumTips();
	for (int i = 0; i < n; i++) {
		height = inTree->getRootToTipDistance(i);
		if (height > maxHeight)
			maxHeight = height;
	}
	
	writeNodeElement(inTree, NULL, maxHeight, 0.0);

	writeCloseElement("tree");
}

void MEXMLCreator::writeTreesElement(METreeList *inTreeList)
{
	int i, n = inTreeList->getNumTrees();
	METree *tree;
	char *name;
	
	writeOpenElement(inTreeList, "trees", NULL);
	
	for (i=0; i<n; i++) {
		tree=inTreeList->getTree(i);
		
		writeTreeElement(tree);
	}
	
	writeCloseElement("trees");
}

void MEXMLCreator::writeLineageElement(MEPhylogeny *inPhylogeny, METaxonList *inTaxonList, 
								long inLineage, long &ioTip)
{
	writeIndent();
	writer->writeString("<node height=\"");
	indent++;
		 
	long descendant = inPhylogeny->getDescendant(inLineage);
	
	if (descendant == -1) {
	
		writer->writeDouble(0.0);
		writer->writeString("\">\r");
		
		writeIndent();
		writer->writeString("<taxon idref=\"");
		writer->writeString(inTaxonList->getTaxonName(ioTip));
		writer->writeString("\"/>\r");
		
		ioTip++;
		
	} else {
	
		double height = inPhylogeny->getBirthTime(descendant);

		if (inLineage == 0)
			writer->writeDouble(inPhylogeny->getHeight());
		else
			writer->writeDouble(inPhylogeny->getHeight(inLineage));
		writer->writeString("\">\r");
		
		long sibling = inPhylogeny->getSibling(descendant);
		
		while (sibling != descendant) {
			writeLineageElement(inPhylogeny, inTaxonList, sibling, ioTip);			
			sibling = inPhylogeny->getSibling(sibling);
		}
	}

	indent--;
	writeIndent();
	writer->writeString("</node>\r");
}

void MEXMLCreator::writeTreeElement(MEPhylogeny *inPhylogeny, METaxonList *inTaxonList)
{
	writeOpenElement(inPhylogeny, "tree", NULL);
	
	long tip = 0;
	writeLineageElement(inPhylogeny, inTaxonList, 0, tip);

	writeCloseElement("tree");
}

void MEXMLCreator::writeTreesElement(MEPhylogenyList *inPhylogenyList, METaxonList *inTaxonList)
{
	int i, n = inPhylogenyList->getNumPhylogenies();
	MEPhylogeny *phylogeny;
	char *name;
	
	writeOpenElement(inPhylogenyList, "trees", NULL);
	
	for (i=0; i<n; i++) {
		phylogeny=inPhylogenyList->getPhylogeny(i);
		
		writeTreeElement(phylogeny, inTaxonList);
	}
	
	writeCloseElement("trees");
}

void MEXMLCreator::openHeader()
{
	if (!hasHeader) {
		writer->writeString("<?xml version=\"1.0\" standalone=\"yes\"?>\r\r");
		hasHeader=true;
	}
}

void MEXMLCreator::createAlignment(MEBasicAlignment *inAlignment)
{
	ThrowIfNull(inAlignment);
	
	openHeader();
	
	if (useTaxaElement) {
		writeTaxaElement(inAlignment);
		writeAlignmentElement(inAlignment);
	} else 
		writeAlignmentElement(inAlignment);
}

//: This function does not actually close the trees block so subsequent
//  calls to "createTree" or "createTreeList" will add the trees to the
//  same block. Creating a different block will close the trees block.
void MEXMLCreator::createTree(METree *inTree)
{
	ThrowIfNull(inTree);
	
	openHeader();
	
	if (useTaxaElement && !hasTaxaElement) {
		METaxonList *taxonList = inTree->getTaxonList();
		if (taxonList != NULL) 
			writeTaxaElement(taxonList);
	}

	writeTreeElement(inTree);
}

//: This function does not actually close the trees block so subsequent
//  calls to "createTree" or "createTreeList" will add the trees to the
//  same block. Creating a different block will close the trees block.
void MEXMLCreator::createTreeList(METreeList *inTreeList)
{
	ThrowIfNull(inTreeList);
	ThrowIf(inTreeList->getNumTrees() < 1);
	
	openHeader();
	
	if (useTaxaElement && !hasTaxaElement) {
		METaxonList *taxonList = inTreeList->getTaxonList();
		if (taxonList != NULL) 
			writeTaxaElement(taxonList);
	}

	writeTreesElement(inTreeList);
}

void MEXMLCreator::finishCreation()
{
}

