/******************************************************************************
 PhyloGenEngine.cpp

				PhyloGenEngine Class
	
	Copyright  2002 Andrew Rambaut, University of Oxford. All rights reserved.

 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <math.h>

#include <fstream>
#include <iostream>

#include "PhyloGenEngine.h"
#include "CProgress.h"

#include "CDataSeries.h"
#include "CDoubleArray.h"
#include "CVariate.h"
#include "CGraph.h"
#include "CLinePlot.h"
#include "CScatterPlot.h"
#include "CLogAxis.h"
#include "CAxis.h"

#include "CReportRecord.h"
#include "CGraphicsRecord.h"
#include "CTextWriter.h"
#include "CSVGReporter.h"

#include "MEPhylogeny.h"
#include "MEPhylogenyList.h"

#include "MEBirthDeathSimulator.h"
#include "MEEpisodicSimulator.h"

#ifdef SPATIAL_MODEL
#include "MESpatialSimulator.h"
#endif

#include "MENexusCreator.h"
#include "MEPhylipCreator.h"

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine
// ---------------------------------------------------------------------------------

PhyloGenEngine::PhyloGenEngine()
{
	mModelType = kUnknownModel;
	mSimulator = NULL;
	
	mWholePhylogenyList = new MEPhylogenyList();
	mReconstructedPhylogenyList = new MEPhylogenyList();
	mSampledPhylogenyList = new MEPhylogenyList();

	mProgress = NULL;
}

// ---------------------------------------------------------------------------------
//		 ~PhyloGenEngine
// ---------------------------------------------------------------------------------

PhyloGenEngine::~PhyloGenEngine()
{
	if (mSimulator != NULL)
		delete mSimulator;		

	if (mWholePhylogenyList != NULL)
		delete mWholePhylogenyList;		

	if (mReconstructedPhylogenyList != NULL)
		delete mReconstructedPhylogenyList;		

	if (mSampledPhylogenyList != NULL)
		delete mSampledPhylogenyList;		

	if (mProgress != NULL)
		delete mProgress;		
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setBirthDeath
// ---------------------------------------------------------------------------------

void PhyloGenEngine::setBirthDeath(BirthDeathRecord &inBirthDeathRecord)
{
	MEBirthDeathSimulator *bSim;

	if (mModelType != kBirthDeathModel) {
		if (mSimulator != NULL) {
			delete mSimulator;
			mSimulator = NULL;
		}		
		
		bSim = new MEBirthDeathSimulator();
		mSimulator = bSim;
		
		mModelType = kBirthDeathModel;
	} else {
		bSim = dynamic_cast<MEBirthDeathSimulator *>(mSimulator);
	}
	
	bSim->setBirthDeathRates(inBirthDeathRecord.mBirthRate, inBirthDeathRecord.mDeathRate);
	bSim->setExtantStop(inBirthDeathRecord.mExtantStop);
	bSim->setTimeStop(inBirthDeathRecord.mTimeStop);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setBirthDeath
// ---------------------------------------------------------------------------------

void PhyloGenEngine::setEpisodic(EpisodicRecord &inEpisodicRecord)
{
	MEEpisodicSimulator *eSim;

	if (mModelType != kEpisodicModel) {
		if (mSimulator != NULL) {
			delete mSimulator;
			mSimulator = NULL;
		}		
		
		eSim = new MEEpisodicSimulator();
		mSimulator = eSim;
		
		mModelType = kEpisodicModel;
	} else {
		eSim = dynamic_cast<MEEpisodicSimulator *>(mSimulator);
	}
	
	eSim->clearEpisodes();
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setAddGrowth
// ---------------------------------------------------------------------------------

void PhyloGenEngine::setAddGrowth(AddGrowthRecord &inAddGrowthRecord)
{
	MEEpisodicSimulator *eSim;

	if (mSimulator == NULL) {
		EpisodicRecord rec;
		setEpisodic(rec);
	}
		
	if (mModelType != kEpisodicModel) {
		throw CException("Episodic model must be selected before an episode can be added");
	} else {
		eSim = dynamic_cast<MEEpisodicSimulator *>(mSimulator);
	}

	MEGrowthEpisode *episode = new MEGrowthEpisode();

	*episode = static_cast<MEGrowthEpisode>(inAddGrowthRecord);
	
	eSim->addEpisode(episode);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setAddConstant
// ---------------------------------------------------------------------------------

void PhyloGenEngine::setAddConstant(AddConstantRecord &inAddConstantRecord)
{
	MEEpisodicSimulator *eSim;

	if (mModelType != kEpisodicModel) {
		throw CException("Episodic model must be selected before an episode can be added");
	} else {
		eSim = dynamic_cast<MEEpisodicSimulator *>(mSimulator);
	}
	
	if (eSim->getNumEpisodes() < 1)
		throw CException("A growth episode must be the first episode in the simulation");

	MEConstantEpisode *episode = new MEConstantEpisode();

	*episode = static_cast<MEConstantEpisode>(inAddConstantRecord);
	
	eSim->addEpisode(episode);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setAddExtinction
// ---------------------------------------------------------------------------------

void PhyloGenEngine::setAddExtinction(AddExtinctionRecord &inAddExtinctionRecord)
{
	MEEpisodicSimulator *eSim;

	if (mModelType != kEpisodicModel) {
		throw CException("Episodic model must be selected before an episode can be added");
	} else {
		eSim = dynamic_cast<MEEpisodicSimulator *>(mSimulator);
	}

	if (eSim->getNumEpisodes() < 1)
			throw CException("A growth episode must be the first episode in the simulation");

	MEExtinctionEpisode *episode = new MEExtinctionEpisode();

	*episode = static_cast<MEExtinctionEpisode>(inAddExtinctionRecord);
	
	eSim->addEpisode(episode);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doClearEpisodes
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doClearEpisodes()
{
	MEEpisodicSimulator *eSim;

	if (mModelType != kEpisodicModel) {
		throw CException("Episodic model has not been selected");
	} else {
		eSim = dynamic_cast<MEEpisodicSimulator *>(mSimulator);
	}

	eSim->clearEpisodes();
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setSpatial
// ---------------------------------------------------------------------------------

void PhyloGenEngine::setSpatial(SpatialRecord &inSpatialRecord)
{
#ifdef SPATIAL_MODEL
	MESpatialSimulator *sSim;
	
	if (mModelType != kSpatialModel) {
		if (mSimulator != NULL)
			delete mSimulator;		
		
		sSim = new MESpatialSimulator();
		mSimulator = sSim;

		mModelType = kSpatialModel;
	} else {
		sSim = dynamic_cast<MESpatialSimulator *>(mSimulator);
	}
	
	sSim->setBirthDeathRates(inSpatialRecord.mBirthRate, inSpatialRecord.mDeathRate);
	sSim->setNumAdjacencies(inSpatialRecord.mAdjacencies);
	sSim->setExtantStop(inSpatialRecord.mExtantStop);
	sSim->setTimeStop(inSpatialRecord.mTimeStop);
#endif
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doGenerate
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doGenerate(GenerateRecord inGenerateRecord, CReportRecord *ioReport)
{
	if (mSimulator == NULL) {
		BirthDeathRecord rec;
		setBirthDeath(rec);
	}
		
	MEPhylogeny *phylogeny;
	
	mWholePhylogenyList->removeAll();
	mReconstructedPhylogenyList->removeAll();
	
	if (inGenerateRecord.mReplicates == 1) {
		mSimulator->setProgress(mProgress);
	} else {
		mSimulator->setProgress(NULL);
		if (mProgress != NULL) {
			mProgress->reset();
			mProgress->description("Generating phylogenies...");
			mProgress->progress(0, inGenerateRecord.mReplicates);
			
			mProgress->setSkip(inGenerateRecord.mReplicates / 100);
		}
	}
		
	int i = 0;
	int count = 0;
	
	int maxAttempts = inGenerateRecord.mAttempts;
	
	if (maxAttempts == 0)
		maxAttempts = inGenerateRecord.mReplicates * 1000;
	
	while (i < inGenerateRecord.mReplicates && count < maxAttempts) {
		try {
			mSimulator->simulate();

			if (!inGenerateRecord.mReconstruct || inGenerateRecord.mKeep) {
				phylogeny = mSimulator->getWholePhylogeny();
				mWholePhylogenyList->add(phylogeny);
			}
			
			if (inGenerateRecord.mReconstruct) {
				phylogeny = mSimulator->getSamplePhylogeny(mSimulator->getNumExtantLineages());	
				mReconstructedPhylogenyList->add(phylogeny);
			}
			
			i++;
			
			if (inGenerateRecord.mReplicates > 1 && mProgress != NULL)
				mProgress->progress(i, inGenerateRecord.mReplicates);
				
		} catch (MEAllExtinctTreeSimError e) {
		}
		
		count++;
	}

	if (i == inGenerateRecord.mReplicates) {
		CParagraphElement* para = new CParagraphElement(kNormalText, kAlignElementDefault, 0);
		para->add("Simulation complete: ");
		para->add(inGenerateRecord.mReplicates);
		para->add(" phylogenies generated.");
		if (count > inGenerateRecord.mReplicates) {
			para->addLineBreak();
			para->add("Attempts: ");
			para->add(count);
			para->add(" with ");
			para->add(count - inGenerateRecord.mReplicates);
			para->add(" phylogenies going extinct.");
		}
		ioReport->add(para);
	} else {
		CParagraphElement* para = new CParagraphElement(kNormalText, kAlignElementDefault, 0);
		para->add("Simulation failed: All phylogenies went extinct in ");
		para->add(count);
		para->add("attempts.");
		ioReport->add(para);
	}
	
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doReconstruct
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doReconstruct(ReconstructRecord &inReconstructRecord, CReportRecord *ioReport)
{
	MEPhylogeny *phylogeny;
	
	if (mWholePhylogenyList->getNumPhylogenies() == 0)
		throw CException("No whole phylogenies to reconstuct. Use GENERATE to create phylogenies.");		

	int n = mWholePhylogenyList->getNumPhylogenies();
	
	mReconstructedPhylogenyList->removeAll();
	
	if (mProgress != NULL) {
		mProgress->reset();
		mProgress->description("Reconstructing phylogenies...");
		mProgress->progress(0, n);
		mProgress->setSkip(n / 100);
	}
		
	int i, j;
	int count = 0;

	for (i = 0; i < n; i++) {
		mSimulator->setPhylogeny(mWholePhylogenyList->getPhylogeny(i));
			
		if (mProgress != NULL)
			mProgress->progress(i, n);

		phylogeny = mSimulator->getSamplePhylogeny(mSimulator->getNumExtantLineages());	
		mReconstructedPhylogenyList->add(phylogeny);
	}

	if (mProgress != NULL)
		mProgress->complete();

	CParagraphElement* para = new CParagraphElement(kNormalText, kAlignElementDefault, 0);
	
	if (n == 1) {
		para->add("Reconstruction complete: 1 phylogeny reconstructed.");
	} else {
		para->add("Reconstruction complete: ");
		para->add(n);
		para->add(" phylogenies reconstructed.");
	}
	
	if (!inReconstructRecord.mKeep) {
		mWholePhylogenyList->removeAll();
		para->add(" Source phylogenies have been cleared from memory. ");
	}
	
	ioReport->add(para);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::setSample
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doSample(SampleRecord &inSampleRecord, CReportRecord *ioReport)
{
	MEPhylogeny *phylogeny;
	
	MEPhylogenyList *list;
	if (inSampleRecord.mSource == SampleRecord::kWholeSource) {
		if (mWholePhylogenyList->getNumPhylogenies() == 0)
			throw CException("No whole phylogenies to sample. Use GENERATE to create phylogenies.");
		list = mWholePhylogenyList;
	} else if (inSampleRecord.mSource == SampleRecord::kReconstructedSource) {
		if (mReconstructedPhylogenyList->getNumPhylogenies() == 0)
			throw CException("No reconstructed phylogenies to sample. Use RECONSTRUCT to reconstruct phylogenies.");
		list = mReconstructedPhylogenyList;
	}
		

	int n = list->getNumPhylogenies();
	
	mSampledPhylogenyList->removeAll();
	
	if (mProgress != NULL) {
		mProgress->reset();
		mProgress->description("Sampling phylogenies...");
		mProgress->progress(0, n * inSampleRecord.mReplicates);
		mProgress->setSkip((n * inSampleRecord.mReplicates) / 100);
	}
		
	int i, j;
	int count = 0;

	for (i = 0; i < n; i++) {
		mSimulator->setPhylogeny(list->getPhylogeny(i));
			
		for (j = 0; j < inSampleRecord.mReplicates; j++) {
			if (mProgress != NULL)
				mProgress->progress(i * inSampleRecord.mReplicates + j, 
									n * inSampleRecord.mReplicates);
			}

			if (inSampleRecord.mType == SampleRecord::kFractionType) {
				long sampleSize = (long)floor(inSampleRecord.mFraction * 
                                                        (double)mSimulator->getNumExtantLineages() + 0.5);
				phylogeny = mSimulator->getSamplePhylogeny(sampleSize);	
				
			} else if (inSampleRecord.mType == SampleRecord::kNumberType) {
			
				phylogeny = mSimulator->getSamplePhylogeny(inSampleRecord.mNumber);
					
			} else if (inSampleRecord.mType == SampleRecord::kProbabilityType) {
				phylogeny = mSimulator->getSamplePhylogeny(inSampleRecord.mProbability);
			}

			mSampledPhylogenyList->add(phylogeny);
			
	}

	if (mProgress != NULL)
		mProgress->complete();

	CParagraphElement* para = new CParagraphElement(kNormalText, kAlignElementDefault, 0);
	
	if (n == 1) {
		para->add("Sampling complete: 1 phylogeny ");
	} else {
		para->add("Sampling complete: ");
		para->add(n);
		para->add(" phylogenies sampled");
	}
	
	if (inSampleRecord.mReplicates == 1) {
		para->add("sampled.");
	} else {
		para->add("sampled ");
		para->add(inSampleRecord.mReplicates);
		para->add(" times.");
	}
	
	if (!inSampleRecord.mKeep) {
		list->removeAll();
		para->add(" Source phylogenies have been cleared from memory. ");
	}
	
	ioReport->add(para);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doStatus
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doStatus(StatusRecord inStatusRecord, CReportRecord *ioReport)
{
	CParagraphElement *para;
	
	if (mSimulator != NULL)
		mSimulator->describe(ioReport);
	else {
		para = new CParagraphElement();
		para->add("No simulation model selected");
		ioReport->add(para);
	}
	
	if (mWholePhylogenyList != NULL) {
		para = new CParagraphElement();
		if (mWholePhylogenyList->getNumPhylogenies() == 0) {
			para->add("No phylogenies in whole phylogeny list");
		} else if (mWholePhylogenyList->getNumPhylogenies() == 1) {
			para->add("1 phylogeny in whole phylogeny list");
		} else {
			para->add(mWholePhylogenyList->getNumPhylogenies());
			para->add(" phylogenies in whole phylogeny list");
		}
		ioReport->add(para);
	}
	
	if (mReconstructedPhylogenyList != NULL) {
		para = new CParagraphElement();
		if (mReconstructedPhylogenyList->getNumPhylogenies() == 0) {
			para->add("No phylogenies in reconstructed phylogeny list");
		} else if (mReconstructedPhylogenyList->getNumPhylogenies() == 1) {
			para->add("1 phylogeny in reconstructed phylogeny list");
		} else {
			para->add(mReconstructedPhylogenyList->getNumPhylogenies());
			para->add(" phylogenies in reconstructed phylogeny list");
		}
		ioReport->add(para);
	}
	
	if (mSampledPhylogenyList != NULL) {
		para = new CParagraphElement();
		if (mSampledPhylogenyList->getNumPhylogenies() == 0) {
			para->add("No phylogenies in sampled phylogeny list");
		} else if (mSampledPhylogenyList->getNumPhylogenies() == 1) {
			para->add("1 phylogeny in sampled phylogeny list");
		} else {
			para->add(mSampledPhylogenyList->getNumPhylogenies());
			para->add(" phylogenies in sampled phylogeny list");
		}
		ioReport->add(para);
	}
	
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doShape
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doShape(ShapeRecord inShapeRecord, CReportRecord *ioReport)
{
	MEPhylogenyList *list;
	if (inShapeRecord.mSource == ShapeRecord::kWholeSource) {
		if (mWholePhylogenyList->getNumPhylogenies() == 0)
			throw CException("No whole phylogenies in memory to get statistics on");
		list = mWholePhylogenyList;
	} else if (inShapeRecord.mSource == ShapeRecord::kReconstructedSource) {
		if (mReconstructedPhylogenyList->getNumPhylogenies() == 0)
			throw CException("No reconstructed phylogenies in memory to get statistics on");
		list = mReconstructedPhylogenyList;
	} else {
		if (mSampledPhylogenyList->getNumPhylogenies() == 0)
			throw CException("No sampled phylogenies in memory to get statistics on");
		list = mSampledPhylogenyList;
	}
		
	MEPhylogeny *phylogeny;
	CVariate height;
	CVariate balance;
	
	int i, n = list->getNumPhylogenies();
	for (i = 0; i < n; i++) {
		phylogeny = list->getPhylogeny(i);
		
		height.add(phylogeny->getHeight());
		balance.add(phylogeny->getBalance());
		
	}
	
	CTableElement *table = new CTableElement(3, 4, 1);
	table->setCell(0, 0, new CTextElement("Statistic"));
	table->setCell(0, 1, new CTextElement("Mean"));
	table->setCell(0, 2, new CTextElement("Variance"));
	table->setCell(0, 3, new CTextElement("S.D."));
	table->setCell(1, 0, new CTextElement("Height"));
	table->setCell(1, 1, new CRealElement(height.getMean()));
	table->setCell(1, 2, new CRealElement(height.getVariance()));
	table->setCell(1, 3, new CRealElement(height.getStandardDeviation()));
	table->setCell(2, 0, new CTextElement("Balance"));
	table->setCell(2, 1, new CRealElement(balance.getMean()));
	table->setCell(2, 2, new CRealElement(balance.getVariance()));
	table->setCell(2, 3, new CRealElement(balance.getStandardDeviation()));
	
	ioReport->add(table);
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doWritePhylogenies
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doWritePhylogenies(WritePhylogeniesRecord inWritePhylogeniesRecord, 
										CWriter *inWriter,
										CReportRecord *ioReport)
{
	long i, n;
	long nw = mWholePhylogenyList->getNumPhylogenies();
	long nr = mReconstructedPhylogenyList->getNumPhylogenies();
	long ns = mSampledPhylogenyList->getNumPhylogenies();
	
	if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kWholeSource) {
		n=nw;
		if (nw == 0)
			throw CException("No whole phylogenies in memory to write");
	}
	
	if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kReconstructedSource) {
		n=nr;
		if (nr == 0)
			throw CException("No reconstructed phylogenies in memory to write");
	}

	if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kSampleSource) {
		n=ns;
		if (ns == 0)
			throw CException("No sampled phylogenies in memory to write");
	}
		
	MECreator *creator;
	if (inWritePhylogeniesRecord.mFormat == WritePhylogeniesRecord::kNexusFormat)
		creator = new MENexusCreator(inWriter);
	else
		creator = new MEPhylipCreator(inWriter);
	
	CParagraphElement* para = new CParagraphElement(kNormalText, kAlignElementDefault, 0);
	para->add("File created: ");

	if (n > 1) {
		if (mProgress != NULL) {
			mProgress->description("Writing phylogenies...");
			mProgress->reset();
			mProgress->progress(0, (int)n);
			mProgress->setSkip(n / 100);
		}

		MEPhylogeny *phylogeny;
		char str[256];
		
		for (i = 0; i < n; i++) {
			if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kWholeSource)
				phylogeny = mWholePhylogenyList->getPhylogeny(i);
			else if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kReconstructedSource)
				phylogeny = mReconstructedPhylogenyList->getPhylogeny(i);
			else
				phylogeny = mSampledPhylogenyList->getPhylogeny(i);
				
			sprintf(str, "phylogeny%d", i+1);
			phylogeny->setName(str);
			
			creator->createPhylogeny(phylogeny);

			if (mProgress != NULL)
				mProgress->progress(i+1, (int)n);
		}
				
		para->add((int)n);
		if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kWholeSource) {
			para->add(" whole phylogenies written.");
		} else if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kReconstructedSource) {
			para->add(" reconstructed phylogenies written.");
		} else {
			para->add(" sampled phylogenies written.");
		}
	} else {
		if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kWholeSource) {
			creator->createPhylogeny(mWholePhylogenyList->getPhylogeny(0));
			para->add("1 whole phylogenies written.");
		} else if (inWritePhylogeniesRecord.mSource == WritePhylogeniesRecord::kReconstructedSource) {
			creator->createPhylogeny(mReconstructedPhylogenyList->getPhylogeny(0));
			para->add("1 reconstructed phylogenies written.");
		} else {
			creator->createPhylogeny(mSampledPhylogenyList->getPhylogeny(0));
			para->add("1 sampled phylogenies written.");
		}
	}
		
	creator->finishCreation();
	
	delete creator;

	ioReport->add(para);
}

// =========================================================================
//	PhyloGenEngine::doDrawLTTPlot
// =========================================================================

void PhyloGenEngine::doDrawLTTPlot(DrawLTTRecord &inDrawLTTRecord, CReportRecord* ioReport)
{
	if (inDrawLTTRecord.mWhole &&
			mWholePhylogenyList->getNumPhylogenies() == 0)
		throw CException("No whole phylogenies to draw plot of");

	if (inDrawLTTRecord.mReconstructed &&
			mReconstructedPhylogenyList->getNumPhylogenies() == 0)
		throw CException("No reconstructed phylogenies to draw plot of");

	if (inDrawLTTRecord.mSampled &&
			mSampledPhylogenyList->getNumPhylogenies() == 0)
		throw CException("No sampled phylogenies to draw plot of");

	MEPhylogeny *phylogeny;

	CDataSeries *data;
	CDoubleArray *xData, *yData;
	
	CGraph *graph = new CGraph();
	graph->SetYAxisLogarithmic(true);
	graph->SetTitle("Lineages Through Time");
	graph->SetXAxisLabel("Time");
	graph->SetYAxisLabel("Number of Lineages");
	
	graph->SetTitleFont(20, "Helvetica");
	graph->SetAxisLabelFont(15, "Helvetica");
	graph->SetTickLabelFont(10, "Helvetica");
	
	CLinePlot *plot;
	
	int i, n, count, plotNo;
	double time;
				

	if (inDrawLTTRecord.mSampled) {
		data = new CDataSeries();
		xData = data->GetXData();
		yData = data->GetYData();
	
		phylogeny = mSampledPhylogenyList->getPhylogeny(inDrawLTTRecord.mPhylogenyNumber - 1);
		
		n = phylogeny->getNumLineages();
		for (i = 0; i < n; i++) {
			if (phylogeny->getDescendant(i) != -1) {
				time = phylogeny->getDeathTime(i);

				xData->add(time);
				yData->add(+1.0);		
			}
		}
		time = phylogeny->getDeathTime(phylogeny->getExtantLineage(0));
		xData->add(time);
		yData->add(0.0);
		
		data->Sort(CDataSeries::kSortAscending, CDataSeries::kSortXData);
		
		n = xData->getNumValues();
		count = 0;
		for (i = 0; i < n; i++) {
			if (yData->getValue(i) > 0.0) {
				count++;
				yData->setValue(i, count);
			} else {
				yData->setValue(i, count);
			}
		}
		
		plot = new CLinePlot();
		CColourRec colour = {kRedColour, 1.0, 1.0, 1.0, 1.0};
		plot->SetLineStyle(1.0, kRoundLineCap, kRoundLineJoin, colour);
		plotNo = graph->AddPlot(plot);
		graph->SetPlotData(plotNo, data);
	}
	
	if (inDrawLTTRecord.mReconstructed) {
		data = new CDataSeries();
		xData = data->GetXData();
		yData = data->GetYData();
	
		phylogeny = mReconstructedPhylogenyList->getPhylogeny(inDrawLTTRecord.mPhylogenyNumber - 1);

		n = phylogeny->getNumLineages();
		for (i = 0; i < n; i++) {
			if (phylogeny->getDescendant(i) != -1) {
				time = phylogeny->getDeathTime(i);

				xData->add(time);
				yData->add(+1.0);		
			}
		}
		time = phylogeny->getDeathTime(phylogeny->getExtantLineage(0));
		xData->add(time);
		yData->add(0.0);
		
		data->Sort(CDataSeries::kSortAscending, CDataSeries::kSortXData);
		
		n = xData->getNumValues();
		count = 1;
		for (i = 0; i < n; i++) {
			if (yData->getValue(i) > 0.0) {
				count++;
				yData->setValue(i, count);
			} else {
				yData->setValue(i, count);
			}
		}
		
		plot = new CLinePlot();
		CColourRec colour = {kBlackColour, 1.0, 1.0, 1.0, 1.0};
		plot->SetLineStyle(1.0, kRoundLineCap, kRoundLineJoin, colour);
		plotNo = graph->AddPlot(plot);
		graph->SetPlotData(plotNo, data);
	}
	
	if (inDrawLTTRecord.mWhole) {
		data = new CDataSeries();
		xData = data->GetXData();
		yData = data->GetYData();
	
		phylogeny = mWholePhylogenyList->getPhylogeny(inDrawLTTRecord.mPhylogenyNumber - 1);

		n = phylogeny->getNumLineages();
		time = phylogeny->getBirthTime(0);
		xData->add(time);
		yData->add(+1.0);
			
		for (i = 0; i < n; i++) {
			time = phylogeny->getDeathTime(i);
			if (phylogeny->getDescendant(i) == -1) {
				if (!phylogeny->getIsExtant(i)) {
					xData->add(time);
					yData->add(-1.0);
				}
			} else {
				xData->add(time);
				yData->add(+1.0);
			}
		}
		time = phylogeny->getDeathTime(phylogeny->getExtantLineage(0));
		xData->add(time);
		yData->add(-1.0);
		
		data->Sort(CDataSeries::kSortAscending, CDataSeries::kSortXData);
		
		n = xData->getNumValues();
		count = 0;
		for (i = 0; i < n; i++) {
			if (yData->getValue(i) > 0.0) {
				count++;
				yData->setValue(i, count);
			} else {
				count--;
				yData->setValue(i, count);
			}
		}
		
		plot = new CLinePlot();
		CColourRec colour = {kBlackColour, 1.0, 1.0, 1.0, 1.0};
		plot->SetLineStyle(2.0, kRoundLineCap, kRoundLineJoin, colour);
		plotNo = graph->AddPlot(plot);
		graph->SetPlotData(plotNo, data);
	}
	
	CFigureElement *f = new CFigureElement("LTT");
	f->setSize(kPixelPageUnits, 560.0, 420.0);
	graph->DrawGraph(f, inDrawLTTRecord.mTolerance, 0.0);

	ioReport->add(f);
}		

// =========================================================================
//	PhyloGenEngine::doDrawBalancePlot
// =========================================================================

void PhyloGenEngine::doDrawBalancePlot(DrawBalancePlotRecord &inDrawBalancePlotRecord, CReportRecord* ioReport)
{
	MEPhylogeny *phylogeny;

	CVariate balance[12];
	double sumWeights[12];
		
	int index, i, j, m, n = mReconstructedPhylogenyList->getNumPhylogenies();

	for (i = 0; i < 12; i++)
		sumWeights[i] = 0.0;

	int S;
	double lnS, I, weight;
	for (i = 0; i < n; i++) {
		phylogeny = mReconstructedPhylogenyList->getPhylogeny(i);
		
//		std::cout << "NodeSize\tBalance\tWeight\tBin" << std::endl;
		m = phylogeny->getNumLineages();
		for (j = 0; j < m; j++) {
			S = phylogeny->getNumTips(j);
			if (S > 3) {
				lnS = log(S);
				I = phylogeny->calcBalanceForLineage(j);
				if (S % 2 == 0) {
					if (I == 0)
						weight = 2.0 * double(S - 1) / S;
					else
						weight = double(S - 1) / S;
				} else
					weight = 1.0;
				
				index = floor(lnS + 0.5);
				balance[index].add(I * weight);
				sumWeights[index] += weight;
				
//				std::cout << S << "\t" << I << "\t" << weight << "\t" << index << "-" << index+1 << std::endl;
			}
		}			
	}
	for (i = 0; i < 12; i++)
		balance[i].setWeight(sumWeights[i]);
	
	CDataSeries *data;
	CDoubleArray *xData, *yData, *yHigh, *yLow;
	
	CGraph *graph = new CGraph();
	graph->SetTitle("Tree Balance");
	graph->SetXAxisLabel("ln(node size)");
	graph->SetYAxisLabel("mean I");
	
	graph->SetTitleFont(20, "Helvetica");
	graph->SetAxisLabelFont(15, "Helvetica");
	graph->SetTickLabelFont(10, "Helvetica");
	
	data = new CDataSeries();
	xData = data->GetXData();
	yData = data->GetYData();
	yHigh = data->GetYHigh();
	yLow = data->GetYLow();

	double minX = 10000.0;
	double maxX = 0.0;
	for (i = 0; i < 12; i++) {
		if (balance[i].getNumValues() > 0) {
			if (i < minX)
				minX = i;
			if (i + 2 > maxX)
				maxX = i + 2;
				
			xData->add(i + 0.5);
			yData->add(balance[i].getMean());		
			yHigh->add(balance[i].getMean() + balance[i].getStandardError());		
			yLow->add(balance[i].getMean() - balance[i].getStandardError());		
		}
	}
	
	CScatterPlot *plot = new CScatterPlot();
	CColourRec colour1 = {kBlackColour, 1.0, 1.0, 1.0, 1.0};
	CColourRec colour2 = {kNoColour, 1.0, 1.0, 1.0, 1.0};
	plot->SetLineStyle(1.0, kRoundLineCap, kRoundLineJoin, colour1);
	plot->SetMarkStyle(kCircleMarks, 10.0, 0.0, colour2, colour1);
	graph->AddPlot(plot);
	graph->SetPlotData(0, data);
	
	graph->SetXRangeManual(minX, maxX);
	graph->SetYRangeManual(0.0, 1.0);
	
	CFigureElement *f = new CFigureElement("Balance");
	f->setSize(kPixelPageUnits, 560.0, 420.0);
	graph->DrawGraph(f, 0.0, 0.0);

	ioReport->add(f);
}		

// =========================================================================
//	PhyloGenEngine::doDrawPhylogeny
// =========================================================================

void PhyloGenEngine::doDrawPhylogeny(DrawPhylogenyRecord &inDrawPhylogenyRecord, CReportRecord* ioReport)
{
/*	MEPhylogeny *phylogeny;

	CGraph *graph = new CGraph();
	graph->SetTitle("Phylogeny");
	graph->SetXAxisLabel("Time");
	
	graph->SetTitleFont(20, "Helvetica");
	graph->SetAxisLabelFont(15, "Helvetica");
	graph->SetTickLabelFont(10, "Helvetica");
	
	CPhylogenyPlot *plot;
	
	phylogeny = mReconstructedPhylogenyList->getPhylogeny(inDrawPhylogenyRecord.mPhylogenyNumber - 1);

	plot = new CPhylogenyPlot();
	plot->SetLineStyle(0.25, "black", kRoundLineCap, kRoundLineJoin);
	int plotNo = graph->AddPlot(plot);	
	plot->SetPhylogeny(phylogeny);
	
	CFigureElement *f = new CFigureElement("Phylogeny");
	f->setSize(kPixelPageUnits, 560.0, 420.0);
	graph->DrawGraph(f);

	ioReport->add(f);*/
}		

// ---------------------------------------------------------------------------------
//		 PhyloGenEngine::doDrawSpatial
// ---------------------------------------------------------------------------------

void PhyloGenEngine::doDrawSpatial(DrawSpatialRecord &inDrawSpatialRecord, CReportRecord *ioReport)
{
#ifdef SPATIAL_MODEL
	MESpatialSimulator *sSim;
	
	if (mModelType == kSpatialModel) {
		sSim = dynamic_cast<MESpatialSimulator *>(mSimulator);
	
		if (sSim->getPhylogeny()->getNumExtantLineages() < 1)
			throw CException("Spatial model has not run yet");
			
		sSim->drawSpatial(ioReport);
	} else {
		throw CException("Spatial model has not been selected");
	}
#endif
}

