/******************************************************************************

Phylogeny Generator (PhyloGen) v1.0
Copyright (C) 2002 Andrew Rambaut, 
Department of Zoology, University of Oxford, UK.

http://evolve.zoo.ox.ac.uk/
mailto:andrew.rambaut@zoo.ox.ac.uk

This file is part of the PhyloGen program.

    PhyloGen is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PhyloGen is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolEvolLib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 PhyloGenEngine.h

		Interface for the PhyloGenEngine Class

 ******************************************************************************/

#ifndef _PHYLOGEN_ENGINE_H_
#define _PHYLOGEN_ENGINE_H_

#define SPATIAL_MODEL 0

#include "CExceptions.h"

#include "MEBirthDeathSimulator.h"
#include "MEEpisodicSimulator.h"

enum ModelType {
	kUnknownModel,
	kBirthDeathModel,
	kEpisodicModel,
	kSpatialModel,
	numModelTypes
};

class BirthDeathRecord {
	public:
		double mBirthRate;
		double mDeathRate;
		int mExtantStop;
		double mTimeStop;
		
		BirthDeathRecord() {
			mBirthRate = 1.0;
			mDeathRate = 0.0;
			mExtantStop = 100;
			mTimeStop = 0.0;
		}
};
		
class EpisodicRecord { // not used at present
	public:
		
		EpisodicRecord() {
		}
};
		
class AddGrowthRecord : public MEGrowthEpisode {
	public:
		
		AddGrowthRecord() {
			mStopType = MEGrowthEpisode::kExtantStopType;
			mBirthRate = 1.0;
			mDeathRate = 0.0;
			mExtantStop = 100;
			mByExtantStop = 0;
			mTimeStop = 0.0;
			mByTimeStop = 0.0;
			mMultipleTimeStop = 0.0;
		}
};
		
class AddConstantRecord : public MEConstantEpisode {
	public:
		
		AddConstantRecord() {			
			mStopType = MEConstantEpisode::kMultipleStopType;
			mStochastic = false;
			mBirthDeathRate = 1.0;
			mExtantStop = 100;
			mByExtantStop = 0;
			mTimeStop = 0.0;
			mByTimeStop = 0.0;
			mMultipleTimeStop = 1.0;
		}
};

class AddExtinctionRecord : public MEExtinctionEpisode {
	public:

		AddExtinctionRecord() {
			mType = kFractionType;

			mFraction  = 1.0;
			mNumber = 0;
			mProbability = 1.0;
		}
};
		
class SpatialRecord {
	public:
		int mAdjacencies;
		double mBirthRate;
		double mDeathRate;
		int mExtantStop;
		double mTimeStop;
		
		SpatialRecord() {
			mAdjacencies = 6;
			mBirthRate = 1.0;
			mDeathRate = 0.0;
			mExtantStop = 100;
			mTimeStop = 0.0;
		}
};

class GenerateRecord {
	public:
		int mReplicates;
		int mAttempts;
		bool mReconstruct;
		bool mKeep;
		
		GenerateRecord() {
			mReplicates = 1;
			mAttempts = 0;
			mReconstruct = true;
			mKeep = true;
		}
};
		
class ReconstructRecord {
	public:
		bool mKeep;
		
		ReconstructRecord() {
			mKeep = true;
		}
};
		
class SampleRecord {
	public:
		enum PhylogenySource {
			kWholeSource,
			kReconstructedSource
		};	

		enum SampleType {
			kFractionType,
			kNumberType,
			kProbabilityType
		};	

		int mReplicates;
		PhylogenySource	mSource;
		bool mKeep;
		SampleType mType;
		double mFraction;
		long mNumber;
		double mProbability;
		
		SampleRecord() {
			mReplicates = 1;
			
			mSource = kReconstructedSource;
			mKeep = true;
			
			mType = kFractionType;

			mFraction  = 1.0;
			mNumber = 0;
			mProbability = 1.0;
		}
};
		
class StatusRecord { // not used at present
	public:
		bool mMemory;
		
		StatusRecord() {
			mMemory = false;
		}
};
		
class ShapeRecord {
	public:
		enum PhylogenySource {
			kWholeSource,
			kReconstructedSource,
			kSampleSource
		};
		
		PhylogenySource mSource;
		
		ShapeRecord() {
			mSource = kReconstructedSource;
		}
};
		
class WritePhylogeniesRecord {
	public:
		enum PhylogenyFormat {
			kPhylipFormat,
			kNexusFormat
		};
		
		enum PhylogenySource {
			kWholeSource,
			kReconstructedSource,
			kSampleSource
		};
		
		char mFileName[1024];
		PhylogenyFormat mFormat;
		PhylogenySource mSource;
		
		WritePhylogeniesRecord() {
			mFileName[0] = '\0';
			mFormat = kPhylipFormat;
			mSource = kReconstructedSource;
		}
};

class DrawLTTRecord {
	public:
		int mPhylogenyNumber;
		bool mWhole;
		bool mReconstructed;
		bool mSampled;
		
		double mTolerance;
		
		DrawLTTRecord() {
			mPhylogenyNumber = 1;
			
			mWhole = true;
			mReconstructed = true;
			mSampled = false;
			
			mTolerance = 0.1;
		}
};

class DrawBalancePlotRecord {
	public:
		
		DrawBalancePlotRecord() {
		}
};

class DrawPhylogenyRecord {
	public:
		int mPhylogenyNumber;		
		DrawPhylogenyRecord() {
			mPhylogenyNumber = 1;
		}
};

class DrawSpatialRecord {
	public:
		
		DrawSpatialRecord() {
		}
};


		
class MEPhylogenyList;
class CReportRecord;
class CWriter;
class CProgress;

// Class Declaration

class PhyloGenEngine {
	protected:
		ModelType					mModelType;
		
		MEPhylogenySimulator		*mSimulator;
		
		MEPhylogenyList				*mWholePhylogenyList;
		MEPhylogenyList				*mReconstructedPhylogenyList;
		MEPhylogenyList				*mSampledPhylogenyList;
				
		CProgress					*mProgress;
				
	public:
									PhyloGenEngine();
		virtual						~PhyloGenEngine();

		virtual void				setProgress(CProgress *inProgress) 
										{ mProgress = inProgress; }
		
		virtual void				setBirthDeath(BirthDeathRecord &inBirthDeathRecord);
		virtual void				setEpisodic(EpisodicRecord &inEpisodicRecord);
		virtual void				setAddGrowth(AddGrowthRecord &inAddGrowthRecord);
		virtual void				setAddConstant(AddConstantRecord &inAddConstantRecord);
		virtual void				setAddExtinction(AddExtinctionRecord &inAddExtinctionRecord);
		virtual void				doClearEpisodes();
		
		virtual void				setSpatial(SpatialRecord &inSpatialRecord);

		virtual void				doGenerate(GenerateRecord inGenerateRecord, CReportRecord *ioReport);

		virtual void				doReconstruct(ReconstructRecord &inReconstructRecord, CReportRecord *ioReport);

		virtual void				doSample(SampleRecord &inSampleRecord, CReportRecord *ioReport);

		virtual void				doStatus(StatusRecord inStatusRecord, CReportRecord *ioReport);

		virtual void				doShape(ShapeRecord inShapeRecord, CReportRecord *ioReport);
		
		virtual void				doWritePhylogenies(WritePhylogeniesRecord inWritePhylogeniesRecord,
														CWriter *inWriter,
														CReportRecord *ioReport);

		virtual void				doDrawLTTPlot(DrawLTTRecord &inDrawLTTRecord, CReportRecord* ioReport);
		virtual void				doDrawBalancePlot(DrawBalancePlotRecord &inDrawBalancePlotRecord, CReportRecord* ioReport);
		virtual void				doDrawPhylogeny(DrawPhylogenyRecord &inDrawPhylogenyRecord, CReportRecord* ioReport);
		virtual void				doDrawSpatial(DrawSpatialRecord &inDrawSpatialRecord, CReportRecord *ioReport);
};

#endif /* _PHYLOGEN_ENGINE_H_ */
