/******************************************************************************
 PhyloGenEngineCaller.cpp

				PhyloGenEngineCaller Class
	
	Copyright  2002 Andrew Rambaut, University of Oxford. All rights reserved.

 ******************************************************************************/

#include <string.h>

#include "CApplication.h"
#include "CWriter.h"
#include "PhyloGenEngineCaller.h"
#include "PhyloGenEngine.h"


#include "CSVGReporter.h"

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::PhyloGenEngineCaller
// ---------------------------------------------------------------------------------

PhyloGenEngineCaller::PhyloGenEngineCaller(CApplication *inApplication, PhyloGenEngine *inEngine) :
	CEngineCaller(inApplication)
{
	mEngine = inEngine;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::executeCommand
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::executeCommand(CCommandRecord *inCommand) 
{
	CReportRecord *report = NULL;

	if (inCommand->compareLabel("BIRTHDEATHMODEL", 1)) {
		report = doBirthDeath(inCommand);
#if SPATIAL_MODEL
	} else if (inCommand->compareLabel("SPATIALMODEL", 2)) {
		report = doSpatial(inCommand);
#endif
	} else if (inCommand->compareLabel("EPISODICMODEL", 1)) {
		report = doEpisodic(inCommand);
	} else if (inCommand->compareLabel("ADDGROWTH", 4)) {
		report = doAddGrowth(inCommand);
	} else if (inCommand->compareLabel("ADDCONSTANT", 4)) {
		report = doAddConstant(inCommand);
	} else if (inCommand->compareLabel("ADDEXTINCTION", 4)) {
		report = doAddExtinction(inCommand);
	} else if (inCommand->compareLabel("CLEAREPISODES", 2)) {
		report = doClearEpisodes(inCommand);
	} else if (inCommand->compareLabel("GENERATE", 1)) {
		report = doGenerate(inCommand);
	} else if (inCommand->compareLabel("RECONSTRUCT", 1)) {
		report = doReconstruct(inCommand);
	} else if (inCommand->compareLabel("SAMPLE", 2)) {
		report = doSample(inCommand);
	} else if (inCommand->compareLabel("STATUS", 2)) {
		report = doStatus(inCommand);
	} else if (inCommand->compareLabel("SHAPE", 2)) {
		report = doShape(inCommand);
	} else if (inCommand->compareLabel("WRITEPHYLOGENIES", 1)) {
		report = doWritePhylogenies(inCommand);
	} else if (inCommand->compareLabel("DRAWLTT", 5)) {
		report = doDrawLTTPlot(inCommand);
	} else if (inCommand->compareLabel("DRAWPHYLOGENY", 5)) {
		report = doDrawPhylogeny(inCommand);
#if SPATIAL_MODEL
	} else if (inCommand->compareLabel("DRAWSPATIAL", 5)) {
		report = doDrawSpatial(inCommand);
	} else if (inCommand->compareLabel("DRAWBALANCE", 5)) {
		report = doDrawBalancePlot(inCommand);
#endif
	} else
		throw CUnknownCommand(inCommand->getLabel());

	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doBirthDeath
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doBirthDeath(CCommandRecord *inCommand)
{
	BirthDeathRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("BIRTHRATE", 1)) {
			rec.mBirthRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("DEATHRATE", 1)) {
			rec.mDeathRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("EXTANTSTOP", 1)) {
			rec.mExtantStop = sub->getValueAsInteger();
		} else if (sub->compareLabel("TIMESTOP", 1)) {
			rec.mTimeStop = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	mEngine->setBirthDeath(rec);
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doSpatial
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doSpatial(CCommandRecord *inCommand)
{
	SpatialRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("BIRTHRATE", 1)) {
			rec.mBirthRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("DEATHRATE", 1)) {
			rec.mDeathRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("ADJACENCIES", 1)) {
			rec.mAdjacencies = sub->getValueAsInteger();
		} else if (sub->compareLabel("EXTANTSTOP", 1)) {
			rec.mExtantStop = sub->getValueAsInteger();
		} else if (sub->compareLabel("TIMESTOP", 1)) {
			rec.mTimeStop = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	mEngine->setSpatial(rec);
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doEpisodic
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doEpisodic(CCommandRecord *inCommand)
{
	EpisodicRecord rec;
	mEngine->setEpisodic(rec);
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doAddGrowth
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doAddGrowth(CCommandRecord *inCommand)
{
	AddGrowthRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("BIRTHRATE", 2)) {
			rec.mBirthRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("DEATHRATE", 1)) {
			rec.mDeathRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("EXTANTSTOP", 1)) {
			rec.mStopType = AddGrowthRecord::kExtantStopType;
			rec.mExtantStop = sub->getValueAsInteger();
		} else if (sub->compareLabel("BYEXTANTSTOP", 3)) {
			rec.mStopType = AddGrowthRecord::kByExtantStopType;
			rec.mByExtantStop = sub->getValueAsInteger();
		} else if (sub->compareLabel("TIMESTOP", 1)) {
			rec.mStopType = AddGrowthRecord::kTimeStopType;
			rec.mTimeStop = sub->getValueAsDouble();
		} else if (sub->compareLabel("BYTIMESTOP", 3)) {
			rec.mStopType = AddGrowthRecord::kByTimeStopType;
			rec.mByTimeStop = sub->getValueAsDouble();
		} else if (sub->compareLabel("MULTIPLESTOP", 1)) {
			rec.mStopType = AddGrowthRecord::kMultipleStopType;
			rec.mMultipleTimeStop = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	mEngine->setAddGrowth(rec);
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doAddConstant
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doAddConstant(CCommandRecord *inCommand)
{
	AddConstantRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("STOCHASTIC", 1)) {
			rec.mStochastic = sub->getValueAsBoolean();
		} else if (sub->compareLabel("RATE", 1)) {
			rec.mBirthDeathRate = sub->getValueAsDouble();
		} else if (sub->compareLabel("EXTANTSTOP", 1)) {
			rec.mStopType = AddConstantRecord::kExtantStopType;
			rec.mExtantStop = sub->getValueAsInteger();
		} else if (sub->compareLabel("BYEXTANTSTOP", 3)) {
			rec.mStopType = AddConstantRecord::kByExtantStopType;
			rec.mByExtantStop = sub->getValueAsInteger();
		} else if (sub->compareLabel("TIMESTOP", 1)) {
			rec.mStopType = AddConstantRecord::kTimeStopType;
			rec.mTimeStop = sub->getValueAsDouble();
		} else if (sub->compareLabel("BYTIMESTOP", 3)) {
			rec.mStopType = AddConstantRecord::kByTimeStopType;
			rec.mByTimeStop = sub->getValueAsDouble();
		} else if (sub->compareLabel("MULTIPLESTOP", 1)) {
			rec.mStopType = AddConstantRecord::kMultipleStopType;
			rec.mMultipleTimeStop = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	mEngine->setAddConstant(rec);
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doAddExtinction
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doAddExtinction(CCommandRecord *inCommand)
{
	AddExtinctionRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("FRACTION", 1)) {
			rec.mType = AddExtinctionRecord::kFractionType;
			rec.mFraction = sub->getValueAsDouble();
		} else if (sub->compareLabel("NUMBER", 1)) {
			rec.mType = AddExtinctionRecord::kNumberType;
			rec.mNumber = sub->getValueAsInteger();
		} else if (sub->compareLabel("PROBABILITY", 1)) {
			rec.mType = AddExtinctionRecord::kProbabilityType;
			rec.mProbability = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 

	mEngine->setAddExtinction(rec);
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doClearEpisodes
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doClearEpisodes(CCommandRecord *inCommand)
{
	mEngine->doClearEpisodes();
	
	return NULL;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doGenerate
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doGenerate(CCommandRecord *inCommand)
{
	GenerateRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("REPLICATES", 3)) {
			rec.mReplicates = sub->getValueAsInteger();
		} else if (sub->compareLabel("ATTEMPTS", 1)) {
			rec.mAttempts = sub->getValueAsInteger();
		} else if (sub->compareLabel("RECONSTRUCT", 3)) {
			rec.mReconstruct = sub->getValueAsBoolean();
		} else if (sub->compareLabel("KEEP", 1)) {
			rec.mKeep = sub->getValueAsBoolean();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doGenerate(rec, report);
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doReconstruct
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doReconstruct(CCommandRecord *inCommand)
{
	ReconstructRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("KEEP", 1)) {
			rec.mKeep = sub->getValueAsBoolean();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doReconstruct(rec, report);
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doSample
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doSample(CCommandRecord *inCommand)
{
	SampleRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("REPLICATES", 1)) {
			rec.mReplicates = sub->getValueAsInteger();
		} else if (sub->compareLabel("SOURCE", 1)) {		
			if (sub->compareValue("WHOLE", 1, false))
				rec.mSource = SampleRecord::kWholeSource;
			else if (sub->compareValue("RECONSTRUCTED", 1, false))
				rec.mSource = SampleRecord::kReconstructedSource;
			else
				throw CUnknownValue(sub->getValue());
		} else if (sub->compareLabel("KEEP", 1)) {
			rec.mKeep = sub->getValueAsBoolean();
		} else if (sub->compareLabel("FRACTION", 1)) {
			rec.mType = SampleRecord::kFractionType;
			rec.mFraction = sub->getValueAsDouble();
		} else if (sub->compareLabel("NUMBER", 1)) {
			rec.mType = SampleRecord::kNumberType;
			rec.mNumber = sub->getValueAsInteger();
		} else if (sub->compareLabel("PROBABILITY", 1)) {
			rec.mType = SampleRecord::kProbabilityType;
			rec.mProbability = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doSample(rec, report);
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doStatus
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doStatus(CCommandRecord *inCommand)
{
	StatusRecord rec;
	SubCommandRecord *sub;

	CReportRecord *report = new CReportRecord();

	mEngine->doStatus(rec, report);
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doShape
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doShape(CCommandRecord *inCommand)
{
	ShapeRecord rec;
	SubCommandRecord *sub;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("SOURCE", 1)) {
		
			if (sub->compareValue("WHOLE", 1, false))
				rec.mSource = ShapeRecord::kWholeSource;
			else if (sub->compareValue("RECONSTRUCTED", 1, false))
				rec.mSource = ShapeRecord::kReconstructedSource;
			else if (sub->compareValue("SAMPLED", 1, false))
				rec.mSource = ShapeRecord::kSampleSource;
			else
				throw CUnknownValue(sub->getValue());
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doShape(rec, report);
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doWritePhylogenies
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doWritePhylogenies(CCommandRecord *inCommand)
{
	WritePhylogeniesRecord rec;
	SubCommandRecord *sub;
	bool hasFileName = false;

	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("FILE", 2)) {
			strncpy(rec.mFileName, sub->getValue(), 1023);
			hasFileName = true;
		} else if (sub->compareLabel("FORMAT", 2)) {
		
			if (sub->compareValue("PHYLIP", 1, false))
				rec.mFormat = WritePhylogeniesRecord::kPhylipFormat;
			else if (sub->compareValue("NEXUS", 1, false))
				rec.mFormat = WritePhylogeniesRecord::kNexusFormat;
				
		} else if (sub->compareLabel("SOURCE", 1)) {
		
			if (sub->compareValue("WHOLE", 1, false))
				rec.mSource = WritePhylogeniesRecord::kWholeSource;
			else if (sub->compareValue("RECONSTRUCTED", 1, false))
				rec.mSource = WritePhylogeniesRecord::kReconstructedSource;
			else if (sub->compareValue("SAMPLED", 1, false))
				rec.mSource = WritePhylogeniesRecord::kSampleSource;
			else
				throw CUnknownValue(sub->getValue());
				
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	if (!hasFileName) {
		if (rec.mSource == WritePhylogeniesRecord::kWholeSource)
			strcpy(rec.mFileName, "whole");
		else if (rec.mSource == WritePhylogeniesRecord::kReconstructedSource)
			strcpy(rec.mFileName, "reconstructed");
		else
			strcpy(rec.mFileName, "sampled");
			
		if (rec.mFormat == WritePhylogeniesRecord::kPhylipFormat)
			strcat(rec.mFileName, ".tf");
		else
			strcat(rec.mFileName, ".nex");
	}
	
	CReportRecord *report = new CReportRecord();

	CFileRequest ioFileRequest;
	strcpy(ioFileRequest.mFileName, rec.mFileName);
	ioFileRequest.mAskAppend = true;
	ioFileRequest.mAskFormat = true;
	ioFileRequest.mFormat = (rec.mFormat == WritePhylogeniesRecord::kNexusFormat ? 0 : 1);
	ioFileRequest.mFormatNames[0] = "Nexus";
	ioFileRequest.mFormatNames[1] = "PHYLIP";
	ioFileRequest.mNumFormats = 2;

	CWriter *writer = mApplication->createFileWriter(&ioFileRequest);

	rec.mFormat = (ioFileRequest.mFormat == 0 ? 
						WritePhylogeniesRecord::kNexusFormat :
						WritePhylogeniesRecord::kPhylipFormat);
	
	mEngine->doWritePhylogenies(rec, writer, report);
	
	delete writer;

	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doDrawLTTPlot
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doDrawLTTPlot(CCommandRecord *inCommand)
{
	DrawLTTRecord rec;
	SubCommandRecord *sub;

	CFileRequest ioFileRequest;
	strcpy(ioFileRequest.mFileName, "LTT.svg");
	bool hasFile = false;
	
	if (inCommand->getHasValue() && inCommand->compareValue("FILE", 1, false))
		hasFile = true;
		
	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("FILE", 1)) {
			strcpy(ioFileRequest.mFileName, sub->getValue());
			hasFile = true;
		} else if (sub->compareLabel("PHYLOGENY", 1)) {
			rec.mPhylogenyNumber = sub->getValueAsInteger();
		} else if (sub->compareLabel("WHOLE", 1)) {
			rec.mWhole = sub->getValueAsBoolean();
		} else if (sub->compareLabel("RECONSTRUCTED", 1)) {
			rec.mReconstructed = sub->getValueAsBoolean();
		} else if (sub->compareLabel("SAMPLE", 1)) {
			rec.mSampled = sub->getValueAsBoolean();
		} else if (sub->compareLabel("TOLERANCE", 1)) {
			rec.mTolerance = sub->getValueAsDouble();
		} else
			throw CUnknownSubCommand(sub->getLabel());

	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doDrawLTTPlot(rec, report);
	
	if (hasFile) {
		ioFileRequest.mAskAppend = false;
		ioFileRequest.mAskFormat = true;

		CWriter *writer = mApplication->createFileWriter(&ioFileRequest);

	 	CSVGReporter* reporter = new CSVGReporter(writer, true);
		reporter->writeReport(report);
		delete reporter;
	}
		
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doDrawBalancePlot
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doDrawBalancePlot(CCommandRecord *inCommand)
{
	DrawBalancePlotRecord rec;
	SubCommandRecord *sub;

	CFileRequest ioFileRequest;
	strcpy(ioFileRequest.mFileName, "balance.svg");
	bool hasFile = false;
	
	if (inCommand->getHasValue() && inCommand->compareValue("FILE", 1, false))
		hasFile = true;
		
	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("FILE", 1)) {
			strcpy(ioFileRequest.mFileName, sub->getValue());
			hasFile = true;
		} else
			throw CUnknownSubCommand(sub->getLabel());

	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doDrawBalancePlot(rec, report);
	
	if (hasFile) {
		ioFileRequest.mAskAppend = false;
		ioFileRequest.mAskFormat = true;

		CWriter *writer = mApplication->createFileWriter(&ioFileRequest);

	 	CSVGReporter* reporter = new CSVGReporter(writer, true);
		reporter->writeReport(report);
		delete reporter;
	}
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doDrawPhylogeny
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doDrawPhylogeny(CCommandRecord *inCommand)
{
	DrawPhylogenyRecord rec;
	SubCommandRecord *sub;


	CFileRequest ioFileRequest;
	strcpy(ioFileRequest.mFileName, "phylogeny.svg");
	bool hasFile = false;
	
	if (inCommand->getHasValue() && inCommand->compareValue("FILE", 1, false))
		hasFile = true;
		
	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("FILE", 1)) {
			strcpy(ioFileRequest.mFileName, sub->getValue());
			hasFile = true;
		} else if (sub->compareLabel("PHYLOGENY", 1)) {
			rec.mPhylogenyNumber = sub->getValueAsInteger();
		} else
			throw CUnknownSubCommand(sub->getLabel());

	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doDrawPhylogeny(rec, report);
	
	if (hasFile) {
		ioFileRequest.mAskAppend = false;
		ioFileRequest.mAskFormat = true;

		CWriter *writer = mApplication->createFileWriter(&ioFileRequest);

	 	CSVGReporter* reporter = new CSVGReporter(writer, true);
		reporter->writeReport(report);
		delete reporter;
	}
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doDrawSpatial
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doDrawSpatial(CCommandRecord *inCommand)
{
	DrawSpatialRecord rec;
	SubCommandRecord *sub;

	CFileRequest ioFileRequest;
	strcpy(ioFileRequest.mFileName, "spatial.svg");
	bool hasFile = false;
	
	if (inCommand->getHasValue() && inCommand->compareValue("FILE", 1, false))
		hasFile = true;
		
	for (int i = 0; i < inCommand->getNumSubCommands(); i++) {
	
		sub = inCommand->getSubCommand(i);

		if (sub->compareLabel("FILE", 1)) {
			strcpy(ioFileRequest.mFileName, sub->getValue());
			hasFile = true;
		} else
			throw CUnknownSubCommand(sub->getLabel());
	} 
	
	CReportRecord *report = new CReportRecord();

	mEngine->doDrawSpatial(rec, report);
	
	if (hasFile) {
		ioFileRequest.mAskAppend = false;
		ioFileRequest.mAskFormat = true;

		CWriter *writer = mApplication->createFileWriter(&ioFileRequest);

	 	CSVGReporter* reporter = new CSVGReporter(writer, true);
		reporter->writeReport(report);
		delete reporter;
	}
	
	return report;
}

// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doHelp
// ---------------------------------------------------------------------------------

CReportRecord* PhyloGenEngineCaller::doHelp(CCommandRecord *inCommand)
{
	CReportRecord *report = NULL;
	
	if (!inCommand->getHasValue() || inCommand->compareValue("ALL", 3, false)) {
		report = new CReportRecord("PhyloGen Help");
		
		CParagraphElement *para = new CParagraphElement();
		para->add("PhyloGen v1.1, 2002 Andrew Rambaut ");
		para->addLineBreak();
		CTextElement *text = new CTextElement("Department of Zoology, University of Oxford, Oxford, UK");
		text->setLink("http://www.zoo.ox.ac.uk/");
		para->add(text);
		para->addLineBreak();
		text = new CTextElement("Website: <http://evolve.zoo.ox.ac.uk>");
		text->setLink("http://evolve.zoo.ox.ac.uk/");
		para->add(text);
		para->addLineBreak();
		text = new CTextElement("Email: andrew.rambaut@zoo.ox.ac.uk");
		text->setLink("mailto:andrew.rambaut@zoo.ox.ac.uk");
		para->add(text);
		para->addLineBreak();
		para->add("Please report any bugs or suggestions to the author.");
		para->addLineBreak();
		report->add(para);

		para = new CParagraphElement(kHeadingLevel2);
		para->add(new CTextElement("Model specification commands:"));
		report->add(para);
		CListElement *list = new CListElement(kDefinitionList, 1);
		list->add(new CTextElement("BIRTHDEATHMODEL", "#BIRTHDEATHMODEL"));
		list->add(new CTextElement("Specify birth-death model and parameters"));

#if SPATIAL_MODEL
		list->add(new CTextElement("SPATIALMODEL", "#SPATIALMODEL"));
		list->add(new CTextElement("Specify spatial model and parameters"));
#endif
		list->add(new CTextElement("EPISODICMODEL", "#EPISODICMODEL"));
		list->add(new CTextElement("Specify episodic model and parameters"));

		list->add(new CTextElement("ADDGROWTH", "#ADDGROWTH"));
		list->add(new CTextElement("Add an episode of growth to the episodic model"));

		list->add(new CTextElement("ADDCONSTANT", "#ADDCONSTANT"));
		list->add(new CTextElement("Add an episode of constant size to the episodic model"));

		list->add(new CTextElement("ADDEXTINCTION", "#ADDEXTINCTION"));
		list->add(new CTextElement("Add a mass extinction to the episodic model"));

		list->add(new CTextElement("CLEAREPISODES", "#CLEAREPISODES"));
		list->add(new CTextElement("Clear currently set episodes from episodic model"));

		report->add(list);

		para = new CParagraphElement(kHeadingLevel2);
		para->add(new CTextElement("Operation commands:"));
		report->add(para);
		list = new CListElement(kDefinitionList, 1);
		list->add(new CTextElement("GENERATE", "#GENERATE"));
		list->add(new CTextElement("Generate phylogenies under the current model"));
		
		list->add(new CTextElement("RECONSTRUCT", "#RECONSTRUCT"));
		list->add(new CTextElement("Reconstruct trees of extant taxa only"));
		
		list->add(new CTextElement("SAMPLE", "#SAMPLE"));
		list->add(new CTextElement("Sample generated phylogenies"));
		
		list->add(new CTextElement("STATUS", "#STATUS"));
		list->add(new CTextElement("Get current simulator status"));
		
		list->add(new CTextElement("SHAPE", "SHAPE"));
		list->add(new CTextElement("Get various statistics about the shape of the phylogenies"));

		list->add(new CTextElement("WRITEPHYLOGENIES", "#WRITEPHYLOGENIES"));
		list->add(new CTextElement("Write phylogenies to file"));
		
		list->add(new CTextElement("DRAWLTT", "#DRAWLTT"));
		list->add(new CTextElement("Draw lineages-through-time plot"));
		
#if SPATIAL_MODEL
		list->add(new CTextElement("DRAWSPATIAL", "#DRAWSPATIAL"));
		list->add(new CTextElement("Draw spatial model grid"));
		list->add(new CTextElement("DRAWBALANCE", "#DRAWBALANCE"));
		list->add(new CTextElement("Draw graph of balance against node size"));
#endif
		report->add(list);

		para = new CParagraphElement(kHeadingLevel2);
		para->add(new CTextElement("General commands:"));
		report->add(para);
		list = new CListElement(kDefinitionList, 1);
		list->add(new CTextElement("LOG"));
		list->add(new CTextElement("Log output to file"));
		list->add(new CTextElement("HELP"));
		list->add(new CTextElement("Use HELP <command> to get help on a command"));
		list->add(new CTextElement("QUIT"));
		list->add(new CTextElement("Quit the application"));
		report->add(list);
		
		if (inCommand->compareValue("ALL", 3, false)) {
			CParagraphElement *heading = new CParagraphElement(kHeadingLevel1);
			heading->add("Details of commands: ");
			
			// fake up commands to write out the rest of the help
			doCommandHelp(&CCommandRecord("HELP", "BIRTHDEATHMODEL"), report);
#if SPATIAL_MODEL
			doCommandHelp(&CCommandRecord("HELP", "SPATIALMODEL"), report);
#endif
			doCommandHelp(&CCommandRecord("HELP", "EPISODICMODEL"), report);
			doCommandHelp(&CCommandRecord("HELP", "ADDGROWTH"), report);
			doCommandHelp(&CCommandRecord("HELP", "ADDCONSTANT"), report);
			doCommandHelp(&CCommandRecord("HELP", "ADDEXTINCTION"), report);
			doCommandHelp(&CCommandRecord("HELP", "CLEAREPISODES"), report);

			doCommandHelp(&CCommandRecord("HELP", "GENERATE"), report);
			doCommandHelp(&CCommandRecord("HELP", "RECONSTRUCT"), report);
			doCommandHelp(&CCommandRecord("HELP", "SAMPLE"), report);
			doCommandHelp(&CCommandRecord("HELP", "STATUS"), report);
			doCommandHelp(&CCommandRecord("HELP", "SHAPE"), report);
			doCommandHelp(&CCommandRecord("HELP", "WRITEPHYLOGENIES"), report);
			doCommandHelp(&CCommandRecord("HELP", "DRAWLTT"), report);
#if SPATIAL_MODEL
			doCommandHelp(&CCommandRecord("HELP", "DRAWSPATIAL"), report);
			doCommandHelp(&CCommandRecord("HELP", "DRAWBALANCE"), report);
#endif
		}
	} else {
		report = new CReportRecord("PhyloGen Help");
		CParagraphElement *heading = new CParagraphElement(kHeadingLevel2);
		heading->add("Help on command: ");
		
		doCommandHelp(inCommand, report);
	}
	
	return report;
}


// ---------------------------------------------------------------------------------
//		 PhyloGenEngineCaller::doCommandHelp
// ---------------------------------------------------------------------------------

void PhyloGenEngineCaller::doCommandHelp(CCommandRecord *inCommand, CReportRecord* ioReport)
{
	CParagraphElement *heading = new CParagraphElement(kHeadingLevel2);
	CParagraphElement *para = new CParagraphElement(kNormalText, kAlignElementDefault, 1);
	CListElement *list = new CListElement(kDefinitionList, 2);

	if (inCommand->compareValue("BIRTHDEATHMODEL", 1, false)) {
		heading->add("BIRTHDEATHMODEL");
		heading->setAnchor("BIRTHDEATHMODEL");
		para->add("Specify birth-death phylogeny generation model. ");
		list->add("BIRTHRATE = <real>");
		list->add("Per lineage birth rate (default: 1.0)");
		list->add("DEATHRATE = <real>");
		list->add("Per lineage death rate (default: 0.0)");
		list->add("EXTANTSTOP = <integer>");
		list->add("Stop when number of extant reaches value (default: 100)");
		list->add("TIMESTOP = <real>");
		list->add("Stop when time reaches value (0 for no stop time) (default: 0.0)");
	} else 	if (inCommand->compareValue("SPATIALMODEL", 1, false)) {
		heading->add("SPATIALMODEL");
		heading->setAnchor("SPATIALMODEL");
		para->add("Specify spatial phylogeny generation model. ");
		list->add("ADJACENCIES = <integer>");
		list->add("Number of adjacencies - 2, 3, 4, 6 or 8 (default: 6)");
		list->add("BIRTHRATE = <real>");
		list->add("Per lineage birth rate (default: 1.0)");
		list->add("DEATHRATE = <real>");
		list->add("Per lineage death rate (default: 0.0)");
		list->add("EXTANTSTOP = <integer>");
		list->add("Stop when number of extant reaches value (default: 100)");
		list->add("TIMESTOP = <real>");
		list->add("Stop when time reaches value (0 for no stop time) (default: 0.0)");
	} else 	if (inCommand->compareValue("EPISODICMODEL", 1, false)) {
		heading->add("EPISODICMODEL");
		heading->setAnchor("EPISODICMODEL");
		para->add("Specify episodic phylogeny generation model. Episodes are added using "
					"the ADDGROWTH, ADDCONSTANT, ADDEXTINCTION commands. A growth "
					"episode should probably be the first episode.");
	} else if (inCommand->compareValue("ADDGROWTH", 4, false)) {
		heading->add("ADDGROWTH");
		heading->setAnchor("ADDGROWTH");
		para->add("Add an episode of growth to the simulation. ");
		list->add("BIRTHRATE = <real>");
		list->add("Per lineage birth rate (default: 1.0)");
		list->add("DEATHRATE = <real>");
		list->add("Per lineage death rate (default: 0.0)");
		list->add("EXTANTSTOP = <integer>");
		list->add("Stop when number of extant reaches value (default: 100)");
		list->add("BYEXTANTSTOP = <integer>");
		list->add("Stop when number of extant changes by value. Used as alternative "
					"to EXTANTSTOP. Will stop when number of extant lineages drops "
					"by specied value. (default: off)");
		list->add("TIMESTOP = <real>");
		list->add("Stop when time reaches value (0 for no stop time) (default: off)");
		list->add("BYTIMESTOP = <real>");
		list->add("Stop when time changes by value. Used as alternative to TIMESTOP. (default: off)");
		list->add("MULTIPLELASTTIME = <real>");
		list->add("Stop when time changes by the given multiple of previous episode's "
					"running time. Used as alternative to TIMESTOP. (default: off)");
	} else if (inCommand->compareValue("ADDCONSTANT", 4, false)) {
		heading->add("ADDCONSTANT");
		heading->setAnchor("ADDCONSTANT");
		para->add("Add an episode of constant size to the simulation. ");
		list->add("STOCHASTIC = TRUE | FALSE");
		list->add("Stochastic fluctuation? (default: FALSE)");
		list->add("RATE = <real>");
		list->add("Per lineage birth/death rate (default: 1.0)");
		list->add("EXTANTSTOP = <integer>");
		list->add("Stop when number of extant reaches value. This can only be used"
					"with the stochcastic mode. (default: off)");
		list->add("BYEXTANTSTOP = <integer>");
		list->add("Stop when number of extant changes by value. This can only be used"
					"with the stochcastic mode. (default: off)");
		list->add("TIMESTOP = <real>");
		list->add("Stop when time reaches value (0 for no stop time) (default: off)");
		list->add("BYTIMESTOP = <real>");
		list->add("Stop when time changes by value. Used as alternative to "
					"TIMESTOP. (default: off)");
		list->add("MULTIPLELASTTIME = <real>");
		list->add("Stop when time changes by the given multiple of previous episode's "
					"running time. Used as alternative to TIMESTOP. (default: 1.0)");
	} else if (inCommand->compareValue("ADDEXTINCTION", 4, false)) {
		heading->add("ADDEXTINCTION");
		heading->setAnchor("ADDEXTINCTION");
		para->add("Add a mass extinction to the simulation. ");
		list->add("FRACTION = <real>");
		list->add("Fraction of extant lineages to go extinct (default: 0.5)");
		list->add("NUMBER = <integer>");
		list->add("Number of extant lineages to go extinct (default: off)");
		list->add("PROBABILITY = <real>");
		list->add("Per lineage probability of extant lineages going extinct (default: off)");
	} else if (inCommand->compareValue("CLEAREPISODES", 2, false)) {
		heading->add("CLEAREPISODES");
		heading->setAnchor("CLEAREPISODES");
		para->add("Clear currently set episodes from episodic model. ");
		delete list;
		list = NULL;
	} else if (inCommand->compareValue("GENERATE", 3, false)) {
		heading->add("GENERATE");
		heading->setAnchor("GENERATE");
		para->add("Generate trees under the current model. ");
		list->add("REPLICATES = <integer>");
		list->add("Number of phylogenies to generate (default: 1)");
		list->add("ATTEMPTS = <integer>");
		list->add("Number of attempts to make in cases of all lineages going extinct."
					" (default: 1000 times number of trees to generate)");
		list->add("RECONSTRUCT = TRUE | FALSE");
		list->add("Create reconstructed phylogenies (extant lineages only)? (default: true)");
		list->add("KEEP = TRUE | FALSE");
		list->add("Keep whole phylogenies after reconstructing? (default: true)");
	} else if (inCommand->compareValue("RECONSTRUCT", 2, false)) {
		heading->add("RECONSTRUCT");
		heading->setAnchor("RECONSTRUCT");
		para->add("Reconstruct trees of extant taxa only. ");
		list->add("KEEP = TRUE | FALSE");
		list->add("Keep whole phylogenies list after reconstructing? (default: TRUE)");
	} else if (inCommand->compareValue("SAMPLE", 2, false)) {
		heading->add("SAMPLE");
		heading->setAnchor("SAMPLE");
		para->add("Specify sampling conditions. ");
		list->add("REPLICATES = <integer>");
		list->add("Number of samples to make of each phylogeny in list (default: 1)");
		list->add("SOURCE = WHOLE | RECONSTRUCTED");
		list->add("Sample from the whole phylogenies or reconstructed (default: RECONSTRUCTED)");
		list->add("KEEPSOURCE = TRUE | FALSE");
		list->add("Keep source phylogenies list after sampling (default: TRUE)");
		list->add("FRACTION = <real>");
		list->add("Fraction of tips to sample (default: 0.5)");
		list->add("NUMBER = <integer>");
		list->add("Number of tips to sample (default: off)");
		list->add("PROBABILITY = <real>");
		list->add("Per tip probability of sampling (default: off)");
	} else if (inCommand->compareValue("STATUS", 2, false)) {
		heading->add("STATUS");
		heading->setAnchor("STATUS");
		para->add("Get current simulator status. ");
		list->add("MEMORY = <boolean>");
		list->add("Report information about memory use (default: off)");
	} else if (inCommand->compareValue("SHAPE", 2, false)) {
		heading->add("SHAPE");
		heading->setAnchor("SHAPE");
		para->add("Get various statistics about the shape of the phylogenies. ");
		list->add("SOURCE = <WHOLE | RECONSTRUCTED | SAMPLED>");
		list->add("Analyse the whole, reconstructed, or sampled trees."
					" (default: RECONSTRUCTED)");
	} else if (inCommand->compareValue("WRITEPHYLOGENIES", 3, false)) {
		heading->add("WRITEPHYLOGENIES");
		heading->setAnchor("WRITEPHYLOGENIES");
		para->add("Write phylogenies to a file. ");
		list->add("FILE = <file name>");
		list->add("Name of file (default: 'whole.tf' | 'reconstructed.tf' | 'sampled.tf)");
		list->add("FORMAT = <PHYLIP | NEXUS>");
		list->add("Format of file to write. (default: PHYLIP)");
		list->add("SOURCE = <WHOLE | RECONSTRUCTED | SAMPLED>");
		list->add("Write the whole, reconstructed, or sampled trees."
					" (default: RECONSTRUCTED)");
//		list->add("RANGE = <range>");
//		list->add("Write a range of trees from the list. Specifiy as '1-10' or "
//					"'ALL' to write all trees. (default: ALL)");
	} else if (inCommand->compareValue("DRAWLTT", 5, false)) {
		heading->add("DRAWLTT");
		heading->setAnchor("DRAWLTT");
		para->add("Draw lineages-through-time plot. For some versions, write to file must be used.");
		list->add("FILE | FILE = <file name>");
		list->add("Write to file ('ltt.svg') or specify name of file to write ");
		list->add("PHYLOGENY = <integer>");
		list->add("Which phylogeny to draw (default: 1)");
		list->add("WHOLE = <boolean>");
		list->add("Draw LTT for whole phylogeny (default: true)");
		list->add("RECONSTRUCTED = <boolean>");
		list->add("Draw LTT for reconstructed phylogeny (default: true)");
		list->add("SAMPLED = <boolean>");
		list->add("Draw LTT for sampled phylogeny, if available (default: true)");
		list->add("TOLERANCE = <real>");
		list->add("Minimum time between points to draw (for very large phylogenies) (default: 0.0)");
	} else if (inCommand->compareValue("DRAWSPATIAL", 5, false)) {
		heading->add("DRAWSPATIAL");
		heading->setAnchor("DRAWSPATIAL");
		para->add("Draw spatial model grid.  For some versions, write to file must be used.");
		list->add("FILE | FILE = <file name>");
		list->add("Write to file ('spatial.svg') or specify name of file to write ");
	} else if (inCommand->compareValue("DRAWBALANCE", 5, false)) {
		heading->add("DRAWBALANCE");
		heading->setAnchor("DRAWBALANCE");
		para->add("Draw graph of balance against node size.  For some versions, write to file must be used.");
		list->add("FILE | FILE = <file name>");
		list->add("Write to file ('balance.svg') or specify name of file to write ");
/*	} else if (inCommand->compareValue("RUN", 3, false)) {
		heading->add("RUN");
		para->setAnchor("RUN");
		para->add("Load a data file and run any commands.");
		para->addLineBreak();
		para->add("Usage: RUN <filename>");
		delete list;
		list = NULL;
	} else if (inCommand->compareValue("LOAD", 4, false)) {
		heading->add("LOAD");
		para->setAnchor("LOAD");
		para->add("Load a data file.");
		para->addLineBreak();
		para->add("Usage: LOAD <filename>");
		delete list;
		list = NULL;*/
	} else if (inCommand->compareValue("LOG", 3, false)) {
		heading->add("LOG");
		para->setAnchor("LOG");
		para->add("Start logging output to a log file.");
		para->addLineBreak();
		para->add("Usage: LOG <filename> - turn log file on");
		para->addLineBreak();
		para->add("Usage: LOG - turn log file off");
		list->add("FORMAT = TEXT | HTML");
		list->add("Format to write log file (default: TEXT)");
	} else if (inCommand->compareValue("HELP", 4, false)) {
		heading->add("HELP");
		para->setAnchor("HELP");
		para->add("Get help on commands.");
		para->addLineBreak();
		para->add("Usage: HELP - list available commands");
		para->addLineBreak();
		para->add("Usage: HELP <command> - get help on specific command");
		para->addLineBreak();
		para->add("Usage: HELP ALL - get entire help document");
	} else if (inCommand->compareValue("QUIT", 1, false)) {
		heading->add("QUIT");
		para->setAnchor("CODES");
		para->add("Quit the program.");
		para->addLineBreak();
		para->add("Usage: QUIT");
		delete list;
		list = NULL;
	} else {
		para->add("Command not known - use HELP to get list of commands");
		delete list;
		list = NULL;
	}
		
	ioReport->add(heading);
	ioReport->add(para);
	
	if (list != NULL) {
		para = new CParagraphElement(kHeadingLevel3);
		para->add("Subcommands:");
		ioReport->add(para);
		ioReport->add(list);
	}
}

